#define TRACE	0

#include "evo.h"

void BigEvolver::AnnotFromGFFRec(const GFFRecord &Rec, gpos_t ChrStartPos,
  gpos_t ChrLength)
	{
	asserta(sizeof(coord_t) == sizeof(gpos_t));

	set<string> InvalidFeatures;
	string AnnotName(Rec.Feature);
	string SeqName(Rec.SeqName);

	if (AnnotName == "microsat")
		return;

	ANNOT_TYPE AnnotType = StrToAnnotType(Rec.Feature);

	if (AnnotType == ANNOT_TYPE_Invalid)
		{
		if (InvalidFeatures.find(Rec.Feature) == InvalidFeatures.end())
			{
			Warning("Invalid feature '%s'", Rec.Feature);
			InvalidFeatures.insert(Rec.Feature);
			}
		return;
		}

	bool GeneIndexRequired = (AnnotType == ANNOT_TYPE_CDS);
	uint16 GeneIndex = GetGeneIndex(Rec.Attrs, GeneIndexRequired);
	uint8 CEC = GetCEC(Rec.Attrs);
	if (CEC == UCHAR_MAX)
		{
		switch (AnnotType)
			{
		case ANNOT_TYPE_CDS:
			CEC = CEC_GenericCDS;
			break;
		case ANNOT_TYPE_NCCE:
			CEC = CEC_GenericNCCE;
			break;
		default:
			break;
			}
		}

	if (AnnotType == ANNOT_TYPE_CDS && GeneIndex == UINT_MAX)
		{
		Log("\n");
		Rec.LogMe();
		Log("\n");
		Die("Exon missing GeneIndex attribute");
		}

	asserta(Rec.End >= Rec.Start);
	coord_t Length = Rec.End - Rec.Start + 1;

	AnnotData Annot;
	Annot.Type = AnnotType;
	Annot.CEC = CEC;
	Annot.Pos = ChrStartPos + Rec.Start - 1;
	Annot.Length = Length;
	Annot.Strand = (Rec.Strand != '-');
	Annot.Frame = SCHAR_MAX;
	Annot.GeneIndex = GeneIndex;
	Annot.GeneArrayIndex = GetGeneArrayIndex(Rec.Attrs);
	Annot.MicrosatMotifLength = UINT8_MAX;

	if (Annot.GetEndPos() > ChrStartPos + ChrLength - 1)
		{
		Log("\n");
		Log("GFFRecord=");
		Rec.LogMe();
		Log("\n");
		Die("Annot overflows chromosome length");
		}

	AddAnnot(Annot);
	}

void MakeChrPosMap(const GenInfo &Genome, map<string, gpos_t> &ChrNameToStartPos,
  map<gpos_t, string> &ChrStartPosToName)
	{
	ChrNameToStartPos.clear();
	gpos_t Pos = 0;
	const unsigned ChrCount = Genome.GetChrCount();
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		{
		const ChrInfo &CI = Genome.GetChrInfoByIndex(ChrIndex);
		const string &Name = CI.GetName();
		if (ChrNameToStartPos.find(Name) != ChrNameToStartPos.end())
			Die("Chr names not unique, %s twice", Name.c_str());

		ChrNameToStartPos[Name] = Pos;
		Pos += CI.GetLength();
		}
	ChrStartPosToName.clear();
	for (map<string, gpos_t>::const_iterator p = ChrNameToStartPos.begin();
	  p != ChrNameToStartPos.end(); ++p)
		{
		const string &Name = p->first;
		gpos_t Pos = p->second;
		ChrStartPosToName[Pos] = Name;
		}
	}

void LogChrPosMap(const map<string, gpos_t> &ChrNameToStartPos,
  const map<gpos_t, string> &ChrStartPosToName)
	{
	Log("\n");
	Log("ChrPosMap:\n");
	Log("         ChrName         Pos\n");
	Log("----------------  ----------\n");
	for (map<string, gpos_t>::const_iterator p = ChrNameToStartPos.begin();
	  p != ChrNameToStartPos.end(); ++p)
		Log("%16.16s  %10"PF"\n", p->first.c_str(), p->second);
	Log("\n");
	Log("ChrPosMap2:\n");
	Log("         ChrName         Pos\n");
	Log("----------------  ----------\n");
	for (map<gpos_t, string>::const_iterator p = ChrStartPosToName.begin();
	  p != ChrStartPosToName.end(); ++p)
		Log("%16.16s  %10"PF"\n", p->second.c_str() , p->first);
	}

void PosToChrPos(gpos_t Pos, const map<gpos_t, string> &ChrStartPosToName,
  string &ChrName, gpos_t &ChrPos)
	{
	if (ChrStartPosToName.empty())
		Die("PosToChrPos, map empty");
	map<gpos_t, string>::const_iterator p = ChrStartPosToName.upper_bound(Pos);
	if (p == ChrStartPosToName.begin())
		Die("PosToChrPos, lower_bound=begin()");
	--p;
	ChrPos = Pos - p->first;
	ChrName = p->second;
	}

gpos_t ChrPosToPos(const string &ChrName, gpos_t ChrPos,
  const map<string, gpos_t> &ChrNameToStartPos)
	{
	map<string, gpos_t>::const_iterator p = ChrNameToStartPos.find(ChrName);
	if (p == ChrNameToStartPos.end())
		Die("ChrPosToPos(%s): not found", ChrName.c_str());
	return p->second + ChrPos;
	}

void LogAnnotChr(const AnnotData &Annot,
  const map<gpos_t, string> &ChrStartPosToName)
	{
	string ChrName;
	gpos_t ChrPos;
	PosToChrPos(Annot.Pos, ChrStartPosToName, ChrName, ChrPos);
	Log("Chr %s ChrPos %"PF" EndPos %"PF" %s Length %"PF"\n",
	  ChrName.c_str(),
	  ChrPos,
	  ChrPos + Annot.Length - 1,
	  Annot.TypeStr(),
	  Annot.Length);
	}

void BigEvolver::AnnotsFromGFF(const char *FileName,
  const map<string, gpos_t> &ChrNameToStartPos,
  const map<gpos_t, string> &ChrStartPosToName, const GenInfo &Genome)
	{
	m_NextGeneIndex = 0;
	m_Annots.clear();
	m_ArrayGeneIndexes.clear();

	FILE *f = OpenStdioFile(FileName);
	unsigned FileSize = (unsigned) GetStdioFileSize(f);
	GFFRecord Rec;
	unsigned RecordCount = 0;
	gpos_t ChrStartPos = 0;
	ProgressStart("Reading GFF records from %.16s", FileName);

#if	TRACE
	{
	Log("\n");
	Log("ChrNameToStartPos:\n");
	Log("       Pos  Name\n");
	Log("==========  ====\n");
	for (map<string, gpos_t>::const_iterator p = ChrNameToStartPos.begin();
	  p != ChrNameToStartPos.end(); ++p)
		Log("%10"PF"  %s\n", p->second, p->first.c_str());
	Log("\n");
	Log("       Chr       Start         End  Length         Pos  Type\n");
	Log("==========  ==========  ==========  ======  ==========  ====\n");
	}
#endif
	string PrevChrName;
	gpos_t ChrLength = GPOS_T_MAX;
	while (GetNextGFFRecord(f, Rec))
		{
		if ((RecordCount++)%1000 == 0)
			{
			unsigned Pos = (unsigned) GetStdioFilePos(f);
			ProgressStep(Pos, FileSize);
			}
		const string ChrName = Rec.SeqName;
		if (ChrName != PrevChrName)
			{
			map<string, gpos_t>::const_iterator p = ChrNameToStartPos.find(ChrName);
			if (p == ChrNameToStartPos.end())
				Die("Sequence label '%s' in '%s' is not a chromosome in '%s'",
				  ChrName.c_str(),
				  FileName,
				  Genome.GetName().c_str());

			ChrStartPos = p->second;
			PrevChrName = ChrName;

			const ChrInfo &CI = Genome.GetChrInfo(ChrName);
			ChrLength = CI.GetLength();
			}
		AnnotFromGFFRec(Rec, ChrStartPos, ChrLength);
		}
	ProgressDone();

	unsigned AnnotCount = SIZE(m_Annots);
	unsigned ArrayGeneCount = SIZE(m_ArrayGeneIndexes);
	unsigned ArrayGeneAnnotCount = SIZE(m_ArrayGeneAnnots);

	Log("\n");
	ProgressNoRes("%u GFF records", AnnotCount);
	ProgressNoRes("%u annotations", AnnotCount);
	ProgressNoRes("%u array genes", ArrayGeneCount);
	ProgressNoRes("%u array gene annots", ArrayGeneAnnotCount);
	Log("\n");

	CloseStdioFile(f);

	FixExons();
	ValidateExons();

// Validate no overlaps
	vector<AnnotBendPos> BPs;
	BPs.reserve(AnnotCount);

	for (map<gpos_t, AnnotData>::const_iterator p = m_Annots.begin();
	  p != m_Annots.end(); ++p)
		{
		const AnnotData &Annot = p->second;
		asserta(Annot.Pos == p->first);
		AnnotBendPos BP;
		BP.Pos = Annot.Pos;
		BP.Annot = &Annot;
		BP.Start = true;
		BPs.push_back(BP);
		BP.Pos = Annot.GetEndPos();
		BP.Start = false;
		BPs.push_back(BP);
		}
	sort(BPs.begin(), BPs.end());
	asserta(SIZE(BPs) == 2*AnnotCount);
	unsigned BPCount = SIZE(BPs);
	unsigned CopyCount = 0;
#if	TRACE
	Log("CopyC         Pos  Start  Annot\n");
	Log("=====  ==========  =====  =====\n");
#endif
	for (unsigned i = 0; i < BPCount; ++i)
		{
		const AnnotBendPos &BP = BPs[i];
#if	TRACE
		Log("%5u  %10"PF"  %5c  ", CopyCount, BP.Pos, tof(BP.Start));
		BP.Annot->LogMe();
		Log("\n");
#endif
		if (BP.Start)
			{
			if (CopyCount != 0)
				{
				const AnnotData &Annot = *(BP.Annot);
				asserta(i > 0);
				const AnnotBendPos &PrevBP = BPs[i-1];
				asserta(PrevBP.Start);
				const AnnotData &PrevAnnot = *(PrevBP.Annot);
				Log("\n");
				LogAnnotChr(PrevAnnot, ChrStartPosToName);
				LogAnnotChr(Annot, ChrStartPosToName);
				Die("Annots overlap");
				}
			++CopyCount;
			}
		else
			{
			if (CopyCount != 1)
				{
				LogAnnots();
				asserta(CopyCount == 1);
				asserta(i > 0);
				const AnnotData &Annot = *(BP.Annot);
				const AnnotBendPos &PrevBP = BPs[i-1];
				asserta(PrevBP.Start);
				const AnnotData &PrevAnnot = *(PrevBP.Annot);
				Log("\n");
				LogAnnotChr(PrevAnnot, ChrStartPosToName);
				LogAnnotChr(Annot, ChrStartPosToName);
				Die("Annots overlap");
				}
			--CopyCount;
			}
		}
	}
