#include "bndfile.h"
#include "utils.h"
#include "vol.h"
#include "progress.h"

void AssertBundleBounds(const VOL &Bundles, unsigned SPCount);
void AssertEdgesBounds(const VOL &Edges, unsigned SPIndex);

#define ReadData(v)		ReadStdioFile(m_f, &v, (unsigned) sizeof(v))
#define WriteData(v)	WriteStdioFile(m_f, &v, (unsigned) sizeof(v))

#define STATE(x, y)	if (m_State != x) Die("BndFile: function call in invalid state"); m_State = y;
#define STATEC(x)	if (m_State != x) Die("BndFile: function call in invalid state");

const char *SPTypeToStr(int Type)
	{
	switch (Type)
		{
	case BndFile::Hit:
		return "hit";
	case BndFile::RepeatBox:
		return "repbox";
	case BndFile::Hole:
		return "hole";
	case BndFile::Indel:
		return "indel";
	case BndFile::MicroChain:
		return "microchain";
	case BndFile::BoundingRectangle:
		return "boundrect";
		}
	static char s[32];
	sprintf(s, "?SPType=%d", Type);
	return s;
	}

const char *SPSubtypeToStr(int Subtype)
	{
	switch (Subtype)
		{
	case BndFile::None:
		return "None";
	case BndFile::Blast: 
		return "Blast";
	case BndFile::DeletedRepeat: 
		return "DeletedRepeat";
	case BndFile::Merged: 
		return "Merged";
	case BndFile::DeletedMerged: 
		return "DeletedMerged";
	case BndFile::HM1:
		return "HM1";
	case BndFile::HM2:
		return "HM2";
	case BndFile::DeletedRealigned:
		return "DeletedRealigned";
	case BndFile::DeletedMaxChain:
		return "DeletedMaxChain";
	case BndFile::TooBig:
		return "TooBig";
		}
	static char s[32];
	sprintf(s, "?SPSubtype=%d", Subtype);
	return s;
 	}

BndFile::BndFile()
	{
	m_State = Closed;
	m_f = 0;
	m_ChrIndex1 = 0;
	m_ChrIndex2 = 0;
	m_SPIndex = 0;
	}

BndFile::~BndFile()
	{
	}

void BndFile::Create(const char *FileName, const char *GenomeName1,
  const char *GenomeName2)
	{
	STATE(Closed, ChrPairWriteEnded)

	if (0 != m_f)
		Die("BndFile::Create while reading or writing");

	m_f = OpenStdioFile(FileName, true);
	m_ChrPairIndex.clear();

// Put invalid magic here, this is a protection against partially
// written files. Correct magic will be written in Close().
	m_Header.Magic = UINT_MAX;
	safestrcpy(m_Header.GenomeName1, GenomeName1, sizeof(m_Header.GenomeName1));
	safestrcpy(m_Header.GenomeName2, GenomeName2, sizeof(m_Header.GenomeName2));

// Placeholders, will overwrite later
	m_Header.ChrNameIndex1StartPos = UINT_MAX;
	m_Header.ChrNameIndex2StartPos = UINT_MAX;
	m_Header.ChrPairIndexStartPos = UINT_MAX;

	WriteStdioFile(m_f, &m_Header, sizeof(m_Header));
	}

void BndFile::StartChrPair(unsigned ChrIndex1, unsigned ChrIndex2, const char *Name1,
  const char *Name2)
	{
	STATE(ChrPairWriteEnded, ChrPairWriteStarted)

	m_ChrNameIndex1.Add(ChrIndex1, Name1);
	m_ChrNameIndex2.Add(ChrIndex2, Name2);

// Write a placeholder & remember position.
// When done with this chr pair, overwrite
// with SPCount.
	m_SPCountPos = (off_t) GetStdioFilePos(m_f);
	WriteData(m_SPIndex);

// Index points to SPCount
	m_ChrPairIndex.Add(ChrIndex1, ChrIndex2, m_SPCountPos);
	m_SPIndex = 0;
	}

void BndFile::WriteSPs(const SPVec &SPs)
	{
	ProgressStart("Writing output");
	const unsigned SPCount = (unsigned) SPs.size();
	for (unsigned SPIndex = 0; SPIndex < SPCount; ++SPIndex)
		{
		if (0 == SPIndex%10000)
			ProgressStep(SPIndex, SPCount);
		const SegPair &SP = *(SPs[SPIndex]);
		WriteSP(SP);
		}
	ProgressDone();
	}

unsigned BndFile::WriteSP(const SegPair &SP)
	{
	STATE(ChrPairWriteStarted, ChrPairWriteStarted)
	switch (SP.Type)
		{
	case OldHit:
	case RepeatBox:
	case Hole:
	case MicroChain:
	case Indel:
	case Hit:
	case BoundingRectangle:
		WriteHit(SP);
		break;
	
	default:
		Die("BndFile::WriteSP invalid SP type %u", SP.Type);
		}

	return m_SPIndex++;
	}

void BndFile::EndChrPair(const VOL &Bundles, const VOL &Edges)
	{
	STATE(ChrPairWriteStarted, ChrPairWriteEnded)
	
#ifndef NDEBUG
	AssertBundleBounds(Bundles, m_SPIndex);
	AssertEdgesBounds(Edges, m_SPIndex);
#endif

	Bundles.ToFile(m_f);
	Edges.ToFile(m_f);

// Overwrite placeholder with number
// of objects written for this chr pair.
	off64_t CurrentPos = GetStdioFilePos(m_f);
	SeekStdioFile(m_f, m_SPCountPos);
	WriteData(m_SPIndex);
	SeekStdioFile(m_f, CurrentPos);
	}

void BndFile::Close()
	{
	if (m_State == ChrPairWriteEnded)
		{
		m_Header.Magic = BndFileMagic;

		m_Header.ChrNameIndex1StartPos = GetStdioFilePos(m_f);
		m_ChrNameIndex1.ToFile(m_f);
		m_ChrNameIndex1.clear();

		m_Header.ChrNameIndex2StartPos = GetStdioFilePos(m_f);
		m_ChrNameIndex2.ToFile(m_f);
		m_ChrNameIndex2.clear();

		m_Header.ChrPairIndexStartPos = GetStdioFilePos(m_f);
		m_ChrPairIndex.ToFile(m_f);
		m_ChrPairIndex.clear();

		WriteStdioFile(m_f, 0, &m_Header, sizeof(m_Header));

		CloseStdioFile(m_f);
		m_f = 0;

		m_State = Closed;
		}
	else if (m_State == Opened)
		{
		CloseStdioFile(m_f);
		m_f = 0;
		}
	else
		Die("BndFile::Close called in invalid state");
	}

void BndFile::Open(const char *FileName)
	{
	STATE(Closed, Opened)

	m_f = OpenStdioFile(FileName);

	ReadStdioFile(m_f, &m_Header, sizeof(m_Header));
	if (m_Header.Magic != BndFileMagic)
		Die("Header magic mismatch");

	SeekStdioFile(m_f, (off64_t) m_Header.ChrNameIndex1StartPos);
	m_ChrNameIndex1.FromFile(m_f);

	SeekStdioFile(m_f, (off64_t) m_Header.ChrNameIndex2StartPos);
	m_ChrNameIndex2.FromFile(m_f);

	SeekStdioFile(m_f, (off64_t) m_Header.ChrPairIndexStartPos);
	m_ChrPairIndex.FromFile(m_f);
	}

const ChrPairIndex &BndFile::GetChrPairIndex() const
	{
	STATEC(Opened)

	return m_ChrPairIndex;
	}

const ChrNameIndex &BndFile::GetChrNameIndex1() const
	{
	STATEC(Opened)

	return m_ChrNameIndex1;
	}

const ChrNameIndex &BndFile::GetChrNameIndex2() const
	{
	STATEC(Opened)

	return m_ChrNameIndex2;
	}

const char *BndFile::GetGenomeName1() const
	{
	STATEC(Opened)

	return m_Header.GenomeName1;
	}

const char *BndFile::GetGenomeName2() const
	{
	STATEC(Opened)

	return m_Header.GenomeName2;
	}

bool BndFile::ReadChrPair(unsigned ChrIndex1, unsigned ChrIndex2, 
  SPVec &SPs, VOL &Bundles, VOL &Edges) const
	{
	off_t SPCountPos;
	bool Found = m_ChrPairIndex.Lookup(ChrIndex1, ChrIndex2, &SPCountPos);
	if (!Found)
		return false;

	SeekStdioFile(m_f, SPCountPos);
	unsigned SPCount;
	ReadData(SPCount);

	SPs.clear();
	SPs.reserve(SPCount);

	for (unsigned SPIndex = 0; SPIndex < SPCount; ++SPIndex)
		{
		SegPair *SP = ReadSP();
		SPs.push_back(SP);
		}
	if (SPs.size() != SPCount)
		Die("SPs.size()=%u SPCount=%u", (unsigned) SPs.size(), SPCount);

	Bundles.FromFile(m_f);
	Edges.FromFile(m_f);

	return true;
	}

SegPair *BndFile::ReadSP() const
	{
	byte Type;
	ReadData(Type);
	SegPair *SP = 0;
	switch (Type)
		{
	case Hit:
		SP = new SegPair;
		SP->Type = Hit;
		ReadHit(*SP);
		break;

	//case OldHit:
	//	SP = new SegPair;
	//	SP->Type = Hit;
	//	SP->AlnLength = 0;
	//	ReadOldHit(*SP);
	//	break;

	case RepeatBox:
		SP = new SegPair;
		SP->Type = RepeatBox;
		SP->AlnLength = 0;
		ReadHit(*SP);
		break;

	case Hole:
		SP = new SegPair;
		SP->Type = Hole;
		SP->AlnLength = 0;
		ReadHit(*SP);
		break;

	case Indel:
		SP = new SegPair;
		SP->Type = Indel;
		SP->AlnLength = 0;
		ReadHit(*SP);
		break;

	case BoundingRectangle:
		SP = new SegPair;
		SP->Type = BoundingRectangle;
		SP->AlnLength = 0;
		ReadHit(*SP);
		break;

	default:
		Die("BndFile::ReadSP, unknown SegPair type %u", Type);
		}

	return SP;
	}

void BndFile::ReadHit(SegPair &SP) const
	{
//	ReadData(SP.Type);	// already read in ReadSP()
	ReadData(SP.Subtype);
	ReadData(SP.Segs[0].lo);
	ReadData(SP.Segs[0].hi);
	ReadData(SP.Segs[1].lo);
	ReadData(SP.Segs[1].hi);
	ReadData(SP.Reverse);
	ReadData(SP.Score);
	ReadData(SP.AlnLength);
	}

void BndFile::WriteHit(const SegPair &SP)
	{
	WriteData(SP.Type);
	WriteData(SP.Subtype);
	WriteData(SP.Segs[0].lo);
	WriteData(SP.Segs[0].hi);
	WriteData(SP.Segs[1].lo);
	WriteData(SP.Segs[1].hi);
	WriteData(SP.Reverse);
	WriteData(SP.Score);
	WriteData(SP.AlnLength);
	}

const char *BndFile::GetChrName1(unsigned ChrIndex) const
	{
	return m_ChrNameIndex1.GetName(ChrIndex);
	}

const char *BndFile::GetChrName2(unsigned ChrIndex) const
	{
	return m_ChrNameIndex2.GetName(ChrIndex);
	}

void FreeSPs(SPVec &SPs)
	{
	for (SPVec::iterator p = SPs.begin(); p != SPs.end(); ++p)
		{
		SegPair *SP = *p;
		delete SP;
		*p = 0;
		}
	}

bool BndFile::IsActiveHit(const SegPair &SP)
	{
	if (BndFile::Hit != SP.Type)
		return false;
	if (BndFile::DeletedMerged == SP.Subtype || BndFile::DeletedRepeat == SP.Subtype)
		return false;
	return true;
	}

bool BndFile::IsDeleted(const SegPair &SP)
	{
#define	DEL(t, subt)											\
	if (SP.Type == BndFile::t && SP.Subtype == BndFile::subt)	\
		return true;

	DEL(Hit, DeletedRepeat)
	DEL(Hit, DeletedMerged)
	DEL(Hit, DeletedRealigned)
	DEL(Hit, DeletedMaxChain)
#undef HIT
	return false;
	}

SegPair *SPCopy(const SegPair &SP)
	{
	SegPair *Copy = new SegPair;
	*Copy = SP;
	return Copy;
	}

void AssertBundleBounds(const VOL &Bundles, unsigned SPCount)
	{
	unsigned BundleCount = Bundles.GetVectorLength();
	for (unsigned BundleIndex = 0; BundleIndex < BundleCount; ++BundleIndex)
		{
		const IntList &SPIndexes = Bundles.GetList(BundleIndex);
		for_CIntList(SPIndexes, p)
			{
			unsigned SPIndex = *p;
			if (SPIndex >= SPCount)
				Die("AssertBundleBounds, SPCount=%u bundle=%u, spindex=%u",
				  SPCount, BundleIndex, SPIndex);
			}
		}
	}

void AssertEdgesBounds(const VOL &Edges, unsigned SPCount)
	{
	if (Edges.GetVectorLength() != SPCount)
		Die("AssertEdgesBounds SPCount=%u Edges=%u",
		  SPCount, Edges.GetVectorLength());

	for (unsigned SPIndex = 0; SPIndex < SPCount ; ++SPIndex)
		{
		const IntList &SPIndexes = Edges.GetList(SPIndex);
		for_CIntList(SPIndexes, p)
			{
			unsigned SPIndex2 = *p;
			if (SPIndex2 >= SPCount)
				Die("AssertEdgeBounds, SPCount=%u spindex=%u, spindex2=%u",
				  SPCount, SPIndex, SPIndex2);
			}
		}
	}

#ifndef NDEBUG
void AssertValidSegPair(const SegPair &SP)
	{
	if (SP.Segs[0].lo > SP.Segs[0].hi)
		{
		LogSPN(SP);
		Die("AssertValidSegPair: lo > hi type=%d", SP.Type);
		}
	if (SP.Segs[1].lo > SP.Segs[1].hi)
		{
		LogSPN(SP);
		Die("AssertValidSegPair: lo > hi type=%d", SP.Type);
		}
	}
#endif	// NDEBUG

void ShrinkSeg(const Seg &InSeg, Seg &OutSeg, unsigned Margin)
	{
	coord_t Mid = (InSeg.lo + InSeg.hi)/2;
	if (InSeg.hi <= Margin)
		{
		OutSeg.lo = Mid;
		OutSeg.hi = Mid;
		return;
		}

	OutSeg.lo += Margin;
	OutSeg.hi -= Margin;

	if (InSeg.lo > InSeg.hi)
		{
		OutSeg.lo = Mid;
		OutSeg.hi = Mid;
		}
	}

void ShrinkSP(const SegPair &InSP, SegPair &OutSP, unsigned Margin)
	{
	OutSP = InSP;
	ShrinkSeg(InSP.Segs[0], OutSP.Segs[0], Margin);
	ShrinkSeg(InSP.Segs[1], OutSP.Segs[1], Margin);
	}
