#include "cmafile.h"
#include "chrnameindex.h"
#include "utils.h"

/***
File format is as follows:

   Header
        <GenomeCount>
        <AlnCount>
        <AlnIndexPos>	// address of <AlnPosStartMagic>
   GenomeName1_Length
   GenomeName1
   GenomeName2_Length
   GenomeName2
   ...
   GenomeName_<GenomeCount>_Length
   GenomeName_<GenomeCount>
   ChrNameIndex_1                
   ChrNameIndex_2               
   ...                          
   ChrNameIndex_<GenomeCount>    
   AlnCoords_1					     
   AlnRowVec_1
   AlnCoords_2					     
   AlnRowVec_2					   
   ...                            
   AlnCoords<AlnCount>              
   AlnRowVec_<AlnCount>
   <AlnPosStartMagic>
   AlnPos_1     
   AlnPos_2						  
   ...								  
   AlnPos_<AlnCount>
   <AlnPosEndMagic>

All structures are fixed-length except for the
alignments themselves (AlnRowVec's).

This format allows alignments to be written
to the file sequentially without having to
keep them in memory.
***/

#define ReadData(v)		ReadStdioFile(m_f, &v, (unsigned) sizeof(v))
#define WriteData(v)	WriteStdioFile(m_f, &v, (unsigned) sizeof(v))

void RevComp(char *S, unsigned L);
void HitToAln(const SegPair &Hit, const char *Path,
  char **ptrRow1, char **ptrRow2);

CMAFile::CMAFile()
	{
	m_f = NULL;
	m_State = STATE_Closed;
	m_GenomeCount = 0;
	m_AlnCount = 0;
	}

CMAFile::~CMAFile()
	{
	Close();
	}

void CMAFile::Close()
	{
	switch (m_State)
		{
	case STATE_Open:
		CloseOpen();
		break;
	case STATE_Creating:
		CloseCreate();
		break;
	case STATE_Closed:
		return;
	default:
		Die("CMAFile::Close called in invalid state %u", (unsigned) m_State);
		}
	m_State = STATE_Closed;
	}

void CMAFile::Open(const char *FileName)
	{
	RequiredState(STATE_Closed);
	m_f = OpenStdioFile(FileName);

// Read and validate header
	ReadData(m_FileHeader);

	if (m_FileHeader.Magic != CMAFileMagic)
		Die("CMAFile::Open, bad magic 0x%x should be 0x%x",
		  m_FileHeader.Magic, CMAFileMagic);
	if (m_FileHeader.Magic2 != CMAFileMagic2)
		Die("CMAFile::Open, bad magic2 0x%x should be 0x%x, cross-platform packing mismatch?",
		  m_FileHeader.Magic2, CMAFileMagic2);
	if (m_FileHeader.GenomeCount < 2)
		Die("CMAFile::Open, genome count = %u, must be >= 2", m_FileHeader.GenomeCount);

	m_GenomeCount = m_FileHeader.GenomeCount;
	m_AlnCount = m_FileHeader.AlnCount;

// Read names
	m_GenomeNames.clear();
	for (unsigned GenomeIndex = 0; GenomeIndex < m_GenomeCount; ++GenomeIndex)
		{
		unsigned NameLength;
		ReadData(NameLength);
		char *Name = new char[NameLength+1];
		ReadStdioFile(m_f, Name, NameLength);
		Name[NameLength] = 0;
		std::string strName(Name);
		m_GenomeNames.push_back(strName);
		delete[] Name;
		}

// Read chromosome name indexes
	m_ChrNameIndexes.clear();
	for (unsigned GenomeIndex = 0; GenomeIndex < m_GenomeCount; ++GenomeIndex)
		{
		ChrNameIndex *Index = new ChrNameIndex;
		Index->FromFile(m_f, true);
		m_ChrNameIndexes.push_back(Index);
		}

	off64_t CoordsPos = GetStdioFilePos(m_f);

	SeekStdioFile(m_f, (off64_t) m_FileHeader.AlnIndexStartPos);
	unsigned Magic;
	ReadData(Magic);
	if (Magic != AlnPosStartMagic)
		Die("CMAFile::Open, AlnPosStartMagic=0x%x, should be 0x%x",
		  Magic, AlnPosStartMagic);

	const unsigned AlnCount = m_FileHeader.AlnCount;
	m_AlnFilePos.clear();
	for (unsigned AlnIndex = 0; AlnIndex < AlnCount; ++AlnIndex)
		{
		uint64 Pos;
		ReadData(Pos);
		m_AlnFilePos.push_back(Pos);
		}

	ReadData(Magic);
	if (Magic != AlnPosEndMagic)
		Die("CMAFile::Open, AlnPosEndMagic=0x%x, should be 0x%x",
		  Magic, AlnPosEndMagic);

// Read coordinates (without saving) to build m_CoordsMap
// index (which is saved).
	SeekStdioFile(m_f, CoordsPos);
	m_CoordsMap.clear();
	for (unsigned AlnIndex = 0; AlnIndex < m_AlnCount; ++AlnIndex)
		{
		uint64 Pos = GetAlnFilePos(AlnIndex);
		SeekStdioFile(m_f, (off64_t) Pos);

		CoordVec Coords;
		Coords.reserve(m_GenomeCount);
		for (unsigned GenomeIndex = 0; GenomeIndex < m_GenomeCount; ++GenomeIndex)
			{
			CoordData c;
			ReadData(c.ChrIndex);
			ReadData(c.StartPos);
			ReadData(c.EndPos);
			ReadData(c.Reversed);
			Coords.push_back(c);
			}
		
		const std::pair<const CoordVec, unsigned> Value(Coords, AlnIndex);
		m_CoordsMap.insert(Value);
		}
	}

const char *CMAFile::GetGenomeName(unsigned GenomeIndex) const
	{
	assert(GenomeIndex < GetGenomeCount());
	return m_GenomeNames[GenomeIndex].c_str();
	}

const ChrNameIndex &CMAFile::GetChrNameIndex(unsigned GenomeIndex) const
	{
	assert(GenomeIndex < GetGenomeCount());
	return *(m_ChrNameIndexes[GenomeIndex]);
	}

unsigned CMAFile::GetCMACount() const
	{
	return m_AlnCount;
	}

void CMAFile::Create(const char *FileName, const CMAFile &SourceFile)
	{
	Create(FileName, SourceFile.GetGenomeNames(), SourceFile.GetChrNameIndexVec());
	}

void CMAFile::Create(const char *FileName, const StdStrVec &GenomeNames,
  const ChrNameIndexVec &ChrNameIndexes)
	{
	RequiredState(STATE_Closed);
	m_f = CreateStdioFile(FileName);

	if (GenomeNames.size() != ChrNameIndexes.size())
		Die("CMAFile::Create, vectors must be same size %u %u",
		  (unsigned) GenomeNames.size(), (unsigned) ChrNameIndexes.size());

	m_GenomeCount = (unsigned) GenomeNames.size();

// Write placeholder header with invalid magics.
// Overwrite with correct header later when everything else
// successfully completed.
	memset(&m_FileHeader, 0, sizeof(m_FileHeader));
	WriteData(m_FileHeader);
	WriteStdioFile(m_f, 0, &m_FileHeader, sizeof(m_FileHeader));

// Write and store genome names
	for (unsigned GenomeIndex = 0; GenomeIndex < m_GenomeCount; ++GenomeIndex)
		{
		const char *Name = GenomeNames[GenomeIndex].c_str();
		unsigned NameLength = (unsigned) strlen(Name);
		WriteData(NameLength);
		WriteStdioFile(m_f, Name, NameLength);
		m_GenomeNames.push_back(Name);
		}

// Write and store chromosome name indexes
	for (unsigned GenomeIndex = 0; GenomeIndex < m_GenomeCount; ++GenomeIndex)
		{
	// hack to cast away const (yuck)
		ChrNameIndex *Index = (ChrNameIndex *) ChrNameIndexes[GenomeIndex];
		Index->ToFile(m_f);
		}

	m_State = STATE_Creating;
	}

void CMAFile::AppendCMA(const CoordVec &Coords, const StrVec &Rows)
	{
#ifndef NDEBUG
	AssertValid(Coords, Rows);
#endif
	RequiredState(STATE_Creating);

// Save file position,
	off64_t Pos = GetStdioFilePos(m_f);
	m_AlnFilePos.push_back(Pos);

// Write coords
	for (unsigned GenomeIndex = 0; GenomeIndex < m_GenomeCount; ++GenomeIndex)
		{
		const CoordData &c = Coords[GenomeIndex];
		WriteData(c.ChrIndex);
		WriteData(c.StartPos);
		WriteData(c.EndPos);
		WriteData(c.Reversed);
		}

// Write rows
// RowLength = # cols
	const unsigned RowLength = (unsigned) strlen(Rows[0]);
	WriteData(RowLength);
	for (unsigned GenomeIndex = 0; GenomeIndex < m_GenomeCount; ++GenomeIndex)
		{
		const char *Row = Rows[GenomeIndex];
		WriteStdioFile(m_f, Row, RowLength);
		}

	++m_AlnCount;
	}

void CMAFile::CloseOpen()
	{
	RequiredState(STATE_Open);
	CloseShared();
	}

void CMAFile::CloseShared()
	{
	m_ChrNameIndexes.clear();
	m_GenomeNames.clear();
	m_AlnFilePos.clear();
	m_CoordsMap.clear();
	CloseStdioFile(m_f);
	m_f = 0;
	m_GenomeCount = 0;
	m_AlnCount = 0;
	m_State = STATE_Closed;
	}

void CMAFile::CloseCreate()
	{
	RequiredState(STATE_Creating);

	off64_t Pos = GetStdioFilePos(m_f);

	memset(&m_FileHeader, 0, sizeof(m_FileHeader));
	m_FileHeader.GenomeCount = m_GenomeCount;
	m_FileHeader.AlnCount = m_AlnCount;
	m_FileHeader.AlnIndexStartPos = (uint64) Pos;
	m_FileHeader.Magic = CMAFileMagic;
	m_FileHeader.Magic2 = CMAFileMagic2;

// Write AlnPos vector
	WriteData(AlnPosStartMagic);
	const unsigned AlnCount = (unsigned) m_AlnFilePos.size();
	for (unsigned AlnIndex = 0; AlnIndex < AlnCount; ++AlnIndex)
		{
		uint64 Pos = (uint64) m_AlnFilePos[AlnIndex];
		WriteData(Pos);
		}
	WriteData(AlnPosEndMagic);

// Write header
	SeekStdioFile(m_f, 0);
	WriteData(m_FileHeader);

	CloseShared();
	}

void CMAFile::RequiredState(STATE s) const
	{
	if (m_State != s)
		Die("CMAFile: function called in invalid state %u", (unsigned) s);
	}

void CMAFile::AssertValid(const CoordVec &Coords, const StrVec &Rows) const
	{
	const unsigned GenomeCount = GetGenomeCount();

	if (Coords.size() != GenomeCount)
		Die("Invalid CMA, Coords.size=%u != GenomeCount=%u",
		  (unsigned) Coords.size(), GenomeCount);

	if (Rows.size() != GenomeCount)
		Die("Invalid CMA, Rows.size=%u != GenomeCount=%u",
		  (unsigned) Rows.size(), GenomeCount);

	const unsigned RowLength0 = (unsigned) strlen(Rows[0]);
	for (unsigned GenomeIndex = 1; GenomeIndex < GenomeCount; ++GenomeIndex)
		{
		unsigned RowLength = (unsigned) strlen(Rows[GenomeIndex]);
		if (RowLength != RowLength0)
			Die("CMAFile: invalid alignment, row %u length differs %u %u",
			  GenomeIndex, RowLength0, RowLength);
		}
	}

bool CMAFile::GetCMA(const CoordsVec &Coords, StrVec &Rows) const
	{
	if (Coords.size() != m_GenomeCount)
		return false;

	CoordsMap::const_iterator p = m_CoordsMap.find(Coords);
	if (p == m_CoordsMap.end())
		return false;
	unsigned CMAIndex = p->second;

	CoordsVec Coords2;
	GetCMA(CMAIndex, Coords2, Rows);

#ifndef NDEBUG
	for (unsigned GenomeIndex = 0; GenomeIndex < m_GenomeCount; ++GenomeIndex)
		if (!(Coords[GenomeIndex] == Coords2[GenomeIndex]))
			Die("CMAFile::GetCMA, coords mismatch");
#endif
	return true;
	}

void CMAFile::GetCMA(unsigned CMAIndex, CoordsVec &Coords, StrVec &Rows) const
	{
	if (CMAIndex >= GetCMACount())
		Die("CMAFile::GetCMA(%u), there are %u CMAs", CMAIndex, GetCMACount());

	uint64 Pos = m_AlnFilePos[CMAIndex];
	SeekStdioFile(m_f, (off64_t) Pos);

// Read coords
	Coords.clear();
	Coords.reserve(m_GenomeCount);
	for (unsigned GenomeIndex = 0; GenomeIndex < m_GenomeCount; ++GenomeIndex)
		{
		CoordData c;
		ReadData(c.ChrIndex);
		ReadData(c.StartPos);
		ReadData(c.EndPos);
		ReadData(c.Reversed);
		Coords.push_back(c);
		}

// Read rows
	unsigned RowLength;
	ReadData(RowLength);

	Rows.clear();
	Rows.reserve(m_GenomeCount);
	for (unsigned GenomeIndex = 0; GenomeIndex < m_GenomeCount; ++GenomeIndex)
		{
		char *Row = myalloc(char, RowLength+1);
		ReadStdioFile(m_f, Row, RowLength);
		Row[RowLength] = 0;
		Rows.push_back(Row);
		}
	}

CMP CoordsCmp(const CoordData &c1, const CoordData &c2)
	{
	if (c1.ChrIndex == c2.ChrIndex)
		{
		if (c1.StartPos == c2.StartPos)
			{
			if (c1.EndPos == c2.EndPos)
				{
				if (c1.Reversed == c2.Reversed)
					return CMP_EQ;
				else
					{
					if (c1.Reversed < c2.Reversed)
						return CMP_LT;
					else
						return CMP_GT;
					}
				}
			else
				{
				if (c1.EndPos < c2.EndPos)
					return CMP_LT;
				else
					return CMP_GT;
				}
			}
		else
			{
			if (c1.StartPos < c2.StartPos)
				return CMP_LT;
			else
				return CMP_GT;
			}
		}
	else
		{
		if (c1.ChrIndex < c2.ChrIndex)
			return CMP_LT;
		else
			return CMP_GT;
		}
	}

bool CoordsVecLTClass::operator()(const CoordVec &c1, const CoordVec &c2) const
	{
	const unsigned GenomeCount = (unsigned) c1.size();
	assert(GenomeCount == c2.size());
	for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
		{
		CMP c = CoordsCmp(c1[GenomeIndex], c2[GenomeIndex]);
		if (c == CMP_LT)
			return true;
		if (c == CMP_GT)
			return false;
		}
	return false;
	}

void FreeStrVec(StrVec &Rows)
	{
	const unsigned GenomeCount = (unsigned) Rows.size();
	for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
		{
		myfree(Rows[GenomeIndex]);
		Rows[GenomeIndex] = 0;
		}
	}

void FreePaths(STRVEC &Paths)
	{
	for (STRVEC::iterator p = Paths.begin(); p != Paths.end(); ++p)
		{
		char *str = *p;
		if (str != 0)
			myfree(str);
		}
	}

void CMAFile::AppendHits(const SPVec &SPs, const STRVEC &Paths)
	{
	const unsigned SPCount = (unsigned) SPs.size();
	for (unsigned SPIndex = 0; SPIndex < SPCount; ++SPIndex)
		AppendHit(*(SPs[SPIndex]), Paths[SPIndex]);
	}

void CMAFile::AppendHit(const SegPair &Hit, const char *Path)
	{
	extern unsigned g_ChrIndex1;
	extern unsigned g_ChrIndex2;

	CoordData c0;
	CoordData c1;

	c0.ChrIndex = g_ChrIndex1;
	c1.ChrIndex = g_ChrIndex2;

	c0.StartPos = Hit.Segs[0].lo;
	c1.StartPos = Hit.Segs[1].lo;

	c0.EndPos = Hit.Segs[0].hi;
	c1.EndPos = Hit.Segs[1].hi;

	c0.Reversed = false;
	c1.Reversed = Hit.Reverse;

	CoordVec Coords;
	Coords.push_back(c0);
	Coords.push_back(c1);

	char *Row1;
	char *Row2;
	HitToAln(Hit, Path, &Row1, &Row2);

	StrVec Rows;
	Rows.push_back(Row1);
	Rows.push_back(Row2);

	AppendCMA(Coords, Rows);

	FreeRows(Rows);
	}

uint64 CMAFile::GetAlnFilePos(unsigned CMAIndex) const
	{
	if (CMAIndex >= m_AlnCount)
		Die("CMAFile::GetAlnFilePos(%u), %u alignments",
		  CMAIndex, m_AlnCount);
	//off64_t OffPos = (off64_t) m_FileHeader.AlnIndexStartPos + (off64_t) (CMAIndex*CMAIndex);
	//SeekStdioFile(m_f, OffPos);

	//uint64 Pos;
	//ReadData(Pos);
	uint64 Pos = m_AlnFilePos[CMAIndex];
	return Pos;
	}

void LogCoords(const CoordVec &Coords)
	{
	const unsigned GenCount = (unsigned) Coords.size();
	for (unsigned GenIndex = 0; GenIndex < GenCount; ++GenIndex)
		{
		const CoordData &CD = Coords[GenIndex];

		if (GenIndex != 0)
			Log("; ");
		Log("Gen=%u Chr=%u lo=%u hi=%u Rev=%c",
		  GenIndex,
		  CD.ChrIndex,
		  (unsigned) CD.StartPos,
		  (unsigned) CD.EndPos,
		  CD.Reversed ? 'T' : 'F');
		}
	}
