#include "utils.h"
#include "revfile.h"
#include "bndfile.h"
#include "seqfile.h"
#include "getopts.h"
#include "progress.h"

#define TRACE	0

struct ChrBundle
	{
	unsigned ChrId0;
	unsigned ChrId1;
	unsigned BundleId;

	bool operator<(const ChrBundle &rhs) const
		{
		if (ChrId0 != rhs.ChrId0)
			return ChrId0 < rhs.ChrId0;
		if (ChrId1 != rhs.ChrId1)
			return ChrId1 < rhs.ChrId1;
		return BundleId < rhs.BundleId;
		}
	};

// Each chromosome pair has a list of bundles.
// Each bundle is an IntList of SP indexes.
typedef std::pair<unsigned, unsigned> ChrPair;

static std::set<ChrPair> ChrPairs;
static std::map<ChrPair, IntSet> g_ChrPairToBundleIndexes;
static std::map<ChrPair, IntList> g_ChrPairToNonBundleSPIndexes;
static std::map<ChrBundle, IntList> g_BundleIdToSPIndexes;
static SPVec g_SPs;

static void Add(const AlnInfo &AI)
	{
	if (AI.Segs.size() != 2)
		Die("Rev file contains AlnInfo with %u segments, must be pairwise",
		  (unsigned) AI.Segs.size());

	const GCSeg Seg0 = AI.GetSeg(0);
	const GCSeg Seg1 = AI.GetSeg(1);

	SegPair &SP = *new SegPair;

	SP.Segs[0].lo = Seg0.lo;
	SP.Segs[1].lo = Seg1.lo;

	SP.Segs[0].hi = Seg0.hi;
	SP.Segs[1].hi = Seg1.hi;

	SP.Type = BndFile::Hit;
	SP.Subtype = BndFile::None;
	SP.Score = (sp_score_t) AI.Score;
	SP.AlnLength = AI.GetAlnLength();
	SP.Reverse = (Seg0.Plus != Seg1.Plus);

	unsigned SPIndex = (unsigned) g_SPs.size();
	g_SPs.push_back(&SP);

	ChrPair CP(Seg0.ChrId, Seg1.ChrId);
	ChrPairs.insert(CP);

	unsigned BundleId = AI.GetBundleId();
	if (BundleId == UINT_MAX)
		{
		std::map<ChrPair, IntList>::iterator p =
		  g_ChrPairToNonBundleSPIndexes.find(CP);
		if (p == g_ChrPairToNonBundleSPIndexes.end())
			{
			IntList SPIndexes;
			SPIndexes.push_back(SPIndex);
			std::map<ChrPair, IntList>::value_type Value(CP, SPIndexes);
			g_ChrPairToNonBundleSPIndexes.insert(Value);
			}
		else
			{
			IntList &SPIndexes = p->second;
			SPIndexes.push_back(SPIndex);
			}
		}
	else
		{
		assert(BundleId != UINT_MAX);
		std::map<ChrPair, IntSet>::iterator p = g_ChrPairToBundleIndexes.find(CP);
		if (p == g_ChrPairToBundleIndexes.end())
			{
			IntSet BundleIndexes;
			BundleIndexes.insert(BundleId);
			std::map<ChrPair, IntSet>::value_type Value(CP, BundleIndexes);
			g_ChrPairToBundleIndexes.insert(Value);
			}
		else
			{
			IntSet &BundleIndexes = p->second;
			BundleIndexes.insert(BundleId);
			}
		}

	ChrBundle BI;
	BI.ChrId0 = Seg0.ChrId;
	BI.ChrId1 = Seg1.ChrId;
	BI.BundleId = BundleId;

	std::map<ChrBundle, IntList>::iterator p = g_BundleIdToSPIndexes.find(BI);
	if (p == g_BundleIdToSPIndexes.end())
		{
		IntList SPIndexes;
		SPIndexes.push_back(SPIndex);
		std::map<ChrBundle, IntList>::value_type Value(BI, SPIndexes);
		g_BundleIdToSPIndexes.insert(Value);
		}
	else
		{
		IntList &SPIndexes = p->second;
		SPIndexes.push_back(SPIndex);
		}
	}

const IntList &GetSPIndexes(const ChrBundle &BI)
	{
	std::map<ChrBundle, IntList>::iterator p = g_BundleIdToSPIndexes.find(BI);
	if (p == g_BundleIdToSPIndexes.end())
		Die("ChrId0=%u, ChrId1=%u, Bundle=%u: not found",
		  BI.ChrId0, BI.ChrId1, BI.BundleId);
	return p->second;
	}

static void Output(BndFile &Bnd, const RevFile &Rev)
	{
	const unsigned SPCount = (unsigned) g_SPs.size();

// For each chromosome pair in Rev file
	Progress("Writing");
	unsigned PairCount = (unsigned) ChrPairs.size();
	unsigned Counter = 0;
	for (std::set<ChrPair>::const_iterator ptrChrPair = ChrPairs.begin();
	  ptrChrPair != ChrPairs.end(); ++ptrChrPair)
		{
		++Counter;

		const ChrPair &CP = *ptrChrPair;

		std::string NameA = Rev.GetChrName(0, CP.first);
		std::string NameB = Rev.GetChrName(1, CP.second);

		double Pct = (double) Counter*100.0 / (double) PairCount;
		Progress("[%u] %s, [%u] %s, %u of %u  %.1f%%",
		  CP.first,
		  NameA.c_str(),
		  CP.second,
		  NameB.c_str(),
		  Counter,
		  PairCount,
		  Pct);

		unsigned ChrId0 = CP.first;
		unsigned ChrId1 = CP.second;

		Bnd.StartChrPair(ChrId0, ChrId1, NameA.c_str(), NameB.c_str());

	// Pass 1: Count SPs & bundles for this chr pair, map
	// global SP indexes to local SP indexes 0, 1 .. max for
	// this chr pair.
		std::map<unsigned, unsigned> GlobalToLocalSPIndex;
		unsigned ChrPairSPCount = 0;
		unsigned MaxBundleIndex = UINT_MAX;

	// Pass 1a: SPs which had bundle indexes assigned
		std::map<ChrPair, IntSet>::iterator p = g_ChrPairToBundleIndexes.find(CP);
		if (p != g_ChrPairToBundleIndexes.end())
			{
			const IntSet &BundleIndexes = p->second;
			for_CIntSet(BundleIndexes, q)
				{
				unsigned BundleId = *q;
				if (MaxBundleIndex == UINT_MAX || BundleId > MaxBundleIndex)
					MaxBundleIndex = BundleId;

				ChrBundle BI;
				BI.ChrId0 = ChrId0;
				BI.ChrId1 = ChrId1;
				BI.BundleId = BundleId;

				const IntList &SPIndexes = GetSPIndexes(BI);
				for_CIntList(SPIndexes, s)
					{
					unsigned GlobalSPIndex = *s;
					GlobalToLocalSPIndex[GlobalSPIndex] = ChrPairSPCount++;
					}
				}
			}
		unsigned ChrPairBundleCount = (MaxBundleIndex == UINT_MAX ? 0 : MaxBundleIndex + 1);

	// Pass 1b: SPs with no bundle index assigned
		std::map<ChrPair, IntList>::iterator p2 = g_ChrPairToNonBundleSPIndexes.find(CP);
		if (p2 != g_ChrPairToNonBundleSPIndexes.end())
			{
			const IntList &SPIndexes = p2->second;
			for_CIntList(SPIndexes, q)
				{
				unsigned GlobalSPIndex = *q;
				GlobalToLocalSPIndex[GlobalSPIndex] = ChrPairSPCount++;
				}
			ChrPairBundleCount += (unsigned) SPIndexes.size();
			}

	// Now we have sizes needed to create the VOLs.
		VOL Bundles;
		VOL Edges;
		Bundles.Create(ChrPairBundleCount, ChrPairSPCount);
		Edges.Create(ChrPairSPCount);

	// Pass 2: Write SPs and build Bundle VOL.
	// Pass 2a: SPs which had bundle indexes assigned
		p = g_ChrPairToBundleIndexes.find(CP);
		if (p != g_ChrPairToBundleIndexes.end())
			{
			const IntSet &BundleIndexes = p->second;
			for_CIntSet(BundleIndexes, q)
				{
				unsigned BundleId = *q;
				assert(BundleId < ChrPairBundleCount);

				ChrBundle BI;
				BI.ChrId0 = ChrId0;
				BI.ChrId1 = ChrId1;
				BI.BundleId = BundleId;

#if	TRACE
				Log("ChrId0=%u ChrId1=%u Bundle=%u\n", ChrId0, ChrId1, BundleId);
#endif
				const IntList &SPIndexes = GetSPIndexes(BI);;
				for_CIntList(SPIndexes, s)
					{
					unsigned GlobalSPIndex = *s;
					assert(GlobalSPIndex < SPCount);

					unsigned ChrPairSPIndex = GlobalToLocalSPIndex[GlobalSPIndex];
					assert(ChrPairSPIndex < ChrPairSPCount);

#if	TRACE
					Log("Write SP Global=%u  ChrPair=%u\n", GlobalSPIndex, ChrPairSPIndex);
#endif

					const SegPair &SP = *(g_SPs[GlobalSPIndex]);
					Bnd.WriteSP(SP);
					Bundles.Append(BundleId, ChrPairSPIndex);
					}
				}
			}

	// Pass 2b: SPs with no bundle index assigned
		unsigned BundleId = (MaxBundleIndex == UINT_MAX ? 0 : MaxBundleIndex + 1);
		p2 = g_ChrPairToNonBundleSPIndexes.find(CP);
		if (p2 != g_ChrPairToNonBundleSPIndexes.end())
			{
			const IntList &SPIndexes = p2->second;
			for_CIntList(SPIndexes, q)
				{
				unsigned GlobalSPIndex = *q;
				unsigned SPIndex = GlobalToLocalSPIndex[GlobalSPIndex];
				assert(SPIndex < ChrPairSPCount);
				const SegPair &SP = *(g_SPs[SPIndex]);
				Bnd.WriteSP(SP);
				Bundles.Append(BundleId++, SPIndex);
				}
			}

		Bnd.EndChrPair(Bundles, Edges);
		}
	Progress("Done");
	}

void CvtRev2Bnd()
	{
	const char *RevFileName = RequiredValueOpt("fromrev");
	const char *BndFileName = RequiredValueOpt("tobnd");

	RevFile Rev;
	BndFile Bnd;

	Progress("Opening %s", RevFileName);
	Rev.Open(RevFileName);

	const std::string &GenomeNameA = Rev.GetGenomeName(0);
	const std::string &GenomeNameB = Rev.GetGenomeName(1);
	
	Progress("Genomes %s, %s", GenomeNameA.c_str(), GenomeNameB.c_str());

	const unsigned AICount = Rev.GetAlnInfoCount();
	ProgressStart("Reading");
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%1000 == 0)
			ProgressStep(AIIndex, AICount);
		AlnInfo AI;
		Rev.GetAlnInfo(AIIndex, AI);
		Add(AI);
		}
	ProgressDone();

	Bnd.Create(BndFileName, GenomeNameA.c_str(), GenomeNameB.c_str());
	Output(Bnd, Rev);
	Bnd.Close();
	}
