#include "utils.h"
#include "revfile.h"
#include "progress.h"

void TruncLabel(std::string &Label);
void LogAIRows(const AlnInfo &AI, const StdStrVec &Rows, const RevFile *Rev = 0);

extern unsigned g_MaxUnstackedSegCount;

extern unsigned g_UnstackedBaseCount;
extern unsigned g_TotalUnstackedBaseCount;
extern unsigned g_UnstackedSegCount;

static coord_t g_SegStart = COORD_T_MAX;
static coord_t g_PrevPos = COORD_T_MAX;
static unsigned g_PrevGenIndex = UINT_MAX;
static unsigned g_PrevChrId = UINT_MAX;
static unsigned g_PrevAIIndex1 = UINT_MAX;
static unsigned g_PrevAIIndex2 = UINT_MAX;

void DumpStackStatsTotal()
	{
	Log("\n");
	Progress("Total %u unstacked bases, %u segs",
	  g_UnstackedBaseCount, g_UnstackedSegCount);
	Log("\n");
	}

static void ClearUnstackedSeg()
	{
	g_PrevGenIndex = UINT_MAX;
	g_PrevChrId = UINT_MAX;
	g_PrevAIIndex1 = UINT_MAX;
	g_PrevAIIndex2 = UINT_MAX;
	}

static void LogUnstackedSeg(const RevFile &Rev)
	{
	if (g_PrevGenIndex == UINT_MAX)
		return;

	++g_UnstackedSegCount;
	if (g_MaxUnstackedSegCount > 0 && g_UnstackedSegCount > g_MaxUnstackedSegCount)
		Die("Max unstacked seg count (%u) exceeded", g_MaxUnstackedSegCount);
	Log("\n");
	Log("Unstacked Gen %u Chr %u From %u To %u AI1=%u AI2=%u\n",
	  g_PrevGenIndex,
	  g_PrevChrId,
	  (unsigned) g_SegStart,
	  (unsigned) g_PrevPos,
	  g_PrevAIIndex1,
	  g_PrevAIIndex2);
	Log("\n");
	Log("Aln1:\n");
	AlnInfo AI;
	unsigned AlnIndex = Rev.GetAlnInfo(g_PrevAIIndex1, AI);
	StdStrVec Rows;
	Rev.GetAlnRows(AlnIndex, Rows);
	LogAIRows(AI, Rows, &Rev);

	Log("\n");
	Log("Aln2:\n");
	AlnIndex = Rev.GetAlnInfo(g_PrevAIIndex2, AI);
	Rev.GetAlnRows(AlnIndex, Rows);
	LogAIRows(AI, Rows, &Rev);
	}

static void AddUnstackedBase(const RevFile &Rev, unsigned GenIndex, unsigned ChrId,
  coord_t Pos, unsigned AIIndex1, unsigned AIIndex2)
	{
	++g_UnstackedBaseCount;

	if (GenIndex == g_PrevGenIndex && ChrId == g_PrevChrId &&
	  AIIndex1 == g_PrevAIIndex1 && AIIndex2 == g_PrevAIIndex2 &&
	  (Pos == g_PrevPos + 1 || Pos == g_PrevPos - 1))
		{
		g_PrevPos = Pos;
		return;
		}

	LogUnstackedSeg(Rev);
	g_PrevGenIndex = GenIndex;
	g_PrevChrId = ChrId;
	g_PrevPos = Pos;
	g_SegStart = Pos;
	g_PrevAIIndex1 = AIIndex1;
	g_PrevAIIndex2 = AIIndex2;
	}

void DumpStackStats(const RevFile &Rev, unsigned GenIndex, unsigned ChrId,
  const AIXVec &AIXs)
	{
	ClearUnstackedSeg();

	coord_t ctChrLength = Rev.GetChrLength(GenIndex, ChrId);
	unsigned ChrLength = (unsigned) ctChrLength;
	if ((coord_t) ChrLength != ctChrLength)
		Die("Chr length overflows unsigned");

// For each base, store AIXIndex of AIX or UINT_MAX if none
	IntVec AIXIndexes;
	AIXIndexes.resize(ChrLength, UINT_MAX);

	unsigned UnstackedBaseCount = 0;
	const unsigned AIXCount = (unsigned) AIXs.size();
	for (unsigned AIXIndex = 0; AIXIndex < AIXCount; ++AIXIndex)
		{
		const AlnInfoX &AIX = AIXs[AIXIndex];
		const unsigned SegCount = (unsigned) AIX.AI.GetSegCount();
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = AIX.AI.Segs[SegIndex];
			if (Seg.GenomeIndex != GenIndex || Seg.ChrId != ChrId)
				continue;

			asserta(Seg.lo <= Seg.hi);
			asserta(Seg.hi < ctChrLength);

			for (coord_t Pos = Seg.lo; Pos <= Seg.hi; ++Pos)
				{
				if (AIXIndexes[Pos] == UINT_MAX)
					AIXIndexes[Pos] = AIXIndex;
				else
					{
					unsigned AIXIndex2 = AIXIndexes[Pos];
					const AlnInfoX &AIX2 = AIXs[AIXIndex2];
					AddUnstackedBase(Rev, GenIndex, ChrId, Pos, AIX.AlnInfoIndex,
					  AIX2.AlnInfoIndex);
					}
				}
			}
		}
	LogUnstackedSeg(Rev);

	if (UnstackedBaseCount > 0)
		{
		g_TotalUnstackedBaseCount += UnstackedBaseCount;

		std::string GenomeName = Rev.GetGenomeName(GenIndex);
		std::string ChrName = Rev.GetChrName(GenIndex, ChrId);
		TruncLabel(GenomeName);
		TruncLabel(ChrName);

		std::string Name = GenomeName + "." + ChrName;
		Log("%32.32s  %u alns, %u overlap bases in %u segs (%u total overlaps)\n",
		  Name.c_str(),
		  AIXCount,
		  UnstackedBaseCount,
		  g_UnstackedSegCount,
		  g_TotalUnstackedBaseCount);
		}

	g_UnstackedBaseCount += UnstackedBaseCount;
	}
