#include "evo.h"
#include "rateglobals.h"

static void CheckNoBig(const char *Msg, const vector<pair<gpos_t, double> > &Rates)
	{
	for (vector<pair<gpos_t, double> >::const_iterator p = Rates.begin();
	  p != Rates.end(); ++p)
		{
		const pair<gpos_t, double> &LengthRate = *p;
		gpos_t Length = LengthRate.first;
		double Rate = LengthRate.second;
		if (Length > P_MicroThreshold && Rate != 0)
			Die("Rate for %s, length %"PF" exceeds micro threshold (%"PF")",
			  Msg, Length, P_MicroThreshold);
		}
	}

static void CheckNoMicro(const char *Msg, const vector<pair<gpos_t, double> > &Rates)
	{
	for (vector<pair<gpos_t, double> >::const_iterator p = Rates.begin();
	  p != Rates.end(); ++p)
		{
		const pair<gpos_t, double> &LengthRate = *p;
		gpos_t Length = LengthRate.first;
		double Rate = LengthRate.second;
		if (Length < P_MicroThreshold && Rate != 0)
			Die("Rate for %s, length %"PF" below micro threshold (%"PF")",
			  Msg, Length, P_MicroThreshold);
		}
	}

#define COMPMICRODIST(Event)												\
	ComputeMicroDist(#Event, g_##Event##Rates, g_Micro##Event##Rates,		\
	  g_IntegratedMicro##Event##Rate, g_Micro##Event##Bins);

#define COMPBIGDIST(Event)													\
	ComputeBigDist(#Event, g_##Event##Rates, g_IntegratedBig##Event##Rate,	\
	  g_Big##Event##LHFs, g_Big##Event##Bins);

#define COMPDIST(Event)	COMPMICRODIST(Event) COMPBIGDIST(Event)

void ComputeDists()
	{
	printf("Del");
	COMPDIST(Delete)
	printf(" Ins");
	COMPMICRODIST(Insert)
	printf(" Inv");
	COMPDIST(Invert)
	printf(" Tan");
	COMPDIST(Tandem)
	printf(" Mov");
	COMPDIST(Move)
	printf(" Cpy");
	COMPDIST(Copy)
	printf(" NCCE");
	COMPDIST(NCCE)

	printf(" IMv");
	COMPBIGDIST(InterChrMove)
	printf(" ICy");
	COMPBIGDIST(InterChrCopy)

	CheckNoBig("Insert", g_InsertRates);
	CheckNoMicro("InterChrMove", g_InterChrMoveRates);
	CheckNoMicro("InterChrCopy", g_InterChrCopyRates);
	printf("\n");
	}

static bool InitDists()
	{
	vector<pair<gpos_t, double> > Rates;

#define RATE(Event, Length, Freq)	\
  g_##Event##Rates.push_back(pair<gpos_t, double>(Length, Freq));

#include "eventrates.h"
#include "nccerates.h"

	return true;
	}
static bool InitDistsDone = InitDists();

void ReadEventDist(FILE *f, const string &Event)
	{
	if (Event == "Delete")
		ReadDist(f, g_DeleteRates);
	else if (Event == "Insert")
		ReadDist(f, g_InsertRates);
	else if (Event == "Invert")
		ReadDist(f, g_InvertRates);
	else if (Event == "Tandem")
		ReadDist(f, g_TandemRates);
	else if (Event == "Copy")
		ReadDist(f, g_CopyRates);
	else if (Event == "Move")
		ReadDist(f, g_MoveRates);
	else if (Event == "InterChrCopy")
		ReadDist(f, g_InterChrCopyRates);
	else if (Event == "InterChrMove")
		ReadDist(f, g_InterChrMoveRates);
	else
		Die("ReadEventDist(%s) invalid", Event.c_str());
	}

void ReadNCCECWeights(FILE *f)
	{
	vector<pair<gpos_t, double> > Weights;
	ReadDist(f, Weights);

	memset(P_NCCECWeights, 0, sizeof(P_NCCECWeights));
	const unsigned N = SIZE(Weights);
	for (unsigned i = 0; i < N; ++i)
		{
		const pair<gpos_t, double> CECWeight = Weights[i];
		gpos_t CEC = CECWeight.first;
		double Weight = CECWeight.second;

		if (CEC < CEC_GenericNCCE || CEC > CEC_GenericNCCE+9)
			Die("Invalid NC CEC %u", CEC);

		if (Weight < 0)
			Die("Invalid weight %g for NC CEC %u", Weight, CEC);
		
		unsigned k = CEC - CEC_GenericNCCE;
		if (P_NCCECWeights[k] != 0)
			Die("Weight for NC CEC %u specified twice", CEC);

		P_NCCECWeights[k] = Weight;
		}
	}

void LogNCCECWeights()
	{
	Log("\n");
	Log("NCCECWeights=\n");
	Log("#         CEC      Weight\n");
	Log("#  ==========  ==========\n");
	unsigned N = sizeof(P_NCCECWeights)/sizeof(P_NCCECWeights[0]);
	for (unsigned i = 0; i < N; ++i)
		{
		double Weight = P_NCCECWeights[i];
		Log("   %10"PF", %10.4g;\n", i + CEC_GenericNCCE, Weight);
		}
	Log("//\n");
	}

void ReadNCCELengthDist(FILE *f)
	{
	ReadDist(f, g_NCCERates);
	}

gpos_t GetNCCELength()
	{
	gpos_t GetMicroNCCELength();
	gpos_t GetBigNCCELength();

	bool f = g_IntegratedMicroNCCERate /
	  (g_IntegratedMicroNCCERate + g_IntegratedBigNCCERate);
	bool Micro = RandTrue(f);
	if (Micro)
		return GetMicroNCCELength();
	else
		return GetBigNCCELength();
	}
