#include "evo.h"

#define TRACE	0

/***
Generate a clustered random sequence of integers.

There is a probability distribution D over integers 0..N
is specified.

The next integer to be emitted is selected from a
history-dependent distribution H which is the sum
of D plus a "context distribution" C from the preceding
letters. The weight of recent letters decreases to
zero over a "decay length" L. The fractional contribution
of C to H is F, which controls the degree of clustering
ranging from F=0, giving no context, to F=1 giving
maximal clusting (all integers in the sequence will
be the same).

The probability distribution is specified as a weight
vector for efficiency.
***/

void GenClusteredSeq(const vector<unsigned> &WeightVec, byte MaxInt,
  unsigned DecayLength, double F, ByteVec &Seq, unsigned SeqLength)
	{
	asserta(F >= 0 && F <= 1);

	Seq.clear();
	Seq.reserve(SeqLength);

	byte NextInt = (byte) WeightedRand(WeightVec);

#if	TRACE
	Log("First int %u\n", NextInt);
#endif
	Seq.push_back(NextInt);

	for (unsigned i = 1; i < SeqLength; ++i)
		{
		bool Context = RandTrue(F);
#if	TRACE
		Log("Context = RandTrue(%.1f) = %c\n", F, tof(Context));
#endif
		if (Context)
			{
			asserta(false);
			vector<double> Weights(MaxInt+1, 0);
			for (int j = DecayLength - 1; j >= 0; --j)
				{
				unsigned Int = UINT_MAX;
				int k = i - (DecayLength - j);
				if (k < 0)
					{
					Int = WeightedRand(WeightVec);
#if	TRACE
					Log("k=%d Int=%u\n", k, Int);
#endif
					}
				else
					{
					assert(k < (int) Seq.size());
					Int = Seq[k];
#if	TRACE
					Log("k=%d Int=Seq[%d] = %u\n", k, k, Int);
#endif
					}
				asserta(Int <= MaxInt);
				Weights[Int] += j;
				}
			NextInt = (byte) RandWeightDist(Weights);
			}
		else
			NextInt = (byte) WeightedRand(WeightVec);

#if	TRACE
		Log("NextInt = %u\n", NextInt);
#endif
		Seq.push_back(NextInt);
		}
	}

#if	TEST
const unsigned L = 64;

static void Test(const vector<double> &Weights, double F)
	{
	vector<unsigned> WeightVec;
	unsigned SetWeightVec(const vector<double> &Weights, vector<unsigned> &WeightVec,
	  unsigned Length);
	SetWeightVec(Weights, WeightVec, 1000);
	IntVec Seq;
	GenClusteredSeq(WeightVec, 3, 5, F, Seq, L);

	asserta(SIZE(Seq) == L);
	Log("F=%4.1f ", F);
	for (unsigned i = 0; i < L; ++i)
		Log("%d", Seq[i]);
	Log("\n");
	}

void TestGenClusteredSeq()
	{
	vector<double> Weights;
	Weights.push_back(1);
	Weights.push_back(2);
	Weights.push_back(3);

	//vector<unsigned> WeightVec;
	//unsigned SetWeightVec(const vector<double> &Weights, vector<unsigned> &WeightVec,
	//  unsigned Length);
	//SetWeightVec(Weights, WeightVec, 1000);
	//for (unsigned i = 0 ; i < 100000; ++i)
	//	{
	//	unsigned j = WeightedRand(WeightVec);
	//	asserta(j < 3);
	//	Counts[j]++;
	//	}

	for (int i = 0; i < 100; ++i)
		Test(Weights, 0);

	//Test(Weights, 0.5);
	//Test(Weights, 0.5);
	//Test(Weights, 0.5);
	//Test(Weights, 0.5);
	//Test(Weights, 0.5);
	//Test(Weights, 0.5);
	//Test(Weights, 0.5);
	//Test(Weights, 0.5);

	//Test(Weights, 1);
	//Test(Weights, 1);
	//Test(Weights, 1);
	//Test(Weights, 1);
	//Test(Weights, 1);
	//Test(Weights, 1);
	//Test(Weights, 1);
	//Test(Weights, 1);

	double Total = Counts[0] + Counts[1] + Counts[2];
	Log("Counts %u %u %u\n", Counts[0], Counts[1], Counts[2]);
	Log("Weights %g %g %g\n", Counts[0]/Total, Counts[1]/Total, Counts[2]/Total);
	}
#endif // TEST
