#include "evo.h"

#define TRACE	0

/***
The MUL file I that is input to MicroEvolve() is the output file
from MacroEvolve(). This is created by BigEvolver.ToFile(), with
MULs in extended ancestral order, meaning that novel and duplicated
MULs are written after the last MUL in the "true" ancestor. A linked
list gives the evolved order.

The .evo file F output by MicroEvolve() is written in a modified
ancestral order in which all family members appear consecutively
at the point where the first family member is found in the ancestral
genome, but the order is otherwise ancestral. This is for efficiency:
each entire family is generated in its entirety when it is encountered
in a sequential read of the ancestral sequence. Again, a linked list
of MULs provides the evolved order.

GetOutMULIndex() returns the MUL index in F given the MUL index in I.

Non-deleted leaves in F are assigned MUL indexes 0 .. (L-1) where L
is the number of leaves, non-deleted internal nodes are assigned MUL
indexes L .. L+N-1, where N is the number of internal nodes. Deleted
MULs do not appear in F (including internal nodes that have no non-
deleted leaves).
***/

static IntVec g_InLeafMULIndexToOut;
static IntVec g_InInternalMULIndexToOut;

void LogInOutMULMap()
	{
	unsigned Count = SIZE(g_InLeafMULIndexToOut);
	asserta(Count == SIZE(g_InInternalMULIndexToOut));
	Log("\n");
	Log("Leaf In->Out map:\n");
	Log("%u MULs\n", Count);
	Log("        In        Leaf    Internal\n");
	Log("==========  ==========  ==========\n");
	for (unsigned i = 0; i < Count; ++i)
		{
		Log("%10u", i);
		unsigned Leaf = g_InLeafMULIndexToOut[i];
		unsigned Internal = g_InInternalMULIndexToOut[i];
		if (Leaf == UINT_MAX)
			Log("           *");
		else
			Log("  %10u", Leaf);
		if (Internal == UINT_MAX)
			Log("           *");
		else
			Log("  %10u", Internal);
		Log("\n");
		}
	}

unsigned GetOutMULIndex(unsigned InMULIndex, bool Leaf)
	{
	if (Leaf)
		{
		asserta(InMULIndex < SIZE(g_InLeafMULIndexToOut));
		return g_InLeafMULIndexToOut[InMULIndex];
		}
	else
		{
		asserta(InMULIndex < SIZE(g_InInternalMULIndexToOut));
		return g_InInternalMULIndexToOut[InMULIndex];
		}
	}

unsigned GetInMULIndex(unsigned OutMULIndex)
	{
	const unsigned N = SIZE(g_InLeafMULIndexToOut);
	for (unsigned InMULIndex = 0; InMULIndex < N; ++InMULIndex)
		if (g_InLeafMULIndexToOut[InMULIndex] == OutMULIndex)
			return InMULIndex;
	return UINT_MAX;
	}

void InOutMULMap(const BigEvolver &Big)
	{
	const unsigned MULCount = Big.GetMULCount();
	const unsigned FamilyCount = Big.GetFamilyCount();

	unsigned LeafMULCount = UINT_MAX;
	unsigned InternalMULCount = UINT_MAX;
	Big.GetEvolvedMULCounts(LeafMULCount, InternalMULCount);

	g_InLeafMULIndexToOut.resize(MULCount, UINT_MAX);
	g_InInternalMULIndexToOut.resize(MULCount, UINT_MAX);

	unsigned OutLeafMULIndex = 0;
	unsigned OutInternalMULIndex = LeafMULCount;

#if	TRACE
	Log("\n");
	Log("InOutMULMap\n");
#endif

	const CoordMap &CM = Big.GetCoordMap();
	BoolVec FamilyDone(FamilyCount, false);
	for (CoordMap::const_iterator p = CM.begin(); p != CM.end(); ++p)
		{
		unsigned InMULIndex = p->second;
		const MULData &MUL = Big.GetMUL(InMULIndex);
		if (MUL.IsME() && !MUL.IsFamilyMember())
			continue;
#if	TRACE
		Log("InMULIndex %u", InMULIndex);
#endif
		if (MUL.IsDeleted())
			{
#if	TRACE
			Log(" deleted\n");
#endif
			continue;
			}
		if (MUL.IsFamilyMember())
			{
			unsigned FamilyIndex = MUL.GetFamily();
#if	TRACE
			Log(" family %u", FamilyIndex);
#endif
			asserta(FamilyIndex < FamilyCount);
			if (FamilyDone[FamilyIndex])
				{
#if	TRACE
				Log(" done\n");
#endif
				}
			else
				{
				FamilyDone[FamilyIndex] = true;
				const SmallTree &Tree = Big.GetFamilyTree(FamilyIndex);
				const vector<SmallTreeNode> &Nodes = Tree.Nodes;
				FORC(Node)
					{
					const SmallTreeNode &Node = Nodes[NodeIndex];
					unsigned InMULIndex = Node.GetMULIndex();
#if	TRACE
					Log(" node %u", NodeIndex);
#endif
					if (Node.IsLeaf())
						{
#if	TRACE
						Log(" leaf %u=%u", InMULIndex, OutLeafMULIndex);
#endif
						asserta(InMULIndex < SIZE(g_InLeafMULIndexToOut));
						asserta(g_InLeafMULIndexToOut[InMULIndex] == UINT_MAX);
						g_InLeafMULIndexToOut[InMULIndex] = OutLeafMULIndex++;
						}
					else 
						{
					// It used to be that deleted MULs could appear in family trees.
					// This is no longer allowed.
						//asserta(
						//  Big.GetNonDeletedInternalCount(FamilyIndex, NodeIndex) > 0);
#if	TRACE
						Log(" int %u=%u", InMULIndex, OutInternalMULIndex);
#endif
						asserta(InMULIndex < SIZE(g_InInternalMULIndexToOut));
						asserta(g_InInternalMULIndexToOut[InMULIndex] == UINT_MAX);
						g_InInternalMULIndexToOut[InMULIndex] = OutInternalMULIndex++;
						}
					}
#if	TRACE
				Log("\n");
#endif
				}
			}
		else
			{
#if	TRACE
			Log(" no family, %u=%u\n", InMULIndex, OutLeafMULIndex);
#endif
			asserta(InMULIndex < SIZE(g_InLeafMULIndexToOut));
			asserta(g_InLeafMULIndexToOut[InMULIndex] == UINT_MAX);
			g_InLeafMULIndexToOut[InMULIndex] = OutLeafMULIndex++;
			}
		}
	asserta(OutLeafMULIndex <= LeafMULCount);
	//asserta(OutLeafMULIndex == LeafMULCount);
	//asserta(OutInternalMULIndex == LeafMULCount + InternalMULCount);
	}
