#include "evo.h"

static IntVec g_CurrentInsertCounts;
static IntVec g_CurrentEnhancedInsertCounts;

const MEInfo &BigEvolver::GetME(unsigned MEIndex) const
	{
	asserta(MEIndex < SIZE(m_MEs));
	return m_MEs[MEIndex];
	}

MEInfo &BigEvolver::GetME(unsigned MEIndex)
	{
	asserta(MEIndex < SIZE(m_MEs));
	return m_MEs[MEIndex];
	}

void BigEvolver::AddME(const MEInfo &ME, const RangeVect &EnhancedRanges)
	{
	m_MEs.push_back(ME);
	m_MEEnhancedRanges.push_back(EnhancedRanges);
	}

void BigEvolver::NonEnhancedInsertME(unsigned MEIndex)
	{
	MEInfo &ME = GetME(MEIndex);
// Arbitrarily try five times
	for (unsigned i = 0; i < 5; ++i)
		{
		gpos_t Pos = GetRandomPos();
		bool Accept = AcceptMutation(BIG_MUT_TYPE_Insert, Pos, ME.Length);
		if (Accept)
			{
			InsertME(Pos, MEIndex);
			if (m_ValidateEvents)
				Validate();
			return;
			}
		}
	}

void BigEvolver::EnhancedInsertME(unsigned MEIndex)
	{
	MEInfo &ME = GetME(MEIndex);
	asserta(MEIndex < SIZE(m_MEEnhancedRanges));
	const RangeVect &Ranges = m_MEEnhancedRanges[MEIndex];
	unsigned RangeCount = SIZE(Ranges);
	if (RangeCount == 0)
		return;

// Arbitrarily try five times
	for (unsigned i = 0; i < 5; ++i)
		{
		unsigned RangeIndex = RandInt(RangeCount);
		gpos_t RangeStart = Ranges[RangeIndex].first;
		gpos_t RangeLength = Ranges[RangeIndex].second;
		gpos_t Offset = RandPos(RangeLength);
		gpos_t Pos = RangeStart + Offset;
		if (!PosExists(Pos))
			continue;

		bool Accept = AcceptMutation(BIG_MUT_TYPE_Insert, Pos, ME.Length);
		if (Accept)
			{
			InsertME(Pos, MEIndex);
			if (m_ValidateEvents)
				Validate();
			return;
			}
		}
	}

void BigEvolver::InsertMEs(double Fract)
	{
	if (Fract == 0)
		InitInsertMEs();

	unsigned MECount = SIZE(m_MEs);
	for (unsigned MEIndex = 0; MEIndex < MECount; ++MEIndex)
		{
		const MEInfo &ME = m_MEs[MEIndex];

	// Non-enhanced inserts
		{
		double NewCount = (double) ME.InsertCount*Fract;
		unsigned IntNewCount = (unsigned) NewCount;
		unsigned OldCount = g_CurrentInsertCounts[MEIndex];
		asserta(IntNewCount >= OldCount);
		unsigned InsertCount = IntNewCount - OldCount;
		for (unsigned i = 0; i < InsertCount; ++i)
			NonEnhancedInsertME(MEIndex);
		g_CurrentInsertCounts[MEIndex] += InsertCount;
		}

	// Enhanced inserts
		{
		double NewCount = (double) ME.EnhancedInsertCount*Fract;
		unsigned IntNewCount = (unsigned) NewCount;
		unsigned OldCount = g_CurrentEnhancedInsertCounts[MEIndex];
		asserta(IntNewCount >= OldCount);
		unsigned InsertCount = IntNewCount - OldCount;
		for (unsigned i = 0; i < InsertCount; ++i)
			EnhancedInsertME(MEIndex);
		g_CurrentEnhancedInsertCounts[MEIndex] += InsertCount;
		}
		}
	}

void BigEvolver::InitInsertMEs()
	{
	unsigned MECount = SIZE(m_MEs);
	asserta(SIZE(m_MEEnhancedRanges) == MECount);

	g_CurrentInsertCounts.resize(MECount, 0);
	g_CurrentEnhancedInsertCounts.resize(MECount, 0);
	}

void BigEvolver::AppendMEMULs()
	{
	unsigned MECount = SIZE(m_MEs);
	for (unsigned MEIndex = 0; MEIndex < MECount; ++MEIndex)
		{
		MEInfo &ME = m_MEs[MEIndex];

		unsigned MULCount = GetMULCount();
		unsigned LastMULIndex = (MULCount == 0 ? UINT_MAX : MULCount - 1);
		MULData MUL;
		MUL.Pos = m_MaxPos + 1;
		ME.Pos = MUL.Pos;
		if (MEIndex == 0)
			m_MEStartPos = MUL.Pos;
		MUL.Plus = true;
		MUL.Length = ME.Length;
		MUL.SetMEIndex(MEIndex);
		InsertMUL(MUL, LastMULIndex);
		m_MEEndPos = MUL.GetEndPos();
		}
	}

void BigEvolver::ValidateMEMULs() const
	{
	unsigned MECount = SIZE(m_MEs);
	for (unsigned MEIndex = 0; MEIndex < MECount; ++MEIndex)
		{
		const MEInfo &ME = m_MEs[MEIndex];
		asserta(RangeIsContiguous(ME.Pos, ME.Length));

		IntVec MULIndexes;
		GetMULIndexes(ME.Pos, ME.Length, MULIndexes);
		unsigned MULCount = SIZE(MULIndexes);
		asserta(MULCount > 0);
		unsigned FirstMULIndex = MULIndexes[0];
		unsigned LastMULIndex = MULIndexes[MULCount-1];
		const MULData &FirstMUL = GetMUL(FirstMULIndex);
		const MULData &LastMUL = GetMUL(LastMULIndex);

		asserta(ME.Pos == FirstMUL.Pos);
		asserta(ME.Pos + ME.Length - 1 == LastMUL.GetEndPos());

		for (unsigned i = 0; i < MULCount; ++i)
			{
			unsigned MULIndex = MULIndexes[i];
			const MULData &MUL = GetMUL(MULIndex);
			asserta(MUL.GetMEIndex() == MEIndex);
			}
		}
	}

void BigEvolver::InsertME(gpos_t Pos, unsigned MEIndex)
	{
	const MEInfo &ME = GetME(MEIndex);

// Don't use BigCopy because not quite the same.
	unsigned FirstMULIndex = UINT_MAX;
	unsigned SecondMULIndex = UINT_MAX;
	Split(Pos, &FirstMULIndex, &SecondMULIndex);
	unsigned TargetMULIndex = FirstMULIndex;

	IntVec MULIndexes;
	GetMULIndexes(ME.Pos, ME.Length, MULIndexes);
	const unsigned MULCount = SIZE(MULIndexes);
	asserta(MULCount > 0);

	for (unsigned i = 0; i < MULCount; ++i)
		{
		unsigned MULIndex = MULIndexes[i];
		TargetMULIndex = CopyMUL(MULIndex, TargetMULIndex);
		MULData &NewMUL = GetMUL(TargetMULIndex);
		NewMUL.SetMEIndex(UINT_MAX);
		}

	ValidateCopyMULLayout(ME.Pos, Pos, ME.Length);//@@
	}
