#include "stack.h"

/***
MakeColOffsetVecs constructs vectors that enable O(1) lookup
for column <-> offset conversions given an aligned sequence.

"Offset" is the position 0, 1, 2 ... relative to the start of
the ungapped sequence, where the start is the first letter
in Row if Plus=true or the last letter in Row if Plus=false.
Offsets increase with increasing column if Plus=true or
with decreasing column if Plus=false.

OffsetToCol[i] is the column containing the letter at offset
i in the ungapped sequence.

ColToOffset[i] is the offset of the letter appearing in column
i of the aligned sequence. If the column contains a gap, then
it is the offset of the letter immediately following the gap
(in order of increasing offset) If the gap ends in the last column
(Plus=true) or first column (Plus=false), ColToOffset[i] contains
L, the (unaligned) sequence length, i.e. the last offset + 1.

ColHasGap[i] is true if column i in the aligned sequence contains
a gap, otherwise false.

Example:

	0 1 2 3 4 5 5 5 6 7 		ColToOffset[]
	0 1 2 3 4     7 8  			OffsetToCol[]
	F F F F F T T T F T         ColHasGap[]

    0 1 2 3 4     5 6   		Offset
	0 1 2 3 4 5 6 7 8 9			Column index

    A C T A G - - T G -			L=7
***/

void MakeColOffsetVecs(const StdStr &Row, IntVec &ColToOffset,
  IntVec &OffsetToCol, BoolVec &ColHasGap, bool Plus)
	{
	OffsetToCol.clear();

	const unsigned ColCount = (unsigned) Row.size();
	ColToOffset.resize(ColCount);
	ColHasGap.resize(ColCount);

	unsigned Offset = 0;
	if (Plus)
		{
		for (unsigned ColIndex = 0; ColIndex < ColCount; ++ColIndex)
			{
			char c = Row[ColIndex];
			ColToOffset[ColIndex] = (unsigned) Offset;
			bool Gap = (c == '-');
			ColHasGap[ColIndex] = Gap;
			if (!Gap)
				{
				assert(OffsetToCol.size() == (unsigned) Offset);
				OffsetToCol.push_back(ColIndex);
				++Offset;
				}
			}
		}
	else
		{
		for (int ColIndex = ColCount - 1; ColIndex >= 0; --ColIndex)
			{
			char c = Row[ColIndex];
			ColToOffset[ColIndex] = (unsigned) Offset;
			bool Gap = (c == '-');
			ColHasGap[ColIndex] = Gap;
			if (!Gap)
				{
				assert(OffsetToCol.size() == (unsigned) Offset);
				OffsetToCol.push_back(ColIndex);
				++Offset;
				}
			}
		}
	}
