#include "stack.h"
#include "timing.h"

#define TRACE			0
#define	TRACE_INPUT		0

/***
The Stacker class implements a "stacking" algorithm.

A "stacked" set of alignments (blocks), or stacked blockset
in TBA terminology, is a set of alignments in which a given
base appears no more than once.

An outline of the algorithm is as follows.

1. Create a graph G in which nodes are aligments and edges
indicate that at least one base is found in both alignments.

2. Compute the set S of connected components (CCs) of G.

Note that a given output alignment can include bases from no more
than one CC, so processing one CC at a time is a very useful time
and space optimization.

3. For each connnected component in S:

3a. Create a graph H in which nodes are bases and edges indicate
that the bases are aligned in at least one column in some alignment
in H. An edge carries a sign to indicate whether its bases are
aligned on the same or opposite strands.

3b. Compute the set T of connected components (CCs) of H. Each CC
is a column in an output alignment (block). Create indexes on these
columns so that the column containing a given position (Genome,
Chromosome, Coordinate) can be found in O(log N) time. Note that
there can be only one such column; this guarantees the stacking
property.

3d. If T is empty, done, otherwise choose an arbitrary column
C from T.

3e. Create an empty output alignment A.

3f. Using the indexes, search for a column D that can be a
successor of C in a multiple alignment. If one is found, append
D to A, let C := D and repeat step 3f. Otherwise, continue to 3g.

3g. Using the indexes, search for a column D that can be a
predecessor of C in a multiple alignment. If one is found,
append D to A, let C := D and repeat step 3f. Otherwise,
continue to step 3h.

3h. Output alignment A, go to step 3e.
***/

void Stacker::Clear()
	{
	Bases.clear();
	BaseInfoToIndex.clear();
	Edges.clear();
	EdgePluses.clear();
	OutputCols.clear();
	OutputColBases.clear();

	GCPToColIndex.clear();
	OutputAI.Clear();

	OutputConfScore = 0;
	InputSegCount = 0;

	OutputBlockCount = 0;
	DeletedBigColCount = 0;
	InconsistentColCount = 0;
	ShortColCount = 0;
	}

void Stacker::BuildAndWriteOutputBlocks()
	{
	const unsigned OutputColCount = SIZE(OutputColBases);
	unsigned ColIndex = 0;
	std::vector<BaseWithSign> Col;
	while (SIZE(OutputColBases) > 0)
		{
#if	TRACE
		Log("\n");
		Log("Top of loop\n");
#endif
		OutputCols.clear();
		ColIndex = 0;
		for (;;)
			{
			if (ColIndex == OutputColCount)
				return;
			if (SIZE(OutputColBases[ColIndex]) > 0)
				break;
			++ColIndex;
			}
#if	TRACE
		Log("Non-empty col %u\n", ColIndex);
#endif
		InitOutputBlock(OutputColBases[ColIndex]);
		DeleteOutputCol(ColIndex);

		for (;;)
			{
			bool Invert = false;
			unsigned NextColIndex = GetAdjacentOutputCol(true, Invert);
			if (NextColIndex == UINT_MAX)
				break;
#if	TRACE
			Log("Next %u\n", NextColIndex);
#endif
			AddOutputCol(NextColIndex, Invert, true);
			}

		for (;;)
			{
			bool Invert = false;
			unsigned PrevColIndex = GetAdjacentOutputCol(false, Invert);
			if (PrevColIndex == UINT_MAX)
				break;
#if	TRACE
			Log("Prev %u\n", PrevColIndex);
#endif
			AddOutputCol(PrevColIndex, Invert, false);
			}

#if	TRACE
		Log("Output:\n");
		LogOutputBlock();
#endif
		WriteOutputBlock();
		}
	}

void Stacker::Stack()
	{
#if	TRACE_INPUT
	Log("\n");
	Log("Stacker::Stack():\n");
	LogAIXs(AIXs);
#endif

	ComputeInputSegCount();

	BuildBaseGraph();

	IntVecVec CCs;
	BoolVec Signs;
	GetConnectedComponents(Edges, EdgePluses, CCs, Signs, false);

	BuildOutputColBases(CCs, Signs);

	BuildAndWriteOutputBlocks();

	ValidateNoOutputOverlaps();
	}

void Stacker::ComputeInputSegCount()
	{
	const AIXVec &AIXs = *ptrAIXs;
	const unsigned AICount = SIZE(AIXs);
	InputSegCount = 0;
	FORIX(AI)
		{
		const AlnInfoX &AIX = AIXs[AIIndex];
		InputSegCount += SIZE(AIX.AI.Segs);
		}
	}
