#include "stack.h"

#define TRACE	0

static unsigned g_MaxGapLength = 8;
static unsigned g_MinBlockLength = 8;
static bool g_LogAIIndex = false;
static bool g_LogScores = false;

static unsigned g_OutputBlockCount = 0;
static unsigned g_ShortBlockCount = 0;
static unsigned g_DoAIIndex = UINT_MAX;

static void OutputBlock(const AlnInfo &AI, const StdStrVec &Rows, unsigned AIIndex,
  unsigned OrigColCount, RevFile &Out)
	{
#if	TRACE
	Log("\n");
	Log("OutputBlock:\n");
	LogAIRows(AI, Rows);
	Log("\n");
#endif
	if (OrigColCount == 0)
		return;

	const unsigned SegCount = SIZE(Rows);
	if (SegCount == 0)
		return;

	unsigned ColCount = SIZE(Rows[0]);
	if (ColCount < g_MinBlockLength)
		{
		++g_ShortBlockCount;
		return;
		}

	float Score = AI.Score;
	double ConfScore = AI.GetFloatAttr(ALN_ATTR_ConfScore);
	
	double Ratio = (double) ColCount / (double) OrigColCount;
	float BlockScore = (float) (Ratio*Score);
	double BlockConfScore = (ConfScore == DBL_MAX ? DBL_MAX : Ratio*ConfScore);

	if (g_LogScores)
		{
		static bool HeaderDone = false;
		if (!HeaderDone)
			{
			HeaderDone = true;
			Log("\n");
			Log("AIIndex   Length  BlockLen  Fract%%    Score  BlockScore  ConfScore  BlockConfScore\n");
			Log("=======  =======  ========  ======  =======  ==========  =========  ==============\n");
			//   1234567  1234567  12345678  12345  1234567  1234567890  123456789  12345678901234
			}
		Log("%7u  %7u  %8u  %5.1f%%  %7g  %10g  %9g  %14g\n",
		  AIIndex,
		  OrigColCount,
		  ColCount,
		  Ratio*100,
		  Score,
		  BlockScore,
		  ConfScore,
		  BlockConfScore);
		}

// Make a copy because we must delete rows that are all gaps
	StdStrVec OutRows;
	AlnInfo OutAI(AI);

	OutAI.Score = BlockScore;
	OutAI.SetFloatAttr(ALN_ATTR_ConfScore, BlockConfScore);

	OutAI.Segs.clear();
	const GCSegVec &Segs = AI.Segs;
	FORIX(Seg)
		{
		const StdStr &Row = Rows[SegIndex];

	// Delete rows that are all gaps
		if (GetUngappedSeqLength(Row) == 0)
			continue;

		const GCSeg &Seg = Segs[SegIndex];
		OutAI.Segs.push_back(Seg);
		OutRows.push_back(Row);
		}

	if (SIZE(OutAI.Segs) > 1)
		{
		++g_OutputBlockCount;
		Out.AddAln(OutAI, OutRows);
		}
	}

static void SplitAln(unsigned AIIndex, const AlnInfo &AI, const StdStrVec &Rows, RevFile &Out)
	{
#if	TRACE
	Log("\n");
	Log("SplitAln:\n");
	LogAIRows(AI, Rows);
#endif

	const GCSegVec &Segs = AI.Segs;
	const unsigned SegCount = SIZE(Segs);
	const unsigned RowCount = SIZE(Rows);
	asserta(SegCount == RowCount);
	if (SegCount == 0)
		return;
	const unsigned ColCount = SIZE(Rows[0]);
	if (ColCount <= g_MaxGapLength)
		{
		Out.AddAln(AI, Rows);
		return;
		}

// Ends[i] is true if column i ends an output block,
// typically because it is the last column of a long gap.
	BoolVec Ends;
	Ends.resize(ColCount, false);

// End of input terminates the last block
	Ends[ColCount-1] = true;

// Mark columns that are ends of long blocks
	FORIX(Seg)
		{
		const StdStr &Row = Rows[SegIndex];
		unsigned GapLength = 0;
		unsigned GapStartColIndex = 0;
		FORIX(Col)
			{
			char c = Row[ColIndex];
			if (c == '-')
				{
				if (GapLength == 0)
					GapStartColIndex = ColIndex;
				++GapLength;
				}
			if (c != '-' || ColIndex + 1 == ColCount)
				{
				if (GapLength > g_MaxGapLength)
					{
					if (GapStartColIndex > 0)
						Ends[GapStartColIndex-1] = true;
					Ends[ColIndex-1] = true;
					}
				GapLength = 0;
				}
			}
		}

#if	TRACE
	Log("\n");
	Log("Ends:\n");
	Log("  Col  E\n");
	Log("=====  =\n");
	FORIX(Col)
		{
		Log("%5u  %c\n", ColIndex, Ends[ColIndex] ? '<' : ' ');
		}
#endif

	AlnInfo OutAI(AI);
	GCSegVec &OutSegs = OutAI.Segs;
	FORIX(Seg)
		{
		GCSeg &OutSeg = OutSegs[SegIndex];
		if (OutSeg.Plus)
			OutSeg.hi = COORD_T_MAX;
		else
			OutSeg.lo = COORD_T_MAX;
		}

	StdStrVec OutRows;
	OutRows.resize(RowCount);
	unsigned BlockLength = 0;
	FORIX(Col)
		{
		++BlockLength;
#if	TRACE
		Log("\n");
		Log("Col %u  block length %u:\n", ColIndex, BlockLength);
#endif
		FORIX(Seg)
			{
			const StdStr &Row = Rows[SegIndex];
			char c = Row[ColIndex];

			StdStr &OutRow = OutRows[SegIndex];
			OutRow.push_back(c);

			GCSeg &OutSeg = OutSegs[SegIndex];
			if (c != '-')
				{
				if (OutSeg.Plus)
					{
					if (OutSeg.hi == COORD_T_MAX)
						OutSeg.hi = OutSeg.lo;
					else
						++OutSeg.hi;
					}
				else
					{
					if (OutSeg.lo == COORD_T_MAX)
						OutSeg.lo = OutSeg.hi;
					else
						--OutSeg.lo;
					}
#if	_DEBUG
				if (OutSeg.hi != COORD_T_MAX && OutSeg.lo != COORD_T_MAX)
					ValidateSegLength(OutSeg, OutRow);
#endif
				}
#if	TRACE
			{
			Log("Seg %u ", SegIndex);
			LogGCSeg(OutSeg);
			Log("  Length=%u", (unsigned) OutSeg.GetLength());
			Log("  '%c'", c);
			Log("  %s\n", OutRow.c_str());
			Log(" UngappedLength=%u", GetUngappedSeqLength(OutRow));
			Log("\n");
			}
#endif
			}

		if (Ends[ColIndex])
			{
#if	TRACE
			Log(">> END BLOCK, length %u\n", BlockLength);
#endif
			if (BlockLength > 0)
				{
				OutputBlock(OutAI, OutRows, AIIndex, ColCount, Out);
				FORIX(Seg)
					{
					OutRows[SegIndex].clear();
					}
				}
			BlockLength = 0;
			FORIX(Seg)
				{
				GCSeg &OutSeg = OutSegs[SegIndex];
				if (OutSeg.Plus)
					{
					if (OutSeg.hi != COORD_T_MAX)
						{
						OutSeg.lo = OutSeg.hi + 1;
						OutSeg.hi = COORD_T_MAX;
						}
					}
				else
					{
					if (OutSeg.lo != COORD_T_MAX)
						{
						OutSeg.hi = OutSeg.lo - 1;
						OutSeg.lo = COORD_T_MAX;
						}
					}
				}
			}
		}
	}

void StackSplit()
	{
	const char *InFileName = RequiredValueOpt("split");
	const char *OutFileName = RequiredValueOpt("out");
	IntOpt("aiindex", &g_DoAIIndex);
	g_LogAIIndex = FlagOpt("logaiindex");
	g_LogScores = FlagOpt("logscores");
	IntOpt("minblocklength", &g_MinBlockLength);
	IntOpt("maxgaplength", &g_MaxGapLength);
	if (g_MaxGapLength < 2)
		Die("max gap length must be >= 2");

	RevFile In;
	RevFile Out;

	In.Open(InFileName);
	const GenInfoVec &Genomes = In.GetGenomes();
	Out.Create(OutFileName, Genomes);

	if (g_DoAIIndex != UINT_MAX)
		{
		AlnInfo AI;
		In.GetAlnInfo(g_DoAIIndex, AI);

		unsigned AlnIndex = In.GetAlnIndex(g_DoAIIndex);

		StdStrVec Rows;
		In.GetAlnRows(AlnIndex, Rows);

		SplitAln(g_DoAIIndex, AI, Rows, Out);

		In.Close();
		Out.Close();
		return;
		}

	AIVec AIs;
	In.GetAllAlnInfos(AIs);
	const unsigned AICount = SIZE(AIs);
	Progress("%10 AIs", AICount);

	ProgressStart("Splitting");
	FORIX(AI)
		{
		if (g_LogAIIndex)
			Log("AI %u", AIIndex);
		if (AIIndex%100 == 0)
			ProgressStep(AIIndex, AICount);
		const AlnInfo &AI = AIs[AIIndex];
		unsigned AlnIndex = In.GetAlnIndex(AIIndex);
		if (AlnIndex == UINT_MAX)
			{
			Out.AddAlnInfo(AI);
			continue;
			}
		StdStrVec Rows;
		In.GetAlnRows(AlnIndex, Rows);

		SplitAln(AIIndex, AI, Rows, Out);
		if (g_LogAIIndex)
			Log(" (%u)\n", g_OutputBlockCount);
		}
	ProgressDone();

	ProgressNoRes("%10u  input blocks", AICount);
	ProgressNoRes("%10u  output blocks", g_OutputBlockCount);
	ProgressNoRes("%10u  short blocks", g_ShortBlockCount);

	In.Close();
	Out.Close();
	}
