#include <vector>
#include <set>
#include <map>
#include <string>
#include <iterator>
#include <algorithm>
#include "revfile.h"
#include "intlist.h"
#include "types.h"
#include "utils.h"
#include "getopts.h"
#include "progress.h"

/* The input files */
RevFile InFiles[2];

/* "Reference" and "Other" are arbitrary terms; the process is symmetric.
 * One genome is picked arbitrarily and the alignments are sorted according
 * to the coordinates of that genome; that is the 'reference genome'
 */

/* RefGenome[i] is the index of the 'reference genome' in the i-th revfile */
int RefGenome[2];
/* OtherGenome[i] is the index of the 'other genome' in the i-th revfile */
int OtherGenome[2];

/* Progress... */
unsigned TotalAlnCount;
unsigned CurrentAlnCount = 0;

/* End of important globals */

VALUE_OPT ValueOpts[] =
{
	{ "rev1",				0 },
	{ "rev2",				0 },
	{ "log",				0 }
};
int ValueOptCount = sizeof(ValueOpts)/sizeof(ValueOpts[0]);

FLAG_OPT FlagOpts[] =
{
	{ "dummy",				0 }
};
int FlagOptCount = sizeof(FlagOpts)/sizeof(FlagOpts[0]);

void Usage()
{
	fprintf(stderr,
"Usage:\n"
"    venn -rev1 <revfile> -rev2 <revfile>\n"
"\n"
	);
}

class ColumnInfo
{
	public:
	
	unsigned RefChrId;
	coord_t RefCoord;
	unsigned OtherChrId;
	coord_t OtherCoord;
	bool RefIsGap;
	bool OtherIsGap;

	bool operator<(const ColumnInfo &other) const
	{
		if (RefChrId != other.RefChrId)
			return RefChrId < other.RefChrId;
		else if (RefCoord != other.RefCoord)
			return RefCoord < other.RefCoord;
		else if (OtherChrId != other.OtherChrId)
			return OtherChrId < other.OtherChrId;
		else if (OtherCoord != other.OtherCoord)
			return (OtherCoord < other.OtherCoord);
		else if (RefIsGap != other.RefIsGap)
			return ((!RefIsGap) && (other.RefIsGap));
		else 
			return ((!OtherIsGap) && (other.OtherIsGap));
	}

	bool operator==(const ColumnInfo &other) const
	{
		return (RefChrId == other.RefChrId)
			&& (RefCoord == other.RefCoord)
			&& (OtherChrId == other.OtherChrId)
			&& (OtherCoord == other.OtherCoord)
			&& (RefIsGap == other.RefIsGap)
			&& (OtherIsGap == other.OtherIsGap);
	}

	bool operator!=(const ColumnInfo &other) const
	{
		return (RefChrId != other.RefChrId)
			|| (RefCoord != other.RefCoord)
			|| (OtherChrId != other.OtherChrId)
			|| (OtherCoord != other.OtherCoord)
			|| (RefIsGap != other.RefIsGap)
			|| (OtherIsGap != other.OtherIsGap);
	}
};

class Stats
{
	public:
	
	unsigned TotalCols;
	unsigned TotalClasses;
	unsigned SingleClasses;

	Stats() : TotalCols(0), TotalClasses(0), SingleClasses(0) {}
};

Stats Unique[2];
Stats Common[2];

void Consume(std::multiset<ColumnInfo>::iterator &iter, const std::multiset<ColumnInfo>::iterator &end_iter, Stats &S)
{
	std::multiset<ColumnInfo>::iterator original_iter = iter;
	++(S.TotalCols);
	++(S.TotalClasses);
	++iter;
	if ((iter == end_iter) || ((*iter) != (*original_iter)))
		++(S.SingleClasses);
	else
		while ((iter != end_iter) && ((*iter) == (*original_iter)))
		{
			++(S.TotalCols);
			++iter;
		}
}

void Venn()
{
	if ((InFiles[0].GetGenomeCount() != 2) || (InFiles[1].GetGenomeCount() != 2))
		Die("This program can only handle two genomes at a time.");

	/* Verify compatibility of the two RevFiles */
	if ((InFiles[0].GetGenomeInfo(0) == InFiles[1].GetGenomeInfo(0))
	    && (InFiles[0].GetGenomeInfo(1) == InFiles[1].GetGenomeInfo(1)))
	{
		RefGenome[0] = 0;
		RefGenome[1] = 0;
		OtherGenome[0] = 1;
		OtherGenome[1] = 1;
	}
	else if ((InFiles[0].GetGenomeInfo(0) == InFiles[1].GetGenomeInfo(1))
	    && (InFiles[0].GetGenomeInfo(1) == InFiles[1].GetGenomeInfo(0)))
	{
		RefGenome[0] = 0;
		RefGenome[1] = 1;
		OtherGenome[0] = 1;
		OtherGenome[1] = 0;
	}
	else
		Die("The genomes do not match.");

	TotalAlnCount = InFiles[0].GetAlnCount() + InFiles[1].GetAlnCount();

	std::multiset<ColumnInfo> Columns[2];

	ProgressStart("Processing...");
	/* For each input file */
	for (unsigned i = 0; i != 2; ++i)
	{
		/* For each alignment */
		unsigned FileAlnCount = InFiles[i].GetAlnInfoCount();
		for (unsigned j = 0; j != FileAlnCount; ++j)
		{
			AlnInfo AI;
			unsigned AlnIndex = InFiles[i].GetAlnInfo(j, AI);
			StdStrVec Rows;
			InFiles[i].GetAlnRows(AlnIndex, Rows);

			std::string RefSeq = Rows[RefGenome[i]];
			std::string OtherSeq = Rows[OtherGenome[i]];
			if (!(AI.Segs[RefGenome[i]].Plus))
			{
				std::reverse(RefSeq.begin(), RefSeq.end());
				std::reverse(OtherSeq.begin(), OtherSeq.end());
			}

			/* Initialize a column */
			ColumnInfo CI;
			CI.RefChrId = AI.Segs[RefGenome[i]].ChrId;
			CI.RefCoord = AI.Segs[RefGenome[i]].lo;
			CI.OtherChrId = AI.Segs[OtherGenome[i]].ChrId;

			bool OtherPlus = ((AI.Segs[RefGenome[i]].Plus) == (AI.Segs[OtherGenome[i]].Plus));
			if (OtherPlus)
				CI.OtherCoord = AI.Segs[OtherGenome[i]].lo;
			else
			{
				CI.OtherCoord = AI.Segs[OtherGenome[i]].hi;
				if (OtherSeq[0] == '-')
					++(CI.OtherCoord);
			}

			/* Go through the alignment and add columns to the sets */
			unsigned k = 0;
			for(;;)
			{
				CI.RefIsGap = (RefSeq[k] == '-');
				CI.OtherIsGap = (OtherSeq[k] == '-');
				Columns[i].insert(CI);
				++k;
				if (k == RefSeq.size())
					break;
				if (!CI.RefIsGap)
					++(CI.RefCoord);
				if (OtherPlus)
				{
					if (!CI.OtherIsGap)
						++(CI.OtherCoord);
				}
				else
				{
					if (OtherSeq[k] != '-')
						--(CI.OtherCoord);
				}
			}

			++CurrentAlnCount;
			if (!(CurrentAlnCount & 0xfff))
				ProgressStep(CurrentAlnCount, TotalAlnCount);
		}
	}
	ProgressDone();


	/* Intersect the sets */
	Progress("Intersecting...");
	std::multiset<ColumnInfo>::iterator iter[2] = {Columns[0].begin(), Columns[1].begin()};

	while ((iter[0] != Columns[0].end()) && (iter[1] != Columns[1].end()))
	{
		if ((*(iter[0])) == (*(iter[1])))
		{
			Consume(iter[0], Columns[0].end(), Common[0]);
			Consume(iter[1], Columns[1].end(), Common[1]);
		}
		else if ((*(iter[0])) < (*(iter[1])))
			Consume(iter[0], Columns[0].end(), Unique[0]);
		else
			Consume(iter[1], Columns[1].end(), Unique[1]);
	}
	while (iter[0] != Columns[0].end())
		Consume(iter[0], Columns[0].end(), Unique[0]);
	while (iter[1] != Columns[1].end())
		Consume(iter[1], Columns[1].end(), Unique[1]);
}

int main(int argc, char *argv[])
{
	ProcessArgVect(argc, argv);
	const char *RevFileName[2];

	RevFileName[0] = RequiredValueOpt("rev1");
	RevFileName[1] = RequiredValueOpt("rev2");

	for (int i = 0; i != 2; ++i)
		InFiles[i].Open(RevFileName[i]);

	Venn();

	for (int i = 0; i != 2; ++i)
		printf("File #%d contains %u unique columns in %u equiv. classes (%u of which are non-singles)\n", i, Unique[i].TotalCols, Unique[i].TotalClasses, Unique[i].TotalClasses - Unique[i].SingleClasses);
	for (int i = 0; i != 2; ++i)
		printf("File #%d contains %u common columns in %u equiv. classes (%u of which are non-singles)\n", i, Common[i].TotalCols, Common[i].TotalClasses, Common[i].TotalClasses - Common[i].SingleClasses);

	return 0;
}
