#include "revfile.h"
#include "getopts.h"
#include "progress.h"

static unsigned g_AncestralCount;
static unsigned g_CopiedCount;

static bool IsGapChar(char c)
	{
	return c == '-';
	}

static bool IsParalog(const AlnInfo &AI)
	{
	unsigned a = AI.GetAttr(ALN_ATTR_Paralog);
	return a == 1;
	}

static double Pcts(double i, double x)
	{
	if (x == 0)
		return 0;
	return 100.0*i/x;
	}

static void GetCounts(const StdStr &Row1, const StdStr &Row2,
  unsigned &BaseCount1, unsigned &BaseCount2, unsigned &MatchCount)
	{
	BaseCount1 = 0;
	BaseCount2 = 0;
	MatchCount = 0;

	const unsigned L = SIZE(Row1);
	asserta(SIZE(Row2) == L);
	for (unsigned i = 0; i < L; ++i)
		{
		bool Gap1 = IsGapChar(Row1[i]);
		bool Gap2 = IsGapChar(Row2[i]);

		if (!Gap1)
			++BaseCount1;
		if (!Gap2)
			++BaseCount2;
		if (!Gap1 && !Gap2)
			++MatchCount;
		}
	}

void AncStats()
	{
	const char *FileName = RequiredValueOpt("ancstats");

	unsigned AncGenomeIndex = 0;
	unsigned EvGenomeIndex = 1;
	IntOpt("ancgix", &AncGenomeIndex);
	IntOpt("evgix", &EvGenomeIndex);

	unsigned TotalBaseCount1 = 0;
	unsigned TotalBaseCount2 = 0;
	unsigned TotalParaBaseCount1 = 0;
	unsigned TotalParaBaseCount2 = 0;

	RevFile Rev;
	Rev.Open(FileName);

	if (Rev.GetGenomeCount() != 2)
		Warning("Rev file not pairwise");

	const GenInfo &AncGen = Rev.GetGenomeInfo(AncGenomeIndex);
	const GenInfo &EvGen = Rev.GetGenomeInfo(EvGenomeIndex);

	unsigned AncLength = AncGen.GetLength();
	unsigned EvLength = EvGen.GetLength();

	unsigned NonPairwiseAlns = 0;
	const unsigned AICount = Rev.GetAlnInfoCount();
	ProgressStart("Processing");
	unsigned GenomesNotFoundCount = 0;
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%100 == 0)
			ProgressStep(AIIndex, AICount);

		AlnInfo AI;
		unsigned AlnIndex = Rev.GetAlnInfo(AIIndex, AI);
		if (AI.GetSegCount() != 2)
			{
			++NonPairwiseAlns;
			continue;
			}

		if (AlnIndex == UINT_MAX)
			continue;

		StdStrVec Rows;
		Rev.GetAlnRows(AlnIndex, Rows);
		asserta(SIZE(Rows) == 2);

		unsigned AncRowIndex = UINT_MAX;
		unsigned EvRowIndex = UINT_MAX;
		for (unsigned i = 0; i < AI.GetSegCount(); ++i)
			{
			const GCSeg &Seg = AI.Segs[i];
			if (Seg.GenomeIndex == AncGenomeIndex)
				AncRowIndex = i;
			else if (Seg.GenomeIndex == EvGenomeIndex)
				EvRowIndex = i;
			}
		if (AncRowIndex == UINT_MAX || EvRowIndex == UINT_MAX)
			{
			++GenomesNotFoundCount;
			continue;
			}

		unsigned BaseCount1;
		unsigned BaseCount2;
		unsigned MatchCount;
		GetCounts(Rows[AncRowIndex], Rows[EvRowIndex], BaseCount1, BaseCount2, MatchCount);

		if (IsParalog(AI))
			{
			g_CopiedCount += MatchCount;
			TotalParaBaseCount1 += BaseCount1;
			TotalParaBaseCount2 += BaseCount2;
			}
		else
			{
			g_AncestralCount += MatchCount;
			TotalBaseCount1 += BaseCount1;
			TotalBaseCount2 += BaseCount2;
			}
		}
	ProgressDone();

	if (NonPairwiseAlns > 0)
		Warning("%u non-pairwise alignments ignored", NonPairwiseAlns);
	if (GenomesNotFoundCount > 0)
		Warning("Anc and/or ev genome not found in %u alignments",
		  GenomesNotFoundCount);

	if (g_AncestralCount > AncLength)
		Warning("%u ancestral bases in evolved genome > ancestral length %u",
		  g_AncestralCount, AncLength);

	unsigned NovelCount = EvLength - g_AncestralCount - g_CopiedCount;
	unsigned DeletedCount = AncLength - g_AncestralCount;

	Log("\n");
	Log("         Feature   Ancestral     Evolved        Diff      Pct\n");
	Log("================  ==========  ==========  ==========  =======\n");
	Log("            Name  %10.10s  %10.10s\n",
	  Rev.GetGenomeName(AncGenomeIndex).c_str(),
	  Rev.GetGenomeName(EvGenomeIndex).c_str());

	double Diff = (double) EvLength - (double) AncLength;
	Log("   Genome length  %10u  %10u  %+10.0f  %6.1f%% (evolved genome length as %% of ancestral length)\n",
	  AncLength, EvLength, Diff, Pcts(EvLength, AncLength));

	Log("\n");
	Log("Aligned non-para  %10u  %10u                      (number of bases in non-paralogous alignments)\n",
	  TotalBaseCount1, TotalBaseCount2);
	Log("\n");
	Log("    Aligned para  %10u  %10u                      (number of bases in paralogous alignments)\n",
	  TotalParaBaseCount1, TotalParaBaseCount2);

	Log("\n");
	Diff = (double) g_AncestralCount - (double) AncLength;
	Log(" Ancestral bases  %10u  %10u  %+10.0f  %6.1f%% (%% of evolved genome that is ancestral)\n",
	  AncLength, g_AncestralCount, Diff, Pcts(g_AncestralCount, EvLength));

	Log("    Copied bases  %10u  %10u              %6.1f%% (%% of evolved genome that is paralogous)\n",
	  0, g_CopiedCount, Pcts(g_CopiedCount, EvLength));

	Log("     Novel bases  %10u  %10u              %6.1f%% (%% of evolved genome that is novel)\n",
	  0, NovelCount, Pcts(NovelCount, EvLength));

	Log("   Deleted bases  %10u  %10u              %6.1f%% (%% of ancestral genome that was deleted)\n",
	  0, DeletedCount, Pcts(DeletedCount, AncLength));

	Rev.Close();
	}
