#include "gff2.h"
#include "getopts.h"
#include "lhist.h"
#include "base.h"
#include "progress.h"
#include "revfile.h"
#include <set>
#include <map>

const unsigned CETypeTandem = CETypeCount + 1;
const unsigned CETypeIsland = CETypeCount + 2;

using namespace std;

CE_TYPE FeatureToCEType(const string &Feature);

static void UpdateIntronDist(const set<GFFRecord2> &Exons, LHistData &IntronDist)
	{
	const GFFRecord2 *LastRec = 0;
	for (set<GFFRecord2>::const_iterator p = Exons.begin(); p != Exons.end(); ++p)
		{
		const GFFRecord2 *Rec = &(*p);
		if (LastRec != 0)
			{
			if (LastRec->SeqName == Rec->SeqName && Rec->Start <= LastRec->End)
				{
				Log("\n");
				Log("LastRec=");
				LastRec->LogMe();
				Log("\n");
				Log("Rec=");
				Rec->LogMe();
				Log("\n");
				Die("Exon records overlap");
				}
			unsigned IntronLength = Rec->Start - LastRec->End - 1;
			if (IntronLength > 0)
				IntronDist.Add(IntronLength);
			}
		LastRec = Rec;
		}
	}

static void UpdateGeneMap(
  map<pair<string, unsigned>, pair<unsigned, unsigned> >
    &ChrNameGeneIndexToGeneLoHi, const string &ChrName,
	unsigned GeneIndex, unsigned lo, unsigned hi)
	{
	pair<string, unsigned> NameIndex(ChrName, GeneIndex);
	map<pair<string, unsigned>, pair<unsigned, unsigned> >::const_iterator p =
	  ChrNameGeneIndexToGeneLoHi.find(NameIndex);
	if (p == ChrNameGeneIndexToGeneLoHi.end())
		ChrNameGeneIndexToGeneLoHi[NameIndex] = pair<unsigned, unsigned>(lo, hi);
	else
		{
		pair<unsigned, unsigned> LoHi = p->second;
		LoHi.first = min(LoHi.first, lo);
		LoHi.second = max(LoHi.second, hi);
		ChrNameGeneIndexToGeneLoHi[NameIndex] = LoHi;
		}
	}

void AnnotStats()
	{
	const char *FileName = RequiredValueOpt("annotstats");
	const char *RevFileName = ValueOpt("seq");
	const char *CSVFileName = ValueOpt("out");
	unsigned GenomeIndex = 0;

	FILE *f = OpenStdioFile(FileName);
	FILE *fCSV = 0;
	if (CSVFileName != 0)
		fCSV = CreateStdioFile(CSVFileName);

	vector<LHistData> LengthDists(CETypeCount);
	for (unsigned i = 0; i < CETypeCount; ++i)
		{
		string Name = CETypeToStr((CE_TYPE) i);
		if (CSVFileName != 0)
			LengthDists[i].Init("Length dist. " + Name);
		else
			LengthDists[i].InitCSV("Length dist. " + Name, 50, 500);
		}

	LHistData IntronDist;
	LHistData IslandDist;
	LHistData TandemDist;
	LHistData GeneDist;

	IntronDist.Init("Length dist. introns");
	IslandDist.Init("Length dist. CpG islands");
	TandemDist.Init("Tandem arrays");
	GeneDist.Init("Length dist. genes");

	set<string> Labels;
	set<string> InvalidFeatures;
	map<pair<string, unsigned>, pair<unsigned, unsigned> >
	  ChrNameGeneIndexToGeneLoHi;
	map<unsigned, set<GFFRecord2> > Exons;

	GFFRecord2 Rec;
	ProgressStartFile(f, "Reading");
	unsigned Count = 0;
	while (GetNextGFFRecord(f, Rec))
		{
		if (Count++%100 == 0)
			ProgressStepFile();

		unsigned GeneIndex = UINT_MAX;
		bool Found = GFFGetIntAttr(Rec.Attrs, "gene_index", GeneIndex);
		if (Found)
			UpdateGeneMap(ChrNameGeneIndexToGeneLoHi,
			  Rec.SeqName, GeneIndex, Rec.Start, Rec.End);

		Labels.insert(Rec.SeqName);
		if (Rec.Feature == "island")
			{
			IslandDist.Add(Rec.GetLength());
			continue;
			}
		else if (Rec.Feature == "tandem")
			{
			TandemDist.Add(Rec.GetLength());
			continue;
			}

		CE_TYPE CEType = FeatureToCEType(Rec.Feature);
		if (CEType == CE_Invalid)
			InvalidFeatures.insert(Rec.Feature);

		if (Rec.Feature == "CDS" || Rec.Feature == "UTR")
			{
			unsigned GeneIndex = GFFGetRequiredIntAttr(Rec.Attrs, "gene_index");
			map<unsigned, set<GFFRecord2> >::iterator p = Exons.find(GeneIndex);
			if (p == Exons.end())
				{
				set<GFFRecord2> Set;
				Set.insert(Rec);
				Exons[GeneIndex] = Set;
				}
			else
				p->second.insert(Rec);
			}

		asserta((unsigned) CEType < CETypeCount);
		LengthDists[CEType].Add(Rec.GetLength());
		}
	ProgressDone();

	Progress("Compute intron distribution");
	for (map<unsigned, set<GFFRecord2> >::const_iterator p = Exons.begin();
	  p != Exons.end(); ++p)
		UpdateIntronDist(p->second, IntronDist);

	Progress("Compute gene distribution");
	for (map<pair<string, unsigned>, pair<unsigned, unsigned> >::const_iterator p =
	  ChrNameGeneIndexToGeneLoHi.begin(); p != ChrNameGeneIndexToGeneLoHi.end(); ++p)
		{
		const pair<unsigned, unsigned> &LoHi = p->second;
		asserta(LoHi.second >= LoHi.first);
		unsigned Length = LoHi.second - LoHi.first + 1;
		GeneDist.Add(Length);
		}

	if (!InvalidFeatures.empty())
		{
		string s;
		for (set<string>::const_iterator p = InvalidFeatures.begin();
		  p != InvalidFeatures.end(); ++p)
			s += " " + *p;
		Log("\n");
		Warning("Skipped these features in GFF file: %s", s.c_str());
		}

	for (unsigned i = 0; i < CETypeCount; ++i)
		{
		const LHistData &LH = LengthDists[i];
		if (LH.GetCount() > 0)
			{
			LH.LogHist(false);
			if (fCSV != 0)
				LH.OutCSV(f);
			}
		}
	IntronDist.LogHist();
	TandemDist.LogHist();
	IslandDist.LogHist();
	GeneDist.LogHist();

	CloseStdioFile(f);

	if (RevFileName == 0)
		return;

	RevFile Rev;
	Rev.Open(RevFileName);

	Log("\n");
	Log("    Length  Chr name\n");
	Log("----------  --------\n");
	unsigned SeqLength = 0;
	for (set<string>::const_iterator p = Labels.begin(); p != Labels.end(); ++p)
		{
		const string &Label = *p;
		unsigned ChrId = Rev.GetChrId(GenomeIndex, Label);
		unsigned ChrLength = Rev.GetChrLength(GenomeIndex, ChrId);
		SeqLength += ChrLength;
		Log("%10u  %s\n", ChrLength, Label.c_str());
		}
	Log("==========\n");
	Log("%10u\n", SeqLength);
	Log("\n");

	unsigned CDSTotal = (unsigned) LengthDists[CE_CDS].GetSum();
	unsigned UTRTotal = (unsigned) LengthDists[CE_UTR].GetSum();
	unsigned NXETotal = (unsigned) LengthDists[CE_NXE].GetSum();
	unsigned NGETotal = (unsigned) LengthDists[CE_NGE].GetSum();
	unsigned GeneTotal = (unsigned) GeneDist.GetSum();
	unsigned IslandTotal = (unsigned) IslandDist.GetSum();
	unsigned TandemTotal = (unsigned) TandemDist.GetSum();
	unsigned IntronTotal = (unsigned) IntronDist.GetSum();
	unsigned GeneConstrainedTotal = CDSTotal + UTRTotal + NXETotal;
	unsigned GeneUnconstrainedTotal = GeneTotal - GeneConstrainedTotal;
	unsigned ConstrainedTotal = CDSTotal + UTRTotal + NXETotal + NGETotal;
	unsigned NeutralTotal = SeqLength - ConstrainedTotal;

	Log("                Type       Bases     Pct\n");
	Log("--------------------  ----------  ------\n");

#define C(x) Log("%20.20s  %10u  %5.1f%%\n", #x, x##Total, Pct(x##Total, SeqLength));
	C(CDS)
	C(Intron)
	C(UTR)
	C(NXE)
	Log("\n");
	Log("%20.20s  %10u  %5.1f%%  (=UTR+CDS+NXE)\n", "Gene constrained",
	  GeneConstrainedTotal, Pct(GeneConstrainedTotal, SeqLength));
	Log("%20.20s  %10u  %5.1f%%\n", "Gene unconstrained",
	  GeneUnconstrainedTotal, Pct(GeneUnconstrainedTotal, SeqLength));
	Log("%20.20s  %10u  %5.1f%%\n", "Gene total", GeneTotal, Pct(GeneTotal, SeqLength));
	Log("\n");
	C(Island)
	C(Tandem)
	Log("\n");
	C(NGE)
	Log("\n");
	Log("%20.20s  %10u  %5.1f%%  (=CDS+UTR+NXE+NGE)\n", "Constrained",
	  ConstrainedTotal, Pct(ConstrainedTotal, SeqLength));
	Log("%20.20s  %10u  %5.1f%%\n", "Neutral", NeutralTotal, Pct(NeutralTotal, SeqLength));
	Log("%20.20s  %10u  %5.1f%%\n", "All", SeqLength, 100.0);
#undef C

	if (fCSV != 0)
		CloseStdioFile(fCSV);
	}
