#include "gff2.h"
#include "progress.h"
#include "myrand.h"
#include "getopts.h"
#include "evomodel.h"
#include "pdf.h"
#include "base.h"

double AccelProb(double p, double f);

void SetModel(const char *FileName);
void GetRandomProbs(unsigned N, double Mean, vector<double> &Probs);
void Clump(const vector<double> &Probs, const vector<unsigned> &ClumpLengths,
  vector<vector<double> > &Clumps, bool RandomizeClumpOrder = true);
double GetSum(const vector<double> &v);

static map<unsigned, vector<GFFRecord2> > g_GeneIndexToCEs;
static vector<GFFRecord2> g_OutRecs;

static RandDist g_NonGeneCEMeanRD;
static RandDist g_CDSMeanRD;

static unsigned g_GeneCount;
static double g_SumGeneMean;
static unsigned g_CDSCount;
static double g_SumCDSMean;
static unsigned g_GeneNCCECount;
static double g_SumGeneNCCEMean;
static unsigned g_NonGeneCECount;
static double g_SumNonGeneCEMean;

static double g_SumCurrentGeneBaseProbs;
static unsigned g_CurrentGeneLength;

uint8 ProbToProbIndex(double Prob)
	{
	unsigned i = (uint8) (Prob*255.0 + 0.5);
	if (i > 255)
		i = 255;
	return (uint8) i;
	}

uint8 GetRandProbIndex(double MeanAccProb)
	{
	asserta(MeanAccProb >= 0 && MeanAccProb <= 1.0);

	RandDist RD;
	GetTruncNormRD(MeanAccProb, RD);
	double Prob = RD.Get();
	uint8 Ix = ProbToProbIndex(Prob);
	if (Ix == NEUTRAL_PROBIX)
		Ix = MAX_NON_NEUTRAL_PROBIX;
	return Ix;
	}

static void AddToOutput(const vector<GFFRecord2> &Recs)
	{
	const unsigned N = SIZE(Recs);
	for (unsigned i = 0; i < N; ++i)
		{
		const GFFRecord2 &Rec = Recs[i];
		g_OutRecs.push_back(Rec);
		}
	}

static double SetProbsAttr(GFFRecord2 &Rec, const vector<double> &Probs)
	{
	GFFDeleteAttr(Rec.Attrs, "probs");
	GFFDeleteAttr(Rec.Attrs, "avg_prob");

	string ProbAttr;
	const unsigned N = SIZE(Probs);
	unsigned Length = (unsigned) Rec.GetLength();
	if (N == 0 || N != Length)
		Die("SetProbsAttr rec len %u, %u probs", Length, N);

	double Sum = 0.0;
	ProbAttr.reserve(2*N);
	for (unsigned i = 0; i < N; ++i)
		{
		double Prob = Probs[i];
		uint8 ProbIndex = ProbToProbIndex(Prob);
		if (ProbIndex == NEUTRAL_PROBIX)
			ProbIndex = MAX_NON_NEUTRAL_PROBIX;
		double Prob2 = g_AcceptProbs[ProbIndex];
		Sum += Prob2;
		char s[8];
		sprintf(s, "%02x", ProbIndex);
		ProbAttr += s;
		}

	double Avg = Sum/N;
	char sAvg[16];
	sprintf(sAvg, "%6.4f", Avg);

	GFFAppendStrAttr(Rec.Attrs, "avgprob", sAvg);
	GFFAppendStrAttr(Rec.Attrs, "probs", ProbAttr);

	return Sum;
	}

static double GetNonGeneCEMean()
	{
	return g_NonGeneCEMeanRD.Get();
	}

static double GetCDSMean()
	{
	return g_CDSMeanRD.Get();
	}

static double GetGeneNonCodingMean(double CDSMean)
	{
	double Mean = AccelProb(CDSMean, P_GeneCDSToNCEAccel);
	return Mean;
	}

static void GetRandomCEClumpLengths(unsigned CELength,
  vector<unsigned> &ClumpLengths)
	{
	RandDist RD;
	GetNormRD(P_CEClumpMeanLength, P_CEClumpLengthStdDev, RD);

	ClumpLengths.clear();
	unsigned Remaining = CELength;
	while (Remaining > 0)
		{
		unsigned Length = (unsigned) RD.Get();
		if (Length == 0)
			Length = 1;
		if (Length > Remaining)
			Length = Remaining;
		ClumpLengths.push_back(Length);
		Remaining -= Length;
		}
	}

// Input vector is assumed approximately sorted in
// increasing order. Re-sort so that values at the
// start are placed at the beginning and end of
// the vector so that small values are at the ends
// and large values are in the middle.
void NCESort(vector<double> &Probs)
	{
	const unsigned N = SIZE(Probs);

	vector<double> Probs2(N);
	vector<bool> Done(N);

	for (unsigned i = 0; i < N; ++i)
		{
		unsigned j = i%2 == 0 ? i/2 : N - i/2 - 1;
		Probs2[j] = Probs[i];
		asserta(!Done[j]);
		Done[j] = true;
		}

	for (unsigned i = 0; i < N; ++i)
		asserta(Done[i]);

	Probs = Probs2;
	}

static double SetCEProbs(GFFRecord2 &Rec, vector<double> &Probs, bool NCE)
	{
	unsigned CELength = (unsigned) Rec.GetLength();
	vector<unsigned> ClumpLengths;
	GetRandomCEClumpLengths(CELength, ClumpLengths);

	vector<vector<double> > Clumps;
	Clump(Probs, ClumpLengths, Clumps, !NCE);

	vector<double> ClumpedProbs;
	unsigned ClumpCount = SIZE(Clumps);
	for (unsigned ClumpIndex = 0; ClumpIndex < ClumpCount; ++ClumpIndex)
		{
		const vector<double> &Clump = Clumps[ClumpIndex];
		const unsigned ClumpLength = SIZE(Clump);
		for (unsigned i = 0; i < ClumpLength; ++i)
			ClumpedProbs.push_back(Clump[i]);
		}

	{
// This block for validation only
	vector<double> ProbsCopy(Probs);
	vector<double> ClumpedProbsCopy(ClumpedProbs);
	
	sort(ProbsCopy.begin(), ProbsCopy.end());
	sort(ClumpedProbsCopy.begin(), ClumpedProbsCopy.end());
	asserta(ProbsCopy == ClumpedProbsCopy);
	}

	if (NCE)
		NCESort(ClumpedProbs);

	double RoundedSum = SetProbsAttr(Rec, ClumpedProbs);

	double Mean = RoundedSum/CELength;
	if (Rec.Feature == "CDS")
		{
		++g_CDSCount;
		g_SumCDSMean += Mean;
		}
	else if (Rec.Feature == "NXE" || Rec.Feature == "UTR")
		{
		++g_GeneNCCECount;
		g_SumGeneNCCEMean += Mean;
		}
	else if (Rec.Feature == "NGE")
		{
		++g_NonGeneCECount;
		g_SumNonGeneCEMean += Mean;
		}
	else
		asserta(false);

	return RoundedSum;
	}

static double SetNonGeneCEProbs(GFFRecord2 &Rec)
	{
	double Mean = GetNonGeneCEMean();

	unsigned CELength = (unsigned) (Rec.GetLength());

	vector<double> Probs;
	GetRandomProbs(CELength, Mean, Probs);

	return SetCEProbs(Rec, Probs, true);
	}

static void SetGeneCEProbs(vector<GFFRecord2> &CEs, double Mean, bool NCE)
	{
	unsigned CECount = SIZE(CEs);
	if (CECount == 0)
		return;

	vector<unsigned> CELengths;
	unsigned TotalLength = 0;
	for (unsigned i = 0; i < CECount; ++i)
		{
		const GFFRecord2 &Rec = CEs[i];
		unsigned CELength = (unsigned) Rec.GetLength();
		TotalLength += CELength;
		CELengths.push_back(CELength);
		}

	vector<double> Probs;
	GetRandomProbs(TotalLength, Mean, Probs);

	vector<vector<double> > Clumps;
	Clump(Probs, CELengths, Clumps, true);

	double RoundedSum = 0;
	for (unsigned i = 0; i < CECount; ++i)
		{
		GFFRecord2 &Rec = CEs[i];
		RoundedSum += SetCEProbs(Rec, Clumps[i], NCE);
		}

	g_CurrentGeneLength += TotalLength;
	g_SumCurrentGeneBaseProbs += RoundedSum;
	}

static void SetGeneCEProbs(const vector<GFFRecord2> &CEs)
	{
	vector<GFFRecord2> CDSs;
	vector<GFFRecord2> NCCEs;

	const unsigned CECount = SIZE(CEs);
	for (unsigned i = 0; i < CECount; ++i)
		{
		const GFFRecord2 &Rec = CEs[i];
		if (Rec.Feature == "CDS")
			CDSs.push_back(Rec);
		else
			NCCEs.push_back(Rec);
		}

	double CDSMean = GetCDSMean();
	double NCCEMean = GetGeneNonCodingMean(CDSMean);

	if (CDSs.empty())
		{
		asserta(!NCCEs.empty());
		const GFFRecord2 &Rec = CEs.front();
		unsigned GeneIndex = GFFGetRequiredIntAttr(Rec.Attrs, "gene_index");
		Die("Gene %u has no CDSs", GeneIndex);
		}

	SetGeneCEProbs(CDSs, CDSMean, false);
	SetGeneCEProbs(NCCEs, NCCEMean, true);

	double GeneMean = g_SumCurrentGeneBaseProbs/g_CurrentGeneLength;

	++g_GeneCount;
	g_SumGeneMean += GeneMean;

	g_SumCurrentGeneBaseProbs = 0;
	g_CurrentGeneLength = 0;

	AddToOutput(CDSs);
	AddToOutput(NCCEs);
	}

void AssignProbs()
	{
	const char *CEFileName = RequiredValueOpt("assprobs");
	const char *OutFileName = RequiredValueOpt("out");
	const char *ModelFileName = ValueOpt("model");

	SetModel(ModelFileName);

	GetTruncNormRDNewton(P_NCCEMeanMean, P_NCCEMeanStdDev, g_NonGeneCEMeanRD);
	GetTruncNormRDNewton(P_CDSMeanMean, P_CDSMeanStdDev, g_CDSMeanRD);

	string SeqName = "";
	unsigned RecordCount = 0;
	FILE *fCEs = OpenStdioFile(CEFileName);
	FILE *fOut = CreateStdioFile(OutFileName);
	GFFRecord2 Rec;
	unsigned FileSize = (unsigned) GetStdioFileSize(fCEs);
	vector<GFFRecord2> CEs;

	unsigned GeneCECount = 0;
	ProgressStart("Reading");
	while (GetNextGFFRecord(fCEs, Rec))
		{
		if ((RecordCount++)%100 == 0)
			{
			unsigned Pos = (unsigned) GetStdioFilePos(fCEs);
			ProgressStep(Pos, FileSize);
			}
		//const string Name(Rec.SeqName);
		//if (SeqName == "")
		//	SeqName = Name;
		//else if (Name != SeqName)
		//	Die("GFF file has more than one sequence name '%s', '%s'",
		//	  SeqName.c_str(), Name.c_str());

		asserta(Rec.End >= Rec.Start);
		unsigned GeneIndex = UINT_MAX;
		bool Found = GFFGetIntAttr(Rec.Attrs, "gene_index", GeneIndex);
		if (!Found)
			{
			if (Rec.Feature == "NGE")
				SetNonGeneCEProbs(Rec);
			g_OutRecs.push_back(Rec);
			continue;
			}

		++GeneCECount;
		map<unsigned, vector<GFFRecord2> >::iterator p =
		  g_GeneIndexToCEs.find(GeneIndex);
		if (p == g_GeneIndexToCEs.end())
			{
			vector<GFFRecord2> CEs;
			CEs.push_back(Rec);
			g_GeneIndexToCEs[GeneIndex] = CEs;
			}
		else
			{
			vector<GFFRecord2> &CEs = p->second;
			CEs.push_back(Rec);
			}
		}
	ProgressDone();

	ProgressStart("Assigning probs to genes");
	unsigned GeneCount = SIZE(g_GeneIndexToCEs);
	unsigned Count = 0;
	for (map<unsigned, vector<GFFRecord2> >::iterator p = g_GeneIndexToCEs.begin();
	  p != g_GeneIndexToCEs.end(); ++p)
		{
		if (Count++%10 == 0)
			ProgressStep(Count, GeneCount);

		vector<GFFRecord2> &CEs = p->second;
		SetGeneCEProbs(CEs);
		}
	ProgressDone();

	Progress("Sorting");
	sort(g_OutRecs.begin(), g_OutRecs.end());

	unsigned OutCount = SIZE(g_OutRecs);
	Count = 0;
	ProgressStart("Writing output");
	for (vector<GFFRecord2>::const_iterator p = g_OutRecs.begin();
	  p != g_OutRecs.end(); ++p)
		{
		if (Count++%100 == 0)
			ProgressStep(Count, OutCount);
		WriteGFFRec(fOut, *p);
		}
	ProgressDone();

	Log("\n");
	Log("        Type      Number    Mean\n");
	Log("============  ==========  ======\n");

	Log("%12.12s", "Gene");
	Log("  %10u", g_GeneCount);
	if (g_GeneCount > 0)
		Log("  %6.4f", g_SumGeneMean/g_GeneCount);
	Log("\n");

	Log("%12.12s", "CDS");
	Log("  %10u", g_CDSCount);
	if (g_CDSCount > 0)
		Log("  %6.4f", g_SumCDSMean/g_CDSCount);
	Log("\n");

	Log("%12.12s", "Gene NCCE");
	Log("  %10u", g_GeneNCCECount);
	if (g_GeneNCCECount > 0)
		Log("  %6.4f", g_SumGeneNCCEMean/g_GeneNCCECount);
	Log("\n");

	Log("%12.12s", "Non-gene CE");
	Log("  %10u", g_NonGeneCECount);
	if (g_NonGeneCECount > 0)
		Log("  %6.4f", g_SumNonGeneCEMean/g_NonGeneCECount);
	Log("\n");

	CloseStdioFile(fCEs);
	CloseStdioFile(fOut);
	}
