#include <ctype.h>
#include "evolver.h"
#include "progress.h"
#include "getopts.h"

bool isacgt(char Letter);

void Evolver::BaseList_Init(const char *Seq, unsigned L,
  const BaseInfo &BaseTemplate, bool ShowProgress)
	{
	m_Bases.clear();
	m_BaseListFwds.clear();
	m_BaseListBwds.clear();
	m_AncPositions.clear();

	m_FirstEvolvedPos = UINT32_MAX;
	m_LastEvolvedPos = UINT32_MAX;
	m_FirstDeletedPos = UINT32_MAX;
	m_LastDeletedPos = UINT32_MAX;

	m_AncestralLength = L;
	m_EvolvedLength = L;

	if (L == 0)
		return;

	BaseInfo Base(BaseTemplate);

// Allocate memory to allow for 2x growth. With re-use of deleted
// extended bases, 2x is the asympotic number of bases in the
// extended genome.
	double ExtendFactor = 2.0;
	FloatOpt("ext", &ExtendFactor);
	size_t ExtendedSize = (size_t) (ExtendFactor*L);

	m_Bases.reserve(ExtendedSize);
	m_BaseListFwds.reserve(ExtendedSize);
	m_BaseListBwds.reserve(ExtendedSize);

	m_AncPositions.reserve(L/2);

	if (ShowProgress)
		ProgressStart("Initializing sequence");
	for (unsigned i = 0; i < L; ++i)
		{
		if (ShowProgress && i%10000 == 0)
			ProgressStep(i, L);

		char Letter = (char) toupper(Seq[i]);
		if (!isacgt(Letter))
			{
			if (isprint(Letter))
				Die("Invalid letter '%c' in sequence", Letter);
			else
				Die("Invalid byte 0x%02x in sequence", (unsigned char) Letter);
			}

		Base.SetLetter(Letter);
		Base.SetAncLetter(Letter);
		m_Bases.push_back(Base);
		m_BaseListFwds.push_back(i + 1 == L ? UINT32_MAX : i + 1);
		m_BaseListBwds.push_back(i == 0 ? UINT32_MAX : i - 1);
		}

	if (ShowProgress)
		ProgressDone();

	m_AncestralLength = L;
	m_FirstEvolvedPos = 0;
	m_LastEvolvedPos = L - 1;
	}

uint32 Evolver::BaseList_Insert(const BaseInfo &Base, uint32 AfterPos,
  uint32 AncPos)
	{
	uint32 ExtendedLength = SIZE32(m_Bases);
	if (AfterPos >= ExtendedLength)
		Die("BaseList_Insert, AfterPos=%u >= extended length=%u",
		  AfterPos, ExtendedLength);

	uint32 Next = m_BaseListFwds[AfterPos];

// If possible, re-use deleted position
	uint32 Pos = m_FirstDeletedPos;
	if (Pos != UINT32_MAX)
		{
		asserta(Pos < ExtendedLength);
		if (!m_Bases[Pos].IsDeleted())
			{
			m_Bases[Pos].LogMe(true);
			Die("m_Bases[%"PF"] not deleted", Pos);
			}
		m_FirstDeletedPos = m_BaseListFwds[Pos];
		if (m_FirstDeletedPos == UINT32_MAX)
			m_LastDeletedPos = UINT32_MAX;
		else
			{
			asserta(m_Bases[m_FirstDeletedPos].IsDeleted());
			m_BaseListBwds[m_FirstDeletedPos] = UINT32_MAX;
			}

		m_Bases[Pos] = Base;
		asserta(Pos - m_AncestralLength < SIZE(m_AncPositions));
		m_AncPositions[Pos - m_AncestralLength] = AncPos;

		m_BaseListFwds[Pos] = Next;
		m_BaseListBwds[Pos] = AfterPos;
		}
	else
		{
		Pos = SIZE(m_Bases);
		m_Bases.push_back(Base);
		m_AncPositions.push_back(AncPos);
		m_BaseListFwds.push_back(Next);
		m_BaseListBwds.push_back(AfterPos);
		}

	++m_EvolvedLength;

	if (AfterPos == UINT32_MAX)
		m_FirstEvolvedPos = Pos;
	else
		m_BaseListFwds[AfterPos] = Pos;

	if (Next == UINT32_MAX)
		m_LastEvolvedPos = Pos;
	else
		m_BaseListBwds[Next] = Pos;

	return Pos;
	}

void Evolver::BaseList_Delete(uint32 Pos)
	{
	if (Pos >= SIZE(m_Bases))
		Die("BaseList_Delete, pos=%u >= evolved length=%u", Pos, SIZE(m_Bases));
	asserta(m_EvolvedLength > 0);
	--m_EvolvedLength;

	m_Bases[Pos].Delete();

	uint32 Next = m_BaseListFwds[Pos];
	uint32 Prev = m_BaseListBwds[Pos];

	if (Next == UINT32_MAX)
		m_LastEvolvedPos = m_BaseListBwds[Pos];
	else
		m_BaseListBwds[Next] = Prev;

	if (Prev == UINT32_MAX)
		m_FirstEvolvedPos = m_BaseListFwds[Pos];
	else
		m_BaseListFwds[Prev] = Next;

// Add to deleted list iff in extended genome, this allows re-use
// of the position.
	if (Pos < m_AncestralLength)
		{
		m_BaseListFwds[Pos] = UINT32_MAX;
		m_BaseListBwds[Pos] = UINT32_MAX;
		return;
		}

	m_BaseListFwds[Pos] = UINT32_MAX;
	m_BaseListBwds[Pos] = m_LastDeletedPos;

	if (m_LastDeletedPos == UINT32_MAX)
		{
		asserta(m_FirstDeletedPos == UINT32_MAX);
		m_FirstDeletedPos = Pos;
		}
	else
		m_BaseListFwds[m_LastDeletedPos] = Pos;
	m_LastDeletedPos = Pos;
	}

void Evolver::BaseList_Validate() const
	{
	const unsigned ExtendedLength = SIZE(m_Bases);
	if (ExtendedLength < m_AncestralLength)
		Die("ExtendedLength=%u, AncestralLength=%u",
		  ExtendedLength, m_AncestralLength);

	asserta(m_BaseListFwds.size() == ExtendedLength);
	asserta(m_BaseListBwds.size() == ExtendedLength);
	asserta(m_AncPositions.size() == ExtendedLength - m_AncestralLength);

	vector<bool> Found(ExtendedLength);

	if (m_FirstEvolvedPos != UINT32_MAX && m_FirstEvolvedPos >= ExtendedLength)
		Die("m_FirstEvolvedPos=%u >= extended length=%u",
		  m_FirstEvolvedPos, ExtendedLength);

	uint32 EvolvedLength = 0;
	uint32 Pos = m_FirstEvolvedPos;
	for (;;)
		{
		if (Pos == UINT32_MAX)
			break;
		++EvolvedLength;

		if (EvolvedLength > ExtendedLength)
			Die("Infinite loop detected in base list");

		if (Found[Pos])
			{
			BaseList_Log();
			Die("Bad base list, pos %u found twice", Pos);
			}
		Found[Pos] = true;

		uint32 NextPos = m_BaseListFwds[Pos];
		if (NextPos == UINT32_MAX)
			{
			if (Pos != m_LastEvolvedPos)
				{
				BaseList_Log();
				Die("Broken base list, m_BaseListFwds[%u]=UINT32_MAX, m_LastEvolvedPos=%u",
				  Pos, m_LastEvolvedPos);
				}
			break;
			}

		if (NextPos >= ExtendedLength)
			{
			BaseList_Log();
			Die("m_BaseListFwds[%u] = %u >= extended length=%u",
			  Pos, NextPos, ExtendedLength);
			}

		if (NextPos != UINT32_MAX && m_BaseListBwds[NextPos] != Pos)
			{
			BaseList_Log();
			Die("Broken base list, m_BaseListFwds[%u]=%u, m_BaseListBwds[%u]=%u",
			  Pos, NextPos, NextPos, m_BaseListBwds[NextPos]);
			}

		Pos = NextPos;
		}

	if (EvolvedLength != m_EvolvedLength)
		Die("EvolvedLength %u != m_EvolvedLength %u",
		  EvolvedLength, m_EvolvedLength);

	if (m_FirstDeletedPos != UINT32_MAX && m_FirstDeletedPos >= ExtendedLength)
		Die("m_FirstDeletedPos=%u >= extended length=%u",
		  m_FirstDeletedPos, ExtendedLength);

	Pos = m_FirstDeletedPos;
	unsigned LoopCounter = 0;
	for (;;)
		{
		if (Pos == UINT32_MAX)
			break;

		if (LoopCounter++ > ExtendedLength)
			Die("Infinite loop detected in deleted base list");

		if (Found[Pos])
			Die("Bad base list, pos %u found twice (del)", Pos);

		Found[Pos] = true;
		const BaseInfo &Base = m_Bases[Pos];
		if (!Base.IsDeleted())
			{
			Base.LogMe(true);
			Die("Bad base list, pos %u in deleted list but base not deleted", Pos);
			}

		uint32 NextPos = m_BaseListFwds[Pos];
		if (NextPos == UINT32_MAX)
			{
			if (Pos != m_LastDeletedPos)
				Die("Broken base list, m_BaseListFwds[%u]=UINT32_MAX, m_LastDeletedPos=%u (del)",
				  Pos, m_LastDeletedPos);
			break;
			}

		if (NextPos >= ExtendedLength)
			{
			BaseList_Log();
			Die("m_BaseListFwds[%u] = %u >= extended length=%u (del)",
			  Pos, NextPos, ExtendedLength);
			}

		if (NextPos != UINT32_MAX && m_BaseListBwds[NextPos] != Pos)
			Die("Broken base list, m_BaseListFwds[%u]=%u, m_BaseListBwds[%u]=%u (del)",
			  Pos, NextPos, NextPos, m_BaseListBwds[NextPos]);

		Pos = NextPos;
		}

	for (uint32 Pos = m_AncestralLength; Pos < ExtendedLength; ++Pos)
		{
		if (!Found[Pos] && !PosIsDeleted(Pos))
			Die("Bad base list, pos %u not found, extended length %u",
			  Pos, ExtendedLength);
		}

	for (uint32 i = 0; i < ExtendedLength - m_AncestralLength; ++i)
		{
		uint32 Pos = m_AncPositions[i];
		if (Pos != UINT32_MAX && Pos >= ExtendedLength)
			Die("Bad m_AncPositions[%u] = %u >= ext. len %u",
			  i, Pos, ExtendedLength);
		}
	}

void Evolver::BaseList_Invert(gpos_t Pos, gpos_t Length)
	{
	if (Length == 0)
		return;

	uint32 ExtendedLength = SIZE(m_BaseListFwds);
	asserta(SIZE(m_BaseListBwds) == ExtendedLength);

	gpos_t Left = GetPrevPos(Pos);
	gpos_t CurrPos = GPOS_T_MAX;
	gpos_t NextPos = Pos;
	gpos_t NextNextPos = m_BaseListFwds[NextPos];
	for (gpos_t i = 0; i + 1 < Length; ++i)
		{
		CurrPos = NextPos;
		NextPos = NextNextPos;
		asserta(NextPos < ExtendedLength);
		NextNextPos = m_BaseListFwds[NextPos];

		if (NextPos == UINT32_MAX)
			Die("BaseList_Invert, overflows end of sequence");

		if (NextPos == UINT32_MAX)
			m_FirstEvolvedPos = CurrPos;
		else
			{
			asserta(NextPos < ExtendedLength);
			m_BaseListFwds[NextPos] = CurrPos;
			}

		asserta(CurrPos < ExtendedLength);
		m_BaseListBwds[CurrPos] = NextPos;
		}

	gpos_t End = NextPos;
	gpos_t Right = NextNextPos;

	m_BaseListFwds[Pos] = Right;

	if (Right == UINT32_MAX)
		m_LastEvolvedPos = Pos;
	else
		m_BaseListBwds[Right] = Pos;

	if (Left == UINT32_MAX)
		m_FirstEvolvedPos = End;
	else
		m_BaseListFwds[Left] = End;

	asserta(End < ExtendedLength);
	m_BaseListBwds[End] = Left;
	}

void Evolver::BaseList_Log() const
	{
	Log("\n");
	Log("Ancestral:\n");
	Log("    AncPos         Bwd         Fwd      Type  +\n");
	Log("==========  ==========  ==========  ========  =\n");
	uint32 L = SIZE(m_Bases);
	for (uint32 Pos = 0; Pos < L; ++Pos)
		{
		uint32 Fwd = m_BaseListFwds[Pos];
		uint32 Bwd = m_BaseListBwds[Pos];
		Log("%10u", Pos);

		if (Bwd == UINT32_MAX)
			Log("           *");	
		else
			Log("  %10u", Bwd);

		if (Fwd == UINT32_MAX)
			Log("           *");	
		else
			Log("  %10u", Fwd);

		const BaseInfo &Base = m_Bases[Pos];
		Log("  %8.8s", BaseTypeToStr((BaseType) Base.Type));
		Log("  %c", pom(Base.Plus));
		Log("\n");
		}

	Log("\n");
	Log("Evolved:\n");
	uint32 EvPos = 0;
	uint32 Pos = m_FirstEvolvedPos;
	Log("     EvPos      AncPos         Bwd         Fwd      Type  +\n");
	Log("==========  ==========  ==========  ==========  ========  =\n");
	for (;;)
		{
		if (Pos == UINT32_MAX)
			break;
		Log("%10u", EvPos++);
		uint32 Fwd = m_BaseListFwds[Pos];
		uint32 Bwd = m_BaseListBwds[Pos];
		Log("  %10u", Pos);

		if (Bwd == UINT32_MAX)
			Log("           *");	
		else
			Log("  %10u", Bwd);

		if (Fwd == UINT32_MAX)
			Log("           *");	
		else
			Log("  %10u", Fwd);

		const BaseInfo &Base = m_Bases[Pos];
		Log("  %8.8s", BaseTypeToStr((BaseType) Base.Type));
		Log("  %c", pom(Base.Plus));
		Log("\n");

		Pos = Fwd;
		if (EvPos > L)
			{
			Log("** LOOP IN LIST **\n");
			break;
			}
		}
	Log("\n");
	}

void Evolver::BaseList_LogSummary() const
	{
	Log("\n");
	Log("Ancestral length     %"PF"\n", m_AncestralLength);
	Log("Evolved length       %"PF"\n", m_EvolvedLength);
	Log("m_Bases size         %"PF"\n", SIZE(m_Bases));
	Log("m_BaseListFwds size  %"PF"\n", SIZE(m_BaseListFwds));
	Log("m_BaseListBwds size  %"PF"\n", SIZE(m_BaseListBwds));
	Log("m_AncPositions size  %"PF"\n", SIZE(m_AncPositions));
	Log("m_Bases cap          %"PF"\n", (unsigned) m_Bases.capacity());
	Log("m_BaseListFwds cap   %"PF"\n", (unsigned) m_BaseListFwds.capacity());
	Log("m_BaseListBwds cap   %"PF"\n", (unsigned) m_BaseListBwds.capacity());
	Log("m_AncPositions cap   %"PF"\n", (unsigned) m_AncPositions.capacity());
	Log("m_BaseListBwds cap   %"PF"\n", (unsigned) m_BaseListBwds.capacity());
	Log("m_FirstEvolvedPos    %"PF"\n", m_FirstEvolvedPos);
	Log("m_LastEvolvedPos     %"PF"\n", m_LastEvolvedPos);
	Log("m_FirstDeletedPos    %"PF"\n", m_FirstDeletedPos);
	Log("m_LastDeletedPos     %"PF"\n", m_LastDeletedPos);
	}

void Evolver::BaseList_LogPos(uint32 Pos) const
	{
	Log("\n");
	Log("BaseList_LogPos(%u)", Pos);
	if (Pos == GPOS_T_MAX)
		{
		Log(" = GPOS_T_MAX\n");
		return;
		}
	if (Pos >= SIZE(m_Bases))
		{
		Log(" ** Out of range\n");
		return;
		}
	LogNeighborsAsTable(Pos);
	}
