#include <stdio.h>
#include <stdlib.h>
#include "bmtypes.h"
#include "bmgraph.h"
#include "bmsolver.h"
#include "utils.h"

#define ABS( x ) ( (x) >= 0 ) ? (x) : -(x)

BMGraph::BMGraph(int TotalNodes, int TotalEdges)
{
	Initialized = false;
	Init(TotalNodes, TotalEdges);
}

BMGraph::BMGraph()
{
	Initialized = false;
}

BMGraph::~BMGraph()
{
	if (Initialized)
		Destroy();
}

void BMGraph::Destroy()
{
	if (!Initialized)
		Die("Invalid destroy() request");

	free(nodes);
	free(arcs);
	free(cap);

	Initialized = false;
}

void BMGraph::Init(int TotalNodes, int TotalEdges)
{
	if (Initialized)
		Die("Graph already initialized");

	Initialized = true;

	pos_current = 0;
	n = TotalNodes;
	m = TotalEdges;
	if ((n <= 0) || (m <= 0))
		Die("Graph requires at least one arc and one node");

	//Allocate
	nodes     = (node *)calloc(n + 2, sizeof(node));
	arcs      =  (arc *)calloc((2 * m) + 1, sizeof(arc));
	cap       =  (int *)calloc(2 * m, sizeof(int)); 
	arc_tail  =  (int *)calloc(2 * m, sizeof(int)); 
	arc_first =  (int *)calloc(n + 2, sizeof(int));
	/* arc_first [ 0 .. n+1 ] = 0 - initialized by calloc */

	if ((nodes == NULL) || (arcs == NULL) || (arc_first == NULL) || (arc_tail == NULL))
		Die("Memory allocation error");
			 
	for (node *in = nodes; in <= nodes + n; ++in)
		 in->excess = 0;

	/* setting pointer to the first arc */
	arc_current = arcs;
	node_max = 0;
	node_min = n;
	m_c = 0;
	total_p = total_n = 0;

	for (node *ndp = nodes; ndp < nodes + n; ++ndp)
		ndp->excess = 0;
}

void BMGraph::FillNode(int i, int excess)
{
	if ( i < 0 || i > n )
		Die("FillNode: Node index out of range");

	( nodes + i ) -> excess = excess;
	if ( excess > 0 ) total_p += excess;
	if ( excess < 0 ) total_n -= excess;
}

void BMGraph::FillEdge(int tail, int head, int low, int acap, price_t cost)
{
	if ( tail < 0 || tail > n || head < 0 || head > n)
		Die("FillEdge: Node index out of range");

	if ( acap < 0 ) {
		acap = MAX_32;
	}

	if ( low < 0 || low > acap )
		Die("FillEdge: Invalid minimum flow");

 /* no of arcs incident to node i is placed in arc_first[i+1] */
	arc_first[tail + 1] ++; 
	arc_first[head + 1] ++;
	node *in = nodes + tail;
	node *jn = nodes + head;

	/* storing information about the arc */
	arc_tail[pos_current]				 = tail;
	arc_tail[pos_current+1]			 = head;
	arc_current				-> head		 = jn;
	arc_current				-> r_cap	 = acap - low;
	cap[pos_current]						 = acap;
	arc_current				-> cost		 = cost;
	arc_current				-> sister  = arc_current + 1;
( arc_current + 1 ) -> head		 = nodes + tail;
( arc_current + 1 ) -> r_cap	 = 0;
	cap[pos_current+1]					 = 0;
( arc_current + 1 ) -> cost		 = -cost;
( arc_current + 1 ) -> sister  = arc_current;

	in -> excess -= low;
	jn -> excess += low;

	/* searching for minimum and maximum node */
	if ( head < node_min ) node_min = head;
	if ( tail < node_min ) node_min = tail;
	if ( head > node_max ) node_max = head;
	if ( tail > node_max ) node_max = tail;

	if ( cost < 0 ) cost = -cost;
	if ( cost > m_c && acap > 0 ) m_c = cost;

	arc_current += 2;
	pos_current += 2;
}

void BMGraph::parse_bm(int *n_ad, int *m_ad, node **nodes_ad, 
					arc **arcs_ad, int *node_min_ad, price_t *m_c_ad, 
					int **cap_ad)
	{

		if ( ABS( total_p - total_n ) > 0.5 ) /* unbalanced problem */
			Die("Unbalanced problem");

	/********** ordering arcs - linear time algorithm ***********/

	/* first arc from the first node */
	( nodes + node_min ) -> first = arcs;

	/* before below loop arc_first[i+1] is the number of arcs outgoing from i;
		 after this loop arc_first[i] is the position of the first 
		 outgoing from node i arcs after they would be ordered;
		 this value is transformed to pointer and written to node.first[i]
		 */
	 
	int i, arc_num, arc_new_num;

	for ( i = node_min + 1; i <= node_max + 1; i ++ ) 
		{
			arc_first[i]					+= arc_first[i-1];
			( nodes + i ) -> first = arcs + arc_first[i];
		}


	for ( i = node_min; i < node_max; i ++ ) /* scanning all the nodes	
																							exept the last*/
		{

			int last = ( ( nodes + i + 1 ) -> first ) - arcs;
															 /* arcs outgoing from i must be cited		
																from position arc_first[i] to the position
																equal to initial value of arc_first[i+1]-1	*/

			for ( arc_num = arc_first[i]; arc_num < last; arc_num ++ )
				{ int tail = arc_tail[arc_num];

					while ( tail != i )
						/* the arc no  arc_num	is not in place because arc cited here
							 must go out from i;
							 we'll put it to its place and continue this process
							 until an arc in this position would go out from i */

						{ arc_new_num  = arc_first[tail];
							arc_current  = arcs + arc_num;
							arc *arc_new			 = arcs + arc_new_num;
							
							/* arc_current must be cited in the position arc_new		
								 swapping these arcs:																	*/

							node *head_p				 = arc_new -> head;
							arc_new -> head			 = arc_current -> head;
							arc_current -> head  = head_p;

							int acap						 = cap[arc_new_num];
							cap[arc_new_num]		 = cap[arc_num];
							cap[arc_num]				 = acap;

							acap								 = arc_new -> r_cap;
							arc_new -> r_cap		 = arc_current -> r_cap;
							arc_current -> r_cap = acap;

							price_t cost				= arc_new -> cost;
							arc_new -> cost			 = arc_current -> cost;
							arc_current -> cost  = cost;

							if ( arc_new != arc_current -> sister )
								{
									arc *arc_tmp								 = arc_new -> sister;
									arc_new  -> sister		 = arc_current -> sister;
									arc_current -> sister  = arc_tmp;

									( arc_current -> sister ) -> sister = arc_current;
									( arc_new			-> sister ) -> sister = arc_new;
								}

							arc_tail[arc_num] = arc_tail[arc_new_num];
							arc_tail[arc_new_num] = tail;

							/* we increase arc_first[tail]	*/
							arc_first[tail] ++ ;

							tail = arc_tail[arc_num];
						}
				}
			/* all arcs outgoing from  i	are in place */
		}				

	/* -----------------------	arcs are ordered	------------------------- */

	/*------------ testing network for possible excess overflow ---------*/

	for (node *ndp = nodes + node_min; ndp <= nodes + node_max; ndp ++ )
	{
		 excess_t cap_in  =	 ( ndp -> excess );
		 excess_t cap_out = - ( ndp -> excess );
		 for ( arc_current = ndp -> first; arc_current != (ndp+1) -> first; 
					 arc_current ++ )
				{
					arc_num = arc_current - arcs;
					if ( cap[arc_num] > 0 ) cap_out += cap[arc_num];
					if ( cap[arc_num] == 0 ) 
						cap_in += cap[( arc_current -> sister )-arcs];
				}

		 //TODO: Check if cap_in == cap_out == 0 or what ?
	}

	if ((node_min < 0) || (node_min > 1)) /* unbalanced problem */
		Die("Error: Lowest connected node index is not 0 or 1");

	/* ----------- assigning output values ------------*/
	*m_ad = m;
	*n_ad = node_max - node_min + 1;
	*node_min_ad = node_min;
	*nodes_ad = nodes + node_min;
	*arcs_ad = arcs;
	*m_c_ad  = m_c;
	*cap_ad		= cap;

	/* free internal memory */
	free ( arc_first ); free ( arc_tail );
}

void BMGraph::Solve()
{
	if (Initialized)
	{
		BMSolver BMS(*this);
	}
	else
		Die("Invalid solve() request");
}

node *BMGraph::GetNodeArray()
{
	if (!Initialized)
		Die("Invalid GetNodeArray() request");

	return nodes;
}

arc *BMGraph::GetArcArray()
{
	if (!Initialized)
		Die("Invalid GetArcArray() request");

	return arcs;
}

int *BMGraph::GetCapArray()
{
	if (!Initialized)
		Die("Invalid GetCapArray() request");

	return cap;
}
