#include <vector>
#include <map>
#include <string>
#include <iterator>
#include <algorithm>
#include "revfile.h"
#include "types.h"
#include "utils.h"
#include "getopts.h"
#include "progress.h"


VALUE_OPT ValueOpts[] =
{
	{ "rev",				0 },
	{ "log",				0 }
};
int ValueOptCount = sizeof(ValueOpts)/sizeof(ValueOpts[0]);

FLAG_OPT FlagOpts[] =
{
	{ "dummy",				0 }
};
int FlagOptCount = sizeof(FlagOpts)/sizeof(FlagOpts[0]);

void Usage()
{
	fprintf(stderr,
"Usage:\n"
"    boolcov -rev file.rev\n"
"\n"
	);
}

// Counts and returns the number of Aligned and Ungapped positions
// in a genome. Requires 2bit*Sizeof(Genome) memory.
//
// Any genomic position that participates in an alignment is Aligned.
//
// Any genomic position that participates in an alignment, where in
//   that column there exists at least one other row with a non-gap
//   character, is Ungapped.
//
// Sanity check: Aligned >= Ungapped
//
std::pair<uint64,uint64> GetUsedBaseCount(RevFile &InFile, unsigned GenomeIndex)
{
	std::vector<bool> Aligned(InFile.GetGenomeInfo(GenomeIndex).GetLength64());
	std::vector<bool> Ungapped(InFile.GetGenomeInfo(GenomeIndex).GetLength64());
	std::map<unsigned, uint64> Offsets;

	uint64 Offset = 0;
	for (unsigned c = 0; c != InFile.GetChrCount(GenomeIndex); ++c)
	{
		unsigned id = InFile.GetChrId(GenomeIndex, c);
		Offsets[id] = Offset;
		Offset += InFile.GetChrLength(GenomeIndex, id);
	}

	ProgressStart("Processing genome %u...", GenomeIndex);
	for (unsigned a = 0; a != InFile.GetAlnInfoCount(); ++a)
	{
		if (!(a & 0xfff))
			ProgressStep(a, InFile.GetAlnInfoCount());

		AlnInfo AI;
		unsigned AlnIndex = InFile.GetAlnInfo(a, AI);
		if (AlnIndex == (unsigned)-1)
			continue;

		StdStrVec Rows;
		InFile.GetAlnRows(AlnIndex, Rows);

		for (unsigned s = 0; s != AI.GetSegCount(); ++s)
		{
			if (AI.Segs[s].GenomeIndex != GenomeIndex)
				continue;
			if (AI.Segs[s].Plus)
			{
				uint64 k = Offsets[AI.Segs[s].ChrId] + AI.Segs[s].lo;
				for (unsigned i = 0; i != Rows[s].size(); ++i)
				{
					if (Rows[s][i] != '-')
					{
						Aligned[k] = true;
						for (unsigned r = 0; r != Rows.size(); ++r)
							if ((r != s) && (Rows[r][i] != '-'))
							{
								Ungapped[k] = true;
								break;
							}
						++k;
					}
				}
			}
			else
			{
				uint64 k = Offsets[AI.Segs[s].ChrId] + AI.Segs[s].hi;
				for (unsigned i = 0; i != Rows[s].size(); ++i)
				{
					if (Rows[s][i] != '-')
					{
						Aligned[k] = true;
						for (unsigned r = 0; r != Rows.size(); ++r)
							if ((r != s) && (Rows[r][i] != '-'))
							{
								Ungapped[k] = true;
								break;
							}
						--k;
					}
				}
			}
		}
	}
	ProgressDone();

	std::pair<uint64,uint64> result(0,0);
	for (std::vector<bool>::const_iterator b = Aligned.begin(); b != Aligned.end(); ++b)
		if (*b)
			++result.first;
	for (std::vector<bool>::const_iterator b = Ungapped.begin(); b != Ungapped.end(); ++b)
		if (*b)
			++result.second;

	return result;
}

int main(int argc, char *argv[])
{
	ProcessArgVect(argc, argv);

	RevFile InFile;
	InFile.Open(RequiredValueOpt("rev"));

	std::vector<uint64> UngappedUsage(InFile.GetGenomeCount());
	std::vector<uint64> Usage(InFile.GetGenomeCount());
	std::vector<uint64> Size(InFile.GetGenomeCount());

	for (unsigned g = 0; g != InFile.GetGenomeCount(); ++g)
	{
		std::pair<uint64,uint64> pu = GetUsedBaseCount(InFile, g);
		Usage[g] = pu.first;
		UngappedUsage[g] = pu.second;
		Size[g] = InFile.GetGenomeInfo(g).GetLength64();
	}

	printf("%12s %12s %12s %6s %12s %6s\n", "Genome", "Size", "Aligned", "Al.%", "Ungapped", "Un.%");
	printf("----------------------------------------------------------------------\n");
	for (unsigned g = 0; g != InFile.GetGenomeCount(); ++g)
		printf("%12s %12"UINT64_PRINTF" %12"UINT64_PRINTF" %5.1f%% %12"UINT64_PRINTF" %5.1f%%\n", InFile.GetGenomeName(g).c_str(), Size[g], Usage[g], (Usage[g] * 100.0 / Size[g]), UngappedUsage[g], (UngappedUsage[g] * 100.0 / Size[g]));

	return 0;
}
