#include "utils.h"
#include "revfile.h"
#include "getopts.h"
#include "progress.h"
#include "uniquebid.h"

void ComputeBundleScores(const AIVec &AIs,
  std::map<UniqueBundleId, ai_score_t> &BundleScores);

void BundleScoresOneFile(const char *InFileName, const char *OutFileName)
	{
	RevFile In;
	RevFile Out;

	In.Open(InFileName);

	const std::vector<GenInfo> &Genomes = In.GetGenomes();
	Out.Create(OutFileName, Genomes);

	Progress("Compute bundle scores: getting AIs");
	AIVec AIs;
	In.GetAllAlnInfos(AIs);

	const unsigned AICount = (unsigned) AIs.size();
	Progress("%u AIs", AICount);

	std::map<UniqueBundleId, ai_score_t> BundleScores;
	ComputeBundleScores(AIs, BundleScores);

	const unsigned BundleCount = (unsigned) BundleScores.size();
	std::vector<ai_score_t> Scores;
	Scores.resize(BundleCount);
	unsigned BundleId = 0;
	for (std::map<UniqueBundleId, ai_score_t>::const_iterator p = BundleScores.begin();
	  p != BundleScores.end(); ++p)
		{
		ai_score_t Score = p->second;
		Scores[BundleId++] = Score;
		}
	std::sort(Scores.begin(), Scores.end());

	unsigned Hi90Index = (BundleCount*90)/100;
	unsigned Lo10Index = (BundleCount*10)/100;

	ai_score_t Hi90Score = Scores[Hi90Index];
	ai_score_t Lo10Score = Scores[Lo10Index];

	ai_score_t Range = Hi90Score - Lo10Score;
	if (Range == 0)
		Range = 1;

	ProgressStart("Assigning bundle scores");
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%1000 == 0)
			ProgressStep(AIIndex, AICount);

		AlnInfo &AI = AIs[AIIndex];
		unsigned BundleId = AI.GetBundleId();
		if (BundleId == UINT_MAX)
			Die("AI missing bundle id");

		UniqueBundleId UB;
		UB.BundleId = BundleId;
		UB.ChrId0 = AI.GetSeg(0).ChrId;
		UB.ChrId1 = AI.GetSeg(1).ChrId;

		std::map<UniqueBundleId, ai_score_t>::const_iterator p =
		  BundleScores.find(UB);
		if (p == BundleScores.end())
			Die("Bundle id not found");

		ai_score_t BundleScore = p->second;
		double ScaledBundleScore = BundleScore;

		if (ScaledBundleScore > Hi90Score)
			ScaledBundleScore = Hi90Score;

		if (ScaledBundleScore < Lo10Score)
			ScaledBundleScore = Lo10Score;

		ScaledBundleScore = ((ScaledBundleScore - Lo10Score)*99)/Range + 1;
		if (ScaledBundleScore > 100)
			ScaledBundleScore = 100;

		unsigned IntScaledBundleScore = (unsigned) ScaledBundleScore;

		AI.SetFloatAttr(ALN_ATTR_BundleScore, BundleScore);
		AI.SetAttr(ALN_ATTR_ScaledBundleScore, IntScaledBundleScore);

		unsigned AlnIndex = In.GetAlnIndex(AIIndex);
		if (AlnIndex != UINT_MAX)
			{
			StdStrVec Rows;
			In.GetAlnRows(AlnIndex, Rows);
			Out.AddAln(AI, Rows);
			}
		else
			Out.AddAlnInfo(AI);
		}
	ProgressDone();

	Progress("Closing files");
	In.Close();
	Out.Close();
	}

void BundleScores()
	{
	const char *InFileName = RequiredValueOpt("bundlescores");
	const char *OutFileName = RequiredValueOpt("out");

	BundleScoresOneFile(InFileName, OutFileName);
	}
