#include "utils.h"
#include "wordhist.h"
#include "fastastr.h"
#include "getopts.h"
#include "progress.h"
#include "xlat.h"
#include <limits.h>
#include <math.h>

uint64 CalcWordCount(ALPHA Alpha, unsigned WordSize);

extern bool g_PatternIsMiddleBase;

double g_ThreshAlpha = 0.0;

struct WordInfo
	{
	uint64 Word;
	COUNT_TYPE Count1;
	float s;
	};

static int Cmp(const void *arg1, const void *arg2)
	{
	WordInfo WI1 = *(WordInfo *) arg1;
	WordInfo WI2 = *(WordInfo *) arg2;
	if (WI1.s > WI2.s)
		return -1;
	else if (WI1.s == WI2.s)
		return 0;
	assert(WI1.s < WI2.s);
	return 1;
	}

static unsigned CalcThresholdPctToMask(unsigned WordSize, WordHist &WH1,
  WordHist &WH2)
	{
	if (!g_PatternIsMiddleBase)
		Die("Cannot use -pcttomask unless pattern is middle base only");

	const char *strPctToMask = RequiredValueOpt("pcttomask");

	unsigned TopN = 20;
	IntOpt("topn", &TopN);

	ALPHA Alpha = WH1.GetAlpha();
	if (WH2.GetAlpha() != Alpha)
		Die("CalcThreshold, different alphabets");

	double PctToMask = atof(strPctToMask);
	if (PctToMask < 0 || PctToMask > 100.0)
		Die("Invalid pcttomask '%s', must be in range 0 .. 100", strPctToMask);

	uint64 GenomeSize = WH1.GetTotalCount();
	uint64 WordCount = CalcWordCount(Alpha, WordSize);

	ProgressNoRes("Memory needed for count sort %u Mb",
	  (unsigned) (sizeof(WordInfo)*WordCount)/1000000 + 1);

	WordInfo *WIs = myalloc(WordInfo, WordCount);
	for (uint64 Word = 0; Word < WordCount; ++Word)
		{
		WordInfo &WI = WIs[Word];

		COUNT_TYPE Count1 = WH1.GetCount(Word);
		COUNT_TYPE Count2 = WH2.GetCount(Word);

		WI.Word = Word;
		WI.Count1 = Count1;
		WI.s = (float) sqrt(Count1*Count2);
		}

// Sort counts in /decreasing/ order
	size_t stWordCount = (size_t) WordCount;
	if (stWordCount != WordCount)
		Die("Word count overflows size_t");

	qsort(WIs, stWordCount, sizeof(WordInfo), Cmp);

	if (TopN > 0)
		{
		if (TopN > WordCount)
			TopN = (unsigned) WordCount;

		ALPHA Alpha = WH1.GetAlpha();

		ProgressNoRes("");
		ProgressNoRes("Top %u words", TopN);
		ProgressNoRes("");
		if (Alpha == ALPHA_Nucleo)
			{
	        ProgressNoRes("Word                                  Count1     Count2   sqrt(C1*C2)");
		    ProgressNoRes("---------------------------------  ---------  ---------  ------------");
			}
		else
			{
	        ProgressNoRes("Word                 Count1     Count2   sqrt(C1*C2)");
		    ProgressNoRes("----------------  ---------  ---------  ------------");
			}

		for (unsigned i = 0; i < TopN; ++i)
			{
			const WordInfo &WI = WIs[i];
			uint64 Word = WI.Word;
			unsigned Count1 = WH1.GetCount(Word);
			if (Count1 != WI.Count1)
				Die("Internal error count1");
			unsigned Count2 = WH2.GetCount(Word);
			double s = sqrt(Count1*Count2);
			if (Alpha == ALPHA_Nucleo)
				{
				char s1[64];
				char s2[64];
				ProgressNoRes("%-16s %-16.16s  %9u  %9u  %12.1f\n",
				  NucWordToStr(WI.Word, WordSize, s1),
				  NucWordToStr(RevCompNucWord(WI.Word, WordSize), WordSize, s2),
				  Count1,
				  Count2,
				  s);
				}
			else
				{
				ProgressNoRes("%-16s  %9u  %9u  %12.1f",
				  AminoWordToStr(WI.Word, WordSize), Count1, Count2, s);
				}
			}
		ProgressNoRes("");
		}

	uint64 TargetSum = (uint64) (PctToMask*GenomeSize/100.0);
	uint64 Sum = 0;
	for (uint64 Word = 0; Word < WordCount; ++Word)
		{
		WordInfo &WI = WIs[Word];

		Sum += WI.Count1;
		if (Sum >= TargetSum)
			{
			ProgressNoRes("Threshold %u masks %.1f %%",
			  (unsigned) WI.s, (Sum*100.0)/GenomeSize);
			return (unsigned) WI.s;
			}
		}
	Warning("Failed to satisfy -pcttomask %.1f, setting threshold to 0", PctToMask);
	return 0;
	}

static unsigned CalcThresholdTimesExpected(unsigned WordSize, WordHist &WH1, WordHist &WH2)
	{
	double Factor = RequiredFloatOpt("timesexpected");

	uint64 GenomeSize1 = WH1.GetGenomeSize();
	uint64 GenomeSize2 = WH2.GetGenomeSize();

	uint64 WordCount1 = CalcWordCount(WH1.GetAlpha(), WordSize);
	uint64 WordCount2 = CalcWordCount(WH2.GetAlpha(), WordSize);

	double Expected1 = ((double) GenomeSize1)/WordCount1;
	double Expected2 = ((double) GenomeSize2)/WordCount2;

	double dThresh = sqrt(Expected1*Expected2)*Factor;
	Log("Expected1=%g Expected2=%g Factor=%g Thresh=%g\n",
	  Expected1, Expected2, Factor, dThresh);
	if (dThresh > COUNT_TYPE_MAX)
		Die("Threshold %g exceeds max %u", dThresh, COUNT_TYPE_MAX);

	ProgressNoRes("Genome size 1 %s, expected count %u",
	  uint64tostr(GenomeSize1), (unsigned) Expected1);
	ProgressNoRes("Genome size 2 %s, expected count %u",
	  uint64tostr(GenomeSize2), (unsigned) Expected2);
	ProgressNoRes("Threshold %u", (unsigned) dThresh);

	return (unsigned) dThresh;
	}

void CheckThresholdArgs()
	{
	if (ValueOpt("threshold") == 0 && ValueOpt("timesexpected") == 0
	  && ValueOpt("pcttomask") == 0 && ValueOpt("threshalpha") == 0)
		Die("Must specify one of -threshold, -timesexptected, -threshalpha or -pcttomask");
	}

unsigned CalcThreshold(unsigned WordSize, WordHist &WH1, WordHist &WH2)
	{
	if (ValueOpt("threshold"))
		return RequiredIntOpt("threshold");

	else if (ValueOpt("timesexpected"))
		return CalcThresholdTimesExpected(WordSize, WH1, WH2);

	else if (ValueOpt("pcttomask"))
		return CalcThresholdPctToMask(WordSize, WH1, WH2);

	else if (ValueOpt("threshalpha"))
		{
		g_ThreshAlpha = atof(ValueOpt("threshalpha"));
		return 0;
		}

	Die("Must specify one of -threshold, -timesexptected, -pcttomask or -threshalpha");
	return 0;
	}
