#include "revfile.h"
#include "utils.h"
#include "getopts.h"
#include "progress.h"
#include <set>

#define TRACE	0

void GetPairwiseCols(unsigned GenomeIndex1, unsigned ChrId1,
  unsigned GenomeIndex2, unsigned ChrId2, 
  const AlnInfo &AI, const vector<string> &Rows,
  vector<PairwiseCol> &Cols);

static bool g_PairOnly = false;

char CompLetter(char c);

static inline bool isgap(char c)
	{
	return c == '-';
	}

static inline bool AllGaps(const string &s)
	{
	for (unsigned i = 0; i < SIZE(s); ++i)
		if (!isgap(s[i]))
			return false;
	return true;
	}

static bool Weq(unsigned i, unsigned j)
	{
	if (i == UINT_MAX || j == UINT_MAX)
		return true;
	return i == j;
	}

static bool WildcardEq(unsigned GenomeIndexA, unsigned ChrIdA,
  unsigned GenomeIndexB, unsigned ChrIdB)
	{
	return Weq(GenomeIndexA, GenomeIndexB) && Weq(ChrIdA, ChrIdB);
	}

static void InitAln(AlnInfo &AI, vector<string> &Rows,
  const PairwiseCol &Col)
	{
	AI.Clear();
	Rows.clear();

	Rows.resize(2);

	GCSeg Seg;
	Seg.GenomeIndex = Col.GenomeIndex1;
	Seg.ChrId = Col.ChrId1;
	Seg.lo = Col.Pos1;
	Seg.hi = Col.Pos1;
	Seg.Plus = true;
	AI.Segs.push_back(Seg);

	Seg.GenomeIndex = Col.GenomeIndex2;
	Seg.ChrId = Col.ChrId2;
	Seg.lo = Col.Pos2;
	Seg.hi = Col.Pos2;
	Seg.Plus = Col.Plus;
	AI.Segs.push_back(Seg);

	Rows[0].push_back(Col.Letter1);
	if (Col.Plus)
		Rows[1].push_back(Col.Letter2);
	else
		Rows[1].push_back(CompLetter(Col.Letter2));
	}

static bool AppendPlusCol(AlnInfo &AI, vector<string> &Rows,
  const PairwiseCol &Col)
	{
	asserta(Col.Plus);

	bool gap1 = isgap(Col.Letter1);
	bool gap2 = isgap(Col.Letter2);

	if (gap1 && gap2)
		return true;

	if (Col.Pos1 != AI.Segs[0].hi + 1)
		return false;
	if (Col.Pos2 != AI.Segs[1].hi + 1)
		return false;

	Rows[0].push_back(Col.Letter1);
	Rows[1].push_back(Col.Letter2);
	if (!gap1)
		++(AI.Segs[0].hi);
	if (!gap2)
		++(AI.Segs[1].hi);
	}

static bool AppendMinusCol(AlnInfo &AI, vector<string> &Rows,
  const PairwiseCol &Col)
	{
	asserta(!Col.Plus);
	bool gap1 = isgap(Col.Letter1);
	bool gap2 = isgap(Col.Letter2);

	if (gap1 && gap2)
		return true;

	if (Col.Pos1 != AI.Segs[0].hi + (gap1 ? 0 : 1))
		return false;
	if (Col.Pos2 != AI.Segs[1].lo - (gap2 ? 0 : 1))
		return false;

	Rows[0].push_back(Col.Letter1);
	Rows[1].push_back(CompLetter(Col.Letter2));
	if (!gap1)
		++(AI.Segs[0].hi);
	if (!gap2)
		--(AI.Segs[1].lo);
	}

static bool AppendCol(AlnInfo &AI, vector<string> &Rows,
  const PairwiseCol &Col)
	{
	if (Col.Plus)
		return AppendPlusCol(AI, Rows, Col);
	else
		return AppendMinusCol(AI, Rows, Col);
	}

static void ColSetToFile(const set<PairwiseCol> &Cols,
  const RevFile &In, const char *FileName)
	{
	if (FileName == 0)
		return;

	const vector<GenInfo> &Genomes = In.GetGenomes();
	RevFile Out;
	Out.Create(FileName, Genomes);

	AlnInfo AI;
	vector<string> Rows(2);
	for (set<PairwiseCol>::const_iterator p = Cols.begin();
	  p != Cols.end(); ++p)
		{
		const PairwiseCol &Col = *p;
		if (p == Cols.begin())
			{
			InitAln(AI, Rows, Col);
			continue;
			}
		if (!AppendCol(AI, Rows, Col))
			{
			if (!AllGaps(Rows[0]) && !AllGaps(Rows[1]))
				Out.AddAln(AI, Rows);

			InitAln(AI, Rows, Col);
			}
		}
	if (!Cols.empty() && !AllGaps(Rows[0]) && !AllGaps(Rows[1]))
		Out.AddAln(AI, Rows);
	}

static void OutAnnot(const char *OutFileName, const RevFile &In,
  unsigned GenomeIndex1, unsigned ChrId1,
  unsigned GenomeIndex2, unsigned ChrId2,
  const set<PairwiseCol> &ColsBoth)
	{
	const vector<GenInfo> &Genomes = In.GetGenomes();
	RevFile Out;
	Out.Create(OutFileName, Genomes);

	const unsigned AICount = In.GetAlnInfoCount();
	ProgressStart("Reading %u AIs", AICount);
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		ProgressStep(AIIndex, AICount);
		unsigned AlnIndex = In.GetAlnIndex(AIIndex);
		if (AlnIndex == UINT_MAX)
			continue;

		AlnInfo AI;
		In.GetAlnInfo(AIIndex, AI);

		const unsigned SegCount = AI.GetSegCount();
		if (SegCount == 0)
			continue;

		StdStrVec Rows;
		In.GetAlnRows(AlnIndex, Rows);
		asserta(SIZE(Rows) == SegCount);

		vector<PairwiseCol> Cols;
		GetPairwiseCols(GenomeIndex1, ChrId1, GenomeIndex2, ChrId2,
		  AI, Rows, Cols);
		if (Cols.empty())
			continue;

		AlnInfo NewAI;
		StdStrVec NewRows;
		NewAI.Score = 0;
		vector<unsigned> SegIndexes;
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = AI.Segs[SegIndex];
			if (!g_PairOnly || WildcardEq(Seg.GenomeIndex, Seg.ChrId, GenomeIndex1, ChrId1) ||
			  WildcardEq(Seg.GenomeIndex, Seg.ChrId, GenomeIndex2, ChrId2))
				{
				NewAI.Segs.push_back(Seg);
				NewRows.push_back(string(""));
				SegIndexes.push_back(SegIndex);
				}
			}
		if (SegIndexes.empty())
			continue;

		const unsigned ColCount = SIZE(Rows[0]);
		asserta(SIZE(Cols) == ColCount);
		for (unsigned ColIndex = 0; ColIndex < ColCount; ++ColIndex)
			{
			const PairwiseCol &Col = Cols[ColIndex];
			bool Upper = (ColsBoth.find(Col) != ColsBoth.end());
			for (unsigned k = 0; k < SIZE(SegIndexes); ++k)
				{
				unsigned SegIndex = SegIndexes[k];
				const GCSeg &Seg = AI.Segs[SegIndex];
				char c = Rows[SegIndex][ColIndex];
				if (Upper)
					c = toupper(c);
				else
					c = tolower(c);
				NewRows[k].push_back(c);
				}
			}

		Out.AddAln(NewAI, NewRows);
		}
	ProgressDone();

	Out.Close();
	}

void CmpRev()
	{
	const char *InFileNameA = RequiredValueOpt("in1");
	const char *InFileNameB = RequiredValueOpt("in2");

	const char *Genome1 = ValueOpt("gen1");
	const char *Genome2 = ValueOpt("gen2");

	const char *Chr1 = ValueOpt("chrname1");
	const char *Chr2 = ValueOpt("chrname2");

	const char *OutFileNameA = ValueOpt("out1");
	const char *OutFileNameB = ValueOpt("out2");

	bool Annot = FlagOpt("annot");
	g_PairOnly = FlagOpt("paironly");

	RevFile InA;
	RevFile InB;

	InA.Open(InFileNameA);
	InB.Open(InFileNameB);

	unsigned GenomeIndexA1 = UINT_MAX;
	unsigned GenomeIndexA2 = UINT_MAX;
	unsigned GenomeIndexB1 = UINT_MAX;
	unsigned GenomeIndexB2 = UINT_MAX;
	unsigned ChrIdA1 = UINT_MAX;
	unsigned ChrIdA2 = UINT_MAX;
	unsigned ChrIdB1 = UINT_MAX;
	unsigned ChrIdB2 = UINT_MAX;

	if (Genome1 != 0)
		GenomeIndexA1 = InA.GetGenomeIndex(Genome1);
	if (Genome2 != 0)
		GenomeIndexA2 = InA.GetGenomeIndex(Genome2);
	if (Chr1 != 0)
		ChrIdA1 = InA.GetChrId(GenomeIndexA1, Chr1);
	if (Chr2 != 0)
		ChrIdA2 = InA.GetChrId(GenomeIndexA2, Chr2);

	if (Genome1 != 0)
		GenomeIndexB1 = InB.GetGenomeIndex(Genome1);
	if (Genome2 != 0)
		GenomeIndexB2 = InB.GetGenomeIndex(Genome2);
	if (Chr1 != 0)
		ChrIdB1 = InB.GetChrId(GenomeIndexB1, Chr1);
	if (Chr2 != 0)
		ChrIdB2 = InB.GetChrId(GenomeIndexB2, Chr2);

	set<PairwiseCol> ColsA;
	set<PairwiseCol> ColsB;
	InA.GetPairwiseCols(GenomeIndexA1, ChrIdA1, GenomeIndexA2, ChrIdA2, ColsA);
	InB.GetPairwiseCols(GenomeIndexB1, ChrIdB1, GenomeIndexB2, ChrIdB2, ColsB);

	set<PairwiseCol> ColsBoth;
	set<PairwiseCol> ColsAOnly;
	set<PairwiseCol> ColsBOnly;

	set<PairwiseCol>::const_iterator pA = ColsA.begin();
	set<PairwiseCol>::const_iterator pB = ColsB.begin();

	unsigned BothCountWithGaps = 0;
	unsigned BothCountNoGaps = 0;
	unsigned AOnlyCountWithGaps = 0;
	unsigned AOnlyCountNoGaps = 0;
	unsigned BOnlyCountWithGaps = 0;
	unsigned BOnlyCountNoGaps = 0;
	unsigned BothMatchCount = 0;
	unsigned AOnlyMatchCount = 0;
	unsigned BOnlyMatchCount = 0;
	unsigned BothInvCount = 0;
	unsigned AOnlyInvCount = 0;
	unsigned BOnlyInvCount = 0;

#if	TRACE
	Log("Genome1  Genome2  Chr1  Chr2        Pos1        Pos2  XX+\n");
	Log("-------  -------  ----  ----  ----------  ----------  ---\n");
#endif
	while (pA != ColsA.end() && pB != ColsB.end())
		{
		const PairwiseCol &ColA = *pA;
		const PairwiseCol &ColB = *pB;
		bool Gap = ColA.HasGap() || ColB.HasGap();
		if (ColA < ColB)
			{
#if	TRACE
			Log("%7u  %7u  %4u  %4u  %10u  %10u  %c%c%c  A only\n",
			  ColA.GenomeIndex1, ColA.GenomeIndex2, ColA.ChrId1, ColA.ChrId2,
			  ColA.Pos1, ColA.Pos2, ColA.Letter1, ColA.Letter2, pom(ColA.Plus));
#endif
			ColsAOnly.insert(ColA);
			++AOnlyCountWithGaps;
			if (ColA.IsMatch())
				++AOnlyMatchCount;
			if (!Gap && !ColA.Plus)
				++AOnlyInvCount;
			if (!Gap)
				++AOnlyCountNoGaps;
			++pA;
			}
		else if (ColB < ColB)
			{
#if	TRACE
			Log("%7u  %7u  %4u  %4u  %10u  %10u  %c%c%c  B only\n",
			  ColB.GenomeIndex1, ColB.GenomeIndex2, ColB.ChrId1, ColB.ChrId2,
			  ColB.Pos1, ColB.Pos2, ColB.Letter1, ColB.Letter2, pom(ColB.Plus));
#endif
			ColsBOnly.insert(ColB);
			++BOnlyCountWithGaps;
			if (ColB.IsMatch())
				++BOnlyMatchCount;
			if (!Gap && !ColB.Plus)
				++BOnlyInvCount;
			if (!Gap)
				++BOnlyCountNoGaps;
			++pB;
			}
		else
			{
#if	TRACE
			Log("%7u  %7u  %4u  %4u  %10u  %10u  %c%c%c  Both\n",
			  ColA.GenomeIndex1, ColA.GenomeIndex2, ColA.ChrId1, ColA.ChrId2,
			  ColA.Pos1, ColA.Pos2, ColA.Letter1, ColA.Letter2, pom(ColA.Plus));
#endif
			ColsBoth.insert(ColA);
			++BothCountWithGaps;
			if (!Gap && !ColA.Plus)
				++BothInvCount;
			if (!Gap)
				{
				if (ColA.IsMatch())
					++BothMatchCount;
				++BothCountNoGaps;
				}
			++pA;
			++pB;
			}
		}

	while (pA != ColsA.end())
		{
		const PairwiseCol &ColA = *pA;
		bool Gap = ColA.HasGap();
#if	TRACE
		Log("%7u  %7u  %4u  %4u  %10u  %10u  %c%c%c  A only\n",
		  ColA.GenomeIndex1, ColA.GenomeIndex2, ColA.ChrId1, ColA.ChrId2,
		  ColA.Pos1, ColA.Pos2, ColA.Letter1, ColA.Letter2, pom(ColA.Plus));
#endif
		ColsAOnly.insert(ColA);
		++AOnlyCountWithGaps;
		if (ColA.IsMatch())
			++AOnlyMatchCount;
		if (!Gap && !ColA.Plus)
			++AOnlyInvCount;
		if (!Gap)
			++AOnlyCountNoGaps;
		++pA;
		}

	while (pB != ColsB.end())
		{
		const PairwiseCol &ColB = *pB;
		bool Gap = ColB.HasGap();
#if	TRACE
		Log("%7u  %7u  %4u  %4u  %10u  %10u  %c%c%c  B only\n",
		  ColB.GenomeIndex1, ColB.GenomeIndex2, ColB.ChrId1, ColB.ChrId2,
		  ColB.Pos1, ColB.Pos2, ColB.Letter1, ColB.Letter2, pom(ColB.Plus));
#endif
		ColsBOnly.insert(ColB);
		++BOnlyCountWithGaps;
		if (ColB.IsMatch())
			++BOnlyMatchCount;
		if (!Gap && !ColB.Plus)
			++BOnlyInvCount;
		if (!Gap)
			++BOnlyCountNoGaps;
		++pB;
		}

	if (Annot)
		{
		OutAnnot(OutFileNameA, InA, GenomeIndexA1, ChrIdA1,
		  GenomeIndexA2, ChrIdA2, ColsBoth);
		OutAnnot(OutFileNameB, InB, GenomeIndexB1, ChrIdB1,
		  GenomeIndexB2, ChrIdB2, ColsBoth);
		}

	if (!Annot)
		{
		ColSetToFile(ColsAOnly, InA, OutFileNameA);
		ColSetToFile(ColsBOnly, InB, OutFileNameB);
		}

	InA.Close();
	InB.Close();

	if (BothCountWithGaps + AOnlyCountWithGaps != SIZE(ColsA))
		Warning("BothCountWithGaps=%u + AOnlyCountWithGaps=%u = %u != SIZE(ColsA)=%u",
		  BothCountWithGaps, AOnlyCountWithGaps, BothCountWithGaps + AOnlyCountWithGaps, SIZE(ColsA));

	if (BothCountWithGaps + BOnlyCountWithGaps != SIZE(ColsB))
		Warning("BothCountWithGaps=%u + BOnlyCountWithGaps=%u = %u != SIZE(ColsB)=%u",
		  BothCountWithGaps, BOnlyCountWithGaps, BothCountWithGaps + BOnlyCountWithGaps, SIZE(ColsB));

	asserta(BothCountWithGaps >= BothCountNoGaps);
	asserta(AOnlyCountWithGaps >= AOnlyCountNoGaps);
	asserta(BOnlyCountWithGaps >= BOnlyCountNoGaps);

	unsigned BothCountGaps = BothCountWithGaps - BothCountNoGaps;
	unsigned AOnlyCountGaps = AOnlyCountWithGaps - AOnlyCountNoGaps;
	unsigned BOnlyCountGaps = BOnlyCountWithGaps - BOnlyCountNoGaps;

	Log("\n");
	Log("File1       %s\n", InFileNameA);
	Log("File2       %s\n", InFileNameB);
	Log("Genome1     %s\n", Genome1 == 0 ? "*" : Genome1);
	Log("Genome2     %s\n", Genome2 == 0 ? "*" : Genome2);
	Log("Chr1        %s\n", Chr1 == 0 ? "*" : Chr1);
	Log("Chr2        %s\n", Chr2 == 0 ? "*" : Chr2);
	Log("\n");
	Log("      Both      A only      B only\n");
	Log("----------  ----------  ----------\n");
	Log("%10u  %10u  %10u  Gapped columns\n",
	  BothCountGaps, AOnlyCountGaps, BOnlyCountGaps);
	Log("%10u  %10u  %10u  Ungapped columns\n",
	  BothCountNoGaps, AOnlyCountNoGaps, BOnlyCountNoGaps);
	Log("%10u  %10u  %10u  Gapped and ungapped columns\n",
	  BothCountWithGaps, AOnlyCountWithGaps, BOnlyCountWithGaps);
	Log("%10u  %10u  %10u  Match columns\n",
	  BothMatchCount, AOnlyMatchCount, BOnlyMatchCount);
	Log("%10u  %10u  %10u  Inverted columns\n",
	  BothInvCount, AOnlyInvCount, BOnlyInvCount);

	double BothPctId = Pct(BothMatchCount, BothCountNoGaps);
	double APctId = Pct(BothMatchCount + AOnlyMatchCount, BothCountNoGaps + AOnlyCountNoGaps);
	double BPctId = Pct(BothMatchCount + BOnlyMatchCount, BothCountNoGaps + BOnlyCountNoGaps);

	double BothPctInv = Pct(BothInvCount, BothCountNoGaps);
	double APctInv = Pct(BothInvCount + AOnlyInvCount, BothCountNoGaps + AOnlyCountNoGaps);
	double BPctInv = Pct(BothInvCount + BOnlyInvCount, BothCountNoGaps + BOnlyCountNoGaps);

	Log("\n");
	Log("      Both           A           B\n");
	Log("----------  ----------  ----------\n");
	Log("%10.1f  %10.1f  %10.1f   Percent identity\n", BothPctId, APctId, BPctId);
	Log("%10.1f  %10.1f  %10.1f   Percent inverted\n", BothPctInv, APctInv, BPctInv);
	}
