#include "evolver.h"
#include "evomodel.h"
#include "myrand.h"

/***
Sweeps must be done often enough that only a small fraction
of CpGs undergo a mutation to TpG. (Denote this fraction by f).
This is because substition does not commute with duplication;
you can estimate the age of a duplication by looking at the
number of substitutions between the two copies.

The input parameter is q=CpG_C_to_T_Ratio, which is defined as:

	q = (C->T rate outside CpG islands) / (C->T rate inside).

We expect typical q to be around 10. The C->T rate in CpG
islands is:

	s = 1 * P(C->T | C substitutes) * (1 - CGRejectProb)

With typical values, s = 0.6*0.76 = 0.46. The rate outside
an island is then Q = qs, with typical value 10*0.46 = 4.6.

A rate of s is produced by standard substitutions, so the
added rate to be implemented by CpG sweeps is:

	R = Q - s

with typical value R = 4.6 - 0.46 = 4.1.

In one sweep representing a duration of time t, the probability
P of a given C->T substitution is then Rt, providing t << R.

The sweep duration is then determined by f = P = Rt, hence:

	t = f/R

We expect R ~ 5, and choose f = 0.1 as a reasonable default;
this gives t = 0.02.

If the branch length b <~ t, then this will result in only a
single sweep and we are back to the problem with duplications.
We therefore set a minimum number m of sweeps, say 10, to ensure
a detectable age distribution over duplications. Then:

	t = min(f/R, b/m)
***/

double Pct(double x, double T);

extern double g_CGRejectProb;
extern FILE *g_fStats;

unsigned g_CpGActualSweepCount = 0;
unsigned g_CpGSubCount = 0;
unsigned g_CpGCodingAcceptCount = 0;
unsigned g_CpGCodingRejectCount = 0;
unsigned g_CpGNCCEAcceptCount = 0;
unsigned g_CpGNCCERejectCount = 0;
unsigned g_CpGNeutralAcceptCount = 0;
double g_SumProb;

bool DoCpGSweeps()
	{
	return P_MinCpGSweeps > 0;
	}

double GetCtoTIslandRate()
	{
	return P_CT_Rate;
	}

double GetCpGCtoTNonIslandRate()
	{
	double IslandRate = GetCtoTIslandRate();
	return IslandRate*P_CpG_C_to_T_Ratio;
	}

double GetCpGCtoTSweepRate()
	{
	double IslandRate = GetCtoTIslandRate();
	double NonIslandRate = GetCpGCtoTNonIslandRate();
	return NonIslandRate - IslandRate;
	}

// t = min(f/R, b/m)
double GetCpGSweepDuration(double BranchLength)
	{
	double f = P_CpGFraction;
	double R = GetCpGCtoTNonIslandRate();
	double MinSweepCount = P_MinCpGSweeps;
	return min(f/R, BranchLength/MinSweepCount);
	}

static inline bool IsCG(char Letter1, char Letter2)
	{
	if (Letter1 == 'C' && Letter2 == 'G')
		return true;
	return false;
	}

void Evolver::InitCpG_Sweeps()
	{
	m_CpGSweepInterval = 0;
	m_CpGSweepEventCounter = 0;
	m_DoCpGSweeps = DoCpGSweeps();
	if (m_DoCpGSweeps)
		{
		double CpGSweepDuration = GetCpGSweepDuration(m_BranchLength);
		unsigned CpGSweepCount = (unsigned) (m_BranchLength/CpGSweepDuration + 0.5);
		if (CpGSweepCount < P_MinCpGSweeps)
			CpGSweepCount = P_MinCpGSweeps;
		if (CpGSweepCount == 0)
			CpGSweepCount = 1;
		m_CpGSweepInterval = m_EventCount/CpGSweepCount;
		if (m_CpGSweepInterval == 0)
			m_CpGSweepInterval = 1;
		}
	}

bool Evolver::InCpG(gpos_t Pos) const
	{
	if (Pos == GPOS_T_MAX)
		return false;

	gpos_t PrevPos = GetPrevPos(Pos);
	if (PrevPos == GPOS_T_MAX)
		return false;

	gpos_t NextPos = GetNextPos(Pos);

	if (NextPos == GPOS_T_MAX)
		return false;

	const BaseInfo &PrevBase = GetBase(PrevPos);
	const BaseInfo &Base = GetBase(Pos);
	const BaseInfo &NextBase = GetBase(NextPos);
	
	char PrevLetter = PrevBase.GetOrientedLetter();
	char Letter = Base.GetOrientedLetter();
	char NextLetter = NextBase.GetOrientedLetter();

	if (IsCG(PrevLetter, Letter))
		return true;
	if (IsCG(Letter, NextLetter))
		return true;
	return false;
	}

void Evolver::CpG_Sweep(double Duration)
	{
	if (Duration == 0)
		return;

	++g_CpGActualSweepCount;

	double SweepRate = GetCpGCtoTSweepRate();
	double Prob = SweepRate*Duration;
	g_SumProb += Prob;

	if (Prob > 1.0)
		Die("CpG sweep, prob > 1, ratio=%.3g, duration=%.3g",
		  P_CpG_C_to_T_Ratio, Duration);

	FOR_EVBASES(p)
		{
		if (InCpG(p))
			CpGBase(p, Prob);
		}
	}

void Evolver::CpGBase(gpos_t Pos, double Prob)
	{
	BaseInfo &Base = GetBase(Pos);
	if (Base.IsIsland())
		return;

	double Subst = RandTrue(Prob);
	if (!Subst)
		return;

	++g_CpGSubCount;
	char Letter = Base.GetLetter();
	asserta(Letter == 'C' || Letter == 'G');
	char NewLetter = (Letter == 'C' ? 'T' : 'A');
	if (Base.IsConstrained())
		{
		if (Base.IsCoding())
			{
			if (!AcceptExonSubst(Pos, NewLetter))
				{
				++g_CpGCodingRejectCount;
				return;
				}
			++g_CpGCodingAcceptCount;
			}
		else 
			{
			if (!AcceptRange(Pos, 1))
				{
				++g_CpGNCCERejectCount;
				return;
				}
			++g_CpGNCCEAcceptCount;
			}
		}
	else
		++g_CpGNeutralAcceptCount;
	Base.SetLetter(NewLetter);
	}

void Evolver::GetCpGCounts(gpos_t &CpGBaseCount, gpos_t &CpGIslandBaseCount) const
	{
	CpGBaseCount = 0;
	CpGIslandBaseCount = 0;

	FOR_EVBASES(p)
		{
		if (!InCpG(p))
			continue;
		const BaseInfo &Base = m_Bases[p];
		++CpGBaseCount;
		if (Base.IsIsland())
			++CpGIslandBaseCount;
		}
	}

void Evolver::GetComposition(gpos_t LetterCounts[4], gpos_t DinucCounts[4][4]) const
	{
	for (unsigned i = 0; i < 4; ++i)
		{
		LetterCounts[i] = 0;
		memset(DinucCounts[i], 0, 4*sizeof(gpos_t));
		}

	unsigned LastLetter = UINT_MAX;
	FOR_EVBASES(p)
		{
		const BaseInfo &Base = m_Bases[p];

		extern unsigned g_LetterIndex[256];
		unsigned LetterIndex = g_LetterIndex[(unsigned char) Base.GetOrientedLetter()];
		if (LetterIndex >= 4)
			Die("Invalid letter %c", Base.GetOrientedLetter());
		++(LetterCounts[LetterIndex]);
		if (LastLetter != UINT_MAX)
			++(DinucCounts[LastLetter][LetterIndex]);
		LastLetter = LetterIndex;
		}
	}

void LogCpGStats(gpos_t AnCpGBaseCount, gpos_t AnIslandCpGCount, gpos_t AnLength,
  gpos_t EvCpGBaseCount, gpos_t EvIslandCpGCount, gpos_t EvLength,
  const gpos_t AnLetterCounts[4], const gpos_t AnDinucCounts[4][4],
  const gpos_t EvLetterCounts[4], const gpos_t EvDinucCounts[4][4])
	{
	gpos_t AnTotalLetterCount = AnLetterCounts[0] + AnLetterCounts[1] +
	  AnLetterCounts[2] + AnLetterCounts[3];
	gpos_t EvTotalLetterCount = EvLetterCounts[0] + EvLetterCounts[1] +
	  EvLetterCounts[2] + EvLetterCounts[3];

	gpos_t AnGCCount = AnLetterCounts[1] + AnLetterCounts[2];
	gpos_t EvGCCount = EvLetterCounts[1] + EvLetterCounts[2];
	double AnGCPct = Pct(AnGCCount, AnTotalLetterCount);
	double EvGCPct = Pct(EvGCCount, EvTotalLetterCount);
	double AnCPct = Pct(AnLetterCounts[1], AnTotalLetterCount);
	double AnGPct = Pct(AnLetterCounts[2], AnTotalLetterCount);
	double EvCPct = Pct(EvLetterCounts[1], EvTotalLetterCount);
	double EvGPct = Pct(EvLetterCounts[2], EvTotalLetterCount);

	Log("\n");
	Log("Letter   Ancestral   Anc.%%     Evolved    Ev.%%\n");
	Log("======  ==========  ======  ==========  ======\n");

	Log("%6s  %10"PF"  %5.1f%%  %10"PF"  %5.1f%%\n",
		"C+G",
		AnGCCount,
		AnGCPct,
		EvGCCount,
		EvGCPct);
	Log("\n");

	Log("%6c  %10"PF"  %5.1f%%  %10"PF"  %5.1f%%\n",
		'A',
		AnLetterCounts[0],
		Pct(AnLetterCounts[0], AnTotalLetterCount),
		EvLetterCounts[0],
		Pct(EvLetterCounts[0], EvTotalLetterCount));

	Log("%6c  %10"PF"  %5.1f%%  %10"PF"  %5.1f%%\n",
		'C',
		AnLetterCounts[1],
		Pct(AnLetterCounts[1], AnTotalLetterCount),
		EvLetterCounts[1],
		Pct(EvLetterCounts[1], EvTotalLetterCount));

	Log("%6c  %10"PF"  %5.1f%%  %10"PF"  %5.1f%%\n",
		'G',
		AnLetterCounts[2],
		Pct(AnLetterCounts[2], AnTotalLetterCount),
		EvLetterCounts[2],
		Pct(EvLetterCounts[2], EvTotalLetterCount));

	Log("%6c  %10"PF"  %5.1f%%  %10"PF"  %5.1f%%\n",
		'T',
		AnLetterCounts[3],
		Pct(AnLetterCounts[3], AnTotalLetterCount),
		EvLetterCounts[3],
		Pct(EvLetterCounts[3], EvTotalLetterCount));
	Log("        ==========          ==========\n");
	Log("        %10"PF"          %10"PF"\n",
	  AnTotalLetterCount, EvTotalLetterCount);

	gpos_t AnDinucCount = AnTotalLetterCount - 1;
	gpos_t EvDinucCount = EvTotalLetterCount - 1;
	Log("\n");
	Log(" Dinuc   Ancestral   Anc.%%     Evolved    Ev.%%\n");
	Log("======  ==========  ======  ==========  ======\n");
	for (unsigned i = 0; i < 16; ++i)
		{
		static char LetterToChar[4] = { 'A', 'C', 'G', 'T' };
		unsigned Letter1 = i/4;
		unsigned Letter2 = i%4;
		char c1 = LetterToChar[Letter1];
		char c2 = LetterToChar[Letter2];
		Log("    %c%c  %10"PF"  %5.1f%%  %10"PF"  %5.1f%%\n",
		  c1,
		  c2,
		  AnDinucCounts[Letter1][Letter2],
		  Pct(AnDinucCounts[Letter1][Letter2], AnDinucCount),
		  EvDinucCounts[Letter1][Letter2],
		  Pct(EvDinucCounts[Letter1][Letter2], EvDinucCount));
		}
	Log("\n");

	if (P_MinCpGSweeps == 0)
		return;

	unsigned CodingCount = g_CpGCodingAcceptCount + g_CpGCodingRejectCount;
	unsigned NCCECount = g_CpGNCCEAcceptCount + g_CpGNCCERejectCount;

	Log("\n");
	Log("CpG sweeps             %10u\n", g_CpGActualSweepCount);
	Log("\n");
	Log("Ancestral CpGs         %10"PF"  (%.1f%%, random %.1f%%)\n",
	  AnCpGBaseCount/2, Pct(AnCpGBaseCount/2, AnLength), AnGPct*AnCPct/100.0);
	Log("Evolved CpGs           %10"PF"  (%.1f%%, random %.1f%%)\n",
	  EvCpGBaseCount/2, Pct(EvCpGBaseCount/2, EvLength), EvGPct*EvCPct/100.0);
	Log("\n");
	Log("Anc. CpG island bases  %10"PF"\n", AnIslandCpGCount);
	Log("Ev. CpG island bases   %10"PF"\n", EvIslandCpGCount);
	Log("\n");
	Log("Proposed CpG subs      %10u\n", g_CpGSubCount);
	Log("CpG coding accepts     %10u  (%.1f%%)\n",
	  g_CpGCodingAcceptCount, Pct(g_CpGCodingAcceptCount, CodingCount));
	Log("CpG coding rejects     %10u\n", g_CpGCodingRejectCount);
	Log("CpG NCCE accepts       %10u  (%.1f%%)\n",
	  g_CpGNCCEAcceptCount, Pct(g_CpGNCCEAcceptCount, NCCECount));
	Log("CpG NCCE rejects       %10u\n", g_CpGNCCERejectCount);
	Log("CpG neutral accepts    %10u\n", g_CpGNeutralAcceptCount);
	if (g_CpGActualSweepCount > 0)
		Log("Avg CpG subst prob     %10.2f\n", g_SumProb/g_CpGActualSweepCount);
	Log("\n");

	if (g_fStats == 0)
		return;

	fprintf(g_fStats, "EV_ACCEPTS;CpG_Sweeps;%u\n", g_CpGActualSweepCount);

	fprintf(g_fStats, "EV_ACCEPTS;CpG_CDS_Subs;%u\n", g_CpGCodingAcceptCount);
	fprintf(g_fStats, "EV_REJECTS;CpG_CDS_Subs;%u\n", g_CpGCodingRejectCount);

	fprintf(g_fStats, "EV_ACCEPTS;CpG_NCCE_Subs;%u\n", g_CpGNCCEAcceptCount);
	fprintf(g_fStats, "EV_REJECTS;CpG_NCCE_Subs;%u\n", g_CpGNCCERejectCount);

	fprintf(g_fStats, "EV_ACCEPTS;CpG_Neutral_Subs;%u\n", g_CpGNeutralAcceptCount);
	}

/***
 StartPos           EndPos
    |                  |
                  <-d-> <-d->
    [-----CpG island---]
                  |          |
                 Pos        Pos
                >----strand---->

d <= MaxDist
***/
bool Evolver::FindNearbyCpGIsland(gpos_t Pos, bool Strand, gpos_t MaxDist,
  gpos_t &StartPos, gpos_t &Length) const
	{
	gpos_t EndPos = Pos;
	for (gpos_t i = 0; i < MaxDist; ++i)
		{
		if (EndPos == GPOS_T_MAX)
			return false;
		const BaseInfo &Base = GetBase(EndPos);
		if (Base.IsIsland())
			break;
		EndPos = GetAdjPos(EndPos, !Strand);
		}

	StartPos = Pos;
	Length = 1;
	for (;;)
		{
		gpos_t PrevPos = GetAdjPos(StartPos, !Strand);
		if (PrevPos == GPOS_T_MAX)
			return true;
		const BaseInfo &Base = GetBase(PrevPos);
		if (!Base.IsIsland())
			return true;
		++Length;
		StartPos = PrevPos;
		}
	}
