#include <map>
#include <set>
#include <vector>
#include "revfile.h"
#include "getopts.h"
#include "gff2.h"
#include "progress.h"
#include "ivtree.h"
#include "lohi.h"
#include "evomodel.h"
#include "myrand.h"

coord_t GetOverlapLength(coord_t loA, coord_t hiA, coord_t loB, coord_t hiB);

using namespace std;

static unsigned g_AltStructureCount;
static unsigned g_OverlapCount;
static unsigned g_BadStartCodonCount;
static unsigned g_BadStopCodonCount;
static unsigned g_NoCDSCount;
static unsigned g_SubrangeCount;
static unsigned g_ShortIntronCount;
static unsigned g_InFrameStopCount;
static bool g_KeepOverlaps = false;

struct ExonStuff
	{
	coord_t Start;
	coord_t Length;

	coord_t GetHi() const
		{
		return Start + Length - 1;
		}

	void LogMe(bool NL = false) const
		{
		Log("Exon start %"COORD_T_PRINTF" length %"COORD_T_PRINTF" end %"COORD_T_PRINTF,
		  Start, Length, GetHi());
		if (NL)
			Log("\n");
		}
	};

struct CDSStuff
	{
	coord_t Start;
	coord_t Length;
	int Frame;

	coord_t GetHi() const
		{
		return Start + Length - 1;
		}

	void LogMe(bool NL = false) const
		{
		Log("CDS start %"COORD_T_PRINTF" length %"COORD_T_PRINTF" end %"COORD_T_PRINTF,
		  Start, Length, GetHi());
		if (NL)
			Log("\n");
		}
	};

struct UTRStuff
	{
	coord_t Start;
	coord_t Length;

	coord_t GetHi() const
		{
		return Start + Length - 1;
		}

	void LogMe(bool NL = false) const
		{
		Log("UTR start %"COORD_T_PRINTF" length %"COORD_T_PRINTF" end %"COORD_T_PRINTF,
		  Start, Length, GetHi());
		if (NL)
			Log("\n");
		}
	};

struct GeneStuff
	{
	unsigned Index;
	bool Bad;
	string Id;
	string TranscriptId;
	coord_t lo;
	coord_t hi;
	bool Plus;
	vector<ExonStuff> Exons;
	vector<CDSStuff> CDSs;
	vector<UTRStuff> FirstUTRs;
	vector<UTRStuff> LastUTRs;

	void LogMe(bool NLStart = true) const
		{
		if (NLStart)
			Log("\n");
		Log("Gene id %s bad %c transcript %s start %"COORD_T_PRINTF" end %"COORD_T_PRINTF" %c\n",
		  Id.c_str(),
		  tof(Bad),
		  TranscriptId.c_str(),
		  lo,
		  hi,
		  pom(Plus));
		}

	void LogMeVerbose() const
		{
		LogMe();
		for (unsigned i = 0; i < SIZE(Exons); ++i)
			Exons[i].LogMe(true);
		for (unsigned i = 0; i < SIZE(FirstUTRs); ++i)
			FirstUTRs[i].LogMe(true);
		for (unsigned i = 0; i < SIZE(CDSs); ++i)
			CDSs[i].LogMe(true);
		for (unsigned i = 0; i < SIZE(LastUTRs); ++i)
			LastUTRs[i].LogMe(true);
		}
	};

static unsigned g_NextGeneIndex;
static FILE *g_fOutAnnots;
static RevFile *g_SeqRev;
static map<string, GeneStuff> g_Genes;
static set<string> g_UnknownFeatures;
static set<string> g_AlternativeTranscriptIds;
static GFFRecord2 g_Rec;
static unsigned g_GenomeIndex;
static unsigned g_ChrId;
static coord_t g_ChrLength;
static IvTree<LoHi> g_AnnotIvs;
static string g_ChrName;

bool IsStartCodon(const string &Codon);
bool IsStopCodon(const string &Codon);
void GetStopCodons(const string &Seq, vector<gpos_t> &Starts);

static vector<pair<coord_t, string> > g_SeqUpdates;

static void GetSeq(string &Seq, coord_t Pos, coord_t Length, bool Plus)
	{
	g_SeqRev->GetSubSeq2(g_GenomeIndex, g_ChrId, Seq, Pos, Length, !Plus);
	for (string::iterator p = Seq.begin(); p != Seq.end(); ++p)
		*p = (char) toupper(*p);
	}

static void LogUnknownFeatures()
	{
	Log("\n");
	Log("Unknown features: ");
	for (set<string>::const_iterator p = g_UnknownFeatures.begin();
	  p != g_UnknownFeatures.end(); ++p)
		{
		const string &Feature = *p;
		Log(" %s", Feature.c_str());
		}
	Log("\n");
	}

static void LogAltTranscripts()
	{
	Log("\n");
	Log("Alt transcripts: ");
	for (set<string>::const_iterator p = g_AlternativeTranscriptIds.begin();
	  p != g_AlternativeTranscriptIds.end(); ++p)
		{
		const string &Name= *p;
		Log(" %s", Name.c_str());
		}
	Log("\n");
	}

static void LogGenes()
	{
	Log("\n");
	Log("Index                Id  Bad        Transcript          lo          hi  +  Exons  CDSs  FUTRs  LUTRs\n");
	Log("=====  ================  ===  ================  ==========  ==========  =  =====  ====  =====  =====\n");
	for (map<string, GeneStuff>::const_iterator p = g_Genes.begin();
	  p != g_Genes.end(); ++p)
		{
		const string &Id = p->first;
		const GeneStuff &Gene = p->second;
		asserta(Gene.Id == Id);
		Log("%5u", Gene.Index);
		Log("  %16.16s", Gene.Id.c_str());
		Log("  %3c", Gene.Bad ? 'Y' : ' ');
		Log("  %16.16s", Gene.TranscriptId.c_str());
		Log("  %10"COORD_T_PRINTF, Gene.lo);
		Log("  %10"COORD_T_PRINTF, Gene.hi);
		Log("  %c", pom(Gene.Plus));
		Log("  %5u", SIZE(Gene.Exons));
		Log("  %4u", SIZE(Gene.CDSs));
		Log("  %5u", SIZE(Gene.FirstUTRs));
		Log("  %5u", SIZE(Gene.LastUTRs));
		Log("\n");
		}
	Log("\n");
	}

static bool IsBadStartCodon(coord_t Start, bool Plus)
	{
	string Codon;
	GetSeq(Codon, Start, 3, Plus);
	if (!IsStartCodon(Codon))
		{
		g_Rec.LogMe();
		++g_BadStartCodonCount;
		Warning("start_codon record points to '%s', expected ATG", Codon.c_str());
		return true;
		}
	return false;
	}

static void SetUTRs(GeneStuff &Gene)
	{
	asserta(!Gene.CDSs.empty());
	Gene.FirstUTRs.clear();
	Gene.LastUTRs.clear();

	const CDSStuff &FirstCDS = Gene.CDSs.front();
	const CDSStuff &LastCDS = Gene.CDSs.back();

	const unsigned ExonCount = SIZE(Gene.Exons);
	for (unsigned i = 0; i < ExonCount; ++i)
		{
		const ExonStuff &Exon = Gene.Exons[i];

	// Exon is 100% CDS?
		if (Exon.Start >= FirstCDS.Start && Exon.GetHi() <= LastCDS.GetHi())
			continue;

	// Exon is 100% leading UTR?
		else if (Exon.GetHi() < FirstCDS.Start)
			{
			UTRStuff FirstUTR;
			FirstUTR.Start = Exon.Start;
			FirstUTR.Length = Exon.Length;
			Gene.FirstUTRs.push_back(FirstUTR);
			}

	// Exon is UTR-CDS-UTR?
		else if (Exon.Start < FirstCDS.Start && Exon.GetHi() > FirstCDS.GetHi())
			{
			asserta(Gene.CDSs.size() == 1);
			asserta(FirstCDS.Start == LastCDS.Start && FirstCDS.Length == LastCDS.Length);

			UTRStuff FirstUTR;
			FirstUTR.Start = Exon.Start;
			FirstUTR.Length = FirstCDS.Start - Exon.Start;
			Gene.FirstUTRs.push_back(FirstUTR);

			UTRStuff LastUTR;
			LastUTR.Start = LastCDS.GetHi() + 1;
			LastUTR.Length = Exon.GetHi() - LastCDS.GetHi();
			Gene.LastUTRs.push_back(LastUTR);
			}

	// Exon is UTR-CDS?
		else if (Exon.Start < FirstCDS.Start && Exon.GetHi() >= FirstCDS.Start)
			{
			asserta(Exon.GetHi() == FirstCDS.GetHi());
			UTRStuff FirstUTR;
			FirstUTR.Start = Exon.Start;
			FirstUTR.Length = FirstCDS.Start - Exon.Start;
			Gene.FirstUTRs.push_back(FirstUTR);
			}

	// Exon is CDS-UTR?
		else if (LastCDS.Start >= Exon.Start && LastCDS.GetHi() < Exon.GetHi())
			{
			UTRStuff LastUTR;
			LastUTR.Start = LastCDS.GetHi() + 1;
			LastUTR.Length = Exon.GetHi() - LastCDS.GetHi();
			Gene.LastUTRs.push_back(LastUTR);
			}

	// Exon is 100% trailing UTR?
		else if (Exon.Start > LastCDS.GetHi())
			{
			UTRStuff LastUTR;
			LastUTR.Start = Exon.Start;
			LastUTR.Length = Exon.Length;
			Gene.LastUTRs.push_back(LastUTR);
			}
		else
			{
			Log("First ");
			FirstCDS.LogMe(true);
			Log("Last ");
			LastCDS.LogMe(true);
			Exon.LogMe();
			Die("SetUTRs: logic error");
			}
		}

// Validate Exons = CDSs + UTRs
	coord_t L_FUTR = 0;
	coord_t L_LUTR = 0;
	coord_t L_Exon = 0;
	coord_t L_CDS = 0;
	const unsigned FirstUTRCount = SIZE(Gene.FirstUTRs);
	const unsigned LastUTRCount = SIZE(Gene.LastUTRs);
	const unsigned CDSCount = SIZE(Gene.CDSs);
	for (unsigned i = 0; i < FirstUTRCount; ++i)
		L_FUTR += Gene.FirstUTRs[i].Length;
	for (unsigned i = 0; i < LastUTRCount; ++i)
		L_LUTR += Gene.LastUTRs[i].Length;
	for (unsigned i = 0; i < ExonCount; ++i)
		L_Exon += Gene.Exons[i].Length;
	for (unsigned i = 0; i < CDSCount; ++i)
		L_CDS += Gene.CDSs[i].Length;

	coord_t Sum = L_CDS + L_FUTR + L_LUTR;
	if (L_Exon != Sum)
		{
		Gene.LogMeVerbose();
		Die("UTR length fail L_CDS=%"COORD_T_PRINTF" L_FUTR=%"COORD_T_PRINTF" L_LUTR=%"COORD_T_PRINTF
		  " Sum=%"COORD_T_PRINTF" L_Exon=%"COORD_T_PRINTF,
		  L_CDS, L_FUTR, L_LUTR, Sum, L_Exon);
		}
	}

static void SetUTRs()
	{
	for (map<string, GeneStuff>::iterator p = g_Genes.begin();
	  p != g_Genes.end(); ++p)
		{
		GeneStuff &Gene = p->second;
		if (Gene.Bad)
			continue;
		SetUTRs(Gene);
		}
	}

static void SetLoHi(GeneStuff &Gene)
	{
	if (Gene.Exons.empty())
		{
		Gene.LogMe();
		Warning("Gene has no exons");
		Gene.Bad = true;
		return;
		}
	const ExonStuff &FirstExon = Gene.Exons.front();
	Gene.lo = FirstExon.Start;
	Gene.hi = FirstExon.GetHi();

	const unsigned ExonCount = SIZE(Gene.Exons);
	const unsigned CDSCount = SIZE(Gene.CDSs);

	for (unsigned i = 1; i < ExonCount; ++i)
		{
		const ExonStuff &Exon = Gene.Exons[i];
		Gene.lo = min(Gene.lo, Exon.Start);
		Gene.hi = max(Gene.hi, Exon.GetHi());
		}

	for (unsigned i = 1; i < CDSCount; ++i)
		{
		const CDSStuff &CDS = Gene.CDSs[i];
		Gene.lo = min(Gene.lo, CDS.Start);
		Gene.hi = max(Gene.hi, CDS.GetHi());
		}
	}

static void SetLoHis()
	{
	for (map<string, GeneStuff>::iterator p = g_Genes.begin();
	  p != g_Genes.end(); ++p)
		{
		GeneStuff &Gene = p->second;
		SetLoHi(Gene);
		}
	}

static bool IsBadCDS(const GeneStuff &Gene, unsigned CDSIndex)
	{
	asserta(CDSIndex < SIZE(Gene.CDSs));
	const CDSStuff &CDS = Gene.CDSs[CDSIndex];

	const unsigned ExonCount = SIZE(Gene.Exons);
	const unsigned CDSCount = SIZE(Gene.CDSs);

	unsigned ExonIndex = UINT_MAX;
	for (ExonIndex = 0; ExonIndex < ExonCount; ++ExonIndex)
		{
		const ExonStuff &Exon = Gene.Exons[ExonIndex];
		if (CDS.Start >= Exon.Start && CDS.GetHi() <= Exon.GetHi())
			break;
		}

	if (ExonIndex == UINT_MAX)
		{
		CDS.LogMe(true);
		Warning("CDS not inside exon");
		return true;
		}

	const ExonStuff &Exon = Gene.Exons[ExonIndex];

	bool InternalCDS = CDSIndex > 0 && CDSIndex < CDSCount - 1;
	if (InternalCDS)
		{
		if (CDS.Start != Exon.Start || CDS.GetHi() != Exon.GetHi())
			{
			Exon.LogMe(true);
			CDS.LogMe(true);
			Warning("Internal CDS does not exactly match exon");
			return true;
			}
		}

	bool FirstCDS = (CDSIndex == 0);
	bool LastCDS = (CDSIndex == CDSCount - 1);
	if (FirstCDS && !LastCDS)
		{
		if (CDS.GetHi() != Exon.GetHi())
			{
			Exon.LogMe(true);
			CDS.LogMe(true);
			Warning("First CDS does not end at exon end");
			return true;
			}
		}

	if (LastCDS && !FirstCDS)
		{
		if (CDS.Start != Exon.Start)
			{
			Exon.LogMe(true);
			CDS.LogMe(true);
			Warning("Last CDS does not start at exon start");
			return true;
			}
		}

	return false;
	}

static bool IsBadGene(const GeneStuff &Gene)
	{
	if (SIZE(Gene.CDSs) == 0)
		{
		Gene.LogMe();
		Log("Gene has no CDS\n");
		++g_NoCDSCount;
		return true;
		}

	if (SIZE(Gene.CDSs) > SIZE(Gene.Exons))
		{
		Gene.LogMe();
		Warning("Gene has more CDSs than exons");
		return true;
		}

	coord_t StartCodonPos = COORD_T_MAX;
	coord_t StopCodonPos = COORD_T_MAX;

	StartCodonPos = Gene.CDSs.front().Start;
	StopCodonPos = Gene.CDSs.back().GetHi() - 2;

	if (!Gene.Plus)
		swap(StartCodonPos, StopCodonPos);

	string StartCodon;
	string StopCodon;
	GetSeq(StartCodon, StartCodonPos, 3, Gene.Plus);
	GetSeq(StopCodon, StopCodonPos, 3, Gene.Plus);

	if (!IsStartCodon(StartCodon))
		{
		Gene.LogMe();
		Log("CDS starts with %s, expected ATG\n", StartCodon.c_str());
		++g_BadStartCodonCount;
		return true;
		}

	if (!IsStopCodon(StopCodon))
		{
		Gene.LogMe();
		Log("CDS ends with %s, expected stop codon\n", StopCodon.c_str());
		++g_BadStopCodonCount;
		return true;
		}
	
	const unsigned N = SIZE(Gene.CDSs);
	string CDSSeq;
	for (unsigned i = 0; i < N; ++i)
		{
		if (IsBadCDS(Gene, i))
			return true;
		const CDSStuff &CDS = Gene.CDSs[i];
		string CurrCDSSeq;
		GetSeq(CurrCDSSeq, CDS.Start, CDS.Length, Gene.Plus);
		if (Gene.Plus)
			CDSSeq += CurrCDSSeq;
		else
			CDSSeq = CurrCDSSeq + CDSSeq;
		}

	if (CDSSeq.size() < 3)
		{
		Gene.LogMe();
		Log("CDS < 3 bases");
		return true;
		}
	if (CDSSeq.size()%3 != 0)
		{
		Gene.LogMe();
		Log("CDS not multiple of 3\n");
		return true;
		}

	StartCodon = CDSSeq.substr(0, 3);
	StopCodon = CDSSeq.substr(CDSSeq.size() - 3, 3);

	asserta(IsStartCodon(StartCodon));
	asserta(IsStopCodon(StopCodon));
	
	unsigned ExonCount = SIZE(Gene.Exons);
	for (unsigned ExonIndex = 0; ExonIndex + 1 < ExonCount; ++ExonIndex)
		{
		const ExonStuff &Exon = Gene.Exons[ExonIndex];
		const ExonStuff &NextExon = Gene.Exons[ExonIndex + 1];
		unsigned IntronLength = (unsigned) (NextExon.Start - Exon.GetHi());
		if (IntronLength < P_MinIntronLength)
			{
			Gene.LogMe();
			Log("Intron %u length %u < min %u\n",
			  ExonIndex, IntronLength, P_MinIntronLength);
			++g_ShortIntronCount;
			return true;
			}
		}

// Strip stop codon
	gpos_t CDSLen = (gpos_t) CDSSeq.size() - 3;
	CDSSeq.resize(CDSLen);

	vector<gpos_t> StopCodonStarts;
	GetStopCodons(CDSSeq, StopCodonStarts);
	if (!StopCodonStarts.empty())
		{
		++g_InFrameStopCount;
		Gene.LogMe();
		Log("Stop codons in CDS\n");
		return true;
		}

	return false;
	}

static unsigned DiscardBadGenes()
	{
	vector<string> BadGeneIds;
	unsigned Count = 0;
	for (map<string, GeneStuff>::iterator p = g_Genes.begin();
	  p != g_Genes.end(); ++p)
		{
		GeneStuff &Gene = p->second;
		if (IsBadGene(Gene))
			{
			Gene.Bad = true;
			++Count;
			}
		}

	//const unsigned N = SIZE(BadGeneIds);
	//for (unsigned i = 0; i < N; ++i)
	//	g_Genes.erase(BadGeneIds[i]);

	return Count;
	}

static GeneStuff &GetGeneByIndex(unsigned GeneIndex)
	{
	for (map<string, GeneStuff>::iterator p = g_Genes.begin();
	  p != g_Genes.end(); ++p)
		{
		GeneStuff &Gene = p->second;
		if (Gene.Index == GeneIndex)
			return Gene;
		}
	Die("GetGeneByIndex(%u): not found", GeneIndex);
	UNREACHABLE_RETURN(*(GeneStuff *) 0);
	}

static unsigned DiscardOverlappingGenes()
	{
	IvTree<LoHi> GeneIvs;

	unsigned BadCount = 0;
	//vector<string> BadGeneIds;
	for (map<string, GeneStuff>::iterator p = g_Genes.begin();
	  p != g_Genes.end(); ++p)
		{
		GeneStuff &Gene = p->second;

		pair<coord_t, coord_t> Iv(Gene.lo, Gene.hi);
		vector<LoHi> Results;
		GeneIvs.GetOverlaps(Iv, Results);
		const GeneStuff *GoodOv = 0;
		for (vector<LoHi>::const_iterator p = Results.begin();
		  p != Results.end(); ++p)
			{
			const LoHi &LH = *p;
			const GeneStuff &OtherGene = GetGeneByIndex(LH.Index);
			if (!OtherGene.Bad)
				{
				GoodOv = &OtherGene;
				break;
				}
			}
		if (GoodOv != 0)
			{
			Log("\n");
			Gene.LogMe(false);
			const GeneStuff &OtherGene = *GoodOv;
			OtherGene.LogMe(false);
			if (OtherGene.lo == Gene.lo && OtherGene.hi == Gene.hi)
				{
				++g_AltStructureCount;
				Log("Gene %s is alternative structure for %s\n",
				  Gene.Id.c_str(), OtherGene.Id.c_str());
				Gene.Bad = true;
				}
			else if ((OtherGene.lo >= Gene.lo && OtherGene.hi <= Gene.hi) ||
			  ((Gene.lo >= OtherGene.lo && Gene.hi <= OtherGene.hi)))
				{
				++g_SubrangeCount;
				Log("Gene %s is subrange for %s\n",
				  Gene.Id.c_str(), OtherGene.Id.c_str());
				if (!g_KeepOverlaps)
					Gene.Bad = true;
				}
			else
				{
				++g_OverlapCount;
				coord_t Ov = GetOverlapLength(Gene.lo, Gene.hi,
				  OtherGene.lo, OtherGene.hi);
				double AvgGeneLength =
				  (Gene.hi - Gene.lo + OtherGene.hi - OtherGene.lo + 2)/2.0;
				double p = Pct(Ov, AvgGeneLength);
				Log("Gene %s (%u-%u) overlaps %s (%u-%u), %u bases, %.1f%%\n",
				  Gene.Id.c_str(),
				  (unsigned) Gene.lo,
				  (unsigned) Gene.hi,
				  OtherGene.Id.c_str(),
				  (unsigned) OtherGene.lo,
				  (unsigned) OtherGene.hi,
				  (unsigned) Ov,
				  p);
				if (!g_KeepOverlaps)
					Gene.Bad = true;
				}
			++BadCount;
			//BadGeneIds.push_back(Gene.Id);
			continue;
			}

		LoHi LH;
		if (LH.Lo > 2)
			LH.Lo = Gene.lo - 2;
		LH.Hi = Gene.hi + 2;
		LH.Index = Gene.Index;
		GeneIvs.Add(LH);
		}

	//const unsigned N = SIZE(BadGeneIds);
	//for (unsigned i = 0; i < N; ++i)
	//	{
	//	const string &Id = BadGeneIds[i];
	//	g_Genes.erase(Id);
	//	}

	return BadCount;
	}

static GeneStuff &GetGene()
	{
	string GeneId;
	GFFGetRequiredAttr(g_Rec.Attrs, "gene_id", GeneId);
	map<string, GeneStuff>::iterator p = g_Genes.find(GeneId);
	if (p == g_Genes.end())
		{
		GeneStuff Gene;
		Gene.Index = g_NextGeneIndex++;
		Gene.Id = GeneId;
		Gene.Plus = (g_Rec.Strand == '+');
		Gene.Bad = false;
		g_Genes[GeneId] = Gene;
		return g_Genes[GeneId];
		}
	return p->second;
	}

static void AddExon(GeneStuff &Gene, coord_t Start, coord_t Length)
	{
	ExonStuff Exon;
	Exon.Start = Start;
	Exon.Length = Length;
	Gene.Exons.push_back(Exon);
	}

static void AddCDS(GeneStuff &Gene, coord_t Start, coord_t Length, int Frame)
	{
	CDSStuff CDS;
	CDS.Start = Start;
	CDS.Length = Length;
	CDS.Frame = Frame;
	Gene.CDSs.push_back(CDS);
	}

static void DoAnnot()
	{
	if (g_Rec.Feature != "start_codon" && g_Rec.Feature != "exon" &&
	  g_Rec.Feature != "CDS")
		{
		Log("Unknown feature: ");
		g_Rec.LogMe(true);
		g_UnknownFeatures.insert(g_Rec.Feature);
		return;
		}

	if (g_Rec.Start == 0)
		{
		g_Rec.LogMe();
		Warning("Start=0");
		return;
		}

	if (g_Rec.End < g_Rec.Start)
		{
		g_Rec.LogMe();
		Warning("End < Start");
		return;
		}

	if (g_Rec.End > g_ChrLength)
		{
		g_Rec.LogMe();
		Warning("End > ChrLength");
		return;
		}

	coord_t Start = g_Rec.Start - 1;
	coord_t Length = g_Rec.GetLength();
	string TranscriptId;
	GFFGetRequiredAttr(g_Rec.Attrs, "transcript_id", TranscriptId);

	GeneStuff &Gene = GetGene();
	if (Gene.TranscriptId == "")
		Gene.TranscriptId = TranscriptId;
	else
		{
		if (TranscriptId != Gene.TranscriptId)
			{
			g_AlternativeTranscriptIds.insert(TranscriptId);
			return;
			}
		}

	bool Plus = (g_Rec.Strand == '+');
	if (Plus != Gene.Plus)
		{
		g_Rec.LogMe();
		Warning("Gene on both + and - strand");
		return;
		}

	if (g_Rec.Feature == "start_codon")
		IsBadStartCodon(Start, g_Rec.Strand == '+');
	else if (g_Rec.Feature == "exon")
		AddExon(Gene, Start, Length);
	else if (g_Rec.Feature == "CDS")
		AddCDS(Gene, Start, Length, g_Rec.Frame);
	}

static void WriteFeature(const char *FeatureName, coord_t Start, coord_t Length,
  unsigned GeneIndex, bool Plus, int Frame = INT_MAX)
	{
	coord_t lo = Start;
	coord_t hi = Start + Length - 1;

	vector<LoHi> Results;
	pair<coord_t, coord_t> Iv(lo, hi);

	if (!g_KeepOverlaps)
		{
		g_AnnotIvs.GetOverlaps(Iv, Results);

		if (!Results.empty())
			{
			Log("\n");
			Log("Feature %s start %"COORD_T_PRINTF" length %"COORD_T_PRINTF" gene %u\n",
			  FeatureName, Start, Length, GeneIndex);
			Log("Overlaps:\n");
			const unsigned N = SIZE(Results);
			for (unsigned i = 0; i < N; ++i)
				{
				const LoHi &Result = Results[i];
				Log("lo=%"COORD_T_PRINTF" hi=%"COORD_T_PRINTF" len=%"COORD_T_PRINTF"\n",
				  Result.Lo, Result.Hi, Result.GetLength());
				}
			Die("Overlapping annots");
			}
		}

	LoHi LH;
	LH.Lo = lo;
	LH.Hi = hi;

	g_AnnotIvs.Add(LH);

	GFFRecord2 Rec;

	Rec.SeqName = g_ChrName;
	Rec.Source = "evo";
	Rec.Start = lo + 1;
	Rec.End = hi + 1;
	Rec.Feature = FeatureName;
	Rec.Score = 0;
	Rec.Strand = pom(Plus);
	Rec.Frame = Frame;

	char Attrs[64];
	sprintf(Attrs, "gene_index %u;", GeneIndex);
	Rec.Attrs = Attrs;

	WriteGFFRec(g_fOutAnnots, Rec);
	}

static void WriteUTR(const UTRStuff &UTR, unsigned GeneIndex, bool Plus)
	{
	WriteFeature("UTR", UTR.Start, UTR.Length, GeneIndex, Plus);
	}

static void WriteCDS(const CDSStuff &CDS, unsigned GeneIndex, bool Plus)
	{
	WriteFeature("CDS", CDS.Start, CDS.Length, GeneIndex, Plus, CDS.Frame);
	}

static void WriteAnnots(const GeneStuff &Gene, unsigned GeneIndex)
	{
	const unsigned FirstUTRCount = SIZE(Gene.FirstUTRs);
	const unsigned LastUTRCount = SIZE(Gene.LastUTRs);
	const unsigned CDSCount = SIZE(Gene.CDSs);

	for (unsigned i = 0; i < FirstUTRCount; ++i)
		WriteUTR(Gene.FirstUTRs[i], GeneIndex, Gene.Plus);

	for (unsigned i = 0; i < CDSCount; ++i)
		WriteCDS(Gene.CDSs[i], GeneIndex, Gene.Plus);

	for (unsigned i = 0; i < LastUTRCount; ++i)
		WriteUTR(Gene.LastUTRs[i], GeneIndex, Gene.Plus);
	}

static void WriteAnnots()
	{
	unsigned GeneIndex = 0;
	for (map<string, GeneStuff>::iterator p = g_Genes.begin();
	  p != g_Genes.end(); ++p)
		{
		const GeneStuff &Gene = p->second;
		if (Gene.Bad)
			continue;
		WriteAnnots(Gene, GeneIndex);
		++GeneIndex;
		}
	}

void CvtAnnots()
	{
	const char *InAnnotFileName = RequiredValueOpt("cvtannots");
	const char *OutAnnotFileName = RequiredValueOpt("out");
	const char *InSeqFileName = RequiredValueOpt("seq");
	g_GenomeIndex = 0;
	IntOpt("genix", &g_GenomeIndex);
	const char *ChrName = ValueOpt("chrname");
	g_KeepOverlaps = FlagOpt("keepov");

	RevFile SeqRev;
	SeqRev.Open(InSeqFileName);
	g_SeqRev = &SeqRev;

	if (ChrName != 0)
		g_ChrId = SeqRev.GetChrId(g_GenomeIndex, ChrName);

	g_ChrLength = SeqRev.GetChrLength(g_GenomeIndex, g_ChrId);

	FILE *fInAnnots = OpenStdioFile(InAnnotFileName);
	g_fOutAnnots = CreateStdioFile(OutAnnotFileName);

	coord_t LastEnd = 0;
	while (GetNextGFFRecord(fInAnnots, g_Rec))
		{
		if (g_ChrName == "")
			g_ChrName = g_Rec.SeqName;
		else if (g_Rec.SeqName != g_ChrName)
			Die("Two sequences in annotation file (%s, %s)",
			  g_ChrName.c_str(), g_Rec.SeqName.c_str());
		if (g_Rec.Start < LastEnd)
			Die("Input records not sorted");

		DoAnnot();
		}

	SetLoHis();
	DiscardBadGenes();
	DiscardOverlappingGenes();
	SetUTRs();

	if (!g_UnknownFeatures.empty())
		{
		Warning("%u unknown GTF feature types", SIZE(g_UnknownFeatures));
		LogUnknownFeatures();
		}

	if (!g_AlternativeTranscriptIds.empty())
		{
		ProgressNoRes("%u alternative transcripts discarded",
		  SIZE(g_AlternativeTranscriptIds));
		LogAltTranscripts();
		}
	else
		ProgressNoRes("No alternative transcripts");

	LogGenes();

	unsigned GoodGeneCount = 0;
	for (map<string, GeneStuff>::const_iterator p = g_Genes.begin();
	  p != g_Genes.end(); ++p)
		{
		const GeneStuff &Gene = p->second;
		if (!Gene.Bad)
			++GoodGeneCount;
		}

	ProgressNoRes("%5u  gene ids", SIZE(g_Genes));
	ProgressNoRes("%5u  no CDS", g_NoCDSCount);
	ProgressNoRes("%5u  alternative structures discarded", g_AltStructureCount);
	ProgressNoRes("%5u  subranges discarded", g_SubrangeCount);
	ProgressNoRes("%5u  overlapping structures discarded", g_OverlapCount);
	ProgressNoRes("%5u  bad start codons", g_BadStartCodonCount);
	ProgressNoRes("%5u  bad stop codons", g_BadStopCodonCount);
	ProgressNoRes("%5u  short introns", g_ShortIntronCount);
	ProgressNoRes("%5u  good genes", GoodGeneCount);

	WriteAnnots();

	SeqRev.Close();
	}
