#include <stdio.h>
#include <string>
#include <sstream>
#include <set>
#include <cmath>
#include "utils.h"
#include "revfile.h"
#include "progress.h"
#include "getopts.h"

#define BUFFER_SIZE 65535

static char ApprovedDNA[256] = {0};

unsigned atouint(const std::string &s)
{
	return strtoul(s.c_str(), NULL, 10);
}

/* Grabs genome info from two revfiles */
void GetGenomesVector(const char *RevFile1, const char *RevFile2,
  std::vector<GenInfo> &Genomes)
{
	Genomes.clear();
	Genomes.reserve(2);

	RevFile InRev1;
	InRev1.Open(RevFile1);
	if (InRev1.GetGenomeCount() != 1)
		Die("File %s does not contain a single genome", RevFile1);
	Genomes.push_back(InRev1.GetGenomeInfo(0));
	InRev1.Close();

	RevFile InRev2;
	InRev2.Open(RevFile2);
	if (InRev2.GetGenomeCount() != 1)
		Die("File %s does not contain a single genome", RevFile2);
	Genomes.push_back(InRev2.GetGenomeInfo(0));
	InRev2.Close();
}

/* Converts blast output to rev */
void Blast2Rev(const char *BlastFileName, const char *OutRevFileName,
  const char *InRevFileName1, const char *InRevFileName2,
  unsigned minperc, unsigned minlength, bool DoRename)
{
	FILE *f = OpenStdioFile(BlastFileName);

	std::vector<GenInfo> Genomes;
	GetGenomesVector(InRevFileName1, InRevFileName2, Genomes);

	RevFile OutFile;
	OutFile.Create(OutRevFileName, Genomes);

	Progress("Converting blast file '%s' to rev file '%s'",
	  BlastFileName, OutRevFileName);

	off64_t StepCount = GetStdioFileSize(f);
	float StepScale = 1.0;
	if (StepCount > INT_MAX)
	{
		StepScale = ((float)INT_MAX) / StepCount;
		StepCount = INT_MAX;
	}

	ProgressStart("Processing...");

	char buf[BUFFER_SIZE];
	size_t i = 0, N = 0;

	unsigned long int LineNumber = 1;
	unsigned long int HSPsWritten = 0, HSPsIgnored = 0;
	int         CurrentFieldIndex;
	std::string CurrentFieldValue, CurrentChr1, CurrentChr2;
	unsigned    CurrentChrID1 = 0, CurrentChrID2 = 0;
	unsigned    PercentID = 0;

	/* Construct a pairwise AI */
	AlnInfo AI;
	AI.Attrs.push_back(ALN_ATTR_AlnLength);
	AI.Values.push_back(0);
	GCSeg Seg;
	Seg.GenomeIndex = 0;
	Seg.Plus = true;
	AI.Segs.push_back(Seg);
	Seg.GenomeIndex = 1;
	AI.Segs.push_back(Seg);
	
	CurrentFieldIndex = 1;
	CurrentFieldValue.reserve(255);

	while ((N = fread(buf, 1, BUFFER_SIZE, f)))
	{
		i = 0;
		ProgressStep((int)(GetStdioFilePos(f)*StepScale), (int)StepCount);

		while (i != N)
		{
			if (buf[i] == '\t')
			{
				switch (CurrentFieldIndex)
				{
					case 1:
					{
						if (DoRename)
						{
							std::string num = CurrentFieldValue.substr(CurrentFieldValue.find_last_of("."));
							if (num.size() <= 1)
								Die("Invalid chromosome name '%s'", CurrentFieldValue.c_str());
							std::stringstream num_stream(num.substr(1));

							num_stream >> CurrentChrID1;
							CurrentChr1 = Genomes[0].GetChrName(CurrentChrID1);
						}
						else
						{
							CurrentChr1 = CurrentFieldValue;
							CurrentChrID1 = Genomes[0].GetChrId(CurrentChr1);
						}
						break;
					}
					case 2:
					{
						if (DoRename)
						{
							std::string num = CurrentFieldValue.substr(CurrentFieldValue.find_last_of("."));
							if (num.size() <= 1)
								Die("Invalid chromosome name '%s'", CurrentFieldValue.c_str());
							std::stringstream num_stream(num.substr(1));

							num_stream >> CurrentChrID2;
							CurrentChr2 = Genomes[1].GetChrName(CurrentChrID2);
						}
						else
						{
							CurrentChr2 = CurrentFieldValue;
							CurrentChrID2 = Genomes[1].GetChrId(CurrentChr2);
						}
						break;
					}
					case 6:
					{
						AI.Score = (float) atof(CurrentFieldValue.c_str());
						break;
					}
					case 7:
					{
						AI.Values[0] = atouint(CurrentFieldValue);
						break;
					}
					case 11:
					{
						PercentID = atouint(CurrentFieldValue);
						break;
					}
					case 17:
					{
						AI.Segs[1].Plus = (atoi(CurrentFieldValue.c_str()) != -1);
						break;
					}
					case 18:
					{
						if (AI.Segs[1].Plus)
							AI.Segs[0].lo = atouint(CurrentFieldValue) - 1;
						else
							AI.Segs[0].hi = atouint(CurrentFieldValue) - 1;
						break;
					}
					case 19:
					{
						if (AI.Segs[1].Plus)
							AI.Segs[0].hi = atouint(CurrentFieldValue) - 1;
						else
							AI.Segs[0].lo = atouint(CurrentFieldValue) - 1;
						break;
					}
					case 21:
					{
						AI.Segs[1].lo = atouint(CurrentFieldValue) - 1;
						break;
					}
				}
				CurrentFieldIndex++;
				CurrentFieldValue.clear();
			}
			else if (buf[i] == '\n')
			{
				if (CurrentFieldIndex != 22)
				{
					Die("Line %lu is of wrong size\n", LineNumber);
				}

				AI.Segs[1].hi = atouint(CurrentFieldValue) - 1;

				if ((PercentID >= minperc) && (AI.Values[0] >= minlength))
				{
					AI.Segs[0].ChrId = CurrentChrID1;
					AI.Segs[1].ChrId = CurrentChrID2;
					OutFile.AddAlnInfo(AI);
					HSPsWritten++;
				}
				else
					HSPsIgnored++;

				LineNumber++;
				CurrentFieldIndex = 1;
				CurrentFieldValue.clear();
			}
			else
				CurrentFieldValue.push_back(buf[i]);
			++i;
		}
	}

	OutFile.Close();
	ProgressDone();
	Progress("%lu HSPs written, %lu HSPs ignored\n", HSPsWritten, HSPsIgnored);
}

/* Converts lastz specific output to rev */
void Lastz2Rev(const char *BlastFileName, const char *OutRevFileName,
  const char *InRevFileName1, const char *InRevFileName2,
  unsigned minperc, unsigned minlength, bool DoRename)
{
	FILE *f = OpenStdioFile(BlastFileName);

	std::vector<GenInfo> Genomes;
	GetGenomesVector(InRevFileName1, InRevFileName2, Genomes);

	RevFile OutFile;
	OutFile.Create(OutRevFileName, Genomes);

	Progress("Converting lastz file '%s' to rev file '%s'", BlastFileName, OutRevFileName);

	off64_t StepCount = GetStdioFileSize(f);
	float StepScale = 1.0;
	if (StepCount > INT_MAX)
	{
		StepScale = ((float)INT_MAX) / StepCount;
		StepCount = INT_MAX;
	}

	ProgressStart("Processing...");

	char buf[BUFFER_SIZE];
	size_t i = 0, N = 0;

	unsigned long int LineNumber = 1;
	unsigned long int HSPsWritten = 0, HSPsIgnored = 0, LinesIgnored = 0;
	int         CurrentFieldIndex;
	std::string CurrentFieldValue, CurrentChr1, CurrentChr2;
	unsigned    CurrentChrID1 = 0, CurrentChrID2 = 0;
	unsigned    PercentID = 0;

	/* Construct a pairwise AI */
	AlnInfo AI;
	/*
	AI.Attrs.push_back(ALN_ATTR_AlnLength);
	AI.Values.push_back(0);
	*/
	GCSeg Seg;
	Seg.GenomeIndex = 0;
	Seg.Plus = true;
	AI.Segs.push_back(Seg);
	Seg.GenomeIndex = 1;
	AI.Segs.push_back(Seg);
	
	std::vector<std::string> Rows;
	Rows.resize(2);

	CurrentFieldIndex = 1;
	CurrentFieldValue.reserve(255);

	bool StartOfLine = true;
	bool IgnoreLine = false;

	while ((N = fread(buf, 1, BUFFER_SIZE, f)))
	{
		i = 0;
		ProgressStep((int)(GetStdioFilePos(f)*StepScale), (int)StepCount);

		while (i != N)
		{
			if (StartOfLine && (buf[i] == '#'))
			{
				IgnoreLine = true;
				LinesIgnored++;
			}

			StartOfLine = false;

			if (!IgnoreLine)
			{
				if (buf[i] == '\t')
				{
					switch (CurrentFieldIndex)
					{
						case 1:
						{
							AI.Score = (float) atof(CurrentFieldValue.c_str());
							break;
						}
						case 2:
						{
							if (DoRename)
							{
								std::string num = CurrentFieldValue.substr(CurrentFieldValue.find_last_of("."));
								if (num.size() <= 1)
									Die("Invalid chromosome name '%s'", CurrentFieldValue.c_str());
								std::stringstream num_stream(num.substr(1));

								num_stream >> CurrentChrID1;
								CurrentChr1 = Genomes[0].GetChrName(CurrentChrID1);
							}
							else
							{
								CurrentChr1 = CurrentFieldValue;
								CurrentChrID1 = Genomes[0].GetChrId(CurrentChr1);
							}
							break;
						}
						case 3:
						{
							if (CurrentFieldValue[0] != '+')
								Die("This program only supports '+' sign for first genome");
							break;
						}
						case 5:
						{
							AI.Segs[0].lo = atouint(CurrentFieldValue);
							break;
						}
						case 6:
						{
							AI.Segs[0].hi = atouint(CurrentFieldValue) - 1;
							break;
						}
						case 7:
						{
							if (DoRename)
							{
								std::string num = CurrentFieldValue.substr(CurrentFieldValue.find_last_of("."));
								if (num.size() <= 1)
									Die("Invalid chromosome name '%s'", CurrentFieldValue.c_str());
								std::stringstream num_stream(num.substr(1));

								num_stream >> CurrentChrID2;
								CurrentChr2 = Genomes[1].GetChrName(CurrentChrID2);
							}
							else
							{
								CurrentChr2 = CurrentFieldValue;
								CurrentChrID2 = Genomes[1].GetChrId(CurrentChr2);
							}
							break;
						}
						/*
						case 7:
						{
							AI.Values[0] = atouint(CurrentFieldValue);
							break;
						}
						*/
						case 8:
						{
							if ((CurrentFieldValue[0] != '-') && (CurrentFieldValue[0] != '+'))
								Die("Invalid sign value: %s", CurrentFieldValue.c_str());
							AI.Segs[1].Plus = (CurrentFieldValue[0] == '+');
							break;
						}
						case 10:
						{
							if (AI.Segs[1].Plus)
								AI.Segs[1].lo = atouint(CurrentFieldValue);
							else
								AI.Segs[1].hi = Genomes[1].GetChrInfo(CurrentChrID2).GetLength() - atouint(CurrentFieldValue) - 1;
							break;
						}
						case 11:
						{
							if (AI.Segs[1].Plus)
								AI.Segs[1].hi = atouint(CurrentFieldValue) - 1;
							else
								AI.Segs[1].lo = Genomes[1].GetChrInfo(CurrentChrID2).GetLength() - atouint(CurrentFieldValue);
							break;
						}
						case 13:
						{
							PercentID = (unsigned)ceilf((float) atof(CurrentFieldValue.c_str()));
							break;
						}
						case 16:
						{
							Rows[0] = CurrentFieldValue;
							break;
						}
					}
					CurrentFieldIndex++;
					CurrentFieldValue.clear();
				}
				else if (buf[i] == '\n')
				{
					if (CurrentFieldIndex != 17)
					{
						Die("Line %lu is of wrong size\n", LineNumber);
					}

					Rows[1] = CurrentFieldValue;

					bool HasUnapprovedChars = false;

					for (unsigned k = 0; k < 2; ++k)
						for (unsigned j = 0; j < Rows[k].size(); j++)
						{
							if (!ApprovedDNA[(unsigned char)Rows[k][j]])
							{
								if (!HasUnapprovedChars)
									Warning("Unapproved DNA letter in line %lu: %c\n", LineNumber, Rows[k][j]);
								HasUnapprovedChars = true;
							}
						}

					if ((!HasUnapprovedChars) && (PercentID >= minperc) && (Rows[0].size() >= minlength))
					{
						AI.Segs[0].ChrId = CurrentChrID1;
						AI.Segs[1].ChrId = CurrentChrID2;
						OutFile.AddAln(AI, Rows);
						HSPsWritten++;
					}
					else
						HSPsIgnored++;

				}
				else
					CurrentFieldValue.push_back(buf[i]);
			}
			if (buf[i] == '\n')
			{
				LineNumber++;
				CurrentFieldIndex = 1;
				CurrentFieldValue.clear();
				StartOfLine = true;
				IgnoreLine = false;
			}
			++i;
		}
	}

	OutFile.Close();
	ProgressDone();
	Progress("%lu HSPs written, %lu HSPs ignored, %lu lines ignored\n", HSPsWritten, HSPsIgnored, LinesIgnored);
}

void CvtBlast2Rev(bool UseLastZFormat)
{
	unsigned minperc = 0;
	unsigned minlength = 0;

	const char *BlastFileName = (UseLastZFormat ? RequiredValueOpt("fromlastz") : RequiredValueOpt("fromblast"));
	const char *OutFileName = RequiredValueOpt("torev");
	const char *SeqFileName1 = RequiredValueOpt("seq1");
	const char *SeqFileName2 = RequiredValueOpt("seq2");

	IntOpt("minpct", &minperc);
	IntOpt("minlength", &minlength);
	bool DoRename = FlagOpt("gendotid");

	ApprovedDNA[(unsigned char)'a'] = 1;
	ApprovedDNA[(unsigned char)'c'] = 1;
	ApprovedDNA[(unsigned char)'g'] = 1;
	ApprovedDNA[(unsigned char)'t'] = 1;
	ApprovedDNA[(unsigned char)'A'] = 1;
	ApprovedDNA[(unsigned char)'C'] = 1;
	ApprovedDNA[(unsigned char)'G'] = 1;
	ApprovedDNA[(unsigned char)'T'] = 1;
	ApprovedDNA[(unsigned char)'N'] = 1;
	ApprovedDNA[(unsigned char)'n'] = 1;
	ApprovedDNA[(unsigned char)'-'] = 1;

	if (UseLastZFormat)
		Lastz2Rev(BlastFileName, OutFileName, SeqFileName1, SeqFileName2, minperc,
			minlength, DoRename);
	else
		Blast2Rev(BlastFileName, OutFileName, SeqFileName1, SeqFileName2, minperc,
			minlength, DoRename);
}

