#include "revfile.h"
#include "getopts.h"
#include "progress.h"
#include "utils.h"

using namespace std;

static unsigned g_PageWidth = 80;
static unsigned g_GenomeIndex;
static StdStrVec g_OutRows;
static IntVec g_OutRowIndexToGenomeIndex;
static BoolVec g_AnyLetters;
static unsigned g_OutColIndex;
static StdStrVec g_GenomeNames;
static unsigned g_MaxNameLength;
static IntVecVec g_OutRowIndexLookup;

static void MakeLookups(const IntVec &Counts)
	{
	const unsigned GenomeCount = SIZE(Counts);
	unsigned OutRowCount = 0;
	for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
		{
		unsigned Count = Counts[GenomeIndex];
		OutRowCount += Count;
		}
	g_OutRowIndexToGenomeIndex.resize(OutRowCount, UINT_MAX);
	unsigned DupeIndex = GenomeCount;
	g_OutRowIndexLookup.clear();
	g_OutRowIndexLookup.resize(GenomeCount);
	for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
		{
		unsigned Count = Counts[GenomeIndex];
		g_OutRowIndexLookup[GenomeIndex].push_back(GenomeIndex);
		unsigned OutRowIndex = GenomeIndex;
		g_OutRowIndexToGenomeIndex[OutRowIndex] = GenomeIndex;
		for (unsigned i = 1; i < Count; ++i)
			{
			unsigned OutRowIndex = DupeIndex++;
			g_OutRowIndexLookup[GenomeIndex].push_back(OutRowIndex);
			g_OutRowIndexToGenomeIndex[OutRowIndex] = GenomeIndex;
			}
		}
	}

static void MakeSegIndexToOutRowIndex(const vector<GCSeg> &Segs,
  IntVec &SegIndexToOutRowIndex)
	{
	const unsigned GenomeCount = SIZE(g_OutRowIndexLookup);
	IntVec Counts(GenomeCount, 0);
	const unsigned SegCount = SIZE(Segs);
	SegIndexToOutRowIndex.clear();
	SegIndexToOutRowIndex.resize(SegCount, UINT_MAX);
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSeg &Seg = Segs[SegIndex];
		unsigned GenomeIndex = Seg.GenomeIndex;
		asserta(GenomeIndex < GenomeCount);
		unsigned Count = Counts[GenomeIndex];
		const IntVec &v = g_OutRowIndexLookup[GenomeIndex];
		asserta(Count < SIZE(v));
		unsigned OutRowIndex = v[Count];
		SegIndexToOutRowIndex[SegIndex] = OutRowIndex;
		++(Counts[GenomeIndex]);
		}
	}

static void ClearOut(unsigned OutRowCount)
	{
	g_OutRows.clear();
	g_AnyLetters.clear();
	g_OutRows.resize(OutRowCount);
	g_AnyLetters.resize(OutRowCount, false);
	g_OutColIndex = 0;
	}

static void AppendCol(const StdStr &Col)
	{
	const unsigned N = SIZE(Col);
	for (unsigned i = 0; i < N; ++i)
		{
		char c = Col[i];
		StdStr &Row = g_OutRows[i];
		Row.push_back(c);
		}

	++g_OutColIndex;
	}

static void OutAncRow(FILE *f, const StdStr &Row)
	{
	for (StdStr::const_iterator p = Row.begin(); p != Row.end(); ++p)
		fputc(*p, f);
	}

static void OutRow(FILE *f, const StdStr *PrevRow, const StdStr &Row)
	{
	if (PrevRow == 0)
		{
		fputs(Row.c_str(), f);
		return;
		}
	const unsigned N = SIZE(Row);
	asserta(SIZE(*PrevRow) == N);
	for (unsigned i = 0; i < N; ++i)
		{
		char pc = (*PrevRow)[i];
		char c = Row[i];
		if (c == pc && c != ' ' && c != '|')
			fputc('.', f);
		else
			fputc(c, f);
		}
	}

static void FlushOut(FILE *f)
	{
	const unsigned GenomeCount = SIZE(g_GenomeNames);
	const unsigned OutRowCount = SIZE(g_OutRows);
	const StdStr *PrevRow = 0;
	for (unsigned OutRowIndex = 0; OutRowIndex < OutRowCount; ++OutRowIndex)
		{
		if (!g_AnyLetters[OutRowIndex] && OutRowIndex > 0)
			continue;
		StdStr &Row = g_OutRows[OutRowIndex];
		unsigned GenomeIndex = g_OutRowIndexToGenomeIndex[OutRowIndex];
		asserta(GenomeIndex < SIZE(g_GenomeNames));
		const StdStr &Name = g_GenomeNames[GenomeIndex];
		char Format[32];
		const char *DupeStr = (OutRowIndex >= GenomeCount ? "++" : "  ");
		sprintf(Format, "%s %%%u.%us  ", DupeStr, g_MaxNameLength, g_MaxNameLength);
		fprintf(f, Format, Name.c_str());
		if (OutRowIndex == 0)
			OutAncRow(f, Row);
		else
			OutRow(f, PrevRow, Row);
		fprintf(f, "\n");
		PrevRow = &Row;
		}
	fprintf(f, "\n");

	ClearOut(OutRowCount);
	}

static void AddBreakpoint()
	{
	const unsigned OutRowCount = SIZE(g_OutRows);
	StdStr Blanks(OutRowCount, ' ');
	StdStr Bars(OutRowCount, '|');
	AppendCol(Blanks);
	AppendCol(Bars);
	AppendCol(Blanks);
	}

static void AddAln(FILE *f, const AlnInfo &AI, const StdStrVec &Rows)
	{
	asserta(SIZE(AI.Segs) == SIZE(Rows));
	asserta(SIZE(AI.Segs) > 0);

	const unsigned OutRowCount = SIZE(g_OutRows);
	const unsigned SegCount = SIZE(AI.Segs);

	IntVec SegIndexToOutRowIndex;
	MakeSegIndexToOutRowIndex(AI.Segs, SegIndexToOutRowIndex);

	const unsigned ColCount = SIZE(Rows[0]);
	for (unsigned ColIndex = 0; ColIndex < ColCount; ++ColIndex)
		{
		if (g_OutColIndex >= g_PageWidth)
			FlushOut(f);
		StdStr Col;
		Col.resize(OutRowCount, ' ');
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const StdStr &Row = Rows[SegIndex];
			unsigned OutRowIndex = SegIndexToOutRowIndex[SegIndex];
			char c = Row[ColIndex];
			if (c != '-')
				g_AnyLetters[OutRowIndex] = true;
			Col[OutRowIndex] = c;
			}
		AppendCol(Col);
		}
	}

static bool LT(const AlnInfoX &AIX1, const AlnInfoX &AIX2)
	{
	const AlnInfo &AI1 = AIX1.AI;
	const AlnInfo &AI2 = AIX2.AI;

	const GCSeg *Seg1 = AI1.GetSegPtr(g_GenomeIndex);
	const GCSeg *Seg2 = AI2.GetSegPtr(g_GenomeIndex);

	if (Seg1 == 0 && Seg2 != 0)
		return false;
	if (Seg2 == 0 && Seg1 != 0)
		return true;
	if (Seg1 == 0 && Seg2 == 0)
		return AI1.Segs[0].lo < AI2.Segs[0].lo;
	return Seg1->lo < Seg2->lo;
	}

// Sort /decreasing/ by length
static bool LTSeg(const GCSeg &Seg1, const GCSeg &Seg2)
	{
	return Seg1.GetLength() > Seg2.GetLength();
	}

void CvtRev2ASA()
	{
	const char *RevFileName = RequiredValueOpt("fromrev");
	const char *ASAFileName = RequiredValueOpt("toasa");
	const char *GenomeName = RequiredValueOpt("genome");
	IntOpt("minlength", &g_PageWidth);

	RevFile Rev;
	Rev.Open(RevFileName);

	StdStrVec GenomeNames;
	const unsigned GenomeCount = Rev.GetGenomeCount();
	bool NameFound = false;
	g_MaxNameLength = 0;
	for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
		{
		const StdStr &Name = Rev.GetGenomeName(GenomeIndex);
		g_MaxNameLength = max(SIZE(Name), g_MaxNameLength);
		g_GenomeNames.push_back(Name);
		GenomeNames.push_back(Name);
		if (Name == GenomeName)
			{
			g_GenomeIndex = GenomeIndex;
			NameFound = true;
			}
		}

	if (!NameFound)
		Die("Genome '%s' not found", GenomeName);

	sort(GenomeNames.begin(), GenomeNames.end());

	IntVec MaxRows(GenomeCount, 0);

	AIXVec AIXs;
	Rev.GetAllAlnInfoXs(AIXs, true);

	const unsigned AIXCount = SIZE(AIXs);
	for (unsigned AIXIndex = 0; AIXIndex < AIXCount; ++AIXIndex)
		{
		AlnInfo &AI = AIXs[AIXIndex].AI;
		sort(AI.Segs.begin(), AI.Segs.end(), LTSeg);
		}

	ProgressStart("Compute max rows / genome");
	for (unsigned AIIndex = 0; AIIndex < AIXCount; ++AIIndex)
		{
		if (AIIndex%1000 == 0)
			ProgressStep(AIIndex, AIXCount);

		IntVec RowCount(GenomeCount, 0);

		const AlnInfoX &AIX = AIXs[AIIndex];
		const unsigned SegCount = AIX.AI.GetSegCount();
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = AIX.AI.Segs[SegIndex];
			unsigned GenomeIndex = Seg.GenomeIndex;
			asserta(GenomeIndex < GenomeCount);
			++(RowCount[GenomeIndex]);
			}

		for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
			MaxRows[GenomeIndex] = max(RowCount[GenomeIndex], MaxRows[GenomeIndex]);
		}
	ProgressDone();

	MakeLookups(MaxRows);

	unsigned OutRowCount = 0;
	for (unsigned GenomeIndex = 0; GenomeIndex < GenomeCount; ++GenomeIndex)
		{
		unsigned RowCount = MaxRows[GenomeIndex];
		ProgressNoRes("Genome %16.16s  Max rows %u",
		  Rev.GetGenomeName(GenomeIndex).c_str(), RowCount);
		OutRowCount += RowCount;
		}

	Progress("Sorting AIs");
	sort(AIXs.begin(), AIXs.end(), LT);

	ProgressStart("Formatting");
	FILE *fOut = CreateStdioFile(ASAFileName);
	ClearOut(OutRowCount);
	for (unsigned AIXIndex = 0; AIXIndex < AIXCount; ++AIXIndex)
		{
		if (AIXIndex%100 == 0)
			ProgressStep(AIXIndex, AIXCount);

		const AlnInfoX &AIX = AIXs[AIXIndex];
		const AlnInfo &AI = AIX.AI;
		StdStrVec Rows;
		unsigned AlnIndex = Rev.GetAlnIndex(AIX.AlnInfoIndex);
		Rev.GetAlnRows(AlnIndex, Rows);

		AddAln(fOut, AI, Rows);
		if (AIXIndex + 1 < AIXCount)
			AddBreakpoint();
		}
	ProgressDone();
	FlushOut(fOut);

	CloseStdioFile(fOut);
	Rev.Close();
	}
