#include "utils.h"
#include "maffile.h"
#include "revfile.h"
#include "getopts.h"
#include "progress.h"

void TruncLabel(std::string &Label);
const StdStr &TraceTypeToStr(unsigned TraceType);
static bool g_NoCmt = false;
static bool g_UseIds = false;
static bool g_UseChrIds = false;
static bool g_NoPairwise = false;

static void Add(const RevFile &Rev, unsigned AIIndex, MAFFile &MAF)
	{
	unsigned AlnIndex = Rev.GetAlnIndex(AIIndex);

	StdStrVec ssRows;
	unsigned AlnInfoId = Rev.GetAlnRows(AlnIndex, ssRows);

	AlnInfo AI;
	unsigned AlnIndex2 = Rev.GetAlnInfo(AlnInfoId, AI);

	if (g_NoPairwise && SIZE(AI.Segs) < 3)
		return;

	if (AlnIndex != AlnIndex2)
		Die("Add assert failed");

	StdStrVec SeqNames;
	StdStrVec Rows;
	PosVec SeqLengths;
	PosVec Los;
	PosVec His;
	BoolVec Pluses;

	const unsigned SegCount = (unsigned) AI.Segs.size();
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSeg &S = AI.Segs[SegIndex];

		const unsigned GenomeIndex = S.GenomeIndex;
		const unsigned ChrId = S.ChrId;

		std::string GenomeName = Rev.GetGenomeName(GenomeIndex);
		const ChrInfo &CI = Rev.GetChrInfo(GenomeIndex, ChrId);

		std::string ChrName = CI.GetName();

		TruncLabel(GenomeName);
		TruncLabel(ChrName);

		size_t gs = GenomeName.size();
		size_t cs = ChrName.size();
		if (gs == 0)
			Die("Genome %u has no name (length=0)", GenomeIndex);
		if (cs == 0)
			Die("Chr %u,%u has no name (length=0)", GenomeIndex, ChrId);

		StdStr SeqName;
		if (g_UseIds)
			{
			char Tmp[32];
			sprintf(Tmp, "%u.%u", GenomeIndex, ChrId);
			SeqName = Tmp;
			}
		else if (g_UseChrIds)
			{
			char Tmp[32];
			sprintf(Tmp, "%s.%u", GenomeName.c_str(), ChrId);
			SeqName = Tmp;
			}
		else
			SeqName = GenomeName + "." + ChrName;

		SeqNames.push_back(SeqName);

		const char *Row = ssRows[SegIndex].c_str();
		Rows.push_back(Row);

		unsigned SeqLength = (unsigned) CI.GetLength();
		if (SeqLength <= S.hi)
			{
			Warning("Expanding seq length to end of hit");
			SeqLength = S.hi + 1;
			}

		SeqLengths.push_back(SeqLength);
		Los.push_back(S.lo);
		His.push_back(S.hi);
		Pluses.push_back(S.Plus);
		}

	unsigned TraceType = AI.GetAttr(ALN_ATTR_TraceType);
	if (TraceType != UINT_MAX && !g_NoCmt)
		{
		const StdStr &StrType = TraceTypeToStr(TraceType);
		MAF.SetSubCmt(StrType);
		}
	else
		MAF.SetSubCmt("");

	char Cmt[1024];
	Cmt[0] = 0;
	const unsigned AttrCount = SIZE(AI.Attrs);
	FORIX(Attr)
		{
		AlnAttr Attr = AI.Attrs[AttrIndex];
		if ((unsigned) Attr >= KnownAttrCount)
			continue;
		const char *AttrName = AlnAttrToStr(Attr);
		if (Cmt[0] != 0)
			strcat(Cmt, " ");
		strcat(Cmt, AttrName);
		strcat(Cmt, "=");
		
		switch (AATypes[Attr])
			{
		case ALN_ATTR_TYPE_UInt32:
			{
			char Str[16];
			sprintf(Str, "%u", AI.GetAttr(Attr));
			strcat(Cmt, Str);
			break;
			}
		case ALN_ATTR_TYPE_Float32:
			{
			char Str[16];
			sprintf(Str, "%g", AI.GetFloatAttr(Attr));
			strcat(Cmt, Str);
			break;
			}
		default:
			strcat(Cmt, "?");
			break;
			}
		}

	MAF.AppendAln(Los, His, Pluses, Rows, SeqNames, SeqLengths, AI.Score, 0,
	  Cmt[0] == 0 ? 0 : Cmt);
	}

void CvtRev2MAF()
	{
	const char *RevFileName = RequiredValueOpt("fromrev");
	const char *MAFFileName = RequiredValueOpt("tomaf");

	const char *CmdLine = ValueOpt("maf_cmdline");
	const char *Scoring = ValueOpt("maf_scoring");
	const char *Program = ValueOpt("maf_program");

	g_NoCmt = FlagOpt("nocmt");
	g_UseIds = FlagOpt("useids");
	g_UseChrIds = FlagOpt("usechrids");
	g_NoPairwise = FlagOpt("nopairwise");
	bool PlusCoords = FlagOpt("pluscoords");

	RevFile Rev;
	Rev.Open(RevFileName);

	MAFFile MAF;
	MAF.Create(MAFFileName, CmdLine, Scoring, Program);
	MAF.SetPlusCoords(PlusCoords);

	const unsigned AICount = Rev.GetAlnInfoCount();
	ProgressStart("Converting");
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%100 == 0)
			ProgressStep(AIIndex, AICount);
		Add(Rev, AIIndex, MAF);
		}
	ProgressDone();

	Rev.Close();
	MAF.Close();
	}
