#include "revfile.h"
#include "getopts.h"
#include "utils.h"
#include "progress.h"

using namespace std;

static bool LettersMatch(char c1, char c2, bool IgnoreCase,
  bool NWild)
	{
	if (c1 == c2)
		return true;
	if (IgnoreCase && toupper(c1) == toupper(c2))
		return true;
	if (NWild && c1 == 'N' || c2 == 'N' || c1 == 'n' || c2 == 'n')
		return true;
	return false;
	}

void DiffChrSeq()
	{
	const char *ChrName = RequiredValueOpt("diffchrseq");
	const char *GenomeName = RequiredValueOpt("genome");
	const char *InFileName1 = RequiredValueOpt("in1");
	const char *InFileName2 = RequiredValueOpt("in2");
	bool IgnoreCase = FlagOpt("ignorecase");
	bool NWild = FlagOpt("nwild");

	unsigned MinLength = 8;
	IntOpt("minlength", &MinLength);

	RevFile In1;
	RevFile In2;

	In1.Open(InFileName1);
	In2.Open(InFileName2);

	unsigned GenomeIndex1 = In1.GetGenomeIndex(GenomeName);
	unsigned GenomeIndex2 = In2.GetGenomeIndex(GenomeName);

	unsigned ChrId1 = In1.GetChrId(GenomeIndex1, ChrName);
	unsigned ChrId2 = In2.GetChrId(GenomeIndex2, ChrName);

	unsigned Length1 = In1.GetSeqLength(GenomeIndex1, ChrId1);
	unsigned Length2 = In2.GetSeqLength(GenomeIndex2, ChrId2);

	if (Length1 != Length2)
		Progress("Lengths differ %u %u", Length1, Length2);

	unsigned Length = min(Length1, Length2);
	if (Length == 0)
		return;

	char *Seq1 = new char[Length];
	char *Seq2 = new char[Length];

	char *Block1 = new char[Length];
	char *Block2 = new char[Length];

	unsigned char *RunLengths = new unsigned char[Length];
	memset(RunLengths, 0, Length);

	In1.GetSeq(GenomeIndex1, ChrId1, Seq1, Length);
	In2.GetSeq(GenomeIndex2, ChrId2, Seq2, Length);

	unsigned StartBlock = 0;
	unsigned char RunLength = 0;
	for (unsigned i = 0; i < Length; ++i)
		{
		char c1 = Seq1[i];
		char c2 = Seq2[i];
		bool Match = LettersMatch(c1, c2, IgnoreCase, NWild);
		if (!Match)
			RunLength = 0;
		else if (RunLength < MinLength)
			++RunLength;
		RunLengths[i] = RunLength;
		}

	for (unsigned i = 1; i < Length; ++i)
		{
		unsigned j = Length - i - 1;
		unsigned char PrevRunLength = RunLengths[j+1];
		if (RunLengths[j] > 0 && PrevRunLength >= MinLength)
			RunLengths[j] = MinLength;
		}

	StartBlock = 0;
	unsigned BlockLength = 0;
	bool Matching = true;
	for (unsigned i = 0; i < Length; ++i)
		{
		char c1 = Seq1[i];
		char c2 = Seq2[i];
		if (RunLengths[i] >= MinLength)
			{
			if (!Matching)
				{
				if (BlockLength > 0)
					{
					Log("\n");
					Log("Mismatch %-5u  %u-%u\n", BlockLength, StartBlock, i-1);

					Block1[BlockLength] = 0;
					Block2[BlockLength] = 0;
					Log("         %s\n", Block1);
					Log("         %s\n", Block2);
					BlockLength = 0;
					}
				}
			Matching = true;
			}
		else
			{
			if (Matching)
				{
				if (BlockLength > 0)
					{
					Log("\n");
					Log("Match    %-5u  %u-%u\n", BlockLength, StartBlock, i-1);

					Block1[BlockLength] = 0;
					Block2[BlockLength] = 0;
					Log("         %s\n", Block1);
					Log("         %s\n", Block2);
					BlockLength = 0;
					}
				StartBlock = i;
				}
			Matching = false;
			}
		Block1[BlockLength] = c1;
		Block2[BlockLength] = c2;
		++BlockLength;
		}

	if (Matching)
		{
		if (BlockLength > 0)
			{
			Log("\n");
			Log("Match    %-5u  %u-%u\n", BlockLength, StartBlock, Length-1);
			Block1[BlockLength] = 0;
			Block2[BlockLength] = 0;
			Log("         %s\n", Block1);
			Log("         %s\n", Block2);
			}
		}
	else
		{
		if (BlockLength > 0)
			{
			Log("\n");
			Log("Mismatch %-5u  %u-%u\n", BlockLength, StartBlock, Length-1);

			Block1[BlockLength] = 0;
			Block2[BlockLength] = 0;
			Log("         %s\n", Block1);
			Log("         %s\n", Block2);
			}
		}
	}
