#include <cairo.h>
#include <cairo-pdf.h>
#include <stack>
#include <cmath>
#include <utility>
#include "geninfo.h"
#include "alninfo.h"
#include "revfile.h"
#include "utils.h"
#include "getopts.h"

VALUE_OPT ValueOpts[] =
{
	{ "fromrev",    0 },
	{ "tocmap",     0 },
	{ "tolmap",     0 },
	{ "blocksize",  0 },
	{ "chr0name",   0 },
	{ "chr0id",     0 },
	{ "chr1name",   0 },
	{ "chr1id",     0 },
	{ "npp",        0 },
	{ "npi",        0 },
	{ "log",        0 }
};
int ValueOptCount = sizeof(ValueOpts)/sizeof(ValueOpts[0]);

FLAG_OPT FlagOpts[] =
{
	{ "dummy",      0 }
};
int FlagOptCount = sizeof(FlagOpts)/sizeof(FlagOpts[0]);

void Usage()
{
	fprintf(stderr,
"Usage:\n"
"    drawrev -fromrev <revfile> -tocmap <pdf> [-blocksize <blocksize>] [-npi <nucleotides_per_inch>]\n"
"    drawrev -fromrev <revfile> -tolmap <png> [-npp <nucleotides_per_pixel>]\n"
"\n"
	);
}

int Colors[][3] =
{
	{102, 102, 102},
	{  0,   2, 240},
	{166, 242, 240},
	{  0, 184,   0},
	{ 41, 255, 199},
	{127, 255,   0},
	{ 99, 204,   2},
	{247, 250,  22},
	{191, 158, 179},
	{211,  84,   7},
	{255, 122, 145},
	{252, 135,  54},
	{255, 102, 245},
	{176,   0, 148},
	{158,  23, 242},
	{240,  20,  10},
	{207, 222, 224},
	{120,   0,   0},
	{129,  13, 240},
	{  0, 201, 143},
	{121, 233,  12},
	{164,  28,  30},
	{  0, 217, 255},
	{237, 237, 146},
};

unsigned ColorCount = sizeof(Colors) / sizeof(Colors[0]);

void cairo_setcolor(cairo_t *cr, unsigned id)
{
	id = id % ColorCount;
	cairo_set_source_rgb(cr, Colors[id][0] / 255.0, Colors[id][1] / 255.0, Colors[id][2] / 255.0);
}

//Counts[genome][chromosome][block][otherchromosome] = # of bases voting for that
std::vector<std::vector<std::vector<unsigned> > > Counts[2];
cairo_surface_t *surface;
cairo_t *cr;
unsigned BlockSize = 100000;
unsigned npp = 1000000;
unsigned npi = 20000000;

double UpperBorder = 0.25*72;
double TitleHeight = 0.25*72;
double ChromLabelHeight = 0.2*72;
double LegendLabelHeight = 0.4*72;
double LegendHeight = 0.25*72;
double LowerBorder = 0.25*72;
double LeftBorder = 0.25*72;
double ChromWidth = 0.25*72;
double ChromOffsetX = 0.1*72;
double RightBorder = 0.25*72;

std::stack<std::pair<double, double> > cairo_stack;

void cairo_push(cairo_t *cr)
{
	double x = 0, y = 0;
	cairo_get_current_point(cr, &x, &y);
	cairo_stack.push(std::make_pair<double, double>(x, y));
}

void cairo_pop(cairo_t *cr)
{
	std::pair<double, double> p = cairo_stack.top();
	cairo_stack.pop();
	cairo_move_to(cr, p.first, p.second);
}

void cairo_top(cairo_t *cr)
{
	std::pair<double, double> p = cairo_stack.top();
	cairo_move_to(cr, p.first, p.second);
}

void cairo_stroke2(cairo_t *cr)
{
	if (cairo_has_current_point(cr))
	{
		cairo_push(cr);
		cairo_stroke(cr);
		cairo_pop(cr);
	}
	else
	{
		cairo_stroke(cr);
	}
}

void cairo_fill2(cairo_t *cr)
{
	if (cairo_has_current_point(cr))
	{
		cairo_push(cr);
		cairo_fill(cr);
		cairo_pop(cr);
	}
	else
	{
		cairo_fill(cr);
	}
}


/* Writes the text "utf" on the cairo "cr", using the largest font size (up to
 * "maxsize") that will make the text fit into a "w"-by-"h" rectangle. The text
 * is centered into that rectangle. Font metrics are also taken into account;
 * (i.e. if you write "oooo" and "OOOO", they will be aligned to the same baseline
 * even though they have different heights).
 *
 * Prior to calling this routine, the cursor needs to be positioned at the bottom
 * left corner of the rectangle.
 *
 * If "debug" is true, it will also paint guide rectangles useful for debugging
 */
void cairo_text(cairo_t *cr, const char *utf, double w, double h, double maxsize, bool debug = false)
{
	double font_size = maxsize;
	for(;;)
	{
		cairo_text_extents_t te;
		cairo_font_extents_t fe;
		cairo_set_font_size(cr, font_size);
		cairo_text_extents(cr, utf, &te);
		cairo_font_extents(cr, &fe);
		if ((te.width <= w) && (fe.ascent + fe.descent <= h))
		{
			cairo_push(cr);
			cairo_rel_move_to(cr, (w-te.width) / 2.0 - te.x_bearing, -(h - fe.ascent - fe.descent) / 2.0 - fe.descent);
			cairo_show_text(cr, utf);
			cairo_pop(cr);
			if (debug)
			{
				cairo_push(cr);
				cairo_rel_move_to(cr, (w-te.width) / 2.0, -(h - fe.ascent - fe.descent) / 2.0);
				cairo_rel_line_to(cr, te.width, 0);
				cairo_rel_line_to(cr, 0, -fe.ascent - fe.descent);
				cairo_rel_line_to(cr, -te.width, 0);
				cairo_close_path(cr);
				cairo_set_line_width(cr, 0.5);
				cairo_stroke(cr);
				cairo_set_line_width(cr, 1.0);
				cairo_pop(cr);
				cairo_push(cr);
				cairo_rel_line_to(cr, w, 0);
				cairo_rel_line_to(cr, 0, -h);
				cairo_rel_line_to(cr, -w, 0);
				cairo_close_path(cr);
				cairo_set_line_width(cr, 0.5);
				cairo_stroke(cr);
				cairo_set_line_width(cr, 1.0);
				cairo_pop(cr);
			}
			break;
		}
		else
		{
			if (font_size > 1)
				font_size = font_size - 1;
			else
				font_size /= 2.0;

			if (font_size < 1e-5)
				break;
		}
	}
}

void HSV2RGB(double h, double s, double v, double *r, double *g, double *b)
{
	h = h * 360.0;
	int hi = ((int)(h / 60.0)) % 6;
	double f = h / 60.0 - (int)(h / 60.0);
	double p = v * (1 - s);
	double q = v * (1 - f*s);
	double t = v * (1 - (1-f) * s);

	switch (hi)
	{
		case 0:
		*r = v; *g = t; *b = p; return;
		case 1:
		*r = q; *g = v; *b = p; return;
		case 2:
		*r = p; *g = v; *b = t; return;
		case 3:
		*r = p; *g = q; *b = v; return;
		case 4:
		*r = t; *g = p; *b = v; return;
		case 5:
		*r = v; *g = p; *b = q; return;
		default:
		Die("Invalid HSV");
	}
}

void StripString(std::string &Haystack, const std::string &Needle)
{
	size_t n;
	while ((n = Haystack.find(Needle)) != std::string::npos)
	{
		Haystack.erase(n, Needle.size());
	}
}

void DrawCounts(const GenInfoVec &Genomes, const char *OutFileName)
{
	surface = cairo_pdf_surface_create(OutFileName, 612, 792);
	if (cairo_surface_status(surface) != CAIRO_STATUS_SUCCESS)
		Die("Cannot create output file");

	cr = cairo_create(surface);
	if (cairo_status(cr) != CAIRO_STATUS_SUCCESS)
		Die("Cannot allocate canvas");

	cairo_select_font_face(cr, "Arial", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);

	double NPD = npi/72.0;

	for (unsigned i = 0; i < 2; ++i)
	{
		unsigned ChrCount = Genomes[i].GetChrCount();
		unsigned OtherChrCount = Genomes[1-i].GetChrCount();
		unsigned MaxLength = 0;
		for (unsigned j = 0; j < ChrCount; ++j)
			if (MaxLength < Counts[i][j].size())
				MaxLength = Counts[i][j].size();
		double MaxChromHeight = MaxLength * BlockSize / NPD;

		double TotalHeight = UpperBorder + TitleHeight + MaxChromHeight + ChromLabelHeight + LegendLabelHeight + LegendHeight + ChromLabelHeight + LowerBorder;
		double TotalWidth = LeftBorder + std::max(ChrCount, OtherChrCount) * (ChromWidth + ChromOffsetX) + RightBorder - ChromOffsetX;

		cairo_pdf_surface_set_size(surface, TotalWidth, TotalHeight);
		cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
		cairo_paint(cr);

		// Upper Border
		cairo_move_to(cr, LeftBorder, UpperBorder);

		// Title
		cairo_rel_move_to(cr, 0, TitleHeight);
		cairo_set_source_rgb(cr, 0, 0, 0);
		cairo_text(cr, Genomes[i].GetName().c_str(), TotalWidth - LeftBorder - RightBorder, TitleHeight, 12);

		// Chromosomes + Labels
		cairo_rel_move_to(cr, 0, MaxChromHeight);
		cairo_push(cr);
		for (unsigned j = 0; j < Counts[i].size(); ++j)
		{
			double ChromHeight = Counts[i][j].size() * BlockSize / NPD;

			for (unsigned k = 0; k < Counts[i][j].size(); ++k)
			{
				unsigned BestChrom = 0;
				unsigned original_k = k;
				for (unsigned m = 1; m < Counts[i][j][k].size(); ++m)
				{
					if (Counts[i][j][k][m] > Counts[i][j][k][BestChrom])
						BestChrom = m;
				}
				while (k < Counts[i][j].size() - 1)
				{
					unsigned BestChrom2 = 0;
					for (unsigned m = 1; m < Counts[i][j][k+1].size(); ++m)
					{
						if (Counts[i][j][k+1][m] > Counts[i][j][k+1][BestChrom2])
							BestChrom2 = m;
					}
					if (BestChrom == BestChrom2)
						++k;
					else
					{
						break;
					}
				}
				cairo_push(cr);
				cairo_rel_move_to(cr, 0, -(original_k * BlockSize / NPD));
				cairo_rel_line_to(cr, ChromWidth, 0);
				cairo_rel_line_to(cr, 0, -((k+1 - original_k) * BlockSize / NPD));
				cairo_rel_line_to(cr, -ChromWidth, 0);
				cairo_close_path(cr);
				/*
				double hue = BestChrom / (double)(Counts[i][j][k].size());
				double R, G, B;
				HSV2RGB(hue, 1, 1, &R, &G, &B);
				cairo_set_source_rgb(cr, R, G, B);
				*/
				cairo_setcolor(cr, BestChrom);
				cairo_set_line_width(cr, 1.0);
				cairo_fill(cr);
				cairo_pop(cr);
			}
			cairo_push(cr);
			cairo_rel_line_to(cr, 0, -ChromHeight);
			cairo_rel_line_to(cr, ChromWidth, 0);
			cairo_rel_line_to(cr, 0, ChromHeight);
			cairo_close_path(cr);
			cairo_set_source_rgb(cr, 0, 0, 0);
			cairo_set_line_width(cr, 0.5);
			cairo_stroke(cr);
			cairo_top(cr);
			cairo_rel_move_to(cr, 0, ChromLabelHeight);
			cairo_text(cr, Genomes[i].GetChrName(j).c_str(), ChromWidth, ChromLabelHeight, 9);
			cairo_pop(cr);
			cairo_rel_move_to(cr, ChromWidth + ChromOffsetX, 0);
		}
		cairo_pop(cr);
		cairo_rel_move_to(cr, 0, ChromLabelHeight);

		// Legend Title
		cairo_rel_move_to(cr, 0, LegendLabelHeight);
		std::string Legend = "Legend (";
		Legend += Genomes[1-i].GetName();
		Legend += ")";
		cairo_text(cr, Legend.c_str(), TotalWidth - LeftBorder - RightBorder, LegendLabelHeight, 9);

		// Legend Chromosomes + Labels
		for (unsigned m = 0; m < Counts[i][0][0].size(); ++m)
		{
			cairo_push(cr);
			cairo_rel_line_to(cr, ChromWidth, 0);
			cairo_rel_line_to(cr, 0, LegendHeight);
			cairo_rel_line_to(cr, -ChromWidth, 0);
			cairo_close_path(cr);
			/*
			double hue = m / (double)(Counts[i][0][0].size());
			double R, G, B;
			HSV2RGB(hue, 1, 1, &R, &G, &B);
			cairo_set_source_rgb(cr, R, G, B);
			*/
			cairo_setcolor(cr, m);
			cairo_set_line_width(cr, 1.0);
			cairo_fill(cr);
			cairo_top(cr);
			cairo_rel_move_to(cr, 0, ChromLabelHeight + LegendHeight);
			cairo_set_source_rgb(cr, 0, 0, 0);
			cairo_text(cr, Genomes[1-i].GetChrName(m).c_str(), ChromWidth, ChromLabelHeight, 9);
			cairo_pop(cr);

			cairo_rel_move_to(cr, ChromWidth + ChromOffsetX, 0);
		}
		cairo_show_page(cr);
	}
	cairo_surface_destroy(surface);
	cairo_destroy(cr);
}

void DrawRevToLmap(const char *RevFileName, const char *OutFileName)
{
	RevFile InFile;

	InFile.Open(RevFileName);

	if (InFile.GetGenomeCount() != 2)
		Die("Linear maps can only be produced for exactly two genomes");

	double HeaderSize = 15;

	double SizeX = ceil(InFile.GetGenomeInfo(0).GetLength64() / (double)npp);
	double SizeY = ceil(InFile.GetGenomeInfo(1).GetLength64() / (double)npp);

	double HeaderOffset = 2*HeaderSize;

	double TotalX = SizeX + 2*HeaderSize;
	double TotalY = SizeY + 2*HeaderSize;

	surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, (int)TotalX, (int)TotalY);
	if (cairo_surface_status(surface) != CAIRO_STATUS_SUCCESS)
		Die("Cannot allocate drawing surface");

	cr = cairo_create(surface);
	if (cairo_status(cr) != CAIRO_STATUS_SUCCESS)
		Die("Cannot allocate canvas");

	cairo_select_font_face(cr, "Arial", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);

	cairo_set_source_rgb(cr, 1, 1, 1);
	cairo_paint(cr);
	cairo_set_line_width(cr, 0.5);

	std::vector<uint64> Offsets[2];

	cairo_set_source_rgb(cr, 0, 0, 0.4);
	cairo_move_to(cr, HeaderSize, HeaderOffset);
	cairo_line_to(cr, HeaderSize, TotalY);
	cairo_stroke(cr);
	cairo_move_to(cr, HeaderOffset, HeaderSize);
	cairo_line_to(cr, TotalX, HeaderSize);
	cairo_stroke(cr);
	for (unsigned i = 0; i < 2; ++i)
	{
		unsigned ChrCount = InFile.GetChrCount(i);
		Offsets[i].resize(ChrCount, 0);

		if (i == 0)
		{
			cairo_move_to(cr, HeaderOffset, HeaderSize);
			cairo_text(cr, InFile.GetGenomeName(i).c_str(), TotalX - HeaderOffset, HeaderSize, 12);
		}
		else
		{
			cairo_move_to(cr, HeaderSize, TotalY);
			cairo_rotate(cr, -M_PI/2.0);
			cairo_text(cr, InFile.GetGenomeName(i).c_str(), TotalY - HeaderOffset, HeaderSize, 12);
			cairo_identity_matrix(cr);
		}

		for (unsigned j = 0; j < ChrCount; ++j)
		{
			if (InFile.GetChrId(i, j) >= ChrCount)
				Die("This program does not support sparse indices");
			if (j > 0)
				Offsets[i][j] = Offsets[i][j-1] + InFile.GetChrLength(i, j-1);
			double h = Offsets[i][j] / (double)npp;
			if (i == 0)
			{
				cairo_move_to(cr, h + HeaderOffset, (j == 0) ? 0 : HeaderSize);
				cairo_line_to(cr, h + HeaderOffset, TotalY);
				cairo_stroke(cr);
				cairo_move_to(cr, h + HeaderOffset, HeaderOffset);
				cairo_text(cr, InFile.GetChrName(i, j).c_str(), InFile.GetChrLength(i, j)/(double)npp, HeaderSize, 12);
			}
			else
			{
				cairo_move_to(cr, (j == 0) ? 0 : HeaderSize, h + HeaderOffset);
				cairo_line_to(cr, TotalX, h + HeaderOffset);
				cairo_stroke(cr);
				cairo_move_to(cr, HeaderOffset, h + HeaderOffset + InFile.GetChrLength(i, j)/(double)npp);
				cairo_rotate(cr, -M_PI/2.0);
				cairo_text(cr, InFile.GetChrName(i, j).c_str(), InFile.GetChrLength(i, j)/(double)npp, HeaderSize, 12);
				cairo_identity_matrix(cr);
			}
		}
	}
	cairo_set_source_rgb(cr, 0, 0, 0);

	unsigned AICount = InFile.GetAlnInfoCount();

	for(unsigned AlnIndex = 0; AlnIndex < AICount; ++AlnIndex)
	{
		AlnInfo AI;
		InFile.GetAlnInfo(AlnIndex, AI);
		if ((AlnIndex & 0x2fff) == 0)
		{
			printf("%.2f%%\r", AlnIndex*100.0/AICount);
		}

		double x_lo = (Offsets[0][AI.GetSeg(0).ChrId] + AI.GetSeg(0).lo) / (double)npp + HeaderOffset;
		double x_hi = (Offsets[0][AI.GetSeg(0).ChrId] + AI.GetSeg(0).hi) / (double)npp + HeaderOffset;
		double y_lo = (Offsets[1][AI.GetSeg(1).ChrId] + AI.GetSeg(1).lo) / (double)npp + HeaderOffset;
		double y_hi = (Offsets[1][AI.GetSeg(1).ChrId] + AI.GetSeg(1).hi) / (double)npp + HeaderOffset;

		if (AI.GetSeg(0).Plus == AI.GetSeg(1).Plus)
		{
			cairo_move_to(cr, x_lo, y_lo);
			cairo_line_to(cr, x_hi, y_hi);
			cairo_stroke(cr);
		}
		else
		{
			cairo_move_to(cr, x_lo, y_hi);
			cairo_line_to(cr, x_hi, y_lo);
			cairo_stroke(cr);
		}
	}
	printf("\n");

	if (cairo_surface_write_to_png(surface, OutFileName) != CAIRO_STATUS_SUCCESS)
		Die("Error writing output file");

	cairo_surface_destroy(surface);
	cairo_destroy(cr);

	InFile.Close();
}

void DrawRevToLmapSingleChrom(const char *RevFileName, const char *OutFileName)
{
	RevFile InFile;

	InFile.Open(RevFileName);

	if (InFile.GetGenomeCount() != 2)
		Die("Linear maps can only be produced for exactly two genomes");

	double HeaderSize = 15;

	unsigned ChrX = 0;
	unsigned ChrY = 0;

	if (ValueOpt("chr0name"))
		ChrX = InFile.GetGenomeInfo(0).GetChrId(ValueOpt("chr0name"));
	else
		ChrX = RequiredIntOpt("chr0id");
	if (ValueOpt("chr1name"))
		ChrY = InFile.GetGenomeInfo(1).GetChrId(ValueOpt("chr1name"));
	else
		ChrY = RequiredIntOpt("chr1id");

	double SizeX = ceil(InFile.GetChrLength(0, ChrX) / (double)npp);
	double SizeY = ceil(InFile.GetChrLength(1, ChrY) / (double)npp);

	double HeaderOffset = 2*HeaderSize;

	double TotalX = SizeX + 2*HeaderSize;
	double TotalY = SizeY + 2*HeaderSize;

	surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, (int)TotalX, (int)TotalY);
	if (cairo_surface_status(surface) != CAIRO_STATUS_SUCCESS)
		Die("Cannot allocate drawing surface");

	cr = cairo_create(surface);
	if (cairo_status(cr) != CAIRO_STATUS_SUCCESS)
		Die("Cannot allocate canvas");

	cairo_select_font_face(cr, "Arial", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);

	cairo_set_source_rgb(cr, 1, 1, 1);
	cairo_paint(cr);
	cairo_set_line_width(cr, 0.5);

	cairo_set_source_rgb(cr, 0, 0, 0.4);
	cairo_move_to(cr, HeaderSize, HeaderOffset);
	cairo_line_to(cr, HeaderSize, TotalY);
	cairo_stroke(cr);
	cairo_move_to(cr, HeaderOffset, HeaderSize);
	cairo_line_to(cr, TotalX, HeaderSize);
	cairo_stroke(cr);
	for (unsigned i = 0; i < 2; ++i)
	{
		if (i == 0)
		{
			cairo_move_to(cr, HeaderOffset, HeaderSize);
			cairo_text(cr, InFile.GetGenomeName(i).c_str(), TotalX - HeaderOffset, HeaderSize, 12);
			cairo_move_to(cr, HeaderOffset, 0);
			cairo_line_to(cr, HeaderOffset, TotalY);
			cairo_stroke(cr);
			cairo_move_to(cr, HeaderOffset, HeaderOffset);
			cairo_text(cr, InFile.GetChrName(i, (i == 0) ? ChrX : ChrY).c_str(), InFile.GetChrLength(i, (i == 0) ? ChrX : ChrY)/(double)npp, HeaderSize, 12);
		}
		else
		{
			cairo_move_to(cr, HeaderSize, TotalY);
			cairo_rotate(cr, -M_PI/2.0);
			cairo_text(cr, InFile.GetGenomeName(i).c_str(), TotalY - HeaderOffset, HeaderSize, 12);
			cairo_identity_matrix(cr);
			cairo_move_to(cr, 0, HeaderOffset);
			cairo_line_to(cr, TotalX, HeaderOffset);
			cairo_stroke(cr);
			cairo_move_to(cr, HeaderOffset, HeaderOffset + InFile.GetChrLength(i, (i == 0) ? ChrX : ChrY)/(double)npp);
			cairo_rotate(cr, -M_PI/2.0);
			cairo_text(cr, InFile.GetChrName(i, (i == 0) ? ChrX : ChrY).c_str(), InFile.GetChrLength(i, (i == 0) ? ChrX : ChrY)/(double)npp, HeaderSize, 12);
			cairo_identity_matrix(cr);
		}
	}
	cairo_set_source_rgb(cr, 0, 0, 0);

	unsigned AICount = InFile.GetAlnInfoCount();

	for(unsigned AlnIndex = 0; AlnIndex < AICount; ++AlnIndex)
	{
		AlnInfo AI;
		InFile.GetAlnInfo(AlnIndex, AI);
		if ((AlnIndex & 0x2fff) == 0)
		{
			printf("%.2f%%\r", AlnIndex*100.0/AICount);
		}
		if ((AI.GetSeg(0).ChrId != ChrX) || (AI.GetSeg(1).ChrId != ChrY))
			continue;

		double x_lo = AI.GetSeg(0).lo / (double)npp + HeaderOffset;
		double x_hi = AI.GetSeg(0).hi / (double)npp + HeaderOffset;
		double y_lo = AI.GetSeg(1).lo / (double)npp + HeaderOffset;
		double y_hi = AI.GetSeg(1).hi / (double)npp + HeaderOffset;

		if (AI.GetSeg(0).Plus == AI.GetSeg(1).Plus)
		{
			cairo_move_to(cr, x_lo, y_lo);
			cairo_line_to(cr, x_hi, y_hi);
			cairo_stroke(cr);
		}
		else
		{
			cairo_move_to(cr, x_lo, y_hi);
			cairo_line_to(cr, x_hi, y_lo);
			cairo_stroke(cr);
		}
	}
	printf("\n");

	if (cairo_surface_write_to_png(surface, OutFileName) != CAIRO_STATUS_SUCCESS)
		Die("Error writing output file");

	cairo_surface_destroy(surface);
	cairo_destroy(cr);

	InFile.Close();
}

void DrawRevToCmap(const char *RevFileName, const char *OutFileName)
{
	RevFile InFile;

	InFile.Open(RevFileName);

	if (InFile.GetGenomeCount() != 2)
		Die("Color maps can only be produced for exactly two genomes");

	for (unsigned i = 0; i < 2; ++i)
	{
		unsigned ChrCount = InFile.GetChrCount(i);
		unsigned OtherChrCount = InFile.GetChrCount(1-i);
		Counts[i].resize(ChrCount);
		for (unsigned j = 0; j < ChrCount; ++j)
		{
			if (!(InFile.GetGenomeInfo(i).GetChrId(j) < ChrCount))
				Die("This program does not support sparse chromosome indexes");
			unsigned BlockCount = InFile.GetChrLength(i, j) / BlockSize;
			if (BlockCount * BlockSize < InFile.GetChrLength(i, j))
				++BlockCount;
			Counts[i][j].resize(BlockCount);

			for (unsigned k = 0; k < BlockCount; ++k)
				Counts[i][j][k].resize(OtherChrCount);

			std::string NewName = InFile.GetChrName(i, j);
			StripString(NewName, "chr");
			InFile.GetModifiableGenomeInfo(i).GetModifiableChrInfoPtr(j)->SetName(NewName);
		}
	}

	unsigned AICount = InFile.GetAlnInfoCount();

	for(unsigned AlnIndex = 0; AlnIndex < AICount; ++AlnIndex)
	{
		AlnInfo AI;
		InFile.GetAlnInfo(AlnIndex, AI);
		if ((AlnIndex & 0x2fff) == 0)
		{
			printf("%.2f%%\r", AlnIndex*100.0/AICount);
		}

		for (unsigned i = 0; i < 2; ++i)
		{
			coord_t lo = AI.Segs[i].lo;
			coord_t hi = AI.Segs[i].hi;
			for (unsigned Block = lo / BlockSize; Block <= hi / BlockSize; ++Block)
			{
				coord_t BlockLo = std::max(lo, Block*BlockSize);
				coord_t BlockHi = std::min(hi, (Block+1)*BlockSize-1);
				Counts[i][AI.Segs[i].ChrId][Block][AI.Segs[1-i].ChrId] += BlockHi - BlockLo + 1;
			}
		}
	}
	printf("\n");

	DrawCounts(InFile.GetGenomes(), OutFileName);

	InFile.Close();
}

int main(int argc, char *argv[])
{
	ProcessArgVect(argc, argv);

	if (ValueOpt("fromrev") && ValueOpt("tocmap"))
	{
		IntOpt("npi", &npi);
		IntOpt("blocksize", &BlockSize);
		DrawRevToCmap(ValueOpt("fromrev"), ValueOpt("tocmap"));
	}
	else if (ValueOpt("fromrev") && ValueOpt("tolmap"))
	{
		if ((ValueOpt("chr0id") || ValueOpt("chr0name")) && (ValueOpt("chr1id") || ValueOpt("chr1name")))
		{
			npp = 100000;
			IntOpt("npp", &npp);
			DrawRevToLmapSingleChrom(ValueOpt("fromrev"), ValueOpt("tolmap"));
		}
		else
		{
			IntOpt("npp", &npp);
			DrawRevToLmap(ValueOpt("fromrev"), ValueOpt("tolmap"));
		}
	}
	else
	{
		Usage();
		return 1;
	}

	return 0;
}
