#include "evolver.h"
#include "ratemgr.h"
#include "intlist.h"
#include "evomodel.h"
#include "myrand.h"
#include "progress.h"

double GetSubstituteRate(double ATFract);

/***
The rate of an event E is the average number times E will occur per
site per unit time.

For tandem expand and contract the number of sites includes only
those annotated as tandems.

Let R be the integrated rate, i.e. the average number of times E
will occur per unit time over the entire genome.

The average time between occurences of an event with rate R is 1/R
(e.g. if the rate is 3 per second, then the time between events is
1/3 of a second).

An event E with integrated rate R will occur k = Rt times in time
t, and if there are N events with integrated rates R_i, i = 1 .. N,
then the total number of events expected in time t is K = S t where
S is the sum of the integrated rates for each event.

The average time between events is 1/S.
***/

void Evolver::ComputeEventRates(double ATFract)
	{
	double SavedSubstRate = 0;
	if (ATFract == DBL_MAX)
		{
		if (m_IntegratedEventRates.empty())
			Die("ComputEventRates(MAX), empty");
		SavedSubstRate = m_IntegratedEventRates[EVENT_TYPE_Substitute];
		}

	BoolVec EventDone(EventTypeCount, false);

	m_IntegratedEventRates.clear();
	m_IntegratedEventRates.resize(EventTypeCount, 0);

	gpos_t ChrLength = GetEvolvedLength();
	gpos_t TandemBaseCount = GetTandemBaseCount();

	unsigned GeneCount = GetGeneCount();
	unsigned NGECount = GetNGECount();
	unsigned NXECount = GetNXECount();
	unsigned CDSCount = GetCDSCount();
	unsigned UTRCount = GetUTRCount();
	unsigned IslandCount = GetIslandCount();

	double MERate = GetMERate();
	double SubstituteRate = 0;
	if (ATFract == DBL_MAX)
		m_IntegratedEventRates[EVENT_TYPE_Substitute] = SavedSubstRate;
	else
		{
		SubstituteRate = GetSubstituteRate(ATFract);
		m_IntegratedEventRates[EVENT_TYPE_Substitute] = SubstituteRate*ChrLength;
		}
	EventDone[EVENT_TYPE_Substitute] = true;

#define R(e)									\
	m_IntegratedEventRates[EVENT_TYPE_##e] =	\
	  ChrLength*g_##e##RateMgr->GetTotalRate();	\
	EventDone[EVENT_TYPE_##e] = true;

	R(Delete)
	R(Invert)
	R(Insert)
	R(Tandem)
	R(Copy)
	R(Move)
#undef R

#define R(e)									\
	m_IntegratedEventRates[EVENT_TYPE_##e] = 0;	\
	EventDone[EVENT_TYPE_##e] = true;

	R(InterChrCopy)
	R(InterChrMove)
	R(ChrFuse)
	R(ChrSplit)
	R(RecipTransloc)
	R(NonRecipTransloc)
#undef R

#define R(e, r, N)										\
	m_IntegratedEventRates[EVENT_TYPE_##e] = (r)*(N);	\
	EventDone[EVENT_TYPE_##e] = true;

//	R(Substitute,				SubstituteRate,					ChrLength)
	R(TandemExpand,				P_TandemExpandRate,				TandemBaseCount)
	R(TandemContract,			P_TandemContractRate,			TandemBaseCount)
	R(MEInsert,					MERate,							ChrLength)

	R(MoveStartCodonIntoUTR,	P_MoveStartCodonIntoUTRRate,	GeneCount)
	R(MoveStartCodonIntoCDS,	P_MoveStartCodonIntoCDSRate,	GeneCount)
	R(MoveStopCodonIntoUTR,		P_MoveStopCodonIntoUTRRate,		GeneCount)
	R(MoveStopCodonIntoCDS,		P_MoveStopCodonIntoCDSRate,		GeneCount)

	R(MoveUTRTerm,				P_MoveUTRTermRate,				GeneCount)

	R(DeleteUTR,				P_DeleteUTRRate,				UTRCount)
	R(CreateNontermUTR,			P_CreateNontermUTRRate,			UTRCount)
	R(CreateTermUTR,			P_CreateTermUTRRate,			UTRCount)

	R(DeleteCDS,				P_DeleteCDSRate,				CDSCount)
	R(CreateCDS,				P_CreateCDSRate,				CDSCount)
	R(MoveDonorIntoCDS,			P_MoveDonorIntoCDSRate,			CDSCount)
	R(MoveCDSDonorIntoIntron,	P_MoveCDSDonorIntoIntronRate,	CDSCount)
	R(MoveAcceptorIntoCDS,		P_MoveAcceptorIntoCDSRate,		CDSCount)
	R(MoveCDSAcceptorIntoIntron,P_MoveCDSAcceptorIntoIntronRate,CDSCount)

	R(MoveDonorIntoUTR,			P_MoveDonorIntoUTRRate,			UTRCount)
	R(MoveUTRDonorIntoIntron,	P_MoveUTRDonorIntoIntronRate,	UTRCount)
	R(MoveAcceptorIntoUTR,		P_MoveAcceptorIntoUTRRate,		UTRCount)
	R(MoveUTRAcceptorIntoIntron,P_MoveUTRAcceptorIntoIntronRate,UTRCount)

	R(DeleteNXE,				P_DeleteNXERate,				NXECount)
	R(CreateNXE,				P_CreateNXERate,				NXECount)
	R(MoveNXE,					P_MoveNXERate,					NXECount)

	R(DeleteNGE,				P_DeleteNGERate,				NGECount)
	R(CreateNGE,				P_CreateNGERate,				NGECount)
	R(MoveNGE,					P_MoveNGERate,					NGECount)

	R(DeleteIsland,				P_DeleteCpGIslandRate,			IslandCount)
	R(CreateIsland,				P_CreateCpGIslandRate,			IslandCount)
	R(MoveIsland,				P_MoveCpGIslandRate,			IslandCount)

	R(ChangeGeneSpeed,			P_ChangeGeneSpeedRate,			GeneCount)
	R(ChangeNGESpeed,			P_ChangeNGESpeedRate,			NGECount)
#undef R

#define E(e)	if (!EventDone[EVENT_TYPE_##e]) Die("Event " #e ": rate not done");
#include "events.h"

	double TotalIntegratedRate = 0;
	for (unsigned i = 0; i < EventTypeCount; ++i)
		{
		double Rate = m_IntegratedEventRates[i];
		TotalIntegratedRate += Rate;
		}

	if (TotalIntegratedRate == 0)
		Die("Total event rate zero");

	m_AverageWaitingTime = 1.0/TotalIntegratedRate;

	GetRandDistHis(m_IntegratedEventRates, m_EventRandDistHis);
	}

void Evolver::GetIntegratedRates(vector<double> &Rates) const
	{
	Rates = m_IntegratedEventRates;
	}

const char *EventTypeToStr(EVENT_TYPE Type)
	{
	switch (Type)
		{
#define E(e)	case EVENT_TYPE_##e: return #e;
#include "events.h"
		}
	static char Str[32];
	sprintf(Str, "EVENT_TYPE_%u", (unsigned) Type);
	return Str;
	}

const char *CountToStr(gpos_t Count)
	{
	const gpos_t Kilo = (gpos_t) 1000;
	const gpos_t Mega = (gpos_t) 1e6;
	const gpos_t Giga = (gpos_t) 1e9;

	static char Str[32];
	double dCount = (double) Count;
	if (Count < Kilo)
		sprintf(Str, "%"PF, Count);
	else if (Count >= Kilo && Count < Mega)
		sprintf(Str, "%.2fk", dCount / Kilo);
	else if (Count >= Mega && Count < Giga)
		sprintf(Str, "%.2fM", dCount / Mega);
	else
		sprintf(Str, "%.2fG", dCount / Giga);
	return Str;
	}

// Select a random event.
EVENT_TYPE Evolver::GetRandomEventType()
	{
	assert(SIZE(m_IntegratedEventRates) == EventTypeCount);
	EVENT_TYPE Type = (EVENT_TYPE) RandWeightDist4(m_EventRandDistHis);
	return Type;
	}

double Evolver::GetRandomWaitingTime() const
	{
	asserta(m_AverageWaitingTime > 0);
	asserta(m_AverageWaitingTime != DBL_MAX);

	return RandExp(m_AverageWaitingTime);
	}

void Evolver::LogEventRates() const
	{
	Log("\n");
	Log("               Event  Integrated Rate     Pct\n");
	Log("====================  ===============  ======\n");

	double TotalRate = 0;
	double SumPct = 0;
	for (unsigned e = 0; e < EventTypeCount; ++e)
		TotalRate += m_IntegratedEventRates[e];

	double r;
	double Pct;
#define E(e)											\
		r = m_IntegratedEventRates[EVENT_TYPE_##e];		\
		Pct = 100.0*r/TotalRate;						\
		SumPct += Pct;									\
		Log("%20.20s  %15.3g  %6.1f%%\n", #e, r, Pct);

#include "events.h"

	Log("                      ===============  ======\n");
	Log("                      %15.3g  %6.1f%%\n", TotalRate, SumPct);
	Log("\n");
	}
