#include "evolver.h"

bool IsStopCodon(const string &Codon);

/***
Non-inverted gene on plus strand:

Base.Letter     A   T   G
Base.Codex		1   2   3
Base.Plus       T   T   T

Non-inverted gene on minus strand:

comp            g   t   a
Base.Letter     C   A   T
Base.Codex     -3  -2  -1
Base.Plus       T   T   T

Inverted gene originally on plus strand,
now on minus strand:

Base.Letter     G   T   A
Base.Codex	   -3  -2  -1
Base.Plus       F   F   F

Inverted gene originally on minus strand,
now on plus strand:

Base.Letter     T   A   C
Base.Codex		1   2   3
Base.Plus       F   F   F
***/

bool IsStartCodon(const string &Codon);

gpos_t Evolver::GetAdjExonPos(gpos_t Pos_, bool Plus) const
	{
	gpos_t Pos = Pos_;
	for (;;)
		{
		Pos = GetAdjPos(Pos, Plus);
		if (Pos == GPOS_T_MAX)
			break;
		const BaseInfo &Base = GetBase(Pos);
		if (Base.IsExon())
			return Pos;
		if (!Base.IsGene())
			break;
		}
	return GPOS_T_MAX;
	}

gpos_t Evolver::GetPrevExonPos(gpos_t Pos_, bool Plus) const
	{
	gpos_t Pos = Pos_;
	for (;;)
		{
		Pos = GetAdjPos(Pos, !Plus);
		if (Pos == GPOS_T_MAX)
			break;
		const BaseInfo &Base = GetBase(Pos);
		if (Base.IsExon())
			return Pos;
		if (!Base.IsGene())
			break;
		}
	return GPOS_T_MAX;
	}

void Evolver::SetTriplet(gpos_t Pos, const string &Letters, bool Strand)
	{
	asserta(Letters.size() == 3);

	gpos_t Pos2 = GetAdjExonPos(Pos, Strand);
	asserta(Pos2 != GPOS_T_MAX);
	gpos_t Pos3 = GetAdjExonPos(Pos2, Strand);
	asserta(Pos3 != GPOS_T_MAX);

	BaseInfo &Base1 = GetBase(Pos);
	BaseInfo &Base2 = GetBase(Pos2);
	BaseInfo &Base3 = GetBase(Pos3);

	Base1.SetStrandLetter(Letters[0], Strand);
	Base2.SetStrandLetter(Letters[1], Strand);
	Base3.SetStrandLetter(Letters[2], Strand);
	}

// Get three bases (triplet) on given strand starting at given position,
// return first position in next triplet. If in CDS, is a codon, but could
// also be in a UTR because we're searching for a Cand start or stop
// codon for a move-start or move-stop event.
gpos_t Evolver::GetNextExonTriplet(gpos_t Pos, string &Letters, bool Strand) const
	{
	gpos_t Pos2 = GetAdjExonPos(Pos, Strand);
	if (Pos2 == GPOS_T_MAX)
		return GPOS_T_MAX;
	gpos_t Pos3 = GetAdjExonPos(Pos2, Strand);
	if (Pos3 == GPOS_T_MAX)
		return GPOS_T_MAX;

	const BaseInfo &Base1 = GetBase(Pos);
	const BaseInfo &Base2 = GetBase(Pos2);
	const BaseInfo &Base3 = GetBase(Pos3);

	Letters.resize(3);
	Letters[0] = Base1.GetStrandLetter(Strand);
	Letters[1] = Base2.GetStrandLetter(Strand);
	Letters[2] = Base3.GetStrandLetter(Strand);
	return GetAdjExonPos(Pos3, Strand);
	}

void Evolver::GetNextPair(gpos_t Pos, string &Letters, bool Strand) const
	{
	gpos_t Pos2 = GetAdjPos(Pos, Strand);
	asserta(Pos2 != GPOS_T_MAX);

	const BaseInfo &Base1 = GetBase(Pos);
	const BaseInfo &Base2 = GetBase(Pos2);

	Letters.resize(2);
	Letters[0] = Base1.GetStrandLetter(Strand);
	Letters[1] = Base2.GetStrandLetter(Strand);
	}

// Get three bases (triplet) on given strand starting at given position,
// return first position in next triplet.
gpos_t Evolver::GetNextIntronTriplet(gpos_t Pos, string &Letters, bool Strand) const
	{
	gpos_t Pos2 = GetAdjPos(Pos, Strand);
	if (Pos2 == GPOS_T_MAX)
		return GPOS_T_MAX;
	gpos_t Pos3 = GetAdjPos(Pos2, Strand);
	if (Pos3 == GPOS_T_MAX)
		return GPOS_T_MAX;

	const BaseInfo &Base1 = GetBase(Pos);
	const BaseInfo &Base2 = GetBase(Pos2);
	const BaseInfo &Base3 = GetBase(Pos3);

	if (!Base1.IsIntron() || !Base2.IsIntron() || !Base3.IsIntron())
		return GPOS_T_MAX;

	Letters.resize(3);
	Letters[0] = Base1.GetStrandLetter(Strand);
	Letters[1] = Base2.GetStrandLetter(Strand);
	Letters[2] = Base3.GetStrandLetter(Strand);
	return GetAdjPos(Pos3, Strand);
	}

// Get three bases (triplet) on given strand /ending/ at given position,
// return /last/ position in next triplet. If in CDS, is a codon, but could
// also be in a UTR because we're searching for a Cand start or stop
// codon for a move-start or move-stop event.
gpos_t Evolver::GetPrevExonTriplet(gpos_t Pos, string &Letters, bool Strand) const
	{
	gpos_t Pos2 = GetPrevExonPos(Pos, Strand);
	if (Pos2 == GPOS_T_MAX)
		return GPOS_T_MAX;
	gpos_t Pos3 = GetPrevExonPos(Pos2, Strand);
	if (Pos3 == GPOS_T_MAX)
		return GPOS_T_MAX;

	const BaseInfo &Base1 = GetBase(Pos);
	const BaseInfo &Base2 = GetBase(Pos2);
	const BaseInfo &Base3 = GetBase(Pos3);

	Letters.resize(3);
	Letters[0] = Base3.GetStrandLetter(Strand);
	Letters[1] = Base2.GetStrandLetter(Strand);
	Letters[2] = Base1.GetStrandLetter(Strand);
	return GetPrevExonPos(Pos3, Strand);
	}

bool Evolver::IsCodon(gpos_t Pos, bool Strand) const
	{
	gpos_t Pos2 = GetAdjExonPos(Pos, Strand);
	if (Pos2 == GPOS_T_MAX)
		return GPOS_T_MAX;
	gpos_t Pos3 = GetAdjExonPos(Pos2, Strand);
	if (Pos3 == GPOS_T_MAX)
		return GPOS_T_MAX;

	const BaseInfo &Base1 = GetBase(Pos);
	const BaseInfo &Base2 = GetBase(Pos2);
	const BaseInfo &Base3 = GetBase(Pos3);
	if (!Base1.IsCDS() || !Base2.IsCDS() || !Base3.IsCDS())
		return false;

	if (Strand)
		return Base1.Codex == 1 && Base2.Codex == 2 && Base3.Codex == 3;
	else
		return Base1.Codex == -1 && Base2.Codex == -2 && Base3.Codex == -3;
	}

void Evolver::LogExonTriplet(gpos_t Pos, bool Strand) const
	{
	gpos_t Pos2 = GetAdjExonPos(Pos, Strand);
	gpos_t Pos3 = GetAdjExonPos(Pos2, Strand);

	const BaseInfo &Base1 = GetBase(Pos);
	const BaseInfo &Base2 = GetBase(Pos2);
	const BaseInfo &Base3 = GetBase(Pos3);

	Log("%"PF" ", Pos);
	Base1.LogMe(true);

	Log("%"PF" ", Pos2);
	Base2.LogMe(true);

	Log("%"PF" ", Pos3);
	Base3.LogMe(true);
	}

bool Evolver::GetFractCodonBeforeIntron(gpos_t Pos, string &Letters,
  bool Strand, bool &StartStop) const
	{
	asserta(PosInIntron(Pos));

	Letters.clear();
	StartStop = false;

	while (PosInIntron(Pos))
		Pos = GetAdjPos(Pos, !Strand);

	const BaseInfo &Base = GetBase(Pos);
	StartStop |= Base.IsStartStop();
	asserta(Base.IsExon());
	if (!Base.IsCDS())
		return false;
	
	switch (Base.Codex)
		{
	case 1:
	case -1:
		Letters.resize(1);
		Letters[0] = Base.GetStrandLetter(Strand);
		return true;

	case 2:
	case -2:
		{
		Letters.resize(2);
		Letters[1] = Base.GetStrandLetter(Strand);
		Pos = GetPrevExonPos(Pos, Strand);
		const BaseInfo &Base0 = GetBase(Pos);
		StartStop |= Base0.IsStartStop();
		asserta(Base0.IsCDS());
		Letters[0] = Base0.GetStrandLetter(Strand);
		return true;
		}

	case 3:
	case -3:
		return true;

	default:
		asserta(false);
		}
	return true;
	}

bool Evolver::GetFractCodonAfterIntron(gpos_t Pos, string &Letters,
  bool Strand, bool &StartStop) const
	{
	asserta(PosInIntron(Pos));
	Letters.clear();
	StartStop = false;

	while (PosInIntron(Pos))
		Pos = GetAdjPos(Pos, Strand);

	const BaseInfo &Base = GetBase(Pos);
	StartStop |= Base.IsStartStop();
	asserta(Base.IsExon());
	if (!Base.IsCDS())
		return false;

	switch (Base.Codex)
		{
	case 1:
	case -1:
		return true;

	case 2:
	case -2:
		{
		Letters.resize(2);
		Letters[0] = Base.GetStrandLetter(Strand);
		Pos = GetAdjExonPos(Pos, Strand);
		const BaseInfo &Base1 = GetBase(Pos);
		StartStop |= Base1.IsStartStop();
		asserta(Base1.IsCDS());
		Letters[1] = Base1.GetStrandLetter(Strand);
		return true;
		}

	case 3:
	case -3:
		Letters.resize(1);
		Letters[0] = Base.GetStrandLetter(Strand);
		return true;

	default:
		asserta(false);
		}
	UNREACHABLE_RETURN(false);
	}

bool Evolver::GetFractCodonExonStart(const ExonInfo &Exon, string &FractCodon,
  bool Strand, bool &StartStop) const
	{
	FractCodon.clear();
	StartStop = false;
	gpos_t FirstExonPos = Strand ? Exon.Pos : GetRangeEndPos(Exon.Pos, Exon.Length);
	const BaseInfo &Base1 = GetBase(FirstExonPos);
	StartStop |= Base1.IsStartStop();
	if (!Base1.IsCDS())
		return false;
	switch (Base1.Codex)
		{
	case 1:
	case -1:
		return true;

	case 2:
	case -2:
		{
		FractCodon.push_back(Base1.GetStrandLetter(Strand));

		gpos_t Pos2 = GetAdjPos(FirstExonPos, Strand);
		if (Pos2 == GPOS_T_MAX)
			return false;
		const BaseInfo &Base2 = GetBase(Pos2);
		StartStop |= Base2.IsStartStop();
		if (!Base2.IsCDS())
			return false;
		FractCodon.push_back(Base2.GetStrandLetter(Strand));
		return true;
		}

	case 3:
	case -3:
		{
		FractCodon.push_back(Base1.GetStrandLetter(Strand));
		return true;
		}

	default:
		asserta(false);
		}
	UNREACHABLE_RETURN(false);
	}

bool Evolver::GetFractCodonExonEnd(const ExonInfo &Exon, string &FractCodon,
  bool Strand, bool &StartStop) const
	{
	StartStop = false;
	FractCodon.clear();
	gpos_t LastExonPos = Strand ? GetRangeEndPos(Exon.Pos, Exon.Length): Exon.Pos;
	const BaseInfo &Base1 = GetBase(LastExonPos);
	StartStop |= Base1.IsStartStop();
	if (!Base1.IsCDS())
		return false;
	switch (Base1.Codex)
		{
	case 1:
	case -1:
		FractCodon.push_back(Base1.GetStrandLetter(Strand));
		return true;

	case 2:
	case -2:
		{
		FractCodon.resize(2);
		FractCodon[1] = Base1.GetStrandLetter(Strand);

		gpos_t Pos2 = GetAdjPos(LastExonPos, !Strand);
		if (Pos2 == GPOS_T_MAX)
			return false;
		const BaseInfo &Base2 = GetBase(Pos2);
		StartStop |= Base2.IsStartStop();
		if (!Base2.IsCDS())
			return false;
		FractCodon[0] = Base2.GetStrandLetter(Strand);
		return true;
		}

	case 3:
	case -3:
		return true;

	default:
		asserta(false);
		}
	UNREACHABLE_RETURN(false);
	}

bool Evolver::HasStopCodon(const string &Seq) const
	{
	unsigned L = SIZE(Seq);
	for (unsigned i = 0; i < L; i += 3)
		{
		string Codon = Seq.substr(i, 3);
		if (IsStopCodon(Codon))
			return true;
		}
	return false;
	}
