#include "evolver.h"

/***
Definition of GFF frame from http://mblab.wustl.edu/GTF22.html:
	0 indicates that the feature begins with a whole codon at the 5' most base.
	1 means that there is one extra base (the third base of a codon) before the
	first whole codon and 2 means that there are two extra bases (the second
	and third bases of the codon) before the first codon. Note that for reverse
	strand features, the 5' most base is the <end> coordinate.
	Frame is calculated as (3 - ((length-frame) mod 3)) mod 3.

The above computation for frame is dangerous in C because length-frame may
be negative, and the % operator is compiler-dependent when an operand is < 0,
therefore I add 3 to ensure the operands are > 0.
***/

// IncFrame increments the frame by Length positions.
uint8 IncFrame(uint8 Frame, gpos_t Length)
	{
	asserta(Frame < 3);
	return (uint8) ((3 - (Length + 3 - Frame)%3)%3);
	}

static int8 FrameToFirstCodexPlus(int Frame)
	{
	switch (Frame)
		{
// Frame 0 means no excess bases at start of frame,
// so first base is first base in codon.
	case 0:
		return 1;

// Frame 1 means one excess bases at start of frame,
// so first base (the excess base) is third base in codon.
	case 1:
		return 3;

// Frame 2 means two excess bases at start of frame,
// so first base is second base in codon.
	case 2:
		return 2;
		}
	asserta(false);
	UNREACHABLE_RETURN(0);
	}

static int FirstCodexToFramePlus(int8 Codex)
	{
	switch (Codex)
		{
	case 1:
		return 0;
	case 2:
		return 2;
	case 3:
		return 1;
		}
	asserta(false);
	UNREACHABLE_RETURN(0);
	}

static int LastCodexToFrameMinus(int8 Codex)
	{
	return FirstCodexToFramePlus(-Codex);
	}

static int8 FrameToLastCodexPlus(int8 Frame, gpos_t Length)
	{
// Use Length + 2 instead of Length - 1 in case Length is zero.
	int8 LastFrame = IncFrame(Frame, Length + 2);
	return FrameToFirstCodexPlus(LastFrame);
	}

static int8 FrameToFirstCodexMinus(int8 Frame, gpos_t Length)
	{
	return (int8) -FrameToLastCodexPlus(Frame, Length);
	}

// Returns codex of first base in plus strand order
int8 FrameToFirstCodex(int8 Frame, bool Strand, gpos_t Length)
	{
	if (Strand)
		return FrameToFirstCodexPlus(Frame);
	else
		return FrameToFirstCodexMinus(Frame, Length);
	}

// (Pos,Length) must exactly specify the range of a CDS.
int Evolver::GetCDSFrame(gpos_t Pos, gpos_t Length) const
	{
	const BaseInfo &Base = GetBase(Pos);
	bool Strand = Base.GetStrand();
	if (Strand)
		return FirstCodexToFramePlus(Base.Codex);
	else
		{
		gpos_t LastPos = GetRangeEndPos(Pos, Length);
		const BaseInfo &LastBase = GetBase(LastPos);
		return LastCodexToFrameMinus(LastBase.Codex);
		}
	}
