#include "evolver.h"

void Evolver::LogEvolvedAsTable() const
	{
	Log("     EvPos         Pos       HmPos  Cat  Pix  Cx  ECIUG  +  AE\n");
	Log("==========  ==========  ==========  ===  ===  ==  =====  =  ==\n");

	gpos_t EvPos = 0;
	FOR_EVBASES(p)
		{
		const BaseInfo &Base = m_Bases[p];
		Log("%10"PF, EvPos);
		Log("  %10"PF, p);
		gpos_t HmPos = GetHomologousPos(p);
		if (HmPos == GPOS_T_MAX)
			Log("           *");
		else
			Log("  %10"PF, HmPos);
		Log("  %3.3s", BaseCategoryToStr(GetCategory(p)));
		Log("   %02X", Base.ProbIx);
		if (Base.IsCoding())
			Log("  %2d", Base.Codex);
		else
			Log("    ");

		Log("  %c%c%c%c%c",
		  tof(Base.IsExon()),
		  tof(Base.IsCoding()),
		  tof(Base.IsIntron()),
		  tof(Base.IsUTR()),
		  tof(Base.IsUntranscribedGene()));

		Log("  %c", pom(Base.Plus));

		char A = Base.GetAncLetter();
		char E = Base.GetLetter();
		Log("  %c", A);
		if (E != A)
			Log("%c", E);
		Log("\n");
		++EvPos;
		}
	}

void Evolver::LogNeighborRegions(gpos_t Pos, unsigned Margin) const
	{
	gpos_t StartPos = Pos;
	gpos_t Length = 1;
	for (unsigned i = 0; i < Margin; ++i)
		{
		gpos_t p = GetPrevPos(StartPos);
		if (p == GPOS_T_MAX)
			break;
		StartPos = p;
		++Length;
		}

	gpos_t EndPos = Pos;
	for (unsigned i = 0; i < Margin; ++i)
		{
		gpos_t p = GetNextPos(EndPos);
		if (p == GPOS_T_MAX)
			break;
		EndPos = p;
		++Length;
		}
	LogRegions(StartPos, Length);
	}

void Evolver::LogNeighborsAsTable(gpos_t Pos, unsigned Margin) const
	{
	gpos_t StartPos = Pos;
	gpos_t Length = 1;
	for (unsigned i = 0; i < Margin; ++i)
		{
		gpos_t p = GetPrevPos(StartPos);
		if (p == GPOS_T_MAX)
			break;
		StartPos = p;
		++Length;
		}

	gpos_t EndPos = Pos;
	for (unsigned i = 0; i < Margin; ++i)
		{
		gpos_t p = GetNextPos(EndPos);
		if (p == GPOS_T_MAX)
			break;
		EndPos = p;
		++Length;
		}
	LogRangeAsTable(StartPos, Length);
	}

void Evolver::LogRegions(gpos_t Pos, gpos_t Length) const
	{
	Log("\n");
	Log("LogRegions(%"PF", %"PF")\n", Pos, Length);
	Log("      From          To      Length  +        Type\n");
	Log("==========  ==========  ==========  =  ==========\n");

	gpos_t StartPos = Pos;
	gpos_t LastPos = GPOS_T_MAX;
	gpos_t RegionLength = 0;
	BaseType2 LastType = (BaseType2) 0;
	bool LastPlus = false;
	for (unsigned i = 0; i < Length; ++i)
		{
		const BaseInfo &Base = GetBase(Pos);
		BaseType2 Type = Base.GetType2();
		gpos_t NextPos = (i + 1 == Length ? Pos : GetNextPos(Pos));

		gpos_t EndPos = GPOS_T_MAX;
		bool EndRegion = false;
		if (Type != LastType || Base.Plus != LastPlus)
			{
			EndPos = LastPos;
			EndRegion = true;
			}
		if (NextPos != Pos + 1 && NextPos + 1 != Pos)
			{
			EndPos = LastPos;
			EndRegion = true;
			}
		if (i + 1 == Length)
			{
			EndPos = Pos;
			EndRegion = true;
			}

		if (i != 0 && EndRegion)
			{
			Log("%10"PF"  %10"PF"  %10"PF"  %c  %10.10s\n",
			  StartPos, EndPos, RegionLength, pom(LastPlus), BaseType2ToStr(LastType));
			RegionLength = 0;
			}

		if (EndRegion)
			StartPos = Pos;
	
		++RegionLength;
		LastPos = Pos;
		LastType = Type;
		Pos = NextPos;
		LastPlus = Base.Plus;
		}
	Log("\n");
	}

void Evolver::LogRangeAsTable(gpos_t Pos, gpos_t Length) const
	{
	Log("\n");
	Log("Pos %"PF" Length %"PF"\n", Pos, Length);
	Log("       Pos       HmPos  Cat  Pix  Cx  +  ECISUG  AE\n");
	Log("==========  ==========  ===  ===  ==  =  ======  ==\n");

	for (gpos_t i = 0; i < Length; ++i)
		{
		if (!PosExists(Pos))
			return;
		const BaseInfo &Base = GetBase(Pos);
		Log("%10"PF, Pos);
		gpos_t HmPos = GetHomologousPos(Pos);
		if (HmPos == GPOS_T_MAX)
			Log("           *");
		else
			Log("  %10"PF, HmPos);
		Log("  %3.3s", BaseCategoryToStr(GetCategory(Pos)));
		Log("  %3u", Base.ProbIx);
		if (Base.IsCoding())
			Log("  %2d", Base.Codex);
		else
			Log("    ");
		Log("  %c", pom(Base.Plus));
		Log("  %c%c%c%c%c%c",
		  Base.IsExon() ? 'E' : ' ',
		  Base.IsCDS() ? 'C' : ' ',
		  Base.IsIntron() ? 'I' : ' ',
		  Base.IsSplice() ? 'S' : ' ',
		  Base.IsUTR() ? 'U' : ' ',
		  Base.IsUntranscribedGene() ? 'G' : ' ');
		Log("  %c%c", Base.GetLetter(), Base.GetAncLetter());
		Log("\n");
		Pos = GetNextPos(Pos);
		}
	}

void Evolver::LogMe() const
	{
	Log("\n");
	if (m_FirstEvolvedPos == UINT32_MAX)
		Log("Evolved head          *\n");
	else
		Log("Evolved head %10"PF"\n", m_FirstEvolvedPos);

	if (m_LastEvolvedPos == UINT32_MAX)
		Log("Evolved tail          *\n");
	else
		Log("Evolved tail %10"PF"\n", m_LastEvolvedPos);

	if (m_FirstDeletedPos == UINT32_MAX)
		Log("Deleted head          *\n");
	else
		Log("Deleted head %10"PF"\n", m_FirstDeletedPos);

	if (m_LastDeletedPos == UINT32_MAX)
		Log("Deleted tail          *\n");
	else
		Log("Deleted tail %10"PF"\n", m_LastDeletedPos);

	Log("m_Bases.size %10u\n", SIZE(m_Bases));

	Log("       Pos       HmPos         Fwd         Bwd  Cat  Pix  Cx  +  ECISUG  AE\n");
	Log("==========  ==========  ==========  ==========  ===  ===  ==  =  ======  ==\n");

	for (gpos_t Pos = 0; Pos < SIZE(m_Bases); ++Pos)
		{
		const BaseInfo &Base = m_Bases[Pos];
		Log("%10"PF, Pos);
		gpos_t HmPos = GetHomologousPos(Pos);
		if (HmPos == GPOS_T_MAX)
			Log("           *");
		else
			Log("  %10"PF, HmPos);
	
		gpos_t Fwd = m_BaseListFwds[Pos];
		if (Fwd == GPOS_T_MAX)
			Log("           *");
		else
			Log("  %10"PF, Fwd);

		gpos_t Bwd = m_BaseListBwds[Pos];
		if (Bwd == GPOS_T_MAX)
			Log("           *");
		else
			Log("  %10"PF, Bwd);

		Log("  %3.3s", BaseCategoryToStr(GetCategory(Pos)));
		Log("  %3u", Base.ProbIx);
		if (Base.IsCoding())
			Log("  %2d", Base.Codex);
		else
			Log("    ");
		Log("  %c", pom(Base.Plus));
		Log("  %c%c%c%c%c%c",
		  Base.IsExon() ? 'E' : ' ',
		  Base.IsCDS() ? 'C' : ' ',
		  Base.IsIntron() ? 'I' : ' ',
		  Base.IsSplice() ? 'S' : ' ',
		  Base.IsUTR() ? 'U' : ' ',
		  Base.IsUntranscribedGene() ? 'G' : ' ');
		Log("  %c%c", Base.GetAncLetter(), Base.GetLetter());
		Log("\n");
		}
	}

void Evolver::LogTandems() const
	{
	Log("\n");
	Log("Tandems:\n");
	Log("    EvPos       AcPos  Length\n");
	Log("---------  ----------  ------\n");
	gpos_t EvPos = 0;
	FOR_EVBASES(p)
		{
		map<gpos_t, gpos_t>::const_iterator q = m_Tandems.find(p);
		if (q != m_Tandems.end())
			Log("%10"PF"  %10"PF"  %6"PF"\n", EvPos, q->first, q->second);
		++EvPos;
		}
	}

void Evolver::LogGeneIndexes() const
	{
	Log("\n");
	Log("GeneIndexes:\n");
	Log("       Pos  Index\n");
	Log("----------  -----\n");
	for (map<gpos_t, unsigned>::const_iterator p = m_GeneIndexes.begin();
	  p != m_GeneIndexes.end(); ++p)
		Log("%10"PF"  %5u\n", p->first, p->second);
	}

void Evolver::LogSizeStats() const
	{
	unsigned N = SIZE(m_Bases);
	if (N == 0)
		return;

	unsigned ListBytes = 2*N*sizeof(unsigned);
	unsigned VectorBytes = N*sizeof(BaseInfo);
	unsigned Total = ListBytes + VectorBytes;

	Log("\n");
	Log("sizeof(unsigned) %u\n", (unsigned) sizeof(unsigned));
	Log("sizeof(void *)   %u\n", (unsigned) sizeof(void *));
	Log("sizeof(float)    %u\n", (unsigned) sizeof(float));
	Log("sizeof(double)   %u\n", (unsigned) sizeof(double));
	Log("sizeof(BaseInfo) %u\n", (unsigned) sizeof(BaseInfo));
	Log("Bases            %u Mb\n", (N + 500000)/1000000);
	Log("Fwd/Bwd          %u Mb\n", (ListBytes + 500000)/1000000);
	Log("m_Bases          %u Mb\n", (VectorBytes + 500000)/1000000);
	Log("Total            %u Mb\n", (Total + 500000)/1000000);
	Log("Bytes/base      %6.1f\n", (float) Total / (float) N);
	Log("\n");
	}
