#include <time.h>
#include "evolver.h"
#include "utils.h"
#include "getopts.h"
#include "progress.h"
#include "revfile.h"
#include "myrand.h"

#ifdef _MSC_VER
#include <process.h>
#else
#include <unistd.h>
#endif

void GotAlignment(const Evolver &Evo, const AlnInfo &AI,
  const StdStrVec &Rows);
void SetModel(const char *FileName);
void LogAverageEventRates();
void LogERDetails();
void LogMatchCounts();
void CalcNucMx();
void LogEventStats(double BranchLength, const vector<double> &InitialRates,
  const vector<double> &FinalRates);
void LogRateSummary(bool Cmt);
void LogCS(const CatStats &AnCS, const CatStats &EvCS);
void LogGeneModelLengthDist(const vector<gpos_t> &InitialIntronLengthDist,
  const vector<gpos_t> &InitialExonLengthDist,
  const vector<gpos_t> &FinalIntronLengthDist,
  const vector<gpos_t> &FinalExonLengthDist);
void LogCpGStats(gpos_t AnCpGCount, gpos_t AnIslandCpGCount, gpos_t AnLength,
  gpos_t EvCpGCount, gpos_t EvIslandCpGCount, gpos_t EvLength,
  const gpos_t AnLetterCounts[4], const gpos_t AnDinucCounts[4][4],
  const gpos_t EvLetterCounts[4], const gpos_t EvDinucCounts[4][4]);
void LogRandCount();

extern set<unsigned> g_AnGeneIndexes;
extern set<unsigned> g_EvGeneIndexes;
extern unsigned g_AlnCount;
extern unsigned g_ShortAlnCount;

char *g_AncSeq;
const char *g_EvSeq;
gpos_t g_AncSeqLength;
gpos_t g_EvSeqLength;
RevFile *g_AlnRev;
bool g_ValAlnSeqs;
unsigned g_ValEvents;
unsigned g_StartValEvents;
unsigned g_ParalogousAlnCount;
unsigned g_FailEvent = UINT_MAX;
FILE *g_fStats = 0;
FILE *g_fCodonSubs = 0;

void Evolve()
	{
	const char *InSeqFileName = ValueOpt("inseq");
	const char *InEvoFileName = ValueOpt("inevo");
	const char *OutSeqFileName = ValueOpt("outseq");
	const char *AlnFileName = ValueOpt("aln");
	const char *OutEvoFileName = ValueOpt("outevo");
	unsigned InGenomeIndex = 0;
	IntOpt("genix", &InGenomeIndex);
	unsigned InChrId = 0;
	IntOpt("chrid", &InChrId);
	const char *InChrName = ValueOpt("chrname");
	const char *OutChrName = ValueOpt("outchrname");
	const char *ModelFile = ValueOpt("model");
	const char *InAnnotFileName = ValueOpt("inannots");
	const char *OutAnnotFileName = ValueOpt("outannots");
	const char *OutGenomeName = ValueOpt("outgenome");
	const char *MEFileName = ValueOpt("mes");
	if (OutGenomeName == 0)
		OutGenomeName = "evolved";
	g_ValAlnSeqs = FlagOpt("valalns");
	bool ValInit = !FlagOpt("novalinit");
	bool ValFinal = !FlagOpt("novalfinal");
	bool DoStats = !FlagOpt("nostats");
	bool LogEvents = FlagOpt("logevents");
	IntOpt("failevent", &g_FailEvent);
	const char *StatsFileName = ValueOpt("statsfile");
	if (StatsFileName != 0)
		g_fStats = CreateStdioFile(StatsFileName);
	const char *CodonSubsFileName = ValueOpt("codonsubs");
	if (CodonSubsFileName != 0)
		g_fCodonSubs = CreateStdioFile(CodonSubsFileName);

	unsigned LogEventFrom = UINT_MAX;
	unsigned LogEventTo = UINT_MAX;
	IntOpt("logeventfrom", &LogEventFrom);
	IntOpt("logeventto", &LogEventTo);

	double BranchLength = DBL_MAX;
	unsigned EventCount = UINT_MAX;
	FloatOpt("branchlength", &BranchLength);
	IntOpt("events", &EventCount);

	g_ValEvents = 0;
	g_StartValEvents = 0;
	IntOpt("valevents", &g_ValEvents);
	IntOpt("startval", &g_StartValEvents);

	if ((InSeqFileName == 0 && InEvoFileName == 0) ||
	    (InSeqFileName != 0 && InEvoFileName != 0))
		Die("Must specify exactly one of -inseq and -inevo");

	SetModel(ModelFile);

	Evolver Evo;

	gpos_t AnLetterCounts[4];
	gpos_t EvLetterCounts[4];

	gpos_t AnDinucCounts[4][4];
	gpos_t EvDinucCounts[4][4];

	if (InSeqFileName != 0)
		{
		///////////////////////////////////
		// Rev file-specific initialization
		///////////////////////////////////
		RevFile In;
		In.Open(InSeqFileName);

		if (InChrName == 0)
			InChrName = strsave(In.GetChrName(InGenomeIndex, InChrId).c_str());
		else
			InChrId = In.GetChrId(InGenomeIndex, InChrName);

		Evo.FromRevFile(In, InGenomeIndex, InChrId);

		if (g_ValAlnSeqs)
			{
			g_AncSeqLength = In.GetSeqLength(InGenomeIndex, InChrId);
			g_AncSeq = (char *) malloc(g_AncSeqLength);
			In.GetSeq(InGenomeIndex, InChrId, g_AncSeq, g_AncSeqLength);
			}

		if (InAnnotFileName != 0)
			Evo.GetAnnots(InAnnotFileName, InChrName);

		Evo.GetComposition(AnLetterCounts, AnDinucCounts);
		gpos_t ATCount = AnLetterCounts[0] + AnLetterCounts[3];
		gpos_t CGCount = AnLetterCounts[1] + AnLetterCounts[2];
		double ATFract = (double) ATCount / (double) (CGCount + ATCount);
		CalcNucMx();

		if ((BranchLength == DBL_MAX && EventCount == UINT_MAX) ||
		    (BranchLength != DBL_MAX && EventCount != UINT_MAX))
			Die("Must specify exactly one of -branchlength or -events");

		if (BranchLength != DBL_MAX)
			Evo.InitBranchLength(BranchLength);
		else if (EventCount != UINT_MAX)
			Evo.InitEventCount(EventCount);

		if (MEFileName != 0)
			Evo.ReadMEs(MEFileName);

		Evo.ComputeEventRates(ATFract);
		InitRand();

		In.Close();
		}
	else if (InEvoFileName != 0)
		{
		///////////////////////////////////
		// Evo file-specific initialization
		///////////////////////////////////
		Evo.FromFile(InEvoFileName);
		InChrId = Evo.GetAncChrId();
		InChrName = strsave(Evo.GetAncGenInfo().GetChrName(InChrId).c_str());
		if (ValueOpt("seed") != 0)
			InitRand();
		}
	else
		asserta(false);

////////////////////////////////////////
// Done with conditional .rev/.evo stuff
// Common initialization.
////////////////////////////////////////
	if (BranchLength != DBL_MAX)
		Evo.InitBranchLength(BranchLength);
	else if (EventCount != UINT_MAX)
		Evo.InitEventCount(EventCount);

	if (InSeqFileName != 0)
		Evo.InitCpG_Sweeps();

	if (OutChrName == 0)
		OutChrName = InChrName;

	CatStats AnCS;
	CatStats EvCS;

	vector<unsigned> InitialIntronLengthDist;
	vector<unsigned> FinalIntronLengthDist;

	vector<unsigned> InitialExonLengthDist;
	vector<unsigned> FinalExonLengthDist;

	vector<double> InitialRates;
	vector<double> FinalRates;

	gpos_t AnCpGCount = 0;
	gpos_t EvCpGCount = 0;
	gpos_t AnIslandCpGCount = 0;
	gpos_t EvIslandCpGCount = 0;

	Progress("Get initial rates");
	Evo.GetIntegratedRates(InitialRates);

	gpos_t AnLength = Evo.GetAncestralLength();
	if (DoStats)
		{
		Progress("Get initial composition");
		Evo.GetComposition(AnLetterCounts, AnDinucCounts);

		Progress("Get initial gene model stats");
		Evo.GetGeneModelLengthDist(InitialIntronLengthDist, InitialExonLengthDist);
		Evo.GetGeneIndexes(g_AnGeneIndexes);

		Progress("Get initial CpG stats");
		Evo.GetCpGCounts(AnCpGCount, AnIslandCpGCount);

		Progress("Get initial base category stats");
		Evo.GetCatStats(AnCS);
		Evo.LogSizeStats();
		}

	if (ValInit)
		Evo.Validate(true);

	Evo.SetLogEventFromTo(LogEventFrom, LogEventTo);
	Evo.SetLogEvents(LogEvents);

//////////////////////////////////////
	Evo.Run();
//////////////////////////////////////

	gpos_t EvLength = Evo.GetEvolvedLength();
	Progress("Evolved length %"PF" bases", EvLength);

	if (OutEvoFileName != 0)
		Evo.ToFile(OutEvoFileName);

	if (DoStats)
		{
		Progress("Get final rates");
		Evo.GetIntegratedRates(FinalRates);

		Progress("Get final gene model stats");
		Evo.GetGeneModelLengthDist(FinalIntronLengthDist, FinalExonLengthDist);
		Evo.GetGeneIndexes(g_EvGeneIndexes);

		Progress("Get final CpG stats");
		Evo.GetCpGCounts(EvCpGCount, EvIslandCpGCount);

		Progress("Get final composition stats");
		Evo.GetComposition(EvLetterCounts, EvDinucCounts);

		Progress("Get final base category stats");
		Evo.GetCatStats(EvCS);
		}

	if (ValFinal)
		{
		Progress("Validating final state");
		Evo.Validate(true);
		}

	string EvSeq;
	if (g_ValAlnSeqs || OutSeqFileName != 0)
		{
		Progress("Construct evolved sequence");
		Evo.GetEvolvedSeq(EvSeq);

		g_EvSeq = EvSeq.c_str();
		g_EvSeqLength = SIZE(EvSeq);
		asserta(EvLength == g_EvSeqLength);
		}

	GenInfoVec SeqGens(1);
	SeqGens[0].SetName(OutGenomeName);
	SeqGens[0].AddChr(OutChrName, EvLength);
	if (OutSeqFileName != 0)
		{
		Progress("Create seq file");
		RevFile SeqRev;
		SeqRev.Create(OutSeqFileName, SeqGens);
		SeqRev.AddSeq(OutGenomeName, OutChrName, EvSeq.c_str(), SIZE(EvSeq));
		SeqRev.Close();
		}

	if (AlnFileName != 0)
		{
		Progress("Generating alignments");

		GenInfoVec AlnGenomes(2);
		AlnGenomes[0] = Evo.GetAncGenInfo();
		AlnGenomes[1] = SeqGens[0];

		RevFile AlnRev;
		AlnRev.Create(AlnFileName, AlnGenomes);
		g_AlnRev = &AlnRev;

		Evo.FindAlignments(GotAlignment);

		if (g_ValAlnSeqs)
			{
			extern unsigned g_ValAlnColCount;
			Progress("%u cols validated", g_ValAlnColCount);
			EvSeq.clear();
			}
		Progress("%u alignments, %u saved, %u too short, %u paralogous",
		  g_AlnCount + g_ShortAlnCount, g_AlnCount, g_ShortAlnCount, g_ParalogousAlnCount);
		Progress("Finalizing aln file");
		AlnRev.Close();
		}

	if (OutAnnotFileName != 0)
		{
		Progress("Generating annotations");
		FILE *fOutAnnots = CreateStdioFile(OutAnnotFileName);
		Evo.GenerateAnnots(fOutAnnots, OutChrName);
		CloseStdioFile(fOutAnnots);
		}

	if (DoStats)
		{
		LogRateSummary(false);
		LogEventStats(BranchLength, InitialRates, FinalRates);

		LogERDetails();
		Evo.LogMEs();
		LogGeneModelLengthDist(InitialIntronLengthDist, InitialExonLengthDist,
		  FinalIntronLengthDist, FinalExonLengthDist);
		LogCS(AnCS, EvCS);
		LogCpGStats(AnCpGCount, AnIslandCpGCount, AnLength, EvCpGCount, EvIslandCpGCount,
		  EvLength, AnLetterCounts, AnDinucCounts, EvLetterCounts, EvDinucCounts);

		Log("\n");
		gpos_t AL = Evo.GetAncestralLength();
		gpos_t EL = Evo.GetEvolvedLength();
		Log("Ancestral length    %10"PF"\n", AL);
		Log("Evolved length      %10"PF"\n", EL);
		Log("Length change       %+10d\n", (int) EL - (int) AL);
		Log("\n");
		}
//	LogRandCount();
	}
