#include "evolver.h"
#include "readfasta.h"
#include "evomodel.h"
#include "myrand.h"
#include "ratemgr.h"
#include "gff2.h"

double Pct(double x, double T);
void SetMERates(const RateVec &Rates);

static Evolver *g_Evo;
static RateVec g_MERates;

/***
FASTA label format is >Name;Attrs, where Attrs are in GTF format:

	rate "<r>"; avgdel "<a>"; stddev "<s>"; pct "<p>";
or,
	rateperbase "<r>"; avgdel "<a>"; stddev "<s>"; pct "<p>";

Total ME insert rate is in units of insertions per tick, where an insertion
means a complete ME.

The Rate field in the ME structure is in the same units, i.e. insertions of
that ME per tick. This value is either the 'rate' attribute from the FASTA
label, or 'rateperbase' / length.
***/

static void ParseMELabel(const string &Label, MEInfo &ME, unsigned SeqLength)
	{
	size_t n = Label.find(";");
	if (n == string::npos)
		Die("Invalid ME label, missing '[:;]<params>'", Label.c_str());

	if (n == 0)
		Die("Invalid ME label, zero-length name in label '%s'", Label.c_str());

	ME.Name = Label.substr(0, n);
	string Attrs = Label.substr(n + 1, string::npos);

	string Value;
	ME.RateSpecified = GFFGetAttr(Attrs, "rate", Value);
	ME.RatePerBaseSpecified = GFFGetAttr(Attrs, "rateperbase", Value);
	if (ME.RateSpecified + ME.RatePerBaseSpecified != 1)
		Die("Invaid ME label, must specify exactly one of rate or rateperbase '%s'",
		  Label.c_str());

	double Rate = atof(Value.c_str());
	ME.Rate = Rate;
	if (ME.RateSpecified)
		ME.RateParam = Rate;
	else if (ME.RatePerBaseSpecified)
		ME.RatePerBaseParam = Rate;

	bool Found = GFFGetAttr(Attrs, "avgdel", Value);
	if (Found)
		{
		ME.MeanDeleteLength = atof(Value.c_str());
		if (ME.MeanDeleteLength <= 0.5)
			Die("Invalid ME label, avgdel '%s'", Value.c_str());

		Found = GFFGetAttr(Attrs, "stddev", Value);
		if (!Found)
			Die("Invalid ME label, 'avgdel' specified but not 'stddev' in '%s'",
			  Label.c_str());
		ME.StdDevDeleteLength = atof(Value.c_str());
		if (ME.StdDevDeleteLength <= 0.0)
			Die("Invalid ME label, stddev '%s'", Value.c_str());

		ME.FractWithDeletion = 1.0;
		Found = GFFGetAttr(Attrs, "pct", Value);
		if (Found)
			{
			ME.FractWithDeletion = atof(Value.c_str())/100.0;
			if (ME.FractWithDeletion < 0 || ME.FractWithDeletion > 1.0)
				Die("Invalid ME label, 'pct' %s", Value.c_str());
			}
		}
	if (ME.RatePerBaseSpecified)
		{
		double AvgInsertedSeqLength = SeqLength - ME.MeanDeleteLength*ME.FractWithDeletion;
		if (AvgInsertedSeqLength < 1.0)
			Die("ME '%s', average inserted length=%g (length-avgdel*pct/100)",
			  ME.Name.c_str(), AvgInsertedSeqLength);
		ME.Rate /= AvgInsertedSeqLength;
		}
	}

void GotMESeq(const string &Label, const string &Seq)
	{
	MEInfo ME;
	ME.Seq = Seq;
	ParseMELabel(Label, ME, SIZE(Seq));
	g_Evo->AddME(ME);
	}

void Evolver::ReadMEs(const char *FileName)
	{
	m_MEs.clear();
	g_MERates.clear();
	g_Evo = this;
	ReadFasta(FileName, GotMESeq);
	SetMERandThreshs();
	SetMERates(g_MERates);
	}

void Evolver::AddME(const MEInfo &ME)
	{
	m_MEs.push_back(ME);
	if (ME.Rate > 0)
		{
		const unsigned N = SIZE(g_MERates);
		for (unsigned i = 0; i < N; ++i)
			{
			pair<gpos_t, double> &LengthRate = g_MERates[i];
			if (LengthRate.first == SIZE(ME.Seq))
				{
				LengthRate.second += ME.Rate;
				return;
				}
			}
		g_MERates.push_back(pair<gpos_t, double>(SIZE(ME.Seq), ME.Rate));
		}
	else
		Warning("ME %s insert rate zero", ME.Name.c_str());
	}

void Evolver::LogMEs() const
	{
	if (m_MEs.empty())
		{
		Log("\n");
		Log("No mobile elements defined.\n");
		Log("\n");
		return;
		}

	Log("\n");
	Log("Mobile elements:\n");
	Log("Index              Name        Rate        rate  rateprbase     pct  avgdel  stddev  Deletes    Accepted   Acc.Bases  Sequence\n");
	Log("=====  ================  ==========  ==========  ==========  ======  ======  ======  =======  ==========  ==========  ========\n");
	const unsigned N = SIZE(m_MEs);
	double TotalRate = GetMERate();
	unsigned TotalBases = 0;
	unsigned TotalAccepts = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		Log("%5u", i);
		const MEInfo &ME = m_MEs[i];
		Log("  %16.16s", ME.Name.c_str());
		Log("  %10.3g", ME.Rate);
		if (ME.RateSpecified)
			Log("  %10.3g", ME.RateParam);
		else
			Log("  %10.10s", "");
		if (ME.RatePerBaseSpecified)
			Log("  %10.3g", ME.RatePerBaseParam);
		else
			Log("  %10.10s", "");
		Log("  %5.1f%%", ME.FractWithDeletion*100.0);
		Log("  %6.0f", ME.MeanDeleteLength);
		Log("  %6.1f", ME.StdDevDeleteLength);
		Log("  %7u", ME.DeleteCount);
		Log("  %10u", ME.AcceptedInsertCount);
		unsigned Bases = ME.InsertedBaseCount;
		Log("  %10u", Bases);
		if (SIZE(ME.Seq) <= 32)
			Log("  %s", ME.Seq.c_str());
		else
			Log("  %32.32s...", ME.Seq.c_str());
		Log("\n");

		TotalBases += Bases;
		TotalAccepts += ME.AcceptedInsertCount;
		}
	Log("=====                    ==========                                                           ==========  ==========\n");
	Log("%5u                    %10.3g                                                           %10u  %10u (%.1f%% of genome)\n",
	  N,
	  TotalRate,
	  TotalAccepts,
	  TotalBases,
	  Pct(TotalBases, GetEvolvedLength()));
	Log("\n");
	}

double Evolver::GetMERate() const
	{
	const unsigned N = SIZE(m_MEs);
	double TotalRate = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		const MEInfo &ME = m_MEs[i];
		TotalRate += ME.Rate;
		}
	return TotalRate;
	}

void Evolver::SetMERandThreshs()
	{
	double TotalRate = GetMERate();

	const unsigned N = SIZE(m_MEs);
	double AccumRate = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		MEInfo &ME = m_MEs[i];
		if (ME.Rate == 0)
			ME.RandThresh = UINT_MAX;
		else
			{
			AccumRate += ME.Rate;
			ME.RandThresh = (unsigned) (AccumRate*UINT_MAX/TotalRate + 0.5);
			unsigned PrevRandThresh = (i == 0 ? 0 : m_MEs[i-1].RandThresh);
			double BinRate = TotalRate*(ME.RandThresh - PrevRandThresh)/(double) UINT_MAX;
			double Pct(double,double);
			double p = Pct(BinRate/ME.Rate - 1, 1);
			if (p > P_MaxRateApproxErrPct)
				Die("Bin err %.1f%% exceeds max=%.1f%% for ME %s",
				  p, P_MaxRateApproxErrPct, ME.Name.c_str());
			}
		}
	}

MEInfo &Evolver::GetRandomME()
	{
	unsigned r = RandInt(UINT_MAX);
	const unsigned N = SIZE(m_MEs);
	for (unsigned i = 0; i < N; ++i)
		{
		MEInfo &ME = m_MEs[i];
		if (r <= ME.RandThresh)
			return ME;
		}
	Die("GetRandomME() failed");
	UNREACHABLE_RETURN(*((MEInfo *) 0));
	}
