#include "evomodel.h"
#include "utils.h"
#include "myrand.h"

/***
I struggled to define a method for accelerating
and decelerating probabilities. We need a function
for which the domain and range are both [0,1].

Qualitatively, the acceleration function F(p)
should look something like this:


1 |                     aaaa
  |             a          d
  |        a              d
  |      a               d
F |    a                d
  |   a              d
  |  a           d
  | a     d  
0 |addd
   __________________________
   0           p            1

where a's are for accelerating and d's are for
decelerating. Preferably we should be able to
show that a random series of accelerations and
decelerations leaves the distribution of a set
of probabilities approximately unchanged.

I provisionally abandoned an attempt to do this
in a principled analytic way and instead chose
the following scheme. A factor a defines the
acceleration (a > 1) or deceleration (a < 1).
In a range starting at the origin, F(p) = fp.
The function is defined to be symmetric about
the line F = 1 - p, which means that F is
composed of two straight lines which are joined
by an apex at F = 1 - p, i.e. at p=1/(1+a)
and F=a/(1+a).

A hand-waving argument claims that this function
multiplies by a given factor in a maximum range
consistent with a symmetry between acceleration
and deceleration. E.g., with a=2, as much of
the range as reasonable becomes twice as fast.

1 |.                 a
  |  .           a
  |    .     a
  |      .a
F |     a  .
  |    a     .
  |   a        .
  |  a           .F=1-p
0 | a F=ap         .
   ___________________
   0           p     1

***/

double AccelProb(double p, double a)
	{
	if (p <= 1/(1 + a))
		return a*p;
	else
		{
		double P = p - 1/(1 + a);
		return a/(1 + a) + P/a;
		}
	}

static double GetAccelFactor()
	{
	return 1 + RandNorm(P_AccelIncrementMean, P_AccelIncrementStdDev);
	}

double GetAccel()
	{
	double a = GetAccelFactor();
	return RandTrue(0.5) ? a : 1/a;
	}

double GetFasterAccel()
	{
	return GetAccelFactor();
	}

double GetSlowerAccel()
	{
	return 1.0/GetAccelFactor();
	}
