#include "utils.h"
#include <ctype.h>
#include "fastastr.h"
#include "xlat.h"

#define TRACE		0

FastaStream::FastaStream()
	{
	m_f = 0;
	m_FilePos = 0;
	m_FileSize = 0;
	m_BytesTransferred = 0;
	m_BytesInBuffer = 0;
	m_BufferPos = 0;
	m_WordLength = 0;
	m_Word = 0;
	m_CharCount = 0;
	m_Label = false;
	m_Buffer = 0;
	m_LastChar = 0;
	m_TotalChars = 0;
// Default is untranslated nucleotide input.
// m_Xlat is true if nucleotide input translated into amino
	m_Xlat = false;
// m_Amino is true if input is amino
	m_Amino = false;
	}

FastaStream::~FastaStream()
	{
	if (m_f != 0)
		Close();
	}

void FastaStream::Close()
	{
	if (m_f != 0)
		CloseStdioFile(m_f);
	m_f = 0;
	m_FileSize = 0;
	m_FilePos = 0;
	m_BytesTransferred = 0;
	m_BytesInBuffer = 0;
	m_BufferPos = 0;
	m_WordLength = 0;
	m_Label = false;
	m_Word = 0;
	m_CharCount = 0;
	m_LastChar = 0;
	if (m_Buffer != 0)
		myfree(m_Buffer);
	m_Buffer = 0;
	}

void FastaStream::Open(const char *FileName, unsigned WordLength)
	{
	//if (WordLength != 0 && WordLength%2 == 0)
	//	Die("FastaStream::Open, word length=%u, must be odd", WordLength);
	if (WordLength > 17)
		Die("FastaStream::Open, word length=%u, max=17");

	m_f = OpenStdioFile(FileName);
	m_FileSize = GetStdioFileSize(m_f);
	if (m_FileSize == 0)
		Die("File %s is empty", FileName);
	m_FilePos = 0;
	m_BytesTransferred = 0;
	m_BytesInBuffer = 0;
	m_BufferPos = 0;
	m_Word = 0;
	m_CharCount = 0;
	m_Label = false;
	m_WordLength = WordLength;
	m_Mask = 0;
	m_20w = pow20(WordLength);
	m_LastChar = 0;
	for (unsigned i = 0; i < m_WordLength; ++i)
		m_Mask |= (0x3 << (2*i));
	m_Buffer = myalloc(char, BUFFER_SIZE);
	}

bool FastaStream::GetNextSeqChar(unsigned char *ptrc)
	{
	for (;;)
		{
		unsigned char c;
		bool Ok = GetNextChar(&c);
		if (!Ok)
			return false;
		if (m_Label)
			{
			if (c == '\n')
				m_Label = false;
			}
		else
			{
			if (c == '>')
				{
				m_Label = true;
				continue;
				}
			else if (isalpha(c))
				{
				*ptrc = c;
				return true;
				}
			else if (!isspace(c))
				{
				if (isprint(c))
					Warning("Invalid char '%c' in sequence data at file position %s bytes",
					  c, uint64tostr(m_FilePos));
				else
					Warning("Invalid byte 0x%02x in sequence data at file position %s bytes",
					  c, uint64tostr(m_FilePos));
				}
			}
		}
	}

bool FastaStream::GetNextChar(unsigned char *ptrc)
	{
	if (m_FileSize == 0)
		return false;

	if (m_BufferPos == m_BytesInBuffer)
		{
		unsigned BytesToRead = BUFFER_SIZE;
		off_t BytesLeft = m_FileSize - m_FilePos;
		if (BytesLeft < (off_t) BytesToRead)
			BytesToRead = (unsigned) BytesLeft;
		if (BytesToRead == 0)
			{
			m_LastChar = (unsigned char) 0xff;
			return false;
			}
		ReadStdioFile(m_f, m_Buffer, BytesToRead);
		m_BufferPos = 0;
		m_BytesInBuffer = BytesToRead;
		}
	unsigned char c = (unsigned char) m_Buffer[m_BufferPos++];
	++m_FilePos;
	++m_TotalChars;
	*ptrc = c;
	m_LastChar = c;
	return true;
	}

bool FastaStream::GetNextNucWord(uint64 *ptrWord)
	{
	if (m_WordLength == 0)
		Die("FastaStream::GetNextWord, word size not specified");

	while (m_CharCount < m_WordLength)
		{
		unsigned char c;
		bool Ok = GetNextSeqChar(&c);
		if (!Ok)
			return false;

		if (!isacgt(c))
			{
			m_Word = 0;
		// Reset loop to start next word
		// past non-AGCT character.
			m_CharCount = 0;
			continue;
			}

		m_Word <<= 2;
		m_Word |= NucCode[c];
		++m_CharCount;
		}

	--m_CharCount;
	m_Word &= m_Mask;
	*ptrWord = m_Word;

#if	TRACE
	{
	Log("%s", NucWordToStr(m_Word, m_WordLength));
	Log("  %x", m_Word);
	Log("\n");
	}
#endif
	return true;
	}

bool FastaStream::GetNextAminoWord(uint64 *ptrWord)
	{
	if (m_WordLength == 0)
		Die("FastaStream::GetNextWord, word size not specified");

	while (m_CharCount < m_WordLength)
		{
		unsigned char c;
		bool Ok = GetNextSeqChar(&c);
		if (!Ok)
			return false;

		if (!isamino(c))
			{
			m_Word = 0;
		// Reset loop to start next word
		// past non-AGCT character.
			m_CharCount = 0;
			continue;
			}

		m_Word *= 20;
		m_Word += AminoCode[c];
		++m_CharCount;
		}

	--m_CharCount;
	m_Word %= m_20w;
	*ptrWord = m_Word;

#if	TRACE
	{
	Log("%s", AminoWordToStr(m_Word, m_WordLength));
	Log("  %x", m_Word);
	Log("\n");
	}
#endif
	return true;
	}

bool FastaStream::GetNextSeq(std::vector<char> &Label, std::vector<unsigned char> &Seq)
	{
	if (m_LastChar == 0)
		{
		unsigned char c;
		bool Ok = GetNextChar(&c);
		if (!Ok)
			return false;
		if (c != '>')
			{
			if (isprint(c))
				Die("FASTA file must start with '>', found '%c'", c);
			else
				Die("FASTA file must start with '>', found 0x02x", c);
			}
		}
	else if (m_LastChar == 0xff)
		return false;
	else
		{
		if (m_LastChar != '>')
			{
			if (isprint(m_LastChar))
				Die("FastaStream::GetNextSeq, LastChar='%c'", m_LastChar);
			else
				Die("FastaStream::GetNextSeq, LastChar=%0x02x", m_LastChar);
			}
		}
	Label.clear();
	Label.reserve(4096);	// 4096 is arbitrary choice to reduce thrashing
	for (;;)
		{
		unsigned char c;
		bool Ok = GetNextChar(&c);
		if (!Ok)
			return false;
		if (c == '\n')
			break;
		Label.push_back(c);
		}

	Seq.clear();
	Seq.reserve(10000000);	// 10 Mb is arbitrary choice to reduce thrashing
	for (;;)
		{
		unsigned char c;
		bool Ok = GetNextChar(&c);
		if (!Ok)
			return Seq.size() > 0;
		if (c == '>')
			break;
		if (isprint(c) && !isspace(c))
			Seq.push_back(c);
		}

	return true;
	}
