#include "revfile.h"
#include "getopts.h"
#include "progress.h"
#include "utils.h"
#include "filter4.h"

static RevFile *g_RevFileIn;

// Genome indexes, fixed by command line,
// do not vary while processing.
static unsigned g_GenomeIndexA = UINT_MAX;
static unsigned g_GenomeIndexB = UINT_MAX;

// Chr ids currently being processed.
static unsigned g_ChrIdA = UINT_MAX;
static unsigned g_ChrIdB = UINT_MAX;

// Bundle currently being processed.
static unsigned g_FilterBundleId = UINT_MAX;

// Value of -bundle <b>, if specified
// (requests that a single bundle be processed).
static unsigned g_CmdLineBundleId = UINT_MAX;

static bool g_InvertMinusBundles = false;

static void DoNothing()
	{
	}

static PTR_VOID_FN_VOID g_ptrBeforeChrPairCallback = DoNothing;
static PTR_VOID_FN_VOID g_ptrAfterChrPairCallback = DoNothing;
static PTR_VOID_FN_VOID g_ptrAfterAllChrPairsCallback = DoNothing;

void SetBeforeChrPairCallback(PTR_VOID_FN_VOID ptrFn)
	{
	g_ptrBeforeChrPairCallback = ptrFn;
	}

void SetAfterChrPairCallback(PTR_VOID_FN_VOID ptrFn)
	{
	g_ptrAfterChrPairCallback = ptrFn;
	}

void SetAfterAllChrPairsCallback(PTR_VOID_FN_VOID ptrFn)
	{
	g_ptrAfterAllChrPairsCallback = ptrFn;
	}

void BeforeChrPair_()
	{
	(*g_ptrBeforeChrPairCallback)();
	}

void AfterChrPair_()
	{
	(*g_ptrAfterChrPairCallback)();
	}

static void AfterAllChrPairs_()
	{
	(*g_ptrAfterAllChrPairsCallback)();
	}

const RevFile &GetFilterInputRevFile()
	{
	if (g_RevFileIn == 0)
		Die("g_RevFileIn == 0");
	return *g_RevFileIn;
	}

unsigned GetFilterBundleId()
	{
	return g_FilterBundleId;
	}

void SetFilterBundleId(unsigned BundleId)
	{
	g_FilterBundleId = BundleId;
	}

unsigned GetCmdLineBundleId()
	{
	return g_CmdLineBundleId;
	}

unsigned GetFilterGenomeIndexA()
	{
	return g_GenomeIndexA;
	}

unsigned GetFilterGenomeIndexB()
	{
	return g_GenomeIndexB;
	}

unsigned GetFilterChrIdA()
	{
	return g_ChrIdA;
	}

unsigned GetFilterChrIdB()
	{
	return g_ChrIdB;
	}

void SetFilterChrIdA(unsigned ChrId)
	{
	g_ChrIdA = ChrId;
	}

void SetFilterChrIdB(unsigned ChrId)
	{
	g_ChrIdB = ChrId;
	}

const ChrInfo &GetFilterChrInfoA()
	{
	unsigned GenomeIndexA = GetFilterGenomeIndexA();
	unsigned ChrIdA = GetFilterChrIdA();
	const RevFile &In = GetFilterInputRevFile();
	return In.GetChrInfo(GenomeIndexA, ChrIdA);
	}

const ChrInfo &GetFilterChrInfoB()
	{
	unsigned GenomeIndexB = GetFilterGenomeIndexB();
	unsigned ChrIdB = GetFilterChrIdB();
	const RevFile &In = GetFilterInputRevFile();
	return In.GetChrInfo(GenomeIndexB, ChrIdB);
	}

void SetInvertMinusBundles(bool Invert)
	{
	g_InvertMinusBundles = Invert;
	}

bool GetInvertMinusBundles()
	{
	return g_InvertMinusBundles;
	}

void FilterMakePairwiseAlnInfo(coord_t Alo, coord_t Ahi, coord_t Blo, coord_t Bhi,
  bool Plus, AlnInfo &AI, AlnAttr Attr, unsigned Value)
	{
	AI.Segs.clear();
	AI.Attrs.clear();
	AI.Values.clear();
	AI.Score = 0;

	GCSeg Seg;
	memset(&Seg, 0, sizeof(Seg));

	Seg.GenomeIndex = GetFilterGenomeIndexA();
	Seg.ChrId = GetFilterChrIdA();
	Seg.lo = Alo;
	Seg.hi = Ahi;
	Seg.Plus = true;

	AI.Segs.push_back(Seg);

	Seg.GenomeIndex = GetFilterGenomeIndexB();
	Seg.ChrId = GetFilterChrIdB();
	Seg.lo = Blo;
	Seg.hi = Bhi;
	Seg.Plus = Plus;

	AI.Segs.push_back(Seg);

	if (Attr != ALN_ATTR_None)
		{
		AI.Attrs.push_back(Attr);
		AI.Values.push_back(Value);
		}
	}

int main(int argc, char *argv[])
	{
	ProcessArgVect(argc, argv);

	const char *InFileName = RequiredValueOpt("in");
	extern bool g_ShowProgress;
	g_ShowProgress = !FlagOpt("quiet");

	IntOpt("bundle", &g_CmdLineBundleId);

	bool OptimizeTime = true;
	const char *Favor = ValueOpt("favor");
	if (Favor != 0)
		{
		if (streq(Favor, "time") || streq(Favor, "speed"))
			OptimizeTime = true;
		else if (streq(Favor, "space"))
			OptimizeTime = false;
		else
			Die("Invalid -favor, must be time or space");
		}

	RevFile In;
	In.Open(InFileName);
	g_RevFileIn = &In;

	g_ChrIdA = UINT_MAX;
	g_ChrIdB = UINT_MAX;
	bool DoAllChrPairs = GetRevChrPairOpts(In, &g_GenomeIndexA, &g_GenomeIndexB,
	  &g_ChrIdA, &g_ChrIdB);

	InitFilter4();

	const std::string &GenomeName1 = In.GetGenomeName(g_GenomeIndexA);
	const std::string &GenomeName2 = In.GetGenomeName(g_GenomeIndexB);

	unsigned ChrCount1 = In.GetChrCount(g_GenomeIndexA);
	unsigned ChrCount2 = In.GetChrCount(g_GenomeIndexB);

	Progress("%.16s has %u chrs", GenomeName1.c_str(), ChrCount1);
	Progress("%.16s has %u chrs", GenomeName2.c_str(), ChrCount2);

	if (!DoAllChrPairs)
		FilterChrPair4(In);
	else if (OptimizeTime)
		FilterAllChrPairs4(In);
	else
		{
		std::set<IntPair> ChrIdPairs;
		Progress("Get active chr pairs");
		In.GetActiveChrPairs(g_GenomeIndexA, g_GenomeIndexB, ChrIdPairs);
		unsigned PairCount = (unsigned) ChrIdPairs.size();
		Progress("%u pairs", PairCount);
		unsigned PairIndex = 0;
		for (std::set<IntPair>::const_iterator p = ChrIdPairs.begin();
		  p != ChrIdPairs.end(); ++p)
			{
			const IntPair &ChrPair = *p;

			g_ChrIdA = ChrPair.first;
			g_ChrIdB = ChrPair.second;

			const std::string &ChrName1 = In.GetChrName(g_GenomeIndexA, g_ChrIdA);
			const std::string &ChrName2 = In.GetChrName(g_GenomeIndexB, g_ChrIdB);

			++PairIndex;
			double Pct = (double) PairIndex*100.0 / (double) PairCount;

			Progress("%.16s[%u], %.16s[%u] %u of %u chr pairs (%5.2f%%)",
			  ChrName1.c_str(),
			  g_ChrIdA,
			  ChrName2.c_str(),
			  g_ChrIdB,
			  PairIndex,
			  PairCount,
			  Pct);
			
			FilterChrPair4(In);

			++PairIndex;
			}
		}

	AfterAllChrPairs_();

	In.Close();
	g_RevFileIn = 0;

	ProgressExit();

	return 0;
	}
