#include "evo.h"
#include "gff2.h"

static FILE *g_fOut;
static FILE *g_fOutDNA;
static FILE *g_fGFF;
static bool g_Xlat = false;
static unsigned g_MinCodons = 30;
static bool g_SameStrand = false;
static unsigned g_MaxORFs = UINT_MAX;
static bool g_Dump = false;
static unsigned g_GeneIndex = 0;

const int BLOCK_LENGTH = 20;

enum CODON2
	{
	CODON2_Undefined,
	CODON2_Unknown,
	CODON2_DoneFrame,
	CODON2_STP,
	CODON2_Start,
	CODON2_Ala,
	CODON2_Cys,
	CODON2_Asp,
	CODON2_Glu,
	CODON2_Phe,
	CODON2_Gly,
	CODON2_His,
	CODON2_Ile,
	CODON2_Lys,
	CODON2_Leu,
	CODON2_Asn,
	CODON2_Pro,
	CODON2_Gln,
	CODON2_Arg,
	CODON2_Ser,
	CODON2_Thr,
	CODON2_Val,
	CODON2_Trp,
	CODON2_Tyr,
	CODON2_Met = CODON2_Start,
	};

char Complement(char c)
	{
	switch (toupper(c))
		{
	case 'A': return 'T';
	case 'T': return 'A';
	case 'G': return 'C';
	case 'C': return 'G';
	case '#': return '#';
		}
	return 'N';
	}

// TTT Phe  TCT Ser  TAT Tyr  TGT Cys  
// TTC Phe  TCC Ser  TAC Tyr  TGC Cys  
// TTA Leu  TCA Ser  TAA STP  TGA STP  
// TTG Leu  TCG Ser  TAG STP  TGG Trp  
// CTT Leu  CCT Pro  CAT His  CGT Arg  
// CTC Leu  CCC Pro  CAC His  CGC Arg  
// CTA Leu  CCA Pro  CAA Gln  CGA Arg  
// CTG Leu  CCG Pro  CAG Gln  CGG Arg  
// ATT Ile  ACT Thr  AAT Asn  AGT Ser  
// ATC Ile  ACC Thr  AAC Asn  AGC Ser  
// ATA Ile  ACA Thr  AAA Lys  AGA Arg  
// ATG Met  ACG Thr  AAG Lys  AGG Arg  
// GTT Val  GCT Ala  GAT Asp  GGT Gly  
// GTC Val  GCC Ala  GAC Asp  GGC Gly  
// GTA Val  GCA Ala  GAA Glu  GGA Gly  
// GTG Val  GCG Ala  GAG Glu  GGG Gly  
static CODON2 GetCodonLetters(char c1, char c2, char c3)
	{
	if (c1 == '#' || c2 == '#' || c3 == '#')
		return CODON2_DoneFrame;
#define THREE_LETTERS(c1, c2, c3)	((c1) | ((c2) << 8) | ((c3) << 16))
	const int i = THREE_LETTERS(toupper(c1), toupper(c2), toupper(c3));
	switch (i)
		{
	case THREE_LETTERS('T', 'T', 'T'): return CODON2_Phe;
	case THREE_LETTERS('T', 'T', 'C'): return CODON2_Phe;
	case THREE_LETTERS('T', 'T', 'A'): return CODON2_Leu;
	case THREE_LETTERS('T', 'T', 'G'): return CODON2_Leu;
	case THREE_LETTERS('C', 'T', 'T'): return CODON2_Leu;
	case THREE_LETTERS('C', 'T', 'C'): return CODON2_Leu;
	case THREE_LETTERS('C', 'T', 'A'): return CODON2_Leu;
	case THREE_LETTERS('C', 'T', 'G'): return CODON2_Leu;
	case THREE_LETTERS('A', 'T', 'T'): return CODON2_Ile;
	case THREE_LETTERS('A', 'T', 'C'): return CODON2_Ile;
	case THREE_LETTERS('A', 'T', 'A'): return CODON2_Ile;
	case THREE_LETTERS('A', 'T', 'G'): return CODON2_Met;
	case THREE_LETTERS('G', 'T', 'T'): return CODON2_Val;
	case THREE_LETTERS('G', 'T', 'C'): return CODON2_Val;
	case THREE_LETTERS('G', 'T', 'A'): return CODON2_Val;
	case THREE_LETTERS('G', 'T', 'G'): return CODON2_Val;
	case THREE_LETTERS('T', 'C', 'T'): return CODON2_Ser;
	case THREE_LETTERS('T', 'C', 'C'): return CODON2_Ser;
	case THREE_LETTERS('T', 'C', 'A'): return CODON2_Ser;
	case THREE_LETTERS('T', 'C', 'G'): return CODON2_Ser;
	case THREE_LETTERS('C', 'C', 'T'): return CODON2_Pro;
	case THREE_LETTERS('C', 'C', 'C'): return CODON2_Pro;
	case THREE_LETTERS('C', 'C', 'A'): return CODON2_Pro;
	case THREE_LETTERS('C', 'C', 'G'): return CODON2_Pro;
	case THREE_LETTERS('A', 'C', 'T'): return CODON2_Thr;
	case THREE_LETTERS('A', 'C', 'C'): return CODON2_Thr;
	case THREE_LETTERS('A', 'C', 'A'): return CODON2_Thr;
	case THREE_LETTERS('A', 'C', 'G'): return CODON2_Thr;
	case THREE_LETTERS('G', 'C', 'T'): return CODON2_Ala;
	case THREE_LETTERS('G', 'C', 'C'): return CODON2_Ala;
	case THREE_LETTERS('G', 'C', 'A'): return CODON2_Ala;
	case THREE_LETTERS('G', 'C', 'G'): return CODON2_Ala;
	case THREE_LETTERS('T', 'A', 'T'): return CODON2_Tyr;
	case THREE_LETTERS('T', 'A', 'C'): return CODON2_Tyr;
	case THREE_LETTERS('T', 'A', 'A'): return CODON2_STP;
	case THREE_LETTERS('T', 'A', 'G'): return CODON2_STP;
	case THREE_LETTERS('C', 'A', 'T'): return CODON2_His;
	case THREE_LETTERS('C', 'A', 'C'): return CODON2_His;
	case THREE_LETTERS('C', 'A', 'A'): return CODON2_Gln;
	case THREE_LETTERS('C', 'A', 'G'): return CODON2_Gln;
	case THREE_LETTERS('A', 'A', 'T'): return CODON2_Asn;
	case THREE_LETTERS('A', 'A', 'C'): return CODON2_Asn;
	case THREE_LETTERS('A', 'A', 'A'): return CODON2_Lys;
	case THREE_LETTERS('A', 'A', 'G'): return CODON2_Lys;
	case THREE_LETTERS('G', 'A', 'T'): return CODON2_Asp;
	case THREE_LETTERS('G', 'A', 'C'): return CODON2_Asp;
	case THREE_LETTERS('G', 'A', 'A'): return CODON2_Glu;
	case THREE_LETTERS('G', 'A', 'G'): return CODON2_Glu;
	case THREE_LETTERS('T', 'G', 'T'): return CODON2_Cys;  
	case THREE_LETTERS('T', 'G', 'C'): return CODON2_Cys;  
	case THREE_LETTERS('T', 'G', 'A'): return CODON2_STP;  
	case THREE_LETTERS('T', 'G', 'G'): return CODON2_Trp;  
	case THREE_LETTERS('C', 'G', 'T'): return CODON2_Arg;  
	case THREE_LETTERS('C', 'G', 'C'): return CODON2_Arg;  
	case THREE_LETTERS('C', 'G', 'A'): return CODON2_Arg;  
	case THREE_LETTERS('C', 'G', 'G'): return CODON2_Arg;  
	case THREE_LETTERS('A', 'G', 'T'): return CODON2_Ser;  
	case THREE_LETTERS('A', 'G', 'C'): return CODON2_Ser;  
	case THREE_LETTERS('A', 'G', 'A'): return CODON2_Arg;  
	case THREE_LETTERS('A', 'G', 'G'): return CODON2_Arg;  
	case THREE_LETTERS('G', 'G', 'T'): return CODON2_Gly;  
	case THREE_LETTERS('G', 'G', 'C'): return CODON2_Gly;  
	case THREE_LETTERS('G', 'G', 'A'): return CODON2_Gly;  
	case THREE_LETTERS('G', 'G', 'G'): return CODON2_Gly;  
		}
#undef THREE_LETTERS
	return CODON2_Unknown;
	}

static char *GetThreeLetters(const char Seq[], int Frame, int Pos)
	{
	static char Str[4];
	assert(Frame >= -3 && Frame <= 2);
	if (Frame < 0)
		{
		assert(Pos >= 2);
		Str[0] = Complement(Seq[Pos]);
		Str[1] = Complement(Seq[Pos-1]);
		Str[2] = Complement(Seq[Pos-2]);
		}
	else
		{
		Str[0] = Seq[Pos];
		Str[1] = Seq[Pos+1];
		Str[2] = Seq[Pos+2];
		}
	return Str;
	}

static CODON2 GetCodon(const char Seq[], int Frame, int Pos)
	{
	const char *s = GetThreeLetters(Seq, Frame, Pos);
	return GetCodonLetters(s[0], s[1], s[2]);
	}

static CODON2 GetCodonStr(const char *s)
	{
	return GetCodonLetters(s[0], s[1], s[2]);
	}

static char *CodonToStr(CODON2 c)
	{
	switch (c)
		{
	case CODON2_Undefined:	return "UND";
	case CODON2_Unknown:		return "UNK";
	case CODON2_STP:			return "STP";
	case CODON2_Ala: return "Ala";
	case CODON2_Cys: return "Cys";
	case CODON2_Asp: return "Asp";
	case CODON2_Glu: return "Glu";
	case CODON2_Phe: return "Phe";
	case CODON2_Gly: return "Gly";
	case CODON2_His: return "His";
	case CODON2_Ile: return "Ile";
	case CODON2_Lys: return "Lys";
	case CODON2_Leu: return "Leu";
	case CODON2_Met: return "Met";
	case CODON2_Asn: return "Asn";
	case CODON2_Pro: return "Pro";
	case CODON2_Gln: return "Gln";
	case CODON2_Arg: return "Arg";
	case CODON2_Ser: return "Ser";
	case CODON2_Thr: return "Thr";
	case CODON2_Val: return "Val";
	case CODON2_Trp: return "Trp";
	case CODON2_Tyr: return "Tyr";
	default:
		break;
		}
	return "ERR";
	}

const char *CharsToAA(char c1, char c2, char c3)
	{
	CODON2 Codon = GetCodonLetters(c1, c2, c3);
	return CodonToStr(Codon);
	}

static char CodonToAminoChar(CODON2 c)
	{
	switch (c)
		{
	case CODON2_Unknown:		return 'X';
	case CODON2_Ala: return 'A';
	case CODON2_Cys: return 'C';
	case CODON2_Asp: return 'D';
	case CODON2_Glu: return 'E';
	case CODON2_Phe: return 'F';
	case CODON2_Gly: return 'G';
	case CODON2_His: return 'H';
	case CODON2_Ile: return 'I';
	case CODON2_Lys: return 'K';
	case CODON2_Leu: return 'L';
	case CODON2_Met: return 'M';
	case CODON2_Asn: return 'N';
	case CODON2_Pro: return 'P';
	case CODON2_Gln: return 'Q';
	case CODON2_Arg: return 'R';
	case CODON2_Ser: return 'S';
	case CODON2_Thr: return 'T';
	case CODON2_Val: return 'V';
	case CODON2_Trp: return 'W';
	case CODON2_Tyr: return 'Y';
	default:
		break;
		}
	Warning("CodonToAminoChar(%s): Invalid codon", CodonToStr(c));
	return '!';
	}

static int GetFrameStart(int Frame, int SeqLength)
	{
	assert(Frame >= -3 && Frame <= 2);
	return Frame < 0 ? SeqLength - (-Frame) : Frame;
	}

static bool InFrame(int Frame, int SeqLength, int i)
	{
	assert(Frame >= -3 && Frame <= 2);
	return Frame < 0 ? (i >= 2) : (i <= SeqLength - 3);
	}

static int NextFramePos(int Frame, int i)
	{
	assert(Frame >= -3 && Frame <= 2);
	return Frame < 0 ? i - 3 : i + 3;
	}

static void FindORFsSeq(const char Seq[], int SeqLength, int *ptrFrame,
  int *ptrStart, int *ptrLength)
	{
	int BestORFFrame = -1;
	int BestORFStart = -1;
	int BestORFLength = -1;

	for (int Frame = -3; Frame <= 2; ++Frame)
		{
		int ORFStart = -1;
		int FrameStart = GetFrameStart(Frame, SeqLength);
		if (g_Dump)
			{
			Log("\n");
			Log("Frame %d start %d\n", Frame, FrameStart);
			}
		for (int i = FrameStart; InFrame(Frame, SeqLength, i);
		  i = NextFramePos(Frame, i))
			{
			CODON2 Codon = GetCodon(Seq, Frame, i);
			if (g_Dump)
				{
				if (Frame < 0)
					Log("Pos %d seq %c%c%c revcomp=%c%c%c",
					  i-2, Seq[i], Seq[i-1], Seq[i-2],
					  Complement(Seq[i-2]), Complement(Seq[i-1]), Complement(Seq[i]));
				else
					Log("Pos %d seq %c%c%c",
					  i, Seq[i], Seq[i+1], Seq[i+2]);
				Log(" %s\n", CodonToStr(Codon));
				}
			if (Codon == CODON2_DoneFrame)
				ORFStart = -1;
			if (CODON2_Start == Codon && -1 == ORFStart)
				{
				if (g_Dump)
					Log("START frame=%d pos=%d\n", Frame, i);
				ORFStart = i;
				}
			else if (CODON2_STP == Codon)
				{
				if (ORFStart != -1)
					{
					int ORFLength = (Frame < 0) ? ORFStart - i : i - ORFStart;
					if (g_Dump)
						Log("STOP frame=%d pos=%d start=%d length=%d bestlength=%d\n",
						  Frame, i, ORFStart, ORFLength, BestORFLength);
					if (ORFLength > BestORFLength)
						{
						BestORFFrame = Frame;
						BestORFStart = ORFStart;
						BestORFLength = ORFLength;
						}
					}
				ORFStart = -1;
				}
			}
		}
	*ptrFrame = BestORFFrame;
	*ptrStart = BestORFStart;
	*ptrLength = BestORFLength;
	}

static void XlatSeqFrame(const char Seq[], int SeqLength, int Frame)
	{
	int iStart = GetFrameStart(Frame, SeqLength);
	for (;;)
		{
		if (!InFrame(Frame, SeqLength, iStart))
			break;

		int Col = 1;
		for (int i = iStart; ; i = NextFramePos(Frame, i))
			{
			if (!InFrame(Frame, SeqLength, i) || Col == BLOCK_LENGTH)
				{
				Log("\n");
				break;
				}
			const char *s = GetThreeLetters(Seq, Frame, i);
			Log(" %s", s);
			++Col;
			}

		Col = 1;
		for (int i = iStart; ; i = NextFramePos(Frame, i))
			{
			if (!InFrame(Frame, SeqLength, i) || Col == BLOCK_LENGTH)
				{
				iStart = i;
				Log("\n");
				Log("\n");
				break;
				}
			const char *s = GetThreeLetters(Seq, Frame, i);
			CODON2 Codon = GetCodonStr(s);
			Log(" %s", CodonToStr(Codon));
			++Col;
			}
		}
	Log("\n");
	}

void GetFrameDNA(const char Seq[], int Frame, int ORFStart, int ORFLength, char s[])
	{
	int Pos = 0;
	for (int i = ORFStart; Pos < ORFLength; i = NextFramePos(Frame, i))
		{
		const char *Codon = GetThreeLetters(Seq, Frame, i);
		s[Pos++] = (char) toupper(Codon[0]);
		s[Pos++] = (char) toupper(Codon[1]);
		s[Pos++] = (char) toupper(Codon[2]);
		}
	s[Pos] = 0;
	}

void XlatFrame(const char Seq[], int Frame, int ORFStart, int ORFLength, char s[])
	{
	int Pos = 0;
	for (int i = ORFStart; Pos < ORFLength; i = NextFramePos(Frame, i))
		{
		CODON2 Codon = GetCodon(Seq, Frame, i);
		const char c = CodonToAminoChar(Codon);
		s[Pos++] = c;
		}
	s[Pos] = 0;
	}

void XlatSeq(const char Seq[], int Frame, int SeqLength, char s[])
	{
	int Pos = 0;
	int i = GetFrameStart(Frame, SeqLength);
	for (;;)
		{
		if (!InFrame(Frame, SeqLength, i))
			break;
		CODON2 Codon = GetCodon(Seq, Frame, i);
		const char c = CodonToAminoChar(Codon);
		s[Pos++] = c;
		i = NextFramePos(Frame, i);
		}
	s[Pos] = 0;
	}

static void OnSeq(const string &Label, const string &Seq_)
	{
	if (g_Xlat)
		{
		for (int Frame = -3; Frame <= 2; ++Frame)
			{
			Log("\n");
			Log("Frame %d\n", Frame);
			XlatSeqFrame(Seq_.c_str(), SIZE(Seq_), Frame);
			}
		}

	string Seq(Seq_);

	int Length = SIZE(Seq);
	unsigned ORFCount = 0;
	bool CommonStrand = true;
	for (;;)
		{
		int ORFFrame = -999;
		int ORFStart = -999;
		int ORFLength = -999;
		FindORFsSeq(Seq.c_str(), SIZE(Seq), &ORFFrame, &ORFStart, &ORFLength);
		if (ORFLength < (int) g_MinCodons*3)
			return;

		if (ORFStart != -999)
			{
			bool Strand = (ORFFrame < 0);
			if (ORFCount == 0)
				CommonStrand = Strand;
			asserta(ORFStart != -999 && ORFLength != -999);
			assert(ORFLength%3 == 0);

			int lo;
			int hi;
			if (ORFFrame >= 0)
				{
				lo = ORFStart;
				hi = lo + ORFLength + 2;
				}
			else
				{
				hi = ORFStart;
				lo = hi - ORFLength - 2;
				}

			if (!g_SameStrand || Strand == CommonStrand)
				{
				char *s = new char[ORFLength+4];
				XlatFrame(Seq.c_str(), ORFFrame, ORFStart, ORFLength/3, s);
				Log("Seq %s SeqLength %d, Frame %d, lo %d, hi %d, length %d\n",
				  Label.c_str(), Length, ORFFrame, lo, hi, ORFLength);
				if (g_fOut != 0)
					fprintf(g_fOut, ">%s;%d;%d-%d\n%s\n",
					  Label.c_str(), ORFFrame, lo, hi, s);

				if (g_fOutDNA)
					{
					GetFrameDNA(Seq.c_str(), ORFFrame, ORFStart, ORFLength+3, s);
					fprintf(g_fOutDNA, ">%s;%d;%d-%d\n%s\n",
					  Label.c_str(), ORFFrame, lo, hi, s);
					}
				delete[] s;

				if (g_fGFF != 0)
					{
					GFFRecord2 Rec;
					Rec.SeqName = Label;
					Rec.Source = "evo";
					Rec.Feature = "CDS";
					Rec.Frame = 0;
					Rec.Strand = pom(ORFFrame >= 0);
					Rec.Score = 0;
					Rec.Start = lo + 1;
					Rec.End = hi + 1;
					char Str[64];
					sprintf(Str, "gene_index %u; Frame \"%+d\";",
					  g_GeneIndex++, (ORFFrame >= 0 ? ORFFrame + 1 : ORFFrame));
					Rec.Attrs = string(Str);
					WriteGFFRec(g_fGFF, Rec);
					}
				}
			int j = 0;
			for (int i = ORFStart; j < ORFLength; ORFFrame >= 0 ? ++i : --i, ++j)
				{
				if (i >= Length)
					Die("i=%d Length=%d", i, Length);
				Seq[i] = '#';
				}
			++ORFCount;
			if (ORFCount >= g_MaxORFs)
				return;
			}
		}
	}

void FindORFs()
	{
	const char *InFileName = RequiredValueOpt("findorfs");
	const char *OutFileName = ValueOpt("out");
	const char *OutDNAFileName = ValueOpt("outdna");
	const char *GFFFileName = ValueOpt("gff");
	IntOpt("mincodons", &g_MinCodons);
	IntOpt("maxorfs", &g_MaxORFs);
	g_Xlat = FlagOpt("xlat");
	g_SameStrand = FlagOpt("samestrand");
	g_Dump = FlagOpt("dump");

	if (OutFileName != 0)
		g_fOut = CreateStdioFile(OutFileName);
	if (OutDNAFileName != 0)
		g_fOutDNA = CreateStdioFile(OutDNAFileName);
	if (GFFFileName != 0)
		g_fGFF = CreateStdioFile(GFFFileName);

	ReadFasta(InFileName, OnSeq);

	if (g_fOut != 0)
		CloseStdioFile(g_fOut);
	if (g_fOutDNA != 0)
		CloseStdioFile(g_fOutDNA);
	if (g_fGFF != 0)
		CloseStdioFile(g_fGFF);
	}
