#include "geninfo.h"
#include "utils.h"
#include "getopts.h"

/***
Structure:
	std::string Name;
	std::vector<ChrInfo> Chromosomes;

Stored as:
	unsigned GenInfoStartMagic
	unsigned NameLength
	char char ... Name
	unsigned ChrCount
		ChrInfo_1
		ChrInfo_2
		...
		ChrInfo_N
	unsigned GenInfoEndMagic

***/

#define ReadData(v)		ReadStdioFile(f, &v, (unsigned) sizeof(v))
#define WriteData(v)	WriteStdioFile(f, &v, (unsigned) sizeof(v))

void GenInfo::Create(const std::string &Name)
	{
	std::vector<ChrInfo> Chrs; // empty
	Create(Name, Chrs);
	}

void GenInfo::Create(const std::string &Name, const std::vector<ChrInfo> &Chrs)
	{
	m_Name = Name;

	const unsigned ChrCount = (unsigned) Chrs.size();
	m_Chrs.clear();
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		m_Chrs.push_back(Chrs[ChrIndex]);
	}

void GenInfo::Clear()
	{
	m_Name.clear();
	m_Chrs.clear();
	m_NameToChrId.clear();
	m_ChrIdToIndex.clear();
	}

void GenInfo::FromFile(FILE *f)
	{
	unsigned StartMagic;
	ReadData(StartMagic);
	if (StartMagic != GenInfoStartMagic)
		Die("Invalid file (GenInfoStartMagic)");

	unsigned NameLength;
	ReadData(NameLength);
	m_Name.reserve(NameLength);
	for (unsigned i = 0; i < NameLength; ++i)
		{
		char c;
		ReadData(c);
		m_Name.push_back(c);
		}

	unsigned ChrCount;
	ReadData(ChrCount);
	m_Chrs.reserve(ChrCount);

	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		{
		ChrInfo CI;
		CI.FromFile(f);

		unsigned ChrId = CI.GetId();
		const std::string &ChrName = CI.GetName();
		coord_t ChrLength = CI.GetLength();

		AddChr(ChrId, ChrName, ChrLength);
		}

	unsigned EndMagic;
	ReadData(EndMagic);
	if (EndMagic != GenInfoEndMagic)
		Die("Invalid file (GenInfoEndMagic)");
	}

void GenInfo::ToFile(FILE *f) const
	{
	WriteData(GenInfoStartMagic);

	const unsigned NameLength = (unsigned) m_Name.size();
	WriteData(NameLength);
	for (unsigned i = 0; i < NameLength; ++i)
		WriteData(m_Name[i]);

	const unsigned ChrCount = GetChrCount();
	WriteData(ChrCount);

	for (unsigned ChrId = 0; ChrId < ChrCount; ++ChrId)
		m_Chrs[ChrId].ToFile(f);

	WriteData(GenInfoEndMagic);
	}

const std::string &GenInfo::GetName() const
	{
	return m_Name;
	}

unsigned GenInfo::GetChrCount() const
	{
	return (unsigned) m_Chrs.size();
	}

void GenInfo::GetChrInfo(unsigned ChrId, ChrInfo &CI) const
	{
	CI = GetChrInfo(ChrId);
	}

const ChrInfo &GenInfo::GetChrInfo(unsigned ChrId) const
	{
	std::map<unsigned, unsigned>::const_iterator p = m_ChrIdToIndex.find(ChrId);
	if (p == m_ChrIdToIndex.end())
		Die("GenInfo::GetChrInfo(%u), id not found", ChrId);
	unsigned ChrIndex = p->second;
	assert(ChrIndex < (unsigned) m_Chrs.size());
	return m_Chrs[ChrIndex];
	}

ChrInfo *GenInfo::GetModifiableChrInfoPtr(unsigned ChrId)
	{
	std::map<unsigned, unsigned>::const_iterator p = m_ChrIdToIndex.find(ChrId);
	if (p == m_ChrIdToIndex.end())
		Die("GenInfo::GetChrInfo(%u), id not found", ChrId);
	unsigned ChrIndex = p->second;
	assert(ChrIndex < (unsigned) m_Chrs.size());
	return &m_Chrs[ChrIndex];
	}

const ChrInfo *GenInfo::GetChrInfoPtr(unsigned ChrId) const
	{
	std::map<unsigned, unsigned>::const_iterator p = m_ChrIdToIndex.find(ChrId);
	if (p == m_ChrIdToIndex.end())
		return 0;
	unsigned ChrIndex = p->second;
	assert(ChrIndex < (unsigned) m_Chrs.size());
	return &m_Chrs[ChrIndex];
	}

const ChrInfo &GenInfo::GetChrInfoByIndex(unsigned ChrIndex) const
	{
	if (ChrIndex >= (unsigned) m_Chrs.size())
		Die("GenInfo::GetChrInfoByIndex(%u), %u chrs",
		  ChrIndex, (unsigned) m_Chrs.size());
	return m_Chrs[ChrIndex];
	}

ChrInfo &GenInfo::GetModifiableChrInfoByIndex(unsigned ChrIndex)
	{
	if (ChrIndex >= (unsigned) m_Chrs.size())
		Die("GenInfo::GetChrInfoByIndex(%u), %u chrs",
		  ChrIndex, (unsigned) m_Chrs.size());
	return m_Chrs[ChrIndex];
	}

const ChrInfo &GenInfo::GetChrInfo(const std::string &ChrName) const
	{
	unsigned ChrId;
	bool Found = GetChrId(ChrName, &ChrId);
	if (!Found)
		Die("Chr '%s' not found", ChrName.c_str());
	return GetChrInfo(ChrId);
	}

ChrInfo *GenInfo::GetModifiableChrInfoPtr(const std::string &ChrName)
	{
	unsigned ChrId;
	bool Found = GetChrId(ChrName, &ChrId);
	if (!Found)
		return 0;
	return GetModifiableChrInfoPtr(ChrId);
	}

bool GenInfo::GetChrId(const std::string &ChrName, unsigned *ptrChrId) const
	{
	std::map<std::string, unsigned>::const_iterator p = m_NameToChrId.find(ChrName);
	if (p == m_NameToChrId.end())
		return false;
	*ptrChrId = p->second;
	return true;
	}

const ChrInfo *GenInfo::GetChrInfoPtr(const std::string &ChrName) const
	{
	std::map<std::string, unsigned>::const_iterator p = m_NameToChrId.find(ChrName);
	if (p == m_NameToChrId.end())
		return 0;
	unsigned ChrId = p->second;

	std::map<unsigned, unsigned>::const_iterator q = m_ChrIdToIndex.find(ChrId);
	if (q == m_ChrIdToIndex.end())
		Die("GenInfo::GetChrInfoPtr, chrid not found");

	unsigned ChrIndex = q->second;
	return &m_Chrs[ChrIndex];
	}

unsigned GenInfo::GetChrId(const std::string &ChrName) const
	{
	unsigned ChrId;
	bool Found = GetChrId(ChrName, &ChrId);
	if (!Found)
		Die("Chr '%s' not found", ChrName.c_str());
	return ChrId;
	}

coord_t GenInfo::GetLength() const
	{
	coord_t Length = 0;
	const unsigned ChrCount = GetChrCount();
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		{
		const ChrInfo &CI = m_Chrs[ChrIndex];
		Length += CI.GetLength();
		}
	return Length;
	}

coord64_t GenInfo::GetLength64() const
{
	coord64_t Length = 0;
	const unsigned ChrCount = GetChrCount();
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
	{
		const ChrInfo &CI = m_Chrs[ChrIndex];
		Length += CI.GetLength();
	}
	return Length;
}

bool GenInfo::operator==(const GenInfo &rhs) const
	{
	return m_Name == rhs.GetName() && m_Chrs == rhs.m_Chrs;
	}

void GenInfo::LogMe() const
	{
	Log("\n");
	Log("GenInfo::LogMe(): Name=%s\n", m_Name.c_str());
	Log("\n");
	Log("  Chr              Name     Id      Length\n");
	Log("=====  ================  =====  ==========\n");
	const unsigned ChrCount = (unsigned) m_Chrs.size();
	coord_t TotalLength = 0;
	for (unsigned i = 0; i < ChrCount; ++i)
		{
		const ChrInfo &Chr = m_Chrs[i];
		Log("%5u  %16.16s  %5u  %10u\n",
		  i,
		  Chr.GetName().c_str(),
		  Chr.GetId(),
		  (unsigned) Chr.GetLength());
		TotalLength += Chr.GetLength();
		}
	Log("                                ==========\n");
	Log("                                %10"COORD_T_PRINTF"\n\n", TotalLength);
	}

unsigned GenInfo::AddChr(const std::string &Name, coord_t Length)
	{
	unsigned ChrId = GetUnusedChrId();
	AddChr(ChrId, Name, Length);
	return ChrId;
	}

void GenInfo::AddChr(const ChrInfo &CI, unsigned Id)
	{
	if (Id == UINT_MAX)
		Id = CI.GetId();
	AddChr(Id, CI.GetName(), CI.GetLength());
	}

void GenInfo::AddChr(unsigned Id, const std::string &Name, coord_t Length)
	{
	const ChrInfo *CI = GetChrInfoPtr(Name);
	if (CI != 0)
		Die("GenInfo::AddChr, name %s already added", Name.c_str());

	const std::string *ptrName2 = GetChrNamePtr(Id);
	if (ptrName2 != 0)
		Die("GenInfo::AddChr, id %u already added", Id);

	ChrInfo CI2;
	CI2.Create(Id, Name, Length);

	unsigned ChrIndex = (unsigned) m_Chrs.size();

	std::map<unsigned, unsigned>::value_type v1(Id, ChrIndex);
	m_ChrIdToIndex.insert(v1);

	std::map<std::string, unsigned>::value_type v2(Name, Id);
	m_NameToChrId.insert(v2);

	m_Chrs.push_back(CI2);
	}

const std::string &GenInfo::GetChrName(unsigned ChrId) const
	{
	std::map<unsigned, unsigned>::const_iterator p = m_ChrIdToIndex.find(ChrId);
	if (p == m_ChrIdToIndex.end())
		Die("GenInfo::GetChrName(%u), not found", ChrId);

	unsigned ChrIndex = p->second;
	assert(ChrIndex < (unsigned) m_Chrs.size());
	const ChrInfo &CI = m_Chrs[ChrIndex];
	return CI.GetName();
	}

const std::string *GenInfo::GetChrNamePtr(unsigned ChrId) const
	{
	std::map<unsigned, unsigned>::const_iterator p = m_ChrIdToIndex.find(ChrId);
	if (p == m_ChrIdToIndex.end())
		return 0;

	unsigned ChrIndex = p->second;
	assert(ChrIndex < (unsigned) m_Chrs.size());
	const ChrInfo &CI = m_Chrs[ChrIndex];
	return &CI.GetName();
	}

unsigned GenInfo::GetUnusedChrId() const
	{
	unsigned ArraySize = (unsigned) m_ChrIdToIndex.size();
	for (unsigned Id = ArraySize; Id < ArraySize + 64000; ++Id)
		{
		std::map<unsigned, unsigned>::const_iterator p = m_ChrIdToIndex.find(Id);
		if (p == m_ChrIdToIndex.end())
			return Id;
		}
	Die("GenInfo::UnusedChrId failed");
	return UINT_MAX;
	}

void GenInfo::SetName(const std::string &Name)
	{
	m_Name = Name;
	}

unsigned GenInfo::GetChrIndex(unsigned ChrId) const
	{
	std::map<unsigned, unsigned>::const_iterator p = m_ChrIdToIndex.find(ChrId);
	if (p == m_ChrIdToIndex.end())
		Die("GenInfo::GetChrIndex(%u), ChrId not found");

	unsigned ChrIndex = p->second;
	return ChrIndex;
	}

unsigned GenInfo::GetChrId(unsigned ChrIndex) const
	{
	const unsigned ChrCount = GetChrCount();
	if (ChrIndex >= ChrCount)
		Die("GenInfo::GetChrId(%u), there are only %u chrs",
		  ChrIndex, ChrCount);
	const ChrInfo &CI = m_Chrs[ChrIndex];
	return CI.GetId();
	}

GenInfo &GenInfo::operator=(const GenInfo &rhs)
	{
	Create(rhs.GetName());
	unsigned ChrCount = rhs.GetChrCount();
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		{
		const ChrInfo &CI = rhs.GetChrInfoByIndex(ChrIndex);
		AddChr(CI.GetName(), CI.GetLength());
		}
	return *this;
	}

GenInfo::GenInfo(const GenInfo &rhs)
	{
	m_Name = rhs.GetName();
	unsigned ChrCount = rhs.GetChrCount();
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		{
		const ChrInfo &CI = rhs.GetChrInfoByIndex(ChrIndex);
		AddChr(CI.GetName(), CI.GetLength());
		}
	}
