#include "ivtree.h"
#include "gff2.h"
#include "ratemgr.h"
#include "progress.h"
#include "myrand.h"
#include "getopts.h"
#include "evomodel.h"
#include "genncces.h"

/***
Generating NCCEs
----------------
Special cases:

	3' and 5' UTRs will be extracted from a genome browser track.

	Splice donor and acceptor base pairs are identified as the
	first and last two bases in each intron and fully constrained.
	
Other NCCEs are generated by a stochastic model, as follows.

The length distribution is specified in exactly the same way
as for a mutational event: the user will provide any number of
(length, weight) pairs and evolver will interpolate any missing
values.

The user provides a GFF file specifying "exclusion zones" which
NCCEs must not overlap. These will include at a minimum CDSs and
UTRs, and might also include some microsatellites and designated
"deserts".

The user specifies a fraction of the sequence to be covered by
generated NCCEs.

Some NCCEs are assigned to genes. This is done as follows.

For each pair of adjacent genes, a maximum distance D is selected
by a procedure explained below. CEs closer than this distance are
assigned to the closest gene:


                   |<--D-->|           |<--D-->|
 [-----Gene A------]  (CE1)  (CE2)      (CE3)  [--Gene B---]


Here, CE1 is assigned to Gene A, CE3 is assigned to Gene B, and
CE2 is not assigned to a gene.

D is chosen so that, on average, it is 1/4 the inter-gene distance.
The fraction of the inter-genic distance is sampled from an approx-
imately normal distribution with mean m (default 1/4) and standard
deviation s (default 1/8).

A maximum value of D will be imposed to cover cases where the inter-
genic distance is very large.

Note that this procedure means that a CE inside the range of a gene
is /always/ assigned to that gene, i.e. there cannot be a non-gene
CE inside a gene.
***/

void SetModel(const char *FileName);
extern RateMgr *g_NCCERateMgr;

static gpos_t g_SeqLength;
static string g_SeqName;
static bool g_LogFails = false;

static gpos_t TryGenerateNCCE(IvTree<LoHi> &IntervalTree, vector<GFFRecord2> &NCCEs)
	{
	gpos_t Start = RandPos(g_SeqLength);
	gpos_t Length = g_NCCERateMgr->GetLength();
	gpos_t End = Start + Length - 1;

	if (End >= g_SeqLength)
		return GPOS_T_MAX;

	vector<LoHi> Results;
	IntervalTree.GetOverlaps(pair<gpos_t, gpos_t>(Start, End), Results);

	if (!Results.empty())
		{
		if (g_LogFails)
			{
			static bool HdrDone = false;
			if (!HdrDone)
				{
				HdrDone = true;
				Log("\n");
				Log("     Start         End      Length     OvStart       OvEnd    OvLength\n");
				Log("==========  ==========  ==========  ==========  ==========  ==========\n");
				}
			for (unsigned i = 0; i < SIZE(Results); ++i)
				{
				const LoHi &LH = Results[i];
				Log("%10"PF, Start);
				Log("  %10"PF, End);
				Log("  %10"PF, Length);
				Log("  %10"PF, LH.Lo);
				Log("  %10"PF, LH.Hi);
				Log("  %10"PF, LH.GetLength());
				Log("\n");
				}
			}
		return GPOS_T_MAX;
		}

	GFFRecord2 Rec;
	Rec.SeqName = g_SeqName;
	Rec.Source = "genncces";
	Rec.Feature = "NCCE";
	Rec.Start = Start + 1;
	Rec.End = End + 1;
	Rec.Score = 0;
	Rec.Strand = '+';
	Rec.Frame = INT_MAX;
	Rec.Attrs = "";

	NCCEs.push_back(Rec);

	gpos_t lo = Start;
	if (lo > 0)
		lo -= 1;
	IntervalTree.Add(LoHi(lo, End+1));

	return Length;
	}

static gpos_t GenerateNCCE(unsigned MaxTries, IvTree<LoHi> &IntervalTree,
  vector<GFFRecord2> &NCCEs)
	{
	for (unsigned Try = 0; ; ++Try)
		{
		if (Try >= MaxTries)
			Die("Max tries exceeded generating NCCE");

		gpos_t Length = TryGenerateNCCE(IntervalTree, NCCEs);
		if (Length != GPOS_T_MAX)
			return Length;
		}
	}

static bool LT(const GFFRecord2 &Rec1, const GFFRecord2 &Rec)
	{
	string Name1(Rec1.SeqName);
	string Name2(Rec.SeqName);
	if (Name1 != Name2)
		return Name1 < Name2;
	return Rec1.Start < Rec.Start;
	}

void GenerateNCCEs()
	{
	const char *ExcludeGFFFileName = RequiredValueOpt("excl_gff");
	const char *OutFileName = RequiredValueOpt("out");
	const char *ModelFileName = ValueOpt("model");
	bool Dump = FlagOpt("dump");
	g_LogFails = FlagOpt("logfails");
	unsigned MaxTries = 100;
	IntOpt("maxtries", &MaxTries);
	g_SeqLength = RequiredIntOpt("length");

	double GenomePct = P_GenomeNCCEPct;
	if (GenomePct <= 0 || GenomePct > 100)
		Die("Invalid -pct");

	SetModel(ModelFileName);

	gpos_t BaseCount = (gpos_t) (g_SeqLength*GenomePct/100.0);

	FILE *f = OpenStdioFile(ExcludeGFFFileName);
	unsigned FileSize = (unsigned) GetStdioFileSize(f);
	GFFRecord2 Rec;
	ProgressStart("Reading exclusion records from %.16s", ExcludeGFFFileName);

// Build interval index of regions to exlude.
// Initially defined by recorde in exclude GFF file,
// NCCEs are added as they are generated.
	IvTree<LoHi> IntervalTree;

	g_SeqName = "";
	unsigned RecordCount = 0;
	while (GetNextGFFRecord(f, Rec))
		{
		if ((RecordCount++)%1000 == 0)
			{
			unsigned Pos = (unsigned) GetStdioFilePos(f);
			ProgressStep(Pos, FileSize);
			}
		const string Name(Rec.SeqName);
		if (g_SeqName == "")
			g_SeqName = Name;
		else if (Name != g_SeqName)
			Die("Annotation file has more than one sequence name '%s', '%s'",
			  g_SeqName.c_str(), Name.c_str());

		gpos_t lo = Rec.Start - 1;
		gpos_t hi = Rec.End - 1;
		IntervalTree.Add(LoHi(lo, hi));
		}
	ProgressDone();
	Progress("%u exclusion records", RecordCount);

	if (Dump)
		{
		Log("\n");
		Log("Initial exclusion index:\n");
		Log("        Lo          Hi      Length\n");
		Log("==========  ==========  ==========\n");
		IntervalTree.Traverse(LogLoHi);
		Log("\n");
		}

	Progress("Generating NCCE records");
	vector<GFFRecord2> NCCEs;
	gpos_t TotalBases = 0;
	RecordCount = 0;
	for (RecordCount = 0; ; ++RecordCount)
		{
		TotalBases += GenerateNCCE(MaxTries, IntervalTree, NCCEs);
		if (TotalBases >= BaseCount)
			break;
		}
	double Pct(double,double);
	Progress("Generated %u records, %"PF" bases, %.1f%%",
	  RecordCount, BaseCount, Pct(BaseCount, g_SeqLength));

	if (Dump)
		{
		Log("\n");
		Log("Final exclusion index:\n");
		Log("        Lo          Hi      Length\n");
		Log("==========  ==========  ==========\n");
		IntervalTree.Traverse(LogLoHi);
		Log("\n");
		}

	Progress("Sorting");
	sort(NCCEs.begin(), NCCEs.end(), LT);

	ProgressStart("Writing to %s", OutFileName);
	FILE *fOut = CreateStdioFile(OutFileName);
	const unsigned NCCECount = SIZE(NCCEs);
	for (unsigned i = 0; i < NCCECount; ++i)
		{
		if (i%1000 == 0)
			ProgressStep(i, NCCECount);
		const GFFRecord2 &Rec = NCCEs[i];
		WriteGFFRec(fOut, Rec);
		}
	ProgressDone();

	CloseStdioFile(fOut);
	}
