#define TRACE	0

#include "revfile.h"
#include "ivtree.h"
#include "lohi.h"
#include "progress.h"

coord_t GetOverlapLength(coord_t loA, coord_t hiA, coord_t loB, coord_t hiB);
void GetColRange(const StdStr &Row, const GCSeg &Seg, coord_t lo, coord_t hi,
  unsigned &ColFrom, unsigned &ColTo);
void GetRangeSeg(const StdStr &Row, unsigned ColFrom, unsigned ColTo,
  const GCSeg &Seg, GCSeg &RangeSeg);
bool SeqIsAllGaps(const std::string &Seq);
void LogAIRows(const AlnInfo &AI, const StdStrVec &Rows, const RevFile *Rev);

/***
Cache interval index on the assumption the typical usage
will be repeated calls with the same RevFile, genome and chr
but different lo, hi values. Actually, it's not bad even if
it's only called once because a sequential search is required
as RevFile currently does not support any indexed searches.
***/
static vector<IvTree<LoHi> > g_Ixs;
static const RevFile *g_ptrRev = 0;
static unsigned g_GenomeIndex = UINT_MAX;

static void MakeIntervalIndex(const RevFile &Rev, unsigned GenomeIndex,
  unsigned ChrId, IvTree<LoHi> &Ix)
	{
#if	TRACE
	Log("\n");
	Log("MakeIntervalIndex\n");
#endif
	IntVec AIIndexes;
//	Progress("Get AIIndexes");
	Rev.GetAppendAlnInfoIndexesChr(GenomeIndex, ChrId, AIIndexes);

//	Progress("Get AIs");
	AIXVec AIXs;
	Rev.GetAlnInfoXs(AIIndexes, AIXs);

	const unsigned N = SIZE(AIXs);
//	Progress("Build iv-ix, %u AIs", N);
	for (unsigned i = 0; i < N; ++i)
		{
		const AlnInfoX &AIX = AIXs[i];
		const GCSegVec &Segs = AIX.AI.Segs;
		const unsigned SegCount = SIZE(Segs);
		for (unsigned j = 0; j < SegCount; ++j)
			{
			const GCSeg &Seg = Segs[j];
			if (Seg.GenomeIndex == GenomeIndex && Seg.ChrId == ChrId)
				{
				LoHi LH;
				LH.Lo = Seg.lo;
				LH.Hi = Seg.hi;
				LH.Index = AIX.AlnInfoIndex;
#if	TRACE
				Log("Add: Seg=");
				Seg.LogMe();
				Log(" LH=");
				LH.LogMe(true);
#endif
				Ix.Add(LH);
				}
			}
		}
//	Progress("Iv-ix done");
	}

void GetAlnRange(const AlnInfo &AI, const StdStrVec &Rows, unsigned GenomeIndex,
  unsigned ChrId, coord_t lo, coord_t hi, AlnInfo &RangeAI, StdStrVec &RangeRows)
	{
#if	TRACE
	Log("\n");
	Log("GetAlnRange\n");
#endif
	RangeAI.Clear();
	RangeRows.clear();

	const GCSegVec &Segs = AI.Segs;
	const unsigned SegCount = SIZE(Segs);
	asserta(SIZE(Rows) == SegCount);

// Find first ("master") segment to overlap the given range (if there
// are more than 1, only the first is used).
	unsigned MasterSegIndex = UINT_MAX;
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSeg &Seg = Segs[SegIndex];
		if (Seg.GenomeIndex == GenomeIndex && Seg.ChrId == ChrId &&
		  GetOverlapLength(Seg.lo, Seg.hi, lo, hi) > 0)
			{
			MasterSegIndex = SegIndex;
			break;
			}
		}
#if	TRACE
	Log("MasterSegIndex=%u\n", MasterSegIndex);
#endif
	if (MasterSegIndex == UINT_MAX)
		return;

	const GCSeg &MasterSeg = Segs[MasterSegIndex];
	const StdStr &MasterRow = Rows[MasterSegIndex];

// Sub-range of (gen,chr,lo,hi) found in this alignment
	coord_t AlnLo = max(lo, MasterSeg.lo);
	coord_t AlnHi = min(hi, MasterSeg.hi);

	unsigned StartCol = UINT_MAX;
	unsigned EndCol = UINT_MAX;
	GetColRange(MasterRow, MasterSeg, AlnLo, AlnHi, StartCol, EndCol);
#if	TRACE
	Log("AlnLo=%u AlnHi=%u StartCol=%u EndCol=%u\n",
	  (unsigned) AlnLo, (unsigned) AlnHi, StartCol, EndCol);
#endif
	if (StartCol == UINT_MAX || EndCol == UINT_MAX || EndCol < StartCol)
		{
		Log("\n");
		if (StartCol == UINT_MAX)
			Log("StartCol=*");
		else
			Log("StartCol=%u", StartCol);
		if (EndCol == UINT_MAX)
			Log(" EndCol=*");
		else
			Log(" EndCol=%u", EndCol);
		Log("\n");
		Log("GetAlnRange(Gen=%u, Chr=%u, lo=%u, hi=%u)\n",
		  GenomeIndex, ChrId, (unsigned) lo, (unsigned) hi);
		Log("AI:\n");
		AI.LogMe();
		Log("MasterSeg=");
		MasterSeg.LogMe(true);
		Warning("GetAlnRange, cols not found or end < start");//@@
		RangeAI.Clear();
		RangeRows.clear();
		return;//@@
		}

	unsigned ColCount = EndCol - StartCol + 1;

	GCSeg RangeMasterSeg(MasterSeg);
	RangeMasterSeg.lo = AlnLo;
	RangeMasterSeg.hi = AlnHi;

	RangeAI.Attrs = AI.Attrs;
	RangeAI.Values = AI.Values;
	RangeAI.Score = AI.Score;

	RangeAI.Segs.push_back(RangeMasterSeg);
	RangeRows.push_back(MasterRow.substr(StartCol, ColCount));

	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		if (SegIndex == MasterSegIndex)
			continue;

		const GCSeg &Seg = Segs[SegIndex];
		const StdStr &Row = Rows[SegIndex];

		const StdStr &RangeRow = Row.substr(StartCol, ColCount);
		if (SeqIsAllGaps(RangeRow))
			continue;

		GCSeg RangeSeg;
		GetRangeSeg(Row, StartCol, EndCol, Seg, RangeSeg);

		RangeAI.Segs.push_back(RangeSeg);
		RangeRows.push_back(RangeRow);
		}
	asserta(SIZE(RangeRows) == SIZE(RangeAI.Segs));

// Special case: if all other segs were all gaps,
// don't return an alignment.
	if (SIZE(RangeRows) == 1)
		{
		RangeAI.Clear();
		RangeRows.clear();
		}
	}

void GetAlnsRange(const RevFile &Rev, unsigned GenomeIndex,
  unsigned ChrId, coord_t lo, coord_t hi,
  AIVec &AIs, vector<StdStrVec> &RowsVec)
	{
#if	TRACE
	Log("\n");
	Log("GetAlnsRange(Gen=%u, Chr=%u, lo=%u, hi=%u)\n",
	  GenomeIndex, ChrId, lo, hi);
#endif
	AIs.clear();
	RowsVec.clear();

// WARNING -- technically a bug.
// This test is not 100% robust as the input Rev might have been delete'd
// and new'd at the same address. But good enough for us and there isn't a
// robust way to do it without changing RevFile to have a per-instance UUID.
	if (&Rev != g_ptrRev || GenomeIndex != g_GenomeIndex)
		{
		Progress("Refresh iv-ix cache gen=%u chr=%u", GenomeIndex, ChrId);
		g_Ixs.clear();
		unsigned ChrCount = Rev.GetChrCount(GenomeIndex);
		g_Ixs.resize(ChrCount);
		for (unsigned ChrId = 0; ChrId < ChrCount; ++ChrId)
			{
			unsigned ChrIndex = Rev.GetChrIndex(GenomeIndex, ChrId);
			if (ChrIndex != ChrId)
				Die("Sparse chr ids not supported");
			MakeIntervalIndex(Rev, GenomeIndex, ChrId, g_Ixs[ChrId]);
			}

		g_ptrRev = &Rev;
		g_GenomeIndex = GenomeIndex;
		}

	vector<LoHi> Overlaps;
	g_Ixs[ChrId].GetOverlaps(pair<coord_t, coord_t>(lo, hi), Overlaps);
#if	TRACE
	Log("%u overlaps\n", SIZE(Overlaps));
#endif

	const unsigned N = SIZE(Overlaps);
	for (unsigned i = 0; i < N; ++i)
		{
		const LoHi &LH = Overlaps[i];
#if	TRACE
		Log("Overlap LH=");
		LH.LogMe(true);
#endif
		unsigned AIIndex = LH.Index;
		unsigned AlnIndex = Rev.GetAlnIndex(AIIndex);
		if (AlnIndex == UINT_MAX)
			Die("%.32s, AI %u has no rows", Rev.GetFileName().c_str(), AIIndex);

		AlnInfo AI;
		Rev.GetAlnInfo(AIIndex, AI);

		StdStrVec Rows;
		Rev.GetAlnRows(AlnIndex, Rows);
#if	TRACE
		Log("Unedited aln:\n");
		LogAIRows(AI, Rows, &Rev);
#endif
		AlnInfo RangeAI;
		StdStrVec RangeRows;
		GetAlnRange(AI, Rows, GenomeIndex, ChrId, lo, hi, RangeAI, RangeRows);
#if	TRACE
		Log("Range aln:\n");
		LogAIRows(RangeAI, RangeRows, &Rev);
#endif

		if (RangeRows.empty())
			continue;

		AIs.push_back(RangeAI);
		RowsVec.push_back(RangeRows);
		}
	}
