#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <ctype.h>
#include "getopts.h"
#include "utils.h"

#ifdef _MSC_VER
#pragma warning(disable: 4702)	// Unreachable code
#include <windows.h>
#endif

static int g_argc;
static char **g_argv;
const char *g_ProgramName;
static bool g_EchoCmdLine = false;
static bool g_LogCmdLineAndTime = true;

static void ArgsFromFile(const char *FileName, char *ProgName,
  int *ptr_argc, char ***ptr_argv);

const char *GetProgramName()
	{
	if (0 == g_ProgramName)
		return "?";
	return g_ProgramName;
	}

void CommandLineError(const char *Format, ...)
	{
	va_list ArgList;
	va_start(ArgList, Format);

	char Str[4096];
	vsprintf(Str, Format, ArgList);
	Usage();
	fprintf(stderr, "\n\n");
	for (int i = 0; i < g_argc; ++i)
		fprintf(stderr, "%s%s", i > 0 ? " " : "", g_argv[i]);
	fprintf(stderr, "\n");
	fprintf(stderr, "%s, bad command line: %s\n\n", GetProgramName(), Str);
#if	_MSC_VER && _DEBUG && WIN32
	if (IsDebuggerPresent())
		DebugBreak();
#endif
	exit(1);
	}

static bool TestSetFlagOpt(const char *Arg)
	{
	for (int i = 0; i < FlagOptCount; ++i)
		if (!strcmp(Arg, FlagOpts[i].m_pstrName))
			{
			FlagOpts[i].m_bSet = true;
			return true;
			}
	return false;
	}

static bool TestSetValueOpt(const char *Arg, const char *Value)
	{
	for (int i = 0; i < ValueOptCount; ++i)
		if (!strcmp(Arg, ValueOpts[i].m_pstrName))
			{
			if (0 == Value)
				CommandLineError("Option -%s must have value\n", Arg);
//			ValueOpts[i].m_pstrValue = strsave(Value);
			if (ValueOpts[i].m_pstrValue != 0)
				CommandLineError("Option -%s defined twice\n", Arg);
			ValueOpts[i].m_pstrValue = Value;
			return true;
			}
	return false;
	}

bool FlagOpt(const char *Name)
	{
	for (int i = 0; i < FlagOptCount; ++i)
		if (!strcmp(Name, FlagOpts[i].m_pstrName))
			return FlagOpts[i].m_bSet;
	CommandLineError("FlagOpt(%s) invalid", Name);
	UNREACHABLE_RETURN(false)
	}

const char *ValueOpt(const char *Name)
	{
	for (int i = 0; i < ValueOptCount; ++i)
		if (!strcmp(Name, ValueOpts[i].m_pstrName))
			return ValueOpts[i].m_pstrValue;
	CommandLineError("ValueOpt(%s) invalid", Name);
	UNREACHABLE_RETURN(0)
	}

const char *RequiredValueOpt(const char *Name)
	{
	const char *s = ValueOpt(Name);
	if (0 == s)
		CommandLineError("Required option -%s not specified\n", Name);
	return s;
	}

std::string &GetCmdLine(std::string &s)
	{
	s.clear();
	for (int ArgIndex = 0; ArgIndex < g_argc; ++ArgIndex)
		{
		if (ArgIndex > 0)
			s += " ";
		const char *a = g_argv[ArgIndex];
		s += a;
		}
	return s;
	}

// Hopefully enough memory to allow Die to execute.
static const size_t RESERVE_BYTES = 10*1024*1024;
static void *g_EmergencyReserveMemory;

static void OnOutOfMemory()
	{
	static unsigned RecurseCount = 0;
	if (RecurseCount++ > 0)
		abort();

	free(g_EmergencyReserveMemory);
	Die("Out of memory, allocated so far %g MB", GetMemUseBytes()/1e6);
	}

void ProcessArgVect(int argc, char *argv[])
	{
	OSInit();

	g_EmergencyReserveMemory = malloc(RESERVE_BYTES);
	std::set_new_handler(OnOutOfMemory);

	g_argc = argc;
	g_argv = argv;

	g_ProgramName = strsave(argv[0]);

	if (argc == 3 && streq(argv[1], "file:"))
		ArgsFromFile(argv[2], argv[0], &argc, &argv);

	g_argc = argc;
	g_argv = argv;

	for (int iArgIndex = 1; iArgIndex < argc; )
		{
		const char *Arg = argv[iArgIndex];
		if (Arg[0] != '-')
			CommandLineError("Command-line option \"%s\" must start with '-'\n", Arg);

		if (strcmp(Arg, "-echocmdline") == 0)
			{
			g_EchoCmdLine = true;
			++iArgIndex;
			continue;
			}

		if (strcmp(Arg, "-nologcmdlineandtime") == 0)
			{
			g_LogCmdLineAndTime = false;
			++iArgIndex;
			continue;
			}

		const char *ArgName = Arg + 1;
		if (TestSetFlagOpt(ArgName))
			{
			++iArgIndex;
			continue;
			}
		
		char *Value = 0;
		if (iArgIndex < argc - 1)
			Value = argv[iArgIndex+1];
		if (TestSetValueOpt(ArgName, Value))
			{
			iArgIndex += 2;
			continue;
			}

		CommandLineError("Invalid command line option \"%s\"\n", ArgName);
		}

	const char *LogFileName = ValueOpt("log");
	if (LogFileName != 0)
		SetLogFile(LogFileName);
	SetDieArgcArgv(argc, argv);

	time_t t = time(0);
	if (g_LogCmdLineAndTime)
	{
		Log("%s", asctime(localtime(&t)));
		for (int i = 0; i < argc; ++i)
			{
			if (i > 0)
				Log(" ");
			Log("%s", argv[i]);
			}
		Log("\n");
		Log("\n");
	}

	if (g_EchoCmdLine)
		{
		for (int ArgIndex = 0; ArgIndex < argc; ++ArgIndex)
			{
			if (ArgIndex > 0)
				fprintf(stderr, " ");
			fprintf(stderr, "%s", argv[ArgIndex]);
			}
		fprintf(stderr, "\n");
		}
	}

bool IntOpt(const char *Name, unsigned *ptrValue)
	{
	const char *Value = ValueOpt(Name);
	if (0 == Value)
		return false;
	int n = sscanf(Value, "%u", ptrValue);
	if (n != 1)
		Die("Integer option %s=%s invalid", Name, Value);
	return true;
	}

bool SignedIntOpt(const char *Name, int *ptrValue)
	{
	const char *Value = ValueOpt(Name);
	if (0 == Value)
		return false;
	*ptrValue = atoi(Value);
	return true;
	}

unsigned RequiredIntOpt(const char *Name)
	{
	const char *Value = RequiredValueOpt(Name);
	return (unsigned) atoi(Value);
	}

bool FloatOpt(const char *Name, double *ptrValue)
	{
	const char *Value = ValueOpt(Name);
	if (0 == Value)
		return false;
	*ptrValue = atof(Value);
	return true;
	}

double RequiredFloatOpt(const char *Name)
	{
	const char *Value = RequiredValueOpt(Name);
	return atof(Value);
	}

static char *g_ArgBuffer = 0;

static void ArgsFromFile(const char *FileName, char *ProgName,
  int *ptr_argc, char ***ptr_argv)
	{
	FILE *f = fopen(FileName, "rb");
	if (f == 0)
		CommandLineError("file:%s, cannot open errno=%d", FileName, errno);

	int Ok = fseek(f, 0, SEEK_END);
	if (Ok < 0)
		CommandLineError("Seek failed in file:%s, errno=%d", FileName, errno);
	long FileSize = ftell(f);
	Ok = fseek(f, 0, SEEK_SET);
	if (Ok < 0)
		CommandLineError("Seek failed in file:%s, errno=%d", FileName, errno);

	g_ArgBuffer = myalloc(char, FileSize + 1);
	off_t BytesRead = (off_t) fread(g_ArgBuffer, 1, FileSize, f);
	if (BytesRead != FileSize)
		CommandLineError("fread failed, attempted %u bytes, read %u bytes, errno=%d",
		  (unsigned) FileSize, (unsigned) BytesRead, errno);
	fclose(f);

// White space to terminate last argument
	g_ArgBuffer[FileSize] = 0;

// First pass to count args and nul-terminate
	bool InArg = false;
	bool InComment = false;
	unsigned ArgCount = 0;
	for (long i = 0; i < FileSize; ++i)
		{
		if (InComment)
			{
			if (g_ArgBuffer[i] == '\n')
				InComment = false;
			}
		else
			{
			if (g_ArgBuffer[i] == '#')
				InComment = true;
			else
				{
				if (InArg)
					{
					if (isspace(g_ArgBuffer[i]))
						InArg = false;
					}
				else
					{
					if (!isspace(g_ArgBuffer[i]))
						{
						InArg = true;
						++ArgCount;
						}
					}
				}
			}
		if (!InArg || InComment)
			g_ArgBuffer[i] = 0;
		}

// Second pass to set argv[]
	char **argv = myalloc(char *, ArgCount + 1);
	argv[0] = ProgName;
	InArg = false;
	InComment = false;
	unsigned ArgIndex = 1;
	for (long i = 0; i < FileSize; ++i)
		{
		if (InArg)
			{
			if (g_ArgBuffer[i] == 0)
				InArg = false;
			}
		else
			{
			if (g_ArgBuffer[i] != 0)
				{
				InArg = true;
				argv[ArgIndex++] = g_ArgBuffer + i;
				}
			}
		}
	assert(ArgIndex == ArgCount + 1);
	*ptr_argc = ArgCount + 1;
	*ptr_argv = argv;
	}
