#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "types.h"
#include "utils.h"
#include "gff.h"
#include <string>

#define	TRACE	0

// GFF fields are:
// <seqname> <source> <feature> <start> <end> <score> <strand> <frame> [attributes] [comments]
//     0         1         2        3      4      5        6       7         8           9

static int GFFLineNr;

// WARNING: Line[] and Fields[]are overwritten
// by each call to GetNextGFFRecord.
// The Rec agument to GetNextGFFRecord returns
// pointers into Line[].
static char Line[MAX_GFF_LINE+1];
static char *Fields[9];

static void UnfixWhite(char *s)
	{
	for (char *p = s; *p != 0; ++p)
		{
		if (*p == '@')
			*p = ' ';
		}
	}

unsigned GFFGetRequiredIntAttr(const std::string &Attrs,
  const std::string &Name)
	{
	unsigned Value = UINT_MAX;
	bool Found = GFFGetIntAttr(Attrs, Name, Value);
	if (!Found)
		Die("Required attr '%s' not found in '%s'",
		  Name.c_str(), Attrs.c_str());
	return Value;
	}

void GFFGetRequiredAttr(const std::string &Attrs, const std::string &Name,
  std::string &Value)
	{
	bool Found = GFFGetAttr(Attrs, Name, Value);
	if (!Found)
		Die("Required attr '%s' not found in '%s'",
		  Name.c_str(), Attrs.c_str());
	}

bool GFFGetIntAttr(const std::string &Attrs, const std::string &Name,
  unsigned &Value)
	{
	std::string strValue;
	bool Found = GFFGetAttr(Attrs, Name, strValue);
	if (!Found)
		return false;
	for (std::string::const_iterator p = strValue.begin();
	  p != strValue.end(); ++p)
		if (!isdigit(*p))
			Die("Invalid integer attribute %s=%s",
			  Name.c_str(), strValue.c_str());
	Value = (unsigned) atoi(strValue.c_str());
	return true;
	}

// This is more tolerant that UCSC GTF format, which is accepted as a subset.
// UCSC requires attributes to end in a semicolon which must then be separated
// from the start of any subsequent attribute by exactly one space character.
// I allow abitrary white space (including none) before and after the semi-colon,
// and also allow end-of-line to terminate the last attribute.
// <Idle><InName><AfterName>("<InStringValue>"|<InNumericValue>)<AfterValue>(;|EndOfLine)<Idle>
bool GFFGetAttr(const std::string &Attrs, const std::string &Name,
  std::string &Value)
	{
	Value.clear();
	enum
		{
		Idle,
		InName,
		AfterName,
		InStringValue,
		InNumericValue,
		} State = Idle;

	std::string CurrentName;
	std::string CurrentValue;
#if	TRACE
	Log("State  c       CurrentName      CurrentValue\n");
	Log("=====  =  ================  ================\n");
#endif
	for (std::string::const_iterator p = Attrs.begin(); p != Attrs.end(); ++p)
		{
		char c = *p;
#if	TRACE
		Log("%5d  %c  %16.16s  %16.16s\n",
		  (int) State, c, CurrentName.c_str(), CurrentValue.c_str());
#endif
		if (State == Idle && CurrentName == Name)
			{
			Value = CurrentValue;
			return true;
			}
		switch (State)
			{
		case Idle:
			if (isspace(c))
				continue;
			else if (c == ';')
				continue;
			else if (isalpha(c))
				{
				CurrentName = c;
				CurrentValue.clear();
				State = InName;
				}
			else
				Die("Parsing GFF rec, unexpected '%c' in attrs '%s", c, Attrs.c_str());
			break;

		case InName:
			if (isspace(c))
				State = AfterName;
			else if (c == '"')
				{
				CurrentValue.clear();
				State = InStringValue;
				}
			else
				CurrentName += c;
			break;

		case AfterName:
			if (isspace(c))
				continue;
			else if (c == '"')
				{
				CurrentValue.clear();
				State = InStringValue;
				}
			else if (isdigit(c) || c == '.')
				{
				CurrentValue = c;
				State = InNumericValue;
				continue;
				}
			else
				Die("Parsing GFF rec, unexpected '%c' after '%s' in attrs '%s'",
				  c, CurrentName.c_str(), Attrs.c_str());
			break;

		case InStringValue:
			if (c == '"')
				{
				State = Idle;
				continue;
				}
			CurrentValue += c;
			break;

		case InNumericValue:
			if (c == ';' || isspace(c))
				{
				State = Idle;
				continue;
				}
			else if (isdigit(c) || c == '.')
				CurrentValue += c;
			else
				Die("Parsing GFF rec, '%c' invalid in numeric attribute '%s'",
				  c, CurrentName.c_str());
			break;

		default:
			asserta(false);
			}
		}
	if (State != Idle && State != InNumericValue)
		Die("Unexpected end-of-line in attributes '%s'", Attrs.c_str());

	if (CurrentName == Name)
		{
		Value = CurrentValue;
		return true;
		}
	return false;
	}

//  Target SeqName 123 456
//  1234567
void ParseTargetAttrs(const char *Attrs, char SeqName[],
  int SeqNameBytes, coord_t *ptrStart, coord_t *ptrEnd)
	{
	const char *ptrTarget = strstr(Attrs, "Target ");
	if (0 == ptrTarget)
		Die("Target attribute not found");

	const char *ptrRest = ptrTarget + 7;
	const char *ptrBlank = strchr(ptrRest, ' ');
	if (0 == ptrBlank)
		Die("Invalid Target attributes '%s', missing blank", Attrs);
	size_t NameLength = ptrBlank - ptrRest;
	if (NameLength >= (size_t) SeqNameBytes)
		Die("Target name length too long %u,%u '%s'",
		  (unsigned) NameLength, (unsigned) SeqNameBytes, Attrs);
	memcpy(SeqName, ptrRest, NameLength);
	SeqName[NameLength] = 0;
	UnfixWhite(SeqName);
	int n = sscanf(ptrBlank+1, "%u %u", ptrStart, ptrEnd);
	if (n != 2)
		Die("Invalid Target start/end attributes '%s', sscanf=%d", ptrBlank+1, n);
	}

//  SP 456
//  1234
bool ParseSPAttrs(const char *Attrs, unsigned *ptrSPIndex)
	{
	const char *ptr = strstr(Attrs, "SP ");
	if (0 == ptr)
		return false;

	int n = sscanf(ptr+3, "%u", ptrSPIndex);
	if (n != 1)
		Die("Invalid SP attributes '%s', sscanf=%d", ptr, n);
	return true;
	}

//  SPs 456 789
//  1234
void ParseSPsAttrs(const char *Attrs, unsigned *ptrSPIndex1, unsigned *ptrSPIndex2)
	{
	const char *ptr = strstr(Attrs, "SPs ");
	if (0 == ptr)
		Die("SPs attribute not found");

	int n = sscanf(ptr+4, "%u %u", ptrSPIndex1, ptrSPIndex2);
	if (n != 2)
		Die("Invalid SPs attributes '%s', sscanf=%d", ptr, n);
	}

//  Bundle 456
//  12345678
bool ParseBundleAttrs(const char *Attrs, unsigned *ptrBundleIndex)
	{
	const char *ptr = strstr(Attrs, "Bundle[");
	if (0 == ptr)
		return false;

	int n = sscanf(ptr+7, "%u", ptrBundleIndex);
	if (n != 1)
		Die("Invalid Bundle attributes '%s', sscanf=%d", ptr, n);
	return true;
	}

//  AlnLen[456]
//  1234567
bool ParseAlnLenAttrs(const char *Attrs, unsigned *ptrAlnLen)
	{
	const char *ptr = strstr(Attrs, "AlnLen[");
	if (0 == ptr)
		return false;

	int n = sscanf(ptr+7, "%u", ptrAlnLen);
	if (n != 1)
		Die("Invalid AlnLen attributes '%s', sscanf=%d", ptr, n);
	return true;
	}

// Destructive read -- pokes nuls onto FS
static int GetFields(char *Line, char **Fields, int MaxFields, char FS)
	{
	char *p = Line;
	for (int FieldIndex = 0; FieldIndex < MaxFields; ++FieldIndex)
		{
		Fields[FieldIndex] = p;
		char *Tab = strchr(p, FS);
		char *End = Tab;
		if (0 == End)
			End = strchr(p, '\0');
		size_t FieldLength = End - p;
		if (FieldLength > MAX_GFF_FEATURE_LENGTH)
			Die("Max GFF field length exceeded, field is %d chars, max=%d, line %d",
			  FieldLength, MAX_GFF_FEATURE_LENGTH, GFFLineNr);
		if (0 == Tab)
			return FieldIndex + 1;
		*Tab = 0;
		p = Tab + 1;
		}
	return MaxFields;
	}

bool GetNextGFFRecord(FILE *f, GFFRecord &Rec)
	{
	for (;;)
		{
		++GFFLineNr;
		char *Ok = fgets(Line, sizeof(Line), f);
		if (NULL == Ok)
			{
			if (feof(f))
				return false;
			Die("Error reading GFF file, line=%d feof=%d ftell=%d ferror=%d errno=%d",
			  GFFLineNr, feof(f), ftell(f), ferror(f), errno);
			}
		if ('#' == Line[0])
			continue;
		size_t n = strlen(Line);
		if (0 == n)
			Die("fgets returned zero-length line");
		if (Line[n-1] != '\n')
			{
			if (n == sizeof(Line) - 1)
				Die("Max line length in GFF file exceeded, line %d is %d chars long, max=%d",
				  GFFLineNr, n - 1, MAX_GFF_LINE);
			else
				Die("Missing end-of-line in GFF file line %d", GFFLineNr);
			}
		Line[n-1] = 0;	// delete newline

		int FieldCount = GetFields(Line, Fields, 9, '\t');
		if (FieldCount < 8)
			Die("GFF record has < 8 fields, line %d", GFFLineNr);

		char *SeqName = Fields[0];
		char *Source = Fields[1];
		char *Feature = Fields[2];
		char *Start = Fields[3];
		char *End = Fields[4];
		char *Score = Fields[5];
		char *Strand = Fields[6];
		char *Frame = Fields[7];
		char *Attrs = (FieldCount == 9 ? Fields[8] : &Line[n-1]);

		UnfixWhite(SeqName);

	// Truncate attrs if comment found
		char *Pound = strchr(Attrs, '#');
		if (0 != Pound)
			*Pound = 0;

		Rec.SeqName = SeqName;
		Rec.Source = Source;
		Rec.Feature = Feature;
		Rec.Start = atoi(Start);
		Rec.End = atoi(End);
		Rec.Score = (float) atof(Score);
		Rec.Strand = Strand[0];
		Rec.Frame = Frame[0] == '.' ? INT_MAX : atoi(Frame);
		Rec.Attrs = Attrs;

		return true;
		}
	}

void SaveGFFStrings(GFFRecord &Rec)
	{
	Rec.SeqName = strsave(Rec.SeqName);
	Rec.Feature = strsave(Rec.Feature);
	Rec.Source = strsave(Rec.Source);
	Rec.Attrs = strsave(Rec.Attrs);
	}

void FreeGFFStrings(GFFRecord &Rec)
	{
	free((void *) Rec.SeqName);
	free((void *) Rec.Feature);
	free((void *) Rec.Source);
	free((void *) Rec.Attrs);

	Rec.SeqName = 0;
	Rec.Feature = 0;
	Rec.Source = 0;
	Rec.Attrs = 0;
	}

bool HasTargetAttrs(const char *Attrs)
	{
// not 100% robust, but should be good enough in practice
	return strstr(Attrs, "Target") != 0;
	}

// GFF fields are:
// <seqname> <source> <feature> <start> <end> <score> <strand> <frame> [attributes] [comments]
//     0         1         2        3      4      5        6       7         8           9
void WriteGFFRec(FILE *f, const GFFRecord &Rec)
	{
	fprintf(f, "%s\t%s\t%s", Rec.SeqName, Rec.Source, Rec.Feature);
	fprintf(f, "\t%"COORD_T_PRINTF, Rec.Start);
	fprintf(f, "\t%"COORD_T_PRINTF, Rec.End);
	fprintf(f, "\t%g\t%c", Rec.Score, Rec.Strand);
	if (Rec.Frame == INT_MAX)
		fprintf(f, "\t.");
	else
		fprintf(f, "\t%d", Rec.Frame);
	fprintf(f, "\t%s\n", Rec.Attrs);
	}

void LogGFFRec(const GFFRecord &Rec)
	{
	Log("%s %s %s", Rec.SeqName, Rec.Source, Rec.Feature);
	Log(" Start=%"COORD_T_PRINTF, Rec.Start);
	Log(" End=%"COORD_T_PRINTF, Rec.End);
	Log(" Score=%g %c", Rec.Score, Rec.Strand);
	if (Rec.Frame == -1)
		Log(" .");
	else
		Log(" %d", Rec.Frame);
	Log(" %s\n", Rec.Attrs);
	}
