#include "utils.h"
#include "revfile.h"
#include "progress.h"
#include "getopts.h"
#include "uniquebid.h"

#define TRACE			0
#define TRACEOV			0

#if	TRACE
static unsigned g_TraceCounter = 0;
const unsigned g_MaxBPsTraced = 0;
const unsigned g_MaxBPsTracedWindow = 0;
#endif

#ifdef _MSC_VER
#pragma warning(disable : 4244)	// score_t->ai_score_t conversion
#endif

/***
Glover: Global Overlap Reduction

Deletes hits by the following rule:

	If hit X is covered by hit Y with respect to any genome,
	and the bundle containing hit X scores sufficiently less
	than the chain containing hit Y, then delete X.

  ^
  |                         / Bundle 2
  |                        / high
  |                         score
  |     Bundle 1 /       /
G |         low / X     /
e |      score         /
n |           /       / Y
o |          /       /
m |             
e |                /
  |               / 
B |              /
  |             /
   ------------------------------------>
                           Genome A

***/

void TruncLabelGenomes(std::vector<GenInfo> &Genomes);
void GetBendpoints(const AIVec &AIs, unsigned GenomeIndex, AIBPVec &BPs,
  unsigned Margin);
coord_t GetOverlapLength(const AlnInfo &AIA, const AlnInfo &AIB,
  unsigned GenomeIndex);
coord_t GetOverlapLength(const GCSeg &SegA, const GCSeg &SegB);
void LogAI(const AlnInfo &AI);

const double DefaultMarginPct = 5.0;

static unsigned g_Margin = 16;
static double g_MarginFract = DefaultMarginPct/100.0;
static double g_MaxBundleScoreRatio = 0.5;
static unsigned g_DeletedHitCount;

AIVec g_AIs;
BoolVec g_Del;	// true if aln is to be deleted
std::map<UniqueBundleId, score_t> g_BundleScores;

RevFile *g_RevDel = 0;

static bool Covers(const AlnInfo &AIA, const AlnInfo &AIB,
  score_t BundleScoreA, score_t BundleScoreB, unsigned GenomeIndex)
	{
#if	TRACE
	Log("Covers BndScoreA=%g BndScoreB=%g", BundleScoreA, BundleScoreB);
#endif
	if (BundleScoreA != 0)
		{
		double Ratio = (double) BundleScoreB / (double) BundleScoreA;
		if (Ratio > g_MaxBundleScoreRatio)
			{
#if	TRACE
			Log(" Ratio %g > MaxRatio %g =false\n", Ratio, g_MaxBundleScoreRatio);
#endif
			return false;
			}
		}

	const GCSeg &SegA = AIA.GetSeg(GenomeIndex);
	const GCSeg &SegB = AIB.GetSeg(GenomeIndex);

	unsigned Margin = g_Margin;
	unsigned FractMargin = (unsigned) ((double) SegA.GetLength()*g_MarginFract);
	if (FractMargin < Margin)
		Margin = FractMargin;

#if	TRACE
	Log(" SegA=");
	LogGCSeg(SegA);
	Log(" SegB=");
	LogGCSeg(SegB);
#endif

	coord_t lo = SegA.lo;
	if (lo >= Margin)
		lo -= Margin;
	else
		lo = 0;

	if (SegB.lo < lo)
		{
#if	TRACE
		Log(" SegB.lo %u < lo %u = false", (unsigned) SegB.lo, (unsigned) lo);
#endif
		return false;
		}

	coord_t hi = SegA.hi + Margin;
	bool Cov = (SegB.hi <= hi);
#if	TRACE
	Log(" SegB.hi %u %s hi %u =%s\n",
	  (unsigned) SegB.hi,
	  (SegB.hi <= hi) ? ">=" : "<",
	  (unsigned) hi,
	  Cov ? "true" : "false");
#endif
	return Cov;
	}

static void LogBundleScores()
	{
	Log("\n");
	Log("g_BundleScores=\n");
	Log("\n");
	Log("  Chr0   Chr1   Bundle  Score\n");
	Log("======  =====  ======  =====\n");
	for (std::map<UniqueBundleId, score_t>::const_iterator p = g_BundleScores.begin();
	  p != g_BundleScores.end(); ++p)
		{
		const UniqueBundleId &UB = p->first;
		sp_score_t Score = (sp_score_t) p->second;
		Log("%5u  %5u  %6u  %s\n", UB.ChrId0, UB.ChrId1, UB.BundleId, SPScoreToStr(Score));
		}
	}

score_t GetBundleScore(const UniqueBundleId &UB)
	{
	std::map<UniqueBundleId, score_t>::const_iterator p = g_BundleScores.find(UB);
	if (p == g_BundleScores.end())
		{
		LogBundleScores();
		Die("GetBundleScore Chr0 %u Chr1 %u Bundle %u not found",
		  UB.ChrId0, UB.ChrId1, UB.BundleId);
		}
	score_t Score = p->second;
	return Score;
	}

score_t GetBundleScore(const AlnInfo &AI)
	{
	unsigned BundleIndex = AI.GetBundleId();
	const GCSeg &Seg0 = AI.GetSeg(0);
	const GCSeg &Seg1 = AI.GetSeg(1);

	UniqueBundleId UB;
	UB.BundleId = BundleIndex;
	UB.ChrId0 = Seg0.ChrId;
	UB.ChrId1 = Seg1.ChrId;
	return GetBundleScore(UB);
	}

static void ReduceOverlaps(unsigned GenomeIndex)
	{
	const unsigned OtherGenomeIndex = (GenomeIndex == 0 ? 1 : 0);

// Get hit bendpoints.
// Can't use a margin here because 'coverage' of hit A
// by hit B requires expanding B but not A.
	std::vector<AIBP> BPs;
	GetBendpoints(g_AIs, GenomeIndex, BPs, 0);

// Sliding window for overlap detection.
// A hit enters a list when its lo point enters the window,
// and exits the list when its hi point exits the window.
// When the hi point for hit H exits, all members of the list
// are candidates for covering H or being covered by H.
	IntList Window;
	const unsigned BPCount = (unsigned) BPs.size();
	const unsigned AICount = (unsigned) g_AIs.size();
	unsigned PrevChrId = UINT_MAX;
	ProgressStart("Reduce overlaps w.r.t. genome %u", GenomeIndex);
	for (unsigned BPIndex = 0; BPIndex < BPCount; ++BPIndex)
		{
		if (BPIndex%10000 == 0)
			ProgressStep(BPIndex, BPCount);
		const AIBP &BP = BPs[BPIndex];
		unsigned AIIndex = BP.AlnInfoIndex;
		assert(AIIndex < AICount);

#if	TRACE
		{
		const GCSeg &Seg = g_AIs[AIIndex].GetSeg(GenomeIndex);
		unsigned lo = (unsigned) Seg.lo;
		unsigned hi = (unsigned) Seg.hi;
		Log("BP: Pos=%u %s AI=%u lo=%u hi=%u del=%c\n",
		  BP.Pos,
		  BP.IsLo ? "lo" : "hi",
		  BP.AlnInfoIndex,
		  lo,
		  hi,
		  g_Del[AIIndex] ? 'Y' : 'N');
		Log("Window: ");
		LogList(Window);
		Log("\n");
		if (g_MaxBPsTraced > 0 && ++g_TraceCounter >= g_MaxBPsTraced)
			Die("Traced %u BPs", g_TraceCounter);
		}
#endif
		if (BP.ChrId != PrevChrId)
			{
			PrevChrId = BP.ChrId;
			Window.clear();
			}

		if (BP.IsLo)
		// lo: add to window
			Window.push_back(AIIndex);
		else
			{
		// hi: test hits in window for covering this hit.
			assert(!BP.IsLo);

			const AlnInfo &AI = g_AIs[AIIndex];
			const GCSeg &SegOther = AI.GetSeg(OtherGenomeIndex);
			unsigned ChrIdOther = SegOther.ChrId;
			unsigned BundleIndex = AI.GetBundleId();
			score_t BundleScore = GetBundleScore(AI);

		// Find current hit during window scan
			IntList::iterator ptrCurrentHit = Window.end();
			for (IntList::iterator p = Window.begin(); p != Window.end(); ++p)
				{
				unsigned AIIndex2 = *p;
				assert(AIIndex2 < AICount);

				if (AIIndex2 == AIIndex)
					{
					ptrCurrentHit = p;
#if	TRACE
					Log("    Skip: self\n");
#endif
					continue;
					}

				const AlnInfo &AI2 = g_AIs[AIIndex2];
				unsigned BundleIndex2 = AI2.GetBundleId();
				const GCSeg &SegOther2 = AI2.GetSeg(OtherGenomeIndex);
				unsigned ChrIdOther2 = SegOther2.ChrId;
				score_t BundleScore2 = GetBundleScore(AI2);
#if	TRACEOV
				{
				if (Window.size() > 1 && ++g_TraceCounter >= g_MaxBPsTracedWindow)
					Die("Traced %u BPs", g_TraceCounter);
				Log("Window: ");
				LogList(Window);
				Log("\n");
				const GCSeg &Seg = AI.GetSeg(GenomeIndex);
				const GCSeg &Seg2 = AI2.GetSeg(GenomeIndex);
				static bool FirstTime = true;
				if (FirstTime)
					{
					FirstTime = false;
Log(" ChrA         loA         hiA   ChrB         loB         hiB     Ov%%  DelA  DelB   BndA   BndB  BndScoreA  BndScoreB  Cov\n");
Log("=====  ==========  ==========  =====  ==========  ==========  ======  ====  ====  =====  =====  =========  =========  ===\n");
					}

				char Cov = ' ';
				if (Covers(AI, AI2, BundleScore, BundleScore2, GenomeIndex))
					Cov = '>';
				else if (Covers(AI2, AI, BundleScore2, BundleScore, GenomeIndex))
					Cov = '<';

				coord_t Ov = GetOverlapLength(Seg, Seg2);
				coord_t Len = Seg.GetLength();
				coord_t Len2 = Seg2.GetLength();
				coord_t MinLen = std::min(Len, Len2);
				assert(MinLen > 0);
				double OvPct = (double) Ov*100.0/ (double) MinLen;
				Log("%5u  %10u  %10u  %5u  %10u  %10u  %6.1f  %4c  %4c  %5u  %5u  %9g  %9g  %3c\n",
				  ChrIdOther,
				  Seg.lo,
				  Seg.hi,
				  ChrIdOther2,
				  Seg2.lo,
				  Seg2.hi,
				  OvPct,
				  g_Del[AIIndex] ? 'Y' : 'N',
				  g_Del[AIIndex2] ? 'Y' : 'N',
				  BundleIndex,
				  BundleIndex2,
				  (double) BundleScore,
				  (double) BundleScore2,
				  Cov);
				}
#endif
#if	TRACE
				{
				const GCSeg &Seg2 = AI2.GetSeg(GenomeIndex);
				Log("  AIIndex2=%u lo=%u hi=%u\n",
				  AIIndex2, (unsigned) Seg2.lo, (unsigned) Seg2.hi);
				}
#endif

			// Don't try to resolve overlaps within a bundle,
			// that should be done by some other algorithm (repfix).
				if (ChrIdOther == ChrIdOther2 && BundleIndex == BundleIndex2)
					{
#if	TRACE
					Log("    Skip: same bundle\n");
#endif
					continue;
					}

				if (Covers(AI, AI2, BundleScore, BundleScore2, GenomeIndex))
					{
#if	TRACE
					Log("    Cov(%u > %u), AIIndex2=%u deleted\n",
					  AIIndex, AIIndex2, AIIndex2);
#endif
					g_Del[AIIndex2] = true;
					}
				else if (Covers(AI2, AI, BundleScore2, BundleScore, GenomeIndex))
					{
#if	TRACE
					Log("    Cov(%u > %u), AIIndex=%u deleted\n",
					  AIIndex2, AIIndex, AIIndex);
#endif
					g_Del[AIIndex] = true;
					}
				else
					{
#if	TRACE
					Log("    No cov\n");
#endif
					;
					}
				}
			
			if (ptrCurrentHit == Window.end())
				Die("GloverGenome, current hit not in window");

			asserta(*ptrCurrentHit == BP.AlnInfoIndex);
		// Delete current hit from list
			Window.erase(ptrCurrentHit);
			}
		}
	ProgressDone();
	}

void Output(const RevFile &In, RevFile &Out)
	{
	const unsigned AICount = (unsigned) g_AIs.size();
	ProgressStart("Building output");
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%100 == 0)
			ProgressStep(AIIndex, AICount);

		bool Del = g_Del[AIIndex];
		if (Del)
			++g_DeletedHitCount;

		const AlnInfo &AI = g_AIs[AIIndex];
		unsigned AlnIndex = In.GetAlnIndex(AIIndex);
		if (AlnIndex == UINT_MAX)
			{
			if (Del)
				{
				if (g_RevDel != 0)
					g_RevDel->AddAlnInfo(AI);
				}
			else
				Out.AddAlnInfo(AI);
			}
		else
			{
			StdStrVec Rows;
			In.GetAlnRows(AlnIndex, Rows);

			if (Del)
				{
				if (g_RevDel != 0)
					g_RevDel->AddAln(AI, Rows);
				}
			else
				Out.AddAln(AI, Rows);
			}
		}
	ProgressDone();
	}

void Glover()
	{
	const char *InFileName = RequiredValueOpt("in");
	const char *OutFileName = RequiredValueOpt("out");
	const char *DelFileName = ValueOpt("del");

	IntOpt("margin", &g_Margin);
	FloatOpt("maxbundlescoreratio", &g_MaxBundleScoreRatio);

	double MarginPct = DefaultMarginPct;
	FloatOpt("marginpct", &MarginPct);
	g_MarginFract = MarginPct/100.0;

	RevFile In;
	RevFile Out;

	In.Open(InFileName);
	if (In.GetGenomeCount() != 2)
		Die("glover requires pairwise alignment, genome count > 2");

	if (DelFileName != 0)
		{
		g_RevDel = new RevFile;
		const std::vector<GenInfo> &Genomes = In.GetGenomes();
		g_RevDel->Create(DelFileName, Genomes);
		}

	std::vector<GenInfo> Genomes = In.GetGenomes();

	TruncLabelGenomes(Genomes);

	Out.Create(OutFileName, Genomes);

	Progress("Get all AIs");
	In.GetAllAlnInfos(g_AIs);

	const unsigned AICount = (unsigned) g_AIs.size();
	g_Del.resize(AICount);
	for (unsigned i = 0; i < AICount; ++i)
		g_Del[i] = false;

	Progress("Compute bundle scores");
	ComputeBundleScores(g_AIs, g_BundleScores);
	ReduceOverlaps(0);
	ReduceOverlaps(1);
	Output(In, Out);

	double HitCount = (double) g_AIs.size();
	double Pct = (double) g_DeletedHitCount * 100.0 / HitCount;
	Progress("%u hits deleted = %.1f%%", g_DeletedHitCount, Pct);

	Progress("Closing files and writing output");
	In.Close();
	Out.Close();

	if (g_RevDel != 0)
		{
		g_RevDel->Close();
		delete g_RevDel;
		g_RevDel = 0;
		}

	ProgressExit();
	}
