#include "utils.h"
#include "revfile.h"
#include "progress.h"
#include "getopts.h"

#define TRACE	0

void LogAI(const AlnInfo &AI);
void GetBendpoints(const std::vector<AlnInfo> &AIs, unsigned GenomeIndex,
  std::vector<AIBP> &BPs, unsigned Margin);
coord_t GetOverlapLength(const AlnInfo &AIA, const AlnInfo &AIB,
  unsigned GenomeIndex);
const char *TruncLabel(const char *Label);

static unsigned g_HitBases[2];
static unsigned g_RepBaseCount[2];

static unsigned g_Count1[2];
static unsigned g_Count2[2];
static unsigned g_Count3To4[2];
static unsigned g_Count5To8[2];
static unsigned g_Count9To16[2];
static unsigned g_Count17To32[2];
static unsigned g_Count33To64[2];
static unsigned g_CountGT64[2];

static AIVec g_AIs;

static unsigned GetSegLength(const AlnInfo &AI, unsigned GenomeIndex)
	{
	const GCSeg *Seg = AI.GetSegPtr(GenomeIndex);
	if (Seg == 0)
		Die("Missing seg");
	coord_t L = Seg->GetLength();
	return (unsigned) L;
	}

static coord_t GetGenomeLength(const RevFile &Rev, unsigned GenomeIndex)
	{
	const GenInfo &GI = Rev.GetGenomeInfo(GenomeIndex);
	return GI.GetLength();
	}

static void LogStats(const RevFile &Rev)
	{
	for (unsigned GenIx = 0; GenIx <= 1; ++GenIx)
		{
		const std::string &Name = Rev.GetGenomeName(GenIx);
		coord_t GenomeLength = GetGenomeLength(Rev, GenIx);

		if (g_HitBases[GenIx] == 0)
			Die("No hit bases for genome %u", GenIx);
		double Pct = g_RepBaseCount[GenIx]*100.0/(double) g_HitBases[GenIx];

		Log("\n");
		Log("Genome [%u] %s\n", GenIx, Name.c_str());
		Log("\n");
		Log("Genome length   %10u\n", (unsigned) GenomeLength);
		Log("Hit cols        %10u\n", g_HitBases[GenIx]);
		Log("Excess cols     %10u  %6.1f%%\n", g_RepBaseCount[GenIx], Pct);
		Log("\n");
		Log("\n");

		unsigned Total =
			g_Count1[GenIx] +
			g_Count2[GenIx] +
			g_Count3To4[GenIx] +
			g_Count5To8[GenIx] +
			g_Count9To16[GenIx] +
			g_Count17To32[GenIx] +
			g_Count33To64[GenIx] +
			g_CountGT64[GenIx];

		Log("Repeat Count Distribution:\n");
		Log("[1]             %10u\n", g_Count1[GenIx]);
		Log("[2]             %10u\n", g_Count2[GenIx]);
		Log("[3-4]           %10u\n", g_Count3To4[GenIx]);
		Log("[5-8]           %10u\n", g_Count5To8[GenIx]);
		Log("[9-16]          %10u\n", g_Count9To16[GenIx]);
		Log("[17-32]         %10u\n", g_Count17To32[GenIx]);
		Log("[33-64]         %10u\n", g_Count33To64[GenIx]);
		Log("[>64]           %10u\n", g_CountGT64[GenIx]);
		Log("                ==========\n");
		Log("                %10u\n", Total);
		}
	}

static void ReadAllAIs(const RevFile &Rev)
	{
	const unsigned AICount = Rev.GetAlnInfoCount();
	g_AIs.resize(AICount);

	ProgressStart("Read AlnInfos");
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%1000 == 0)
			ProgressStep(AIIndex, AICount);
		Rev.GetAlnInfo(AIIndex, g_AIs[AIIndex]);
		}
	ProgressDone();
	Progress("%u hits", (unsigned) g_AIs.size());
	}

static void GlobalOverlapStatsGenome(const std::vector<AlnInfo> &AIs,
  unsigned GenomeIndex)
	{
	g_HitBases[GenomeIndex] = 0;
	g_RepBaseCount[GenomeIndex] = 0;

	const unsigned AICount = (unsigned) AIs.size();
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		const AlnInfo &AI = AIs[AIIndex];
		unsigned SegLength = GetSegLength(AI, GenomeIndex);
		g_HitBases[GenomeIndex] += SegLength;
		}

	std::vector<AIBP> BPs;
	GetBendpoints(AIs, GenomeIndex, BPs, 0);

/***
Goal is to compute RepBaseCount. Each base has a "copy count",
i.e. the number of hits that cover the base. RepBaseCount is
the sum of over all bases that are covered by more than one hit
of (copy count - 1). This is a measure of the "excess" due
to overlaps.

This is accomplished by sorting bendpoints and scanning from
left to right, tracking the copy count.
***/

	int CopyCount = 0;
	unsigned RepBaseCount = 0;
	coord_t LastBendpointPos = 0;
	unsigned LastChrId = UINT_MAX;
	const unsigned BPCount = (unsigned) BPs.size();
#if	TRACE
	Log("\n");
	Log(" CC  BaseC  RepBaseC     LastPos  lh         Pos    ChrId\n");
	Log("===  =====  ========  ==========  ==  ==========  =======\n");
#endif
	for (unsigned BPIndex = 0; BPIndex < BPCount; ++BPIndex)
		{
		const AIBP &BP = BPs[BPIndex];
		unsigned AIIndex = BP.AlnInfoIndex;
		assert(AIIndex < AICount);

		const AlnInfo &AI = AIs[AIIndex];
		const GCSeg &Seg = AI.GetSeg(GenomeIndex);
		unsigned BaseCount = BP.Pos - LastBendpointPos + 1;
#if	TRACE
		{
		Log("%3d  %5u  %8d  %10u  %2s  %10u  %7u\n",
		  CopyCount,
		  BaseCount,
		  RepBaseCount,
		  (unsigned) LastBendpointPos,
		  (BP.IsLo ? "lo" : "hi"),
		  (unsigned) BP.Pos,
		  Seg.ChrId);
		}
#endif
		if (Seg.ChrId != LastChrId)
			{
			LastChrId = Seg.ChrId;
			CopyCount = 0;
			LastBendpointPos = COORD_T_MAX;
			}

		if (CopyCount > 0)
			{
			unsigned ExcessAlnColCount = BaseCount*(CopyCount - 1);
			RepBaseCount += ExcessAlnColCount;

			if (CopyCount == 1)
				g_Count1[GenomeIndex] += BaseCount;
			else if (CopyCount == 2)
				g_Count2[GenomeIndex] += BaseCount;
			else if (CopyCount <= 4)
				g_Count3To4[GenomeIndex] += BaseCount;
			else if (CopyCount <= 8)
				g_Count5To8[GenomeIndex] += BaseCount;
			else if (CopyCount <= 16)
				g_Count9To16[GenomeIndex] += BaseCount;
			else if (CopyCount <= 32)
				g_Count17To32[GenomeIndex] += BaseCount;
			else if (CopyCount <= 64)
				g_Count33To64[GenomeIndex] += BaseCount;
			else
				g_CountGT64[GenomeIndex] += BaseCount;
			}

		if (BP.IsLo)
			++CopyCount;
		else
			{
			if (CopyCount <= 0)
				Die("CopyCount <= 0");
			--CopyCount;
			}

		LastBendpointPos = BP.Pos;
		}

	if (CopyCount != 0)
		Die("CopyCount=%d on exit, should be zero");

	g_RepBaseCount[GenomeIndex] = RepBaseCount;
	}

// Write Global Overlap Statistics to log.
void GlobalOverlapStats()
	{
	const char *RevFileName = RequiredValueOpt("stats");

	RevFile Rev;
	Rev.Open(RevFileName);
	if (Rev.GetGenomeCount() != 2)
		Die("glover -stats requires pairwise alignment, genome count > 2");

	ReadAllAIs(Rev);

	for (unsigned GenomeIndex = 0; GenomeIndex <= 1; ++GenomeIndex)
		{
		Progress("Genome [%u] %s", GenomeIndex, Rev.GetGenomeName(GenomeIndex).c_str());
		GlobalOverlapStatsGenome(g_AIs, GenomeIndex);
		}

	LogStats(Rev);
	}
