#include "ice.h"
#include "iovecs.h"
#include "progress.h"
#include "myrand.h"
#include "ratemgr.h"

void WriteStr(FILE *f, const string &s)
	{
	const unsigned L = SIZE(s);
	WRITEV(L);
	for (unsigned j = 0; j < L; ++j)
		WRITEV(s[j]);
	}

void ReadStr(FILE *f, string &s)
	{
	s.clear();
	unsigned L;
	READV(L);
	s.reserve(L);
	for (unsigned j = 0; j < L; ++j)
		{
		char c;
		READV(c);
		s.push_back(c);
		}
	}

void InterChrEvolver::FromFile(const char *FileName)
	{
	Progress("Reading state from %.32s", FileName);
	Clear();

	FILE *f = OpenStdioFile(FileName);

	unsigned Magic = 0;
	READV(Magic);
	if (Magic != IceFileMagic)
		Die("Invalid ice file");

	READV(m_Duration);
	READV(m_EventCount);
	READV(m_EventIndex);
	READV(m_LastAcceptedEventIndex);
	READV(m_LastAcceptedEventPos);
	READV(m_CurrentEventPos);
	READV(m_LastAcceptedEvent);
	READV(m_CurrentEvent);
	READV(m_NextNewGeneIndex);

	ReadVec(f, m_IntegratedEventRates);
	ReadVec(f, m_AncChrLengths);
	ReadVec(f, m_ChrLengths);
	Progress("%u events, %u anc chrs, %u ev chrs",
	  SIZE(m_IntegratedEventRates),
	  SIZE(m_AncChrLengths),
	  SIZE(m_ChrLengths));

	g_InterChrMoveRateMgr->FromFile(f);
	g_InterChrCopyRateMgr->FromFile(f);
	ReadRandState(f);

	unsigned ContigCount;
	READV(ContigCount);
	ProgressStart("%u contigs", ContigCount);
	for (unsigned i = 0; i < ContigCount; ++i)
		{
		if (i%1000 == 0)
			ProgressStep(i, ContigCount);
		ContigData Contig;
		READV(Contig);
		m_Contigs.insert(Contig);
		}
	ProgressDone();
	
	unsigned ChrCount;
	READV(ChrCount);
	Progress("%u chr names", ChrCount);
	for (unsigned i = 0; i < ChrCount; ++i)
		{
		string Name;
		ReadStr(f, Name);
		m_ChrNames.push_back(Name);
		}

	unsigned AncChrCount;
	READV(AncChrCount);
	Progress("%u anc chr names", ChrCount);
	for (unsigned i = 0; i < AncChrCount; ++i)
		{
		string Name;
		ReadStr(f, Name);
		m_AncChrNames.push_back(Name);
		}

	unsigned AncAnnotCount;
	READV(AncAnnotCount);
	ProgressStart("%u anc annots", AncAnnotCount);
	for (unsigned i = 0; i < AncAnnotCount; ++i)
		{
		if (i%1000 == 0)
			ProgressStep(i, AncAnnotCount);
		ChrPosData CP;
		GFFRecord2 Rec;

		READV(CP);
		bool Ok = GetNextGFFRecord(f, Rec);
		if (!Ok)
			Die("End-of-file reading GFF record in ice file");

		m_AncAnnots[CP] = Rec;
		}
	ProgressDone();

	unsigned GeneAccelCount;
	READV(GeneAccelCount);
	Progress("%u gene accels", AncAnnotCount);
	for (unsigned i = 0; i < GeneAccelCount; ++i)
		{
		unsigned GeneIndex;
		double Rate;

		READV(GeneIndex);
		READV(Rate);

		m_GeneAccels[GeneIndex] = Rate;
		}

	CloseStdioFile(f);
	}

void InterChrEvolver::ToFile(const char *FileName) const
	{
	Progress("Writing state to %.32s", FileName);
	FILE *f = CreateStdioFile(FileName);
	Log("Writing state to %.32s f=%p\n", FileName, f);

	unsigned MagicPlaceholder = 0;
	WRITEV(MagicPlaceholder);

	WRITEV(m_Duration);
	WRITEV(m_EventCount);
	WRITEV(m_EventIndex);
	WRITEV(m_LastAcceptedEventIndex);
	WRITEV(m_LastAcceptedEventPos);
	WRITEV(m_CurrentEventPos);
	WRITEV(m_LastAcceptedEvent);
	WRITEV(m_CurrentEvent);
	WRITEV(m_NextNewGeneIndex);

	WriteVec(f, m_IntegratedEventRates);
	WriteVec(f, m_AncChrLengths);
	WriteVec(f, m_ChrLengths);

	g_InterChrMoveRateMgr->ToFile(f);
	g_InterChrCopyRateMgr->ToFile(f);
	WriteRandState(f);

	const unsigned ContigCount = SIZE(m_Contigs);
	WRITEV(ContigCount);
	for (ContigSet::const_iterator p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		WRITEV(Contig);
		}
	
	const unsigned ChrCount = SIZE(m_ChrNames);
	WRITEV(ChrCount);
	for (unsigned i = 0; i < ChrCount; ++i)
		{
		const string &Name = m_ChrNames[i];
		WriteStr(f, Name);
		}

	const unsigned AncChrCount = SIZE(m_AncChrNames);
	WRITEV(AncChrCount);
	for (unsigned i = 0; i < AncChrCount; ++i)
		{
		const string &Name = m_AncChrNames[i];
		WriteStr(f, Name);
		}

	const unsigned AncAnnotCount = SIZE(m_AncAnnots);
	WRITEV(AncAnnotCount);
	for (map<ChrPosData, GFFRecord2>::const_iterator p = m_AncAnnots.begin();
	  p != m_AncAnnots.end(); ++p)
		{
		const ChrPosData &CP = p->first;
		const GFFRecord2 &Rec = p->second;

		WRITEV(CP);
		WriteGFFRec(f, Rec);
		}

	const unsigned GeneAccelCount = SIZE(m_GeneAccels);
	WRITEV(GeneAccelCount);
	for (map<unsigned, double>::const_iterator p = m_GeneAccels.begin();
	  p != m_GeneAccels.end(); ++p)
		{
		unsigned GeneIndex = p->first;
		double Rate = p->second;

		WRITEV(GeneIndex);
		WRITEV(Rate);
		}

	SetStdioFilePos(f, 0);
	unsigned Magic = IceFileMagic;
	WRITEV(Magic);

	Log("close %p\n", f);
	CloseStdioFile(f);
	}
