#include <math.h>
#include "ice.h"
#include "myrand.h"
#include "evomodel.h"

#define TRACE	0

extern unsigned g_GeneDupeCount;
extern unsigned g_ActiveGeneDupeCount;

bool IsActiveResult(GENE_DUPE_RESULT Result);
double GetFasterAccel();
double GetSlowerAccel();

void InterChrEvolver::GeneDupe(unsigned FromChrIndex, gpos_t FromPos,
  unsigned ToChrIndex, gpos_t ToPos, gpos_t CopyLength, bool DoInvert)
	{
#if	TRACE
	Log("\n");
	Log("GeneDupe(FromChrIndex=%u, FromPos=%"PF", ToChrIndex=%u, ToPos=%"PF", CopyLength=%"PF", DoInvert=%c)\n",
	  FromChrIndex, FromPos, ToChrIndex, ToPos, CopyLength, tof(DoInvert));
#endif

// Cannot duplicate gene if inserted into existing gene
	if (PosInGene(ToChrIndex, ToPos))
		{
#if	TRACE
		Log("Pos in gene, quit GeneDupe().\n");
#endif
		return;
		}

	map<unsigned, pair<gpos_t, gpos_t> > LoHis;
	GetCompleteGeneRanges(FromChrIndex, FromPos, CopyLength, LoHis);
#if	TRACE
	Log("%u complete gene ranges\n", SIZE(LoHis));
#endif
	if (LoHis.empty())
		return;

	ContigSet Contigs;
	GetContigs(FromChrIndex, FromPos, CopyLength, Contigs);

	for (map<unsigned, pair<gpos_t, gpos_t> >::const_iterator p = LoHis.begin();
	  p != LoHis.end(); ++p)
		{
		unsigned GeneIndex = p->first;
		++g_GeneDupeCount;
#if	TRACE
		Log("\n");
		Log("GeneIndex %u Lo=%"PF" Hi=%"PF"\n", GeneIndex, p->second.first, p->second.second);
#endif

		double OldAccel = 0;
		double NewAccel = 0;
		GENE_DUPE_RESULT OldResult;
		GENE_DUPE_RESULT NewResult;
		GetGeneDupeResults(OldResult, NewResult);
#if	TRACE
		Log("Old=%s New %s\n",
		  GeneDupeResultToStr(OldResult),
		  GeneDupeResultToStr(NewResult));
#endif
		ContigSet OldGene;
		GetGene(GeneIndex, Contigs, OldGene);
#if	TRACE
		Log("Old gene:\n");
		LogContigs(OldGene);
#endif
		if (OldResult != GDR_Same)
			{
			OldAccel = OldResult == GDR_Faster ? GetFasterAccel() : GetSlowerAccel();
			m_GeneAccels[GeneIndex] = OldAccel;
			if (OldResult == GDR_Lost)
				DeleteGene(OldGene);
			}

		bool NewActive = IsActiveResult(NewResult);
		if (NewActive)
			{
			++g_ActiveGeneDupeCount;

			unsigned NewGeneIndex = ActivateCopiedGene(OldGene, FromChrIndex, FromPos,
			  ToChrIndex, ToPos, CopyLength, DoInvert);

			NewAccel = NewResult == GDR_Faster ? GetFasterAccel() : GetSlowerAccel();
			m_GeneAccels[NewGeneIndex] = NewAccel;
			}
		Log("@GENE_DUPE Gene %u Old %s(%.2f) New %s(%.2f)\n",
		  GeneIndex,
		  GeneDupeResultToStr(OldResult),
		  OldAccel,
		  GeneDupeResultToStr(NewResult),
		  NewAccel);
		}
	}

unsigned InterChrEvolver::ActivateCopiedGene(const ContigSet &Gene,
  unsigned /*FromChrIndex*/, gpos_t /*FromPos*/, unsigned ToChrIndex,
  gpos_t ToPos, gpos_t CopyLength, bool /*DoInvert*/)
	{
#if TRACE
	Log("ActivateCopiedGene(ToChrIndex=%u, ToPos=%"PF", CopyLength=%"PF")\n",
	  ToChrIndex, ToPos, CopyLength);
	Log("Gene:\n");
	LogContigs(Gene);
#endif
	asserta(!Gene.empty());

	unsigned NewGeneIndex = m_NextNewGeneIndex++;
	ContigSet TargetContigs;
	GetContigs(ToChrIndex, ToPos+1, CopyLength, TargetContigs);
#if	TRACE
	Log("New gene index %u\n", NewGeneIndex);
	Log("Target contigs:\n");
	LogContigs(TargetContigs);
#endif
	unsigned GeneIndex = UINT_MAX;
	ContigSet NewContigs;
	for (ContigSetCIter p = Gene.begin(); p != Gene.end(); ++p)
		{
		const ContigData &GeneContig = *p;
		if (GeneIndex == UINT_MAX)
			GeneIndex = GeneContig.GeneIndex;
		else
			asserta(GeneContig.GeneIndex == GeneIndex);
		bool Found = false;
		for (ContigSetCIter q = TargetContigs.begin();
		  q != TargetContigs.end(); ++q)
			{
			const ContigData &TargetContig = *q;
			if (TargetContig.AncChrIndex == GeneContig.AncChrIndex &&
			  TargetContig.AncPos == GeneContig.AncPos)
				{
				Found = true;
				asserta(TargetContig.Length == GeneContig.Length);
				ContigData NewContig(TargetContig);
				size_t n = m_Contigs.erase(TargetContig);
				asserta(n == 1);

				NewContig.GeneIndex = NewGeneIndex;
				NewContig.Type = GeneContig.Type;
				NewContigs.insert(NewContig);

				break;
				}
			}
		if (!Found)
			{
			Log("\n");
			GeneContig.LogMe(true);
			Die("ActivateCopiedGene failed to find GeneContig");
			}
		}
	m_Contigs.insert(NewContigs.begin(), NewContigs.end());

	Validate("End of ActivateCopiedGene");
	return NewGeneIndex;
	}

void InterChrEvolver::DeleteGene(const ContigSet &Gene)
	{
	ContigData Range;
	GetContigsRange(Gene, Range);

	ContigSet DeletedContigs;
	DeleteContigs(Range.EvChrIndex, Range.EvPos, Range.Length, DeletedContigs);

	Deconstrain(DeletedContigs, true);
	InsertContigs(Range.EvChrIndex, Range.EvPos-1, DeletedContigs);

	Validate("End of DeleteGene");
	}
