#include "ice.h"
#include "gff2.h"
#include "progress.h"

extern bool g_XferTandems;

// (chr,intra_geneindex) to inter_geneindex mapping,
// where inter_geneindex is unique over genome.
static map<pair<string, unsigned>, unsigned> g_ChrGixToGix;

static unsigned IntraToInterGeneIndex(const string &ChrName,
  unsigned IntraGeneIndex)
	{
	if (IntraGeneIndex == UINT_MAX)
		return UINT_MAX;

	pair<string, unsigned> ChrGix(ChrName, IntraGeneIndex);
	map<pair<string, unsigned>, unsigned>::const_iterator p =
	  g_ChrGixToGix.find(ChrGix);
	if (p != g_ChrGixToGix.end())
		return p->second;
	unsigned InterGeneIndex = SIZE(g_ChrGixToGix);
	g_ChrGixToGix[ChrGix] = InterGeneIndex;
	asserta(SIZE(g_ChrGixToGix) == InterGeneIndex + 1);
	return InterGeneIndex;
	}

void InterChrEvolver::ReadAnnots(const char *FileName,
  vector<GFFRecord2> &IslandAndTandemRecs)
	{
	g_ChrGixToGix.clear();
	IslandAndTandemRecs.clear();

	FILE *f = OpenStdioFile(FileName);

	map<string, unsigned> ChrNameToIndex;
	for (unsigned i = 0; i < SIZE(m_ChrNames); ++i)
		{
		const string &Name = m_ChrNames[i];
		if (ChrNameToIndex.find(Name) != ChrNameToIndex.end())
			Die("Duplicate chr name %s", Name.c_str());
		ChrNameToIndex[Name] = i;
		}

	ContigData Contig;
	GFFRecord2 Rec;
	ProgressStartFile(f, "Read annotations");
	unsigned Count = 0;
	set<string> UnknownFeatures;
	while (GetNextGFFRecord(f, Rec))
		{
		if (Count++%100 == 0)
			ProgressStepFile();
		map<string, unsigned>::const_iterator p = ChrNameToIndex.find(Rec.SeqName);
		if (p == ChrNameToIndex.end())
			Die("Unknown chr in annot file: %s", Rec.SeqName.c_str());
		unsigned ChrIndex = p->second;

		asserta(Rec.Start > 0);
		asserta(Rec.Start <= Rec.End);
		gpos_t Pos = Rec.Start - 1;
		gpos_t Length = Rec.End - Rec.Start + 1;

		unsigned IntraGeneIndex = UINT_MAX;
		bool Found = GFFGetIntAttr(Rec.Attrs, "gene_index", IntraGeneIndex);
		unsigned InterGeneIndex = IntraToInterGeneIndex(Rec.SeqName, IntraGeneIndex);
		if (Found)
			{
			GFFDeleteAttr(Rec.Attrs, "gene_index");
			GFFAppendIntAttr(Rec.Attrs, "gene_index", InterGeneIndex);
			}
		Contig.GeneIndex = InterGeneIndex;
		if (Rec.Feature == "CDS")
			{
			if (Contig.GeneIndex == UINT_MAX)
				{
				Rec.LogMe();
				Die("CDS record missing gene index");
				}
			Contig.Type = CT_CDS;
			}
		else if (Rec.Feature == "UTR")
			{
			if (Contig.GeneIndex == UINT_MAX)
				{
				Rec.LogMe();
				Die("UTR record missing gene index");
				}
			Contig.Type = CT_UTR;
			}
		else if (Rec.Feature == "NXE")
			{
			if (Contig.GeneIndex == UINT_MAX)
				{
				Rec.LogMe();
				Die("NXE record missing gene index");
				}
			Contig.Type = CT_NXE;
			}
		else if (Rec.Feature == "NGE")
			{
			if (Contig.GeneIndex != UINT_MAX)
				{
				Rec.LogMe();
				Die("NGE record has gene index");
				}
			Contig.Type = CT_NGE;
			}
		else if (Rec.Feature == "island" || Rec.Feature == "tandem")
			{
			if (Rec.Feature == "tandem" && !g_XferTandems)
				continue;
			IslandAndTandemRecs.push_back(Rec);
			continue;
			}
		else
			{
			UnknownFeatures.insert(Rec.Feature);
			continue;
			}

		ChrPosData ChrPos;
		ChrPos.ChrIndex = GetAncChrIndex(Rec.SeqName);
		ChrPos.Pos = Pos;

		if (m_AncAnnots.find(ChrPos) != m_AncAnnots.end())
			Die("Two annots at same position %s(%"PF")",
			  Rec.SeqName.c_str(), Pos);

		m_AncAnnots[ChrPos] = Rec;

		Contig.AncChrIndex = ChrIndex;
		Contig.EvChrIndex = ChrIndex;
		Contig.AncPos = Pos;
		Contig.EvPos = Pos;
		Contig.Length = Length;
		Contig.Plus = true;
		Contig.Copy = false;

		if (Contig.GeneIndex != UINT_MAX && Contig.GeneIndex >= m_NextNewGeneIndex)
			m_NextNewGeneIndex = Contig.GeneIndex + 1;

		m_Contigs.insert(Contig);
		}
	ProgressDone();

	if (!UnknownFeatures.empty())
		{
		string s;
		for (set<string>::const_iterator p = UnknownFeatures.begin();
		  p != UnknownFeatures.end(); ++p)
			s += " " + *p;
		Warning("Unknown features in GFF file: %s", s.c_str());
		}

	CloseStdioFile(f);
	g_ChrGixToGix.clear();
	}

void InterChrEvolver::ReadChrLengths(const char *FileName)
	{
	FILE *f = OpenStdioFile(FileName);
	GFFRecord2 Rec;
	set<string> ChrNameSet;
	vector<string> ChrNames;
	vector<gpos_t> ChrLengths;
	while (GetNextGFFRecord(f, Rec))
		{
		const string &Name = Rec.SeqName;
		if (ChrNameSet.find(Name) != ChrNameSet.end())
			Die("Duplicate chr name: %s", Name.c_str());

		asserta(Rec.End >= Rec.Start);
		gpos_t Length = Rec.End - Rec.Start + 1;

		ChrNames.push_back(Name);
		ChrLengths.push_back(Length);
		}
	CloseStdioFile(f);
	InitChrs(ChrNames, ChrLengths);
	}
