#include "ice.h"
#include "revfile.h"
#include "progress.h"

#define TRACE_ALNS	0

void RevComp(string &Seq);

void InterChrEvolver::GetChrInfo(vector<ChrInfo> &CIs) const
	{
	const unsigned ChrCount = GetChrCount();
	const vector<gpos_t> &ChrLengths = GetChrLengths();
	const vector<string> &ChrNames = GetChrNames();
	asserta(SIZE(ChrLengths) == ChrCount);
	asserta(SIZE(ChrNames) == ChrCount);

	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		{
		ChrInfo CI;
		CI.Create(ChrIndex, ChrNames[ChrIndex], ChrLengths[ChrIndex]);
		CIs.push_back(CI);
		}
	}

void InterChrEvolver::FromRev(const RevFile &Rev, unsigned GenIndex)
	{
	vector<gpos_t> ChrLengths;
	vector<string> ChrNames;

	const unsigned ChrCount = Rev.GetChrCount(GenIndex);
	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		{
		gpos_t ChrLength = Rev.GetChrLength(GenIndex, ChrIndex);
		const string &ChrName = Rev.GetChrName(GenIndex, ChrIndex);

		ChrLengths.push_back(ChrLength);
		ChrNames.push_back(ChrName);
		}

	InitChrs(ChrNames, ChrLengths);
	}

void InterChrEvolver::AlnToRev(const char *AlnFileName, const RevFile &AnSeqRev,
  unsigned AncGenIndex, const char *EvGenomeName, bool Dump)
	{
	vector<ChrInfo> CIs;
	GetChrInfo(CIs);

	GenInfo EvGenome;
	EvGenome.Create(EvGenomeName, CIs);

	const GenInfo &AncGenome = AnSeqRev.GetGenomeInfo(AncGenIndex);

	vector<GenInfo> Genomes;
	Genomes.push_back(AncGenome);
	Genomes.push_back(EvGenome);

	RevFile AlnRev;
	AlnRev.Create(AlnFileName, Genomes);

	ContigSet Contigs;
	GetMaximalContigs(Contigs);

	if (Dump)
		{
		Log("\n");
		Log("AlnToRev, maximal contigs:\n");
		LogContigs(Contigs);
		}

	unsigned ContigCount = SIZE(Contigs);
	unsigned Counter = 0;
	ProgressStart("Creating alignments");
	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		if (Counter++%10 == 0)
			ProgressStep(Counter, ContigCount);

		const ContigData &Contig = *p;

		unsigned AncChrId = AncGenome.GetChrId(Contig.AncChrIndex);
		AlnInfo AI;

		GCSeg AnSeg;
		AnSeg.GenomeIndex = 0;
		AnSeg.ChrId = AncChrId;
		AnSeg.lo = (coord_t) Contig.AncPos;
		AnSeg.hi = (coord_t) Contig.GetAncHi();
		AnSeg.Plus = true;
		AI.Segs.push_back(AnSeg);

		GCSeg EvSeg;
		EvSeg.GenomeIndex = 1;
		EvSeg.ChrId = Contig.EvChrIndex;
		EvSeg.lo = (coord_t) Contig.EvPos;
		EvSeg.hi = (coord_t) Contig.GetEvHi();
		EvSeg.Plus = Contig.Plus;
		AI.Segs.push_back(EvSeg);

		if (Contig.Copy)
			AI.SetAttr(ALN_ATTR_Paralog, 1);

		string Row;
		AnSeqRev.GetSubSeq3(AnSeg, Row);
		
		StdStrVec Rows;
		Rows.push_back(Row);
		Rows.push_back(Row);

#if	TRACE_ALNS
		unsigned AIIndex = AlnRev.AddAln(AI, Rows);
		unsigned GetUngappedSeqLength(const std::string &Seq);
		Contig.LogMe();
		Log(" AIIndex=%u, Row /seg lengths %u,%u %u,%u\n",
		  AIIndex,
		  GetUngappedSeqLength(Rows[0]),
		  AI.Segs[0].GetLength(),
		  GetUngappedSeqLength(Rows[1]),
		  AI.Segs[0].GetLength());
		Log("  ");
		AI.LogMe();
		Log("\n");
#else
		AlnRev.AddAln(AI, Rows);
#endif
		}
	ProgressDone();

	Progress("Closing aln file");
	AlnRev.Close();
	}

void InterChrEvolver::SeqToRev(const char *FileName, const RevFile &AncSeqRev,
  unsigned AncGenIndex, const char *EvGenomeName) const
	{
	vector<ChrInfo> CIs;
	GetChrInfo(CIs);

	GenInfo OutGenome;
	OutGenome.Create(EvGenomeName, CIs);

	RevFile EvSeqRev;
	vector<GenInfo> SeqOutGenomes;
	SeqOutGenomes.push_back(OutGenome);
	EvSeqRev.Create(FileName, SeqOutGenomes);

	const string AncGenName = AncSeqRev.GetGenomeName(AncGenIndex);

	ContigSet Contigs;
	GetMaximalContigs(Contigs);

	unsigned LastChrIndex = UINT_MAX;
	string ChrSeq;
	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (LastChrIndex != Contig.EvChrIndex && LastChrIndex != UINT_MAX)
			{
			asserta(LastChrIndex < SIZE(CIs));
			const string &ChrName = CIs[LastChrIndex].GetName();
			EvSeqRev.AddSeq(EvGenomeName, ChrName, ChrSeq.c_str(),
			  (coord_t) ChrSeq.size());
			ChrSeq.clear();
			}
		unsigned AncPos = (unsigned) Contig.AncPos;
		unsigned Length = (unsigned) Contig.Length;
		unsigned ChrId = AncSeqRev.GetChrId(AncGenIndex, Contig.AncChrIndex);
		string Seq;
		AncSeqRev.GetSubSeq2(AncGenIndex, ChrId, Seq, AncPos, Length);
		if (!Contig.Plus)
			RevComp(Seq);
		ChrSeq += Seq;
		LastChrIndex = Contig.EvChrIndex;
		}
	if (!ChrSeq.empty() && LastChrIndex != UINT_MAX)
		{
		const string &ChrName = CIs[LastChrIndex].GetName();
		EvSeqRev.AddSeq(EvGenomeName, ChrName, ChrSeq.c_str(),
		  (coord_t) ChrSeq.size());
		}
	EvSeqRev.Close();
	}
