#include "ice.h"
#include "gff2.h"

double AccelProb(double Prob, double Accel);
extern double g_AcceptProbs[256];
extern uint8 ProbToProbIndex(double Prob);

void InterChrEvolver::ToGFF(const char *FileName) const
	{
	FILE *f = CreateStdioFile(FileName);

	for (ContigSetCIter p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		ContigToGFF(f, Contig);
		}

	CloseStdioFile(f);
	}

const GFFRecord2 &InterChrEvolver::GetAncAnnot(unsigned ChrIndex, gpos_t Pos) const
	{
	ChrPosData ChrPos;
	ChrPos.ChrIndex = ChrIndex;
	ChrPos.Pos = Pos;
	map<ChrPosData, GFFRecord2>::const_iterator p = m_AncAnnots.find(ChrPos);
	if (p == m_AncAnnots.end())
		Die("GetAncAnnot(%u, %"PF"), not found", ChrIndex, Pos);
	return p->second;
	}

static void AccelProbsAttr(string &Attrs, double Accel)
	{
	string Probs;
	GFFGetRequiredAttr(Attrs, "probs", Probs);
	const unsigned L = SIZE(Probs);
	if (L%2 != 0)
		Die("Probs attr L=%u not multiple of 2", L);
	if (L == 0)
		Die("Empty probs attr");

	double SumOldProbs = 0;
	double SumNewProbs = 0;
	string NewProbs;
	for (unsigned i = 0; i < L; i += 2)
		{
		char c1 = Probs[i];
		char c2 = Probs[i+1];
		if (!isxdigit(c1) || !isxdigit(c2))
			{
			if (isprint(c1) && isprint(c2))
				Die("Invalid hex digits '%c%c' in probs", c1, c2);
			else
				Die("Invalid hex digits %02x%02x in probs", (byte) c1, (byte) c2);
			}
		static char s[3];
		s[0] = c1;
		s[1] = c2;
		char *NotUsed;
		unsigned ProbIndex = strtoul(s, &NotUsed, 16);
		
		double Prob = g_AcceptProbs[ProbIndex];
		double NewProb = AccelProb(Prob, Accel);

		SumOldProbs += Prob;
		SumNewProbs += NewProb;

		uint8 NewProbIndex = ProbToProbIndex(NewProb);
		if (NewProbIndex == (uint8) 0xff)
			NewProbIndex = (uint8) 0xfe;
		sprintf(s, "%02x", NewProbIndex);
		NewProbs += s;
		}

	GFFDeleteAttr(Attrs, "probs");
	GFFDeleteAttr(Attrs, "avg_prob");
	GFFDeleteAttr(Attrs, "gene_speed");
	GFFDeleteAttr(Attrs, "ce_speed");

	char Str[64];
	unsigned BaseCount = L/2;

	sprintf(Str, "%.4f", SumNewProbs/BaseCount);
	GFFAppendStrAttr(Attrs, "avg_prob", Str);

	sprintf(Str, "%.4f", Accel);
	GFFAppendStrAttr(Attrs, "accel", Str);

	sprintf(Str, "%.4f", SumOldProbs/BaseCount);
	GFFAppendStrAttr(Attrs, "old_avg_prob", Str);

	asserta(SIZE(NewProbs) == L);
	GFFAppendStrAttr(Attrs, "probs", NewProbs);
	}

void InterChrEvolver::ContigToGFF(FILE *f, const ContigData &Contig) const
	{
	asserta(Contig.EvChrIndex < SIZE(m_ChrNames));
	const string &ChrName = m_ChrNames[Contig.EvChrIndex];

	GFFRecord2 Rec;
	Rec.SeqName = ChrName;
	Rec.Source = "ice";
	Rec.Feature = ContigTypeToStr(Contig.Type);
	Rec.Start = Contig.EvPos + 1;
	Rec.End = Rec.Start + Contig.Length - 1;
	Rec.Strand = '.';
	Rec.Score = 0;

	if (Contig.IsCE())
		{
		const GFFRecord2 &AncRec = GetAncAnnot(Contig.AncChrIndex, Contig.AncPos);

		if (Contig.Plus)
			Rec.Strand = AncRec.Strand;
		else
			Rec.Strand = (AncRec.Strand == '+' ? '-' : '+');

		if (AncRec.Feature == "CDS")
			{
			asserta(Contig.Type == CT_CDS);
			Rec.Frame = AncRec.Frame;
			}
		else if (AncRec.Feature == "UTR")
			asserta(Contig.Type == CT_UTR);
		else if (AncRec.Feature == "NXE")
			asserta(Contig.Type == CT_NXE);
		else if (AncRec.Feature == "NGE")
			asserta(Contig.Type == CT_NGE);
		else
			asserta(false);

		asserta(AncRec.GetLength() == Contig.Length);
		Rec.Attrs = AncRec.Attrs;
		}

	asserta(Contig.AncChrIndex < SIZE(m_AncChrNames));
	const string &AncChrName = m_AncChrNames[Contig.AncChrIndex];
	GFFAppendStrAttr(Rec.Attrs, "anc_chr", AncChrName);

	if (Contig.IsGeneCE())
		{
		unsigned AttrGeneIndex = GFFGetRequiredIntAttr(Rec.Attrs, "gene_index");
		unsigned GeneIndex = Contig.GeneIndex;
		map<unsigned, double>::const_iterator p = m_GeneAccels.find(GeneIndex);
		if (p != m_GeneAccels.end())
			{
			double Accel = p->second;
			AccelProbsAttr(Rec.Attrs, Accel);
			}
		if (Contig.Copy)
			{
			asserta(AttrGeneIndex != Contig.GeneIndex);
			GFFDeleteAttr(Rec.Attrs, "gene_index");
			GFFPrependIntAttr(Rec.Attrs, "gene_index", Contig.GeneIndex);
			}
		else
			{
			if (AttrGeneIndex != Contig.GeneIndex)
				{
				Log("\n");
				Rec.LogMe(true);
				Contig.LogMe(true);
				Die("AttrGeneIndex=%u != Contig.GeneIndex=%u",
				  AttrGeneIndex, Contig.GeneIndex);
				}
			}
		}

	char Str[64];
	sprintf(Str, " anc_start %"PF"; copy \"%c\"; reloc_count %u;",
	  Contig.AncPos + 1,
	  tof(Contig.Copy),
	  Contig.RelocCount);
	Rec.Attrs += Str;

	Rec.Attrs += (Contig.Plus ? " Plus \"T\";" : " Plus \"F\";");

	Rec.Attrs += " Target \"" + AncChrName;
	sprintf(Str, " %"PF" %"PF"\";", Contig.AncPos + 1, Contig.GetAncHi() + 1);
	Rec.Attrs += Str;

	if (Contig.IsCE())
		{
		string Probs;
		GFFGetRequiredAttr(Rec.Attrs, "probs", Probs);
		asserta(SIZE(Probs) == Contig.Length*2);
		}

	WriteGFFRec(f, Rec);
	}

void InterChrEvolver::AnnotsToGFF(const char *FileName) const
	{
	FILE *f = CreateStdioFile(FileName);

	for (ContigSetCIter p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (Contig.IsCE())
			ContigToGFF(f, Contig);
		}

	CloseStdioFile(f);
	}

void InterChrEvolver::WriteChrNames(const char *OutChrNamesFileName) const
	{
	FILE *f = CreateStdioFile(OutChrNamesFileName);

	for (vector<string>::const_iterator p = m_ChrNames.begin();
	  p != m_ChrNames.end(); ++p)
		fprintf(f, "%s\n", (*p).c_str());

	CloseStdioFile(f);
	}

