#include "utils.h"
#include "ice.h"
#include "progress.h"

void InterChrEvolver::PrintAndDie(const char *Format, ...) const
	{
	string Msg;
	va_list ArgList;
	va_start(ArgList, Format);
	VStrPrintf(Msg, Format, ArgList);
	va_end(ArgList);
	if (m_CurrentEvent != EVENT_TYPE_Invalid)
		Log("Current event %u %s pos %"PF"\n",
		  m_EventIndex,
		  EventTypeToStr(m_CurrentEvent),
		  m_CurrentEventPos);
	if (m_LastAcceptedEvent != EVENT_TYPE_Invalid)
		Log("Last accepted event %u %s pos %"PF"\n",
		  m_LastAcceptedEventIndex,
		  EventTypeToStr(m_LastAcceptedEvent),
		  m_LastAcceptedEventPos);
	Log("\nInterChrEvolver::Die\n");
	if (!m_ValidateWhy.empty())
		Log("Validate(%s)\n", m_ValidateWhy.c_str());
	::Die("%s", Msg.c_str());
	}

void InterChrEvolver::Validate(const string &Why) const
	{
	m_ValidateWhy = Why;
	if (m_EventIndex != 0 && m_EventIndex != UINT_MAX)
		Progress("Validate [%u]", m_EventIndex);
	else
		Progress("Validate", m_EventIndex);
	AssertContiguous(m_Contigs);
	ValidateIntact();
	ValidateChrLengths();
	ValidateGenes();
	Progress("Validate %s OK", Why.c_str());
	m_ValidateWhy.clear();
	}

void InterChrEvolver::AssertContiguous(const set<ContigData> &Contigs) const
	{
	const ContigData *LastContig = 0;
	for (ContigSetCIter p = Contigs.begin(); p != Contigs.end(); ++p)
		{
		const ContigData *Contig = &(*p);
		if (LastContig != 0 && LastContig->EvChrIndex == Contig->EvChrIndex)
			{
			if (LastContig->EvPos + LastContig->Length != Contig->EvPos)
				{
				LogContigs(Contigs);
				Log("Last contig: ");
				LastContig->LogMe(true);
				Log("This contig: ");
				Contig->LogMe(true);
				Die("Not contiguous");
				}
			}
		LastContig = Contig;
		}
	}

void InterChrEvolver::ValidateChrLengths() const
	{
	const unsigned ChrCount = GetChrCount();
	asserta(SIZE(m_ChrLengths) == ChrCount);

	vector<gpos_t> ChrLengths(ChrCount, 0);
	unsigned LastChrIndex = UINT_MAX;
	for (ContigSetCIter p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		if (Contig.EvChrIndex != LastChrIndex)
			{
			if (Contig.EvPos != 0)
				{
				LogMe();
				Die("First contig in chr %u has EvPos=%"PF,
				  Contig.EvChrIndex, Contig.EvPos);
				}
			}
		asserta(Contig.EvChrIndex < ChrCount);
		ChrLengths[Contig.EvChrIndex] += Contig.Length;
		LastChrIndex = Contig.EvChrIndex;
		}

	for (unsigned ChrIndex = 0; ChrIndex < ChrCount; ++ChrIndex)
		if (ChrLengths[ChrIndex] != m_ChrLengths[ChrIndex])
			{
			LogMe();
			Die("Chr %u length mismatch, measured %"PF" cached %"PF,
			  ChrIndex,
			  ChrLengths[ChrIndex],
			  m_ChrLengths[ChrIndex]);
			}
	}

// Validates each gene has a contiguous range
// and at least one CDS.
void InterChrEvolver::ValidateGenes() const
	{
	map<unsigned, unsigned> GeneIndexToCDSCount;
	map<unsigned, gpos_t> GeneIndexToHiContigPos;
	map<unsigned, unsigned> GeneIndexToChrIndex;
	unsigned PrevGeneIndex = UINT_MAX;
	unsigned PrevChrIndex = UINT_MAX;
	for (ContigSet::const_iterator p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		unsigned GeneIndex = Contig.GeneIndex;
		if (Contig.IsGeneCE() && GeneIndex == UINT_MAX)
			{
			Contig.LogMe(true);
			Die("Contig is gene CE but no gene index");
			}

		if (GeneIndex != UINT_MAX)
			{
			if (GeneIndexToCDSCount.find(GeneIndex) == GeneIndexToCDSCount.end())
				GeneIndexToCDSCount[GeneIndex] = 0;

			if (Contig.Type == CT_CDS)
				++(GeneIndexToCDSCount[GeneIndex]);
			}

		unsigned ChrIndex = Contig.EvChrIndex;
		if (GeneIndex != UINT_MAX && GeneIndex == PrevGeneIndex &&
		  ChrIndex != PrevChrIndex)
			{
			Log("\n");
			Die("Gene %u split over chrs %u, %u",
			  GeneIndex, PrevChrIndex, ChrIndex);
			}

		if (PrevGeneIndex != UINT_MAX && PrevGeneIndex != GeneIndex)
			{
			if (GeneIndexToHiContigPos.find(PrevGeneIndex) !=
			  GeneIndexToHiContigPos.end())
				{
				Log("\n");
				Log("HiPos = %"PF"\n", GeneIndexToHiContigPos[PrevGeneIndex]);
				Contig.LogMe(true);
				Die("Gene %u not contiguous", PrevGeneIndex);
				}
			GeneIndexToHiContigPos[PrevGeneIndex] = Contig.EvPos - 1;
			}
		PrevGeneIndex = GeneIndex;
		PrevChrIndex = ChrIndex;
		}
	
	for (map<unsigned, unsigned>::const_iterator p = GeneIndexToCDSCount.begin();
	  p != GeneIndexToCDSCount.end(); ++p)
	  	{
		unsigned GeneIndex = p->first;
		unsigned CDSCount = p->second;
		if (CDSCount == 0)
			Die("Gene %u has not CDS", GeneIndex);
		}
	}

// Validate ancestral genome covered exactly once by
// non-copy contigs.
void InterChrEvolver::ValidateIntact() const
	{
	const unsigned AncChrCount = SIZE(m_AncChrLengths);
	vector<gpos_t> AncChrLengths(AncChrCount);	

	vector<map<gpos_t, const ContigData *> > AncOrderVec(AncChrCount);

	for (ContigSetCIter p = m_Contigs.begin(); p != m_Contigs.end(); ++p)
		{
		const ContigData &Contig = *p;
		asserta(Contig.AncChrIndex < AncChrCount);
		asserta(Contig.GetAncHi() < m_AncChrLengths[Contig.AncChrIndex]);
		if (Contig.Copy)
			continue;
		AncChrLengths[Contig.AncChrIndex] += Contig.Length;
		map<gpos_t, const ContigData *> &AncOrder = AncOrderVec[Contig.AncChrIndex];
		AncOrder[Contig.AncPos] = &Contig;
		}

	gpos_t OldAncLength = 0;
	gpos_t NewAncLength = 0;
	for (unsigned ChrIndex = 0; ChrIndex < AncChrCount; ++ChrIndex)
		{
		OldAncLength += m_AncChrLengths[ChrIndex];
		NewAncLength += AncChrLengths[ChrIndex];
		}
	if (OldAncLength != NewAncLength)
		Die("Anc length changed, should be %"PF", is %"PF", diff %+d",
		  OldAncLength, NewAncLength, (int) NewAncLength - (int) OldAncLength);

	for (unsigned ChrIndex = 0; ChrIndex < AncChrCount; ++ChrIndex)
		{
		if (AncChrLengths[ChrIndex] != m_AncChrLengths[ChrIndex])
			Die("Anc chr length mismatch, chr %u should be %"PF" is %"PF,
			  ChrIndex, m_AncChrLengths[ChrIndex], AncChrLengths[ChrIndex]);
		}

	for (unsigned ChrIndex = 0; ChrIndex < AncChrCount; ++ChrIndex)
		{
		const map<gpos_t, const ContigData *> &AncOrder = AncOrderVec[ChrIndex];
		gpos_t CurrPos = 0;
		const ContigData *LastContig = 0;
		for (map<gpos_t, const ContigData *>::const_iterator p = AncOrder.begin();
		  p != AncOrder.end(); ++p)
			{
			gpos_t Pos = p->first;
			const ContigData *Contig = p->second;
			if (LastContig != 0)
				asserta(Contig->AncChrIndex == LastContig->AncChrIndex);
			gpos_t Length = Contig->Length;
			if (CurrPos != Pos)
				{
				LogMe();
				Log("\n");
				Log("Prev contig: ");
				if (LastContig == 0)
					Log("NULL\n");
				else
					LastContig->LogMe(true);
				Log("This contig: ");
				Contig->LogMe(true);
				if (Pos > CurrPos)
					Die("Ancestral chromosome %u not contiguous, missing %"PF" - %"PF" (L=%"PF")",
					  ChrIndex, CurrPos, Pos-1, Pos - CurrPos);
				else
					Die("Ancestral chromosome %u not contiguous, duplicate %"PF" - %"PF" (L=%"PF")",
					  ChrIndex, Pos, CurrPos-1, CurrPos - Pos);
				}
			LastContig = Contig;
			CurrPos += Length;
			}
		}
	}
