#include "ice.h"
#include "getopts.h"
#include "revfile.h"
#include "progress.h"
#include "myrand.h"

extern unsigned g_ValEvents;
extern unsigned g_StartValEvents;
extern set<unsigned> g_AnGeneIndexes;
extern set<unsigned> g_EvGeneIndexes;
extern FILE *g_fStats;

bool g_XferTandems = false;

void SetModel(const char *FileName);
void LogAverageEventRates();
void LogERDetails();
void LogRateSummary(bool Cmt);
void LogEventStats(double BranchLength, const vector<double> &InitialRates,
  const vector<double> &FinalRates);
double Pct(double, double);
void XferAnnotsVec(const vector<GFFRecord2> &InRecs, const RevFile &AlnRev,
  unsigned FromGenomeIndex, unsigned ToGenomeIndex, FILE *f);

void InterChr()
	{
	const char *AnSeqFileName = RequiredValueOpt("interchr");
	const char *ICEFileName = ValueOpt("ice");
	const char *InAnnotsFileName = ValueOpt("inannots");
	const char *AlnFileName = ValueOpt("aln");
	const char *OutSeqFileName = ValueOpt("outseq");
	const char *OutGFFFileName = ValueOpt("outannots");
	const char *OutChrNamesFileName = ValueOpt("outchrnames");
	const char *ModelFile = ValueOpt("model");
	bool LogEvents = FlagOpt("logevents");
	bool Dump = FlagOpt("dump");
	bool ValInit = !FlagOpt("novalinit");
	bool ValFinal = !FlagOpt("novalfinal");
	unsigned GenIndex = 0;
	IntOpt("genix", &GenIndex);
	const char *OutGenomeName = ValueOpt("outgenome");
	if (OutGenomeName == 0)
		OutGenomeName = "evolved";
	const char *OutContigsFileName = ValueOpt("outcontigs");
	const char *StatsFileName = ValueOpt("statsfile");
	if (StatsFileName != 0)
		g_fStats = CreateStdioFile(StatsFileName);
	const char *SaveFinalFileName = ValueOpt("savefinal");
	g_XferTandems = FlagOpt("xfertandems");

	g_ValEvents = 0;
	g_StartValEvents = 0;
	IntOpt("valevents", &g_ValEvents);
	IntOpt("startval", &g_StartValEvents);

	if (ValueOpt("seed") != 0 || ICEFileName == 0)
		InitRand();

	SetModel(ModelFile);

	if (ValueOpt("targetgenecount"))
		{
		void DeleteRandGenes(const char *GeneFileName, const char *OutFileName, int N);
		const char *strCount = RequiredValueOpt("targetgenecount");
		int N = atoi(strCount);
		
		const char *AnnotsMinusFileName = RequiredValueOpt("annotsminus");
		DeleteRandGenes(InAnnotsFileName, AnnotsMinusFileName, N);
		InAnnotsFileName = AnnotsMinusFileName;
		}

	vector<double> InitialRates;
	vector<double> FinalRates;

	RevFile AnSeqRev;
	AnSeqRev.Open(AnSeqFileName);

	vector<GFFRecord2> IslandAndTandemRecs;
	InterChrEvolver ICE;
	if (ICEFileName != 0)
		ICE.FromFile(ICEFileName);
	else
		{
		ICE.FromRev(AnSeqRev, GenIndex);
		ICE.ReadAnnots(InAnnotsFileName, IslandAndTandemRecs);
		ICE.InitNeutrals();
		ICE.ComputeEventRates();

		if (ValueOpt("branchlength"))
			{
			if (ValueOpt("events"))
				Die("Cannot specify both -branchlength and -events");
			ICE.m_Duration = RequiredFloatOpt("branchlength");
			ICE.SetEventCount();
			}
		else if (ValueOpt("events"))
			{
			if (ValueOpt("branchlength"))
				Die("Cannot specify both -branchlength and -events");
			unsigned EventCount = RequiredIntOpt("events");
			ICE.SetEventCount2(EventCount);
			}
		else
			Die("Must specify either -branchlength or -events");
		}

	if (ValInit)
		ICE.Validate("valinit");

	ICE.SetLogEvents(LogEvents);

	if (Dump)
		{
		Log("Initial state:\n");
		ICE.LogMe();
		}

	ICE.GetEventRates(InitialRates);

	Progress("Get initial gene indexes");
	ICE.GetGeneIndexes(g_AnGeneIndexes);

	gpos_t AncestralLength = ICE.GetLength();
////////////////////
	ICE.Run();
////////////////////
	gpos_t EvolvedLength = ICE.GetLength();

	if (SaveFinalFileName != 0)
		ICE.ToFile(SaveFinalFileName);

	if (ValFinal)
		ICE.Validate("valfinal");

	Progress("Merge neutrals");
	ICE.MergeContiguousNeutrals();

	Progress("Get final gene indexes");
	ICE.GetGeneIndexes(g_EvGeneIndexes);

	if (OutSeqFileName != 0)
		{
		Progress("Creating evolved sequence");
		ICE.SeqToRev(OutSeqFileName, AnSeqRev, GenIndex, OutGenomeName);
		}

	if (AlnFileName != 0)
		ICE.AlnToRev(AlnFileName, AnSeqRev, GenIndex, OutGenomeName, Dump);

	Progress("Closing anc seq rev file");
	AnSeqRev.Close();

	if (OutGFFFileName != 0)
		{
		Progress("Creating annotations");
		ICE.AnnotsToGFF(OutGFFFileName);
		}

	Progress("Transferring island and tandem annotations");

	if (!IslandAndTandemRecs.empty() && AlnFileName != 0)
		{
		FILE *fOut = fopen(OutGFFFileName, "a");
		if (fOut == 0)
			Die("Cannot open %s for appending", OutGFFFileName);

		RevFile AlnRev;
		AlnRev.Open(AlnFileName);
		XferAnnotsVec(IslandAndTandemRecs, AlnRev, 0, 1, fOut);
		AlnRev.Close();
		fclose(fOut);
		}

	if (OutContigsFileName != 0)
		ICE.ToGFF(OutContigsFileName);

	if (OutChrNamesFileName != 0)
		ICE.WriteChrNames(OutChrNamesFileName);

	if (Dump)
		ICE.LogMe();

	ICE.GetEventRates(FinalRates);
	LogRateSummary(false);
	LogEventStats(ICE.m_Duration, InitialRates, FinalRates);
	LogERDetails();

	Log("\n");
	Log("Ancestral length    %10"PF"\n", AncestralLength);
	Log("Evolved length      %10"PF"\n", EvolvedLength);
	Log("Length change       %+10d\n", (int) EvolvedLength - (int) AncestralLength);
	Log("\n");
	}
