#ifndef KFILE_H
#define KFILE_H

#include <sys/mman.h>
#include <algorithm>
#include <vector>
#include <functional>
#include "geninfo.h"
#include "utils.h"
#include "types.h"

template <typename Entry>
class KFile;

/* Specific KFile instances */

typedef signed int kmer;

struct KDecomp
{
	kmer k;
	uint32 idx;
};

class KDecompCompareByIndex
{
	public:
	KDecompCompareByIndex() { }

	bool operator()(const KDecomp &KA, const KDecomp &KB)
	{
		return KA.idx < KB.idx;
	}
};

class KDecompCompareByKmer
{
	public:
	KDecompCompareByKmer() { }

	bool operator()(const KDecomp &KA, const KDecomp &KB)
	{
		return abs(KA.k) < abs(KB.k);
	}
};

typedef KFile<KDecomp> KFileDecomp;

class KPair
{
	public:

	kmer k[2];
	uint32 score;

	bool EqualAbsTo(KPair &rhs)
	{
		return ((abs(k[0]) == abs(rhs.k[0])) && (abs(k[1]) == abs(rhs.k[1])));
	}
};

class KPairCompare
{
	public:

	KPairCompare(unsigned GenomeIndex) : m_GenomeIndex(GenomeIndex) { }
	bool operator()(const KPair &KPA, const KPair &KPB) const
	{
		if (abs(KPA.k[m_GenomeIndex]) != abs(KPB.k[m_GenomeIndex]))
			return (abs(KPA.k[m_GenomeIndex]) < abs(KPB.k[m_GenomeIndex]));
		else if (abs(KPA.k[1 - m_GenomeIndex]) != abs(KPB.k[1 - m_GenomeIndex]))
			return (abs(KPA.k[1 - m_GenomeIndex]) < abs(KPB.k[1 - m_GenomeIndex]));
		else
			return (KPA.score < KPB.score);
	}

	private:

	unsigned m_GenomeIndex;
};

class KPairPartialCompare
{
	public:

	KPairPartialCompare(unsigned GenomeIndex) : m_GenomeIndex(GenomeIndex) { }
	bool operator()(const KPair &KPA, const KPair &KPB)
	{
		return (abs(KPA.k[m_GenomeIndex]) < abs(KPB.k[m_GenomeIndex]));
	}

	private:

	unsigned m_GenomeIndex;
};

typedef KFile<KPair> KFilePair;

#define KFileCheckState(X) if (m_State != (X)) Die("KFile: Not in \"" #X "\" state (%u)", m_State)

template <typename Entry>
class KFile
{
	public:

	typedef std::vector<Entry> EntryVec;
	
	KFile();
	~KFile();

	/* Read an existing file */
	void Open(const std::string &FileName);
	/* Create a new file */
	void Create(const std::string &FileName, const GenInfoVec &Genomes);
	/* Close a file accessed with Open() or Create() */
	void Close();
	/* Mmap-Sort an existing file */
	template <typename EntryCmp>
	void Sort(const std::string &FileName, EntryCmp ECmp = EntryCmp());

	const GenInfoVec &GetGenomes() const;
	unsigned GetGenomeCount() const;
	const GenInfo &GetGenomeInfo(unsigned GenomeIndex) const;
	/*
	// Not yet implemented
	const std::string &GetGenomeName(unsigned GenomeIndex) const;
	unsigned GetGenomeIndex(const std::string &GenomeName) const;
	unsigned GetChrCount(unsigned GenomeIndex) const;
	const ChrInfo &GetChrInfo(unsigned GenomeIndex, unsigned ChrId) const;
	const ChrInfo *GetChrInfo(unsigned GenomeIndex, const std::string &ChrName) const;
	const std::string &GetChrName(unsigned GenomeIndex, unsigned ChrId) const;
	unsigned GetChrId(unsigned GenomeIndex, const std::string &ChrName) const;
	unsigned GetChrId(unsigned GenomeIndex, unsigned ChrIndex) const;
	unsigned GetChrIndex(unsigned GenomeIndex, unsigned ChrId) const;
	coord_t GetChrLength(unsigned GenomeIndex, unsigned ChrId) const;
	*/

	/* Methods for reading a file */
	unsigned GetEntryCount() const;
	void GetNextEntry(Entry &E) const;
	void GetEntry(Entry &E, unsigned EntryIndex) const;
	void GetAllEntries(EntryVec &EVec) const;
	void Rewind() const;
	const Entry *MemoryMapRO();
	void MemoryUnmap();

	/* Methods for writing a file */
	void AddEntry(const Entry &E);
	void AddEntries(const EntryVec &EVec);

	protected:

	enum KFileState {Idle = 1, Reading, Writing};
	KFileState m_State;
	GenInfoVec m_Genomes;
	off_t m_EntryCountPos;
	off_t m_EntryArrayPos;
	FILE *m_fp;
	uint32 m_Magic;
	uint32 m_EntryCount;
	void *m_MapAddr;
	size_t m_MapSize;

	void ReadHeaderGenomes();
	void WriteHeaderGenomes() const;
	void ReadEntryHeader();
	void WriteEntryHeader() const;
	void Init();
};


template <typename Entry>
KFile<Entry>::KFile()
{
	m_Magic = 0x1f2e3d4a;
	m_State = Idle;
	m_EntryCount = 0;
	m_MapAddr = NULL;
}

template <typename Entry>
KFile<Entry>::~KFile()
{
	if (m_MapAddr != NULL)
		MemoryUnmap();

	if (m_State != Idle)
		Close();
}

template <typename Entry>
void KFile<Entry>::Open(const std::string &FileName)
{
	KFileCheckState(Idle);
	m_fp = OpenStdioFile(FileName.c_str());
	m_State = Reading;
	ReadHeaderGenomes();
	m_EntryCountPos = GetStdioFilePos(m_fp);
	ReadEntryHeader();
	m_EntryArrayPos = GetStdioFilePos(m_fp);
}

template <typename Entry>
void KFile<Entry>::Create(const std::string &FileName, const GenInfoVec &Genomes)
{
	KFileCheckState(Idle);
	/*
	if (Genomes.size() == 0)
		Die("KFile: Invalid genome count");
	*/
	m_fp = CreateStdioFile(FileName.c_str());
	m_State = Writing;
	m_Genomes = Genomes;
	WriteHeaderGenomes();
	m_EntryCountPos = GetStdioFilePos(m_fp);
	m_EntryCount = (unsigned)-1; //False entry count that will be corrected when closing file
	WriteEntryHeader();
	m_EntryCount = 0;
	m_EntryArrayPos = GetStdioFilePos(m_fp);
}

template <typename Entry>
void KFile<Entry>::Close()
{
	if (m_State == Reading)
	{
		CloseStdioFile(m_fp);
	}
	else if (m_State == Writing)
	{
		WriteEntryHeader();
		WriteStdioFile(m_fp, 0, &m_Magic, sizeof(m_Magic));
		CloseStdioFile(m_fp);
	}
	else
		Die("KFile: Invalid close request");

	Init();
}

template <typename Entry>
template <typename EntryCmp>
void KFile<Entry>::Sort(const std::string &FileName, EntryCmp ECmp)
{
	KFileCheckState(Idle);
	m_fp = fopen(FileName.c_str(), "r+");
	if (m_fp == NULL)
		Die("KFile: fopen failed");
	m_State = Reading;
	ReadHeaderGenomes();
	m_EntryCountPos = GetStdioFilePos(m_fp);
	ReadEntryHeader();
	m_EntryArrayPos = GetStdioFilePos(m_fp);

	off64_t FileSize = GetStdioFileSize(m_fp);

	if (FileSize != (off64_t)(size_t)FileSize)
	{
		Die("KFile: This executable does not support >4GB mapped files");
	}

	char *FileMem = (char *)mmap(NULL, FileSize, PROT_READ | PROT_WRITE, MAP_SHARED, fileno(m_fp), 0);
	if (FileMem == MAP_FAILED)
		Die("KFile: Memory mapping failure");

	Entry *EntryPtr = (Entry *)(FileMem + m_EntryArrayPos);
	std::sort(EntryPtr, EntryPtr + m_EntryCount, ECmp);
	int r = msync(FileMem, FileSize, MS_SYNC);
	if (r != 0)
		Die("msync() failed: errno=%d", errno);

	if (munmap(FileMem, FileSize) == -1)
		Die("KFile: Memory unmapping failure");

	Close();
}


template <typename Entry>
const GenInfoVec &KFile<Entry>::GetGenomes() const
{
	return m_Genomes;
}

template <typename Entry>
unsigned KFile<Entry>::GetGenomeCount() const
{
	return m_Genomes.size();
}

template <typename Entry>
const GenInfo &KFile<Entry>::GetGenomeInfo(unsigned GenomeIndex) const
{
	if (GenomeIndex >= m_Genomes.size())
		Die("Invalid genome index");
	return m_Genomes[GenomeIndex];
}

template <typename Entry>
unsigned KFile<Entry>::GetEntryCount() const
{
	return m_EntryCount;
}

template <typename Entry>
void KFile<Entry>::GetNextEntry(Entry &E) const
{
	/* No EOF checking -- will die anyway */
	KFileCheckState(Reading);
	ReadStdioFile(m_fp, &E, sizeof(Entry));
}

template <typename Entry>
void KFile<Entry>::GetEntry(Entry &E, unsigned EntryIndex) const
{
	KFileCheckState(Reading);
	if (EntryIndex >= m_EntryCount)
		Die("KFile: Entry index (%u) out of bounds (%u)", EntryIndex, m_EntryCount);

	ReadStdioFile(m_fp, m_EntryArrayPos + EntryIndex * sizeof(Entry), &E, sizeof(Entry));
}

template <typename Entry>
void KFile<Entry>::GetAllEntries(EntryVec &EVec) const
{
	KFileCheckState(Reading);
	EVec.clear();
	EVec.resize(m_EntryCount);
	SeekStdioFile(m_fp, m_EntryArrayPos);
	for (typename EntryVec::iterator i = EVec.begin(); i != EVec.end(); ++i)
		ReadStdioFile(m_fp, &*i, sizeof(Entry));
}

template <typename Entry>
void KFile<Entry>::Rewind() const
{
	KFileCheckState(Reading);
	SeekStdioFile(m_fp, m_EntryArrayPos);
}

template <typename Entry>
const Entry * KFile<Entry>::MemoryMapRO()
{
	KFileCheckState(Reading);

	if (m_MapAddr != NULL)
		Die("Double memory map request");

	m_MapSize = GetStdioFileSize(m_fp);
	m_MapAddr = mmap(NULL, m_MapSize, PROT_READ, MAP_SHARED, fileno(m_fp), 0);
	if (m_MapAddr == MAP_FAILED)
		Die("Memory mapping failed, errno=%d %s", errno, strerror(errno));

  return (const Entry *)((char *)m_MapAddr + m_EntryArrayPos);
}

template <typename Entry>
void KFile<Entry>::MemoryUnmap()
{
	if (m_MapAddr == NULL)
		Die("MemoryUnmap called with no existing memory mapping");

	if (munmap(m_MapAddr, m_MapSize) != 0)
		Die("Memory unmapping failed, errno=%d %s", errno, strerror(errno));

	m_MapAddr = NULL;
}

template <typename Entry>
void KFile<Entry>::AddEntry(const Entry &E)
{
	KFileCheckState(Writing);
	++m_EntryCount;
	WriteStdioFile(m_fp, &E, sizeof(Entry));
}

template <typename Entry>
void KFile<Entry>::AddEntries(const EntryVec &EVec)
{
	KFileCheckState(Writing);
	for (typename EntryVec::const_iterator i = EVec.begin(); i != EVec.end(); ++i)
	{
		++m_EntryCount;
		WriteStdioFile(m_fp, &*i, sizeof(Entry));
	}
}

template <typename Entry>
void KFile<Entry>::ReadHeaderGenomes()
{
	uint32 Magic;
	ReadStdioFile(m_fp, &Magic, sizeof(Magic));
	if (Magic != m_Magic)
		Die("KFile: Invalid Magic");

	uint32 GenomeCount;
	ReadStdioFile(m_fp, &GenomeCount, sizeof(GenomeCount));
	/*
	if (GenomeCount == 0)
		Die("KFile: Invalid genome count");
	*/

	m_Genomes.resize(GenomeCount);
	for (unsigned GenomeIndex = 0; GenomeIndex != GenomeCount; ++GenomeIndex)
		m_Genomes[GenomeIndex].FromFile(m_fp);
}

template <typename Entry>
void KFile<Entry>::WriteHeaderGenomes() const
{
	/* False magic will be corrected when closing the file */
	uint32 FalseMagic = 0x01020304;
	WriteStdioFile(m_fp, &FalseMagic, sizeof(FalseMagic));

	uint32 GenomeCount = GetGenomeCount();
	WriteStdioFile(m_fp, &GenomeCount, sizeof(GenomeCount));

	for (unsigned GenomeIndex = 0; GenomeIndex != GenomeCount; ++GenomeIndex)
		m_Genomes[GenomeIndex].ToFile(m_fp);
}

template <typename Entry>
void KFile<Entry>::ReadEntryHeader()
{
	ReadStdioFile(m_fp, &m_EntryCount, sizeof(m_EntryCount));
	if (m_EntryCount == (uint32)-1)
		Die("KFile: Invalid pair count");
}

template <typename Entry>
void KFile<Entry>::WriteEntryHeader() const
{
	WriteStdioFile(m_fp, m_EntryCountPos, &m_EntryCount, sizeof(m_EntryCount));
}

template <typename Entry>
void KFile<Entry>::Init()
{
	m_Genomes.clear();
	m_EntryCount = 0;
	m_State = Idle;
	m_fp = NULL;
}

#endif
