#ifndef KVFILE_H
#define KVFILE_H

#include <sys/mman.h>
#include <stdio.h>
#include <algorithm>
#include <vector>
#include <functional>
#include "geninfo.h"
#include "utils.h"
#include "types.h"

/* WARNING: This file uses off_t rather than off64_t */

template <typename Entry>
class defEntryDiskFunc
{
	public:
	
	void operator()(const byte *Buffer, unsigned BufferSize, Entry &E) const
	{
		E(Buffer, BufferSize);
	}

	unsigned operator()(byte **BufferPtr, const Entry &E)
	{
		return E(BufferPtr);
	}
};

template <typename Entry, typename EntryDiskFunc = defEntryDiskFunc<Entry> >
class KVFile;

/* Specific KVFile instances */

typedef signed int kmer;

struct GKmer
{
	public:

	unsigned G;
	kmer k;

	GKmer() {}
	GKmer(unsigned G_, kmer k_) : G(G_), k(k_) {}
	bool IsAdjacentTo(const GKmer &rhs) const
	{
		return ((G == rhs.G) && ((k == rhs.k) || (k + 1 == rhs.k) || (k - 1 == rhs.k)));
	}
};

class AlnNode
{
	public:

	std::vector<GKmer> GKmers;

	unsigned GetGKmerCount() const
	{
		return GKmers.size();
	}

	const GKmer &GetGKmer(unsigned GKmerIndex) const
	{
		if (GKmerIndex >= GKmers.size())
			Die("Invalid GKmer Index (%u)", GKmerIndex);
		
		return GKmers[GKmerIndex];
	}

	kmer GetKmerByGenome(unsigned GenomeIndex) const
	{
		for (std::vector<GKmer>::const_iterator i = GKmers.begin(); i != GKmers.end(); ++i)
		{
			if (i->G == GenomeIndex)
				return i->k;
		}
		Die("Invalid GenomeIndex in GetKmerByGenome");
		return kmer(0);
	}

	bool GetKmerByGenome(unsigned GenomeIndex, kmer &k) const
	{
		for (std::vector<GKmer>::const_iterator i = GKmers.begin(); i != GKmers.end(); ++i)
		{
			if (i->G == GenomeIndex)
			{
				k = i->k;
				return true;
			}
		}
		return false;
	}

	bool HasGenome(unsigned GenomeIndex) const
	{
		for (std::vector<GKmer>::const_iterator i = GKmers.begin(); i != GKmers.end(); ++i)
		{
			if (i->G == GenomeIndex)
				return true;
		}
		return false;
	}

	void SetGKmers(const std::vector<GKmer> &GKs)
	{
		GKmers = GKs;
	}

	void operator()(const byte *Buffer, unsigned BufferSize)
	{
		if ((BufferSize % sizeof(GKmer)) != 0)
			Die("Invalid Buffer Size");

		unsigned GKmerCount = BufferSize / sizeof(GKmer);
		GKmers.clear();
		GKmers.resize(GKmerCount);

		const GKmer *GKmerPtr = (const GKmer *)Buffer;
		for (unsigned i = 0; i != GKmerCount; ++i)
		{
			GKmers[i] = GKmerPtr[i];
		}
	}

	unsigned operator()(byte **BufferPtr) const
	{
		if (BufferPtr == NULL)
			Die("Invalid BufferPtr");

		unsigned BufferSize = GKmers.size() * sizeof(GKmer);
		*BufferPtr = new byte[BufferSize];

		GKmer *GKmerPtr = (GKmer *)(*BufferPtr);
		for (unsigned i = 0; i != GKmers.size(); ++i)
		{
			GKmerPtr[i] = GKmers[i];
		}

		return BufferSize;
	}

	//Assumes sorted GKmers
	bool operator<(const AlnNode &rhs) const
	{
		std::vector<GKmer>::const_iterator lhs_iter = GKmers.begin(), rhs_iter = rhs.GKmers.begin();

		while ((lhs_iter != GKmers.end()) && (rhs_iter != rhs.GKmers.end()))
		{
			if (lhs_iter->G != rhs_iter->G)
				return (lhs_iter->G < rhs_iter->G);

			if (lhs_iter->k != rhs_iter->k)
				return (lhs_iter->k < rhs_iter->k);

			++lhs_iter;
			++rhs_iter;
		}

		return (lhs_iter != GKmers.end());
	}

	bool IsAdjacentTo(const AlnNode &rhs) const
	{
		std::vector<GKmer>::const_iterator lhs_iter = GKmers.begin(), rhs_iter = rhs.GKmers.begin();

		while ((lhs_iter != GKmers.end()) && (rhs_iter != rhs.GKmers.end()))
		{
			if (lhs_iter->G != rhs_iter->G)
				return false;

			if (!((lhs_iter->k + 1 == rhs_iter->k) || (lhs_iter->k - 1 == rhs_iter->k) || (lhs_iter->k == rhs_iter->k)))
				return false;

			++lhs_iter;
			++rhs_iter;
		}
		return ((lhs_iter == GKmers.end()) && (rhs_iter == rhs.GKmers.end()));
	}

};

typedef KVFile<AlnNode> AlnNodeFile;

#define KVFileCheckState(X) if (m_State != (X)) Die("KVFile: Not in \"" #X "\" state (%u)", m_State)

template <typename Entry, typename EntryDiskFunc>
class KVFile
{
	public:

	typedef std::vector<Entry> EntryVec;
	
	KVFile();
	~KVFile();

	/* Read an existing file */
	void Open(const std::string &FileName);
	void OpenFP(FILE *fp);
	/* Create a new file */
	void Create(const std::string &FileName, const GenInfoVec &Genomes);
	void CreateFP(FILE *fp, const GenInfoVec &Genomes);
	/* Close a file accessed with Open() or Create() */
	void Close();

	const GenInfoVec &GetGenomes() const;
	unsigned GetGenomeCount() const;
	const GenInfo &GetGenomeInfo(unsigned GenomeIndex) const;

	/* Methods for reading a file */
	unsigned GetEntryCount() const;
	void GetEntry(Entry &E, unsigned EntryIndex) const;
	void GetEntries(EntryVec &EVec, unsigned FromIndex, unsigned ToIndex) const;
	void GetAllEntries(EntryVec &EVec) const;

	/* Methods for writing a file */
	void AddEntry(const Entry &E);
	void AddEntries(const EntryVec &EVec);

	protected:

	enum KVFileState {Idle = 1, Reading, Writing};
	KVFileState m_State;
	GenInfoVec m_Genomes;
	off_t m_EntryCountPos;
	off_t m_EntryIndexPos;
	off_t m_EntryArrayPos;
	FILE *m_fp;
	bool m_HaveExternalFP;
	uint32 m_Magic;
	uint32 m_EntryCount;
	EntryDiskFunc m_EntryDiskFunc;

	std::vector<off_t> m_Index;

	void ReadHeaderGenomes();
	void WriteHeaderGenomes() const;
	void ReadEntryHeader();
	void WriteEntryHeader() const;
	void WriteIndex();
	void Init();
};


template <typename Entry, typename EntryDiskFunc>
KVFile<Entry, EntryDiskFunc>::KVFile()
{
	m_Magic = 0x16263646;
	m_State = Idle;
	m_EntryCount = 0;
}

template <typename Entry, typename EntryDiskFunc>
KVFile<Entry, EntryDiskFunc>::~KVFile()
{
	if (m_State != Idle)
		Close();
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::Open(const std::string &FileName)
{
	KVFileCheckState(Idle);
	m_fp = OpenStdioFile(FileName.c_str());
	m_HaveExternalFP = false;
	m_State = Reading;
	ReadHeaderGenomes();
	m_EntryCountPos = GetStdioFilePos(m_fp);
	ReadEntryHeader();
	m_EntryArrayPos = GetStdioFilePos(m_fp);
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::OpenFP(FILE *fp)
{
	KVFileCheckState(Idle);
	m_fp = fp;
	m_HaveExternalFP = true;
	SeekStdioFile(m_fp, 0);
	m_State = Reading;
	ReadHeaderGenomes();
	m_EntryCountPos = GetStdioFilePos(m_fp);
	ReadEntryHeader();
	m_EntryArrayPos = GetStdioFilePos(m_fp);
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::Create(const std::string &FileName, const GenInfoVec &Genomes)
{
	KVFileCheckState(Idle);
	/*
	if (Genomes.size() == 0)
		Die("KVFile: Invalid genome count");
	*/
	m_fp = CreateStdioFile(FileName.c_str());
	m_HaveExternalFP = false;
	m_State = Writing;
	m_Genomes = Genomes;
	WriteHeaderGenomes();
	m_EntryCountPos = GetStdioFilePos(m_fp);
	m_EntryCount = (unsigned)-1; //False entry count that will be corrected when closing file
	m_EntryIndexPos = 0; //False Entry index pos
	WriteEntryHeader();
	m_EntryCount = 0;
	m_EntryArrayPos = GetStdioFilePos(m_fp);
	m_Index.clear();
	m_Index.push_back(m_EntryArrayPos);
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::CreateFP(FILE *fp, const GenInfoVec &Genomes)
{
	KVFileCheckState(Idle);
	/*
	if (Genomes.size() == 0)
		Die("KVFile: Invalid genome count");
	*/
	m_fp = fp;
	m_HaveExternalFP = true;
	SeekStdioFile(m_fp, 0);
	m_State = Writing;
	m_Genomes = Genomes;
	WriteHeaderGenomes();
	m_EntryCountPos = GetStdioFilePos(m_fp);
	m_EntryCount = (unsigned)-1; //False entry count that will be corrected when closing file
	m_EntryIndexPos = 0; //False Entry index pos
	WriteEntryHeader();
	m_EntryCount = 0;
	m_EntryArrayPos = GetStdioFilePos(m_fp);
	m_Index.clear();
	m_Index.push_back(m_EntryArrayPos);
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::WriteIndex()
{
	KVFileCheckState(Writing);
	m_EntryIndexPos = GetStdioFilePos(m_fp);
	for (typename std::vector<off_t>::const_iterator i = m_Index.begin(); i != m_Index.end(); ++i)
	{
		WriteStdioFile(m_fp, &*i, sizeof(off_t));
	}
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::Close()
{
	if (m_State == Reading)
	{
		if (!m_HaveExternalFP)
			CloseStdioFile(m_fp);
	}
	else if (m_State == Writing)
	{
		WriteIndex();
		WriteEntryHeader();
		WriteStdioFile(m_fp, 0, &m_Magic, sizeof(m_Magic));
		if (!m_HaveExternalFP)
			CloseStdioFile(m_fp);
	}
	else
		Die("KVFile: Invalid close request");

	Init();
}

template <typename Entry, typename EntryDiskFunc>
const GenInfoVec &KVFile<Entry, EntryDiskFunc>::GetGenomes() const
{
	return m_Genomes;
}

template <typename Entry, typename EntryDiskFunc>
unsigned KVFile<Entry, EntryDiskFunc>::GetGenomeCount() const
{
	return m_Genomes.size();
}

template <typename Entry, typename EntryDiskFunc>
const GenInfo &KVFile<Entry, EntryDiskFunc>::GetGenomeInfo(unsigned GenomeIndex) const
{
	if (GenomeIndex >= m_Genomes.size())
		Die("Invalid genome index");
	return m_Genomes[GenomeIndex];
}

template <typename Entry, typename EntryDiskFunc>
unsigned KVFile<Entry, EntryDiskFunc>::GetEntryCount() const
{
	return m_EntryCount;
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::GetEntry(Entry &E, unsigned EntryIndex) const
{
	KVFileCheckState(Reading);
	if (EntryIndex >= m_EntryCount)
		Die("KVFile: Entry index (%u) out of bounds (%u)", EntryIndex, m_EntryCount);

	off_t EntryOffset;
	off_t NextEntryOffset;
	ReadStdioFile(m_fp, m_EntryIndexPos + EntryIndex * sizeof(off_t), &EntryOffset, sizeof(off_t));
	ReadStdioFile(m_fp, m_EntryIndexPos + (EntryIndex + 1)* sizeof(off_t), &NextEntryOffset, sizeof(off_t));

	assert(EntryOffset > 0);
	assert(NextEntryOffset > EntryOffset);

	unsigned BufferSize = unsigned(NextEntryOffset - EntryOffset);

	byte *Buffer = new byte[BufferSize];
	ReadStdioFile(m_fp, EntryOffset, Buffer, BufferSize);
	m_EntryDiskFunc(Buffer, BufferSize, E);
	delete Buffer;
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::GetEntries(EntryVec &EVec, unsigned FromIndex, unsigned ToIndex) const
{
	KVFileCheckState(Reading);
	assert(FromIndex <= ToIndex);
	assert(ToIndex < m_EntryCount);
	unsigned Count = ToIndex - FromIndex + 1;
	EVec.clear();
	EVec.resize(Count);

	off_t EntryOffset;
	off_t NextEntryOffset;
	ReadStdioFile(m_fp, m_EntryIndexPos + FromIndex * sizeof(off_t), &EntryOffset, sizeof(off_t));
	assert(EntryOffset > 0);
	for (unsigned i = FromIndex; i <= ToIndex; ++i)
	{
		ReadStdioFile(m_fp, m_EntryIndexPos + (i + 1) * sizeof(off_t), &NextEntryOffset, sizeof(off_t));
		assert(NextEntryOffset > EntryOffset);

		unsigned BufferSize = unsigned(NextEntryOffset - EntryOffset);

		byte *Buffer = new byte[BufferSize];
		ReadStdioFile(m_fp, EntryOffset, Buffer, BufferSize);
		m_EntryDiskFunc(Buffer, BufferSize, EVec[i-FromIndex]);
		delete Buffer;
		EntryOffset = NextEntryOffset;
	}
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::GetAllEntries(EntryVec &EVec) const
{
	KVFileCheckState(Reading);
	EVec.clear();
	EVec.resize(m_EntryCount);

	off_t EntryOffset;
	off_t NextEntryOffset;
	ReadStdioFile(m_fp, m_EntryIndexPos, &EntryOffset, sizeof(off_t));
	assert(EntryOffset > 0);
	for (unsigned i = 0; i != m_EntryCount; ++i)
	{
		ReadStdioFile(m_fp, m_EntryIndexPos + (i + 1) * sizeof(off_t), &NextEntryOffset, sizeof(off_t));
		assert(NextEntryOffset > EntryOffset);

		unsigned BufferSize = unsigned(NextEntryOffset - EntryOffset);

		byte *Buffer = new byte[BufferSize];
		ReadStdioFile(m_fp, EntryOffset, Buffer, BufferSize);
		m_EntryDiskFunc(Buffer, BufferSize, EVec[i]);
		delete Buffer;
		EntryOffset = NextEntryOffset;
	}
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::AddEntry(const Entry &E)
{
	KVFileCheckState(Writing);
	++m_EntryCount;
	byte *BufferPtr;
	unsigned BufferSize = m_EntryDiskFunc(&BufferPtr, E);
	WriteStdioFile(m_fp, BufferPtr, BufferSize);
	off_t CurrentOffset = GetStdioFilePos(m_fp);
	m_Index.push_back(CurrentOffset);

	delete BufferPtr;
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::AddEntries(const EntryVec &EVec)
{
	KVFileCheckState(Writing);
	for (typename EntryVec::const_iterator i = EVec.begin(); i != EVec.end(); ++i)
	{
		AddEntry(*i);
	}
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::ReadHeaderGenomes()
{
	uint32 Magic;
	ReadStdioFile(m_fp, &Magic, sizeof(Magic));
	if (Magic != m_Magic)
		Die("KVFile: Invalid Magic");

	uint32 GenomeCount;
	ReadStdioFile(m_fp, &GenomeCount, sizeof(GenomeCount));
	/*
	if (GenomeCount == 0)
		Die("KVFile: Invalid genome count");
	*/

	m_Genomes.resize(GenomeCount);
	for (unsigned GenomeIndex = 0; GenomeIndex != GenomeCount; ++GenomeIndex)
		m_Genomes[GenomeIndex].FromFile(m_fp);
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::WriteHeaderGenomes() const
{
	/* False magic will be corrected when closing the file */
	uint32 FalseMagic = 0x01020304;
	WriteStdioFile(m_fp, &FalseMagic, sizeof(FalseMagic));

	uint32 GenomeCount = GetGenomeCount();
	WriteStdioFile(m_fp, &GenomeCount, sizeof(GenomeCount));

	for (unsigned GenomeIndex = 0; GenomeIndex != GenomeCount; ++GenomeIndex)
		m_Genomes[GenomeIndex].ToFile(m_fp);
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::ReadEntryHeader()
{
	ReadStdioFile(m_fp, &m_EntryCount, sizeof(m_EntryCount));
	if (m_EntryCount == (uint32)-1)
		Die("KVFile: Invalid pair count");
	ReadStdioFile(m_fp, &m_EntryIndexPos, sizeof(m_EntryIndexPos));
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::WriteEntryHeader() const
{
	WriteStdioFile(m_fp, m_EntryCountPos, &m_EntryCount, sizeof(m_EntryCount));
	WriteStdioFile(m_fp, m_EntryCountPos + sizeof(m_EntryCount), &m_EntryIndexPos, sizeof(m_EntryIndexPos));
}

template <typename Entry, typename EntryDiskFunc>
void KVFile<Entry, EntryDiskFunc>::Init()
{
	m_Genomes.clear();
	m_EntryCount = 0;
	m_State = Idle;
	m_fp = NULL;
	m_HaveExternalFP = false;
	m_Index.clear();
}

#endif
