#ifndef lhist_h
#define lhist_h

#include <math.h>
#include <string>
#include <vector>

using namespace std;

class LHistData
	{
public:
	LHistData()
		{
		Clear();
		}

	void Clear()
		{
		m_Name.clear();
		m_Counts.clear();
		m_BinLos.clear();
		m_Sum = 0;
		m_MinVal = UINT_MAX;
		m_MaxVal = 0;
		}

	void Init(const vector<unsigned> &BinLos, const string &Name)
		{
		Clear();
		m_Name = Name;
		m_BinLos = BinLos;
		sort(m_BinLos.begin(), m_BinLos.end());
		m_Counts.resize(GetBinCount());
		}

	void Init(const string &Name, unsigned From, unsigned To,
	  unsigned BinWidth)
		{
		Clear();
		m_Name = Name;
		asserta(From <= To);
		if (From != 0)
			m_BinLos.push_back(0);
		for (unsigned i = From; i <= To; i += BinWidth)
			m_BinLos.push_back(i);
		sort(m_BinLos.begin(), m_BinLos.end());
		m_Counts.resize(GetBinCount());
		}

	void InitPctId(const string &Name)
		{
		Clear();
		m_Name = Name;
		GetBinLosPctId(m_BinLos);
		sort(m_BinLos.begin(), m_BinLos.end());
		m_Counts.resize(GetBinCount());
		}

	void Init(const string &Name)
		{
		Clear();
		m_Name = Name;
		GetBinLosLengths(m_BinLos);
		sort(m_BinLos.begin(), m_BinLos.end());
		m_Counts.resize(GetBinCount());
		}

	void InitCSV(const string &Name, unsigned Step, unsigned MaxLen)
		{
		Clear();
		m_Name = Name;
		GetBinLosCSV(m_BinLos, Step, MaxLen);
		sort(m_BinLos.begin(), m_BinLos.end());
		m_Counts.resize(GetBinCount());
		}

	unsigned GetBinCount() const
		{
		return SIZE(m_BinLos);
		}

	const char *BinName(unsigned BinIndex) const
		{
		static char s[32];
		unsigned N = SIZE(m_BinLos);
		asserta(BinIndex < N);
		unsigned Lo = UINT_MAX;
		unsigned Hi = UINT_MAX;
		if (BinIndex + 1 == N)
			{
			Lo = m_BinLos[N-1];
			Hi = m_MaxVal;
			}
		else
			{
			Lo = m_BinLos[BinIndex];
			Hi = m_BinLos[BinIndex+1] - 1;
			}
		if (Lo == Hi)
			sprintf(s, "%u", Lo);
		else
			sprintf(s, "%u - %u", Lo, Hi);
		return s;
		}

	unsigned BinIndex(unsigned Length) const
		{
		unsigned N = SIZE(m_BinLos);
		asserta(N > 0);
		if (Length == 0)
			{
			if (m_BinLos[0] == 0)
				return 0;
			Die("LHistData::Add, L=0 but no zero bin");
			}
		for (unsigned BinIndex = 0; BinIndex < N - 1; ++BinIndex)
			{
			unsigned NextLo = m_BinLos[BinIndex+1];
			if (Length < NextLo)
				return BinIndex;
			}
		return N - 1;
		}

	void Add(const vector<unsigned> &Lengths)
		{
		const unsigned N = SIZE(Lengths);
		for (unsigned i = 0; i < N; ++i)
			Add(Lengths[i]);
		}

	void Add(unsigned Length, unsigned Count = 1)
		{
		unsigned Bin = BinIndex(Length);
		m_Counts[Bin] += Count;
		m_Sum += Count*Length;
		if (m_MinVal == UINT_MAX && m_MaxVal == 0)
			{
			m_MinVal = Length;
			m_MaxVal = Length;
			}
		else
			{
			m_MinVal = min(Length, m_MinVal);
			m_MaxVal = max(Length, m_MaxVal);
			}
		}

	double GetCount() const
		{
		double N = 0;
		for (unsigned i = 0; i < GetBinCount(); ++i)
			N += m_Counts[i];
		return N;
		}

	double GetMean() const
		{
		double N = GetCount();
		if (N == 0)
			return 0;
		return m_Sum/(double) N;
		}

	double GetSum() const
		{
		return m_Sum;
		}

	unsigned GetBinWidth(unsigned BinIndex) const
		{
		unsigned N = SIZE(m_Counts);
		asserta(N > 0 && BinIndex < N);
		if (BinIndex == N - 1)
			{
			if (m_MaxVal < m_BinLos[N-1])
				return 1;
			return m_MaxVal - m_BinLos[N-1] + 1;
			}
		return m_BinLos[BinIndex+1] - m_BinLos[BinIndex];
		}

	double GetAvgCount(unsigned BinIndex) const
		{
		asserta(BinIndex < SIZE(m_Counts));
		unsigned Count = m_Counts[BinIndex];
		unsigned Width = GetBinWidth(BinIndex);
		return (double) Count / (double) Width;
		}

	void OutCSV(FILE *f) const
		{
		const unsigned BinCount = GetBinCount();
		for (unsigned BinIndex = 0; BinIndex < BinCount; ++BinIndex)
			{
			unsigned Count = m_Counts[BinIndex];
			unsigned Lo = m_BinLos[BinIndex];
			fprintf(f, "%s;%u;%u\n", m_Name.c_str(), Lo, Count);
			}
		}

	void LogHist(bool LogarithmicScale = false) const
		{
		Log("\n");
		if (!m_Name.empty())
			Log("%s:\n", m_Name.c_str());
		Log("Avg %.1f, N=%u, total %u\n",
		  GetMean(), (unsigned) GetCount(), (unsigned) m_Sum);

		double MaxAvgCount = 0;
		unsigned MinNonEmptyBin = UINT_MAX;
		unsigned MaxNonEmptyBin = UINT_MAX;
		const unsigned BinCount = GetBinCount();
		for (unsigned i = 0; i < BinCount; ++i)
			{
			double AvgCount = GetAvgCount(i);
			if (AvgCount > 0)
				{
				if (MinNonEmptyBin == UINT_MAX)
					MinNonEmptyBin = i;
				MaxNonEmptyBin = i;
				}
			MaxAvgCount = max(AvgCount, MaxAvgCount);
			}
		double Total = GetCount();

		if (MinNonEmptyBin == UINT_MAX)
			return;
		if (LogarithmicScale)
			Log("Log scale\n");

		Log("                Range          Nr    Pct\n");
		Log("=====================  ==========  =====\n");
		for (unsigned i = MinNonEmptyBin; i <= MaxNonEmptyBin; ++i)
			{
		TopOfLoop:
			unsigned Count = m_Counts[i];
			if (Count == 0)
				{
				unsigned EmptyBinCount = 0;
				unsigned NextNonEmptyBin = i + 1;
				for (unsigned j = i + 1; j <= MaxNonEmptyBin; ++j)
					{
					if (m_Counts[j] != 0)
						{
						NextNonEmptyBin = j;
						break;
						}
					++EmptyBinCount;
					}
				if (EmptyBinCount > 2)
					{
					Log("...(%u empty bins)...\n", EmptyBinCount);
					i = NextNonEmptyBin;
					goto TopOfLoop;
					}
				}

			Log("%21.21s", BinName(i));
			Log("  %10u", m_Counts[i]);
			if (Total > 0)
				Log("  %5.1f%%  ", Count*100.0/Total);

			double AvgCount = GetAvgCount(i);
			double RelCount = 0;
			if (LogarithmicScale)
				{
				RelCount = log(AvgCount*10)/log(MaxAvgCount*10);
				if (RelCount < 0)
					RelCount = 0;
				}
			else
				RelCount = AvgCount/MaxAvgCount;
			double Stars = 64.0*RelCount;
			unsigned iStars = (unsigned) (Stars + 0.5);
			if (iStars > 64)
				iStars = 64;
			for (unsigned j = 0; j < iStars; ++j)
				Log("*");
			Log("\n");
			}
		}

	void GetBinLosPctId(vector<unsigned> &BinLos) const
		{
		BinLos.clear();
		BinLos.push_back(0);
		BinLos.push_back(1);
		BinLos.push_back(21);
		BinLos.push_back(21);
		BinLos.push_back(41);
		BinLos.push_back(51);
		BinLos.push_back(71);
		BinLos.push_back(81);
		BinLos.push_back(91);
		BinLos.push_back(101);
		}

	void GetBinLosCSV(vector<unsigned> &BinLos, unsigned Step, unsigned MaxLen) const
		{
		BinLos.clear();
		for (unsigned L = 0; L <= MaxLen; L += Step)
			BinLos.push_back(L);
		}

	void GetBinLosLengths(vector<unsigned> &BinLos) const
		{
		BinLos.clear();
		BinLos.push_back(0);
		BinLos.push_back(1);
		BinLos.push_back(2);
		BinLos.push_back(3);
		BinLos.push_back(4);
		BinLos.push_back(5);
		BinLos.push_back(6);
		BinLos.push_back(7);
		BinLos.push_back(8);
		BinLos.push_back(9);
		BinLos.push_back(10);
		BinLos.push_back(20);
		BinLos.push_back(30);
		BinLos.push_back(40);
		BinLos.push_back(50);
		BinLos.push_back(70);
		BinLos.push_back(80);
		BinLos.push_back(90);
		BinLos.push_back(100);
		BinLos.push_back(200);
		BinLos.push_back(300);
		BinLos.push_back(400);
		BinLos.push_back(500);
		BinLos.push_back(600);
		BinLos.push_back(700);
		BinLos.push_back(800);
		BinLos.push_back(900);
		BinLos.push_back(1000);
		BinLos.push_back(10000);
		BinLos.push_back(100000);
		BinLos.push_back(1000000);
		BinLos.push_back(10000000);
		BinLos.push_back(100000000);
		}

private:
	string m_Name;
	vector<unsigned> m_Counts;
	vector<unsigned> m_BinLos;
	double m_Sum;
	unsigned m_MinVal;
	unsigned m_MaxVal;
	};

#endif // lhist_h
