#include "maffile.h"
#include "utils.h"

// MAF file format described here:
// http://genome.ucsc.edu/FAQ/FAQformat#format5

void Split(const std::string &Str, std::vector<std::string> &Fields);

MAFFile::MAFFile()
	{
	m_f = 0;
	m_PlusCoords = false;
	}

MAFFile::~MAFFile()
	{
	if (m_f != 0)
		{
		CloseStdioFile(m_f);
		m_f = 0;
		}
	}

void MAFFile::Create(const char *FileName, const char *CmdLine,
  const char *Scoring, const char *Program)
	{
	m_f = CreateStdioFile(FileName);

	WriteStdioFileStr(m_f, "##maf version=1");
	if (Scoring != 0 && !streq(Scoring, ""))
		{
		WriteStdioFileStr(m_f, " scoring=");
		WriteStdioFileStr(m_f, Scoring);
		}
	if (Program != 0&& !streq(Program, ""))
		{
		WriteStdioFileStr(m_f, " program=");
		WriteStdioFileStr(m_f, Program);
		}
	WriteStdioFileStr(m_f, "\n");

	if (CmdLine != 0&& !streq(CmdLine, ""))
		{
		WriteStdioFileStr(m_f, "# ");
		WriteStdioFileStr(m_f, CmdLine);
		WriteStdioFileStr(m_f, "\n");
		}
	WriteStdioFileStr(m_f, "\n");
	}

static void FormatScore(char *s, double Score)
	{
	if (Score > 1 && Score < 1e6)
		sprintf(s, "%.1f", Score);
	else if (Score > 0.001 && Score <= 1)
		sprintf(s, "%.5f", Score);
	else
		sprintf(s, "%g", Score);
	}

// Los[] and His[] are zero-based relative to the original sequence.
// This differs from the MAF convention, where Start/End pos is
// relative to the reverse-complented sequence if Strand='-'.
void MAFFile::AppendAln(const PosVec &Los, const PosVec &His, const BoolVec &Pluses,
  const StdStrVec &Rows, const StdStrVec &SeqNames, const PosVec &SeqLengths,
  double Score, unsigned Pass, const char *Comment)
	{
	const unsigned RowCount = SIZE(Los);
	asserta(SIZE(His) == RowCount);
	asserta(SIZE(Pluses) == RowCount);
	asserta(SIZE(Rows) == RowCount);
	asserta(SIZE(SeqNames) == RowCount);
	asserta(SIZE(SeqLengths) == RowCount);

	if (0 == RowCount)
		return;

/***
Alignment Block Lines (lines starting with 'a' -- parameters for
a new alignment block.

a score=23262.0

Each alignment begins with an 'a' line that set variables for the
entire alignment block. The 'a' is followed by name=value pairs.
There are no required name=value pairs. The currently defined variables are: 

score -- Optional. Floating point score. If this is present, it is good
practice to also define scoring in the first line. 

pass -- Optional. Positive integer value. For programs that do multiple
pass alignments such as blastz, this shows which pass this alignment
came from. Typically, pass 1 will find the strongest alignments genome-wide,
and pass 2 will find weaker alignments between two first-pass alignments. 
***/
	if (Comment != 0)
		{
		WriteStdioFileStr(m_f, "# ");
		WriteStdioFileStr(m_f, Comment);
		WriteStdioFileStr(m_f, "\n");
		}
	WriteStdioFileStr(m_f, "a");
	if (Score != NO_SCORE)
		{
		WriteStdioFileStr(m_f, " score=");

		char s[94];
		FormatScore(s, Score);
		WriteStdioFileStr(m_f, s);
		}
	if (Pass > 0)
		{
		char s[64];
		sprintf(s, " pass=%u", Pass);
		WriteStdioFileStr(m_f, s);
		}
	if (RowCount == 2)
		{
		double PctId = ComputePctId(Rows);
		char s[64];
		sprintf(s, " pctid=%.1f", PctId);
		WriteStdioFileStr(m_f, s);
		}
	WriteStdioFileStr(m_f, "\n");

/***
Lines starting with 's' -- a sequence within an alignment block 

s hg16.chr7    27707221 13 + 158545518 gcagctgaaaaca
s panTro1.chr6 28869787 13 + 161576975 gcagctgaaaaca
s baboon         249182 13 +   4622798 gcagctgaaaaca
s mm4.chr6     53310102 13 + 151104725 ACAGCTGAAAATA

The 's' lines together with the 'a' lines define a multiple alignment.
The 's' lines have the following fields which are defined by position
rather than name=value pairs. 

src -- The name of one of the source sequences for the alignment.
For sequences that are resident in a browser assembly, the form
'database.chromosome' allows automatic creation of links to other
assemblies. Non-browser sequences are typically referenced by the
species name alone. 

start -- The start of the aligning region in the source sequence.
This is a zero-based number. If the strand field is '-' then this is the
start relative to the reverse-complemented source sequence. 

size -- The size of the aligning region in the source sequence.
This number is equal to the number of non-dash characters in the
alignment text field below. 

strand -- Either '+' or '-'. If '-', then the alignment is to the
reverse-complemented source. 

srcSize -- The size of the entire source sequence, not just the parts
involved in the alignment. 

text -- The nucleotides (or amino acids) in the alignment and any
insertions (dashes) as well. All sequence columns and rows must
contain at least one nucleotide (no full columns or rows of insertions). 
***/

	const unsigned ColCount = SIZE(Rows[0]);
	for (unsigned RowIndex = 1; RowIndex < RowCount; ++RowIndex)
		if (SIZE(Rows[RowIndex]) != ColCount)
			Die("MAFFile::AppendAln, row lengths differ");

// Flag rows and columns that are all-gaps
	IntVec NonGapColCounts;
	BoolVec EmptyCols;

	NonGapColCounts.reserve(RowCount);
	EmptyCols.reserve(ColCount);

	for (unsigned ColIndex = 0; ColIndex < ColCount; ++ColIndex)
		{
		bool AllGaps = true;
		for (unsigned RowIndex = 0; RowIndex < RowCount; ++RowIndex)
			{
			char c = Rows[RowIndex][ColIndex];
			if (c != '-' && c != '.')
				{
				AllGaps = false;
				break;
				}
			}
		EmptyCols.push_back(AllGaps);
		}

	unsigned EmptyRowCount = 0;
	for (unsigned RowIndex = 0; RowIndex < RowCount; ++RowIndex)
		{
		unsigned NonGapCharCount = 0;
		const StdStr &Row = Rows[RowIndex];
		for (unsigned ColIndex = 0; ColIndex < ColCount; ++ColIndex)
			{
			if (EmptyCols[ColIndex])
				continue;
			if (Row[ColIndex] != '-' && Row[ColIndex] != '.')
				++NonGapCharCount;
			}
		if (NonGapCharCount == 0)
			++EmptyRowCount;
		NonGapColCounts.push_back(NonGapCharCount);
		}

	if (EmptyRowCount == RowCount)
		return;

	PosVec StartPoses;
	StartPoses.reserve(RowCount);
	for (unsigned RowIndex = 0; RowIndex < RowCount; ++RowIndex)
		{
		if (NonGapColCounts[RowIndex] == 0)
			StartPoses.push_back(0);
		else
			{
			coord_t lo = Los[RowIndex];
			bool Plus = Pluses[RowIndex];
			if (Plus || m_PlusCoords)
				StartPoses.push_back(lo);
			else
				{
				coord_t SeqLength = SeqLengths[RowIndex];
				coord_t hi = His[RowIndex];
				asserta(SeqLength >= (hi + 1));

				coord_t StartPos = SeqLength - hi - 1;
				StartPoses.push_back(StartPos);
				}
			}
		}

// Determine length of longest fields
	unsigned MaxNameLen = 0;
	unsigned MaxStartWidth = 0;
	unsigned MaxSizeWidth = 0;
	unsigned MaxSeqLenWidth = 0;
	for (unsigned RowIndex = 0; RowIndex < RowCount; ++RowIndex)
		{
		if (NonGapColCounts[RowIndex] == 0)
			continue;

		unsigned Width = SIZE(SeqNames[RowIndex]);
		MaxNameLen = std::max(MaxNameLen, Width);

		char s[64];
		sprintf(s, "%u", StartPoses[RowIndex]);
		Width = (unsigned) strlen(s);
		MaxStartWidth = std::max(MaxStartWidth, Width);

		sprintf(s, "%u", NonGapColCounts[RowIndex]);
		Width = (unsigned) strlen(s);
		MaxSizeWidth = std::max(MaxSizeWidth, Width);

		sprintf(s, "%u", SeqLengths[RowIndex]);
		Width = (unsigned) strlen(s);
		MaxSeqLenWidth = std::max(MaxSeqLenWidth, Width);
		}

// s hg16.chr7    27707221 13 + 158545518 gcagctgaaaaca
	unsigned RowCol = 2 + MaxNameLen + 1 + MaxStartWidth + 1 + MaxSizeWidth +
	  1 + MaxSeqLenWidth + 3;

// Output rows
	for (unsigned RowIndex = 0; RowIndex < RowCount; ++RowIndex)
		{
		if (NonGapColCounts[RowIndex] == 0)
			continue;

		WriteStdioFileStr(m_f, "s ");

		const StdStr &Name = SeqNames[RowIndex];
		WriteStdioFileStr(m_f, Name.c_str());
		unsigned Width = SIZE(Name);
		for (unsigned i = Width ; i < MaxNameLen; ++i)
			WriteStdioFileStr(m_f, " ");

		coord_t Pos = StartPoses[RowIndex];
		unsigned uPos = (unsigned) Pos;
		if (Pos != (coord_t) uPos)
			Die("Coord overflowed in MAFFile::AppendAln");
		char s[64];
		sprintf(s, "%u", uPos);
		Width = (unsigned) strlen(s);
		WriteStdioFileStr(m_f, " ");
		for (unsigned i = Width; i < MaxStartWidth; ++i)
			WriteStdioFileStr(m_f, " ");
		WriteStdioFileStr(m_f, s);

		unsigned Size = NonGapColCounts[RowIndex];
		sprintf(s, "%u", Size);
		WriteStdioFileStr(m_f, " ");
		Width = (unsigned) strlen(s);
		for (unsigned i = Width; i < MaxSizeWidth; ++i)
			WriteStdioFileStr(m_f, " ");
		WriteStdioFileStr(m_f, s);

		WriteStdioFileStr(m_f, Pluses[RowIndex] ? " +" : " -");

		unsigned SeqLength = SeqLengths[RowIndex];
		sprintf(s, "%u", SeqLength);
		Width = (unsigned) strlen(s);
		WriteStdioFileStr(m_f, " ");
		for (unsigned i = Width; i < MaxSeqLenWidth; ++i)
			WriteStdioFileStr(m_f, " ");
		WriteStdioFileStr(m_f, s);

		WriteStdioFileStr(m_f, " ");
		s[1] = 0;
		const StdStr &Row = Rows[RowIndex];
		for (unsigned ColIndex = 0; ColIndex < ColCount; ++ColIndex)
			{
			if (EmptyCols[ColIndex])
				continue;
			s[0] = Row[ColIndex];
			WriteStdioFileStr(m_f, s);
			}
		WriteStdioFileStr(m_f, "\n");
		}

	if (m_SubCmt.size() > 0)
		{
		WriteStdioFileStr(m_f, "# ");
		for (unsigned i = 0; i < RowCol - 2; ++i)
			WriteStdioFileStr(m_f, " ");
		WriteStdioFileStr(m_f, m_SubCmt.c_str());
		WriteStdioFileStr(m_f, "\n");
		m_SubCmt.clear();
		}

// Alignment block ended by blank line
	WriteStdioFileStr(m_f, "\n");
	}

void MAFFile::Close()
	{
	CloseStdioFile(m_f);
	m_f = 0;
	}

void MAFFile::Open(const char *FileName)
	{
	m_f = OpenStdioFile(FileName);
	}

bool MAFFile::GetNextAln(PosVec &Los, PosVec &His, BoolVec &Pluses,
  StdStrVec &Rows, StdStrVec &SeqNames, PosVec &SeqLengths, double &Score)
	{
	His.clear();
	Los.clear();
	Pluses.clear();
	Rows.clear();
	SeqNames.clear();
	SeqLengths.clear();
	Score = 0;

	const size_t BUFF_SIZE = 1000000;
	char *Buffer = myalloc(char, BUFF_SIZE);

	for (;;)
		{
		bool EndOfFile = ReadLineStdioFile(m_f, Buffer, BUFF_SIZE);
		if (EndOfFile)
			return false;

		if (Buffer[0] == 0)
			continue;
		if (Buffer[0] == 'a')
			{
			Parse_a_Line(Buffer, Score);
			break;
			}
		if (Buffer[0] == '#')
			continue;
		Buffer[10] = 0;
		Die("Expected 'a' or '#', got: %.10s", Buffer);
		}

	for (;;)
		{
		off64_t Pos = GetStdioFilePos(m_f);
		bool EndOfFile = ReadLineStdioFile(m_f, Buffer, BUFF_SIZE);
		if (EndOfFile)
			break;
		if (Buffer[0] == 0)
			continue;
		if (Buffer[0] == 'a')
			{
			SeekStdioFile(m_f, Pos);
			break;
			}

		if (Buffer[0] == '#')
			continue;
		if (Buffer[0] == 'i')
			continue;
		if (Buffer[0] == 'e')
			continue;

		if (Buffer[0] != 's')
			Die("Expected 's', got: %.10s", Buffer);

		Parse_s_Line(Buffer, SeqNames, SeqLengths, Los, His, Pluses, Rows);
		}

	myfree(Buffer);
	return true;
	}

// Parse s-line, /appending/ data to arrays.
void MAFFile::Parse_s_Line(const char *Buffer, StdStrVec &SeqNames,
  PosVec &SeqLengths, PosVec &Los, IntVec &His, BoolVec &Pluses, StdStrVec &Rows)
	{
	if (Buffer[0] != 's')
		Die("MAFFile::Parse_s_Line, must be 's' line got: %.10s", Buffer);

	StdStrVec Fields;
	Split(Buffer, Fields);

	if (Fields[0] != "s")
		Die("Parse_s_line: Field[0]='%s', expected 's'", Fields[0].c_str());

	if (Fields[4] != "+" && Fields[4] != "-")
		Die("Parse_s_line: Field[4]='%s', expected '+' or '-'", Fields[0].c_str());

	StdStr &SeqName = Fields[1];
	StdStr &Row = Fields[6];

	coord_t StartPos = StrToCoord(Fields[2].c_str());
	coord_t LetterCount = StrToCoord(Fields[3].c_str());
	coord_t SeqLength = StrToCoord(Fields[5].c_str());
	bool Plus = (Fields[4] == "+");

	coord_t lo = COORD_T_MAX;
	coord_t hi = COORD_T_MAX;
	if (Plus)
		{
		lo = StartPos;
		hi = StartPos + LetterCount - 1;
		}
	else
		{
		asserta(SeqLength >= (StartPos + 1));
		hi = SeqLength - StartPos - 1;

		asserta(hi + 1 >= LetterCount);
		lo = (hi + 1) - LetterCount;
		}

	Los.push_back(lo);
	His.push_back(hi);
	Pluses.push_back(Plus);
	SeqNames.push_back(SeqName);
	Rows.push_back(Row);
	SeqLengths.push_back(SeqLength);
	}

/***
a score=23262.0

Each alignment begins with an 'a' line that sets variables for the entire
alignment block. The 'a' is followed by name=value pairs. There are no
required name=value pairs. The currently defined variables are: 

score -- Optional. Floating point score.
	If this is present, it is good practice to also define
	scoring in the first line.

pass -- Optional. Positive integer value.
	For programs that do multiple pass alignments such as blastz, this shows
	which pass this alignment came from. Typically, pass 1 will find the
	strongest alignments genome-wide, and pass 2 will find weaker alignments
	between two first-pass alignments.
***/
void MAFFile::Parse_a_Line(const char *Buffer, double &Score)
	{
	StdStrVec Fields;
	Split(Buffer, Fields);
	if (Fields[0] != "a")
		Die("Parse_a_line(%s): expected 'a' as first field, got '%s'",
		  Buffer, Fields[0].c_str());
	unsigned FieldCount = SIZE(Fields);
	for (unsigned FieldIndex = 1; FieldIndex < FieldCount; ++FieldIndex)
		{
		const StdStr &NameEqValue = Fields[FieldIndex];
		std::string::size_type n = NameEqValue.find('=');
		if (n == std::string::npos || n == 0 || n == SIZE(NameEqValue) - 1)
			Die("Parse_a_line(%s), expected name=value, got '%s'",
			  Buffer, NameEqValue.c_str());
		StdStr Name = NameEqValue.substr(0, n);
		StdStr Value = NameEqValue.substr(n+1);
		if (Name == "score")
			Score = atof(Value.c_str());
		}
	}

off64_t MAFFile::GetFileSize() const
	{
	return GetStdioFileSize(m_f);
	}

off64_t MAFFile::GetFilePos() const
	{
	return GetStdioFilePos(m_f);
	}

void MAFFile::Rewind()
	{
	SeekStdioFile(m_f, 0);
	}

double MAFFile::ComputePctId(const StdStrVec &Rows) const
	{
	const unsigned RowCount = SIZE(Rows);
	if (RowCount != 2)
		Die("MAFFile::ComputePctId, %u rows, must be 2", RowCount);

	const unsigned ColCount = SIZE(Rows[0]);
	if (ColCount == 0)
		return 0;

	const StdStr &Row0 = Rows[0];
	const StdStr &Row1 = Rows[1];
	unsigned SameCount = 0;
	unsigned NonGappedCount = 0;
	FORIX(Col)
		{
		char c0 = Row0[ColIndex];
		if (c0 == '-')
			continue;
		char c1 = Row1[ColIndex];
		if (c1 == '-')
			continue;
		++NonGappedCount;
		if ((c0 == c1) || (isalpha(c0) && isalpha(c1) && (toupper(c0) == toupper(c1))))
			++SameCount;
		}
	if (NonGappedCount == 0)
		return 0;
	return (double) SameCount*100.0/NonGappedCount;
	}
