#include "align4.h"

#define TRACE	0

struct GapInfo
	{
	coord_t lo;
	coord_t hi;
	unsigned LoAIIndex;
	unsigned HiAIIndex;
	};

void LogGap(const GapInfo &Gap)
	{
	Log("Lo=%7d  Hi=%7d  LoAIx=%d  HiAIx=%d",
	  Gap.lo, Gap.hi, Gap.LoAIIndex, Gap.HiAIIndex);
	}

#ifndef NDEBUG
static void AssertBPsSorted(const std::vector<AIBP> &BPs)
	{
	const unsigned N = (unsigned) BPs.size();
	for (unsigned i = 1; i < N; ++i)
		{
		assert(BPs[i-1].Pos <= BPs[i].Pos);
		}
	}
#else
#define AssertBPsSorted(BPs) /* empty */
#endif

typedef std::vector<GapInfo> GapVec;

/***
Gaps are identified by considering bendpoints in order of increasing genome
position, adding 1 to the coverage count for each Lo end and subtracting 1
for each Hi end. A gap starts when the copy count drops to zero, and ends
when the copy count goes from zero to 1.
***/
static void GetGaps(const std::vector<AIBP> &BPs, coord_t ChrLength,
  GapVec &Gaps, IntVec &HitToLoGap, IntVec &HitToHiGap)
	{
#if	TRACE
	Log("GetGaps\n");
	LogBPs(BPs);
#endif

	AssertBPsSorted(BPs);

	Gaps.clear();
	HitToLoGap.clear();
	HitToHiGap.clear();

	const unsigned BPCount = (unsigned) BPs.size();
	assert(BPCount%2 == 0);
	const unsigned AICount = BPCount/2;

	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		HitToLoGap.push_back(UINT_MAX);
		HitToHiGap.push_back(UINT_MAX);
		}

	if (BPs.size() > 0)
		{
		{
	// Lo end gap
		const AIBP &FirstBP = BPs.front();
		assert(FirstBP.IsLo);
		if (FirstBP.Pos > 0 && g_EndHoleSize > 0)
			{
			coord_t FirstGapLo =
			  (FirstBP.Pos >= g_EndHoleSize)? FirstBP.Pos - g_EndHoleSize : 0;
			GapInfo FirstGap;
			FirstGap.hi = FirstBP.Pos - 1;
			FirstGap.lo = FirstGapLo;
			FirstGap.LoAIIndex = UINT_MAX;
			FirstGap.HiAIIndex = FirstBP.AlnInfoIndex;
			HitToLoGap[FirstBP.AlnInfoIndex] = (unsigned) Gaps.size();
			Gaps.push_back(FirstGap);
			}
		}

	// Hi end gap
		const AIBP &LastBP = BPs.back();
		assert(!LastBP.IsLo);
		GapInfo LastGap;
		if (LastBP.Pos < ChrLength - 1 && g_EndHoleSize > 0)
			{
			LastGap.lo = LastBP.Pos + 1;
			LastGap.hi = LastBP.Pos + g_EndHoleSize;
			if (LastGap.hi >= ChrLength)
				LastGap.hi = ChrLength - 1;
			LastGap.LoAIIndex = LastBP.AlnInfoIndex;
			LastGap.HiAIIndex = UINT_MAX;
			HitToHiGap[LastBP.AlnInfoIndex] = (unsigned) Gaps.size();
			Gaps.push_back(LastGap);
			}
		}

	unsigned CopyCount = 0;
	coord_t GapStartPos = 0;
	unsigned GapStartAIIndex = 0;
	for (unsigned BPIndex = 0; BPIndex < BPCount; ++BPIndex)
		{
		const AIBP &BP = BPs[BPIndex];
#if	TRACE
		Log("\n");
		Log("Loop: BP[%3u] ", BPIndex);
		LogBP(BP);
		Log(" CopyCount=%u GapStartAIIndex=%u GapStartPos=%u\n",
		  CopyCount, GapStartAIIndex, (unsigned) GapStartPos);
#endif
		if (BP.IsLo)
			{
			if (CopyCount == 0 && BPIndex > 0)
				{
				GapInfo Gap;
				Gap.lo = GapStartPos;
				assert(BP.Pos > 0);
				Gap.hi = BP.Pos - 1;
				Gap.HiAIIndex = BP.AlnInfoIndex;
				Gap.LoAIIndex = GapStartAIIndex;

				unsigned GapIndex = (unsigned) Gaps.size();

				if (Gap.hi > Gap.lo)
					{
#if	TRACE
					Log("Add gap %u: ", GapIndex);
					LogGap(Gap);
					Log("\n");
#endif
					HitToLoGap[BP.AlnInfoIndex] = GapIndex;
					HitToHiGap[GapStartAIIndex] = GapIndex;
					Gaps.push_back(Gap);
					}
				}
			++CopyCount;
			}
		else
			{
			if (CopyCount == 0)
				Die("GetGaps, copy count < 0");
			--CopyCount;
			if (CopyCount == 0)
				{
				GapStartPos = BP.Pos + 1;
				GapStartAIIndex  = BP.AlnInfoIndex;
#if	TRACE
				Log("CopyCount -> 0, GapStartPos=%u, GapStartAIIndex=%u\n",
				  GapStartPos, GapStartAIIndex);
#endif
				}
			}
		}
	}

/***
MakeHoles version 4.
Make holes given a set of hits.
In outline, the method is as follows.

- Identify gaps w.r.t. each genome, i.e. regions not covered by any hit,
  while doing so
	* Record flanking (Lo and Hi) hits for each gap in HoleToLo/HiHit[]
	* Record flanking (Lo and Hi) gaps for each hit in HitToLo/HiHole[]

The Hole/HitToLo/HiHole[] vectors are used for O(1) lookups.

- For each hit, if it has a Lo gap w.r.t. both genomes, then it has a hole
  at its Lo end, the coordinates of the hole are the coordinates of the gaps.

- For each hit, do the same for Hi gaps, except there is now an additional
  check for duplication (in the typical case, a Hi hole will be the Lo hole
  for the following hit, in which case no new hole is added).
***/
void MakeHoles4(const AIVec &InHits, IntVec &HitToLoHole, IntVec &HitToHiHole,
  IntVec &HoleToLoHit, IntVec &HoleToHiHit, AIVec &HoleAIs)
	{
#if	TRACE
	{
	Log("\n");
	Log("MakeHoles4\n");
	Log("InHits:\n");
	for (size_t i = 0; i < InHits.size(); ++i)
		{
		Log("[%3d]  ", i);
		LogAI(InHits[i]);
		Log("\n");
		}
	Log("\n");
	}
#endif

	HoleAIs.clear();

	HitToLoHole.clear();
	HitToHiHole.clear();

	HoleToLoHit.clear();
	HoleToHiHit.clear();

	unsigned GenomeIndexA = GetFilterGenomeIndexA();
	unsigned GenomeIndexB = GetFilterGenomeIndexB();

	const ChrInfo &CIA = GetFilterChrInfoA();
	const ChrInfo &CIB = GetFilterChrInfoB();

	coord_t ChrLengthA = CIA.GetLength();
	coord_t ChrLengthB = CIB.GetLength();

	const unsigned InAICount = (unsigned) InHits.size();

	if (InAICount == 0)
		return;

	const unsigned BPCount = 2*InAICount;

	std::vector<AIBP> BPAs;
	std::vector<AIBP> BPBs;

	GetBendpoints(InHits, GenomeIndexA, BPAs, 0);
	GetBendpoints(InHits, GenomeIndexB, BPBs, 0);

	unsigned BPCountA = (unsigned) BPAs.size();
	unsigned BPCountB = (unsigned) BPBs.size();

	if (BPCountA != BPCount || BPCountB != BPCount)
		Die("MakeHoles4, BPCount0=%u, BPCount1=%u, BPCount=%u, input is probably not stripped",
		  BPCountA, BPCountB, BPCount);

	std::sort(BPAs.begin(), BPAs.end());
	std::sort(BPBs.begin(), BPBs.end());

#if	TRACE
	{
	Log("BPAs:\n");
	for (size_t i = 0; i < BPAs.size(); ++i)
		{
		Log("[%3d]  ", i);
		LogBP(BPAs[i]);
		Log("\n");
		}
	Log("\n");
	Log("BPBs:\n");
	for (size_t i = 0; i < BPBs.size(); ++i)
		{
		Log("[%3d]  ", i);
		LogBP(BPBs[i]);
		Log("\n");
		}
	Log("\n");
	}
#endif

// Map AI (hit) index to index of gap at lo / hi end of hit.
	IntVec AIToLoGapA;
	IntVec AIToLoGapB;
	IntVec AIToHiGapA;
	IntVec AIToHiGapB;

	for (unsigned AIIndex = 0; AIIndex < InAICount; ++AIIndex)
		{
		AIToHiGapA.push_back(UINT_MAX);
		AIToHiGapB.push_back(UINT_MAX);
		AIToLoGapA.push_back(UINT_MAX);
		AIToLoGapB.push_back(UINT_MAX);

		HitToLoHole.push_back(UINT_MAX);
		HitToHiHole.push_back(UINT_MAX);
		}

	for (unsigned BPIndex = 0; BPIndex < BPCount; ++BPIndex)
		{
		HoleToLoHit.push_back(UINT_MAX);
		HoleToHiHit.push_back(UINT_MAX);
		}

	std::vector<GapInfo> GapsA;
	std::vector<GapInfo> GapsB;

#if	TRACE
	Log("GetGaps A:\n");
#endif
	GetGaps(BPAs, ChrLengthA, GapsA, AIToLoGapA, AIToHiGapA);
#if	TRACE
	Log("GetGaps B:\n");
#endif
	GetGaps(BPBs, ChrLengthB, GapsB, AIToLoGapB, AIToHiGapB);

#if	TRACE
	{
	Log("\n");
	Log("GapsA:\n");
	for (size_t i = 0; i < GapsA.size(); ++i)
		{
		Log("[%3d]  ", i);
		LogGap(GapsA[i]);
		Log("\n");
		}
	Log("GapsB:\n");
	for (size_t i = 0; i < GapsB.size(); ++i)
		{
		Log("[%3d]  ", i);
		LogGap(GapsB[i]);
		Log("\n");
		}
	Log("\n");
	Log("AIToLoGapA:\n");
	for (size_t i = 0; i < AIToLoGapA.size(); ++i)
		Log("[%3d]=%d  \n", i, AIToLoGapA[i]);
	Log("\n");
	Log("AIToLoGapB:\n");
	for (size_t i = 0; i < AIToLoGapB.size(); ++i)
		Log("[%3d]=%d  \n", i, AIToLoGapB[i]);
	Log("\n");
	Log("AIToHiGapA:\n");
	for (size_t i = 0; i < AIToHiGapA.size(); ++i)
		Log("[%3d]=%d  \n", i, AIToHiGapA[i]);
	Log("\n");
	Log("AIToHiGapB:\n");
	for (size_t i = 0; i < AIToHiGapB.size(); ++i)
		Log("[%3d]=%d  \n", i, AIToHiGapB[i]);
	}
#endif
	
	assert(AIToLoGapA.size() == (size_t) InAICount);
	assert(AIToLoGapB.size() == (size_t) InAICount);
	assert(AIToHiGapA.size() == (size_t) InAICount);
	assert(AIToHiGapB.size() == (size_t) InAICount);

// For each endpoint of each hit, add hole if there is one.
	for (unsigned AIIndex = 0; AIIndex < InAICount; ++AIIndex)
		{
#if	TRACE
		Log("\n");
		Log("=========\n");
		Log("AI %d ", AIIndex);
		Log("\n");
		Log("=========\n");
		Log("LO\n");
		Log("--\n");
		LogAI(InHits[AIIndex]);
		Log("\n");
		Log("HitToLoHole=%d\n", HitToLoHole[AIIndex]);
#endif

		assert(HitToLoHole[AIIndex] == UINT_MAX);

		unsigned LoGapIndexA = AIToLoGapA[AIIndex];
		unsigned LoGapIndexB = AIToLoGapB[AIIndex];

#if	TRACE
		Log("LoGapIndexA=%d LoGapIndexB=%d\n", LoGapIndexA, LoGapIndexB);
#endif

	// Is there a gap in both genomes?
		if (LoGapIndexA == UINT_MAX || LoGapIndexB == UINT_MAX)
			continue;

		const GapInfo &LoGapA = GapsA[LoGapIndexA];
		const GapInfo &LoGapB = GapsB[LoGapIndexB];

		if (LoGapA.HiAIIndex != AIIndex)
			Die("GapsA[%u].HiAIIndex=%d != AIIndex=%u",
			  LoGapIndexA, LoGapA.HiAIIndex, AIIndex);
		if (LoGapB.HiAIIndex != AIIndex)
			Die("GapsB[%u].HiAIIndex=%d != AIIndex=%u",
			  LoGapIndexB, LoGapB.HiAIIndex, AIIndex);

#if	TRACE
		Log("LoGapA=");
		LogGap(LoGapA);
		Log("\n");
		Log("LoGapB=");
		LogGap(LoGapB);
		Log("\n");
#endif

	// Create hole
		AlnInfo HoleAI;

		FilterMakePairwiseAlnInfo(LoGapA.lo, LoGapA.hi, LoGapB.lo, LoGapB.hi, true,
		  HoleAI);

		HoleAI.Attrs.push_back(ALN_ATTR_TraceType);
		HoleAI.Values.push_back(TraceType_Hole);

		unsigned HoleIndex = (unsigned) HoleAIs.size();
		HoleAIs.push_back(HoleAI);

		HitToLoHole[AIIndex] = HoleIndex;
		HoleToHiHit[HoleIndex] = AIIndex;

#if	TRACE
		Log("Add hole [%u] = ", HoleIndex);
		LogAI(HoleAI);
		Log("\n");
		Log("HitToLoHole[%d]=%d\n", AIIndex, HitToLoHole[AIIndex]);
		Log("HoleToHiHit[%d]=%d\n", HoleIndex, HoleToHiHit[HoleIndex]);
#endif

	// Is predecessor w.r.t. genome 0 same as predecessor
	// w.r.t genome 1? If so, hole joins the two hits.
		unsigned PredA = LoGapA.LoAIIndex;
		unsigned PredB = LoGapB.LoAIIndex;
#if	TRACE
		Log("PredA=%d PredB=%d\n", PredA, PredB);
#endif
		if (PredA == PredB && PredA != UINT_MAX)
			{
			HoleToLoHit[HoleIndex] = PredA;
			HitToHiHole[PredA] = HoleIndex;
#if	TRACE
			Log("HoleToLoHit[%d] = %d\n", HoleIndex, PredA);
#endif
			}
		else
			{
#if	TRACE
			Log("PredA != PredB, not joined\n");
#endif
			;
			}
		}

	for (unsigned AIIndex = 0; AIIndex < InAICount; ++AIIndex)
		{
	// Hole for hi end of hit
	// Has hole already been created (due to lo end of another hit?)
#if	TRACE
		Log("\n");
		Log("HI\n");
		Log("--\n");
		Log("=========\n");
		Log("AI %d ", AIIndex);
		Log("\n");
		Log("=========\n");
		Log("HitToHiHole=%d\n", HitToHiHole[AIIndex]);
#endif
		if (HitToHiHole[AIIndex] != UINT_MAX)
			continue;

		unsigned HiGapIndexA = AIToHiGapA[AIIndex];
		unsigned HiGapIndexB = AIToHiGapB[AIIndex];

#if	TRACE
		Log("HiGapIndexA=%d HiGapIndexB=%d\n", HiGapIndexA, HiGapIndexB);
#endif

	// Is there a gap in both genomes?
		if (HiGapIndexA == UINT_MAX || HiGapIndexB == UINT_MAX)
			continue;

		const GapInfo &HiGapA = GapsA[HiGapIndexA];
		const GapInfo &HiGapB = GapsB[HiGapIndexB];

#if TRACE
		Log("HiGapA=");
		LogGap(HiGapA);
		Log("\n");
		Log("HiGapB=");
		LogGap(HiGapB);
		Log("\n");
#endif

		assert(HiGapA.LoAIIndex == AIIndex);
		assert(HiGapB.LoAIIndex == AIIndex);

	// Create hole
		AlnInfo HoleAI;

		FilterMakePairwiseAlnInfo(HiGapA.lo, HiGapA.hi, HiGapB.lo, HiGapB.hi, true,
		  HoleAI);

		HoleAI.Attrs.push_back(ALN_ATTR_TraceType);
		HoleAI.Values.push_back(TraceType_Hole);

		unsigned HoleIndex = (unsigned) HoleAIs.size();
		HoleAIs.push_back(HoleAI);

		HitToHiHole[AIIndex] = HoleIndex;
		HoleToLoHit[HoleIndex] = AIIndex;

#if	TRACE
		Log("Add hole %u ", HoleIndex);
		LogAI(HoleAI);
		Log("\n");
		Log("HitToHiHole[%d]=%d\n", AIIndex, HitToHiHole[AIIndex]);
		Log("HoleToLoHit[%d]=%d\n", HoleIndex, HoleToLoHit[HoleIndex]);
#endif

	// Is successor w.r.t. genome 0 same as successor
	// w.r.t genome 1? If so, hole joins the two hits.
		unsigned Succ0 = HiGapA.HiAIIndex;
		unsigned Succ1 = HiGapB.HiAIIndex;
#if	TRACE
		Log("Succ0=%d Succ1=%d\n", Succ0, Succ1);
#endif
		if (Succ0 == Succ1 && Succ0 != UINT_MAX)
			{
			HoleToHiHit[HoleIndex] = Succ0;
			HitToLoHole[Succ0] = HoleIndex;
#if	TRACE
			Log("HoleToHiHit[%d] = %d\n", HoleIndex, Succ0);
#endif
			}
		else
			{
			;
#if	TRACE
			Log("Succ0 != Succ1, not joined\n");
#endif
			}
		}

#if	TRACE
	{
	Log("\n");
	Log("Result of MakeHoles4:\n");
	Log("\n");
	Log("  Hit  PHole  NHole\n");
	Log("=====  =====  =====\n");
	for (unsigned AIIndex = 0; AIIndex < InAICount; ++AIIndex)
		{
		unsigned PrevHoleIndex = HitToLoHole[AIIndex];
		unsigned NextHoleIndex = HitToHiHole[AIIndex];

		Log("%5u", AIIndex);
		Log("  %5s", UIntToStr(PrevHoleIndex));
		Log("  %5s", UIntToStr(NextHoleIndex));
		Log("\n");
		}

	Log("\n");
	Log(" Hole   PHit   NHit\n");
	Log("=====  =====  =====\n");
	unsigned HoleCount = (unsigned) HoleAIs.size();
	for (unsigned HoleIndex = 0; HoleIndex < HoleCount; ++HoleIndex)
		{
		unsigned PrevHitIndex = HoleToLoHit[HoleIndex];
		unsigned NextHitIndex = HoleToHiHit[HoleIndex];

		Log("%5u", HoleIndex);
		Log("  %5s", UIntToStr(PrevHitIndex));
		Log("  %5s", UIntToStr(NextHitIndex));
		Log("\n");
		}
	}
#endif

// Validate
	for (unsigned AIIndex = 0; AIIndex < InAICount; ++AIIndex)
		{
		unsigned PrevHoleIndex = HitToLoHole[AIIndex];
		unsigned NextHoleIndex = HitToHiHole[AIIndex];
		if (PrevHoleIndex != UINT_MAX)
			{
			if (HoleToHiHit[PrevHoleIndex] != AIIndex)
				Die("HoleToHiHit[PrevHoleIndex=%u] = %u, != AIIndex = %u",
				  PrevHoleIndex, HoleToHiHit[PrevHoleIndex], AIIndex);
			}
		if (NextHoleIndex != UINT_MAX)
			{
			if (HoleToLoHit[NextHoleIndex] != AIIndex)
				Die("HoleToLoHit[NextHoleIndex=%u] = %u, != AIIndex = %u",
				  PrevHoleIndex, HoleToLoHit[NextHoleIndex], AIIndex);
			}
		}

	const unsigned HoleCount = (unsigned) HoleAIs.size();
	for (unsigned HoleIndex = 0; HoleIndex < HoleCount; ++HoleIndex)
		{
		unsigned PrevHitIndex = HoleToLoHit[HoleIndex];
		unsigned NextHitIndex = HoleToHiHit[HoleIndex];
		if (PrevHitIndex != UINT_MAX)
			{
			if (HitToHiHole[PrevHitIndex] != HoleIndex)
				Die("HitToHiHole[PrevHitIndex=%u]=%u != HoleIndex=%u",
				  PrevHitIndex, HitToHiHole[PrevHitIndex], HoleIndex);
			}
		if (NextHitIndex != UINT_MAX)
			{
			if (HitToLoHole[NextHitIndex] != HoleIndex)
				Die("HitToLoHole[PrevHitIndex=%u]=%u != HoleIndex=%u",
				  NextHitIndex, HitToLoHole[NextHitIndex], HoleIndex);
			}
		}
	}
