#include "utils.h"
#include "getopts.h"
#include "fastastr.h"
#include "masktab.h"
#include "progress.h"
#include "xlat.h"

void MaskSeq(std::vector<unsigned char> &Seq, ALPHA SeqAlpha, const MaskTable &MT);
void WriteFasta(FILE *f, const char *SeqData, unsigned Length);

bool g_HardMask;
bool g_Overwrite;
std::vector<bool> g_Pattern;
bool g_PatternIsMiddleBase = true;

static bool PatternIsMiddleBase(const std::vector<bool> &Pattern)
	{
	const unsigned PatternLength = (unsigned) Pattern.size();
	if (PatternLength%2 == 0)
		return false;

	unsigned MiddlePos = PatternLength/2;
	for (unsigned i = 0; i < PatternLength; ++i)
		{
		if (i == MiddlePos)
			{
			if (!Pattern[i])
				return false;
			}
		else
			{
			if (Pattern[i])
				return false;
			}
		}
	return true;
	}

static void SetPattern(unsigned WordSize)
	{
	g_Pattern.resize(WordSize);
	for (unsigned i = 0; i < WordSize; ++i)
		g_Pattern[i] = false;
	const char *Pattern = ValueOpt("pattern");
	if (Pattern == 0)
		{
		unsigned Middle = WordSize/2;
		g_Pattern[Middle] = true;
		g_PatternIsMiddleBase = true;
		}
	else
		{
		unsigned PatternLength = (unsigned) strlen(Pattern);
		if (PatternLength != WordSize)
			Die("Pattern length %u different from word size %u",
			  PatternLength, WordSize);
		unsigned TrueCount = 0;
		for (unsigned i = 0; i < WordSize; ++i)
			{
			char c = Pattern[i];
			if (c == '1')
				{
				g_Pattern[i] = true;
				++TrueCount;
				}
			else if (c == '0')
				g_Pattern[i] = false;
			else
				Die("Invalid character '%c' in pattern, must be 1s and 0s", c);
			}
		if (TrueCount == 0)
			Die("Pattern is all zeroes");

		g_PatternIsMiddleBase = PatternIsMiddleBase(g_Pattern);
		}
	}

void DoMask(const char *InFileName, const char *OutFileName,
  const char *TabFileName, ALPHA SeqAlpha)
	{
	g_HardMask = FlagOpt("hardmask");
	g_Overwrite = FlagOpt("overwrite");

	Progress("Reading mask table file %s", TabFileName);

	MaskTable MT;
	MT.FromFile(TabFileName);

	const unsigned WordSize = MT.GetWordSize();
	SetPattern(WordSize);

	FastaStream FS;
	FS.Open(InFileName, WordSize);

	FILE *fOut = CreateStdioFile(OutFileName);

	std::vector<char> Label;
	std::vector<unsigned char> Seq;

	uint64 GenomeLength = 1;
	ProgressStart("Writing %s", OutFileName);
	unsigned Steps = (unsigned) (FS.GetFileSize()/1000000 + 1);
	unsigned PrevStep = UINT_MAX;
	for (;;)
		{
		bool Ok = FS.GetNextSeq(Label, Seq);
		if (!Ok)
			break;

		unsigned Step = (unsigned) (FS.GetFilePos()/1000000);
		if (Step != PrevStep)
			{
			ProgressStep(Step, Steps);
			PrevStep = Step;
			}

		const char *ptrLabel = (char *) &(Label.front());
		WriteStdioFileStr(fOut, ">");
		WriteStdioFile(fOut, ptrLabel, (off_t) Label.size());
		WriteStdioFileStr(fOut, "\n");

		MaskSeq(Seq, SeqAlpha, MT);

		const char *ptrSeq = (char *) &(Seq.front());
		WriteFasta(fOut, ptrSeq, (unsigned) Seq.size());

		GenomeLength += (uint64) Seq.size();
		}
	ProgressDone();

	CloseStdioFile(fOut);
	fOut = 0;

	FS.Close();
	}

void MaskTab()
	{
	const char *InFileName = RequiredValueOpt("masktab");
	const char *OutFileName = RequiredValueOpt("out");
	const char *TabFileName = RequiredValueOpt("tab");

	bool Amino = FlagOpt("aminoseq");
	ALPHA SeqAlpha = (Amino ? ALPHA_Amino : ALPHA_Nucleo);

	DoMask(InFileName, OutFileName, TabFileName, SeqAlpha);
	}
