#include "utils.h"
#include "revfile.h"
#include "progress.h"
#include "getopts.h"
#include <string>

using namespace std;

void ParseStrVec(const string &In, vector<string> &Out, char c = ',');

static vector<RevFile> g_RevFiles;

// Merge a set of rev files each containing an alignment
// (genA) <-> (genB,chrB) where there is exactly one
// unique chromosome chrB in each file.
void MergeChrs()
	{
	const char *cstrRevFileNames = RequiredValueOpt("mergechrs");
	const char *OutFileName = RequiredValueOpt("out");

	vector<string> RevFileNames;
	ParseStrVec(cstrRevFileNames, RevFileNames);

	if (RevFileNames.empty())
		Die("No rev files specified");

	const unsigned RevFileCount = (unsigned) RevFileNames.size();

	g_RevFiles.resize(RevFileCount);
	for (unsigned RevFileIndex = 0; RevFileIndex < RevFileCount; ++RevFileIndex)
		{
		RevFile &RevFile = g_RevFiles[RevFileIndex];
		const string &FileName = RevFileNames[RevFileIndex];
		Progress("Opening %.32s", FileName.c_str());
		RevFile.Open(FileName);

		unsigned GenomeCount = RevFile.GetGenomeCount();
		if (GenomeCount != 2)
			Die("Files must be pairwise");

		const GenInfo &GI0_0 = g_RevFiles[0].GetGenomeInfo(0);
		const GenInfo &GI0_i = g_RevFiles[RevFileIndex].GetGenomeInfo(0);
		if (GI0_i != GI0_0)
			Die("Genome %.32s must be the same in each file", GI0_0.GetName().c_str());

		const GenInfo &GI1_i = g_RevFiles[RevFileIndex].GetGenomeInfo(1);
		unsigned ChrCount = GI1_i.GetChrCount();
		if (ChrCount != 1)
			Die("File %.32s has %u chromosomes in 2nd genome, must be exactly one",
			  FileName.c_str(), ChrCount);
		}

	for (unsigned RevFileIndex1 = 0; RevFileIndex1 < RevFileCount; ++RevFileIndex1)
		{
		const RevFile &RevFile1 = g_RevFiles[RevFileIndex1];
		const GenInfo &GI1_1 = RevFile1.GetGenomeInfo(1);
		const string &ChrName1_1 = GI1_1.GetChrName(0);

		for (unsigned RevFileIndex2 = RevFileIndex1 + 1; RevFileIndex2 < RevFileCount;
		  ++RevFileIndex2)
			{
			const RevFile &RevFile2 = g_RevFiles[RevFileIndex2];
			const GenInfo &GI2_1 = RevFile2.GetGenomeInfo(1);
			const string &ChrName2_1 = GI2_1.GetChrName(0);
			if (ChrName1_1 == ChrName2_1)
				Die("Chr name %.32s in two different files", ChrName1_1.c_str());
			}
		}

	const RevFile &Rev0 = g_RevFiles[0];

	GenInfoVec OutGenomes(2);
	OutGenomes[0] = Rev0.GetGenomeInfo(0);
	GenInfo &CombinedGI1 = OutGenomes[1];

	CombinedGI1.Create(Rev0.GetGenomeName(1));

	for (unsigned RevFileIndex = 0; RevFileIndex < RevFileCount; ++RevFileIndex)
		{
		const RevFile &RevFile = g_RevFiles[RevFileIndex];
		const ChrInfo &Chr1 = RevFile.GetChrInfo(1, 0);
		CombinedGI1.AddChr(Chr1, RevFileIndex);
		}

	Log("\n");
	Log("CombinedGI0:\n");
	OutGenomes[0].LogMe();
	Log("\n");
	Log("CombinedGI1:\n");
	OutGenomes[1].LogMe();

	RevFile Out;
	Out.Create(OutFileName, OutGenomes);

	for (unsigned RevFileIndex = 0; RevFileIndex < RevFileCount; ++RevFileIndex)
		{
		const RevFile &Rev = g_RevFiles[RevFileIndex];

		AIXVec AIXs;
		Rev.GetAllAlnInfoXs(AIXs);

		ProgressStart("Copying file %u", RevFileIndex);
		const unsigned AICount = SIZE(AIXs);
		for (unsigned i = 0; i < AICount; ++i)
			{
			if (i%100 == 0)
				ProgressStep(i, AICount);
			AlnInfoX &AIX = AIXs[i];
			AlnInfo &AI = AIX.AI;
			GCSegVec &Segs = AI.Segs;
			const unsigned SegCount = SIZE(Segs);
			for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
				{
				GCSeg &Seg = Segs[SegIndex];
				if (Seg.GenomeIndex == 1)
					{
					asserta(Seg.ChrId == 0);
					Seg.ChrId = RevFileIndex;
					}
				}
			unsigned AlnIndex = AIX.AlnIndex;
			if (AlnIndex == UINT_MAX)
				Out.AddAlnInfo(AI);
			else
				{
				StdStrVec Rows;
				Rev.GetAlnRows(AlnIndex, Rows);
				Out.AddAln(AI, Rows);
				}
			}
		ProgressDone();
		}

	Out.Close();
	}
