#include "nw.h"

#undef DPM
#undef DPD
#undef DPI

#define	TRACE	0

void Band(unsigned i, unsigned LA, unsigned LB, unsigned r,
   unsigned *ptrMidj, unsigned *ptrMinJ, unsigned *ptrMaxJ)
	{
	const float Gradient = float(LB-1)/float(LA-1);
	unsigned midj = unsigned(float(i)*Gradient + 0.5);
	assert(midj < LB);

// Typical case not near top or bottom of band
	unsigned minJ = 0;
	unsigned maxJ = 2*r;

// Exception near top (small i)
	if (midj < r)
		minJ = r - midj;

// Exception near bottom (large i)
	if (midj + r > LB - 1)
		maxJ = r + (LB - 1 - midj);

	assert(minJ <= 2*r);
	assert(maxJ <= 2*r);
	assert(minJ <= maxJ);

	*ptrMidj = midj;
	*ptrMinJ = minJ;
	*ptrMaxJ = maxJ;
	}

#if	TRACE
static void LogDP(score_t **DPM, const unsigned char *A, 
  const unsigned char *B, unsigned LA, unsigned LB, unsigned r)
	{
	Log("        ");
	for (unsigned k = 0; k < LB; ++k)
		{
		char c = B[k];
		Log(" %4u:%c", k, c);
		}
	Log("\n");
	for (unsigned i = 0; i < LA; ++i)
		{
		unsigned midj;
		unsigned minJ;
		unsigned maxJ;
		Band(i, LA, LB, r, &midj, &minJ, &maxJ);
// j = midj + J - r
// J = j + r - midj
		unsigned minj = (midj + minJ) - r;
		unsigned maxj = (midj + maxJ) - r;

		char c = A[i];
		Log("%4u:%c  ", i, c);
		for (unsigned j = 0; j < minj; ++j)
			Log("       ");
		for (unsigned j = minj; j <= maxj; ++j)
			{
			unsigned J = (j + r) - midj;
			assert(J <= 2*r);
			Log(" %s", ScoreToStr(DPM[i][J]));
			}
		Log("\n");
		}
	}
#endif	// TRACE

score_t NWBanded(const unsigned char *A, unsigned LA, const unsigned char *B,
  unsigned LB, unsigned r, std::string &Path)
	{
	Path.clear();
	if (LA == 0)
		{
		for (unsigned i = 0; i < LB; ++i)
			Path.push_back('I');
		return 0;
		}
	if (LB == 0)
		{
		for (unsigned i = 0; i < LA; ++i)
			Path.push_back('D');
		return 0;
		}

// Allocate DP matrices
	const unsigned Diameter = 2*r + 1;
	const size_t BandArea = LA*Diameter;

// Allocate one buffer for each matrix.
// This is to avoid too many calls to new, which
// is too slow in MSC.
	score_t *DPMBuffer = myalloc(score_t, BandArea);
	score_t *DPDBuffer = myalloc(score_t, BandArea);
	score_t *DPIBuffer = myalloc(score_t, BandArea);

// One pointer for each row.
	score_t **DPM = myalloc(score_t *, LA);
	score_t **DPD = myalloc(score_t *, LA);
	score_t **DPI = myalloc(score_t *, LA);

// Initialize row pointers
	for (unsigned i = 0; i < LA; ++i)
		{
		DPM[i] = DPMBuffer + i*Diameter;
		DPD[i] = DPDBuffer + i*Diameter;
		DPI[i] = DPIBuffer + i*Diameter;
		}

#ifndef NDEBUG
	for (unsigned i = 0; i < LA; ++i)
		for (unsigned J = 0; J < Diameter; ++J)
			{
			DPM[i][J] = SCORE_UNINIT;
			DPD[i][J] = SCORE_UNINIT;
			DPI[i][J] = SCORE_UNINIT;
			}
#endif

// ==========================================
// Main DP loop
// i is index into A = DP matrix row.
// j is index into B, J is index into DP row,
// j=0..LB-1, J=0..Diameter-1.
// Midpoint of band is midj, which is a
// function of i following the diagonal as
// closely as possible with integer rounding.
// Conversion is j = midj + J - r
// Inverse       J = j + r - midj
// ==========================================
	unsigned Prev_midj = UINT_MAX;
	unsigned Prev_minJ = UINT_MAX;
	unsigned Prev_maxJ = UINT_MAX;

// RowShift is change in midj relative to previous row.
	unsigned RowShift = UINT_MAX;

// Prev_J means index into DP arrays of same
// column in previous row.
	unsigned Prev_J = UINT_MAX;

#if	TRACE
	Log("  i  midj  minJ  maxJ   Shift  Prev_midj  Prev_minJ  Prev_maxJ\n");
	Log("===  ====  ====  ====  ======  =========  =========  =========\n");
#endif
	for (unsigned i = 0; i < LA; ++i)
		{
		const unsigned char cA = A[i];

		unsigned midj;
		unsigned minJ;
		unsigned maxJ;
		Band(i, LA, LB, r, &midj, &minJ, &maxJ);
		if (i > 0)
			{
			assert(midj >= Prev_midj);
			RowShift = midj - Prev_midj;
			}
#if	TRACE
		Log("%3u  %4u  %4u  %4u  %6u  %9u  %9u  %9u\n",
		  i, midj, minJ, maxJ, RowShift, Prev_midj, Prev_minJ, Prev_maxJ);
#endif

		for (unsigned J = minJ; J <= maxJ; ++J)
			{
			const unsigned j = midj + J - r;
			assert(j < LB);
			const unsigned char cB = B[j];

		// DPx[i-1][Prev_J] is same column in full DP matrix
		// as DPx[i-1][J].
			Prev_J = J + RowShift;

			{
		// Match M=LetterA+LetterB
			score_t MM = MINUS_INF;
			score_t DM = MINUS_INF;
			score_t IM = MINUS_INF;
			if (i == 0)
				{
				if (j == 0)
					MM = 0;
				else
					IM = TERMGAP + (j - 1)*GAPEXTEND;
				}
			else
				{
				if (j > 0)
					{
				// This test is Prev_J - 1 >= Prev_minJ, move 1
				// to rhs because unsigned.
					if (Prev_J >= Prev_minJ + 1)
						{
					// move 1 to rhs here also.
						if (Prev_J <= Prev_maxJ + 1)
							{
							MM = DPM[i-1][Prev_J - 1];
							DM = DPD[i-1][Prev_J - 1];
							IM = DPI[i-1][Prev_J - 1];
							}
						}
					}
				else
					{
					assert(i > 0 && j == 0);
					DM = TERMGAP + (i - 1)*GAPEXTEND;
					}
				}

			score_t Best = MM;
			if (DM > Best)
				Best = DM;
			if (IM > Best)
				Best = IM;
			DPM[i][J] = Best + SUBST(cA, cB);
			}

			{
		// Delete D=LetterA+GapB
			score_t MD = MINUS_INF;
			score_t DD = MINUS_INF;
			if (i > 0)
				{
				if (Prev_J <= Prev_maxJ)
					{
					assert(Prev_J <= Prev_maxJ);
					assert(Prev_J >= Prev_minJ);
					MD = DPM[i-1][Prev_J] + (j == LB - 1 ? TERMGAP : GAPOPEN);
					DD = DPD[i-1][Prev_J] + GAPEXTEND;
					}
				}
			DPD[i][J] = std::max(MD, DD);
			}

		// Insert I=GapA+LetterB
			{
			score_t MI = MINUS_INF;
			score_t II = MINUS_INF;
			if (J > minJ)
				{
				MI = DPM[i][J-1] + (i == LA - 1 ? TERMGAP : GAPOPEN);
				II = DPI[i][J-1] + GAPEXTEND;
				}

			DPI[i][J] = std::max(MI, II);
			}
			}

		Prev_midj = midj;
		Prev_minJ = minJ;
		Prev_maxJ = maxJ;
		}

#if TRACE
	Log("DPM:\n");
	LogDP(DPM, A, B, LA, LB, r);
	Log("DPD:\n");
	LogDP(DPD, A, B, LA, LB, r);
	Log("DPI:\n");
	LogDP(DPI, A, B, LA, LB, r);
#endif

	score_t Score = TraceBackBanded(A, LA, B, LB, r, DPM, DPD, DPI, Path);

	myfree(DPMBuffer);
	myfree(DPDBuffer);
	myfree(DPIBuffer);

	myfree(DPM);
	myfree(DPD);
	myfree(DPI);

	return Score;
	}
