#include "utils.h"
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <list>

static time_t g_StartTimeStep;
static char *g_Msg = strsave("");
static bool g_NL = false;
bool g_ShowProgress = true;
static int g_ProgressDepth = 0;
static FILE *g_ProgFile;
static off64_t g_ProgFileSize;
static std::list<bool> g_ShowProgressStack;
static bool g_LogProgress = true;

void SetLogProgress(bool On)
	{
	g_LogProgress = On;
	}

void ShowProgress(bool Show)
	{
	g_ShowProgress = Show;
	}

void PushShowProgress(bool Show)
	{
	g_ShowProgressStack.push_front(g_ShowProgress);
	g_ShowProgress = Show;
	}

void PopShowProgress()
	{
	if (g_ShowProgressStack.empty())
		Die("PopShowProgress, no matching Push");
	g_ShowProgress = g_ShowProgressStack.front();
	g_ShowProgressStack.pop_front();
	}

const char *SecsToHHMMSS(int Secs)
	{
	int HH = Secs/3600;
	int MM = (Secs - HH*3600)/60;
	int SS = Secs%60;
	static char Str[16];
	if (HH == 0)
		sprintf(Str, "%02d:%02d", MM, SS);
	else
		sprintf(Str, "%02d:%02d:%02d", HH, MM, SS);
	return Str;
	}

static char *Resources()
	{
	static char Str[64];

	int ElapsedSecs = GetElapsedSecs();
	double MemUseMB = (GetMemUseBytes() + 500000.0)/1000000.0;

	sprintf(Str, "%s %4.0f Mb", SecsToHHMMSS(ElapsedSecs), MemUseMB);
	return Str;
	}

void SetProgressMsg(const char *Format, ...)
	{
	if (!IsTerminal(2))
		return;
	if (!g_ShowProgress)
		return;
	char Str[4096];
	va_list ArgList;
	va_start(ArgList, Format);
	vsprintf(Str, Format, ArgList);
	if (g_Msg != 0)
		{
		size_t OldMsgLength = strlen(g_Msg);
		size_t NewMsgLength = strlen(Str);
		if (OldMsgLength > NewMsgLength)
			{
			size_t Diff = OldMsgLength - NewMsgLength;
			char *Padding = (char *) malloc(Diff + 1);
			asserta(Padding != 0);
			memset(Padding, ' ', Diff);
			Padding[Diff] = 0;
			strcat(Str, Padding);
			free(Padding);
			}
		}
	free(g_Msg);
	g_Msg = strsave(Str);
	}

void ProgressStep(int StepIndex, int StepCount)
	{
	if (!g_ShowProgress)
		return;
	if (!IsTerminal(2))
		return;

	if (g_ProgressDepth != 1 || g_Msg == 0)
		Die("ProgressStep without ProgressStart");

	double Pct = (double) StepIndex*100.0 / (double) StepCount;

	fprintf(stderr, "\r%s %6.2f%%  %s", Resources(), Pct, g_Msg);
	g_NL = false;
	}

void ProgressStepFile()
	{
	if (!g_ShowProgress)
		return;
	if (!IsTerminal(2))
		return;

	if (g_ProgressDepth != 1 || g_Msg == 0)
		Die("ProgressStep without ProgressStart");

	double Pct = (double) GetStdioFilePos(g_ProgFile)*100.0 / (double) g_ProgFileSize;

	fprintf(stderr, "\r%s %6.2f%%  %s", Resources(), Pct, g_Msg);
	g_NL = false;
	}

void ProgressStartFile(FILE *f, const char *Format, ...)
	{
	if (!g_ShowProgress)
		return;
	if (!IsTerminal(2))
		return;

	g_ProgFile = f;
	g_ProgFileSize = GetStdioFileSize(f) + 1; // +1 to avoid div by zero

	if (g_ProgressDepth > 0)
		Die("Nested ProgressStart: '%s' '%s'", g_Msg, Format);
	++g_ProgressDepth;

	g_StartTimeStep = time(0);
	char Str[4096];
	va_list ArgList;
	va_start(ArgList, Format);
	vsprintf(Str, Format, ArgList);
	if (g_Msg != 0)
		free(g_Msg);
	g_Msg = strsave(Str);

	if (!g_NL)
		fprintf(stderr, "\n");
	ProgressStep(0, 1);
	}

void ProgressStart(const char *Format, ...)
	{
	if (!g_ShowProgress)
		return;
	if (!IsTerminal(2))
		return;

	if (g_ProgressDepth > 0)
		Die("Nested ProgressStart: '%s' '%s'", g_Msg, Format);
	++g_ProgressDepth;

	g_StartTimeStep = time(0);
	char Str[4096];
	va_list ArgList;
	va_start(ArgList, Format);
	vsprintf(Str, Format, ArgList);
	if (g_Msg != 0)
		free(g_Msg);
	g_Msg = strsave(Str);

	if (!g_ShowProgress)
		return;

	if (!g_NL)
		fprintf(stderr, "\n");
	ProgressStep(0, 1);
	}

void ProgressDone()
	{
	if (!g_ShowProgress)
		return;
	if (!IsTerminal(2))
		return;

	if (g_ProgressDepth != 1 || g_Msg == 0)
		Die("Unmatched call to ProgressDone");

	--g_ProgressDepth;

	if (0 == g_Msg)
		return;

	if (g_LogProgress)
		Log("%s %s\n", Resources(), g_Msg);

	fprintf(stderr, "\r%s %6.2f%%  %s\n", Resources(), 100.0, g_Msg);
	g_NL = true;

	free(g_Msg);
	g_Msg = 0;
	}

void Progress(const char *Format, ...)
	{
	if (!g_ShowProgress)
		return;
	if (!IsTerminal(2))
		return;

	char Str[4096];
	va_list ArgList;
	va_start(ArgList, Format);
	vsprintf(Str, Format, ArgList);
	if (g_LogProgress)
		Log("%s %s\n", Resources(), Str);

	if (g_ShowProgress)
		{
		if (!g_NL)
			fprintf(stderr, "\n");
		fprintf(stderr, "%s %s\n", Resources(), Str);
		g_NL = true;
		}
	}

void ProgressNoRes(const char *Format, ...)
	{
	if (!g_ShowProgress)
		return;
	if (!IsTerminal(2))
		return;

	char Str[4096];
	va_list ArgList;
	va_start(ArgList, Format);
	vsprintf(Str, Format, ArgList);

	if (g_LogProgress)
		Log("%s\n", Str);

	if (g_ShowProgress)
		{
		if (!g_NL)
			fprintf(stderr, "\n");
		fprintf(stderr, "%s\n", Str);
		g_NL = true;
		}
	}

unsigned GetProgressDepth()
	{
	return g_ProgressDepth;
	}

void ProgressExit()
	{
	if (g_LogProgress)
		{
		time_t t = time(0);
		Log("\n");
		Log("%s\n", Resources());
		Log("Total time %s\n", SecsToHHMMSS(GetElapsedSecs()));
		Log("Max %.0f Mb\n", GetPeakMemUseBytes()/1000000);
		Log("Finished %s", asctime(localtime(&t)));
		}
	}
