#include "utils.h"
#include "gff.h"
#include "getopts.h"
#include "progress.h"
#include "revfile.h"

static unsigned g_BlockLength = 80;

static char FeatureToChar(const string &Feature)
	{
	if (Feature == "CDS")
		return 'c';
	else if (Feature == "UTR")
		return 'u';
	else if (Feature == "NGE")
		return 'n';
	else if (Feature == "NXE")
		return 'x';
	//else if (Feature == "island")
	//	return 'i';
	//else if (Feature == "tandem")
	//	return 't';
	//else if (Feature == "")
	//	return ' ';
	else
		return ' ';
	}

void Project()
	{
	const char *InFileName = RequiredValueOpt("project");
	const char *Genome = ValueOpt("genome");
	const char *GFFFileName = ValueOpt("gff");
	const char *GFFFileName2 = ValueOpt("gff2");
	const char *SeqFileName = ValueOpt("seq1");
	bool DoRevComp = FlagOpt("revcomp");

	IntOpt("block", &g_BlockLength);

	RevFile In;
	In.Open(InFileName);

	unsigned GenomeIndex = 0;
	if (Genome != 0)
		GenomeIndex = In.GetGenomeIndex(Genome);

	vector<AlnInfo> AIs;
	In.GetAllAlnInfos(AIs);

	unsigned Lo = UINT_MAX;
	unsigned Hi = UINT_MAX;
	unsigned OtherLo = UINT_MAX;
	unsigned OtherHi = UINT_MAX;
	unsigned ChrId = UINT_MAX;
	unsigned OtherGenomeIndex = UINT_MAX;
	unsigned OtherChrId = UINT_MAX;
	const unsigned AICount = SIZE(AIs);
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		unsigned AlnIndex = In.GetAlnIndex(AIIndex);
		if (AlnIndex == UINT_MAX)
			continue;

		const AlnInfo &AI = AIs[AIIndex];
		const unsigned SegCount = AI.GetSegCount();
		if (SegCount != 2)
			Die("Must be strictly pairwise alignment");

		unsigned Count = 0;
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = AI.Segs[SegIndex];
			if (Seg.GenomeIndex == GenomeIndex)
				{
				++Count;
				if (Lo == UINT_MAX)
					{
					Lo = Seg.lo;
					Hi = Seg.hi;
					ChrId = Seg.ChrId;
					}
				else
					{
					Lo = min(Lo, Seg.lo);
					Hi = max(Hi, Seg.hi);
					if (Seg.ChrId != ChrId)
						Die("Must be unique chromosome");
					}
				}
			else
				{
				if (OtherLo == UINT_MAX)
					{
					OtherLo = Seg.lo;
					OtherHi = Seg.hi;
					}
				else
					{
					OtherLo = min(OtherLo, Seg.lo);
					OtherHi = max(OtherHi, Seg.hi);
					}
				if (OtherChrId == UINT_MAX)
					{
					OtherGenomeIndex = Seg.GenomeIndex;
					OtherChrId = Seg.ChrId;
					}
				else
					{
					if (Seg.GenomeIndex != OtherGenomeIndex)
						Die("Must be pair-wise unique genomes");
					if (Seg.ChrId != OtherChrId)
						Die("Must be pair-wise unique genomes");
					}
				}
			if (Count != 1)
				Die("Must be strictly pairwise alignment (%u segs)", Count);
			}
		}

	if (Lo == UINT_MAX)
		Die("No alignments");

	const unsigned L = Hi - Lo + 1;
	const unsigned OtherL = OtherHi - OtherLo + 1;

	const string &GenomeName = In.GetGenomeName(GenomeIndex);
	const string &OtherGenomeName = In.GetGenomeName(OtherGenomeIndex);
	const string &ChrName = In.GetChrName(GenomeIndex, ChrId);
	const string &OtherChrName = In.GetChrName(OtherGenomeIndex, OtherChrId);

	Log("\n");
	Log("Genome1   %s\n", GenomeName.c_str());
	Log("Chr1      %s\n", ChrName.c_str());
	Log("Genome2   %s\n", OtherGenomeName.c_str());
	Log("Chr2      %s\n", OtherChrName.c_str());
	Log("Lo1       %u\n", Lo);
	Log("Hi1       %u\n", Hi);
	Log("Length1   %u\n", Hi - Lo + 1);
	Log("Lo2       %u\n", OtherLo);
	Log("Hi2       %u\n", OtherHi);
	Log("Length2   %u\n", OtherHi - OtherLo + 1);

	//string Seq(L, '?');

	string Seq;
	if (SeqFileName == 0)
		Seq.resize(L, '?');
	else
		{
		RevFile SeqRev;
		SeqRev.Open(SeqFileName);
		unsigned GenomeIndexSeq = SeqRev.GetGenomeIndex(GenomeName);
		unsigned ChrIdSeq = SeqRev.GetChrId(GenomeIndex, ChrName);
		unsigned L = Hi - Lo + 1;
		SeqRev.GetSubSeq2(GenomeIndexSeq, ChrIdSeq, Seq, Lo, L);
		asserta(SIZE(Seq) == L);
		SeqRev.Close();
		}

	string OtherSeq(L, '-');

	vector<unsigned> OtherPosVec(L, UINT_MAX);
	vector<unsigned> OtherChrVec(L, UINT_MAX);
	vector<bool> OtherStrand(L, true);
	vector<string> Inserts(L);
	string Annots(L, '.');
	string OtherAnnots(OtherL, '.');

	if (GFFFileName != 0)
		{
		FILE *fGFF = OpenStdioFile(GFFFileName);
		GFFRecord Rec;
		while (GetNextGFFRecord(fGFF, Rec))
			{
			assert(Rec.Start > 0);
			for (unsigned Pos = Rec.Start - 1; Pos <= Rec.End - 1; ++Pos)
				{
				if (Pos < Lo || Pos > Hi)
					continue;
				Annots[Pos-Lo] = FeatureToChar(Rec.Feature);
				}
			}
		CloseStdioFile(fGFF);
		}

	if (GFFFileName2 != 0)
		{
		FILE *fGFF = OpenStdioFile(GFFFileName2);
		GFFRecord Rec;
		while (GetNextGFFRecord(fGFF, Rec))
			{
			assert(Rec.Start > 0);
			for (unsigned Pos = Rec.Start - 1; Pos <= Rec.End - 1; ++Pos)
				{
				if (Pos < OtherLo || Pos > OtherHi)
					continue;
				if (strlen(Rec.Feature) == 0)
					continue;
				OtherAnnots[Pos-OtherLo] = FeatureToChar(Rec.Feature);
				}
			}
		CloseStdioFile(fGFF);
		}

	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		unsigned AlnIndex = In.GetAlnIndex(AIIndex);
		if (AlnIndex == UINT_MAX)
			continue;

		const AlnInfo &AI = AIs[AIIndex];
		StdStrVec Rows;
		In.GetAlnRows(AlnIndex, Rows);

		unsigned SegIndex = AI.GetSegIndex(GenomeIndex);
		unsigned OtherSegIndex = (SegIndex == 0 ? 1 : 0);

		const GCSeg &Seg = AI.Segs[SegIndex];
		const GCSeg &OtherSeg = AI.Segs[OtherSegIndex];
		asserta(Seg.GenomeIndex == GenomeIndex);
		asserta(Seg.lo >= Lo && Seg.hi <= Hi);

		const string &Row = Rows[SegIndex];
		const string &OtherRow = Rows[OtherSegIndex];
		const unsigned ColCount = SIZE(Row);
		asserta(SIZE(OtherRow) == ColCount);

		unsigned Pos = (Seg.Plus ? Seg.lo : Seg.hi);
		unsigned OtherPos = (OtherSeg.Plus ? OtherSeg.lo : OtherSeg.hi);
		for (unsigned ColIndex = 0; ColIndex < ColCount; ++ColIndex)
			{
			char Letter = Row[ColIndex];
			char OtherLetter = OtherRow[ColIndex];

			if (OtherLetter == '-')
				{
				OtherPosVec[Pos-Lo] = UINT_MAX;
				OtherSeq[Pos-Lo] = '-';
				}
			else
				{
				OtherSeq[Pos-Lo] = OtherLetter;
				OtherPosVec[Pos-Lo] = OtherPos;
				OtherStrand[Pos-Lo] = (Seg.Plus == OtherSeg.Plus);
				}

			if (Letter != '-')
				{
				Seq[Pos-Lo] = Letter;
				if (Seg.Plus)
					++Pos;
				else
					--Pos;
				}

			if (OtherLetter != '-')
				{
				if (Letter == '-')
					Inserts[Pos-Lo].push_back(OtherLetter);

				if (OtherSeg.Plus)
					++OtherPos;
				else
					--OtherPos;
				}
			}
		}

	string Str;
	string OtherStr;
	string Sim;
	string AnnotsStr;
	string OtherAnnotsStr;

	Str[0]= Seq[0];
	OtherStr[0] = OtherSeq[0];
	for (unsigned i = 1; i < L; ++i)
		{
		const string &Ins = Inserts[i];
		if (Ins != "")
			{
			for (unsigned k = 0; k < SIZE(Ins); ++k)
				{
				Sim.push_back(' ');
				Str.push_back('-');
				OtherStr.push_back(Ins[k]);
				AnnotsStr.push_back(' ');
				OtherAnnotsStr.push_back(' ');
				}
			}

		unsigned OtherPos = OtherPosVec[i];

		if (OtherPosVec[i] != UINT_MAX && OtherPosVec[i-1] != UINT_MAX && 
		  OtherStrand[i] != OtherStrand[i-1])
			{
			Sim.push_back(' ');
			Str.push_back(' ');
			OtherStr.push_back(OtherStrand[i] ? '>' : '<');
			AnnotsStr.push_back(' ');
			OtherAnnotsStr.push_back(' ');
			}

		if (OtherStrand[i])
			{
			if (OtherPosVec[i] != UINT_MAX && OtherPosVec[i-1] + 1 != OtherPosVec[i])
				{
				Sim.push_back(' ');
				Str.push_back(' ');
				OtherStr.push_back('|');
				AnnotsStr.push_back(' ');
				OtherAnnotsStr.push_back(' ');
				}
			}
		else 
			{
			if (OtherPosVec[i] != UINT_MAX && OtherPosVec[i-1] != OtherPosVec[i] + 1)
				{
				Sim.push_back(' ');
				Str.push_back(' ');
				OtherStr.push_back('|');
				AnnotsStr.push_back(' ');
				OtherAnnotsStr.push_back(' ');
				}
			}

		char Letter = Seq[i];
		char OtherLetter = OtherSeq[i];
		Str.push_back(Letter);
		OtherStr.push_back(OtherLetter);
		AnnotsStr.push_back(Annots[i]);

		if (OtherPos == UINT_MAX)
			OtherAnnotsStr.push_back(' ');
		else
			OtherAnnotsStr.push_back(OtherAnnots[OtherPos-OtherLo]);

		if (Letter != '-' && toupper(Letter) == toupper(OtherLetter))
			Sim.push_back('*');
		else
			Sim.push_back(' ');
		}

	const unsigned N = SIZE(Str);
	asserta(SIZE(OtherStr) == N);
	asserta(SIZE(Sim) == N);
	asserta(SIZE(AnnotsStr) == N);
	asserta(SIZE(OtherAnnotsStr) == N);

	Log("\n");
	unsigned BlockCount = (N + g_BlockLength - 1)/g_BlockLength;
	string GenDotChr = GenomeName + string(" ") + ChrName;
	string OtherGenDotChr = OtherGenomeName + string(" ") + OtherChrName;
	unsigned MaxNameLen = max(SIZE(GenDotChr), SIZE(OtherGenDotChr));

	if (DoRevComp)
		{
		void RevComp(string &Seq);
		RevComp(Str);
		RevComp(OtherStr);
		reverse(AnnotsStr.begin(), AnnotsStr.end());
		reverse(OtherAnnotsStr.begin(), OtherAnnotsStr.end());
		}

	for (unsigned BlockIndex = 0; BlockIndex < BlockCount; ++BlockIndex)
		{
		unsigned Start = BlockIndex*g_BlockLength;
		unsigned End = (BlockIndex + 1)*g_BlockLength;
		if (End > N)
			End = N;

		//Log("%*.*s  ", MaxNameLen, MaxNameLen, "");
		//for (unsigned i = Start; i < End; ++i)
			//Log("%c", Sim[i]);
		//Log("\n");
		//

		for (unsigned i = Start; i < End; ++i)
			{
			char c = Str[i];
			char c2 = OtherStr[i];
			c = tolower(c);
			c2 = tolower(c2);
			if (isalpha(c) && isalpha(c2) && c == c2)
				{
				c = toupper(c);
				c2 = toupper(c2);
				}
			Str[i] = c;
			OtherStr[i] = c2;
			}

		Log("%*.*s  ", MaxNameLen, MaxNameLen, OtherGenDotChr.c_str());
		for (unsigned i = Start; i < End; ++i)
			Log("%c", OtherStr[i]);
		Log("\n");

		Log("%*.*s  ", MaxNameLen, MaxNameLen, GenDotChr.c_str());
		for (unsigned i = Start; i < End; ++i)
			Log("%c", Str[i]);
		Log("\n");

		Log("%*.*s  ", MaxNameLen, MaxNameLen, OtherGenDotChr.c_str());
		for (unsigned i = Start; i < End; ++i)
			Log("%c", OtherAnnotsStr[i]);
		Log("\n");

		Log("%*.*s  ", MaxNameLen, MaxNameLen, GenDotChr.c_str());
		for (unsigned i = Start; i < End; ++i)
			Log("%c", AnnotsStr[i]);
		Log("\n");
		Log("\n");
		}

	Log("\n");
	}
