#include <math.h>
#include <stdlib.h>
#include <vector>
#include <algorithm>
#include <time.h>
#include "getopts.h"
#include "utils.h"
#include "pdf.h"

#ifdef _MSC_VER
#include <process.h>
#else
#include <unistd.h>
#endif

#undef RandInt
#undef RandInt64
#undef RandExp
#undef RandTrue

using namespace std;

static double g_RandCount;

void LogRandCount()
	{
	Log("%g calls to RandWrapper().\n", g_RandCount);
	}

static inline uint32 RandWrapper()
	{
	++g_RandCount;
	return RandInt32();
	}

unsigned GetStochasticSeed()
	{
	return (unsigned) (time(0)*getpid());
	}

void InitRand()
	{
	unsigned Seed = GetStochasticSeed();
	const char *strSeed = ValueOpt("seed");
	if (strSeed != 0 && !streq(strSeed, "stochastic"))
		{
		Seed = GetStochasticSeed();
		IntOpt("seed", &Seed);
		}
	Log("Seed=%u\n", Seed);
	InitRand(Seed);
	}

void RandBuffer(unsigned char Buffer[], unsigned Bytes)
	{
	unsigned N = Bytes/sizeof(unsigned);
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned *p = (unsigned *) &Buffer[i];
		*p = RandWrapper();
		}
	unsigned M = N + Bytes%sizeof(unsigned);
	for (unsigned i = N; i < M; ++i)
		{
		unsigned char Byte = (unsigned char) RandWrapper();
		Buffer[i] = Byte;
		}
	}

unsigned RandInt(unsigned N)
	{
	asserta(N > 0);
	unsigned u = RandWrapper();
	return u%N;
	}

uint64 RandInt64(uint64 N)
	{
	asserta(N > 0);
	uint64 u;
	RandBuffer((unsigned char *) &u, sizeof(u));
	return u%N;
	}

coord_t RandCoord(coord_t N)
	{
	if (sizeof(coord_t) == sizeof(unsigned))
		return (coord_t) RandInt(N);
	else if (sizeof(coord_t) == sizeof(uint64))
		return (coord_t) RandInt64(N);
	coord_t c;
	RandBuffer((unsigned char *) &c, sizeof(c));
	return c%N;
	}

gpos_t RandPos(gpos_t N)
	{
	gpos_t c;
	RandBuffer((unsigned char *) &c, sizeof(c));
	return c%N;
	}

unsigned RandWeightDist(const vector<double> &Weights)
	{
	const unsigned N = SIZE(Weights);
	if (N == 0)
		Die("RandWeightDist N=0");
	double SumWeights = 0;
	unsigned NonZeroWeight = UINT_MAX;
	for (unsigned i = 0; i < N; ++i)
		{
		double w = Weights[i];
		if (w < 0)
			Die("RandWeightDist, weight < 0");
		if (w != 0)
			NonZeroWeight = i;
		SumWeights += w;
		}
	if (SumWeights == 0)
		return RandInt(N);

	unsigned r = RandInt(UINT_MAX);

// Make d in range 0..SumWeights
	double d = (r*SumWeights)/(double) UINT_MAX;

	double Sum = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		Sum += Weights[i];
		if (d <= Sum && Weights[i] != 0)
			return i;
		}
	if (Sum == 0)
		Die("RandWeightDist, zero weight vector");
	Warning("RandWeightDist: rounding error");
// should never get here except by rounding error
	return NonZeroWeight;
	}

void GetRandDistHis(const vector<double> &Weights, vector<unsigned> &His)
	{
	His.clear();
	const unsigned N = SIZE(Weights);
	double SumWeights = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		double w = Weights[i];
		if (w < 0)
			Die("RandWeightDist, weight < 0");
		SumWeights += w;
		}

	double Sum2 = 0.0;

	His.resize(N);
	for (unsigned i = 0; i < N; ++i)
		{
		double w = Weights[i];
		Sum2 += w;
		if (w == 0)
			His[i] = 0;
		else
			His[i] = (unsigned) ((Sum2*UINT_MAX + 0.5)/SumWeights);
		}
	}

static void LogHis(const vector<double> &Weights, const vector<unsigned> &His)
	{
	const unsigned N = SIZE(His);
	double SumWeights = 0;
	for (unsigned i = 0; i < N; ++i)
		SumWeights += Weights[i];

	Log("\n");
	Log("  i          Lo        Diff      Weight           w\n");
	Log("===  ==========  ==========  ==========  ==========\n");
	double SumWeight = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		Log("%3u  %10u", i, His[i]);
		unsigned PrevLo = i == 0 ? 0 : His[i-1];
		unsigned Diff = His[i] - PrevLo;
		double Weight = (double) Diff / (double) UINT_MAX;
		SumWeight += Weight;
		Log("  %10u  %10.3g  %10.3g", Diff, Weight, Weights[i]/SumWeights);
		Log("\n");
		}
	Log("SumWeight = %.3g\n", SumWeight);
	}

unsigned RandWeightDist4(const vector<unsigned> &His)
	{
	const unsigned N = SIZE(His);
	unsigned r = RandInt(UINT_MAX);
	for (unsigned i = 0; i < N; ++i)
		if (r < His[i])
			return i;

	Die("RandWeightDist4: fall through");
	UNREACHABLE_RETURN(0);
	}

unsigned RandWeightDist2(const vector<double> &Weights)
	{
	const unsigned N = SIZE(Weights);
	if (N == 0)
		Die("RandWeightDist N=0");
	double SumWeights = 0;
	unsigned NonZeroWeight = UINT_MAX;
	for (unsigned i = 0; i < N; ++i)
		{
		double w = Weights[i];
		if (w < 0)
			Die("RandWeightDist, weight < 0");
		if (w != 0)
			NonZeroWeight = i;
		SumWeights += w;
		}
	if (SumWeights == 0)
		return RandInt(N);

	double Sum2 = 0;
	vector<unsigned> His;
	His.resize(N);
	for (unsigned i = 0; i < N; ++i)
		{
		double w = Weights[i];
		Sum2 += w;
		if (w == 0)
			His[i] = UINT_MAX;
		else
			His[i] = (unsigned) ((Sum2*UINT_MAX + 0.5)/SumWeights);
		}

	unsigned r = RandInt(UINT_MAX);
	for (unsigned i = 0; i < N; ++i)
		if (r <= His[i])
			return i;

	Warning("RandWeightDist: fall through");
// should never get here except by rounding error
	return NonZeroWeight;
	}

unsigned RandWeightDist3(const vector<double> &Weights)
	{
	const unsigned N = SIZE(Weights);
	if (N == 0)
		Die("RandWeightDist N=0");
	double SumWeights = 0;
	unsigned NonZeroWeight = UINT_MAX;
	for (unsigned i = 0; i < N; ++i)
		{
		double w = Weights[i];
		if (w < 0)
			Die("RandWeightDist, weight < 0");
		if (w != 0)
			NonZeroWeight = i;
		SumWeights += w;
		}
	if (SumWeights == 0)
		return RandInt(N);

	unsigned r = RandInt(UINT_MAX);
	double Sum2 = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		double w = Weights[i];
		if (w == 0)
			continue;
		Sum2 += w;
		if (r <= (unsigned) ((Sum2*UINT_MAX + 0.5)/SumWeights))
			return i;
		}

	Warning("RandWeightDist: fall through");
// should never get here except by rounding error
	return NonZeroWeight;
	}

unsigned RandWeightDist(const double Weights[], unsigned N)
	{
	if (N == 0)
		Die("RandWeightDist N=0");
	double SumWeights = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		double w = Weights[i];
		if (w < 0)
			Die("RandWeightDist, weight < 0");
		SumWeights += Weights[i];
		}
	if (SumWeights == 0)
		Die("RandWeightDist, sum=0");

// Get r in range 0..UINT_MAX
	double r = (double) RandInt(UINT_MAX);

// Make d in range 0..SumWeights
	double d = (r*SumWeights)/(double) UINT_MAX;

	double Sum = 0;
	for (unsigned i = 0; i < N; ++i)
		{
	// Nasty case if Weights[j] == 0, it is possible
	// to return j if we don't skip.
		if (Weights[i] == 0)
			continue;
		Sum += Weights[i];
		if (d <= Sum)
			return i;
		}
// should never get here except by rounding error
	return N;
	}

// Sample from an exponential distribution exp(-x/m).
double RandExp(double m)
	{
	for (;;)
		{
		int ir = RandWrapper();
		if (ir > 0)
			{
			double dr = (double) ir/(double) RAND_MAX;
			return -m*log(dr);
			}
		}
	}

// Fill WeightVec with values 0 .. (Weights.size() - 1) such that the
// number of times value i appears approximates the relative value of Weights[i].
// Attempt to make WeightVec.size() be TargetLength, but with rounding
// error this won't happen exactly.
void SetWeightVec(const vector<double> &Weights, vector<unsigned> &WeightVec,
  unsigned TargetLength)
	{
	WeightVec.clear();
	const size_t N = Weights.size();
	if (N == 0)
		return;

	double SumWeights = 0.0;
	for (size_t i = 0; i < N; ++i)
		{
		double w = Weights[i];
		asserta(w >= 0);
		SumWeights += w;
		}

// Compute length of each segment.
	vector<unsigned> Lengths;
	Lengths.resize(N);
	unsigned SumLengths = 0;
	for (size_t i = 0; i < N; ++i)
		{
		double w = Weights[i];
		unsigned L = (unsigned) ((w/SumWeights)*TargetLength + 0.5);
		if (Weights[i] > 0 && L == 0)
			L = 1;
		Lengths[i] = L;
		SumLengths += L;
		}

	for (unsigned i = 0; i < N; ++i)
		{
		unsigned L = Lengths[i];
		for (unsigned j = 0; j < L; ++j)
			WeightVec.push_back(i);
		}
	}

char RandLetter()
	{
	static char Letters[4] = { 'A', 'C', 'G', 'T' };
	unsigned i = RandInt(4);
	return Letters[i];
	}

void RandSeq(string &Seq, unsigned Length)
	{
	Seq.clear();
	for (unsigned i = 0; i < Length; ++i)
		Seq.push_back(RandLetter());
	}

unsigned WeightedRand(const vector<unsigned> &WeightVec)
	{
	const size_t N = WeightVec.size();
	size_t Index = RandWrapper()%N;
	return WeightVec[Index];
	}

bool RandTrue(double Prob)
	{
	assert(Prob <= 1);
	if (Prob == 0)
		return 0;
	const unsigned N = 1000000;
	unsigned i = RandInt(N);
	return i < Prob*N;
	}

void Normalize(vector<double> &Weights)
	{
	const size_t N = Weights.size();
	if (N == 0)
		return;
	double Sum = 0;
	for (size_t i = 0; i < N; ++i)
		{
		double w = Weights[i];
		asserta(w >= 0);
		Sum += w;
		}
	if (Sum == 0)
		Die("Normalize: Weights vector sums to zero");

	for (size_t i = 0; i < N; ++i)
		Weights[i] /= Sum;
	}

void Normalize(vector<double> &Weights, double Value)
	{
	const size_t N = Weights.size();
	if (N == 0)
		return;
	double Sum = 0;
	for (size_t i = 0; i < N; ++i)
		{
		double w = Weights[i];
		asserta(w >= 0);
		Sum += w;
		}
	if (Sum == 0)
		Die("Normalize: Weights vector sums to zero");

	double Factor = Value/Sum;

	for (size_t i = 0; i < N; ++i)
		Weights[i] *= Factor;
	}

void LogWeightVec(const vector<double> &Weights, const vector<unsigned> &WeightVec)
	{
	const size_t N = Weights.size();
	const size_t L = WeightVec.size();

	vector<double> WeightVecWeights(N, 0);
	for (size_t i = 0; i < L; ++i)
		{
		unsigned Index = WeightVec[i];
		asserta(Index < N);
		++(WeightVecWeights[Index]);
		}
	vector<double> NormalizedWeights(Weights);
	Normalize(NormalizedWeights);
	Normalize(WeightVecWeights);

	Log("\n");
	Log("Weight  WeightVecWeight\n");
	Log("======  ===============\n");
	//   123456  123456789012345
	for (size_t i = 0; i < N; ++i)
		Log("%6.3f  %15.3f\n", NormalizedWeights[i], WeightVecWeights[i]);
	}

void GetWeightsFromWeightVec(const vector<unsigned> &WeightVec,
  vector<double> &Weights)
	{
	Weights.clear();
	if (WeightVec.empty())
		return;
	unsigned MaxSub = 0;
	const unsigned N = SIZE(WeightVec);
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned Sub = WeightVec[i];
		if (Sub > MaxSub)
			MaxSub = Sub;
		}
	unsigned M = MaxSub + 1;
	Weights.resize(M, 0);
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned Sub = WeightVec[i];
		++(Weights[Sub]);
		}
	Normalize(Weights);
	}

void TestRandExp()
	{
	static int Bins[200];
	int N = 10000;
	double Sum = 0;
	for (int i = 0; i < N; ++i)
		{
		double r = RandExp(100.0);
		int ir = (int) (r + 0.5);
		if (ir >= 0 && ir < 200)
			++(Bins[ir]);
		Sum += r;
		Log("%g\n", r);
		}
	Log("N=%d Avg=%g\n", N, Sum/N);
	Log("\n");
	for (int i = 0; i < 200; ++i)
		{
		Log("Bin[%2d] = %5d  ", i, Bins[i]);
		int k = Bins[i];
		for (int j = 0; j < k; ++j)
			Log("*");
		Log("\n");
		}
	}

void TestWeightVec()
	{
	unsigned SAMPLES = 10000000;
	unsigned N = 3;

	vector<double> Weights;
	Weights.resize(N);
	Weights[0] = 2;
	Weights[1] = 1;
	Weights[2] = 3;

	vector<unsigned> WeightVec;
	SetWeightVec(Weights, WeightVec, 1001);
	LogWeightVec(Weights, WeightVec);

	vector<unsigned> Counts(N, 0);
	for (unsigned i = 0; i < SAMPLES; ++i)
		{
		unsigned j = WeightedRand(WeightVec);
		asserta(j < N);
		++(Counts[j]);
		}
	vector<double> NormalizedWeights(Weights);
	Normalize(NormalizedWeights);

	Log("\n");
	Log("Weight  Sampled weight\n");
	Log("======  ==============\n");
	//   123456  12345678901234
	for (unsigned i = 0; i < N; ++i)
		Log("%6.3f  %14.3f\n", NormalizedWeights[i], (double) Counts[i]/SAMPLES);
	}

void RandomizeVec(vector<unsigned> &Vec)
	{
	const unsigned N = SIZE(Vec);
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned j = RandInt(N-i);
		swap(Vec[j], Vec[N-i-1]);
		}
	}

void RandomizeVec(vector<double> &Vec)
	{
	const unsigned N = SIZE(Vec);
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned j = RandInt(N-i);
		swap(Vec[j], Vec[N-i-1]);
		}
	}

void RandomizeVecWeighted(vector<unsigned> &Vec, const vector<double> &Weights)
	{
	vector<double> w(Weights);

	const unsigned N = SIZE(Vec);
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned j = RandWeightDist(w);
		swap(Vec[j], Vec[N-i-1]);
		swap(w[j], w[N-i-1]);
		w.resize(N-1-i);
		}
	}

double RandNorm(double Mean, double StdDev)
	{
	NormalPDF NPDF(Mean, StdDev);
	RandDist RD;
	RD.Init(NPDF, 0, 2*(Mean + StdDev));
	return RD.Get();
	}

void TestRandOld()
	{
	const unsigned N = 100000;
	const unsigned BINS = 20;
	const unsigned Mean = 100000;

	vector<unsigned> Bins;
	Bins.resize(BINS);

	unsigned MaxBin = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned x = (unsigned) RandExp(Mean);
		unsigned k = (x*BINS)/(2*Mean);
		if (k >= BINS)
			continue;
		unsigned b = Bins[k] + 1;
		Bins[k] = b;
		if (b > MaxBin)
			MaxBin = b;
		}

	for (unsigned i = 0; i < BINS; ++i)
		{
		unsigned Min = i*(Mean/5);
		unsigned Max = (i+1)*(Mean/5) - 1;
		printf("%7u-%7u  %u  ", Min, Max, Bins[i]);
		unsigned L = (Bins[i]*32)/MaxBin;
		for (unsigned j = 0; j < L; ++j)
			printf("*");
		printf("\n");
		}
	}

void TestRandOld2()
	{
	const unsigned N = 100000;
	const unsigned BINS = 30;
	const unsigned Mean = 3;

	vector<unsigned> Bins;
	Bins.resize(BINS);

	unsigned MaxBin = 0;
	for (unsigned i = 0; i < N; ++i)
		{
		double x = RandExp(Mean);
		unsigned k = (unsigned) (x*1.0);
		if (k == 0)
			continue;
		if (k >= BINS)
			continue;
		unsigned b = Bins[k] + 1;
		Bins[k] = b;
		if (b > MaxBin)
			MaxBin = b;
		}

	for (unsigned i = 0; i < BINS; ++i)
		{
		printf("%.3f - %.3f  %6u  ", i/1.0, (i+1)/1.0 - 0.001, Bins[i]);
		unsigned L = (Bins[i]*32)/MaxBin;
		for (unsigned j = 0; j < L; ++j)
			printf("*");
		printf("\n");
		}
	}

void TestRand2()
	{
	InitRand();

	unsigned SAMPLES = UINT_MAX/100;

	//const unsigned Lo = 4291919210;
	//const unsigned Hi = 4291919932;
	//unsigned Range = Hi - Lo + 1;
	//unsigned Count = 0;
	//for (unsigned i = 0; i < SAMPLES; ++i)
	//	{
	//	unsigned r = RandInt(UINT_MAX);
	//	if (r >= Lo && r <= Hi)
	//		++Count;
	//	}
	//double dExpected = ((double) Range*SAMPLES) / (double) UINT_MAX;
	//Log("Range = %u Count = %u, expected %.3g\n", Range, Count, dExpected);

	vector<double> Weights;
	Weights.resize(43);
	Weights[0] = 7.02e+006;
	Weights[1] = 2.58e+006;
	Weights[2] = 2.58e+005;
	Weights[3] = 1.41e+006;
	Weights[4] = 2.7e+006;
	Weights[5] = 4.11e+003;
	Weights[6] = 4.11e+003;
	Weights[7] = 4.69e+007;
	Weights[8] = 2.72e+005;
	Weights[9] = 0;
	Weights[10] = 0;
	Weights[11] = 0;
	Weights[12] = 0;
	Weights[13] = 0;
	Weights[14] = 2.69;
	Weights[15] = 0.537;
	Weights[16] = 2.15;
	Weights[17] = 1.74;
	Weights[18] = 1.74;
	Weights[19] = 10.3;
	Weights[20] = 10.3;
	Weights[21] = 10.3;
	Weights[22] = 10.3;
	Weights[23] = 103;
	Weights[24] = 3.47;
	Weights[25] = 5.37;
	Weights[26] = 3.47;
	Weights[27] = 5.37;
	Weights[28] = 3.47;
	Weights[29] = 5.37;
	Weights[30] = 3.47;
	Weights[31] = 5.37;
	Weights[32] = 103;
	Weights[33] = 2.4e+004;
	Weights[34] = 2.45e+003;
	Weights[35] = 2.45e+003;
	Weights[36] = 1.23e+004;
	Weights[37] = 481;
	Weights[38] = 481;
	Weights[39] = 961;
	Weights[40] = 5.37;
	Weights[41] = 5.37;
	Weights[42] = 10.7;
	unsigned N = SIZE(Weights);

	vector<unsigned> His;
	GetRandDistHis(Weights, His);
	LogHis(Weights, His);

	//time_t t1 = time(0);
	//for (unsigned i = 0; i < SAMPLES; ++i)
	//	RandWeightDist4(His);
	//time_t t2 = time(0);
	//for (unsigned i = 0; i < SAMPLES; ++i)
	//	RandWeightDist2(Weights);
	//time_t t3 = time(0);

	vector<unsigned> Counts(N, 0);
	for (unsigned i = 0; i < SAMPLES; ++i)
		{
		unsigned j = RandWeightDist4(His);
		asserta(j < N);
		++(Counts[j]);
		}
	vector<double> NormalizedWeights(Weights);
	Normalize(NormalizedWeights);

	Log("\n");
	Log("  i          Weight       Count  ExpectedCt    Error    Norm. weight  Sampled weight\n");
	Log("===  ==============  ==========  ==========  =======  ==============  ==============\n");
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned ExpectedCount = (unsigned) (SAMPLES*NormalizedWeights[i] + 0.5);
		unsigned Count = Counts[i];
		double Pct(double,double);
		double Error = Pct(2*Count, ExpectedCount + Count) - 100.0;
		Log("%3u  %14.3f  %10u  %10u  %6.1f%%  %14.3g  %14.3g\n",
		  i,
		  Weights[i],
		  Count,
		  ExpectedCount,
		  Error,
		  NormalizedWeights[i],
		  (double) Counts[i]/SAMPLES);
		}
	//Log("\n");
	//Log("Dist2: %u secs\n", (unsigned) (t3 - t2));
	//Log("Dist4: %u secs\n", (unsigned) (t2 - t1));
	}

void TestRand()
	{
	InitRand();
	for (int i = 0; i < 300000; ++i)
		Log("%u\n", RandInt(30000));
	}

unsigned RandInt_Trace(const char *FileName, int LineNr, unsigned N)
	{
	unsigned i = RandInt(N);
	Log("%s:%d RandInt(%u)=%u\n", FileName, LineNr, N, i);
	return i;
	}

uint64 RandInt64_Trace(const char *FileName, int LineNr, uint64 N)
	{
	uint64 i = RandInt64(N);
	Log("%s:%d RandInt(%"UINT64_PRINTF")=%"UINT64_PRINTF"\n", FileName, LineNr, N);
	return i;
	}

double RandExp_Trace(const char *FileName, int LineNr, double m)
	{
	double x = RandExp(m);
	Log("%s:%d RandExp(%g)=%g\n", FileName, LineNr, m, x);
	return x;
	}

bool RandTrue_Trace(const char *FileName, int LineNr, double p)
	{
	bool b = RandTrue(p);
	Log("%s:%d RandTrue(%g)=%c\n", FileName, LineNr, p, tof(b));
	return b;
	}
