#include "ratemgr.h"
#include "evolver.h"

extern FILE *g_fStats;

void ReadModel(const char *FileName);
void ReadDist(FILE *f, vector<pair<gpos_t, double> > &Rates);

static RateMgr g_RateMgrs[RateTypeCount];
static vector<vector<pair<gpos_t, double> > > g_RateVecVec(RateTypeCount);

#define R(e, r)	RateMgr *g_##e##RateMgr = &g_RateMgrs[RT_##r];

R(Delete, Delete)
R(Invert, Invert)
R(Move, Move)
R(Copy, Copy)
R(Tandem, Tandem)
R(TandemExpand, TandemExpand)
R(TandemContract, TandemContract)
R(Insert, Insert)
R(MEInsert, MEInsert)
R(NCCE, NCCE)

#undef R


RateMgr *g_InterChrMoveRateMgr = &g_RateMgrs[RT_InterChrMove];
RateMgr *g_InterChrCopyRateMgr = &g_RateMgrs[RT_InterChrCopy];

int GetRateTypeSign(RATE_TYPE Type)
	{
	switch (Type)
		{
	case RT_Delete:			return -1;
	case RT_Insert:			return +1;
	case RT_Invert:			return  0;
	case RT_Tandem:			return +1;
	case RT_TandemExpand:	return +1;
	case RT_TandemContract:	return -1;
	case RT_Copy:			return +1;
	case RT_Move:			return  0;
	case RT_InterChrMove:	return  0;
	case RT_InterChrCopy:	return +1;
	case RT_NCCE:			return  0;
	case RT_MEInsert:		return +1;
	default:
		break;
		}
	Warning("GetRateTypeSign(%d), unknown type", Type);
	return 0;
	}

const char *RateTypeToStr(RATE_TYPE Type)
	{
	switch (Type)
		{
#define R(r)	case RT_##r: return #r;
#include "ratetypes.h"
		}
	static char Str[16];
	sprintf(Str, "RT_%u", Type);
	return Str;
	}

bool IsInterChrRateType(RATE_TYPE Type)
	{
	return Type == RT_InterChrCopy || Type == RT_InterChrMove;
	}

//static bool InitDists()
//	{
//	vector<pair<gpos_t, double> > Rates;
//
//#define RATE(Event, Length, Freq)												\
//	{																			\
//	vector<pair<gpos_t, double> > &Rates = g_RateVecVec[RT_##Event];			\
//	Rates.push_back(pair<gpos_t, double>(Length, Freq));						\
//	}
//
//#include "eventrates.h"
//#include "nccerates.h"
//
//	return true;
//	}
//static bool InitDistsDone = InitDists();

void LogDists()
	{
	for (unsigned i = 0; i < RateTypeCount; ++i)
		{
		const RateMgr &RM = g_RateMgrs[i];
		RM.LogMe();
		}
	}

void LogRateSummary(bool Cmt)
	{
	const char *sCmt = (Cmt ? "# " : "");
	Log("\n");
	Log("%s       Event   Rate 1-10   Rate 1-99   Rate 100+   TotalRate  Bases 1-10  Bases 1-99  Bases 100+  TotalBases\n",
	  sCmt);
	Log("%s============  ==========  ==========  ==========  ==========  ==========  ==========  ==========  ==========\n",
	  sCmt);
	for (unsigned i = 0; i < RateTypeCount; ++i)
		{
		RateMgr &RM = g_RateMgrs[i];
		if (RM.GetType() == RT_NCCE || RM.GetType() == RT_MEInsert)
			continue;
		RateStats Stats;
		RM.GetStats(Stats);
		Log("%s%12.12s", sCmt, RM.TypeStr());
		Log("  %10.3g", Stats.Rate1to10);
		Log("  %10.3g", Stats.Rate1to99);
		Log("  %10.3g", Stats.Rate100plus);
		Log("  %10.3g", Stats.Rate1to99 + Stats.Rate100plus);
		Log("  %10.3g", Stats.Bases1to10);
		Log("  %10.3g", Stats.Bases1to99);
		Log("  %10.3g", Stats.Bases100plus);
		Log("  %10.3g", Stats.Bases1to99 + Stats.Bases100plus);
		Log("\n");
		}
	}

void LogRateParams()
	{
	for (unsigned i = 0; i < RateTypeCount; ++i)
		{
		RateMgr &RM = g_RateMgrs[i];
		if (RM.GetType() == RT_MEInsert)
			continue;
		RM.LogParams();
		}
	LogRateSummary(true);
	}

void SetModel(const char *FileName)
	{
	if (FileName != 0)
		ReadModel(FileName);

	for (unsigned i = 0; i < RateTypeCount; ++i)
		{
		RateMgr &RM = g_RateMgrs[i];
		RATE_TYPE RT = (RATE_TYPE) i;
		RM.Init(RT, g_RateVecVec[i]);
		}
	}

const char *BinToStr(unsigned Bin)
	{
	switch (Bin)
		{
	case 0:
		return "0";
	case 1:
		return "1-";
	case 2:
		return "10-";
	case 3:
		return "100-";
	case 4:
		return "1000-";
	case 5:
		return "10,000-";
	case 6:
		return "100,000-";
	case 7:
		return "1,000,000-";
	case 8:
		return "10,000,000-";
	case 9:
		return "100,000,000-";
		}
	return "Huge!?";
	}

void LogLengthCounts()
	{
	Log("\n");
	Log("       Event       Range       Fails     Rejects     Accepts       Total\n");
	Log("============  ==========  ==========  ==========  ==========  ==========\n");
	for (unsigned i = 0; i < RateTypeCount; ++i)
		{
		bool Any = false;
		const RateMgr &RM = g_RateMgrs[i];
		RATE_TYPE RT = (RATE_TYPE) i;
		const unsigned *FailCounts = RM.GetResultCounts(ER_Fail);
		const unsigned *RejectCounts = RM.GetResultCounts(ER_Reject);
		const unsigned *AcceptCounts = RM.GetResultCounts(ER_Accept);
		for (unsigned Bin = 1; Bin < 16; ++Bin)
			{
			unsigned FailCount = FailCounts[Bin];
			unsigned RejectCount = RejectCounts[Bin];
			unsigned AcceptCount = AcceptCounts[Bin];
			unsigned TotalCount = FailCount + RejectCount + AcceptCount;
			if (TotalCount == 0)
				continue;
			Any = true;
			const char *RateType = RateTypeToStr(RT);
			const char *BinStr = BinToStr(Bin);
			Log("%12.12s", RateType);
			Log("  %10.10s", BinStr);
			Log("  %10u", FailCount);
			Log("  %10u", RejectCount);
			Log("  %10u", AcceptCount);
			Log("  %10u", TotalCount);
			Log("\n");
			if (g_fStats != 0)
				{
				if (FailCount > 0)
					fprintf(g_fStats, "EV_RANGE_FAIL;%s;%s;%u\n", RateType, BinStr, FailCount);

				if (RejectCount > 0)
					fprintf(g_fStats, "EV_RANGE_REJECT;%s;%s;%u\n", RateType, BinStr, RejectCount);

				if (AcceptCount > 0)
					fprintf(g_fStats, "EV_RANGE_ACCEPT;%s;%s;%u\n", RateType, BinStr, AcceptCount);
				}
			}
		if (Any)
			Log("\n");
		}

	Log("\n");
	Log("       Event   Rej.Bases   Acc.Bases  +\n");
	Log("============  ==========  ==========  =\n");
	int BaseCountDelta = 0;
	for (unsigned i = 0; i < RateTypeCount; ++i)
		{
		const RateMgr &RM = g_RateMgrs[i];
		RATE_TYPE RT = (RATE_TYPE) i;
		const char *RateType = RateTypeToStr(RT);
		int Sign = GetRateTypeSign(RT);
		unsigned Rej = RM.GetTotalRejectedBases();
		unsigned Acc = RM.GetTotalAcceptedBases();
		BaseCountDelta += (Sign*(int) Acc);
		if (Rej == 0 && Acc == 0)
			continue;
		Log("%12.12s", RateType);
		char cSign = '?';
		if (Sign > 0)
			cSign = '+';
		else if (Sign < 0)
			cSign = '-';
		else if (Sign == 0)
			cSign = '0';
		Log("  %10u  %10u  %c\n", Rej, Acc, cSign);
		if (g_fStats != 0)
			{
			if (Acc > 0)
				fprintf(g_fStats, "EV_ACC_BASES;%s;%u\n", RateType, Acc);

			if (Rej > 0)
				fprintf(g_fStats, "EV_REJ_BASES;%s;%u\n", RateType, Rej);
			}
		}

	Log("                          ==========\n");
	Log("Total x sign              %+10d  (sequence length change)\n",
	  BaseCountDelta);
	Log("\n");
	}

void ReadEventDist(FILE *f, const string &Event)
	{
	if (Event == "Delete")
		ReadDist(f, g_RateVecVec[RT_Delete]);
	else if (Event == "Insert")
		ReadDist(f, g_RateVecVec[RT_Insert]);
	else if (Event == "Invert")
		ReadDist(f, g_RateVecVec[RT_Invert]);
	else if (Event == "Tandem")
		ReadDist(f, g_RateVecVec[RT_Tandem]);
	else if (Event == "Copy")
		ReadDist(f, g_RateVecVec[RT_Copy]);
	else if (Event == "Move")
		ReadDist(f, g_RateVecVec[RT_Move]);
	else if (Event == "InterChrCopy")
		ReadDist(f, g_RateVecVec[RT_InterChrCopy]);
	else if (Event == "InterChrMove")
		ReadDist(f, g_RateVecVec[RT_InterChrMove]);
	else if (Event == "NCCE")
		ReadDist(f, g_RateVecVec[RT_NCCE]);
	else
		Die("ReadEventDist(%s) invalid", Event.c_str());
	}

void SetMERates(const RateVec &Rates)
	{
	g_MEInsertRateMgr->Init(RT_MEInsert, Rates);
	}

#if	0
void Test()
	{
	RateMgr &RM = *(g_DeleteRateMgr);
	for (unsigned i = 0; i < 100; ++i)
		{
		gpos_t Length = RM.GetLength();
		Log("%"PF"\n", Length);
		}
	}
#endif
