#ifndef REVFILE_H
#define REVFILE_H

#include <stdio.h>
#include <map>
#include <vector>
#include "geninfo.h"
#include "intlist.h"
#include "revfile_disk.h"
#include "alninfo.h"
#include "iface.h"

#define VALIDATE	0
#define PROGRESS	0

#ifdef _MSC_VER
#pragma warning(disable : 4800)	// int->bool conversion
#endif

void RevComp(char *S, unsigned L);
void RevComp(StdStr &S);

// for getpid():
#ifdef	_MSC_VER
#include <process.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif

#define ReadData_(v)		ReadStdioFile(m_f, &v, (unsigned) sizeof(v))
#define WriteData_(v)	WriteStdioFile(m_f, &v, (unsigned) sizeof(v))

#define STATE(x, y)	if (m_State != x) Die("RevFileBase: function call in invalid state"); m_State = y;
#define STATEC(x)	if (m_State != x) Die("RevFileBase: function call in invalid state");

typedef vector<pair<unsigned, unsigned> > IntPairVec;
typedef map<GenChr, IntVec> GenChrToIntVecMap;
typedef vector<GenInfo> GenInfoVec;

struct IxLoHi
	{
	unsigned Lo;
	unsigned Hi;
	bool operator<(const IxLoHi &rhs) const
		{
		if (Lo < rhs.Lo)
			return true;
		return Hi < rhs.Hi;
		}
	};

class RevFileBase
	{
protected:
	enum RevFileState
		{
		Closed,
		Reading,
		Writing
		};

public:
	RevFileBase();
	virtual ~RevFileBase();
	void Close();
	const string &GetFileName() const;

// Members for reading existing file
// =================================
	void Open(const string &FileName);
	const GenInfoVec &GetGenomes() const;

	virtual unsigned GetMagic() const = 0;
	virtual bool CanCompressAI(const AlnInfo &AI) const = 0;

// Get AlnInfo, return AlnIndex.
	virtual unsigned GetAlnInfo(unsigned AlnInfoIndex, AlnInfo &AI) const = 0;
	virtual unsigned GetAlnInfoX(unsigned AlnInfoIndex, AlnInfoX &AIX) const = 0;

// GetGenomeCount/Info are redundant because could get
// via GetGenomes; provided for convenience.
	unsigned GetGenomeCount() const;
	const GenInfo &GetGenomeInfo(unsigned GenomeIndex) const;

	GenInfo &GetModifiableGenomeInfo(unsigned GenomeIndex);
	GenInfo *GetModifiableGenomeInfoPtr(const string &GenomeName,
	  unsigned *ptrGenomeIndex);

// GetGenomeName, GetChrCount and GetChrInfo are redundant because
// could get via GenInfo; provided for convenience.
	const string &GetGenomeName(unsigned GenomeIndex) const;
	unsigned GetGenomeIndex(const string &GenomeName) const;
	unsigned GetChrCount(unsigned GenomeIndex) const;
	const ChrInfo &GetChrInfo(unsigned GenomeIndex, unsigned ChrId) const;
	const ChrInfo *GetChrInfo(unsigned GenomeIndex, const string &ChrName) const;

	const string &GetChrName(unsigned GenomeIndex, unsigned ChrId) const;
	unsigned GetChrId(unsigned GenomeIndex, const string &ChrName) const;
	unsigned GetChrId(unsigned GenomeIndex, unsigned ChrIndex) const;
	unsigned GetChrIndex(unsigned GenomeIndex, unsigned ChrId) const;
	coord_t GetChrLength(unsigned GenomeIndex, unsigned ChrId) const;

// Members for enumerating sequence data.
	unsigned GetSeqCount() const;
	void GetSeqByIndex(unsigned SeqIndex, char *Seq, unsigned Bytes,
	  unsigned *ptrGenomeIndex, unsigned *ptrChrId) const;
	void GetSeqGenChrByIndex(unsigned SeqIndex, unsigned *ptrGenomeIndex,
	  unsigned *ptrChrId) const;
	coord_t GetSeqLengthByIndex(unsigned SeqIndex) const;

// Members for getting sequence data by genome & chromosome
	bool IsSeqPresent(unsigned GenomeIndex, unsigned ChrId) const;
	coord_t GetSeqLength(unsigned GenomeIndex, unsigned ChrId) const;
	void GetSeq(unsigned GenomeIndex, unsigned ChrId, char *Seq, unsigned Bytes) const;
	void GetSubSeq(unsigned GenomeIndex, unsigned ChrId, char *Seq,
	  unsigned From, unsigned Length, bool RevComp = false) const;
	void GetSubSeq2(unsigned GenomeIndex, unsigned ChrId, StdStr &Seq,
	  unsigned From, unsigned Length, bool RevComp = false) const;
	void GetSubSeq3(const GCSeg &Seg, StdStr &Seq) const;

// Not necessarily 1-1 correspondence between AlnInfo's & AlnRows's.
// Can have an AlnInfo without an explicit alignment, i.e. without
// rows, but not vice versa.
	unsigned GetAlnInfoCount() const;

// Basic query:
//		Input is (gen1, chr1), (gen2, chr2) ... (genN, chrN)
//			Each pair must be present, so is (gen1, chr1) AND (gen2, chr2) AND ...
//		Output is vector of AlnInfoIndexes.
	void GetAppendAlnInfoIndexesChr(unsigned GenIndex, unsigned ChrId,
	  IntSet &AlnInfoIndexes) const;
	void GetAppendAlnInfoIndexesChr(unsigned GenIndex, unsigned ChrId,
	  IntVec &AlnInfoIndexes) const;

// Get AlnInfo[X]s given vector of AlnInfoIndexes.
// Current implemented by naive loop over GetAlnInfo(), but
// could be optimized for speed in future if needed.
	void GetAlnInfos(const IntVec &AIIndexes, AIVec &AIs) const;
	void GetAlnInfoXs(const IntVec &AIIndexes, AIXVec &AIXs) const;
	void GetAlnInfos(const IntSet &AIIndexes, AIVec &AIs) const;

// Special case: get all AlnInfo[X]s.
// Commonly needed operation, allows internal optimization later if desired.
	void GetAllAlnInfos(AIVec &AIs, bool ShowProgress = true) const;
	void GetAllAlnInfoXs(AIXVec &AIXs, bool ShowProgress = true) const;

// Convert AlnInfoIndex <-> AlnIndex.
	virtual unsigned GetAlnIndex(unsigned AlnInfoIndex) const = 0;

	void GetActiveChrPairs(unsigned GenomeIndexA, unsigned GenomeIndexB,
	  set<IntPair> &ChrIdPairs) const;
	void GetActiveChrPairsSlow(unsigned GenomeIndexA, unsigned GenomeIndexB,
	  set<IntPair> &ChrIdPairs) const;

// Members for creating new file
// =============================
	void Create(const string &FileName, const GenInfoVec &Genomes);
	void Create(const string &FileName);

	unsigned AddGenome(const string &Name);

// Add a new alignment, return AlnInfoIndex (not AlnIndex!).
// AlnInfoIndex can be used to get AlnIndex if needed.
	unsigned AddAln(const AlnInfo &AI, const StdStrVec &Rows);

// Add a new sequence, return SeqIndex
	unsigned AddSeq(const string &GenomeName, const string &ChrName,
	  const char *Seq, coord_t Length);

// Check if segment lengths in AI.Segs[] match (ungapped) sequence
// lengths in Rows[], if not die with error message.
	void ValidateSegLengths(const AlnInfo &AI, const StdStrVec &Rows);

protected:
	void CloseReading();
	void CloseWriting();
	void AddHashEntry(unsigned GenomeIndex, unsigned ChrId, uint64 Pos,
	  unsigned AlnCount);
	HashTableEntry *FindHashEntry(unsigned GenomeIndex, unsigned ChrId) const;
	uint64 StoreBitRows(const StdStrVec &Rows);
	void GetAlnRowsInPlace(const DiskAlnInfoTOCEntryC &DATEC, StdStrVec &Rows) const;

	virtual void AddToIndexList(unsigned GenomeIndex, unsigned ChrId,
	  unsigned AlnInfoIndex) = 0;
	virtual void WriteHashTable() = 0;
	virtual void WriteAlnInfoTOC() = 0;

	void WriteGenomes();
	void WriteAlnInfoSegs();
	void WriteAlnTOC();
	void WriteAlns();
	void WriteSeqs();

protected:
	string m_FileName;
	RevFileState m_State;
	GenInfoVec m_Genomes;
	unsigned m_AlnInfoCount;
	unsigned m_AlnCount;
	unsigned m_HashTableEntryCount;
	HashTableEntry *m_Hash;

// Valid when writing only
// Tmp files
	string m_AlnInfoTOCFileName;
	string m_AlnInfoSegsFileName;
	string m_AlnTOCFileName;
	string m_AlnsFileName;
	string m_SeqsFileName;
	FILE *m_fAlnInfoTOC;
	FILE *m_fAlnInfoSegs;
	FILE *m_fAlnTOC;
	FILE *m_fAlns;
	FILE *m_fSeqs;

// Map (GenIndex,ChrId) to SeqIndex.
	map<IntPair, unsigned> m_SeqTOC;
	vector<LongIntPair> m_SeqPosLength;

	RevFileHdr m_Hdr;
	FILE *m_f;

public:
	mutable unsigned m_AICacheHitCount;
	mutable unsigned m_AICacheRefreshCount;
	unsigned m_CompressedAICount;
	unsigned m_InPlaceRowsCount;
	};

void Trunc64(byte *Target, uint64 i64, unsigned Bytes);
void Untrunc64(const byte *Buffer, unsigned Bytes, uint64 &i64);

#endif	// REVFILE_H
