#include <map>
#include <list>
#include <algorithm>
#include "types.h"
#include "segbagai.h"
#include "utils.h"

void SegBag::InsertSeg(const GCSeg &s)
	{
	assert(s.hi >= s.lo);
	SegMapValue v(s.GetMidpoint(), s);
	m_Segs.insert(v);
	}

void SegBag::DeleteSeg(coord_t Pos)
	{
	m_Segs.erase(Pos);
	}

void SegBag::DeleteSeg(const GCSeg &s)
	{
	m_Segs.erase(s.GetMidpoint());
	}

bool SegBag::GetSeg(coord_t Pos, GCSeg *ptrs) const
	{
	SegMapCIter p = m_Segs.find(Pos);
	if (p == m_Segs.end())
		return false;
	SegMapValue v = *p;
	*ptrs = v.second;
	return true;
	}

bool SegBag::GetNearestSeg(coord_t Pos, SegMapCIter *ptrIter) const
	{
	if (m_Segs.size() == 0)
		return false;

	SegMapCIter pGE = m_Segs.lower_bound(Pos);

	if (pGE == m_Segs.begin())
		{
		*ptrIter = pGE;
		return true;
		}

	if (pGE == m_Segs.end())
		{
		SegMapCIter p = pGE;
		assert(p != m_Segs.begin());
		--p;
		*ptrIter = p;
		return true;
		}

	if (pGE->first == Pos)
		{
		*ptrIter = pGE;
		return true;
		}

	SegMapCIter pPred = pGE;
	SegMapCIter pSucc = pGE;
	
	assert(pPred != m_Segs.begin());
	assert(pSucc != m_Segs.end());

	--pPred;

	coord_t Pred = pPred->first;
	coord_t Succ = pSucc->first;

	if (Pred >= Pos || Pos >= Succ)
		{
		Log("Pred=%u Pos=%u Succ=%u\n", Pred, Pos, Succ);
		LogMe();
		Die("SegBag::GetNearestSeg");
		}
	assert(Pred < Pos);
	assert(Pos < Succ);

	if (Pos - Pred < Succ - Pos)
		*ptrIter = pPred;
	else
		*ptrIter = pSucc;

	return true;
	}

bool SegBag::GetNearestSeg(coord_t Pos, coord_t *ptrSegPos) const
	{
	SegMapCIter p;
	bool Found = GetNearestSeg(Pos, &p);
	if (!Found)
		return false;
	*ptrSegPos = p->first;
	return true;
	}

unsigned SegBag::GetSegCount() const
	{
	return (unsigned) m_Segs.size();
	}

void SegBag::LogMe() const
	{
	for (SegMapCIter p = m_Segs.begin(); p != m_Segs.end(); ++p)
		{
		const GCSeg &Seg = p->second;
		Log(" (%u,%u %u-%u m=%u L=%u)",
		  Seg.GenomeIndex,
		  Seg.ChrId,
		  Seg.lo,
		  Seg.hi,
		  Seg.GetMidpoint(),
		  Seg.GetLength());
		}
	}

/***
"Subtract" segment s from the bag of segments.
Consider a segment B in the bag. There are four cases:
	1. Split in two if B covers all of s,
	2. Delete B if s covers all of B,
	3. Reduce size of B if B is partially covered by s, or
	4. No change if there is no overlap of B and s.
***/
void SegBag::Subtract(const GCSeg &s, coord_t MinLength, double CovFract)
	{
	if (m_Segs.size() == 0)
		return;

// Too messy to update m_Segs while scanning it;
// iterators are invalidated each time erase()
// is called. So scan without updating, then
// do the updates.
	std::list<coord_t> DeleteList;
	std::list<GCSeg> InsertList;

	coord_t Mid = s.GetMidpoint();

	SegMapCIter ps;
	bool Found = GetNearestSeg(Mid, &ps);
	assert(Found);
	if (!Found)
		return;

// Scan leftwards towards beginning.
	SegMapCIter p = ps;
	bool Covers(const GCSeg &SA, const GCSeg &SB, double MinFract);
	for (;;)
		{
		coord_t SegPos = p->first;
		const GCSeg &B = p->second;

	// End of leftwards scan when no overlap
	// (Relies on sort order of bag).
		if (B.hi < s.lo)
			break;

	// 1. Split in two if B covers all of s
		if (Covers(B, s, CovFract))
			{
			DeleteList.push_back(SegPos);

			if (s.lo > B.lo && s.lo - B.lo + 1 >= MinLength)
				{
				GCSeg PredSeg;
				PredSeg.GenomeIndex = B.GenomeIndex;
				PredSeg.ChrId = B.ChrId;
				PredSeg.lo = B.lo;
				PredSeg.hi = s.lo - 1;
				PredSeg.Plus = B.Plus;
				InsertList.push_back(PredSeg);
				}

			if (s.hi < B.hi && B.hi - s.hi + 1 >= MinLength)
				{
				GCSeg SuccSeg;
				SuccSeg.GenomeIndex = B.GenomeIndex;
				SuccSeg.ChrId = B.ChrId;
				SuccSeg.lo = s.hi + 1;
				SuccSeg.hi = B.hi;
				SuccSeg.Plus = B.Plus;
				InsertList.push_back(SuccSeg);
				}
			}

	// 2. Delete B if s covers all of B
		else if (Covers(s, B, CovFract))
			DeleteList.push_back(SegPos);

		// 3. Reduce size of B if B is partially covered by s
		else
			{
			if (s.lo > B.lo && s.lo < B.hi && s.lo - B.lo >= MinLength)
				{
				DeleteList.push_back(SegPos);

				GCSeg NewSeg = B;
				NewSeg.hi = s.lo;
				InsertList.push_back(NewSeg);
				}
			}

		if (p == m_Segs.begin())
			break;
		--p;
		}

// Scan rightwards towards end.
	p = ps;
	for (;;)
		{
		coord_t SegPos = p->first;
		const GCSeg &B = p->second;

	// End of rightwards scan when no overlap
	// (Relies on sort order of bag).
		if (B.lo > s.hi)
			break;

	// 1. Split in two if B covers all of s
		if (Covers(B, s, CovFract))
			{
			DeleteList.push_back(SegPos);

			if (s.lo > B.lo && s.lo - B.lo + 1 >= MinLength)
				{
				GCSeg PredSeg;
				PredSeg.GenomeIndex = B.GenomeIndex;
				PredSeg.ChrId = B.ChrId;
				PredSeg.lo = B.lo;
				PredSeg.hi = s.lo - 1;
				PredSeg.Plus = B.Plus;
				InsertList.push_back(PredSeg);
				}

			if (s.hi < B.hi && B.hi - s.hi + 1 >= MinLength)
				{
				GCSeg SuccSeg;
				SuccSeg.GenomeIndex = B.GenomeIndex;
				SuccSeg.ChrId = B.ChrId;
				SuccSeg.lo = s.hi + 1;
				SuccSeg.hi = B.hi;
				SuccSeg.Plus = B.Plus;
				InsertList.push_back(SuccSeg);
				}
			}

	// 2. Delete B if s covers all of B
		else if (Covers(s, B, CovFract))
			DeleteList.push_back(SegPos);

		// 3. Reduce size of B if B is partially covered by s
		else
			{
			if (s.hi > B.lo && s.hi < B.hi && B.hi - s.hi > MinLength)
				{
				DeleteList.push_back(SegPos);

				GCSeg NewSeg = B;
				NewSeg.lo = s.hi;
				InsertList.push_back(NewSeg);
				}
			}

		++p;
		if (p == m_Segs.end())
			break;
		}

// Updates: first delete...
	for (std::list<coord_t>::const_iterator p = DeleteList.begin();
	  p != DeleteList.end(); ++p)
		{
		coord_t Pos = *p;
		DeleteSeg(Pos);
		}

// ...then insert.
	for (std::list<GCSeg>::const_iterator p = InsertList.begin();
	  p != InsertList.end(); ++p)
		{
		const GCSeg &A = *p;
		SegMapValue v(A.GetMidpoint(), A);
		m_Segs.insert(v);
		}
	}

// Compute the coverage of segment s versus the bag.
// This is the number of bases in s covered
// by segments in the bag.
coord_t SegBag::Coverage(const GCSeg &s) const
	{
	const coord_t slen = s.GetLength();
	if (m_Segs.size() == 0 || s.GetLength() == 0)
		return slen;

	SegMapCIter p;
	bool Found = GetNearestSeg(s.lo, &p);
	assert(Found);
	if (!Found)
		return slen;

	coord_t Coverage = 0;

	for (; p != m_Segs.end(); ++p)
		{
		GCSeg B = p->second;
		if (B.lo > s.hi)
			break;

	// Truncate segment if partial overlap
		coord_t lo = std::max(B.lo, s.lo);
		coord_t hi = std::min(B.hi, s.hi);

	// Nearest gap may not overlap
		if (hi < lo)
			continue;

		coord_t len = hi - lo + 1;
		Coverage += len;
		}

	assert(Coverage <= slen);
	return Coverage;
	}
