#include "types.h"
#include "revfile.h"
#include "getopts.h"
#include "utils.h"
#include <string>

class CompostStats
{
	public:

	std::string GenomeName;
	uint64 nuc[256];
	uint64 total;
	uint64 dinuc[256][256];

	CompostStats() { }

	void SetStats(const std::string &Name, const uint64 n[256], const uint64 d[256][256], uint64 tot)
	{
		for (unsigned i = 0; i < 256; ++i)
			nuc[i] = n[i];

		for (unsigned i = 0; i < 256; ++i)
			for (unsigned j = 0; j < 256; ++j)
				dinuc[i][j] = d[i][j];

		total = tot;
		GenomeName = Name;
	}
};

static void GetCompostStats(const RevFile &InFile, CompostStats &Stats)
{
	uint64 nuc[256] = {0};
	uint64 dinuc[256][256] = {{0}};

	if (InFile.GetGenomeCount() != 1)
		Die("This program requires one genome per revfile");

	unsigned ChrCount = InFile.GetChrCount(0);

	if (InFile.GetSeqCount() != ChrCount)
		Die("Missing sequence in revfile");

	for (unsigned int i = 0; i < ChrCount; ++i)
	{
		unsigned Len = InFile.GetSeqLengthByIndex(i);
		unsigned GenomeIndex = 0;
		unsigned ChrId = 0;
		unsigned char *Seq = new unsigned char[Len];
		InFile.GetSeqByIndex(i, (char *)Seq, Len, &GenomeIndex, &ChrId);

		nuc[Seq[0] & 0xDF]++;
		for (unsigned j = 1; j < Len; ++j)
		{
			nuc[Seq[j] & 0xDF]++;
			dinuc[Seq[j-1] & 0xDF][Seq[j] & 0xDF]++;
		}
		delete[] Seq;
	}

	Stats.SetStats(InFile.GetGenomeName(0), nuc, dinuc, (uint64)InFile.GetGenomeInfo(0).GetLength64());
}

static std::string GetPctChg(double From, double To)
{
	if (From == 0)
	{
		if (To == 0)
			return std::string("+0.0%");
		else
			return std::string("inf");
	}
	double pc = (To - From)*100.0/From;
	char buf[20];
	sprintf(buf, "%+6.1f%%", pc);
	return std::string(buf);
}

static void LogCompostLine(std::string Title, double CountA, double TotalA, double CountB, double TotalB)
{
	double PctA = CountA * 100.0 / TotalA;
	double PctB = CountB * 100.0 / TotalB;
	double dBases = CountB - CountA;
	double dPct = PctB - PctA;
	std::string PctChg = GetPctChg(CountA, CountB);

	Log("%6s  %10.0f  %5.1f%%  %10.0f  %5.1f%%  %+10.0f  %+5.1f%%  %7s\n",
				 Title.c_str(), CountA, PctA, CountB, PctB, dBases, dPct, PctChg.c_str());
}

static unsigned IsSelfDual(const std::string &s)
{
	return s == "AT" || s == "TA" || s == "GC" || s == "CG";
}

static void ShowCompost2(const char *InFileNameA, const char *InFileNameB)
{
	RevFile InFileA;
	RevFile InFileB;
	InFileA.Open(InFileNameA);
	InFileB.Open(InFileNameB);

// Allocating a CompostStats object on the stack crashes under Microsoft C++,
// so allocate dynamically.
	static CompostStats &StatsA = *new CompostStats;
	static CompostStats &StatsB = *new CompostStats;
	GetCompostStats(InFileA, StatsA);
	GetCompostStats(InFileB, StatsB);

	Log("\n");
	Log("\n");
	Log("Nucleotide Composition Report\n");
	Log("-----------------------------\n");
	Log("\n");
	Log("Genome 1: %s\n", InFileA.GetGenomeName(0).c_str());
	Log("Genome 2: %s\n", InFileB.GetGenomeName(0).c_str());
	Log("\n");
	Log("\n");
	Log("           Genome1    Gen1.    Genome2    Gen2.                            \n");
	Log("Letter       Bases     Pct       Bases     Pct      dBases    dPct    %% Chg\n");
	Log("======  ==========  ======  ==========  ======  ==========  ======  =======\n");

	for (unsigned int i = 0; i < 256; ++i)
		if ((StatsA.nuc[i] != 0) || (StatsB.nuc[i] != 0) || i == 'A' || i == 'C' || i == 'G' || i == 'T')
		{
			LogCompostLine(
				std::string(1, (unsigned char)i),
				double(StatsA.nuc[i]),
				double(StatsA.total),
				double(StatsB.nuc[i]),
				double(StatsB.total)
			);
		}

	Log("\n");
	LogCompostLine(
		"A+T",
		double(StatsA.nuc[(unsigned char)'A']) + double(StatsA.nuc[(unsigned char)'T']),
		double(StatsA.total),
		double(StatsB.nuc[(unsigned char)'A']) + double(StatsB.nuc[(unsigned char)'T']),
		double(StatsB.total)
	);
	LogCompostLine(
		"G+C",
		double(StatsA.nuc[(unsigned char)'G']) + double(StatsA.nuc[(unsigned char)'C']),
		double(StatsA.total),
		double(StatsB.nuc[(unsigned char)'G']) + double(StatsB.nuc[(unsigned char)'C']),
		double(StatsB.total)
	);

	Log("\n");
	LogCompostLine("Total", (double) StatsA.total, (double) StatsA.total,
	  (double) StatsB.total, (double) StatsB.total);

	Log("\n");
	Log("\n");
	Log("           Genome1    Gen1.    Genome2    Gen2.                            \n");
	Log(" Dinuc       Count     Pct       Count     Pct      dCount    dPct    %% Chg\n");
	Log("======  ==========  ======  ==========  ======  ==========  ======  =======\n");

	double TotA = 0;
	double TotB = 0;
	for (unsigned int i = 0; i < 256; ++i)
		for (unsigned int j = 0; j < 256; ++j)
		{
			TotA += StatsA.dinuc[i][j];
			TotB += StatsB.dinuc[i][j];
		}

	for (unsigned SelfDual = 0; SelfDual < 2; ++SelfDual)
	{
		if (SelfDual > 0)
			Log("\n");
		for (unsigned int i = 0; i < 256; ++i)
		{
			for (unsigned int j = 0; j < 256; ++j)
			{
				if (((i == 'A' || i == 'C' || i == 'G' || i == 'T') && (j == 'A' || j == 'C' || j == 'G' || j == 'T'))
				   || ((StatsA.dinuc[i][j] != 0) && (StatsB.dinuc[i][j] != 0)))
				{
					std::string s(1, (unsigned char)i);
					s += (unsigned char)j;
					if (SelfDual == IsSelfDual(s))
						LogCompostLine(s, double(StatsA.dinuc[i][j]), TotA, double(StatsB.dinuc[i][j]), TotB);
				}
			}
		}
	}

	Log("\n");
	LogCompostLine("AA+TT",
		double(StatsA.dinuc[(unsigned char)'A'][(unsigned char)'A']) + double(StatsA.dinuc[(unsigned char)'T'][(unsigned char)'T']), TotA,
		double(StatsB.dinuc[(unsigned char)'A'][(unsigned char)'A']) + double(StatsB.dinuc[(unsigned char)'T'][(unsigned char)'T']), TotB);
	LogCompostLine("AC+GT",
		double(StatsA.dinuc[(unsigned char)'A'][(unsigned char)'C']) + double(StatsA.dinuc[(unsigned char)'G'][(unsigned char)'T']), TotA,
		double(StatsB.dinuc[(unsigned char)'A'][(unsigned char)'C']) + double(StatsB.dinuc[(unsigned char)'G'][(unsigned char)'T']), TotB);
	LogCompostLine("AG+CT",
		double(StatsA.dinuc[(unsigned char)'A'][(unsigned char)'G']) + double(StatsA.dinuc[(unsigned char)'C'][(unsigned char)'T']), TotA,
		double(StatsB.dinuc[(unsigned char)'A'][(unsigned char)'G']) + double(StatsB.dinuc[(unsigned char)'C'][(unsigned char)'T']), TotB);
	LogCompostLine("CA+TG",
		double(StatsA.dinuc[(unsigned char)'C'][(unsigned char)'A']) + double(StatsA.dinuc[(unsigned char)'T'][(unsigned char)'G']), TotA,
		double(StatsB.dinuc[(unsigned char)'C'][(unsigned char)'A']) + double(StatsB.dinuc[(unsigned char)'T'][(unsigned char)'G']), TotB);
	LogCompostLine("CC+GG",
		double(StatsA.dinuc[(unsigned char)'C'][(unsigned char)'C']) + double(StatsA.dinuc[(unsigned char)'G'][(unsigned char)'G']), TotA,
		double(StatsB.dinuc[(unsigned char)'C'][(unsigned char)'C']) + double(StatsB.dinuc[(unsigned char)'G'][(unsigned char)'G']), TotB);
	LogCompostLine("GA+TC",
		double(StatsA.dinuc[(unsigned char)'G'][(unsigned char)'A']) + double(StatsA.dinuc[(unsigned char)'T'][(unsigned char)'C']), TotA,
		double(StatsB.dinuc[(unsigned char)'G'][(unsigned char)'A']) + double(StatsB.dinuc[(unsigned char)'T'][(unsigned char)'C']), TotB);
	Log("\n");
	Log("\n");

	InFileA.Close();
	InFileB.Close();
}

void ShowCompost()
{
	const char *InFileNameA = RequiredValueOpt("compost1");
	const char *InFileNameB = RequiredValueOpt("compost2");

	ShowCompost2(InFileNameA, InFileNameB);
}

