#ifndef TIDS_H
#define TIDS_H

#include <vector>
#include <algorithm>
#include <map>
#include "revfile.h"

typedef std::vector<std::vector<unsigned> > OffsetsVec;

typedef std::map<int, std::map<int, int> > TMap;

void FillOffsets(OffsetsVec &Offsets, const RevFile &Rev, unsigned TIDLEN)
{
	unsigned GenomeCount = Rev.GetGenomeCount();
	Offsets.clear();
	Offsets.resize(GenomeCount);
	for (unsigned GenomeIndex = 0; GenomeIndex != GenomeCount; ++GenomeIndex)
	{
		const GenInfo &GI = Rev.GetGenomeInfo(GenomeIndex);
		unsigned ChrCount = GI.GetChrCount();
		unsigned CurOffset = 0;
		Offsets[GenomeIndex].resize(ChrCount);
		for (unsigned ChrIndex = 0; ChrIndex != ChrCount; ++ChrIndex)
		{
			unsigned ChrId = GI.GetChrId(ChrIndex);
			unsigned ChrLength = GI.GetChrInfo(ChrId).GetLength();
			if (ChrId >= ChrCount)
				Die("Sparse ChrId index");
			Offsets[GenomeIndex][ChrId] = CurOffset;
			unsigned Diff = ChrLength / TIDLEN;
			if (Diff * TIDLEN < ChrLength)
				++Diff;
			++Diff;
			CurOffset += Diff;
		}
	}
}

void FillMap(OffsetsVec &Offsets, TMap &Map, const RevFile &Rev, int RefGenome, unsigned TIDLEN)
{
	FillOffsets(Offsets, Rev, TIDLEN);

	unsigned AlnCount = Rev.GetAlnInfoCount();

	ProgressStart("Filling map");
	for (unsigned i = 0; i != AlnCount; ++i)
	{
		if ((i & 0xfffff) == 0)
			ProgressStep(i, AlnCount);

		AlnInfo AI;
		std::vector<std::string> Rows;

		Rev.GetAlnInfo(Rev.GetAlnRows(i, Rows), AI);

		if (Rows.size() != 2)
			Die("Alignment of !=2 rows");

		int AIScore = (AI.Score > 10) ? (int)AI.Score : 10;
		if (AI.Segs[1].Plus)
		{
			unsigned Offs[2];
			unsigned Res[2];
			for (unsigned j = 0; j != 2; ++j)
			{
				Offs[j] = Offsets[AI.Segs[j].GenomeIndex][AI.Segs[j].ChrId] + AI.Segs[j].lo / TIDLEN;
				Res[j] = AI.Segs[j].lo % TIDLEN;
			}

			for (unsigned j = 0; j != Rows[0].size(); ++j)
			{
				for (unsigned k = 0; k != 2; ++k)
				{
					if (Rows[k][j] != '-')
					{
						++Res[k];
						if (Res[k] == TIDLEN)
						{
							Res[k] = 0;
							++Offs[k];
						}
					}
				}
				Map[Offs[RefGenome]][int(Offs[1-RefGenome])] = AIScore;
			}
		}
		else
		{
			unsigned Offs[2];
			unsigned Res[2];
			Offs[0] = Offsets[AI.Segs[0].GenomeIndex][AI.Segs[0].ChrId] + AI.Segs[0].lo / TIDLEN;
			Res[0] = AI.Segs[0].lo % TIDLEN;
			Offs[1] = Offsets[AI.Segs[1].GenomeIndex][AI.Segs[1].ChrId] + AI.Segs[1].hi / TIDLEN;
			Res[1] = TIDLEN - (AI.Segs[1].hi % TIDLEN) - 1;

			for (unsigned j = 0; j != Rows[0].size(); ++j)
			{
				if (Rows[0][j] != '-')
				{
					++Res[0];
					if (Res[0] == TIDLEN)
					{
						Res[0] = 0;
						++Offs[0];
					}
				}
				if (Rows[1][j] != '-')
				{
					++Res[1];
					if (Res[1] == TIDLEN)
					{
						Res[1] = 0;
						--Offs[1];
					}
				}
				Map[Offs[RefGenome]][int(-Offs[1-RefGenome])] = AIScore;
			}
		}
	}
	ProgressDone();
}

void RevToKF(const RevFile &Rev, KFilePair &KF, unsigned RefGenome, unsigned TIDLEN)
{
	OffsetsVec Offsets;

	FillOffsets(Offsets, Rev, TIDLEN);

	unsigned AlnCount = Rev.GetAlnInfoCount();

	ProgressStart("Converting");
	for (unsigned i = 0; i != AlnCount; ++i)
	{
		if ((i & 0xfffff) == 0)
			ProgressStep(i, AlnCount);

		AlnInfo AI;
		std::vector<std::string> Rows;

		Rev.GetAlnInfo(Rev.GetAlnRows(i, Rows), AI);

		if (Rows.size() != 2)
			Die("Alignment of !=2 rows");

		int AIScore = (AI.Score > 10) ? (int)AI.Score : 10;
		unsigned Offs[2];
		unsigned Res[2];
		bool Dirty = true;
		int RefIndex = (AI.Segs[0].GenomeIndex == RefGenome) ? 0 : 1;
		if (AI.Segs[0].Plus == false)
		{
			std::reverse(Rows[0].begin(), Rows[0].end());
			std::reverse(Rows[1].begin(), Rows[1].end());
		}
		if (AI.Segs[0].Plus == AI.Segs[1].Plus)
		{
			for (unsigned j = 0; j != 2; ++j)
			{
				Offs[j] = Offsets[AI.Segs[j].GenomeIndex][AI.Segs[j].ChrId] + AI.Segs[j].lo / TIDLEN;
				Res[j] = AI.Segs[j].lo % TIDLEN;
			}

			for (unsigned j = 0; j != Rows[0].size(); ++j)
			{
				for (unsigned k = 0; k != 2; ++k)
				{
					if (Rows[k][j] != '-')
					{
						++Res[k];
						if (Res[k] == TIDLEN)
						{
							Res[k] = 0;
							++Offs[k];
							Dirty = true;
						}
					}
				}
				if (Dirty)
				{
					KPair KP;
					KP.k[0] = Offs[RefIndex];
					KP.k[1] = Offs[1-RefIndex];
					KP.score = AIScore;
					KF.AddEntry(KP);
				}
				Dirty = false;
			}
		}
		else
		{
			Offs[0] = Offsets[AI.Segs[0].GenomeIndex][AI.Segs[0].ChrId] + AI.Segs[0].lo / TIDLEN;
			Res[0] = AI.Segs[0].lo % TIDLEN;
			Offs[1] = Offsets[AI.Segs[1].GenomeIndex][AI.Segs[1].ChrId] + AI.Segs[1].hi / TIDLEN;
			Res[1] = TIDLEN - (AI.Segs[1].hi % TIDLEN) - 1;

			for (unsigned j = 0; j != Rows[0].size(); ++j)
			{
				if (Rows[0][j] != '-')
				{
					++Res[0];
					if (Res[0] == TIDLEN)
					{
						Res[0] = 0;
						++Offs[0];
						Dirty = true;
					}
				}
				if (Rows[1][j] != '-')
				{
					++Res[1];
					if (Res[1] == TIDLEN)
					{
						Res[1] = 0;
						--Offs[1];
						Dirty = true;
					}
				}
				if (Dirty)
				{
					KPair KP;
					KP.k[0] = Offs[RefIndex];
					KP.k[1] = -Offs[1-RefIndex];
					KP.score = AIScore;
					KF.AddEntry(KP);
				}
				Dirty = false;
			}
		}
	}
	ProgressDone();
}



#endif
