#include <vector>
#include <map>
#include <string>
#include <iterator>
#include <algorithm>
#include "revfile.h"
#include "intlist.h"
#include "types.h"
#include "utils.h"
#include "getopts.h"
#include "progress.h"

/* Some globals */

/* The input files */
RevFile InFiles[2];

/* The output file */
RevFile OutFile;

/* RefGenome[i] is the index of the 'reference genome' in the i-th revfile */
int RefGenome[2];
/* OtherGenome[i] is the index of the 'other genome' in the i-th revfile */
int OtherGenome[2];
/* Since each file is expected to have two genomes, RefGenome[i] + OtherGenome[i] == 1 */

bool g_NoParalogy = false;

/* End of important globals */

void RevComp(std::string &Seq);

VALUE_OPT ValueOpts[] =
{
	{ "in1",				0 },
	{ "in2",				0 },
	{ "out",				0 },
	{ "log",				0 }
};
int ValueOptCount = sizeof(ValueOpts)/sizeof(ValueOpts[0]);

FLAG_OPT FlagOpts[] =
{
	{ "noparalogy",			0 }
};
int FlagOptCount = sizeof(FlagOpts)/sizeof(FlagOpts[0]);

void Usage()
{
	fprintf(stderr,
"Usage:\n"
"    transalign -in1 <revfile> -in2 <revfile> -out <revfile> [-noparalogy]\n"
"\n"
	);
}

class SeqAIX
{
	public:

	/* AlnInfo and Sequence */
	AlnInfo AI;
	StdStrVec Rows;
};

/* Add an output alignment to the output file */
void OutputAln(StdStrVec &OutRows, coord_t (&OtherStartOffset)[2], coord_t (&OtherStopOffset)[2], const GCSeg &OtherSeg0, const GCSeg &OtherSeg1, bool (&OtherPlus)[2], std::vector<AlnAttr> &AlnAttrs, std::vector<unsigned> &AlnVals)
{
	AlnInfo AI;
	AI.Score = 0;
	AI.Segs.resize(2);
	AI.Segs[0].GenomeIndex = 0;
	AI.Segs[1].GenomeIndex = 1;
	AI.Segs[0].ChrId = OtherSeg0.ChrId;
	AI.Segs[1].ChrId = OtherSeg1.ChrId;
	AI.Segs[0].Plus = true; //Strand normalization
	AI.Segs[1].Plus = (OtherPlus[0] == OtherPlus[1]);
	AI.Attrs = AlnAttrs;
	AI.Values = AlnVals;
	if (!OtherPlus[0])
	{
		RevComp(OutRows[0]);
		RevComp(OutRows[1]);
		AI.Segs[0].hi = OtherSeg0.hi - OtherStartOffset[0];
		AI.Segs[0].lo = OtherSeg0.hi - OtherStopOffset[0] + 1;
	}
	else
	{
		AI.Segs[0].lo = OtherSeg0.lo + OtherStartOffset[0];
		AI.Segs[0].hi = OtherSeg0.lo + OtherStopOffset[0] - 1;
	}
	if (OtherPlus[1])
	{
		AI.Segs[1].lo = OtherSeg1.lo + OtherStartOffset[1];
		AI.Segs[1].hi = OtherSeg1.lo + OtherStopOffset[1] - 1;
	}
	else
	{
		AI.Segs[1].hi = OtherSeg1.hi - OtherStartOffset[1];
		AI.Segs[1].lo = OtherSeg1.hi - OtherStopOffset[1] + 1;
	}

	OutFile.AddAln(AI, OutRows);
}

/* Transitively combine two alignments */
void ProcessAln(std::list<SeqAIX>::iterator ptr0, std::list<SeqAIX>::iterator ptr1)
{
	const GCSeg &RefSeg0 = ptr0->AI.GetSeg(RefGenome[0]);
	const GCSeg &RefSeg1 = ptr1->AI.GetSeg(RefGenome[1]);
	const GCSeg &OtherSeg0 = ptr0->AI.GetSeg(OtherGenome[0]);
	const GCSeg &OtherSeg1 = ptr1->AI.GetSeg(OtherGenome[1]);
	bool OtherPlus[2] = { OtherSeg0.Plus == RefSeg0.Plus, OtherSeg1.Plus == RefSeg1.Plus };
	std::string &RefRow0 = ptr0->Rows[ptr0->AI.GetSegIndex(RefGenome[0])];
	std::string &RefRow1 = ptr1->Rows[ptr1->AI.GetSegIndex(RefGenome[1])];
	std::string &OtherRow0 = ptr0->Rows[ptr0->AI.GetSegIndex(OtherGenome[0])];
	std::string &OtherRow1 = ptr1->Rows[ptr1->AI.GetSegIndex(OtherGenome[1])];
	StdStrVec OutRows(2);

	/* Make sure they overlap */
	if ((RefSeg0.lo > RefSeg1.hi) || (RefSeg1.lo > RefSeg0.hi))
		return;

	/* Prepare union of attributes */
	std::vector<AlnAttr> OutAlnAttrs;
	std::vector<unsigned> OutAlnVals;
	{
		std::map<AlnAttr, unsigned> AttrMap;
		bool IsParalog = false;
		for (unsigned i = 0; i != ptr0->AI.Attrs.size(); ++i)
		{
			AttrMap[ptr0->AI.Attrs[i]] = ptr0->AI.Values[i];
			if ((ptr0->AI.Attrs[i] == ALN_ATTR_Paralog) && (ptr0->AI.Values[i] == 1))
				IsParalog = true;
		}
		for (unsigned i = 0; i != ptr1->AI.Attrs.size(); ++i)
		{
			AttrMap[ptr1->AI.Attrs[i]] = ptr1->AI.Values[i];
			if ((ptr1->AI.Attrs[i] == ALN_ATTR_Paralog) && (ptr1->AI.Values[i] == 1))
				IsParalog = true;
		}
		/* Paralog heuristic */
		if (IsParalog)
		{
			if (g_NoParalogy)
				return;
			AttrMap[ALN_ATTR_Paralog] = 1;
		}
		for (std::map<AlnAttr, unsigned>::const_iterator i = AttrMap.begin(); i != AttrMap.end(); ++i)
		{
			OutAlnAttrs.push_back(i->first);
			OutAlnVals.push_back(i->second);
		}
	}

	coord_t RefLo[2] = { RefSeg0.lo, RefSeg1.lo };
	coord_t RefHi[2] = { RefSeg0.hi, RefSeg1.hi };
	coord_t OtherStartOffset[2] = {0,0}, OtherStopOffset[2] = {0,0};
	unsigned Col[2] = { 0, 0 };
	unsigned EndCol[2] = { 0, 0 };

	/* Syncronize the beginnings */
	while (RefLo[0] < RefLo[1])
	{
		if (RefRow0[Col[0]] != '-')
			++RefLo[0];
		if (OtherRow0[Col[0]] != '-')
			++OtherStartOffset[0];
		++Col[0];
	}
	while (RefLo[1] < RefLo[0])
	{
		if (RefRow1[Col[1]] != '-')
			++RefLo[1];
		if (OtherRow1[Col[1]] != '-')
			++OtherStartOffset[1];
		++Col[1];
	}

	if (RefHi[0] > RefHi[1])
		RefHi[0] = RefHi[1];
	else
		RefHi[1] = RefHi[0];

	RefHi[0] = RefHi[0] - RefLo[0] + 1;
	RefHi[1] = RefHi[1] - RefLo[1] + 1;

	EndCol[0] = Col[0];
	EndCol[1] = Col[1];

	while (RefHi[0] > 0)
	{
		if (RefRow0[EndCol[0]] != '-')
			--RefHi[0];
		++EndCol[0];
	}

	while (RefHi[1] > 0)
	{
		if (RefRow1[EndCol[1]] != '-')
			--RefHi[1];
		++EndCol[1];
	}

	bool AvoidingBubble = true;

	OtherStopOffset[0] = OtherStartOffset[0];
	OtherStopOffset[1] = OtherStartOffset[1];


	/* Run through the alignment */
	while ((Col[0] != EndCol[0]) && (Col[1] != EndCol[1]))
	{
		char r0 = RefRow0[Col[0]];
		char r1 = RefRow1[Col[1]];
		char o0 = OtherRow0[Col[0]];
		char o1 = OtherRow1[Col[1]];
		if ((r0 != '-') && (r1 != '-'))
		{
			/* Ungapped in both */
			assert(r0 == r1);
			if (o0 != '-')
				++OtherStopOffset[0];
			if (o1 != '-')
				++OtherStopOffset[1];
			if ((o0 != '-') || (o1 != '-'))
			{
				OutRows[0].push_back(o0);
				OutRows[1].push_back(o1);
			}
			++Col[0];
			++Col[1];
			AvoidingBubble = false;
		}
		else if ((r0 == '-') && (r1 != '-'))
		{
			/* Gapped in one */
			assert(o0 != '-');
			OutRows[0].push_back(o0);
			OutRows[1].push_back('-');
			++OtherStopOffset[0];
			++Col[0];
			if (AvoidingBubble)
			{
				OtherStartOffset[0] = OtherStopOffset[0];
				OtherStartOffset[1] = OtherStopOffset[1];
				OutRows[0].clear();
				OutRows[1].clear();
			}
		}
		else if ((r0 != '-') && (r1 == '-'))
		{
			/* Gapped in the other */
			assert(o1 != '-');
			OutRows[0].push_back('-');
			OutRows[1].push_back(o1);
			++OtherStopOffset[1];
			++Col[1];
			if (AvoidingBubble)
			{
				OtherStartOffset[0] = OtherStopOffset[0];
				OtherStartOffset[1] = OtherStopOffset[1];
				OutRows[0].clear();
				OutRows[1].clear();
			}
		}
		else
		{
			/* Gap in both reference sequences; this is a bubble, so the alignment is broken. */
			if ((OtherStopOffset[0] > OtherStartOffset[0]) && (OtherStopOffset[1] > OtherStartOffset[1]))
				OutputAln(OutRows, OtherStartOffset, OtherStopOffset, OtherSeg0, OtherSeg1, OtherPlus, OutAlnAttrs, OutAlnVals);
			assert(o0 != '-');
			assert(o1 != '-');
			++OtherStopOffset[0];
			++OtherStopOffset[1];
			++Col[0];
			++Col[1];
			OtherStartOffset[0] = OtherStopOffset[0];
			OtherStartOffset[1] = OtherStopOffset[1];
			OutRows[0].clear();
			OutRows[1].clear();
			AvoidingBubble = true;
		}
	}
	if ((OtherStopOffset[0] > OtherStartOffset[0]) && (OtherStopOffset[1] > OtherStartOffset[1]))
		OutputAln(OutRows, OtherStartOffset, OtherStopOffset, OtherSeg0, OtherSeg1, OtherPlus, OutAlnAttrs, OutAlnVals);
}

/* Compare alignments of a fixed chromosome pair */
void CombineAlignments(IntPairVec (&AlnInfoIndices)[2])
{
	/* Make vector of <RefGenomeLoCoord, AlnInfoIndex> */
	typedef std::pair<coord_t, unsigned> LoAIIx;
	std::vector<LoAIIx> LoAIIxVec[2];
	for (int i = 0; i != 2; ++i)
	{
		unsigned Count = 0;
		for (IntPairVec::const_iterator j = AlnInfoIndices[i].begin(); j != AlnInfoIndices[i].end(); ++j)
		{
			Count += j->second - j->first + 1;
		}
		LoAIIxVec[i].reserve(Count);
	}

	Progress("Reading and Sorting AlnInfos...");
	/* Populate vector */
	AlnInfo AI;
	for (int i = 0; i != 2; ++i)
	{
		for (IntPairVec::const_iterator j = AlnInfoIndices[i].begin(); j != AlnInfoIndices[i].end(); ++j)
		{
			for (unsigned k = j->first; k <= j->second; ++k)
			{
				InFiles[i].GetAlnInfo(k, AI);
				LoAIIxVec[i].push_back(LoAIIx(AI.GetSeg(RefGenome[i]).lo, k));
			}
		}
		AlnInfoIndices[i].clear();
		std::sort(LoAIIxVec[i].begin(), LoAIIxVec[i].end());
	}

	ProgressStart("Combining alignments...");
	/* Go through all the alignments, sorted by the reference genome coordinate. The following
	 * algorithm sweeps through the reference genome and keeps two sets (a set of alignments
	 * overlapping the current position for each genome). Any new alignment in either genome is
	 * combined with the set of alignments in the other genome.
	 */

	/* Pointer to the next alignment in each revfile */
	std::vector<std::pair<unsigned, coord_t> >::iterator nextptr[2] = {LoAIIxVec[0].begin(), LoAIIxVec[1].begin()};
	/* The alignments overlapping the current coordinate */
	std::list<SeqAIX> OverlapAIXs[2];
	/* Coordinate of next (closest, in increasing lo coordinate) alignment */
	coord_t NextCoord;

	while ((nextptr[0] != LoAIIxVec[0].end()) && (nextptr[1] != LoAIIxVec[1].end()))
	{
		/* Get next coordinate */
		NextCoord = std::min(nextptr[0]->first, nextptr[1]->first);

		if (((nextptr[0] - LoAIIxVec[0].begin()) & 0xfff) == 0)
			ProgressStep(nextptr[0] - LoAIIxVec[0].begin(), LoAIIxVec[0].size());

		/* Retrieve alignments that start at that coordinate (there could be many for each file) */
		std::list<SeqAIX> NewAIXs[2];
		for (int i = 0; i != 2; ++i)
		{
			while (NextCoord == nextptr[i]->first)
			{
				NewAIXs[i].push_front(SeqAIX());
				unsigned AlnIndex = InFiles[i].GetAlnInfo(nextptr[i]->second, NewAIXs[i].front().AI);
				if (AlnIndex == (uint32)-1)
					Die("Alignment missing for AlnInfoIndex %u\n", nextptr[i]->second);
				InFiles[i].GetAlnRows(AlnIndex, NewAIXs[i].front().Rows);
				if (NewAIXs[i].front().AI.GetSeg(RefGenome[i]).Plus == false)
				{
					/* Warning: This doesn't change the Plus to true since set members are const.
					 * But implicitly, after this all RefGenome segments are always Plus; and
					 * OtherGenome segments are "!(RefGenome.Plus XOR OtherGenome.Plus)" */
					RevComp(NewAIXs[i].front().Rows[0]);
					RevComp(NewAIXs[i].front().Rows[1]);
				}
				++nextptr[i];
				if (nextptr[i] == LoAIIxVec[i].end())
					break;
			}
		}

		/* First, process all pairs within NewAIXs */
		for (std::list<SeqAIX>::iterator ptr0 = NewAIXs[0].begin(); ptr0 != NewAIXs[0].end(); ++ptr0)
			for (std::list<SeqAIX>::iterator ptr1 = NewAIXs[1].begin(); ptr1 != NewAIXs[1].end(); ++ptr1)
				ProcessAln(ptr0, ptr1);

		/* Then, process all pairs of NewAIXs vs OverlapAIXs */
		for (std::list<SeqAIX>::iterator ptr0 = NewAIXs[0].begin(); ptr0 != NewAIXs[0].end(); ++ptr0)
			for (std::list<SeqAIX>::iterator ptr1 = OverlapAIXs[1].begin(); ptr1 != OverlapAIXs[1].end(); ++ptr1)
				ProcessAln(ptr0, ptr1);
		for (std::list<SeqAIX>::iterator ptr0 = OverlapAIXs[0].begin(); ptr0 != OverlapAIXs[0].end(); ++ptr0)
			for (std::list<SeqAIX>::iterator ptr1 = NewAIXs[1].begin(); ptr1 != NewAIXs[1].end(); ++ptr1)
				ProcessAln(ptr0, ptr1);

		/* Finally, update OverlapAIXs */
		for (int i = 0; i != 2; ++i)
		{
			std::list<SeqAIX>::iterator ptr = OverlapAIXs[i].begin();
			while (ptr != OverlapAIXs[i].end())
			{
				if (ptr->AI.GetSeg(RefGenome[i]).hi < NextCoord)
					ptr = OverlapAIXs[i].erase(ptr);
				else
					++ptr;
			}
			OverlapAIXs[i].splice(OverlapAIXs[i].end(), NewAIXs[i]);
		}
	}

	/* Do the same for the remaining of a file, even if the other is exhausted, with some optimizations */
	while ((nextptr[0] != LoAIIxVec[0].end()) && (!(OverlapAIXs[1].empty())))
	{
		NextCoord = nextptr[0]->first;
		std::list<SeqAIX> NewAIXs[2];

		/* Fill NewAIXs[0] */
		{
			int i = 0;
			while (NextCoord == nextptr[i]->first)
			{
				NewAIXs[i].push_front(SeqAIX());
				unsigned AlnIndex = InFiles[i].GetAlnInfo(nextptr[i]->second, NewAIXs[i].front().AI);
				if (AlnIndex == (uint32)-1)
					Die("Alignment missing for AlnInfoIndex %u\n", nextptr[i]->second);
				InFiles[i].GetAlnRows(AlnIndex, NewAIXs[i].front().Rows);
				if (NewAIXs[i].front().AI.GetSeg(RefGenome[i]).Plus == false)
				{
					RevComp(NewAIXs[i].front().Rows[0]);
					RevComp(NewAIXs[i].front().Rows[1]);
				}
				++nextptr[i];
				if (nextptr[i] == LoAIIxVec[i].end())
					break;
			}
		}

		for (std::list<SeqAIX>::iterator ptr0 = NewAIXs[0].begin(); ptr0 != NewAIXs[0].end(); ++ptr0)
			for (std::list<SeqAIX>::iterator ptr1 = OverlapAIXs[1].begin(); ptr1 != OverlapAIXs[1].end(); ++ptr1)
				ProcessAln(ptr0, ptr1);

		/* Update OverlapAIXs[1] */
		{
			int i = 1;
			std::list<SeqAIX>::iterator ptr = OverlapAIXs[i].begin();
			while (ptr != OverlapAIXs[i].end())
			{
				if (ptr->AI.GetSeg(RefGenome[i]).hi < NextCoord)
					ptr = OverlapAIXs[i].erase(ptr);
				else
					++ptr;
			}
		}
	}

	while ((nextptr[1] != LoAIIxVec[1].end()) && (!(OverlapAIXs[0].empty())))
	{
		NextCoord = nextptr[1]->first;
		std::list<SeqAIX> NewAIXs[2];

		/* Fill NewAIXs[1] */
		{
			int i = 1;
			while (NextCoord == nextptr[i]->first)
			{
				NewAIXs[i].push_front(SeqAIX());
				unsigned AlnIndex = InFiles[i].GetAlnInfo(nextptr[i]->second, NewAIXs[i].front().AI);
				if (AlnIndex == (uint32)-1)
					Die("Alignment missing for AlnInfoIndex %u\n", nextptr[i]->second);
				InFiles[i].GetAlnRows(AlnIndex, NewAIXs[i].front().Rows);
				if (NewAIXs[i].front().AI.GetSeg(RefGenome[i]).Plus == false)
				{
					RevComp(NewAIXs[i].front().Rows[0]);
					RevComp(NewAIXs[i].front().Rows[1]);
				}
				++nextptr[i];
				if (nextptr[i] == LoAIIxVec[i].end())
					break;
			}
		}

		for (std::list<SeqAIX>::iterator ptr0 = OverlapAIXs[0].begin(); ptr0 != OverlapAIXs[0].end(); ++ptr0)
			for (std::list<SeqAIX>::iterator ptr1 = NewAIXs[1].begin(); ptr1 != NewAIXs[1].end(); ++ptr1)
				ProcessAln(ptr0, ptr1);

		/* Update OverlapAIXs[0] */
		{
			int i = 0;
			std::list<SeqAIX>::iterator ptr = OverlapAIXs[i].begin();
			while (ptr != OverlapAIXs[i].end())
			{
				if (ptr->AI.GetSeg(RefGenome[i]).hi < NextCoord)
					ptr = OverlapAIXs[i].erase(ptr);
				else
					++ptr;
			}
		}
	}
	ProgressDone();
}

/* Choose the reference genome */
void FindRefGenome()
{
	if ((InFiles[0].GetGenomeCount() != 2) || (InFiles[1].GetGenomeCount() != 2))
		Die("This program can only handle two genomes at a time.");

	for (int i = 0; i != 2; ++i)
		for (int j = 0; j != 2; ++j)
			if (InFiles[0].GetGenomeInfo(i) == InFiles[1].GetGenomeInfo(j))
			{
				RefGenome[0] = i;
				RefGenome[1] = j;
				OtherGenome[0] = 1 - i;
				OtherGenome[1] = 1 - j;
				if (InFiles[0].GetGenomeName(1-i) == InFiles[1].GetGenomeName(1-j))
				{
					Warning("Genome name collision! Using '%s' as reference, and renaming '%s' to '%s2'...",
						InFiles[0].GetGenomeName(i).c_str(), InFiles[1].GetGenomeName(1-j).c_str(), InFiles[1].GetGenomeName(1-j).c_str());
				}
				return;
			}

	Die("No common genome between the files.");
}

void TransAlign()
{

	const unsigned RefGenomeChrCount = InFiles[0].GetChrCount(RefGenome[0]);

	/* For every chromosome of the 'reference genome'... */
	for (unsigned ChrIndex = 0; ChrIndex != RefGenomeChrCount; ++ChrIndex)
	{
		/* Retrieve alignments of that reference chromosome */
		IntPairVec AlnInfoIndices[2];
		unsigned ChrId[2];
		for (int idx = 0; idx != 2; ++idx)
		{
			ChrId[idx] = InFiles[idx].GetGenomeInfo(RefGenome[idx]).GetChrId(ChrIndex);
			AlnInfoIndices[idx].reserve(4096);
			InFiles[idx].GetAppendAIIxRangesChr(RefGenome[idx], ChrId[idx], AlnInfoIndices[idx]);
		}
		assert(ChrId[0] == ChrId[1]);

		/* Combine the alignments */
		if ((AlnInfoIndices[0].size() > 0) && (AlnInfoIndices[1].size() > 0))
		{
			Progress("Processing chromosome %.6s...", InFiles[0].GetChrName(RefGenome[0], ChrId[0]).c_str());
			CombineAlignments(AlnInfoIndices);
		}
	}
}

int main(int argc, char *argv[])
{
	ProcessArgVect(argc, argv);
	const char *InFileName[2];
	const char *OutFileName;

	InFileName[0] = RequiredValueOpt("in1");
	InFileName[1] = RequiredValueOpt("in2");
	OutFileName = ValueOpt("out");

	g_NoParalogy = FlagOpt("noparalogy");

	/* Open input Rev Files */
	for (int i = 0; i != 2; ++i)
		InFiles[i].Open(InFileName[i]);

	/* Prepare output Rev File */
	FindRefGenome();
	GenInfoVec OutGenomes;
	for (int i = 0; i != 2; ++i)
		OutGenomes.push_back(InFiles[i].GetGenomes()[OtherGenome[i]]);
	if (OutGenomes[0].GetName() == OutGenomes[1].GetName())
	{
		std::string Name = OutGenomes[1].GetName();
		Name += "2";
		OutGenomes[1].SetName(Name);
	}
	OutFile.Create(OutFileName, OutGenomes);

	/* Perform processing */
	TransAlign();

	for (int i = 0; i != 2; ++i)
		InFiles[i].Close();
	OutFile.Close();

	return 0;
}
