#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "pdf.h"
#include "newton.h"
#include "getopts.h"

double Pct(double, double);
double Normal(double x, double Mean, double StdDev);

// Method for computing erf() found on the web, approximation
// attributed to Nathan Cheung, U.C. Berkeley.
double erf(double x)
	{
	int sign = (x >= 0 ? 1 : -1);
	if (x < 0)
		x = -x;
	const double P = 0.3275911;
	const double a1 = 0.254829592;
	const double a2 = -0.284496736;
	const double a3 = 1.421413741;
	const double a4 = -1.453152027; 
	const double a5 =1.061405429;
	double T = 1.0/(1.0 + P*x);
	double T2 = T*T;
	double T3 = T*T*T;
	double T4 = T*T*T*T;
	double T5 = T*T*T*T*T;
	return sign*(1.0 - (a1*T + a2*T2 + a3*T3 + a4*T4 + a5*T5)*exp(-x*x));
	}

double NormalCDF(double x, double Mean, double StdDev)
	{
	const double Sqrt2 = sqrt(2.0);
	return 0.5*(1.0 + erf((x - Mean)/(Sqrt2*StdDev)));
	}

double GetTruncatedMean(double Mean, double StdDev)
	{
	const double A = 0;
	const double B = 1;
	
	double xA = (A - Mean)/StdDev;
	double xB = (B - Mean)/StdDev;

	double PHI1 = NormalCDF(xA, 0, 1);
	double PHI2 = NormalCDF(xB, 0, 1);
	if (fabs(PHI1 - PHI2) < 1e-6)
		return Mean;

	double phi1 = Normal((A - Mean)/StdDev, 0, 1);
	double phi2 = Normal((B - Mean)/StdDev, 0, 1);

	double T = PHI2 - PHI1;
	double U = phi1 - phi2;
	return Mean + StdDev*U/T;
	}

static struct FO : public FuncObj
	{
	virtual double f(double x)
		{
		return GetTruncatedMean(x, StdDev);
		}

public:
	double StdDev;
	} F;

static double GetTruncNormalMean(double Mu, double Sigma)
	{
	NormalPDF NPDF(Mu, Sigma);

	RandDist RD;
	RD.Init(NPDF, 0, 1);

	return RD.ComputeMean(1000);
	}

static double GetMean(double TruncMean, double StdDev)
	{
	F.StdDev = StdDev;
	return Newton(F, TruncMean, TruncMean, TruncMean+0.1, 1e-6, false);
	}

void GetTruncNormRDNewton(double Mean, double StdDev, RandDist &RD)
	{
// Egregious hack to avoid underflow.
// Arbitrary limits; not analyzed properly.
	bool SmallMean = (Mean >= 0.0 && Mean <= 0.1);
	bool BigMean = (Mean >= 0.9 && Mean <= 1.0);
	bool SmallStdDev = StdDev < 0.1;
	if ((SmallMean || BigMean) && SmallStdDev)
		StdDev = 0.1;

	double Mu = GetMean(Mean, StdDev);
	NormalPDF NPDF(Mu, StdDev);
	RD.Init(NPDF, 0.0, 1.0);
	}

void TruncMeanTable()
	{
// Mu[i], Sigma[i] i = 1 .. 254 are a constant multiple of the mean and
// standard deviation respectively of the untruncated normal distribution
// giving a truncated mean M=i/255
// Mu is multiplied by 256, Sigma is multiplied by 512, so the mean is
// Mu/256 and the standard deviation is Sigma/512.
// Mu is chosen to maximize Sigma.
	vector<int> Mu;
	vector<int> Sigma;

	for (int i = 0; i < 256; ++i)
		{
		Mu.push_back(INT_MIN);
		Sigma.push_back(INT_MIN);
		}

	for (double m = 0.001; m < 1.0; m += 0.001)
		{
		int im = (int) (256*m + 0.5);
		for (double s = 0.001; s < 0.25; s += 0.001)
			{
			int is = (int) (512*s + 0.5);
			double tm = GetTruncatedMean(m, s);
			int itm = (int) (256*tm + 0.5);
			if (itm >= 1 && itm <= 254)
				{
				if (Mu[itm] == INT_MIN || is > Sigma[itm])
					{
					Mu[itm] = im;
					Sigma[itm] = is;
					}
				}
			}
		}

	Log("  n     Mean    Comp.      Err%%       Mu   iSigma    Sigma     zmin     zmax\n");
	Log("===  =======  =======  ========  =======  =======  =======  =======  =======\n");
	for (int i = 1; i <= 254; ++i)
		{
		double m = i/256.;
		double mu = Mu[i]/256.;
		double s = Sigma[i]/512.;
		double c = GetTruncNormalMean(mu, s);
		double e = Pct(c - m, m);
		double zmin = -mu/s;
		double zmax = (1 - mu)/s;
		Log("%3d  %7.4f  %7.4f  %7.4f%%  %7.4f  %7d  %7.4f  %7.4f  %7.4f\n",
		  i, m, c, e, mu, Sigma[i], s, zmin, zmax);
		}

	Log("static MuSigmaValues g_MuSigma[256] =\n");
	Log("\t{\n");
	Log("//\t       Mu   Sigma          i    Mean  imu  isg    Err%%\n");
	Log("//\t   ======  ======        ===  ======  ===  ===  ======\n");
	for (int i = 1; i <= 254; ++i)
		{
		double m = i/256.;
		double mu = Mu[i]/256.;
		double s = Sigma[i]/512.;
		double c = GetTruncNormalMean(mu, s);
		double e = Pct(c - m, m);
		Log("\t{ %6.4f, %6.4f }, //   %3d  %6.4f  %3d  %3d  %5.1f%%\n",
		  mu, s, i, m, Mu[i], Sigma[i], e);
		}
	Log("\t};\n");
	Log("\n");
	}

static void TestTM(double Mean, double StdDev)
	{
	NormalPDF NPDF(Mean, StdDev);

	RandDist RD;
	RD.Init(NPDF, 0, 1);

	double c = RD.ComputeMean(1000);

	double TM = GetTruncatedMean(Mean, StdDev);
	printf("Mean %.3g StdDev %.3g TruncMean %.3g Computed %.3g\n",
	  Mean, StdDev, TM, c);
	}

void TestErf()
	{
	for (double x = -2.0; x < 2.0; x += 0.05)
		{
		double y = erf(x);
		printf("%6.2f  %9.6f  ", x, y);
		int Stars = (int) ((y+1)*32);
		for (int i = 0; i < Stars; ++i)
			printf("*");
		printf("\n");
		}

	printf("\n");
	for (double x = 0.05; x < 2.0; x += 0.05)
		{
		double y = NormalCDF(x, 0.5, 0.1);
		printf("%6.2f  %9.6f  ", x, y);
		int Stars = (int) (y*32);
		for (int i = 0; i < Stars; ++i)
			printf("*");
		printf("\n");
		}
	printf("\n");
	}

void TestTM()
	{
	TestTM(0.1, 0.01);
	TestTM(0.5, 0.3);
	TestTM(0.7, 0.3);
	TestTM(0.8, 0.3);
	TestTM(0.9, 0.3);
	TestTM(0.01, 0.001);
	TestTM(0.00, 0.01);
	}

static void TestTruncNormalRD(double Mean, double StdDev)
	{
	RandDist RD;
	GetTruncNormRDNewton(Mean, StdDev, RD);
	double CalcMean = RD.ComputeMean(100000);
	Log("Mean = %.4f StdDev = %.4f CalcMean = %.4f\n",
	  Mean, StdDev, CalcMean);
	}

void TestX()
	{
	TestTruncNormalRD(0.5, 0.2);
	TestTruncNormalRD(0.8, 0.2);
	TestTruncNormalRD(0.2, 0.2);
	TestTruncNormalRD(0.5, 0.4);
	}

void TruncNorm()
	{
	double Mean = RequiredFloatOpt("truncnorm");
	double StdDev = RequiredFloatOpt("stddev");

	if (Mean < 0.01 || Mean > 0.99)
		Die("Mean must be in range 0.01 to 0.99");

	//TestX();
	//TestTruncNormalRD(Mean, StdDev);

// Egregious hack to avoid underflow.
// Arbitrary values; not analyzed properly.
	bool SmallMean = (Mean >= 0.0 && Mean <= 0.1);
	bool BigMean = (Mean >= 0.9 && Mean <= 1.0);
	bool SmallStdDev = StdDev < 0.1;
	double Value = 0.0;
	if ((SmallMean || BigMean) && SmallStdDev)
		Value = Mean;
	else
		{
		RandDist RD;
		GetTruncNormRDNewton(Mean, StdDev, RD);
		Value = RD.Get();
		}
	printf("%g", Value);
	Log("%g\n", Value);
	}
