#define TRACE	0

#include "evolver.h"
#include "myrand.h"
#include "evomodel.h"

uint8 GetRandProbIndex(double MAC);

EVENT_RESULT Evolver::Try_CreateNontermUTR()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	const unsigned GeneCount = SIZE(Genes);
	vector<gpos_t> UTRLengths;
	for (unsigned i = 0; i < GeneCount; ++i)
		{
		const GeneInfo &Gene = Genes[i];
		const unsigned UTRCount = SIZE(Gene.UTRs);
		for (unsigned j = 0; j < UTRCount; ++j)
			{
			const UTRInfo &UTR = Gene.UTRs[j];
			gpos_t Length = UTR.Length;
			UTRLengths.push_back(Length);
			}
		}

	const unsigned N = SIZE(UTRLengths);
	unsigned r = RandInt(N);
	gpos_t TargetUTRLength = UTRLengths[r];

	GeneInfo Gene;
	gpos_t NewDonorPos = GPOS_T_MAX;
	gpos_t NewAcceptorPos = GPOS_T_MAX;
	gpos_t UTRLength = GPOS_T_MAX;
	bool Found = false;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			break;

		Found = GetCandNontermUTR(Gene, TargetUTRLength, UTRLength,
		  NewDonorPos, NewAcceptorPos);
		if (Found)
			break;
		}
	if (!Found)
		EV_FAILN(CreateNontermUTR, NoCandidates)

	bool Strand = Gene.Plus;
	m_CurrentEventPos = NewDonorPos;

// Fully constrain new acceptor
	BaseInfo &FirstAcceptorBase = GetBase(NewAcceptorPos);
	asserta(FirstAcceptorBase.IsIntron());
	FirstAcceptorBase.SetSplice();

	gpos_t SecondAcceptorPos = GetAdjPos(NewAcceptorPos, Strand);
	BaseInfo &SecondAcceptorBase = GetBase(SecondAcceptorPos);
	asserta(SecondAcceptorBase.IsIntron());
	SecondAcceptorBase.SetSplice();

// First pos in new UTR
	gpos_t Pos = GetAdjPos(SecondAcceptorPos, Strand);
	unsigned BaseCount = 0;
#if	TRACE
	Gene.LogMe();
	Log("NAP %"PF" NDP %"PF"\n", NewAcceptorPos, NewDonorPos);
#endif
	for (;;)
		{
		if (Pos == NewDonorPos)
			break;
		++BaseCount;
		BaseInfo &Base = GetBase(Pos);
#if	TRACE
		Log("Convert base %u before: ", BaseCount);
		Base.LogMe();
#endif
		asserta(Base.IsIntron());
		Base.ProbIx = GetRandProbIndex(Gene.MeanAccProb);
		Base.SetUTR();
#if TRACE
		Log(" .. after ");
		Base.LogMe(true);
#endif
		Pos = GetAdjPos(Pos, Strand);
		}
	asserta(BaseCount > 0);

// Fully constrain new Donor
	BaseInfo &FirstDonorBase = GetBase(NewDonorPos);
	asserta(FirstDonorBase.IsIntron());
	FirstDonorBase.SetSplice();

	gpos_t SecondDonorPos = GetAdjPos(NewDonorPos, Strand);
	BaseInfo &SecondDonorBase = GetBase(SecondDonorPos);
	asserta(SecondDonorBase.IsIntron());
	SecondDonorBase.SetSplice();

	GeneInfo Gene2;
	GetGene(NewAcceptorPos, Gene2);
	unsigned ErrCount = ValidateGene(Gene2);
	if (ErrCount > 0)
		Die("Try_CreateNontermUTR: ValidateGene failed");

	EV_ACCEPTN(CreateCDS, NewAcceptorPos)
	}

bool Evolver::GetCandNontermUTR(const GeneInfo &Gene, gpos_t TargetUTRLength,
  gpos_t &UTRLength, gpos_t &NewDonorPos, gpos_t &NewAcceptorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetCandNontermUTR target len %"PF"\n", TargetUTRLength);
	Gene.LogMe();
	LogGene(Gene);
#endif
	const unsigned IntronCount = SIZE(Gene.Introns);
	const unsigned ExonCount = SIZE(Gene.Exons);
	asserta(ExonCount == IntronCount + 1);

// Try introns in random order, otherwise bias towards
// 5' end of gene.
	vector<unsigned> IntronIndexes;
	for (unsigned i = 0; i < IntronCount; ++i)
		IntronIndexes.push_back(i);
	RandomizeVec(IntronIndexes);

	for (unsigned j = 0; j < IntronCount; ++j)
		if (GetCandNontermUTRIntron(Gene, IntronIndexes[j], TargetUTRLength,
		  UTRLength, NewDonorPos, NewAcceptorPos))
			return true;
	return false;
	}

/***
|<------UTR----->|<-------Intron--------->|<----UTR---->|
                      GT|<--UTR-->|AG

The requirements are:

	New UTR must be target length +/- slop.
	New introns must be >= minimum length.
	Splice sites must match current bases.
***/
bool Evolver::GetCandNontermUTRIntron(const GeneInfo &Gene, unsigned IntronIndex,
  gpos_t TargetUTRLength, gpos_t &UTRLength, gpos_t &NewDonorPos,
  gpos_t &NewAcceptorPos) const
	{
	bool Strand = Gene.Plus;

	asserta(IntronIndex + 1 < SIZE(Gene.Exons));
	asserta(IntronIndex < SIZE(Gene.Introns));
	const IntronInfo &Intron = Gene.Introns[IntronIndex];
	const ExonInfo &Exon1 = Gene.Exons[IntronIndex];
	const ExonInfo &Exon2 = Gene.Exons[IntronIndex+1];
#if	TRACE
	Log("Exon1: ");
	Exon1.LogMe();
	Log(" Exon2: ");
	Exon2.LogMe(true);
#endif
	if (!Exon1.AllUTR || Exon2.AllUTR)
		return false;

	if (Intron.Length < 3*P_MinIntronLength)
		{
#if	TRACE
		Log("FAIL: Short intron\n");
#endif
		return false;
		}

	unsigned MinNewUTRLength = 0;
	if (TargetUTRLength >= P_CreateExonLengthSlop)
		MinNewUTRLength = (unsigned) TargetUTRLength - P_CreateExonLengthSlop;

	unsigned MaxNewUTRLength = (unsigned) TargetUTRLength + P_CreateExonLengthSlop;
	asserta(MaxNewUTRLength >= MinNewUTRLength);

// Avoid zero and very short lengths
	if (MinNewUTRLength < 8)
		MinNewUTRLength = 8;

	string IntronSeq;
	GetSubSeq(Intron.Pos, Intron.Length, IntronSeq, Strand);

	unsigned IntronLength = (unsigned) Intron.Length;

// Donor is first two bases
	string DonorPair;
	DonorPair = IntronSeq.substr(0, 2);

// Acceptor is last two bases
	string AcceptorPair;
	AcceptorPair = IntronSeq.substr(Intron.Length - 2, 2);

/***

...Prev Exon-><------------Existing intron------------><-Next exon...
...Prev Exon-><-Intron(1)->GT<-New UTR->AG<-Intron(2)-><-Next exon...
              <--IntrLen--><><--UTRLen-><><--IntrLen-->
			               |            |
           NewAcceptorOffset            NewDonorOffset
		                   <-----------D-------------->
              <-------------Intron.Length------------->
***/
	unsigned MinAcceptorOffset = P_MinIntronLength;
	unsigned MaxDonorOffset = IntronLength - P_MinIntronLength - 2;
	unsigned D = P_MinIntronLength + MinNewUTRLength + 4;

#if	TRACE
	Log("\n");
	Log("GetCandNontermUTRIntron target length %"PF"\n", TargetUTRLength);
	Log("Min/MaxNewUTRLen %"PF" %"PF"\n", MinNewUTRLength, MaxNewUTRLength);
	Log("Donor %s Acceptor %s\n", DonorPair.c_str(), AcceptorPair.c_str());
	Log("MinAccOff %u, MaxAccOff %u, MaxDonorOff %u, D %u, IntronLength %u\n",
	  MinAcceptorOffset, IntronLength - D, MaxDonorOffset, D, IntronLength);
#endif

	if (IntronLength <= D)
		return false;
	unsigned MaxAcceptorOffset = IntronLength - D;

	vector<unsigned> AcceptorOffsets;
	for (unsigned AcceptorOffset = MinAcceptorOffset;
	  AcceptorOffset <= MaxAcceptorOffset; ++AcceptorOffset)
		if (IntronSeq.substr(AcceptorOffset, 2) == AcceptorPair)
			AcceptorOffsets.push_back(AcceptorOffset);

#if	TRACE
	Log("%u acceptors found\n", SIZE(AcceptorOffsets));
#endif

	RandomizeVec(AcceptorOffsets);
	const unsigned N = SIZE(AcceptorOffsets);
	for (unsigned i = 0; i < N; ++i)
		{
		unsigned AcceptorOffset = AcceptorOffsets[i];
		unsigned ThisMinDonorOffset = AcceptorOffset + MinNewUTRLength + 2;
		unsigned ThisMaxDonorOffset = AcceptorOffset + MaxNewUTRLength + 2;
		if (ThisMaxDonorOffset > MaxDonorOffset)
			ThisMaxDonorOffset = MaxDonorOffset;
#if	TRACE
		Log("Acc off %u acc %u-%u\n",
		  AcceptorOffset, ThisMinDonorOffset, ThisMaxDonorOffset);
#endif
		for (unsigned DonorOffset = ThisMinDonorOffset;
		  DonorOffset <= ThisMaxDonorOffset; ++DonorOffset)
			{
			if (IntronSeq.substr(DonorOffset, 2) == DonorPair)
				{
				UTRLength = DonorOffset - AcceptorOffset - 2;
#if	TRACE
				Log("FOUND: DonorOff %u, AccOff %u, UTRLen %u\n",
				  DonorOffset, AcceptorOffset, (unsigned) UTRLength);
#endif
				asserta(UTRLength + P_CreateExonLengthSlop >= TargetUTRLength);
				asserta(UTRLength <= TargetUTRLength + P_CreateExonLengthSlop);

				gpos_t StartPos =
				  Strand ? Intron.Pos : GetRangeEndPos(Intron.Pos, Intron.Length);
				NewDonorPos = StartPos;
				for (unsigned k = 0; k < DonorOffset; ++k)
					NewDonorPos = GetAdjPos(NewDonorPos, Strand);

				NewAcceptorPos = StartPos;
				for (unsigned k = 0; k < AcceptorOffset; ++k)
					NewAcceptorPos = GetAdjPos(NewAcceptorPos, Strand);

				string NewDonorPair;
				string NewAcceptorPair;
				if (Strand)
					{
					GetSubSeq(NewDonorPos, 2, NewDonorPair, true);
					GetSubSeq(NewAcceptorPos, 2, NewAcceptorPair, true);
					}
				else
					{
					GetSubSeq(GetPrevPos(NewDonorPos), 2, NewDonorPair, false);
					GetSubSeq(GetPrevPos(NewAcceptorPos), 2, NewAcceptorPair, false);
					}
#if TRACE
				Log("DonorOff %u DonorPos %"PF" Donor %s %s\n",
				  DonorOffset, NewDonorPos, DonorPair.c_str(), NewDonorPair.c_str());
				Log("AcceptorOff %u AcceptorPos %"PF" Acceptor %s %s\n",
				  AcceptorOffset, NewAcceptorPos, AcceptorPair.c_str(), NewAcceptorPair.c_str());
#endif
				asserta(NewDonorPair == DonorPair);
				asserta(NewAcceptorPair == AcceptorPair);

				return true;
				}
			}
		}

	return false;
	}
