#define TRACE	0

#include "evolver.h"
#include "ratemgr.h"
#include "myrand.h"
#include "evomodel.h"

bool IsStopCodon(const string &Codon);
int8 IncCodexStrand(int8 Codex, bool Strand);

EVENT_RESULT Evolver::Try_MoveAcceptorIntoUTR()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	GeneInfo Gene;
	gpos_t OldAcceptorPos = GPOS_T_MAX;
	gpos_t NewAcceptorPos = GPOS_T_MAX;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(MoveAcceptorIntoUTR, NoAcceptors)

		bool Found = GetMoveAcceptorIntoUTRCand(Gene, OldAcceptorPos, NewAcceptorPos);
		if (Found)
			break;
		}

	m_CurrentEventPos = OldAcceptorPos;
	bool Strand = Gene.Plus;

	string OldPair;
	string NewPair;

	gpos_t Old2 = GetAdjPos(OldAcceptorPos, Strand);
	gpos_t New2 = GetAdjPos(NewAcceptorPos, Strand);

	BaseInfo &OldBase1 = GetBase(OldAcceptorPos);
	BaseInfo &NewBase1 = GetBase(NewAcceptorPos);

	BaseInfo &OldBase2 = GetBase(Old2);
	BaseInfo &NewBase2 = GetBase(New2);

	asserta(OldBase1.IsIntron());
	asserta(OldBase2.IsIntron());

	asserta(OldBase1.IsSplice());
	asserta(OldBase2.IsSplice());

	asserta(NewBase1.IsUTR());
	asserta(NewBase2.IsUTR());

	OldPair.push_back(OldBase1.GetStrandLetter(Strand));
	NewPair.push_back(NewBase1.GetStrandLetter(Strand));

	OldPair.push_back(OldBase2.GetStrandLetter(Strand));
	NewPair.push_back(NewBase2.GetStrandLetter(Strand));

#if	TRACE
	Log("OAP %"PF" NAP %"PF" Old %s New %s\n",
	  OldAcceptorPos, NewAcceptorPos, OldPair.c_str(), NewPair.c_str());
#endif

	asserta(OldPair == NewPair);

	unsigned BaseCount = 0;

	OldBase1.SetIntron();
	OldBase2.SetIntron();
	OldBase1.Deconstrain();
	OldBase2.Deconstrain();

	gpos_t FirstUTRPos = GetAdjPos(Old2, Strand);
	const BaseInfo &FirstUTRBase = GetBase(FirstUTRPos);
#if	TRACE
	Log("FirstUTRPos %"PF" Base: ", FirstUTRPos);
	FirstUTRBase.LogMe(true);
#endif
	asserta(FirstUTRBase.IsUTR());
	gpos_t Pos = FirstUTRPos;
	for (;;)
		{
		if (Pos == NewAcceptorPos)
			break;
		BaseInfo &Base = GetBase(Pos);
#if	TRACE
		Log("Convert base %u, before: ", BaseCount + 1);
		Base.LogMe();
#endif
		asserta(Base.IsUTR());
		Base.SetIntron();
		Base.Deconstrain();
#if	TRACE
		Log(" .. after: ");
		Base.LogMe(true);
#endif
		++BaseCount;
		Pos = GetAdjPos(Pos, Strand);
		}
	asserta(OldAcceptorPos != NewAcceptorPos);
	//asserta(BaseCount >= 0); // 0 is ok here
	asserta(BaseCount <= P_MaxMoveSpliceBases);

// No mutations allowed in new Acceptor site
	NewBase1.SetSplice();
	NewBase2.SetSplice();

	GeneInfo NewGene;
	GetGene(NewAcceptorPos, NewGene);
	unsigned ErrCount = ValidateGene(NewGene);
	if (ErrCount > 0)
		Die("Try_MoveAcceptorIntoUTR: Validate gene failed");

	EV_ACCEPTN(MoveAcceptorIntoUTR, OldAcceptorPos)
	}

bool Evolver::GetMoveAcceptorIntoUTRCand(const GeneInfo &Gene, gpos_t &OldAcceptorPos,
  gpos_t &NewAcceptorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetMoveAcceptorIntoUTRCand\n");
	Log("Gene: ");
	Gene.LogMe();
#endif
	const unsigned ExonCount = SIZE(Gene.Exons);
	if (ExonCount == 1)
		return false;

	vector<unsigned> ExonIndexes;
	GetRandomizedExonIndexes(Gene, ExonIndexes);

	for (unsigned j = 0; j < ExonCount; ++j)
		{
		unsigned ExonIndex = ExonIndexes[j];
		bool Found =
		  GetMoveAcceptorIntoUTRCandExon(Gene, ExonIndex, OldAcceptorPos, NewAcceptorPos);
		if (Found)
			return true;
		}

	return false;
	}

/***
Acceptor is pair of bases at end of intron, DD=donor,
AA=acceptor:

   <---Exon---> DD Intron AA <---Exon---> DD Intron AA <---Exon--->

Acceptor at beginning of all exons, except first exon in
transcription (X) order.
***/
bool Evolver::GetMoveAcceptorIntoUTRCandExon(const GeneInfo &Gene, unsigned ExonIndex,
  gpos_t &OldAcceptorPos, gpos_t &NewAcceptorPos) const
	{
	bool Strand = Gene.Plus;

	if (Strand && ExonIndex == 0)
		return false;
	if (!Strand && ExonIndex + 1 == SIZE(Gene.Exons))
		return false;

	const ExonInfo &Exon = Gene.Exons[ExonIndex];
#if TRACE
	Log("\n");
	Log("GetMoveAcceptorIntoUTRCandExon ExonIndex %u Count %u\n",
	  ExonIndex, SIZE(Gene.Exons));
	Log("Exon: ");
	Exon.LogMe(true);
#endif
	if (!Exon.AllUTR)
		return false;

	gpos_t UTRBaseCount = GetUTRBaseCount(Gene, ExonIndex);

	gpos_t FirstExonPos =
	  Strand ? Exon.Pos : GetRangeEndPos(Exon.Pos, Exon.Length);

	const BaseInfo &FirstExonBase = GetBase(FirstExonPos);
	asserta(FirstExonBase.IsExon());

	gpos_t SecondOldAcceptorPos = GetAdjPos(FirstExonPos, !Strand);
	OldAcceptorPos = GetAdjPos(SecondOldAcceptorPos, !Strand);

	const BaseInfo &FirstOldAcceptorBase = GetBase(OldAcceptorPos);
	const BaseInfo &SecondOldAcceptorBase = GetBase(SecondOldAcceptorPos);

	asserta(FirstOldAcceptorBase.IsIntron());
	asserta(SecondOldAcceptorBase.IsIntron());

	asserta(FirstOldAcceptorBase.IsSplice());
	asserta(SecondOldAcceptorBase.IsSplice());

	string OldAcceptorPair;
	OldAcceptorPair.push_back(FirstOldAcceptorBase.GetStrandLetter(Strand));
	OldAcceptorPair.push_back(SecondOldAcceptorBase.GetStrandLetter(Strand));

#if TRACE
	{
	Log("\n");
	Log("Exon: ");
	Exon.LogMe(true);
	Log("OldAcceptorPos %"PF"\n", OldAcceptorPos);
	Log("FirstExonBase: ");
	FirstExonBase.LogMe(true);
	Log("FirstOldAcceptorBase: ");
	FirstOldAcceptorBase.LogMe(true);
	Log("SecondOldAcceptorBase: ");
	SecondOldAcceptorBase.LogMe(true);
	Log("OldAcceptorPair %s\n", OldAcceptorPair.c_str());
	}
#endif
	asserta(FirstOldAcceptorBase.IsIntron());
	asserta(SecondOldAcceptorBase.IsIntron());

	gpos_t MaxDist = P_MaxMoveSpliceBases;
	if (MaxDist + 8 >= Exon.Length)
		{
		if (Exon.Length <= 16)
			return false;
		MaxDist = Exon.Length - 8;
		}

	gpos_t Pos1 = GetAdjPos(SecondOldAcceptorPos, Strand);
	for (gpos_t Dist = 0; Dist < MaxDist; ++Dist)
		{
		const BaseInfo &Base1 = GetBase(Pos1);
		if (!Base1.IsUTR())
			return false;
		gpos_t Pos2 = GetAdjPos(Pos1, Strand);
		if (Pos2 == GPOS_T_MAX)
			return false;
		const BaseInfo &Base2 = GetBase(Pos2);
		if (!Base2.IsUTR())
			return false;
#if	TRACE
		Log("Pos1 %"PF"\n", Pos1);
		Log("Base1: ");
		Base1.LogMe(true);
		Log("Base2: ");
		Base2.LogMe(true);
#endif

		string Pair;
		Pair.push_back(Base1.GetStrandLetter(Strand));
		Pair.push_back(Base2.GetStrandLetter(Strand));
#if	TRACE
		Log("Pair %s Old %s\n", Pair.c_str(), OldAcceptorPair.c_str());
#endif
		if (Pair == OldAcceptorPair)
			{
			if (UTRBaseCount < P_MinUTRBases + Dist)
				return false;

			NewAcceptorPos = Pos1;
#if	TRACE
			Log("FOUND NDP %"PF"\n", NewAcceptorPos);
#endif
			return true;
			}

		Pos1 = Pos2;
		}
	return false;
	}
