#define TRACE	0

#include "evolver.h"
#include "ratemgr.h"
#include "myrand.h"
#include "evomodel.h"

bool IsStopCodon(const string &Codon);
int8 IncCodexStrand(int8 Codex, bool Strand);
int8 DecCodexStrand(int8 Codex, bool Strand);
uint8 GetRandProbIndex(double MAC);

EVENT_RESULT Evolver::Try_MoveCDSAcceptorIntoIntron()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	GeneInfo Gene;
	gpos_t OldAcceptorPos = GPOS_T_MAX;
	gpos_t NewAcceptorPos = GPOS_T_MAX;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(MoveCDSAcceptorIntoIntron, NoAcceptors)

		bool Found = GetMoveCDSAcceptorIntronCand(Gene, OldAcceptorPos, NewAcceptorPos);
		if (Found)
			break;
		}

	m_CurrentEventPos = OldAcceptorPos;
	bool Strand = Gene.Plus;
#if	TRACE
	Log("Select gene %u old pos %"PF" new pos %"PF"\n",
	  Gene.Index, OldAcceptorPos, NewAcceptorPos);
	Gene.LogMe();
#endif

	string OldPair;
	string NewPair;

	GetNextPair(OldAcceptorPos, OldPair, Strand);
	GetNextPair(NewAcceptorPos, NewPair, Strand);
	asserta(OldPair == NewPair);

	gpos_t OldSecondAcceptorPos = GetAdjPos(OldAcceptorPos, Strand);
	gpos_t FirstCDSPos = GetAdjPos(OldSecondAcceptorPos, Strand);
	const BaseInfo &FirstCDSBase = GetBase(FirstCDSPos);
#if	TRACE
	Log("\n");
	Log("OldSecondAcceptorPos=%"PF"\n", OldSecondAcceptorPos);
	Log("FirstCDSPos=%"PF"\n", FirstCDSPos);
	Log("FirstCDSBase=");
	FirstCDSBase.LogMe(true);
#endif
	asserta(FirstCDSBase.IsCDS());
	int8 Codex = FirstCDSBase.Codex;
	gpos_t Pos = NewAcceptorPos;

// No mutations allowed in new Acceptor site
	BaseInfo &Base1 = GetBase(Pos);
#if	TRACE
	Log("Pos=NewAcceptorPos=%"PF"\n", Pos);
	Log("Base1=");
	Base1.LogMe(true);
#endif
	asserta(Base1.IsIntron());
	Base1.SetSplice();

	Pos = GetAdjPos(Pos, Strand);
	BaseInfo &Base2 = GetBase(Pos);
#if	TRACE
	Log("Pos=GetAdjPos(%c)=%"PF"\n", pom(Strand), Pos);
	Log("Base2=");
	Base2.LogMe(true);
#endif
	asserta(Base2.IsIntron());
	Base2.SetSplice();

	unsigned BaseCount = 0;
	for (;;)
		{
		Pos = GetAdjPos(Pos, Strand);
		BaseInfo &Base = GetBase(Pos);
#if	TRACE
		Log("Convert base %u, pos %"PF" before: ", BaseCount + 1, Pos);
		Base.LogMe();
#endif
		asserta(Base.Type == BT_Intron || Base.Type == BT_Splice);
		Base.Type = BT_CDS;
		Base.Codex = Codex;
		Base.ProbIx = GetRandProbIndex(Gene.MeanAccProb);
		Codex = IncCodexStrand(Codex, Strand);
#if	TRACE
		Log(" .. after: ");
		Base.LogMe(true);
#endif
		++BaseCount;
		if (Pos == OldSecondAcceptorPos)
			break;
		}
	asserta(BaseCount > 0);
	asserta(BaseCount%3 == 0);
	asserta(BaseCount <= P_MaxMoveSpliceBases);

	GeneInfo NewGene;
	GetGene(NewAcceptorPos, NewGene);
	unsigned ErrCount = ValidateGene(NewGene);
	if (ErrCount > 0)
		Die("Try_MoveCDSAcceptorIntoIntron: Validate gene failed");

	EV_ACCEPTN(MoveCDSAcceptorIntoIntron, NewAcceptorPos)
	}

bool Evolver::GetMoveCDSAcceptorIntronCand(const GeneInfo &Gene,
  gpos_t &OldAcceptorPos, gpos_t &NewAcceptorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetMoveAcceptorIntronCand\n");
	Log("Gene: ");
	Gene.LogMe();
#endif

	vector<unsigned> ExonIndexes;
	GetRandomizedExonIndexes(Gene, ExonIndexes);

	const unsigned ExonCount = SIZE(Gene.Exons);
	asserta(SIZE(ExonIndexes) == ExonCount);
	for (unsigned i = 0; i < ExonCount; ++i)
		if (GetMoveCDSAcceptorIntronCandExon(Gene, ExonIndexes[i], OldAcceptorPos,
		  NewAcceptorPos))
			return true;

	return false;
	}

// Acceptor is sequence at end of intron.
// There is an acceptor immediately preceding all exons except
// the first in transcription (X) order.
// Find sequence: XX<3n bases with no in-frame STOP codon><exon start>
// where XX is the last two bases in the current intron.
bool Evolver::GetMoveCDSAcceptorIntronCandExon(const GeneInfo &Gene,
  unsigned ExonIndex, gpos_t &OldAcceptorPos, gpos_t &NewAcceptorPos) const
	{
	bool Strand = Gene.Plus;
	const unsigned CDSCount = SIZE(Gene.CDSs);

	const ExonInfo &Exon = Gene.Exons[ExonIndex];
	if (!Exon.AllCoding)
		return false;
	
	unsigned CDSIndex = Exon.CDSIndex;
	asserta(CDSIndex != UINT_MAX);

// Never do the first or last CDS (because adjacent to START / STOP).
	if (CDSIndex == 0 || CDSIndex + 1 == CDSCount)
		return false;

	unsigned PrevCDSIndex = (Strand ? CDSIndex - 1 : CDSIndex + 1);
	const CDSInfo &PrevCDS = Gene.CDSs[PrevCDSIndex];
#if TRACE
	Log("PrevCDS[%u]:\n", PrevCDSIndex);
	PrevCDS.LogMe(true);
#endif
	if (PrevCDS.Length < 2)
		return false;

	unsigned PrevExonIndex = (Strand ? ExonIndex - 1 : ExonIndex + 1);
	const ExonInfo &PrevExon = Gene.Exons[PrevExonIndex];

	if (!PrevExon.AllCoding)
		return false;

	if (PrevExonIndex >= SIZE(Gene.Introns))
		return false;

	const IntronInfo &Intron = Gene.Introns[PrevExonIndex];

	gpos_t LastIntronPos =
		Strand ? GetRangeEndPos(Intron.Pos, Intron.Length) : Intron.Pos;
	OldAcceptorPos = GetAdjPos(LastIntronPos, !Strand);

	const BaseInfo &OldFirstAcceptorBase = GetBase(OldAcceptorPos);
	asserta(OldFirstAcceptorBase.IsIntron());

	const BaseInfo &OldSecondAcceptorBase = GetBase(LastIntronPos);
	if (!OldSecondAcceptorBase.IsIntron())
		return false;

	asserta(OldFirstAcceptorBase.IsSplice());
	asserta(OldSecondAcceptorBase.IsSplice());

	string OldAcceptorPair;
	OldAcceptorPair.push_back(OldFirstAcceptorBase.GetStrandLetter(Strand));
	OldAcceptorPair.push_back(OldSecondAcceptorBase.GetStrandLetter(Strand));

	string FractCodonPrefix;
	string FractCodonSuffix;

	bool StartStop;
	GetFractCodonAfterIntron(OldAcceptorPos, FractCodonSuffix, Strand, StartStop);
	if (StartStop)
		return false;
	unsigned FractCodonSuffixLength = SIZE(FractCodonSuffix);
#if TRACE
	Log("Fract codon suffix='%s'\n", FractCodonSuffix.c_str());
#endif
	if (FractCodonSuffixLength > 0)
		{
		string Codon;
		if (FractCodonSuffixLength == 1)
			Codon = OldAcceptorPair + FractCodonSuffix;
		else if (FractCodonSuffixLength == 2)
			Codon = OldAcceptorPair[1] + FractCodonSuffix;
		else
			asserta(false);
		asserta(SIZE(Codon) == 3);
#if	TRACE
		Log("Codon(1)=%s\n", Codon.c_str());
#endif
		if (IsStopCodon(Codon))
			return false;;
		}

	GetFractCodonBeforeIntron(OldAcceptorPos, FractCodonPrefix, Strand, StartStop);
	if (StartStop)
		return false;
#if	TRACE
	Log("\n");
	Log("Fract codon prefix='%s'\n", FractCodonPrefix.c_str());
	Log("Intron: ");
	Intron.LogMe(true);
	Log("OldFirstAcceptorBase: ");
	OldFirstAcceptorBase.LogMe(true);
	Log("OldSecondAcceptorBase: ");
	OldSecondAcceptorBase.LogMe(true);
	Log("OldPair: %s\n", OldAcceptorPair.c_str());
#endif
	unsigned SumLengths = SIZE(FractCodonPrefix) + SIZE(FractCodonSuffix);
	asserta(SumLengths == 0 || SumLengths == 3);

	NewAcceptorPos = OldAcceptorPos;
	gpos_t Pos3 = GPOS_T_MAX;
	unsigned CodonsMovedCount = 0;
	for ( ; CodonsMovedCount < P_MaxMoveSpliceBases/3; ++CodonsMovedCount)
		{
		Pos3 = GetAdjPos(NewAcceptorPos, !Strand);
		const BaseInfo &Base3 = GetBase(Pos3);
		if (!Base3.IsIntron() || Base3.IsStartStop())
			return false;;

		gpos_t Pos2 = GetAdjPos(Pos3, !Strand);
		const BaseInfo &Base2 = GetBase(Pos2);
		if (!Base2.IsIntron() || Base2.IsStartStop())
			return false;;

		NewAcceptorPos = GetAdjPos(Pos2, !Strand);
		const BaseInfo &Base1 = GetBase(NewAcceptorPos);
		if (!Base1.IsIntron() || Base1.IsStartStop())
			return false;;

		string Pair;
		Pair.push_back(Base1.GetStrandLetter(Strand));
		Pair.push_back(Base2.GetStrandLetter(Strand));
#if	TRACE
		Log("\n");
		Log("NewAcceptorPos %"PF"\n", NewAcceptorPos);
		Log("Base1 Pos=%"PF": ", NewAcceptorPos);
		Base1.LogMe(true);
		Log("Base2 Pos=%"PF": ", Pos2);
		Base2.LogMe(true);
		Log("Base3 Pos=%"PF": ", Pos3);
		Base3.LogMe(true);
		Log("Pair %s OldAcceptor %s\n", Pair.c_str(), OldAcceptorPair.c_str());
#endif
		if (Pair == OldAcceptorPair)
			break;
		}
	if (CodonsMovedCount >= P_MaxMoveSpliceBases/3)
		return false;

	gpos_t BaseCount = CodonsMovedCount*3;
	asserta(BaseCount < Intron.Length);
	if (Intron.Length - BaseCount < P_MinIntronLength)
		return false;

	asserta(Pos3 != GPOS_T_MAX);
	gpos_t Pos = Pos3;
	unsigned FractCodonPrefixLength = SIZE(FractCodonPrefix);
	if (FractCodonPrefixLength > 0)
		{
		string Codon(FractCodonPrefix);
		for (unsigned i = FractCodonPrefixLength; i < 3; ++i)
			{
			asserta(Pos != GPOS_T_MAX);

			const BaseInfo &Base = GetBase(Pos);
			if (Base.IsStartStop())
				return false;
			Codon.push_back(Base.GetStrandLetter(Strand));
			Pos = GetAdjPos(Pos, Strand);
			}
		asserta(Codon.size() == 3);
#if	TRACE
		Log("Codon(2)=%s\n", Codon.c_str());
#endif
		if (IsStopCodon(Codon))
			return false;
		}

	for (;;)
		{
		asserta(Pos != GPOS_T_MAX);

#if	TRACE
		Log("Pos=%u ", (unsigned) Pos);
#endif
		string Codon;
		Pos = GetNextTriplet(Pos, Codon, Strand);
#if	TRACE
		Log("Codon(3)=%s\n", Codon.c_str());
#endif
		if (IsStopCodon(Codon))
			return false;;
		if (Pos == OldAcceptorPos)
			break;
		}

#if	TRACE
	Log("FOUND Exon %u CDS %u Old %"PF" New %"PF"\n",
	  ExonIndex, CDSIndex, OldAcceptorPos, NewAcceptorPos);
#endif
	return true;
	}

void Evolver::GetRandomizedExonIndexes(const GeneInfo &Gene,
  vector<unsigned> &ExonIndexes) const
	{
	const unsigned ExonCount = SIZE(Gene.Exons);

	ExonIndexes.clear();
	vector<double> MACs;
	for (unsigned i = 0; i < ExonCount; ++i)
		{
		const ExonInfo &Exon = Gene.Exons[i];
		MACs.push_back(Exon.MeanAccProb);
		ExonIndexes.push_back(i);
		}
	RandomizeVecWeighted(ExonIndexes, MACs);
	}
