#define TRACE	0

#include "evolver.h"
#include "ratemgr.h"
#include "myrand.h"
#include "evomodel.h"

bool IsStopCodon(const string &Codon);
int8 IncCodexStrand(int8 Codex, bool Strand);

EVENT_RESULT Evolver::Try_MoveDonorIntoUTR()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	GeneInfo Gene;
	gpos_t OldDonorPos = GPOS_T_MAX;
	gpos_t NewDonorPos = GPOS_T_MAX;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(MoveDonorIntoUTR, NoDonors)

		bool Found = GetMoveDonorIntoUTRCand(Gene, OldDonorPos, NewDonorPos);
		if (Found)
			break;
		}

	m_CurrentEventPos = OldDonorPos;
	bool Strand = Gene.Plus;

	string OldPair;
	string NewPair;

	gpos_t Old2 = GetAdjPos(OldDonorPos, Strand);
	gpos_t New2 = GetAdjPos(NewDonorPos, Strand);

	BaseInfo &OldBase1 = GetBase(OldDonorPos);
	BaseInfo &NewBase1 = GetBase(NewDonorPos);

	BaseInfo &OldBase2 = GetBase(Old2);
	BaseInfo &NewBase2 = GetBase(New2);

	asserta(OldBase1.IsIntron());
	asserta(OldBase2.IsIntron());

	asserta(NewBase1.IsUTR());
	asserta(NewBase2.IsUTR());

	OldPair.push_back(OldBase1.GetStrandLetter(Strand));
	NewPair.push_back(NewBase1.GetStrandLetter(Strand));

	OldPair.push_back(OldBase2.GetStrandLetter(Strand));
	NewPair.push_back(NewBase2.GetStrandLetter(Strand));

#if	TRACE
	Log("ODP %"PF" NDP %"PF" Old %s New %s\n",
	  OldDonorPos, NewDonorPos, OldPair.c_str(), NewPair.c_str());
#endif

	asserta(OldPair == NewPair);

	OldBase1.SetIntron();
	OldBase2.SetIntron();
	OldBase1.Deconstrain();
	OldBase2.Deconstrain();

	unsigned BaseCount = 0;
	gpos_t LastUTRPos = GetAdjPos(OldDonorPos, !Strand);
	const BaseInfo &LastUTRBase = GetBase(LastUTRPos);
	asserta(LastUTRBase.IsUTR());
	gpos_t Pos = NewDonorPos;
	for (;;)
		{
		BaseInfo &Base = GetBase(Pos);
#if	TRACE
		Log("Convert base %u, before: ", BaseCount + 1);
		Base.LogMe();
#endif
		asserta(Base.IsUTR());
		Base.SetIntron();
		Base.Deconstrain();
#if	TRACE
		Log(" .. after: ");
		Base.LogMe(true);
#endif
		++BaseCount;
		Pos = GetAdjPos(Pos, Strand);
		if (Pos == OldDonorPos)
			break;
		}
	asserta(BaseCount > 0);
	asserta(BaseCount <= P_MaxMoveSpliceBases);

	NewBase1.SetSplice();
	NewBase2.SetSplice();

	GeneInfo NewGene;
	GetGene(NewDonorPos, NewGene);
	unsigned ErrCount = ValidateGene(NewGene);
	if (ErrCount > 0)
		Die("Try_MoveDonorIntoUTR: Validate gene failed");

	EV_ACCEPTN(MoveDonorIntoUTR, OldDonorPos)
	}

bool Evolver::GetMoveDonorIntoUTRCand(const GeneInfo &Gene, gpos_t &OldDonorPos,
  gpos_t &NewDonorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetMoveDonorIntoUTRCand\n");
	Log("Gene: ");
	Gene.LogMe();
#endif

	vector<unsigned> ExonIndexes;
	GetRandomizedExonIndexes(Gene, ExonIndexes);

	for (unsigned j = 0; j < SIZE(ExonIndexes); ++j)
		if (GetMoveDonorIntoUTRCandExon(Gene, ExonIndexes[j],
			  OldDonorPos, NewDonorPos))
			return true;

	return false;
	}

/***
Donor is pair of bases at start of intron, DD=donor,
AA=acceptor:

   <---Exon---> DD Intron AA <---Exon---> DD Intron AA <---Exon--->
***/
bool Evolver::GetMoveDonorIntoUTRCandExon(const GeneInfo &Gene, unsigned ExonIndex,
  gpos_t &OldDonorPos, gpos_t &NewDonorPos) const
	{
	const ExonInfo &Exon = Gene.Exons[ExonIndex];
#if TRACE
	Log("\n");
	Log("GetMoveDonorIntoUTRCandExon\n");
	Log("Exon: ");
	Exon.LogMe(true);
#endif
	if (!Exon.AllUTR)
		return false;

//There is a donor at end of all exons, except last exon in
//transcription (X) order.
	bool Strand = Gene.Plus;
	const unsigned ExonCount = SIZE(Gene.Exons);
	if (Strand && ExonIndex + 1 == ExonCount)
		return false;
	if (!Strand && ExonIndex == 0)
		return false;

	gpos_t LastExonPos =
	  Strand ? GetRangeEndPos(Exon.Pos, Exon.Length) : Exon.Pos;

	gpos_t UTRBaseCount = GetUTRBaseCount(Gene, ExonIndex);

	const BaseInfo &LastExonBase = GetBase(LastExonPos);
	asserta(LastExonBase.IsExon());

	OldDonorPos = GetAdjPos(LastExonPos, Strand);
	gpos_t SecondOldDonorPos = GetAdjPos(OldDonorPos, Strand);

	const BaseInfo &FirstOldDonorBase = GetBase(OldDonorPos);
	const BaseInfo &SecondOldDonorBase = GetBase(SecondOldDonorPos);

	asserta(FirstOldDonorBase.IsIntron());
	asserta(SecondOldDonorBase.IsIntron());

	asserta(FirstOldDonorBase.IsSplice());
	asserta(SecondOldDonorBase.IsSplice());

	string OldDonorPair;
	OldDonorPair.push_back(FirstOldDonorBase.GetStrandLetter(Strand));
	OldDonorPair.push_back(SecondOldDonorBase.GetStrandLetter(Strand));

#if TRACE
	{
	Log("\n");
	Log("Exon: ");
	Exon.LogMe(true);
	Log("OldDonorPos %"PF"\n", OldDonorPos);
	Log("LastExonBase: ");
	LastExonBase.LogMe(true);
	Log("FirstOldDonorBase: ");
	FirstOldDonorBase.LogMe(true);
	Log("SecondOldDonorBase: ");
	SecondOldDonorBase.LogMe(true);
	Log("OldDonorPair %s\n", OldDonorPair.c_str());
	}
#endif
	asserta(FirstOldDonorBase.IsIntron());
	asserta(SecondOldDonorBase.IsIntron());

	asserta(FirstOldDonorBase.IsSplice());
	asserta(SecondOldDonorBase.IsSplice());

	gpos_t MaxDist = P_MaxMoveSpliceBases;
	if (MaxDist + 8 >= Exon.Length)
		{
		if (Exon.Length < 16)
			return false;
		MaxDist = Exon.Length - 8;
		}
	
	gpos_t Pos2 = GetAdjPos(OldDonorPos, !Strand);
	for (gpos_t Dist = 1; Dist < MaxDist; ++Dist)
		{
		const BaseInfo &Base2 = GetBase(Pos2);
		if (!Base2.IsUTR())
			return false;
		gpos_t Pos1 = GetAdjPos(Pos2, !Strand);
		if (Pos1 == GPOS_T_MAX)
			return false;
		const BaseInfo &Base1 = GetBase(Pos1);
		if (!Base1.IsUTR())
			return false;
#if	TRACE
		Log("Pos1 %"PF" NewDonorPos %"PF"\n", Pos1, NewDonorPos);
		Log("Base1: ");
		Base1.LogMe(true);
		Log("Base2: ");
		Base2.LogMe(true);
#endif

		string Pair;
		Pair.push_back(Base1.GetStrandLetter(Strand));
		Pair.push_back(Base2.GetStrandLetter(Strand));
#if	TRACE
		Log("Pair %s Old %s\n", Pair.c_str(), OldDonorPair.c_str());
#endif
		if (Pair == OldDonorPair)
			{
			if (UTRBaseCount < P_MinUTRBases + Dist)
				return false;

			NewDonorPos = Pos1;
#if	TRACE
			Log("FOUND NDP %"PF"\n", NewDonorPos);
#endif
			return true;
			}

		Pos2 = Pos1;
		}
	return false;
	}
