#define TRACE	0

#include "evolver.h"
#include "ratemgr.h"
#include "myrand.h"
#include "evomodel.h"

bool IsStopCodon(const string &Codon);
int8 IncCodexStrand(int8 Codex, bool Strand);
uint8 GetRandProbIndex(double MAC);

EVENT_RESULT Evolver::Try_MoveUTRAcceptorIntoIntron()
	{
	vector<GeneInfo> Genes;
	vector<double> MACs;
	InitTryGenes(Genes, MACs);

	GeneInfo Gene;
	gpos_t OldAcceptorPos = GPOS_T_MAX;
	gpos_t NewAcceptorPos = GPOS_T_MAX;
	for (;;)
		{
		bool Any = NextTryGene(Genes, MACs, Gene);
		if (!Any)
			EV_FAILN(MoveUTRAcceptorIntoIntron, NoAcceptors)

		bool Found = GetMoveUTRAcceptorIntoIntronCand(Gene, OldAcceptorPos,
		  NewAcceptorPos);
		if (Found)
			break;
		}

	m_CurrentEventPos = OldAcceptorPos;
	bool Strand = Gene.Plus;

	string OldPair;
	string NewPair;

	gpos_t Old2 = GetAdjPos(OldAcceptorPos, Strand);
	gpos_t New2 = GetAdjPos(NewAcceptorPos, Strand);

	BaseInfo &OldBase1 = GetBase(OldAcceptorPos);
	BaseInfo &NewBase1 = GetBase(NewAcceptorPos);

	BaseInfo &OldBase2 = GetBase(Old2);
	BaseInfo &NewBase2 = GetBase(New2);

	asserta(OldBase1.IsIntron());
	asserta(OldBase2.IsIntron());

	asserta(NewBase1.IsIntron());
	asserta(NewBase2.IsIntron());

	OldPair.push_back(OldBase1.GetStrandLetter(Strand));
	NewPair.push_back(NewBase1.GetStrandLetter(Strand));

	OldPair.push_back(OldBase2.GetStrandLetter(Strand));
	NewPair.push_back(NewBase2.GetStrandLetter(Strand));

#if	TRACE
	Log("ODP %"PF" NDP %"PF" Old %s New %s\n",
	  OldAcceptorPos, NewAcceptorPos, OldPair.c_str(), NewPair.c_str());
#endif

	asserta(OldPair == NewPair);

	unsigned BaseCount = 0;
	gpos_t Pos = New2;
	for (;;)
		{
		Pos = GetAdjPos(Pos, Strand);
		BaseInfo &Base = GetBase(Pos);
		++BaseCount;
#if	TRACE
		Log("Convert base %u, before: ", BaseCount);
		Base.LogMe();
#endif
		asserta(Base.IsIntron());
		Base.SetUTR();
		Base.ProbIx = GetRandProbIndex(Gene.MeanAccProb);
#if	TRACE
		Log(" .. after: ");
		Base.LogMe(true);
#endif
		if (Pos == Old2)
			break;
		}
	asserta(BaseCount > 0);
	asserta(BaseCount <= P_MaxMoveSpliceBases);

	NewBase1.SetSplice();
	NewBase2.SetSplice();

	GeneInfo NewGene;
	GetGene(NewAcceptorPos, NewGene);
	unsigned ErrCount = ValidateGene(NewGene);
	if (ErrCount > 0)
		Die("Try_MoveUTRAcceptorIntoIntron: Validate gene failed");

	EV_ACCEPTN(MoveUTRAcceptorIntoIntron, OldAcceptorPos)
	}

bool Evolver::GetMoveUTRAcceptorIntoIntronCand(const GeneInfo &Gene,
  gpos_t &OldAcceptorPos, gpos_t &NewAcceptorPos) const
	{
#if	TRACE
	Log("\n");
	Log("GetMoveUTRAcceptorIntoIntronCand\n");
	Log("Gene: ");
	Gene.LogMe();
#endif

	vector<unsigned> ExonIndexes;
	GetRandomizedExonIndexes(Gene, ExonIndexes);
	for (unsigned j = 0; j < SIZE(ExonIndexes); ++j)
		if (GetMoveUTRAcceptorIntoIntronCandExon(Gene, ExonIndexes[j],
		   OldAcceptorPos, NewAcceptorPos))
			return true;

	return false;
	}

/***
Acceptor is pair of bases at end of intron, DD=Acceptor, AA=acceptor:

   <---Exon---> DD Intron AA <---Exon---> DD Intron AA <---Exon--->
***/
bool Evolver::GetMoveUTRAcceptorIntoIntronCandExon(const GeneInfo &Gene,
  unsigned ExonIndex, gpos_t &OldAcceptorPos, gpos_t &NewAcceptorPos) const
	{
	const unsigned ExonCount = SIZE(Gene.Exons);
	bool Strand = Gene.Plus;

//There is a Acceptor preceding start of all exons, except first exon in
//transcription (X) order.
	if (Strand && ExonIndex == 0)
		return false;
	if (!Strand && ExonIndex + 1 == ExonCount)
		return false;

	const ExonInfo &Exon = Gene.Exons[ExonIndex];
	unsigned IntronIndex = (Strand ? ExonIndex  - 1: ExonIndex);
	asserta(IntronIndex < SIZE(Gene.Introns));
	const IntronInfo &Intron = Gene.Introns[IntronIndex];
	if (Intron.Length <= P_MinIntronLength)
		return false;

#if TRACE
	Log("\n");
	Log("GetMoveUTRAcceptorIntoIntronCandExon\n");
	Log("Exon: ");
	Exon.LogMe(true);
	Log("Intron: ");
	Intron.LogMe(true);
#endif

	if (!Exon.AllUTR)
		return false;

	gpos_t FirstExonPos =
	  Strand ? Exon.Pos : GetRangeEndPos(Exon.Pos, Exon.Length);

	const BaseInfo &FirstExonBase = GetBase(FirstExonPos);
	asserta(FirstExonBase.IsExon());

	gpos_t SecondOldAcceptorPos = GetAdjPos(FirstExonPos, !Strand);
	OldAcceptorPos = GetAdjPos(SecondOldAcceptorPos, !Strand);

	const BaseInfo &FirstOldAcceptorBase = GetBase(OldAcceptorPos);
	const BaseInfo &SecondOldAcceptorBase = GetBase(SecondOldAcceptorPos);

	asserta(FirstOldAcceptorBase.IsIntron());
	asserta(SecondOldAcceptorBase.IsIntron());

	asserta(FirstOldAcceptorBase.IsSplice());
	asserta(SecondOldAcceptorBase.IsSplice());

	string OldAcceptorPair;
	OldAcceptorPair.push_back(FirstOldAcceptorBase.GetStrandLetter(Strand));
	OldAcceptorPair.push_back(SecondOldAcceptorBase.GetStrandLetter(Strand));

#if TRACE
	{
	Log("\n");
	Log("OldAcceptorPos %"PF"\n", OldAcceptorPos);
	Log("FirstExonBase: ");
	FirstExonBase.LogMe(true);
	Log("FirstOldAcceptorBase: ");
	FirstOldAcceptorBase.LogMe(true);
	Log("SecondOldAcceptorBase: ");
	SecondOldAcceptorBase.LogMe(true);
	Log("OldAcceptorPair %s\n", OldAcceptorPair.c_str());
	}
#endif
	asserta(FirstOldAcceptorBase.IsIntron());
	asserta(SecondOldAcceptorBase.IsIntron());

	gpos_t MaxDist = P_MaxMoveSpliceBases;
	if (Intron.Length < P_MinIntronLength + MaxDist)
		MaxDist = Intron.Length - P_MinIntronLength;
#if	TRACE
	Log("Intron length %"PF" MinLen %"PF" MaxDist %"PF"\n",
	  Intron.Length, P_MinIntronLength, MaxDist);
#endif
	gpos_t Pos2 = GetAdjPos(OldAcceptorPos, !Strand);
	for (gpos_t Dist = 1; Dist < MaxDist; ++Dist)
		{
		const BaseInfo &Base2 = GetBase(Pos2);
		gpos_t Pos1 = GetAdjPos(Pos2, !Strand);
		if (Pos1 == GPOS_T_MAX)
			return false;
		const BaseInfo &Base1 = GetBase(Pos1);
#if	TRACE
		Log("Pos1 %"PF" NewAcceptorPos %"PF"\n", Pos1, NewAcceptorPos);
		Log("Base1: ");
		Base1.LogMe(true);
		Log("Base2: ");
		Base2.LogMe(true);
#endif
		asserta(Base1.IsIntron());
		asserta(Base2.IsIntron());

		string Pair;
		Pair.push_back(Base1.GetStrandLetter(Strand));
		Pair.push_back(Base2.GetStrandLetter(Strand));
#if	TRACE
		Log("Pair %s Old %s\n", Pair.c_str(), OldAcceptorPair.c_str());
#endif
		if (Pair == OldAcceptorPair)
			{
			NewAcceptorPos = Pos1;
#if	TRACE
			Log("FOUND NDP %"PF" Dist %"PF"\n", NewAcceptorPos, Dist);
#endif
			return true;
			}

		Pos2 = Pos1;
		}
	return false;
	}
