#include "evolver.h"
#include "myrand.h"
#include "evomodel.h"

/***
Substitution works as follows. 

The model defines the following 6 rates:

	A->C, A->G, A->T, C->A, C->G, C->T

Complementary substitutions are assumed to be equal, so e.g.:

	rate(T->G) = rate(A->T)

We cannot efficiently "choose a random A", but we can "choose
a random base". 

The algorithm for Substitute is as follows.

	1. Choose a random base, call its letter X.
	2. If P=XFailProb > 0, fail the event with probability P.
	3. Substitute X->Y, where P(Y|X) = rate(X->Y) / sum_i rate(X->i).

One of g_CGFailProb or g_ATFailProb is > 0. The FailProb value is
chosen so that the ratio sum_i rate(A->i) to sum_i rate(C->i)
is correct.

See nucmx.cpp for more.
***/

extern double g_SubstProbs[4][4];
extern double g_CGFailProb;
extern double g_ATFailProb;

unsigned g_LetterIndex[256];
static bool InitTables()
	{
	for (unsigned i = 0; i < 256; ++i)
		g_LetterIndex[i] = UINT_MAX;
	g_LetterIndex[(unsigned char) 'A'] = 0;
	g_LetterIndex[(unsigned char) 'C'] = 1;
	g_LetterIndex[(unsigned char) 'G'] = 2;
	g_LetterIndex[(unsigned char) 'T'] = 3;
	g_LetterIndex[(unsigned char) 'a'] = 0;
	g_LetterIndex[(unsigned char) 'c'] = 1;
	g_LetterIndex[(unsigned char) 'g'] = 2;
	g_LetterIndex[(unsigned char) 't'] = 3;
	return true;
	}
static bool TablesInitialized = InitTables();

static char LetterToChar[4] = { 'A', 'C', 'G', 'T' };

static char GetSubstLetter(char OldLetter)
	{
	unsigned OldLetterIndex = g_LetterIndex[(unsigned char) OldLetter];
	if (OldLetterIndex > 3)
		Die("Invalid letter 0x%02x", OldLetter);
	const double *Probs = g_SubstProbs[OldLetterIndex];
	if (Probs[OldLetterIndex] != 0)
		Die("Invalid subst probs, %c -> %c is not zero",
		  OldLetter, OldLetter);
	unsigned NewLetterIndex = RandWeightDist(Probs, 4);
	asserta(NewLetterIndex < 4);
	return LetterToChar[NewLetterIndex];
	}

EVENT_RESULT Evolver::Try_Substitute()
	{
	gpos_t Pos = GetRandomPos();
	if (Pos == GPOS_T_MAX)
		EV_FAILN(Substitute, GetRandomPos)

	m_CurrentEventPos = Pos;
	BaseInfo &Base = GetBase(Pos);

	char OldLetter = Base.GetLetter();
	char NewLetter = GetSubstLetter(OldLetter);
	asserta(NewLetter != OldLetter);

	if (g_CGFailProb > 0 && (OldLetter == 'C' || OldLetter == 'G'))
		{
		if (RandTrue(g_CGFailProb))
			EV_FAILN(Substitute, GCReject);
		}

	if (g_ATFailProb > 0 && (OldLetter == 'A' || OldLetter == 'T'))
		{
		if (RandTrue(g_ATFailProb))
			EV_FAILN(Substitute, ATReject);
		}

	if (Base.IsCoding())
		{
		if (!AcceptExonSubst(Pos, NewLetter))
			EV_REJECTN(Substitute, CodingConstraint)
		}
	else
		{
		double Prob = Base.GetAcceptProb();
		if (!RandTrue(Prob))
			EV_REJECTN(Substitute, Substitute)
		}

	Base.SetLetter(NewLetter);

	EV_ACCEPTN(Substitute, Pos)
	}
