#include "utils.h"
#include "revfile.h"
#include "getopts.h"
#include "progress.h"

#define	TRACE	0

void RevComp(const std::string &Seq, std::string &RevCompSeq);
void ParseStrVec(const std::string &In, std::vector<std::string> &Out, char c = ',');
void StripGaps(const std::string &In, std::string &Out);

static unsigned g_MaxErrors = 10;
static unsigned g_ErrorCount = 0;
static unsigned g_NoErrorCount = 0;
static unsigned g_SegCount = 0;
static unsigned g_AlnCount = 0;

struct SeqInfo
	{
	unsigned SeqFileIndex;
	unsigned SeqIndex;
	unsigned GenIndex;
	unsigned ChrId;
	};

struct SegInfo
	{
	unsigned GenIndex;
	unsigned ChrId;
	unsigned AIIndex;
	unsigned SegIndex;

	bool operator<(const SegInfo &rhs) const
		{
		if (GenIndex != rhs.GenIndex)
			return GenIndex < rhs.GenIndex;
		return ChrId < rhs.ChrId;
		}
	};

static std::vector<RevFile> g_SeqFiles;
static std::map<std::pair<std::string, std::string>, SeqInfo> g_GenChrNamesToSeqInfo;
static unsigned g_GenIndex = UINT_MAX;
static unsigned g_ChrId = UINT_MAX;
static unsigned g_SeqLength = 0;
static char *g_Seq = 0;

static void LogSeqMap()
	{
	Log("\n");
	Log("     Genome Name          Chr Name  File  Gen   ChrId\n");
	Log("================  ================  ====  ===  ======\n");
	for (std::map<std::pair<std::string, std::string>, SeqInfo>::const_iterator p =
	  g_GenChrNamesToSeqInfo.begin(); p != g_GenChrNamesToSeqInfo.end(); ++p)
		{
		const std::pair<std::string, std::string> &GenChrNames = p->first;
		const SeqInfo &SI = p->second;

		Log("%16.16s", GenChrNames.first.c_str());
		Log("  %16.16s", GenChrNames.second.c_str());
		Log("  %4u", SI.SeqFileIndex);
		Log("  %3u", SI.GenIndex);
		Log("  %6u", SI.ChrId);
		Log("\n");
		}
	}

static void NoErr()
	{
	++g_NoErrorCount;
	}

static void LogNoErrs()
	{
	if (g_NoErrorCount > 0)
		{
		Log("\n");
		Log("[ ... %u with no errors ... ]\n", g_NoErrorCount);
		g_NoErrorCount = 0;
		}
	}

static void LogSeq(const std::string &Seq, unsigned SeqLength)
	{
	unsigned SeqPos = 0;
	unsigned ColCount = 0;
	unsigned GapCount = 0;
	for (;;)
		{
		if (SeqPos >= SeqLength)
			break;
		char c = Seq[SeqPos++];
		if (c == '-')
			++GapCount;
		else
			{
			if (GapCount > 0)
				{
				if (GapCount > 8)
					{
					Log("-(%u)-",  GapCount);
					ColCount += 6;
					}
				else
					{
					for (unsigned i = 0; i < GapCount; ++i)
						Log("-");
					ColCount += GapCount;
					}
				GapCount = 0;
				}
			Log("%c", c);
			++ColCount;
			if (ColCount == 60)
				{
				Log("...");
				break;
				}
			}
		}
	if (GapCount > 0)
		{
		if (GapCount > 8)
			{
			Log("-(%u)-",  GapCount);
			ColCount += 6;
			}
		else
			{
			for (unsigned i = 0; i < GapCount; ++i)
				Log("-");
			ColCount += GapCount;
			}
		}
	}

static void LogSeqContext(const char *Msg, const std::string &Seq, unsigned Pos)
	{
	Log("%s", Msg);

	unsigned Len = SIZE(Seq);
	unsigned PrefixStartPos = Pos < 8 ? 8 : Pos - 8;
	unsigned PrefixEndPos = Pos == 0 ? 0 : Pos - 1;
	unsigned SuffixStartPos = Pos + 1;
	unsigned SuffixEndPos = Pos + 8;
	for (unsigned i = PrefixStartPos; i <= PrefixEndPos; ++i)
		Log("%c", Seq[i]);
	Log(" > %c < ", Seq[Pos]);
	for (unsigned i = SuffixStartPos; i < Len && i <= SuffixEndPos; ++i)
		Log("%c", Seq[i]);
	Log("\n");
	}

static void Err(unsigned AIIndex, const AlnInfo &AI, const GCSeg &Seg,
  const std::string &Row, unsigned RowPos = UINT_MAX, unsigned SeqPos = UINT_MAX)
	{
	++g_ErrorCount;
	LogNoErrs();
	Log("\n");
	Log("\n");
	Log("ERROR\n");
	Log("AI(%u): ", AIIndex);
	LogAI(AI);
	Log("\n");
	Log("Gen=%u Chr=%u lo=%u hi=%u %c\n",
	  Seg.GenomeIndex,
	  Seg.ChrId,
	  (unsigned) Seg.lo,
	  (unsigned) Seg.hi,
	  Seg.Plus ? '+' : '-');

	unsigned RowLength = (unsigned) Row.size();
	unsigned SeqLength = (unsigned) (Seg.hi - Seg.lo + 1);

	if (Seg.Plus)
		{
		Log("Row:  ");
		LogSeq(Row, RowLength);
		Log("\n");
		}
	else
		{
		Log("Row+: ");
		std::string RevCompRow;
		RevComp(Row, RevCompRow);
		LogSeq(RevCompRow, RowLength);
		Log("\n");
		Log("Row-: ");
		LogSeq(Row, RowLength);
		Log("\n");
		}

	Log("Seq:  ");
	if (Seg.lo >= g_SeqLength)
		Log("Seg.lo=%u >= L=%u", Seg.lo, g_SeqLength);
	else
		{
		if (Seg.lo + SeqLength > g_SeqLength)
			{
			unsigned L = g_SeqLength - Seg.lo;
			LogSeq(&g_Seq[Seg.lo], L);
			Log(" (overflow)\n");
			}
		else
			LogSeq(&g_Seq[Seg.lo], SeqLength);
		}
	Log("\n");

	if (RowPos != UINT_MAX && SeqPos != UINT_MAX)
		{
		LogSeqContext("Seq:  ", g_Seq, SeqPos);
		LogSeqContext("Row:  ", Row, RowPos);
		}

	if (g_MaxErrors > 0 && g_ErrorCount >= g_MaxErrors)
		{
		Log("\n");
		Die("Max errors (%u) reached", g_MaxErrors);
		}
	}

static void AddSeqsToMap(const RevFile &SeqFile, unsigned SeqFileIndex)
	{
	const unsigned SeqCount = SeqFile.GetSeqCount();
	for (unsigned SeqIndex = 0; SeqIndex < SeqCount; ++SeqIndex)
		{
		unsigned GenIndex;
		unsigned ChrId;
		SeqFile.GetSeqGenChrByIndex(SeqIndex, &GenIndex, &ChrId);

		std::string GenName = SeqFile.GetGenomeName(GenIndex);
		std::string ChrName = SeqFile.GetChrName(GenIndex, ChrId);

		SeqInfo SI;
		SI.SeqFileIndex = SeqFileIndex;
		SI.SeqIndex = SeqIndex;
		SI.GenIndex = GenIndex;
		SI.ChrId = ChrId;

		std::map<std::pair<std::string, std::string>, SeqInfo>::value_type
		  v(std::pair<std::string, std::string>(GenName, ChrName), SI);

		g_GenChrNamesToSeqInfo.insert(v);
		}
	}

static void GetChr(const std::string &GenName, const std::string &ChrName,
  char **ptrSeq, unsigned *ptrLength)
	{
	std::pair<std::string, std::string> GenChrNames(GenName, ChrName);

	std::map<std::pair<std::string, std::string>, SeqInfo>::const_iterator p =
	  g_GenChrNamesToSeqInfo.find(GenChrNames);

	if (p == g_GenChrNamesToSeqInfo.end())
		{
		LogSeqMap();
		Die("GetChr(%s,%s): not found", GenName.c_str(), ChrName.c_str());
		}

	const SeqInfo &SI = p->second;

	unsigned SeqFileIndex = SI.SeqFileIndex;
	unsigned SeqIndex = SI.SeqIndex;

	assert(SeqFileIndex < (unsigned) g_SeqFiles.size());

	const RevFile &SeqFile = g_SeqFiles[SeqFileIndex];

	coord_t ctSeqLength = (unsigned) SeqFile.GetSeqLengthByIndex(SeqIndex);

	unsigned SeqLength = (unsigned) ctSeqLength;
	if ((coord_t) SeqLength != ctSeqLength)
		Die("SeqLength overflows unsigned int");

	*ptrSeq = myalloc(char, SeqLength);

	unsigned TmpGenIndex;
	unsigned TmpChrId;
	SeqFile.GetSeqByIndex(SeqIndex, *ptrSeq, SeqLength, &TmpGenIndex, &TmpChrId);

	*ptrLength = SeqLength;
	}

static void GetChrSeq(const RevFile &In, unsigned GenIndex, unsigned ChrId)
	{
	if (GenIndex == g_GenIndex && ChrId == g_ChrId)
		return;

	g_GenIndex = GenIndex;
	g_ChrId = ChrId;

	if (g_Seq != 0)
		{
		myfree(g_Seq);
		g_Seq = 0;
		}

	std::string GenName = In.GetGenomeName(GenIndex);
	std::string ChrName = In.GetChrName(GenIndex, ChrId);

	GetChr(GenName, ChrName, &g_Seq, &g_SeqLength);
	}

static void ValidateSeg(const RevFile &In, unsigned SIIndex,
  std::vector<SegInfo> &SIs, const AIVec &AIs)
	{
#if	TRACE
	Log("\n");
	Log("ValidateSeg(SIIndex=%u)\n", SIIndex);
#endif
	++g_SegCount;

	const unsigned SICount = (unsigned) SIs.size();
	const unsigned AICount = (unsigned) AIs.size();
//	const unsigned AlnCount = (unsigned) In.GetAlnCount();

	if (SIIndex >= SICount)
		Die("SIIndex>=SICount");

	const SegInfo &SI = SIs[SIIndex];

	unsigned AIIndex = SI.AIIndex;
	if (AIIndex >= AICount)
		Die("AIIndex>=AICount");

	unsigned AlnIndex = In.GetAlnIndex(AIIndex);
	//if (AlnIndex >= AlnCount)
	//	Die("AlnIndex>=AlnCount");

	const AlnInfo &AI = AIs[AIIndex];
	const unsigned SegCount = (unsigned) AI.Segs.size();

	unsigned SegIndex = SI.SegIndex;
	if (SegIndex >= SegCount)
		Die("SegIndex>=SegCount");

	const GCSeg &Seg = AI.Segs[SegIndex];

#if	TRACE
	Log("Seg=");
	LogGCSeg(Seg);
	Log("\n");
#endif

	StdStrVec Rows;
	In.GetAlnRows(AlnIndex, Rows);

	if ((unsigned) Rows.size() != SegCount)
		Die("Rows.size() != SegCount");

	std::string &Row = Rows[SegIndex];

	unsigned GenIndex = Seg.GenomeIndex;
	unsigned ChrId = Seg.ChrId;

	GetChrSeq(In, GenIndex, ChrId);

	coord_t ctlo = Seg.lo;
	coord_t cthi = Seg.hi;
	
	unsigned lo = (unsigned) ctlo;
	unsigned hi = (unsigned) cthi;

	if ((coord_t) lo != ctlo || (coord_t) hi != cthi)
		Die("lo/hi overflows unsigned int");

	if (lo > hi)
		Die("lo>hi");

	if (hi >= g_SeqLength)
		{
		Log("hi=%u > seq length=%u\n", (unsigned) hi, (unsigned) g_SeqLength);
		Err(AIIndex, AI, Seg, Row);
		return;
		}

	unsigned RowLength = (unsigned) Row.size();

	if (!Seg.Plus)
		{
		std::string RevCompRow;
#if	TRACE
		{
		StdStr StrippedRow;
		StripGaps(Row, StrippedRow);
		Log("Row- %.64s%s\n", StrippedRow.c_str(), SIZE(StrippedRow) > 64 ? "..." : "");
		}
#endif
		RevComp(Row, RevCompRow);
		Row = RevCompRow;
		}

#if	TRACE
	{
	StdStr StrippedRow;
	StripGaps(Row, StrippedRow);
	Log("Row: %.64s%s\n", StrippedRow.c_str(), SIZE(StrippedRow) > 64 ? "..." : "");
	Log("Seq: ");
	for (unsigned i = lo; i <= hi && i <= lo + 63; ++i)
		Log("%c", g_Seq[i]);
	if (Seg.GetLength() > 64)
		Log("...");
	Log("\n");
	}
#endif

	unsigned SeqPos = lo;
	for (unsigned RowPos = 0; RowPos < RowLength; ++RowPos)
		{
		char cRow = Row[RowPos];
		if (cRow == '-')
			continue;

		char cSeq = g_Seq[SeqPos];

		if (toupper(cRow) != toupper(cSeq))
			{
			Err(AIIndex, AI, Seg, Row, RowPos, SeqPos);
			Log("Row[%u] = %c != Seq[%u] = %c\n",
			  RowPos, cRow, SeqPos, cSeq);
			return;
			}
		++SeqPos;
		}
	if (SeqPos != hi + 1)
		{
		Err(AIIndex, AI, Seg, Row);
		Log("SeqPos=%u != hi=%u\n", SeqPos, hi);

		unsigned GetUngappedSeqLength(const string &Seq);
		Log("RowLength=%u UngappedLength=%u SegLength=%u\n",
		  RowLength, GetUngappedSeqLength(Row), Seg.GetLength());
		return;
		}
	NoErr();
	}

void ValidateAlns()
	{
	const char *InFileName = RequiredValueOpt("valalns");
	const char *cstrSeqFileNames = RequiredValueOpt("seqfiles");
	IntOpt("maxerrs", &g_MaxErrors);

	RevFile In;
	In.Open(InFileName);

	std::vector<std::string> SeqFileNames;
	ParseStrVec(cstrSeqFileNames, SeqFileNames);

	const unsigned SeqFileCount = (unsigned) SeqFileNames.size();

	g_SeqFiles.resize(SeqFileCount);
	for (unsigned SeqFileIndex = 0; SeqFileIndex < SeqFileCount; ++SeqFileIndex)
		{
		RevFile &SeqFile = g_SeqFiles[SeqFileIndex];
		const std::string &FileName = SeqFileNames[SeqFileIndex];
		Progress("Opening seq file %s", FileName.c_str());
		SeqFile.Open(FileName);
		AddSeqsToMap(SeqFile, SeqFileIndex);
		}

	AIVec AIs;
	In.GetAllAlnInfos(AIs);

	const unsigned AICount = (unsigned) AIs.size();
//	const unsigned AlnCount = In.GetAlnCount();

	std::vector<SegInfo> SIs;
	SIs.reserve(3*AICount);

	Progress("Building SIs");
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		unsigned AlnIndex = In.GetAlnIndex(AIIndex);
		if (AlnIndex == UINT_MAX)
			continue;

		++g_AlnCount;
		const AlnInfo &AI = AIs[AIIndex];
		const unsigned SegCount = (unsigned) AI.Segs.size();
		for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
			{
			const GCSeg &Seg = AI.Segs[SegIndex];

			SegInfo SI;
			SI.AIIndex = AIIndex;
			SI.SegIndex = SegIndex;
			SI.GenIndex = Seg.GenomeIndex;
			SI.ChrId = Seg.ChrId;

			SIs.push_back(SI);
			}
		}
	const unsigned SICount = (unsigned) SIs.size();

	Progress("%u SIs", SICount);

// Sort to avoid disk thrashing, just one chromsome sequence is kept in memory.
	Progress("Sorting SIs");
	std::sort(SIs.begin(), SIs.end());

#if	TRACE
	{
	Log("\n");
	Log("SegInfos:\n");
//struct SegInfo
//	{
//	unsigned GenIndex;
//	unsigned ChrId;
//	unsigned AIIndex;
//	unsigned SegIndex;
	Log("Index  Gen    Chr  AIIndex  SegIndex          lo          hi  +\n");
	Log("=====  ===  =====  =======  ========  ==========  ==========  =\n");
	for (unsigned SIIndex = 0; SIIndex < SICount; ++SIIndex)
		{
		const SegInfo &SI = SIs[SIIndex];
		const AlnInfo &AI = AIs[SI.AIIndex];
		const GCSeg &Seg = AI.Segs[SI.SegIndex];
		asserta(SI.GenIndex == Seg.GenomeIndex);
		asserta(SI.ChrId == Seg.ChrId);
		Log("%5u", SIIndex);
		Log("  %3u", SI.GenIndex);
		Log("  %5u", SI.ChrId);
		Log("  %7u", SI.AIIndex);
		Log("  %8u", SI.SegIndex);
		Log("  %10u", (unsigned) Seg.lo);
		Log("  %10u", (unsigned) Seg.hi);
		Log("  %c", Seg.Plus ? '+' : '-');
		Log("\n");
		}
	}
#endif

	ProgressStart("Validating");
	unsigned LastErrCount = 0;
	for (unsigned SIIndex = 0; SIIndex < SICount; ++SIIndex)
		{
		if (g_ErrorCount > LastErrCount)
			{
			LastErrCount = g_ErrorCount;
			SetProgressMsg("Validating, %u errors", g_ErrorCount);
			ProgressStep(SIIndex, SICount);
			}
		else if (SIIndex%20 == 0)
			ProgressStep(SIIndex, SICount);

		ValidateSeg(In, SIIndex, SIs, AIs);
		}

	ProgressDone();

	LogNoErrs();

	Progress("%u alns, %u segs, %u errors",
	  g_AlnCount,
	  g_SegCount,
	  g_ErrorCount);
	}
