#include <vector>
#include <map>
#include <string>
#include <iterator>
#include <algorithm>
#include "revfile.h"
#include "intlist.h"
#include "types.h"
#include "utils.h"
#include "getopts.h"
#include "progress.h"

// Singly-linked list optimization for GNU C++
#ifdef _MSC_VER
#include <list>
#define STDList std::list
#else
#include <ext/slist>
#define STDList __gnu_cxx::slist
#endif

/* Some globals */

/* The Common and Unique counts of columns */
class Stats
{
	public:
	
	unsigned TotalCols;
	unsigned TotalClasses;
	unsigned SingleClasses;

	Stats() : TotalCols(0), TotalClasses(0), SingleClasses(0) {}
};

Stats Unique[2];
Stats Common[2];

/* The input files */
RevFile InFiles[2];

/* The output files */
RevFile OutFiles[2];
bool HaveWriteOpt[2] = {false, false};

/* "Reference" and "Other" are arbitrary terms; the process is symmetric.
 * One genome is picked arbitrarily and the alignments are sorted according
 * to the coordinates of that genome; that is the 'reference genome'.
 */

/* RefGenome[i] is the index of the 'reference genome' in the i-th revfile */
int RefGenome[2];
/* OtherGenome[i] is the index of the 'other genome' in the i-th revfile */
int OtherGenome[2];
/* (We expect these indices to be constant; i.e. for any alignment in a
 * revfile, AI.Segs[j].Genome = j) */

/* Progress... */
unsigned TotalAlnCount;
unsigned CurrentAlnCount = 0;

bool WithGaps = false;

/* End of important globals */

void RevComp(std::string &Seq);

VALUE_OPT ValueOpts[] =
{
	{ "in1",				0 },
	{ "in2",				0 },
	{ "out1",				0 },
	{ "out2",				0 },
	{ "log",				0 }
};
int ValueOptCount = sizeof(ValueOpts)/sizeof(ValueOpts[0]);

FLAG_OPT FlagOpts[] =
{
	{ "dummy",				0 },
	{ "withgaps",			0 },
	{ "naive",				0 }
};
int FlagOptCount = sizeof(FlagOpts)/sizeof(FlagOpts[0]);

void Usage()
{
	fprintf(stderr,
"Usage:\n"
"    venn -in1 <revfile> -in2 <revfile> [-out1 <revfile> -out2 <revfile>] [-naive] [-withgaps]\n"
"\n"
	);
}

/* Comparison function to sort AIXs by the coordinate in the reference genome */
class AIXRefGenomeLoCoordCmp
{
	unsigned idx;

	public:
	AIXRefGenomeLoCoordCmp(unsigned RefGenomeIndex) : idx(RefGenomeIndex) { }

	bool operator()(const AlnInfoX &AIX1, const AlnInfoX &AIX2)
	{
		if (AIX1.AI.Segs[idx].lo != AIX2.AI.Segs[idx].lo)
			return (AIX1.AI.Segs[idx].lo < AIX2.AI.Segs[idx].lo);
		else if (AIX1.AI.Segs[idx].hi != AIX2.AI.Segs[idx].hi)
			return (AIX1.AI.Segs[idx].hi < AIX2.AI.Segs[idx].hi);
		else
			return (AIX1.AlnInfoIndex < AIX2.AlnInfoIndex);
	}
};

/* For each pairwise alignment, the following structure contains the
 * needed info (RefSeq and OtherSeq are the two rows, RefCoord and OtherCoord
 * are the "current" coordinates used in a column comparison, Col is the column
 * index being currently compared, and OtherPlus is whether the other is in the
 * plus or minus strand (the Ref is always in the Plus).
 */

class SeqPair
{
	public:

	std::string RefSeq;
	std::string OtherSeq;
	coord_t RefCoord;
	coord_t OtherCoord;
	unsigned Col;
	bool OtherPlus;
	BoolVec CommonBits;
	const AlnInfo &AI;

	SeqPair(const AlnInfo &ai) : AI(ai) {}
};

class ColumnInfo
{
	public:
	
	coord_t OtherCoord;
	bool OtherGapped;
	std::list<SeqPair>::iterator SPiter;
	unsigned Col;

	ColumnInfo(coord_t c, bool b, std::list<SeqPair>::iterator i, unsigned u) : OtherCoord(c), OtherGapped(b), SPiter(i), Col(u) {}

	bool operator<(const ColumnInfo &ci) const
	{
		if (OtherCoord != ci.OtherCoord)
			return (OtherCoord < ci.OtherCoord);
		else
			return (OtherGapped < ci.OtherGapped);
	}

	bool operator==(const ColumnInfo &ci) const
	{
		return ((OtherCoord == ci.OtherCoord) && (OtherGapped == ci.OtherGapped));
	}

	bool operator!=(const ColumnInfo &ci) const
	{
		return ((OtherCoord != ci.OtherCoord) || (OtherGapped != ci.OtherGapped));
	}
};

class NaiveColumnInfo
{
	public:
	
	unsigned RefChrId;
	coord_t RefCoord;
	unsigned OtherChrId;
	coord_t OtherCoord;
	bool RefIsGap;
	bool OtherIsGap;

	bool operator<(const NaiveColumnInfo &other) const
	{
		if (RefChrId != other.RefChrId)
			return RefChrId < other.RefChrId;
		else if (RefCoord != other.RefCoord)
			return RefCoord < other.RefCoord;
		else if (OtherChrId != other.OtherChrId)
			return OtherChrId < other.OtherChrId;
		else if (OtherCoord != other.OtherCoord)
			return (OtherCoord < other.OtherCoord);
		else if (RefIsGap != other.RefIsGap)
			return ((!RefIsGap) && (other.RefIsGap));
		else 
			return ((!OtherIsGap) && (other.OtherIsGap));
	}

	bool operator==(const NaiveColumnInfo &other) const
	{
		return (RefChrId == other.RefChrId)
			&& (RefCoord == other.RefCoord)
			&& (OtherChrId == other.OtherChrId)
			&& (OtherCoord == other.OtherCoord)
			&& (RefIsGap == other.RefIsGap)
			&& (OtherIsGap == other.OtherIsGap);
	}

	bool operator!=(const NaiveColumnInfo &other) const
	{
		return (RefChrId != other.RefChrId)
			|| (RefCoord != other.RefCoord)
			|| (OtherChrId != other.OtherChrId)
			|| (OtherCoord != other.OtherCoord)
			|| (RefIsGap != other.RefIsGap)
			|| (OtherIsGap != other.OtherIsGap);
	}
};

void WriteOut(int idx, SeqPair &SP)
{
	AlnInfo AIOut = SP.AI;
	StdStrVec Rows(2);

	/* Normalize the orientations */
	bool OtherPlus = ((SP.AI.Segs[RefGenome[idx]].Plus) == (SP.AI.Segs[OtherGenome[idx]].Plus));
	AIOut.Segs[RefGenome[idx]].Plus = true;
	AIOut.Segs[OtherGenome[idx]].Plus = OtherPlus;
	AIOut.Segs[RefGenome[idx]].lo = AIOut.Segs[RefGenome[idx]].hi = SP.AI.Segs[RefGenome[idx]].lo;
	if (OtherPlus)
		AIOut.Segs[OtherGenome[idx]].lo = AIOut.Segs[OtherGenome[idx]].hi = SP.AI.Segs[OtherGenome[idx]].lo;
	else
		AIOut.Segs[OtherGenome[idx]].lo = AIOut.Segs[OtherGenome[idx]].hi = SP.AI.Segs[OtherGenome[idx]].hi;

	/* Go through the vector */
	assert(SP.CommonBits.size() > 0);
	unsigned RefCount = (SP.RefSeq[0] != '-') ? 1 : 0;
	unsigned OtherCount = (SP.OtherSeq[0] != '-') ? 1 : 0;
	unsigned Start = 0;
	for (unsigned j = 1; j != SP.CommonBits.size(); ++j)
	{
		if ((!WithGaps) && ((SP.RefSeq[j] == '-') || (SP.OtherSeq[j] == '-')))
			SP.CommonBits[j] = true;
		if (SP.CommonBits[j] != SP.CommonBits[Start])
		{
			//Output block from Start to j-1
			Rows[RefGenome[idx]] = SP.RefSeq.substr(Start, j - Start);
			Rows[OtherGenome[idx]] = SP.OtherSeq.substr(Start, j - Start);
			//This is wrong if RefCount/OtherCount is 0, but then we won't output that block anyway
			AIOut.Segs[RefGenome[idx]].hi = AIOut.Segs[RefGenome[idx]].lo + RefCount - 1;
			if (OtherPlus)
				AIOut.Segs[OtherGenome[idx]].hi = AIOut.Segs[OtherGenome[idx]].lo + OtherCount - 1;
			else
				AIOut.Segs[OtherGenome[idx]].lo = AIOut.Segs[OtherGenome[idx]].hi - OtherCount + 1;
			//TODO: Set some attribute based on CommonBits[Start]
			if ((RefCount > 0) && (OtherCount > 0) && !(SP.CommonBits[Start]))
				OutFiles[idx].AddAln(AIOut, Rows);
			AIOut.Segs[RefGenome[idx]].lo += RefCount;
			if (OtherPlus)
				AIOut.Segs[OtherGenome[idx]].lo += OtherCount;
			else
				AIOut.Segs[OtherGenome[idx]].hi -= OtherCount;
			RefCount = OtherCount = 0;
			Start = j;
		}
		if (SP.RefSeq[j] != '-')
			++RefCount;
		if (SP.OtherSeq[j] != '-')
			++OtherCount;
	}
	//Output block from Start to size()-1
	Rows[RefGenome[idx]] = SP.RefSeq.substr(Start, SP.CommonBits.size() - Start);
	Rows[OtherGenome[idx]] = SP.OtherSeq.substr(Start, SP.CommonBits.size() - Start);
	AIOut.Segs[RefGenome[idx]].hi = AIOut.Segs[RefGenome[idx]].lo + RefCount - 1;
	if (OtherPlus)
		AIOut.Segs[OtherGenome[idx]].hi = AIOut.Segs[OtherGenome[idx]].lo + OtherCount - 1;
	else
		AIOut.Segs[OtherGenome[idx]].lo = AIOut.Segs[OtherGenome[idx]].hi - OtherCount + 1;
	//TODO: Set some attribute based on CommonBits[Start]
	if ((RefCount > 0) && (OtherCount > 0) && !(SP.CommonBits[Start]))
		OutFiles[idx].AddAln(AIOut, Rows);
}

/* Remove a sequence of identical entities from a sorted multiset, and update the stats */
void inline ConsumeUnique(std::multiset<ColumnInfo>::iterator &iter, const std::multiset<ColumnInfo>::iterator &end_iter, Stats &S)
{
	std::multiset<ColumnInfo>::iterator original_iter = iter;
	++(S.TotalCols);
	++(S.TotalClasses);
	++iter;
	if ((iter == end_iter) || ((*iter) != (*original_iter)))
		++(S.SingleClasses);
	else
		while ((iter != end_iter) && ((*iter) == (*original_iter)))
		{
			++(S.TotalCols);
			++iter;
		}
}
void inline ConsumeCommon(std::multiset<ColumnInfo>::iterator &iter, const std::multiset<ColumnInfo>::iterator &end_iter, Stats &S)
{
	std::multiset<ColumnInfo>::iterator original_iter = iter;
	++(S.TotalCols);
	++(S.TotalClasses);
	iter->SPiter->CommonBits[iter->Col] = true;
	++iter;
	if ((iter == end_iter) || ((*iter) != (*original_iter)))
		++(S.SingleClasses);
	else
		while ((iter != end_iter) && ((*iter) == (*original_iter)))
		{
			++(S.TotalCols);
			iter->SPiter->CommonBits[iter->Col] = true;
			++iter;
		}
}

void NaiveConsume(std::multiset<NaiveColumnInfo>::iterator &iter, const std::multiset<NaiveColumnInfo>::iterator &end_iter, Stats &S)
{
	std::multiset<NaiveColumnInfo>::iterator original_iter = iter;
	++(S.TotalCols);
	++(S.TotalClasses);
	++iter;
	if ((iter == end_iter) || ((*iter) != (*original_iter)))
		++(S.SingleClasses);
	else
		while ((iter != end_iter) && ((*iter) == (*original_iter)))
		{
			++(S.TotalCols);
			++iter;
		}
}

/* Compare alignments of a fixed chromosome pair */
void CompareAlignments(STDList<unsigned> &AlnList1, STDList<unsigned> &AlnList2)
{
	typedef std::set<AlnInfoX, AIXRefGenomeLoCoordCmp> AIXSet;

	AIXSet AIXs[2] = {
		AIXSet(AIXRefGenomeLoCoordCmp(RefGenome[0])),
		AIXSet(AIXRefGenomeLoCoordCmp(RefGenome[1]))};

	/* Retrieve all AlnInfoXs */
	for (STDList<unsigned>::const_iterator i = AlnList1.begin(); i != AlnList1.end(); ++i)
	{
		AlnInfoX AIX;
		InFiles[0].GetAlnInfoX(*i, AIX);
		AIXs[0].insert(AIX);
	}
	AlnList1.clear();
	for (STDList<unsigned>::const_iterator i = AlnList2.begin(); i != AlnList2.end(); ++i)
	{
		AlnInfoX AIX;
		InFiles[1].GetAlnInfoX(*i, AIX);
		AIXs[1].insert(AIX);
	}
	AlnList2.clear();

	/* Go through all the alignments, sorted by the reference genome coordinate. The following
	 * algorithm sweeps through the reference genome and for each position, it populates two
	 * multisets with the coordinates in the other genome, then compares those sets.
	 *
	 * For each reference genome coordinate, say X, it runs two steps: one for the gapped
	 * pieces that correspond to (X-1,X), and one step for the ungapped columns that correspond
	 * to X. For example, consider this alignment:
	 *
	 * 12--3
	 * 56789
	 *
	 * For the reference coordinate "3", the first step will process the (-,7) and (-,8) columns
	 * and the second step will process the (3,9) column. These steps are distinguished by the
	 * GapBefore variable.
	 *
	 * The current position in the reference genome is kept in CurrentCoord. All alignments
	 * overlapping the current coordinate are stored in Seqs. Each such alignment has some
	 * book-keeping (current column in the row, etc.). As CurrentCoord increases, we move to
	 * the next column of each alignment and fill in the multiset.
	 *
	 * The alignments overlapping the current coordinate are populated in Seqs when needed;
	 * initially all sorted AlnInfos are in AIXs, and their alignments are extracted to Seqs
	 * when needed.
	 */

	/* Pointer to the next alignment in each revfile */
	AIXSet::const_iterator nextptr[2] = {AIXs[0].begin(), AIXs[1].begin()};
	/* The alignments overlapping the current coordinate */
	std::list<SeqPair> Seqs[2];
	/* Helper list of SeqPairs to be deleted from Seqs */
	std::list<SeqPair> SeqsToDel[2];
	/* Current coordinate, and coordinate of next (closest, in increasing lo coordinate) alignment */
	coord_t CurrentCoord = 0, NextCoord = 0;
	/* The multisets that are filled and compared for every column */
	std::multiset<ColumnInfo> OtherCoords[2];

	bool GapBefore = true, HaveNextCoord = true, EmptySeqs = true;

	for(;;)
	{
		/* Calculate next coordinate */
		if ((nextptr[0] == AIXs[0].end()) && (nextptr[1] == AIXs[1].end()))
			HaveNextCoord = false;
		else if ((nextptr[0] == AIXs[0].end()) && (nextptr[1] != AIXs[1].end()))
			NextCoord = nextptr[1]->AI.Segs[RefGenome[1]].lo;
		else if ((nextptr[0] != AIXs[0].end()) && (nextptr[1] == AIXs[1].end()))
			NextCoord = nextptr[0]->AI.Segs[RefGenome[0]].lo;
		else
			NextCoord = std::min(nextptr[0]->AI.Segs[RefGenome[0]].lo, nextptr[1]->AI.Segs[RefGenome[1]].lo);

		/* Process the current alignments till the next coordinate */
		while (!(EmptySeqs || (HaveNextCoord && (CurrentCoord == NextCoord))))
		{
			/* Populate the multisets */
			for (int i = 0; i != 2; ++i)
			{
				OtherCoords[i].clear();
				std::list<SeqPair>::iterator p = Seqs[i].begin();
				if (GapBefore)
				{
					while (p != Seqs[i].end())
					{
						assert(p->RefCoord == CurrentCoord);
						while (p->RefSeq[p->Col] == '-')
						{
							assert(p->OtherSeq[p->Col] != '-');
							if (WithGaps)
								OtherCoords[i].insert(ColumnInfo(p->OtherCoord, true, p, p->Col));
							++(p->Col);
							if (p->Col == p->RefSeq.size())
							{
								std::list<SeqPair>::iterator p_prev = p;
								++p;
								SeqsToDel[i].splice(SeqsToDel[i].end(), Seqs[i], p_prev);
								goto loopend;
							}
							if (p->OtherPlus)
								++(p->OtherCoord);
							else if (p->OtherSeq[p->Col] != '-')
								--(p->OtherCoord);
						}
						++p;
						loopend:
						;
					}
				}
				else
				{
					while (p != Seqs[i].end())
					{
						assert(p->RefSeq[p->Col] != '-');
						assert(p->RefCoord == CurrentCoord);
						bool OtherGapped = (p->OtherSeq[p->Col] == '-');
						if (!OtherGapped || WithGaps)
							OtherCoords[i].insert(ColumnInfo(p->OtherCoord, OtherGapped, p, p->Col));
						++(p->Col);
						if (p->Col == p->RefSeq.size())
						{
							std::list<SeqPair>::iterator p_prev = p;
							++p;
							SeqsToDel[i].splice(SeqsToDel[i].end(), Seqs[i], p_prev);
							continue;
						}
						++(p->RefCoord);
						if (p->OtherPlus)
						{
							if (!OtherGapped)
								++(p->OtherCoord);
						}
						else if (p->OtherSeq[p->Col] != '-')
							--(p->OtherCoord);
						++p;
					}
				}
			}

			/* Calculate unique and common columns */
			std::multiset<ColumnInfo>::iterator iter[2] = {OtherCoords[0].begin(), OtherCoords[1].begin()};

			while ((iter[0] != OtherCoords[0].end()) && (iter[1] != OtherCoords[1].end()))
			{
				if ((*(iter[0])) == (*(iter[1])))
				{
					ConsumeCommon(iter[0], OtherCoords[0].end(), Common[0]);
					ConsumeCommon(iter[1], OtherCoords[1].end(), Common[1]);
				}
				else if ((*(iter[0])) < (*(iter[1])))
					ConsumeUnique(iter[0], OtherCoords[0].end(), Unique[0]);
				else
					ConsumeUnique(iter[1], OtherCoords[1].end(), Unique[1]);
			}
			while (iter[0] != OtherCoords[0].end())
				ConsumeUnique(iter[0], OtherCoords[0].end(), Unique[0]);
			while (iter[1] != OtherCoords[1].end())
				ConsumeUnique(iter[1], OtherCoords[1].end(), Unique[1]);

			/* Remove sequences no longer needed */
			for (int i = 0; i != 2; ++i)
			{
				if (HaveWriteOpt[i])
					for (std::list<SeqPair>::iterator p = SeqsToDel[i].begin(); !(SeqsToDel[i].empty()); p = SeqsToDel[i].erase(p))
						WriteOut(i, *p);
				else
					SeqsToDel[i].clear();
			}

			/* Either switch from step 1 (GapBefore == true) to step 2 (GapBefore = false) within
			 * the same coordinate, or move to the next coordinate */
			if (!GapBefore)
				++CurrentCoord;
			GapBefore = !GapBefore;

			EmptySeqs = (Seqs[0].empty() && Seqs[1].empty());
		}

		if (!HaveNextCoord)
			break;

		/* Retrieve the next alignment */
		for (int i = 0; i != 2; ++i)
		{
			while ((nextptr[i] != AIXs[i].end()) && (nextptr[i]->AI.Segs[RefGenome[i]].lo == NextCoord))
			{
				const AlnInfo &AI = nextptr[i]->AI;
				std::list<SeqPair>::iterator sp = Seqs[i].insert(Seqs[i].end(), SeqPair(AI));
				StdStrVec Rows;
				InFiles[i].GetAlnRows(nextptr[i]->AlnIndex, Rows);
				sp->RefSeq = Rows[RefGenome[i]];
				sp->OtherSeq = Rows[OtherGenome[i]];
				sp->CommonBits.resize(sp->RefSeq.size(), false);
				if (!(AI.Segs[RefGenome[i]].Plus))
				{
					RevComp(sp->RefSeq);
					RevComp(sp->OtherSeq);
				}
				sp->RefCoord = AI.Segs[RefGenome[i]].lo;
				if ((AI.Segs[RefGenome[i]].Plus) != (AI.Segs[OtherGenome[i]].Plus))
				{
					sp->OtherCoord = AI.Segs[OtherGenome[i]].hi;
					sp->OtherPlus = false;
					if (sp->OtherSeq[0] == '-')
						++(sp->OtherCoord);
				}
				else
				{
					sp->OtherCoord = AI.Segs[OtherGenome[i]].lo;
					sp->OtherPlus = true;
				}
				sp->Col = 0;
				++(nextptr[i]);
				++CurrentAlnCount;
				if (!(CurrentAlnCount & 0xffff))
					ProgressStep(CurrentAlnCount, TotalAlnCount);
			}
		}

		if (EmptySeqs)
		{
			CurrentCoord = NextCoord;
			GapBefore = true;
		}

		EmptySeqs = false;
	}
}

void VennNaive()
{
	std::multiset<NaiveColumnInfo> Columns[2];

	ProgressStart("Processing...");
	/* For each input file */
	for (unsigned i = 0; i != 2; ++i)
	{
		/* For each alignment */
		unsigned FileAlnCount = InFiles[i].GetAlnInfoCount();
		for (unsigned j = 0; j != FileAlnCount; ++j)
		{
			AlnInfo AI;
			unsigned AlnIndex = InFiles[i].GetAlnInfo(j, AI);
			StdStrVec Rows;
			InFiles[i].GetAlnRows(AlnIndex, Rows);

			std::string RefSeq = Rows[RefGenome[i]];
			std::string OtherSeq = Rows[OtherGenome[i]];
			if (!(AI.Segs[RefGenome[i]].Plus))
			{
				std::reverse(RefSeq.begin(), RefSeq.end());
				std::reverse(OtherSeq.begin(), OtherSeq.end());
			}

			/* Initialize a column */
			NaiveColumnInfo CI;
			CI.RefChrId = AI.Segs[RefGenome[i]].ChrId;
			CI.RefCoord = AI.Segs[RefGenome[i]].lo;
			CI.OtherChrId = AI.Segs[OtherGenome[i]].ChrId;

			bool OtherPlus = ((AI.Segs[RefGenome[i]].Plus) == (AI.Segs[OtherGenome[i]].Plus));
			if (OtherPlus)
				CI.OtherCoord = AI.Segs[OtherGenome[i]].lo;
			else
			{
				CI.OtherCoord = AI.Segs[OtherGenome[i]].hi;
				if (OtherSeq[0] == '-')
					++(CI.OtherCoord);
			}

			/* Go through the alignment and add columns to the sets */
			unsigned k = 0;
			for(;;)
			{
				CI.RefIsGap = (RefSeq[k] == '-');
				CI.OtherIsGap = (OtherSeq[k] == '-');
				if ((!(CI.RefIsGap || CI.OtherIsGap)) || WithGaps)
					Columns[i].insert(CI);
				++k;
				if (k == RefSeq.size())
					break;
				if (!CI.RefIsGap)
					++(CI.RefCoord);
				if (OtherPlus)
				{
					if (!CI.OtherIsGap)
						++(CI.OtherCoord);
				}
				else
				{
					if (OtherSeq[k] != '-')
						--(CI.OtherCoord);
				}
			}

			++CurrentAlnCount;
			if (!(CurrentAlnCount & 0xffff))
				ProgressStep(CurrentAlnCount, TotalAlnCount);
		}
	}
	ProgressDone();


	/* Intersect the sets */
	Progress("Intersecting...");
	std::multiset<NaiveColumnInfo>::iterator iter[2] = {Columns[0].begin(), Columns[1].begin()};

	while ((iter[0] != Columns[0].end()) && (iter[1] != Columns[1].end()))
	{
		if ((*(iter[0])) == (*(iter[1])))
		{
			NaiveConsume(iter[0], Columns[0].end(), Common[0]);
			NaiveConsume(iter[1], Columns[1].end(), Common[1]);
		}
		else if ((*(iter[0])) < (*(iter[1])))
			NaiveConsume(iter[0], Columns[0].end(), Unique[0]);
		else
			NaiveConsume(iter[1], Columns[1].end(), Unique[1]);
	}
	while (iter[0] != Columns[0].end())
		NaiveConsume(iter[0], Columns[0].end(), Unique[0]);
	while (iter[1] != Columns[1].end())
		NaiveConsume(iter[1], Columns[1].end(), Unique[1]);
}

void Venn()
{
	if ((InFiles[0].GetGenomeCount() != 2) || (InFiles[1].GetGenomeCount() != 2))
		Die("This program can only handle two genomes at a time.");

	/* Verify compatibility of the two RevFiles */
	if ((InFiles[0].GetGenomeInfo(0) == InFiles[1].GetGenomeInfo(0))
	    && (InFiles[0].GetGenomeInfo(1) == InFiles[1].GetGenomeInfo(1)))
	{
		RefGenome[0] = 0;
		RefGenome[1] = 0;
		OtherGenome[0] = 1;
		OtherGenome[1] = 1;
	}
	else if ((InFiles[0].GetGenomeInfo(0) == InFiles[1].GetGenomeInfo(1))
	    && (InFiles[0].GetGenomeInfo(1) == InFiles[1].GetGenomeInfo(0)))
	{
		RefGenome[0] = 0;
		RefGenome[1] = 1;
		OtherGenome[0] = 1;
		OtherGenome[1] = 0;
	}
	else
		Die("The genomes do not match.");

	TotalAlnCount = InFiles[0].GetAlnInfoCount() + InFiles[1].GetAlnInfoCount();

	if (FlagOpt("naive"))
	{
		VennNaive();
		return;
	}

	const unsigned RefGenomeChrCount = InFiles[0].GetChrCount(RefGenome[0]);

	ProgressStart("Processing...");
	/* For every chromosome of the 'reference genome'... */
	for (unsigned ChrIndex = 0; ChrIndex != RefGenomeChrCount; ++ChrIndex)
	{
		/* Retrieve alignments of that reference chromosome */
		IntVec AlnInfoIndices[2];
		for (int idx = 0; idx != 2; ++idx)
		{
			unsigned ChrId = InFiles[idx].GetGenomeInfo(RefGenome[idx]).GetChrId(ChrIndex);
			AlnInfoIndices[idx].reserve(4096);
			InFiles[idx].GetAppendAlnInfoIndexesChr(RefGenome[idx], ChrId, AlnInfoIndices[idx]);
		}

		/* Separate alignments by the other chromosome */
		std::map<unsigned, STDList<unsigned> > AlnMap[2];
		std::map<unsigned, STDList<unsigned> >::iterator ptr[2];
		for (int idx = 0; idx != 2; ++idx)
		{
			for_CIntVec(AlnInfoIndices[idx], p)
			{
				AlnInfo AI;
				InFiles[idx].GetAlnInfo(*p, AI);
				AlnMap[idx][AI.Segs[OtherGenome[idx]].ChrId].push_front(*p);
			}
			ptr[idx] = AlnMap[idx].begin();
			AlnInfoIndices[idx].clear();
		}

		STDList<unsigned> DummyList;
		/* Process each std::pair of chromosomes that are common in both files */
		while ((ptr[0] != AlnMap[0].end()) && (ptr[1] != AlnMap[1].end()))
		{
			if (ptr[0]->first == ptr[1]->first)
			{
				CompareAlignments(ptr[0]->second, ptr[1]->second);
				++ptr[0];
				++ptr[1];
			}
			else if (ptr[0]->first < ptr[1]->first)
			{
				CompareAlignments(ptr[0]->second, DummyList);
				++ptr[0];
			}
			else
			{
				CompareAlignments(DummyList, ptr[1]->second);
				++ptr[1];
			}
		}
		while (ptr[0] != AlnMap[0].end())
		{
			CompareAlignments(ptr[0]->second, DummyList);
			++ptr[0];
		}
		while (ptr[1] != AlnMap[1].end())
		{
			CompareAlignments(DummyList, ptr[1]->second);
			++ptr[1];
		}
	}
	ProgressDone();
}

void PrintStats()
{
	for (int i = 0; i != 2; ++i)
		printf("File #%d contains %u unique columns in %u equiv. classes (%u of which are non-singles)\n",
            i+1, Unique[i].TotalCols, Unique[i].TotalClasses, Unique[i].TotalClasses - Unique[i].SingleClasses);

	for (int i = 0; i != 2; ++i)
		printf("File #%d contains %u common columns in %u equiv. classes (%u of which are non-singles)\n",
		        i+1, Common[i].TotalCols, Common[i].TotalClasses, Common[i].TotalClasses - Common[i].SingleClasses);
}

int main(int argc, char *argv[])
{
	ProcessArgVect(argc, argv);
	const char *InFileName[2];
	const char *OutFileName[2];

	InFileName[0] = RequiredValueOpt("in1");
	InFileName[1] = RequiredValueOpt("in2");
	OutFileName[0] = ValueOpt("out1");
	OutFileName[1] = ValueOpt("out2");

	WithGaps = FlagOpt("withgaps");

	for (int i = 0; i != 2; ++i)
	{
		InFiles[i].Open(InFileName[i]);
		if (OutFileName[i] != NULL)
		{
			if (FlagOpt("naive"))
				Die("-naive does not support -out");
			OutFiles[i].Create(OutFileName[i], InFiles[i].GetGenomes());
			HaveWriteOpt[i] = true;
		}
	}

	Venn();
	PrintStats();

	for (int i = 0; i != 2; ++i)
	{
		InFiles[i].Close();
		if (HaveWriteOpt[i])
			OutFiles[i].Close();
	}

	return 0;
}
