#include "revfile.h"
#include "getopts.h"
#include "utils.h"
#include "progress.h"
#include "gff.h"
#include <set>

void WriteFasta(FILE *f, const char *SeqData, unsigned Length);
void RevComp(char *s, unsigned L);

static std::set<std::string> g_MissingLabels;
static std::set<std::string> g_MissingSeqs;

void WarnMissingLabel(const std::string &Label)
	{
	if (g_MissingLabels.find(Label) != g_MissingLabels.end())
		return;
	Warning("Sequence label '%s' not found in rev file", Label.c_str());
	g_MissingLabels.insert(Label);
	}

void WarnMissingSeq(const std::string &Label)
	{
	if (g_MissingSeqs.find(Label) != g_MissingSeqs.end())
		return;
	Warning("Sequence '%s' not found in rev file", Label.c_str());
	g_MissingSeqs.insert(Label);
	}

void XGFFSeqs()
	{
	const char *InFileName = RequiredValueOpt("xgffseqs");
	const char *OutFileName = RequiredValueOpt("out");
	const char *GFFFileName = RequiredValueOpt("gff");
	const char *GenomeName = RequiredValueOpt("genome");

	FILE *fGFF = OpenStdioFile(GFFFileName);

	RevFile In;

	In.Open(InFileName);
	//const std::vector<GenInfo> &Genomes = In.GetGenomes();
	unsigned GenomeIndex = In.GetGenomeIndex(GenomeName);

	FILE *fOut = CreateStdioFile(OutFileName);

	GFFRecord Rec;
	off64_t FileSize = GetStdioFileSize(fGFF);
	unsigned RecordCount = 0;
	ProgressStart("Processing");
	while (GetNextGFFRecord(fGFF, Rec))
		{
		if (RecordCount%100 == 0)
			{
			off64_t FilePos = GetStdioFilePos(fGFF);
			double Fract = (double) FilePos / (double) FileSize;
			unsigned Mills = (unsigned) (Fract*10000);
			ProgressStep(Mills, 10000);
			}
		++RecordCount;
		asserta(Rec.Start > 0);
		asserta(Rec.End >= Rec.Start);

		const ChrInfo *CI = In.GetChrInfo(GenomeIndex, Rec.SeqName);
		if (CI == 0)
			{
			WarnMissingLabel(Rec.SeqName);
			continue;
			}

		unsigned ChrId = CI->GetId();
		if (!In.IsSeqPresent(GenomeIndex, ChrId))
			{
			WarnMissingSeq(Rec.SeqName);
			continue;
			}

		coord_t lo = Rec.Start - 1;
		coord_t hi = Rec.End - 1;
		coord_t Length = hi - lo + 1;
		coord_t SeqLength = In.GetSeqLength(GenomeIndex, ChrId);
		if (hi >= SeqLength)
			{
			Log("\n");
			Rec.LogMe();
			Log("\n");
			Warning("Coordinate past end of sequence in rev file");
			continue;
			}
		char *Seq = new char[Length];
		In.GetSubSeq(GenomeIndex, ChrId, Seq, lo, Length);
		if (Rec.Strand == '-')
			RevComp(Seq, Length);
		fprintf(fOut, ">%s[%"COORD_T_PRINTF"-%"COORD_T_PRINTF"]%c;%s",
		  Rec.SeqName,
		  lo,
		  hi,
		  Rec.Strand,
		  Rec.Feature);
		if (Rec.Frame == INT_MAX)
			fprintf(fOut, ";.");
		else
			fprintf(fOut, ";%d", Rec.Frame);
		fprintf(fOut, ";%s", Rec.Attrs);
		fprintf(fOut, "\n");
		WriteFasta(fOut, Seq, Length);
		}
	ProgressDone();

	CloseStdioFile(fGFF);
	CloseStdioFile(fOut);
	In.Close();
	}
