#include "utils.h"
#include <ctype.h>
#include "fastastr.h"
#include "xlat.h"

#define TRACE		0
#define VERBOSE		0

void WriteFasta(FILE *f, const char *SeqData, unsigned Length);
char CharsToAmino(char c1, char c2, char c3);

extern unsigned g_MinFrameLength;
extern const char *g_SepStr;

extern unsigned g_PlusORFCount;
extern unsigned g_MinusORFCount;
extern uint64 g_PlusORFBaseCount;
extern uint64 g_MinusORFBaseCount;

extern bool g_Stops;

enum CODON_TYPE
	{
	START_CODON,
	STOP_CODON,
	OTHER_CODON,
	};

static CODON_TYPE CodonType(char c1, char c2, char c3)
	{
	c1 = toupper(c1);
	c2 = toupper(c2);
	c3 = toupper(c3);

// Start = ATG
	if (c1 == 'A' && c2 == 'T' && c3 == 'G')
		return START_CODON;

// Stop = TAA, TAG or TGA
	if (c1 == 'T' && c2 == 'A' && c3 == 'A')
		return STOP_CODON;

	if (c1 == 'T' && c2 == 'A' && c3 == 'G')
		return STOP_CODON;

	if (c1 == 'T' && c2 == 'G' && c3 == 'A')
		return STOP_CODON;

	return OTHER_CODON;
	}

#ifndef NDEBUG
static char *CodonTypeToStr(CODON_TYPE t)
	{
	switch (t)
		{
	case START_CODON:
		return "START";
	case STOP_CODON:
		return "STOP ";
	case OTHER_CODON:
		return "Other";
		}
	return "?CodonType?";
	}
#endif

static char CompLetter[256];
static bool CompLetterInit()
	{
	CompLetter[(unsigned char) 'A'] = 'T';
	CompLetter[(unsigned char) 'C'] = 'G';
	CompLetter[(unsigned char) 'G'] = 'C';
	CompLetter[(unsigned char) 'T'] = 'A';
	CompLetter[(unsigned char) 'N'] = 'N';

	CompLetter[(unsigned char) 'a'] = 't';
	CompLetter[(unsigned char) 'c'] = 'g';
	CompLetter[(unsigned char) 'g'] = 'c';
	CompLetter[(unsigned char) 't'] = 'a';
	CompLetter[(unsigned char) 'n'] = 'n';

	return true;
	}
static bool CompLetterInitDone = CompLetterInit();

static CODON_TYPE GetCodon(const std::vector<unsigned char> &Seq, size_t Pos)
	{
	assert(Pos + 2 < Seq.size());

	char c1 = Seq[Pos];
	char c2 = Seq[Pos+1];
	char c3 = Seq[Pos+2];

	return CodonType(c1, c2, c3);
	}

static CODON_TYPE GetCodonRC(const std::vector<unsigned char> &Seq, size_t Pos)
	{
	assert(Pos >= 2);

	char c1 = CompLetter[(unsigned char) Seq[Pos]];
	char c2 = CompLetter[(unsigned char) Seq[Pos-1]];
	char c3 = CompLetter[(unsigned char) Seq[Pos-2]];

	return CodonType(c1, c2, c3);
	}

#ifndef NDEBUG
static void AssertFrame(const std::vector<unsigned char> &Seq, size_t Pos,
  unsigned CodonCount)
	{
	assert(CodonCount > 0);
	CODON_TYPE Codon = GetCodon(Seq, Pos);
	assert(Codon == START_CODON);
	for (unsigned CodonIndex = 1; CodonIndex < CodonCount - 1; ++CodonIndex)
		{
		if (GetCodon(Seq, Pos + 3*CodonIndex) == STOP_CODON)
			{
			for (unsigned i = 0; i < CodonCount - 1; ++i)
				{
				Codon = GetCodon(Seq, Pos + 3*i);
				Log("  Pos  %9u  Codon %5u  %c%c%c  %s\n",
				  Pos + 3*i,
				  i,
				  CompLetter[Seq[Pos + 3*i]],
				  CompLetter[Seq[Pos + 3*i + 1]],
				  CompLetter[Seq[Pos + 3*i + 2]],
				  CodonTypeToStr(Codon));
				}
			Die("AssertFrame: found STOP in frame starting at %u, codon %u",
			  (unsigned) Pos, CodonIndex);
			}
		}
	Codon = GetCodon(Seq, Pos + 3*(CodonCount - 1));
	assert(Codon == STOP_CODON);
	}

static void AssertFrameRC(const std::vector<unsigned char> &Seq, size_t Pos,
  unsigned CodonCount)
	{
	assert(CodonCount > 0);
	CODON_TYPE Codon = GetCodonRC(Seq, Pos);
	assert(Codon == START_CODON);
	for (unsigned CodonIndex = 0; CodonIndex < CodonCount - 1; ++CodonIndex)
		{
		if (GetCodonRC(Seq, Pos - 3*CodonIndex) == STOP_CODON)
			{
			for (unsigned i = 0; i < CodonCount - 1; ++i)
				{
				Codon = GetCodonRC(Seq, Pos - 3*i);
				Log("  Pos  %9u  Codon %5u  %c%c%c  %s\n",
				  Pos - 3*i,
				  i,
				  CompLetter[Seq[Pos - 3*i]],
				  CompLetter[Seq[Pos - 3*i - 1]],
				  CompLetter[Seq[Pos - 3*i - 2]],
				  CodonTypeToStr(Codon));
				}
			
			Die("AssertFrameRC: found STOP in frame starting at %u, codon %u",
			  (unsigned) Pos, CodonIndex);
			}
		}
	Codon = GetCodonRC(Seq, Pos - 3*(CodonCount - 1));
	if (Codon != STOP_CODON)
		{
		Log("Frame starting at %u expected stop at pos %u codon %u, got %c%c%c %s",
		  (unsigned) Pos,
		  (unsigned) (Pos - 3*(CodonCount - 1)),
		  CodonCount - 1, 
		  CompLetter[Seq[Pos - 3*(CodonCount-1)]],
		  CompLetter[Seq[Pos - 3*(CodonCount-1) - 1]],
		  CompLetter[Seq[Pos - 3*(CodonCount-1) - 2]],
		  CodonTypeToStr(Codon));
		Die("AssertFrameRC");
		}
	}
#endif

static void WriteORF(FILE *f, const std::vector<unsigned char> &Seq,
  const std::vector<char> &Label, int Frame, size_t FrameLo,
  size_t FrameHi, ALPHA OutAlpha)
	{
	if ((unsigned) FrameLo != FrameLo || (unsigned) FrameHi != FrameHi)
		Die("WriteORF: overflow");

	size_t LabelLength = Label.size();
	char *LabelCopy = myalloc(char, LabelLength+1);
	for (unsigned i = 0; i < LabelLength; ++i)
		LabelCopy[i] = Label[i];
	LabelCopy[LabelLength] = 0;

	size_t LabelLengthPadded = Label.size() + strlen(g_SepStr) + 128;
	char *UpdatedLabel = myalloc(char, LabelLengthPadded);
	sprintf(UpdatedLabel, "%s%s[%u,%u]%c",
	  LabelCopy,
	  g_SepStr,
	  (unsigned) (FrameHi + 1),
	  (unsigned) (FrameLo + 1),
	  Frame >= 0 ? '+' : '-');

	fprintf(f, ">%s\n", UpdatedLabel);

	if (g_Stops)
		{
		if (Frame >= 0)
			FrameHi += 3;
		else
			FrameLo -= 3;
		}

	assert(FrameLo < FrameHi);
	size_t FrameLength = FrameHi - FrameLo + 1;
	size_t CodonCount = FrameLength/3;
	if (FrameLength%3 != 0)
		Die("WriteORF, frame length not multiple of 3");

	char *SeqCopy = myalloc(char, FrameLength);
	if (Frame >= 0)
		{
		++g_PlusORFCount;
		g_PlusORFBaseCount += (uint64) FrameLength;

		if (OutAlpha == ALPHA_Nucleo)
			{
			for (size_t i = 0; i < FrameLength; ++i)
				SeqCopy[i] = Seq[FrameLo+i];
			}
		else if (OutAlpha == ALPHA_Amino)
			{
			for (size_t i = 0; i < CodonCount; ++i)
				{
				size_t Base = FrameLo + 3*i;
				char c1 = Seq[Base];
				char c2 = Seq[Base+1];
				char c3 = Seq[Base+2];
				char c = CharsToAmino(c1, c2, c3);
				SeqCopy[i] = c;
				}
			}
		else
			Die("Alpha %d not supported", OutAlpha);
		}
	else
		{
		++g_MinusORFCount;
		g_MinusORFBaseCount += (uint64) FrameLength;

		if (OutAlpha == ALPHA_Nucleo)
			{
			for (size_t i = 0; i < FrameLength; ++i)
				SeqCopy[FrameLength - 1 - i] = CompLetter[Seq[FrameLo + i]];
			}
		else if (OutAlpha == ALPHA_Amino)
			{
			for (size_t i = 0; i < CodonCount; ++i)
				{
				size_t Base = FrameHi - 3*i;
				char c1 = CompLetter[Seq[Base]];
				char c2 = CompLetter[Seq[Base-1]];
				char c3 = CompLetter[Seq[Base-2]];
				char c = CharsToAmino(c1, c2, c3);
				SeqCopy[i] = c;
				}
			}
		else
			Die("Alpha %d not supported", OutAlpha);
		}

	if (OutAlpha == ALPHA_Nucleo)
		WriteFasta(f, SeqCopy, (unsigned) FrameLength);
	else
		WriteFasta(f, SeqCopy, (unsigned) CodonCount);
	
	myfree(UpdatedLabel);
	myfree(SeqCopy);

	UpdatedLabel = 0;
	SeqCopy = 0;
	}

static void ExtractORFsFrame(FILE *f, const std::vector<unsigned char> &Seq,
  const std::vector<char> &Label, int Frame, ALPHA OutAlpha)
	{
	assert(Frame >= -3 && Frame <= 2);
	size_t SeqLen = Seq.size();
	if (Frame >= 0)
		{
		size_t Pos = Frame;
		bool InFrame = false;
		size_t FrameStart = 0;
		for (;;)
			{
			if (Pos + 2 >= SeqLen)
				break;

			CODON_TYPE Codon = GetCodon(Seq, Pos);
#if TRACE
#if	VERBOSE
			Log("%2d  %9u  %c%c%c  %s  %s\n",
			  Frame,
			  Pos,
			  CompLetter[Seq[Pos]],
			  CompLetter[Seq[Pos+1]],
			  CompLetter[Seq[Pos+2]],
			  CodonTypeToStr(Codon),
			  InFrame ? "ORF" : "");
#endif
#endif
			switch (Codon)
				{
			case START_CODON:
				if (InFrame)
					break;
				else
					{
					InFrame = true;
					FrameStart = Pos;
					}
				break;

			case STOP_CODON:
				if (!InFrame)
					break;
				else
					{
					size_t FrameLength = Pos - FrameStart;
#if	TRACE
					Log("Plus frame=%2d  ORF from=%7u  to=%7u  length=%6u\n",
					  Frame, (unsigned) Pos, (unsigned) FrameStart,
					  (unsigned) FrameLength);
#endif
					if (FrameLength >= g_MinFrameLength)
						{
					
					// SRT XXX ... XXX STP
					// ^               ^
					// FrameStart      Pos
#ifndef NDEBUG
						assert(FrameStart < Pos);
					// Codon count includes SRT & STP
						unsigned CodonCount = (unsigned) (Pos - FrameStart)/3 + 1;
						AssertFrame(Seq, FrameStart, CodonCount);
#endif
						WriteORF(f, Seq, Label, Frame, FrameStart, Pos - 1, OutAlpha);
						}
					InFrame = false;
					}
				break;

			case OTHER_CODON:
				break;
				}
			Pos += 3;
			}
		}
	else
		{
		assert(Frame < 0);

		int Delta = -Frame - 1;
		assert(Delta >= 0 && Delta <= 2);

		size_t Pos = SeqLen - (size_t) Delta;
		bool InFrame = false;
		size_t FrameStart = 0;
		for (;;)
			{
			CODON_TYPE Codon = GetCodonRC(Seq, Pos);
#if TRACE
#if	VERBOSE
			Log("%2d  %9u  %c%c%c  %s  %s\n",
			  Frame,
			  Pos,
			  CompLetter[Seq[Pos]],
			  CompLetter[Seq[Pos-1]],
			  CompLetter[Seq[Pos-2]],
			  CodonTypeToStr(Codon),
			  InFrame ? "ORF" : "");
#endif
#endif
			switch (Codon)
				{
			case START_CODON:
				if (InFrame)
					break;
				else
					{
					InFrame = true;
					FrameStart = Pos;
					}
				break;

			case STOP_CODON:
				if (!InFrame)
					break;
				else
					{
					size_t FrameLength = FrameStart - Pos;
#if	TRACE
					Log("Minus frame=%2d  ORF from=%7u  to=%7u  length=%6u\n",
					  Frame, (unsigned) Pos, (unsigned) FrameStart,
					  (unsigned) FrameLength);
#endif
					if (FrameLength >= g_MinFrameLength)
						{
					// PTS XXX ... XXX TRS
					//   ^               ^
					// Pos      FrameStart
#ifndef NDEBUG
						assert(FrameStart > Pos);
						unsigned CodonCount = (unsigned) (FrameStart - Pos)/3 + 1;
					// Codon count includes SRT & STP
						AssertFrameRC(Seq, FrameStart, CodonCount);
#endif
						WriteORF(f, Seq, Label, Frame, Pos + 1, FrameStart, OutAlpha);
						}
					InFrame = false;
					}
				break;

			case OTHER_CODON:
				break;
				}
			if (Pos < 6)
				break;
			Pos -= 3;
			}
		}
	}

void ExtractORFsSeq(FILE *f, const std::vector<char> &Label,
  const std::vector<unsigned char> &Seq, ALPHA OutAlpha)
	{
	ExtractORFsFrame(f, Seq, Label, 0, OutAlpha);
	ExtractORFsFrame(f, Seq, Label, 1, OutAlpha);
	ExtractORFsFrame(f, Seq, Label, 2, OutAlpha);
	ExtractORFsFrame(f, Seq, Label, -1, OutAlpha);
	ExtractORFsFrame(f, Seq, Label, -2, OutAlpha);
	ExtractORFsFrame(f, Seq, Label, -3, OutAlpha);
	}
