#include "utils.h"
#include "getopts.h"
#include "revfile.h"
#include "progress.h"

const unsigned MAX_RANGE_LINE = 4096;

coord_t GetOverlapLength(coord_t loA, coord_t hiA, coord_t loB, coord_t hiB);
void GetAlnRange(const AlnInfo &AI, const StdStrVec &Rows, unsigned GenomeIndex,
  unsigned ChrId, coord_t lo, coord_t hi, AlnInfo &RangeAI, StdStrVec &RangeRows);

static bool g_AND = false;	// Default is OR

struct Range
	{
	unsigned GenId;
	unsigned ChrId;
	coord_t Lo;
	coord_t Hi;
	};
static std::vector<Range> Ranges;

static bool InRange(const AlnInfo &AI, const Range &R)
	{
	const unsigned SegCount = AI.GetSegCount();
	for (unsigned SegIndex = 0; SegIndex < SegCount; ++SegIndex)
		{
		const GCSeg &Seg = AI.Segs[SegIndex];
		if (Seg.GenomeIndex == R.GenId && Seg.ChrId == R.ChrId &&
		  GetOverlapLength(R.Lo, R.Hi, Seg.lo, Seg.hi) > 0)
			return true;
		}
	return false;
	}

static bool InRanges(const AlnInfo &AI)
	{
	const unsigned RangeCount = (unsigned) Ranges.size();
	bool Any = false;
	for (unsigned RangeIndex = 0; RangeIndex < RangeCount; ++RangeIndex)
		{
		bool Found = InRange(AI, Ranges[RangeIndex]);
		if (Found)
			{
			Any = true;
			if (!g_AND)
				return true;
			}
		else
			{
			if (g_AND)
				return false;
			}
		}
	if (g_AND)
		{
		if (Any)
			return true;
		else
			Die("AND logic error (RangeCount=%u Any=false)", RangeCount);
		}
	return false;
	}

static void ReadRangeFile()
	{
	const char *RangeFileName = ValueOpt("rangefile");
	if (RangeFileName == 0)
		return;

	FILE *fRange = OpenStdioFile(RangeFileName);

	for (;;)
		{
		char Line[MAX_RANGE_LINE];
		bool EndOfFile = ReadLineStdioFile(fRange, Line, sizeof(Line));
		if (EndOfFile)
			break;

		Range R;
		int n = sscanf(Line, "%u %u %u %u", &R.GenId, &R.ChrId, &R.Lo, &R.Hi);
		if (n != 4)
			Die("Invalid range '%s'", Line);

		Ranges.push_back(R);
		}
	CloseStdioFile(fRange);
	}

static void GetRangeOpts()
	{
	const char *strGenId = ValueOpt("genid");
	if (strGenId == 0)
		return;
	const char *strChrId = RequiredValueOpt("chrid");
	const char *strLo = RequiredValueOpt("lo");
	const char *strHi = RequiredValueOpt("hi");

	Range R;
	R.GenId = (unsigned) atoi(strGenId);
	R.ChrId = (unsigned) atoi(strChrId);
	R.Lo = StrToCoord(strLo);
	R.Hi = StrToCoord(strHi);

	Ranges.push_back(R);
	}

static void DoTrunc(RevFile &Out, const AlnInfo &AI, const StdStrVec &Rows)
	{
	const unsigned RangeCount = SIZE(Ranges);
	for (unsigned i = 0; i < RangeCount; ++i)
		{
		const Range &R = Ranges[i];

		AlnInfo RangeAI;
		StdStrVec RangeRows;
		GetAlnRange(AI, Rows, R.GenId, R.ChrId, R.Lo, R.Hi, RangeAI, RangeRows);

		if (!RangeRows.empty())
			Out.AddAln(RangeAI, RangeRows);
		}
	}

void XRange()
	{
	const char *InFileName = RequiredValueOpt("xrange");
	const char *OutFileName = RequiredValueOpt("out");
	bool Trunc = FlagOpt("trunc");
	bool LogAIIndexes = FlagOpt("logaiindexes");
	g_AND = FlagOpt("and");

	RevFile In;
	RevFile Out;
	In.Open(InFileName);
	const std::vector<GenInfo> &Genomes = In.GetGenomes();
	Out.Create(OutFileName, Genomes);

	GetRangeOpts();
	ReadRangeFile();

	unsigned FoundCount = 0;
	const unsigned AICount = In.GetAlnInfoCount();
	ProgressStart("Searching");
	for (unsigned AIIndex = 0; AIIndex < AICount; ++AIIndex)
		{
		if (AIIndex%1000 == 0)
			ProgressStep(AIIndex, AICount);

		AlnInfo AI;
		unsigned AlnIndex = In.GetAlnInfo(AIIndex, AI);

		if (!InRanges(AI))
			continue;

		if (LogAIIndexes)
			Log("%u\n", AIIndex);

		StdStrVec Rows;
		In.GetAlnRows(AlnIndex, Rows);

		if (Trunc)
			DoTrunc(Out, AI, Rows);
		else
			Out.AddAln(AI, Rows);
		++FoundCount;
		}
	ProgressDone();
	Progress("%u alignments found", FoundCount);
	}
